/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "ECLogger.h"
#include <locale.h>

#include "config.h"
#if HAVE_SYSLOG_H
#include <syslog.h>
#endif
#include <libgen.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

using namespace std;

ECLogger::ECLogger(int max_ll) {
	max_loglevel = max_ll;
	msgbuffer = new char[_LOG_BUFSIZE];
	// set timestamps in startup locales
	setlocale(LC_TIME, "");
}

ECLogger::~ECLogger() {
	delete [] msgbuffer;
}

void ECLogger::SetLoglevel(int max_ll) {
	max_loglevel = max_ll;
}

char* ECLogger::MakeTimestamp() {
	time_t now = time(NULL);
	tm local;

	localtime_r(&now, &local);
	strftime(timestring, _LOG_TSSIZE, "%c", &local);
	return timestring;
}

bool ECLogger::Log(int loglevel) {
	return loglevel <= max_loglevel;
}


ECLogger_Null::ECLogger_Null() : ECLogger(EC_LOGLEVEL_NONE) {}
ECLogger_Null::~ECLogger_Null() {}
void ECLogger_Null::Reset() {}
void ECLogger_Null::Log(int loglevel, const string &message) {}
void ECLogger_Null::Log(int loglevel, const char *format, ...) {}


ECLogger_File::ECLogger_File(int max_ll, int add_timestamp, const char *filename) : ECLogger(max_ll) {
	pthread_mutex_init(&filelock, NULL);
	logname = strdup(filename);
	timestamp = add_timestamp;

	prevcount = 0;
	prevmsg.clear();

	if (strcmp(logname, "-") == 0)
		log = stderr;
	else
		log = fopen(logname, "a");
}

ECLogger_File::~ECLogger_File() {
	if (prevcount > 1) {
		if (timestamp)
			fprintf(log, "%s: ", MakeTimestamp());
		fprintf(log, "Previous message logged %d times\n", prevcount);
	}
	if (strcmp(logname, "-") != 0 && log)
		fclose(log);
	pthread_mutex_destroy(&filelock);
	if(logname)
		free(logname);
}

void ECLogger_File::Reset() {
	if (log == stderr)
		return;

	pthread_mutex_lock(&filelock);
	if(log)
		fclose(log);
	log = fopen(logname, "a");
	pthread_mutex_unlock(&filelock);
}

bool ECLogger_File::DupFilter(const std::string &message) {
	if (prevmsg == message) {
		prevcount++;
		return true;
	}
	if (prevcount > 1) {
		if (timestamp)
			fprintf(log, "%s: ", MakeTimestamp());
		fprintf(log, "Previous message logged %d times\n", prevcount);
	}
	prevmsg = message;
	prevcount = 0;
	return false;
}

void ECLogger_File::Log(int loglevel, const string &message) {
	if (!log)
		return;
	if (!ECLogger::Log(loglevel))
		return;

	pthread_mutex_lock(&filelock);
	if (!DupFilter(message)) {

		if (timestamp)
			fprintf(log, "%s: ", MakeTimestamp());
		fprintf(log, "%s\n", message.c_str());
		fflush(log);
	}
	pthread_mutex_unlock(&filelock);
}

void ECLogger_File::Log(int loglevel, const char *format, ...) {
	va_list va;

	if (!log)
		return;
	if (!ECLogger::Log(loglevel))
		return;

	va_start(va, format);
	_vsnprintf(msgbuffer, _LOG_BUFSIZE, format, va);
	
	pthread_mutex_lock(&filelock);

	if (!DupFilter(msgbuffer)) {

		if (timestamp)
			fprintf(log, "%s: ", MakeTimestamp());
		fprintf(log, "%s\n", msgbuffer);
		fflush(log);
	}
	pthread_mutex_unlock(&filelock);
	va_end(va);
}

ECLogger_Syslog::ECLogger_Syslog(int max_ll, const char *ident) : ECLogger(max_ll) {
	openlog(ident, LOG_PID, LOG_MAIL);
	levelmap[EC_LOGLEVEL_NONE] = LOG_DEBUG;
	levelmap[EC_LOGLEVEL_FATAL] = LOG_CRIT;
	levelmap[EC_LOGLEVEL_ERROR] = LOG_ERR;
	levelmap[EC_LOGLEVEL_WARNING] = LOG_WARNING;
	levelmap[EC_LOGLEVEL_INFO] = LOG_NOTICE;
	levelmap[EC_LOGLEVEL_ANY] = LOG_INFO;
	levelmap[EC_LOGLEVEL_DEBUG] = LOG_DEBUG;
}

ECLogger_Syslog::~ECLogger_Syslog() {
	closelog();
}

void ECLogger_Syslog::Reset() {
	// not needed.
}

void ECLogger_Syslog::Log(int loglevel, const string &message) {
	if (!ECLogger::Log(loglevel))
		return;
	syslog(levelmap[loglevel], "%s", message.c_str());
}

void ECLogger_Syslog::Log(int loglevel, const char *format, ...) {
	va_list va;

	if (!ECLogger::Log(loglevel))
		return;

	va_start(va, format);
	_vsnprintf(msgbuffer, _LOG_BUFSIZE, format, va);
	syslog(levelmap[loglevel], "%s", msgbuffer);
	va_end(va);
}



// create ECLogger object from configuration.
// lpszServiceName == service name for windows event logger
ECLogger* CreateLogger(ECConfig *lpConfig, char *argv0, const char *lpszServiceName) {
	ECLogger *lpLogger = NULL;

	if (stricmp(lpConfig->GetSetting("log_method"), "syslog") == 0) {
		lpLogger = new ECLogger_Syslog(atoi(lpConfig->GetSetting("log_level")), basename(argv0));
	} else if (stricmp(lpConfig->GetSetting("log_method"), "eventlog") == 0) {
		fprintf(stderr, "eventlog logging is only available on windows.\n");
	} else if (stricmp(lpConfig->GetSetting("log_method"), "file") == 0) {
		int ret = 0;
		struct passwd *pw = NULL;
		struct group *gr = NULL;
		if (strcmp(lpConfig->GetSetting("log_file"), "-")) {
			if (lpConfig->GetSetting("run_as_user") && strcmp(lpConfig->GetSetting("run_as_user"),""))
				pw = (struct passwd *) getpwnam(lpConfig->GetSetting("run_as_user"));
			else
				pw = (struct passwd *) getpwuid(getuid());
			if (lpConfig->GetSetting("run_as_group") && strcmp(lpConfig->GetSetting("run_as_group"),""))
				gr = (struct group *) getgrnam(lpConfig->GetSetting("run_as_group"));
			else
				gr = (struct group *) getgrgid(getgid());

			// see if we can open the file as the user we're supposed to run as
			if (pw || gr) {
				ret = fork();
				if (ret == 0) {
					// client test program
					if (gr)
						setgid(gr->gr_gid);
					if (pw)
						setuid(pw->pw_uid);
					FILE *test = fopen(lpConfig->GetSetting("log_file"), "a");
					if (!test) {
						fprintf(stderr, "Unable to open logfile '%s' as user '%s'\n",
								lpConfig->GetSetting("log_file"), pw->pw_name);
						_exit(1);
					} else {
						fclose(test);
					}
					_exit(0);
				}
				if (ret > 0) {	// correct parent, (fork != -1)
					wait(&ret);
					ret = WEXITSTATUS(ret);
				}
			}
		}
		if (ret == 0) {
			lpLogger = new ECLogger_File(atoi(lpConfig->GetSetting("log_level")), atoi(lpConfig->GetSetting("log_timestamp")), lpConfig->GetSetting("log_file"));
			// chown file
			if (pw || gr) {
				uid_t uid = -1;
				gid_t gid = -1;
				if (pw)
					uid = pw->pw_uid;
				if (gr)
					gid = gr->gr_gid;
				chown(lpConfig->GetSetting("log_file"), uid, gid);
			}
		} else {
			fprintf(stderr, "Not enough permissions to append logfile '%s'. Reverting to stderr.\n", lpConfig->GetSetting("log_file"));
			lpLogger = new ECLogger_File(atoi(lpConfig->GetSetting("log_level")), atoi(lpConfig->GetSetting("log_timestamp")), "-");
		}
	}
	if (!lpLogger) {
		fprintf(stderr, "Incorrect logging method selected. Reverting to stderr.\n");
		lpLogger = new ECLogger_File(atoi(lpConfig->GetSetting("log_level")), atoi(lpConfig->GetSetting("log_timestamp")), (char*)"-");
	}

	return lpLogger;
}

void LogConfigErrors(ECConfig *lpConfig, ECLogger *lpLogger) {
	list<string> *strings;
	list<string>::iterator i;

	if (!lpConfig || !lpLogger)
		return;

	strings = lpConfig->GetWarnings();
	for (i = strings->begin() ; i != strings->end(); i++)
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Config warning: " + *i);

	strings = lpConfig->GetErrors();
	for (i = strings->begin() ; i != strings->end(); i++)
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Config error: " + *i);
}
