/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "UnixUtil.h"

#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <string>
using namespace std;

int unix_runas(ECConfig *lpConfig, ECLogger *lpLogger) {
	if (strcmp(lpConfig->GetSetting("run_as_group"),"")) {
		struct group *gr;
		gr = (struct group *) getgrnam(lpConfig->GetSetting("run_as_group"));
		if (!gr) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to run as group '%s'", lpConfig->GetSetting("run_as_group"));
			return -1;
		}
		if (getgid() != gr->gr_gid) {
			if (setgid(gr->gr_gid) != 0) {
				switch (errno) {
				case EPERM:
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Not enough permissions to change to group '%s'.", gr->gr_name);
					break;
				default:
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown error while setting user id: %d", errno);
					break;
				};
				return -1;
			}
		}
	}

	if (strcmp(lpConfig->GetSetting("run_as_user"),"")) {
		struct passwd *pw;
		pw = (struct passwd *) getpwnam(lpConfig->GetSetting("run_as_user"));
		if (!pw) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to run as user '%s'", lpConfig->GetSetting("run_as_user"));
			return -1;
		}
		if (getuid() != pw->pw_uid) {
			if (setuid(pw->pw_uid) != 0) {
				switch (errno) {
				case EAGAIN:
					lpLogger->Log(EC_LOGLEVEL_FATAL, "EGAIN error while setting user id.");
					break;
				case EPERM:
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Not enough permissions to change to user '%s'.", pw->pw_name);
					break;
				default:
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown error while setting user id: %d", errno);
					break;
				};
				return -1;
			}
		}
	}

	return 0;
}

int unix_chown(const char *filename, const char *username, const char *groupname) {
	struct group *gr = NULL;
	struct passwd *pw = NULL;
	uid_t uid;
	gid_t gid;

	gid = getgid();

	if (strcmp(groupname,"")) {
		gr = (struct group *) getgrnam(groupname);
		if (gr)
			gid = gr->gr_gid;
	}

	uid = getuid();

	if (strcmp(username,"")) {
		pw = (struct passwd *) getpwnam(username);
		if (pw)
			uid = pw->pw_uid;
	}

	return chown(filename, uid, gid);
}

int unix_create_pidfile(char *argv0, ECConfig *lpConfig, ECLogger *lpLogger, bool bForce) {
	string pidfilename = string("/var/run/")+string(argv0)+string(".pid");
	FILE *pidfile;
	int oldpid;
	char tmp[255];
	bool running = false;

	if (strcmp(lpConfig->GetSetting("pid_file"), "")) {
		pidfilename = lpConfig->GetSetting("pid_file");
	}

	// test for existing and running process
	pidfile = fopen(pidfilename.c_str(), "r");
	if (pidfile) {
		fscanf(pidfile, "%d", &oldpid);
		fclose(pidfile);

		snprintf(tmp, 255, "/proc/%d/cmdline", oldpid);
		pidfile = fopen(tmp, "r");
		if (pidfile) {
			fscanf(pidfile, "%s", tmp);
			fclose(pidfile);

			if (strlen(tmp) < strlen(argv0)) {
				if (strstr(argv0, tmp))
					running = true;
			} else {
				if (strstr(tmp, argv0))
					running = true;
			}

			if (running) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Warning: Process %s is probably already running.", argv0);
				if (!bForce) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "If you are sure the process is stopped, please remove pidfile %s", pidfilename.c_str());
					return -1;
				}
			}
		}
	}

	pidfile = fopen(pidfilename.c_str(), "w");
	if (!pidfile) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open pidfile '%s'", pidfilename.c_str());
		return 1;
	}

	fprintf(pidfile, "%d\n", getpid());
	fclose(pidfile);
	return 0;
}

int unix_daemonize(ECConfig *lpConfig, ECLogger *lpLogger) {
	int ret;
	char *path = lpConfig->GetSetting("running_path");

	ret = fork();
	if (ret == -1) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Daemonizing failed on 1st step");
		return -1;
	}
	if (ret)
		_exit(0);				// close parent process

	setsid();					// start new session

	ret = fork();
	if (ret == -1) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Daemonizing failed on 2nd step");
		return -1;
	}
	if (ret)
		_exit(0);				// close parent process

	// close output to console. a logger which logged to the console is now diverted to /dev/null
	fclose(stdin);
	freopen("/dev/null", "a+", stdout);
	freopen("/dev/null", "a+", stderr);

	// make sure we daemonize in an always existing directory
	if (path) {
		ret = chdir(path);
		if (ret)
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to run in given path '%s': %s", path, strerror(errno));
	}

	if (!path || ret)
		chdir("/");

	return 0;
}

int unix_popen_rw(const char *lpszCommand, int *lpulIn, int *lpulOut)
{
#define FP_READ		0
#define FP_WRITE	1
	int ulIn[2];
	int ulOut[2];
	pid_t pid;

	if (!lpszCommand || !lpulIn || !lpulOut)
		return -1;

	if (pipe(ulIn) || pipe(ulOut))
		return -1;

	pid = vfork();
	switch (pid) {
	case -1:
		return -1;
	case 0:
		close(ulIn[FP_WRITE]);
		dup2(ulIn[FP_READ], FP_READ);
		close(ulOut[FP_READ]);
		dup2(ulOut[FP_WRITE], FP_WRITE);

		execl("/bin/sh", "sh", "-c", lpszCommand, NULL);
		exit(EXIT_SUCCESS);
		break;
	}

	*lpulIn = ulIn[FP_WRITE];
	close(ulIn[FP_READ]);

	*lpulOut = ulOut[FP_READ];
	close(ulOut[FP_WRITE]);

	if (pid <= 0)
		return -1;
	return 0;
}
