/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "stringutil.h"
#include <sstream>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

std::string stringify(unsigned int x, bool usehex) {
	char szBuff[33];

	if(usehex)
		sprintf(szBuff, "0x%08X", x);
	else
		sprintf(szBuff, "%u", x);
	
	return szBuff;
}

std::string stringify_int64(long long x, bool usehex) {
	std::ostringstream s;

	if (usehex) {
		s.flags(std::ios::showbase);
		s.setf(std::ios::hex, std::ios::basefield);	// showbase && basefield: add 0x prefix
		s.setf(std::ios::uppercase);
	}
	s << x;

	return s.str();
}

std::string stringify_float(float x) {
	std::ostringstream s;

	s << x;

	return s.str();
}

std::string stringify_double(double x, int prec) {
	std::ostringstream s;

	s.precision(prec);
	s << x;

	return s.str();
}

std::string stringify_datetime(time_t x) {
	char date[128];
	struct tm *tm;
	
	tm = gmtime(&x);
	if(!tm){
		x = 0;
		tm = gmtime(&x);
	}
	
	//strftime(date, 128, "%Y-%m-%d %H:%M:%S", tm);
	snprintf(date,128,"%d-%02d-%02d %.2d:%.2d:%.2d",tm->tm_year+1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
	
	return date;
}

char *stristr(char *haystack, char *needle) {
	char *find = needle;
	char *start = haystack;

	while(*haystack != 0) {
		if(*find == 0) {
			// Found the needle
			return start;
		}
		if(tolower(*haystack) == tolower(*find)) {
			find++;
		} else {
			find = needle;
			start = haystack;
		}
		haystack++;
	}

	return *find == 0 ? start : NULL;
}

int memsubstr(const void* haystack, size_t haystackSize, const void* needle, size_t needleSize)
{
	size_t pos = 0;
	size_t match = 0;
	BYTE* searchbuf = (BYTE*)needle;
	BYTE* databuf = (BYTE*)haystack;

	if(haystackSize < needleSize)
		return (haystackSize-needleSize);

	while(pos < haystackSize)
	{
		if(*databuf == *searchbuf){
			searchbuf++;
			match++;

			if(match == needleSize)
				return 0;
		}else{
			databuf -= match;
			pos -= match;

			searchbuf = (BYTE*)needle;
			match = 0;
		}

		databuf++;
		pos++;
	}

	return 1;
}

std::string str_storage(unsigned long long ulBytes, bool bUnlimited) {
	static unsigned int KB = 1024;
	static unsigned int MB = 1024 * KB;
	static unsigned int KiB = 1000;
	static unsigned int MiB = 1000 * KiB;
	static unsigned int GiB = 1000 * MiB;

	if (ulBytes == 0 && bUnlimited)
		return "unlimited";

	/*
	 * Don't show more then 6 digits in string
	 * (so switch at 1000 * 1000 instead of 1024 * 1024)
	 */
	if (ulBytes >= GiB)
		return stringify_int64(ulBytes / MB) + " MB";

	if (ulBytes >= MiB)
		return stringify_int64(ulBytes / KB) + " KB";

	return stringify_int64(ulBytes) + " B";
}

std::string PrettyIP(long unsigned int ip) {
    std::string strPretty;
    
    strPretty += stringify((ip >> 24)&0xFF);
    strPretty += ".";
    strPretty += stringify((ip >> 16)&0xFF);
    strPretty += ".";
    strPretty += stringify((ip >> 8)&0xFF);
    strPretty += ".";
    strPretty += stringify(ip&0xFF);
    
    return strPretty;
}

void strMakeUnicode(const char *lpszA, WCHAR **lppszW)
{
	WCHAR *lpszW = NULL;
	int length = 0;

	length = mbstounicode(NULL, lpszA, strlen(lpszA));
	lpszW = new WCHAR[length+1];
	memset(lpszW, 0, sizeof(WCHAR)*(length+1));
	mbstounicode(lpszW, lpszA, length);

	*lppszW = lpszW;
}

std::string shell_escape(std::string str)
{
	std::string::iterator start;
	std::string::iterator ptr;
	std::string escaped;

	start = ptr = str.begin();
	while (ptr != str.end()) {
		while (ptr != str.end() && *ptr != '\'')
			ptr++;

		escaped += std::string(start, ptr);
		if (ptr == str.end())
			break;

		start = ++ptr;          // skip single quote
		escaped += "'\\''";     // shell escape sequence
	}

	return escaped;
}

// libboost would be nice too..
std::vector<std::string> tokenize(const char *input, const char sep) {
	char *begin, *end = NULL;
	std::vector<std::string> vct;

	begin = (char*)input;
	while (begin && *begin != '\0') {
		end = strchr(begin, sep);
		if (!end) {
			vct.push_back(begin);
			break;
		}
		vct.push_back(std::string(begin,end));
		begin = end+1;
	}

	return vct;
}

std::vector<std::string> tokenize(const std::string &str, const std::string &delimiters)
{
	std::vector<std::string> tokens;
    	
	// skip delimiters at beginning.
   	std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    	
	// find first "non-delimiter".
   	std::string::size_type pos = str.find_first_of(delimiters, lastPos);

   	while (std::string::npos != pos || std::string::npos != lastPos)
   	{
       	// found a token, add it to the std::vector.
       	tokens.push_back(str.substr(lastPos, pos - lastPos));
		
       	// skip delimiters.  Note the "not_of"
       	lastPos = str.find_first_not_of(delimiters, pos);
		
       	// find next "non-delimiter"
       	pos = str.find_first_of(delimiters, lastPos);
   	}

	return tokens;
}

std::string trim(const std::string &strInput)
{
	if (!strInput.empty() && ((strInput[0] == ' ') || (strInput[strInput.size() - 1] == ' ')))
		return strInput.substr(strInput.find_first_not_of(' '), strInput.find_last_not_of(' '));
	return strInput;
}

// replaces %## values by ascii values
// i.e Amsterdam%2C -> Amsterdam,
std::string strUnEscapeHex(std::string strIn)
{
	std::string strOut;
	std::string strTemp;
	int ulescaped = 0;

	for (int i = 0; i < strIn.length(); i++) 
	{
		if (strIn.at(i) == '%' && strIn.length() > i + 2)
		{
			strTemp = "0x";
			strTemp += strIn.at(i+1);
			strTemp += strIn.at(i+2);
			ulescaped = strtol(strTemp.c_str(), NULL, 16);
			strOut += ulescaped;
			i += 2;
		} 
		else 
			strOut += strIn.at(i);
	}

	return strOut;
}
