/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include <signal.h>


#include <inetmapi.h>

#include <mapi.h>
#include <mapix.h>
#include <mapidefs.h>
#include <mapicode.h>
#include <mapiext.h>
#include <mapiguid.h>

#include <CommonUtil.h>

#include <iostream>
#include <stdio.h>
#include <string>
#include <vector>
#include <algorithm>
#include <stdlib.h>
#include <errno.h>

#include "ECLogger.h"
#include "ECConfig.h"
#include "UnixUtil.h"
#include "my_getopt.h"

#include "ECChannel.h"
#include "POP3.h"
#include "IMAP.h"
#include "ecversion.h"

#include "SSLUtil.h"

#ifdef max
#undef max
#endif

#define max(x,y)	(x > y ? x : y)

unsigned int thread_counter = 0;			// id for log
int daemonize = 1;
int quit = 0;
char *szPath = NULL;
ECLogger *lpLogger = NULL;
ECConfig *lpConfig = NULL;
pthread_t mainthread;

void sigterm(int) {
	quit = 1;
}

void sigpipe(int) {
	signal(SIGPIPE, sigpipe);
}

void sighup(int sig) {
	// In Win32, the signal is sent in a seperate, special signal thread. So this test is
	// not needed or required.
	if (pthread_equal(pthread_self(), mainthread)==0)
		return;
	if (lpConfig)
		lpConfig->ReloadSettings();

	if (lpLogger) {
		if (lpConfig) {
			char *ll = lpConfig->GetSetting("log_level");
			int new_ll = ll ? atoi(ll) : 2;
			lpLogger->SetLoglevel(new_ll);
		}

		lpLogger->Reset();
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Log connection was reset");
	}
}

// SIGSEGV catcher
#include <execinfo.h>

void sigsegv(int signr)
{
	void *bt[64];
	int i, n;
	char **btsymbols;

	if(!lpLogger)
		goto exit;

	lpLogger->Log(EC_LOGLEVEL_FATAL, "Caught SIGSEGV (%d), traceback:", signr);

	n = backtrace(bt, 64);

	btsymbols = backtrace_symbols(bt, n);

	for (i = 0; i < n; i++) {
		if (btsymbols)
			lpLogger->Log(EC_LOGLEVEL_FATAL, "%016p %s", bt[i], btsymbols[i]);
		else
			lpLogger->Log(EC_LOGLEVEL_FATAL, "%016p", bt[i]);
	}

	lpLogger->Log(EC_LOGLEVEL_FATAL, "When reporting this traceback, please include Linux distribution name, system architecture and Zarafa version.");

exit:
	kill(0, SIGSEGV);
}

HRESULT running_service(ECConfig *lpConfig, ECLogger *lpLogger, char *szPath, char *servicename);

void print_help(char *name) {
	cout << "Usage:\n" << endl;
	cout << name << " [-F] [-h|--host <serverpath>] [-c|--config <configfile>]" << endl;
	cout << "  -F\t\tDo not run in the background" << endl;
	cout << "  -h path\tUse alternate connect path (e.g. file:///var/run/socket).\n\t\tDefault: file:///var/run/zarafa" << endl;
	cout << "  -V Print version info." << endl;
	cout << "  -c filename\tUse alternate config file (e.g. /etc/zarafa-gateway.cfg)\n\t\tDefault: /etc/zarafa/gateway.cfg" << endl;
	cout << endl;
}



struct HandlerArgs {
	ECChannel *lpChannel;
	ECLogger *lpLogger;
	ECConfig *lpConfig;
	bool bUseSSL;
	unsigned int tc;
};

void *HandlerPOP3(void *lpArg) {
	HandlerArgs *lpHandlerArgs = (HandlerArgs *) lpArg;
	ECChannel *lpChannel = lpHandlerArgs->lpChannel;
	ECLogger *lpLogger = lpHandlerArgs->lpLogger;
	unsigned int local_thread_counter = lpHandlerArgs->tc;
	bool bUseSSL = lpHandlerArgs->bUseSSL;

	signal(SIGPIPE, sigpipe);

    if (bUseSSL) {
        if (lpChannel->HrEnableTLS() != hrSuccess) {
            lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to negotiate SSL connection");
        }
    }

	// szPath is global, pointing to argv variable, or lpConfig variable
	POP3 pop3(lpChannel, szPath, lpLogger, lpHandlerArgs->lpConfig, local_thread_counter);

	std::string inBuffer;
	HRESULT hr;
	bool bPop3Quit = false;
	delete lpHandlerArgs;
	int timeouts = 0;

	// Send hello message
	pop3.HrResponseOk("Zarafa POP3 gateway ready");

	// Main command loop
	while (!bPop3Quit && !quit) {
		std::vector<std::string> strvResult;
		POP3_Command eCommand;

		// check for data
		hr = lpChannel->HrSelect(60);
		if (hr == MAPI_E_TIMEOUT) {
			if (pop3.isLoggedIn() && timeouts < 4) {
				// ignore select() timeout for 5 minutes
				timeouts++;
				continue;
			}
			lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Connection closed because of timeout", local_thread_counter);
			bPop3Quit = true;
			break;
		} else if (hr == MAPI_E_NETWORK_ERROR) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Socket error: %s", local_thread_counter, strerror(errno));
			bPop3Quit = true;
			break;
		}

		timeouts = 0;

		// Read data
		inBuffer = "";
		hr = lpChannel->HrReadLine(&inBuffer);
		if (hr != hrSuccess) {
			if (errno)
				lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Failed to read line: %s", local_thread_counter, strerror(errno));
			else
				lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Client disconnected", local_thread_counter);
			bPop3Quit = true;
			break;
		}

		if (quit) {
			// All threads quiting
			bPop3Quit = true;
			hr = MAPI_E_CALL_FAILED;
			break;
		}

		// Split data on spaces, first result is the command
		pop3.HrSplitInput(inBuffer, strvResult);
		if (strvResult.empty()) {
			lpLogger->Log(EC_LOGLEVEL_WARNING, "[%08X] Empty line received", local_thread_counter);
			continue;
		}

		lpLogger->Log(EC_LOGLEVEL_DEBUG, "[%08X] Command received: %s", local_thread_counter, strvResult[0].c_str());
		// Parse pop3 command
		hr = pop3.HrGetCommand(strvResult[0], eCommand);
		if (hr != hrSuccess) {
			pop3.HrResponseErr("Command not recognized");
			continue;
		}

		switch (eCommand) {
		case POP3_CMD_USER:
			if (strvResult.size() != 2) {
				pop3.HrResponseErr("User command must have 1 argument");
			} else {
				pop3.HrCmdUser(strvResult[1]);
			}
			break;
		case POP3_CMD_PASS:
			// password may contain space character
			if (strvResult.size() < 2) {
				pop3.HrResponseErr("Pass command must have 1 argument");
			} else {
				string pass = strvResult[1];
				ULONG p = 2;
				while (p < strvResult.size()) {
					pass += " " + strvResult[p];
					p++;
				}
				pop3.HrCmdPass(pass);
			}
			break;
		case POP3_CMD_STAT:
			if (strvResult.size() != 1) {
				pop3.HrResponseErr("Stat command has no arguments");
			} else {
				pop3.HrCmdStat();
			}
			break;
		case POP3_CMD_LIST:
			if (strvResult.size() > 2) {
				pop3.HrResponseErr("List must have 0 or 1 arguments");
			} else if (strvResult.size() == 2) {
				pop3.HrCmdList(strtoul(strvResult[1].c_str(), NULL, 0));
			} else {
				pop3.HrCmdList();
			}
			break;
		case POP3_CMD_RETR:
			if (strvResult.size() != 2) {
				pop3.HrResponseErr("RETR must have 1 argument");
			} else {
				pop3.HrCmdRetr(strtoul(strvResult[1].c_str(), NULL, 0));
			}
			break;
		case POP3_CMD_DELE:
			if (strvResult.size() != 2) {
				pop3.HrResponseErr("DELE must have 1 argument");
			} else {
				pop3.HrCmdDele(strtoul(strvResult[1].c_str(), NULL, 0));
			}
			break;
		case POP3_CMD_NOOP:
			if (strvResult.size() > 1) {
				pop3.HrResponseErr("NOOP must have 0 arguments");
			} else {
				pop3.HrCmdNoop();
			}
			break;
		case POP3_CMD_RSET:
			if (strvResult.size() > 1) {
				pop3.HrResponseErr("RSET must have 0 arguments");
			} else {
				pop3.HrCmdRset();
			}
			break;
		case POP3_CMD_UIDL:
			if (strvResult.size() > 2) {
				pop3.HrResponseErr("UIDL must have 0 or 1 arguments");
			} else if (strvResult.size() == 2) {
				pop3.HrCmdUidl(strtoul(strvResult[1].c_str(), NULL, 0));
			} else {
				pop3.HrCmdUidl();
			}
			break;
		case POP3_CMD_TOP:
			if (strvResult.size() != 3) {
				pop3.HrResponseErr("Top must have 2 arguments");
			} else {
				pop3.HrCmdTop(strtoul(strvResult[1].c_str(), NULL, 0), strtoul(strvResult[2].c_str(), NULL, 0));
			}
			break;
		case POP3_CMD_QUIT:
			pop3.HrCmdQuit();
			bPop3Quit = true;
			break;
		default:
			pop3.HrResponseErr("Function not (yet) implemented");
			lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] not existing function called: %s", local_thread_counter, strvResult[0].c_str());
		}
	}

	lpLogger->Log(EC_LOGLEVEL_FATAL, "[%08X] POP3 thread exiting", local_thread_counter);

	delete lpChannel;

	// Do not pthread_exit() because linuxthreads is broken and will not free any objects
	// pthread_exit(NULL);
	return NULL;
}

void *HandlerIMAP(void *lpArg) {
	HandlerArgs *lpHandlerArgs = (HandlerArgs *) lpArg;
	ECChannel *lpChannel = lpHandlerArgs->lpChannel;
	ECLogger *lpLogger = lpHandlerArgs->lpLogger;
	ECConfig *lpConfig = lpHandlerArgs->lpConfig;
	unsigned int local_thread_counter = lpHandlerArgs->tc;
	bool bUseSSL = lpHandlerArgs->bUseSSL;

	signal(SIGPIPE, sigpipe);

	if (bUseSSL) {
		if (lpChannel->HrEnableTLS() != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to negotiate SSL connection");
		}
	}

	IMAP imap(lpChannel, szPath, lpLogger, lpConfig,
			  (strcmp(lpConfig->GetSetting("imap_only_mailfolders"), "yes") == 0), local_thread_counter, (strcmp(lpConfig->GetSetting("imap_public_folders"), "yes") == 0));

	std::string inBuffer;
	HRESULT hr;
	bool bImapQuit = false;
	delete lpHandlerArgs;
	int timeouts = 0;

	imap.HrResponse("OK Zarafa IMAP gateway ready");

	// Main command loop
	while (!bImapQuit && !quit) {
		std::string strTag, strCommand;
		std::vector<std::string> strvResult;
		ULONG ulByteCount;
		IMAP_Command eCommand;

		// check for data
		hr = lpChannel->HrSelect(60);
		if (hr == MAPI_E_TIMEOUT) {
			if (imap.isLoggedIn() && timeouts < 29) {
				// ignore select() timeout for 30 minutes
				timeouts++;
				continue;
			}
			// close idle first, so we don't have a race condition with the channel
			imap.HrResponse("BYE Connection closed because of timeout");
			lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Connection closed because of timeout", local_thread_counter);
			bImapQuit = true;
			break;
		} else if (hr == MAPI_E_NETWORK_ERROR) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Socket error: %s", local_thread_counter, strerror(errno));
			bImapQuit = true;
			break;
		}

		timeouts = 0;

		inBuffer = "";
		hr = lpChannel->HrReadLine(&inBuffer);
		if (hr != hrSuccess) {
			if (errno)
				lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Failed to read line: %s", local_thread_counter, strerror(errno));
			else
				lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Client disconnected", local_thread_counter);
			bImapQuit = true;
			break;
		}

		if (quit) {
			imap.HrResponse("BYE server shutting down");
			hr = MAPI_E_CALL_FAILED;
			bImapQuit = true;
			break;
		}

		lpLogger->Log(EC_LOGLEVEL_DEBUG, "[%08X]< %s", local_thread_counter, inBuffer.c_str());

		imap.HrSplitInput(inBuffer, strvResult);
		std::string strByteTag;
		if (strvResult.empty()) {
			imap.HrResponse("BAD No command given");
			continue;
		}
		if (strvResult.size() == 1) {
			// DONE is not really a command, but the end of the IDLE command by the client marker
			transform(strvResult[0].begin(), strvResult[0].end(), strvResult[0].begin(), ::toupper);
			if (strvResult[0] == "DONE" && imap.isIdle()) {
				imap.HrDone(true);
				continue;
			}

			// all commands should start with an ID
			imap.HrResponse("BAD Command not recognized");
			continue;
		}

		// we did not receive done, so we should not be idle
		if (imap.isIdle()) {
			imap.HrResponse("BAD still in idle state");
			// we're done anyway?
			//imap.HrDone(false); // false for no output
			continue;			
		}

		strTag = strvResult.front();
		strvResult.erase(strvResult.begin());

		strCommand = strvResult.front();
		strvResult.erase(strvResult.begin());

		// Parse imap line
		hr = imap.HrGetCommand(strCommand, eCommand);
		if (hr != hrSuccess) {
			imap.HrResultBad(strTag, "Command not recognized");
			continue;
		}

		// last argument can be {number} where number is the number of bytes to come
		// TODO: make strvResult[strvResult.size() - 1] into strvResult.end()--;
		while (hr == hrSuccess && strvResult.size() > 1 &&
			   strvResult[strvResult.size() - 1].size() > 2 &&				// last word in data is longer than 2 (incl term?)
			   strvResult[strvResult.size() - 1].substr(0, 1) == "{") {		// and starts with an {

			strByteTag = strvResult[strvResult.size() - 1].substr(1);
			strByteTag = strByteTag.substr(0, strByteTag.size() - 1);
			ulByteCount = strtoul(strByteTag.c_str(), NULL, 0);

			hr = imap.HrResponseContinuation("Ready for literal data");
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] error sending", local_thread_counter);
				break;
			}

			hr = lpChannel->HrReadBytes(&inBuffer, ulByteCount);
			if (hr != hrSuccess)
				break;
			strvResult[strvResult.size() - 1] = inBuffer;

			hr = lpChannel->HrReadLine(&inBuffer);
			if (hr != hrSuccess) {
				if (errno)
					lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Failed to read line: %s", local_thread_counter, strerror(errno));
				else
					lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] Client disconnected", local_thread_counter);
				bImapQuit = true;
				break;
			}

			imap.HrSplitInput(inBuffer, strvResult);

			if (quit) {
				imap.HrResponse("BYE server shutting down");
				hr = MAPI_E_CALL_FAILED;
				bImapQuit = true;
				break;
			}
		}
		if (hr != hrSuccess)
			continue;

		switch (eCommand) {
		case IMAP_CMD_CAPABILITY:
			if (!strvResult.empty()) {
				imap.HrResultBad(strTag, "CAPABILITY must have 0 arguments");
			} else {
				hr = imap.HrCmdCapability(strTag);
			}
			break;
		case IMAP_CMD_NOOP:
			if (!strvResult.empty()) {
				imap.HrResultBad(strTag, "NOOP must have 0 arguments");
			} else {
				hr = imap.HrCmdNoop(strTag);
			}
			break;
		case IMAP_CMD_LOGOUT:
			if (!strvResult.empty()) {
				imap.HrResultBad(strTag, "LOGOUT must have 0 arguments");
			} else {
				imap.HrCmdLogout(strTag);
				bImapQuit = true;
			}
			break;
		case IMAP_CMD_STARTTLS:
			if (!strvResult.empty()) {
				imap.HrResultBad(strTag, "STARTTLS must have 0 arguments");
			} else {
				hr = imap.HrCmdStarttls(strTag);
				if (hr != hrSuccess) {
					bImapQuit = true;
				}
			}
			break;
		case IMAP_CMD_AUTHENTICATE:
			if (strvResult.size() != 1) {
				imap.HrResultBad(strTag, "AUTHENTICATE must have 1 argument");
			} else {
				hr = imap.HrCmdAuthenticate(strTag, strvResult[0]);
			}
			break;
		case IMAP_CMD_LOGIN:
			if (strvResult.size() != 2) {
				imap.HrResultBad(strTag, "LOGIN must have 2 arguments");
			} else {
				hr = imap.HrCmdLogin(strTag, strvResult[0], strvResult[1]);
			}
			break;
		case IMAP_CMD_SELECT:
			if (strvResult.size() != 1) {
				imap.HrResultBad(strTag, "SELECT must have 1 argument");
			} else {
				hr = imap.HrCmdSelect(strTag, strvResult[0]);
			}
			break;
		case IMAP_CMD_EXAMINE:
			if (strvResult.size() != 1) {
				imap.HrResultBad(strTag, "EXAMINE must have 1 argument");
			} else {
				hr = imap.HrCmdExamine(strTag, strvResult[0]);
			}
			break;
		case IMAP_CMD_CREATE:
			if (strvResult.size() != 1) {
				imap.HrResultBad(strTag, "CREATE must have 1 argument");
			} else {
				hr = imap.HrCmdCreate(strTag, strvResult[0]);
			}
			break;
		case IMAP_CMD_DELETE:
			if (strvResult.size() != 1) {
				imap.HrResultBad(strTag, "DELETE must have 1 argument");
			} else {
				hr = imap.HrCmdDelete(strTag, strvResult[0]);
			}
			break;
		case IMAP_CMD_RENAME:
			if (strvResult.size() != 2) {
				imap.HrResultBad(strTag, "RENAME must have 2 arguments");
			} else {
				hr = imap.HrCmdRename(strTag, strvResult[0], strvResult[1]);
			}
			break;
		case IMAP_CMD_SUBSCRIBE:
			if (strvResult.size() != 1) {
				imap.HrResultBad(strTag, "SUBSCRIBE must have 1 arguments");
			} else {
				hr = imap.HrCmdSubscribe(strTag, strvResult[0], true);
			}
			break;
		case IMAP_CMD_UNSUBSCRIBE:
			if (strvResult.size() != 1) {
				imap.HrResultBad(strTag, "UNSUBSCRIBE must have 1 arguments");
			} else {
				hr = imap.HrCmdSubscribe(strTag, strvResult[0], false);
			}
			break;
		case IMAP_CMD_LIST:
			if (strvResult.size() != 2) {
				imap.HrResultBad(strTag, "LIST must have 2 arguments");
			} else {
				hr = imap.HrCmdList(strTag, strvResult[0], strvResult[1], false);
			}
			break;
		case IMAP_CMD_LSUB:
			if (strvResult.size() != 2) {
				imap.HrResultBad(strTag, "LSUB must have 2 arguments");
			} else {
				hr = imap.HrCmdList(strTag, strvResult[0], strvResult[1], true);
			}
			break;
		case IMAP_CMD_STATUS:
			if (strvResult.size() != 2) {
				imap.HrResultBad(strTag, "STATUS must have 2 arguments");
			} else {
				hr = imap.HrCmdStatus(strTag, strvResult[0], strvResult[1]);
			}
			break;
		case IMAP_CMD_APPEND:
			if (strvResult.size() == 2) {
				hr = imap.HrCmdAppend(strTag, strvResult[0], strvResult[1]);
			} else if (strvResult.size() == 3) {
				hr = imap.HrCmdAppend(strTag, strvResult[0], strvResult[2], strvResult[1]);
			} else if (strvResult.size() == 4) {
				hr = imap.HrCmdAppend(strTag, strvResult[0], strvResult[3], strvResult[1], strvResult[2]);
			} else {
				imap.HrResultBad(strTag, "APPEND must have 2, 3 or 4 arguments");
			}
			break;
		case IMAP_CMD_CHECK:
			if (!strvResult.empty()) {
				imap.HrResultBad(strTag, "CHECK must have 0 arguments");
			} else {
				hr = imap.HrCmdCheck(strTag);
			}
			break;
		case IMAP_CMD_CLOSE:
			if (!strvResult.empty()) {
				imap.HrResultBad(strTag, "CLOSE must have 0 arguments");
			} else {
				hr = imap.HrCmdClose(strTag);
			}
			break;
		case IMAP_CMD_EXPUNGE:
			if (!strvResult.empty()) {
				imap.HrResultBad(strTag, "EXPUNGE must have 0 arguments");
			} else {
				hr = imap.HrCmdExpunge(strTag);
			}
			break;
		case IMAP_CMD_SEARCH:
			if (strvResult.empty()) {
				imap.HrResultBad(strTag, "SEARCH must have 1 or more arguments");
			} else {
				hr = imap.HrCmdSearch(strTag, strvResult, false);
			}
			break;
		case IMAP_CMD_FETCH:
			if (strvResult.size() != 2) {
				imap.HrResultBad(strTag, "FETCH must have 2 arguments");
			} else {
				hr = imap.HrCmdFetch(strTag, strvResult[0], strvResult[1], false);
			}
			break;
		case IMAP_CMD_STORE:
			if (strvResult.size() != 3) {
				imap.HrResultBad(strTag, "STORE must have 3 arguments");
			} else {
				hr = imap.HrCmdStore(strTag, strvResult[0], strvResult[1], strvResult[2], false);
			}
			break;
		case IMAP_CMD_COPY:
			if (strvResult.size() != 2) {
				imap.HrResultBad(strTag, "COPY must have 2 arguments");
			} else {
				hr = imap.HrCmdCopy(strTag, strvResult[0], strvResult[1], false);
			}
			break;
		case IMAP_CMD_IDLE:
			if (!strvResult.empty()) {
				imap.HrResultBad(strTag, "IDLE must have 0 arguments");
			} else {
				hr = imap.HrCmdIdle(strTag);
			}
			break;
		case IMAP_CMD_UID:
			if (strvResult.empty()) {
				imap.HrResultBad(strTag, "UID must have a command");
			} else {
				strCommand = strvResult.front();
				strvResult.erase(strvResult.begin());

				hr = imap.HrGetCommand(strCommand, eCommand);
				if (hr != hrSuccess) {
					imap.HrResultBad(strTag, "UID command parse error");
				} else {
					switch (eCommand) {
					case IMAP_CMD_SEARCH:
						if (strvResult.empty()) {
							imap.HrResultBad(strTag, "UID SEARCH must have 1 or more arguments");
						} else {
							hr = imap.HrCmdSearch(strTag, strvResult, true);
						}
						break;
					case IMAP_CMD_FETCH:
						if (strvResult.size() != 2) {
							imap.HrResultBad(strTag, "UID FETCH must have 2 arguments");
						} else {
							hr = imap.HrCmdFetch(strTag, strvResult[0], strvResult[1], true);
						}
						break;
					case IMAP_CMD_STORE:
						if (strvResult.size() != 3) {
							imap.HrResultBad(strTag, "UID STORE must have 3 arguments");
						} else {
							hr = imap.HrCmdStore(strTag, strvResult[0], strvResult[1], strvResult[2], true);
						}
						break;
					case IMAP_CMD_COPY:
						if (strvResult.size() != 2) {
							imap.HrResultBad(strTag, "UID COPY must have 2 arguments");
						} else {
							hr = imap.HrCmdCopy(strTag, strvResult[0], strvResult[1], true);
						}
						break;
                    case IMAP_CMD_XAOL_MOVE:
						if (strvResult.size() != 2) {
							imap.HrResultBad(strTag, "UID XAOL-MOVE must have 2 arguments");
						} else {
							hr = imap.HrCmdUidXaolMove(strTag, strvResult[0], strvResult[1]);
						}
                        break;
					default:
						imap.HrResultBad(strTag, "Command not supported");
					}
				}
			}
			break;

		default:
			// currently, we can never get here ... dead code
			imap.HrResultBad(strTag, "Command not (yet) implemented");
			lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] requested not existing function: %s", local_thread_counter, strCommand.c_str());
		}
		if (hr == MAPI_E_END_OF_SESSION || hr == MAPI_E_NETWORK_ERROR) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "[%08X] MAPI Connection error, disconnecting client.", local_thread_counter);
			bImapQuit = true;
		}
	} // while (!bImapquit)

	lpLogger->Log(EC_LOGLEVEL_FATAL, "[%08X] IMAP thread exiting", local_thread_counter);

	imap.HrDone(false);	// HrDone does not send an error string to the client
	delete lpChannel;

	// Do not pthread_exit() because linuxthreads is broken and will not free any objects
	// pthread_exit(NULL);
	return NULL;
}

int main(int argc, char *argv[]) {
	HRESULT hr = hrSuccess;
	int c = 0;

	ssl_threading_setup();

	char *szConfig = "/etc/zarafa/gateway.cfg";

	const configsetting_t lpDefaults[] = {
		{ "server_bind", "0.0.0.0" },
		{ "run_as_user", "" },
		{ "run_as_group", "" },
		{ "pid_file", "/var/run/zarafa-gateway.pid" },
		{ "running_path", "/" },
		{ "pop3_enable", "yes" },
		{ "pop3_port", "110" },
		{ "pop3s_enable", "no" },
		{ "pop3s_port", "995" },
		{ "imap_enable", "yes" },
		{ "imap_port", "143" },
		{ "imaps_enable", "no" },
		{ "imaps_port", "993" },
		{ "imap_only_mailfolders", "yes", CONFIGSETTING_RELOADABLE },
		{ "imap_public_folders", "yes", CONFIGSETTING_RELOADABLE },
		{ "imap_capability_idle", "yes", CONFIGSETTING_RELOADABLE },
		{ "server_socket", "http://localhost:236/zarafa" },
		{ "ssl_private_key_file", "/etc/zarafa/gateway/privkey.pem" },
		{ "ssl_certificate_file", "/etc/zarafa/gateway/cert.pem" },
		{ "ssl_verify_client", "no" },
		{ "ssl_verify_file", "" },
		{ "ssl_verify_path", "" },
		{ "log_method", "file" },
		{ "log_file", "-" },
		{ "log_level", "2", CONFIGSETTING_RELOADABLE },
		{ "log_timestamp", "1" },
		{ NULL, NULL },
	};
	enum {
		OPT_HELP,
		OPT_HOST,
		OPT_CONFIG,
		OPT_FOREGROUND
	};
	struct option long_options[] = {
		{"help", 0, NULL, OPT_HELP},
		{"host", 1, NULL, OPT_HOST},
		{"config", 1, NULL, OPT_CONFIG},
		{"foreground", 1, NULL, OPT_FOREGROUND},
		{NULL, 0, NULL, 0}
	};

	// Get commandline options
	while (1) {
		c = my_getopt_long(argc, argv, "c:h:iuFV", long_options, NULL);

		if (c == -1)
			break;

		switch (c) {
		case OPT_CONFIG:
		case 'c':
			szConfig = my_optarg;
			break;
		case OPT_HOST:
		case 'h':
			szPath = my_optarg;
			break;
		case 'i':				// Install service
		case 'u':				// Uninstall service
			break;
		case OPT_FOREGROUND:
		case 'F':
			daemonize = 0;
			break;
		case 'V':
			cout << "Product version:\t" <<  PROJECT_VERSION_GATEWAY_STR << endl
				 << "File version:\t\t" << PROJECT_SVN_REV_STR << endl;
			return 1;
		case OPT_HELP:
		default:
			print_help(argv[0]);
			return 1;
		}
	}
	// Setup config
	lpConfig = new ECConfig(lpDefaults);

	if (!lpConfig->LoadSettings(szConfig)) {
		cerr << "Unable to open configuration file " << szConfig << endl;
		cerr << "Continuing with defaults" << endl;
	}

	if (lpConfig->HasErrors()) {
		lpLogger = new ECLogger_File(EC_LOGLEVEL_FATAL, 0, "-");	// create fatal logger without a timestamp to stderr
		LogConfigErrors(lpConfig, lpLogger);
		hr = E_FAIL;
		goto exit;
	}
	// Setup logging
	lpLogger = CreateLogger(lpConfig, argv[0], "ZarafaGateway");

	if (lpConfig->HasWarnings())
		LogConfigErrors(lpConfig, lpLogger);

	mainthread = pthread_self();

	if (!szPath)
		szPath = lpConfig->GetSetting("server_socket");


	hr = running_service(lpConfig, lpLogger, szPath, argv[0]);

exit:
    ssl_threading_cleanup();
    
	if (lpConfig)
		delete lpConfig;

	if (lpLogger)
		delete lpLogger;

	return hr == hrSuccess ? 0 : 1;
}

HRESULT running_service(ECConfig *lpConfig, ECLogger *lpLogger, char *szPath, char *servicename) {
	HRESULT hr = hrSuccess;
	int ulListenPOP3 = 0, ulListenPOP3s = 0;
	int ulListenIMAP = 0, ulListenIMAPs = 0;
	bool bListenPOP3, bListenPOP3s;
	bool bListenIMAP, bListenIMAPs;
	fd_set readfds;
	int err = 0;
	pthread_attr_t ThreadAttr;

	pthread_attr_init(&ThreadAttr);
	if (pthread_attr_setdetachstate(&ThreadAttr, PTHREAD_CREATE_DETACHED) != 0) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not set thread attribute to detached");
		goto exit;
	}
	// 1Mb of stack space per thread
	if (pthread_attr_setstacksize(&ThreadAttr, 1024 * 1024)) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not set thread stack size to 1Mb");
		goto exit;
	}


	bListenPOP3 = (strcmp(lpConfig->GetSetting("pop3_enable"), "yes") == 0);
	bListenPOP3s = (strcmp(lpConfig->GetSetting("pop3s_enable"), "yes") == 0);
	bListenIMAP = (strcmp(lpConfig->GetSetting("imap_enable"), "yes") == 0);
	bListenIMAPs = (strcmp(lpConfig->GetSetting("imaps_enable"), "yes") == 0);


	// Setup ssl context
	if ((bListenPOP3s || bListenIMAPs) && ECChannel::HrSetCtx(lpConfig, lpLogger) != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Error loading SSL context, POP3S and IMAPS will be disabled");
		bListenPOP3s = false;
		bListenIMAPs = false;
	}
	
	if (!bListenPOP3 && !bListenPOP3s && !bListenIMAP && !bListenIMAPs) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "POP3, POP3S, IMAP and IMAPS are all four disabled");
		hr = E_FAIL;
		goto exit;
	}

	// Setup sockets
	if (bListenPOP3) {
		hr = HrListen(lpLogger, lpConfig->GetSetting("server_bind"), atoi(lpConfig->GetSetting("pop3_port")), &ulListenPOP3);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to listen on port %s", lpConfig->GetSetting("pop3_port"));
			hr = E_FAIL;
			goto exit;
		}
		lpLogger->Log(EC_LOGLEVEL_INFO, "Listening on port %s for POP3", lpConfig->GetSetting("pop3_port"));
	}

	if (bListenPOP3s) {
		hr = HrListen(lpLogger, lpConfig->GetSetting("server_bind"), atoi(lpConfig->GetSetting("pop3s_port")), &ulListenPOP3s);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to listen on port %s", lpConfig->GetSetting("pop3s_port"));
			hr = E_FAIL;
			goto exit;
		}
		lpLogger->Log(EC_LOGLEVEL_INFO, "Listening on port %s for POP3s", lpConfig->GetSetting("pop3s_port"));
	}

	if (bListenIMAP) {
		hr = HrListen(lpLogger, lpConfig->GetSetting("server_bind"), atoi(lpConfig->GetSetting("imap_port")), &ulListenIMAP);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to listen on port %s", lpConfig->GetSetting("imap_port"));
			hr = E_FAIL;
			goto exit;
		}
		lpLogger->Log(EC_LOGLEVEL_INFO, "Listening on port %s for IMAP", lpConfig->GetSetting("imap_port"));
	}

	if (bListenIMAPs) {
		hr = HrListen(lpLogger, lpConfig->GetSetting("server_bind"), atoi(lpConfig->GetSetting("imaps_port")), &ulListenIMAPs);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to listen on port %s", lpConfig->GetSetting("imaps_port"));
			hr = E_FAIL;
			goto exit;
		}
		lpLogger->Log(EC_LOGLEVEL_INFO, "Listening on port %s for IMAPs", lpConfig->GetSetting("imaps_port"));
	}

	// Setup signals
	// signal(SIGCHLD, SIG_IGN);
	signal(SIGTERM, sigterm);
	signal(SIGINT, sigterm);
	signal(SIGHUP, sighup);		// logrotate

    // SIGSEGV backtrace support
    stack_t st;
    struct sigaction act;
  
    memset(&st, 0, sizeof(st));
    memset(&act, 0, sizeof(act));
  
    st.ss_sp = malloc(65536);
    st.ss_flags = 0;
    st.ss_size = 65536;
  
    act.sa_handler = sigsegv;
    act.sa_flags = SA_ONSTACK | SA_RESETHAND;
  
    sigaltstack(&st, NULL);
    sigaction(SIGSEGV, &act, NULL);
    sigaction(SIGBUS, &act, NULL);
    sigaction(SIGABRT, &act, NULL);

	// Setup MAPI
	hr = MAPIInitialize(NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to initialize MAPI");
		goto exit;
	}
    // Set max open file descriptors to FD_SETSIZE .. higher than this number
    // is a bad idea, as it will start breaking select() calls.
    struct rlimit file_limit;
    file_limit.rlim_cur = FD_SETSIZE;
    file_limit.rlim_max = FD_SETSIZE;
    
    if(setrlimit(RLIMIT_NOFILE, &file_limit) < 0) {
        lpLogger->Log(EC_LOGLEVEL_FATAL, "WARNING: setrlimit(RLIMIT_NOFILE, %d) failed, you will only be able to connect up to %d sockets. Either start the process as root, or increase user limits for open file descriptors", FD_SETSIZE, getdtablesize());
    }        

	// fork if needed and drop privileges as requested.
	// this must be done before we do anything with pthreads
	if (daemonize && unix_daemonize(lpConfig, lpLogger))
		goto exit;
	if (!daemonize)
		setsid();
	unix_create_pidfile(servicename, lpConfig, lpLogger);
	if (unix_runas(lpConfig, lpLogger))
		goto exit;

	lpLogger->Log(EC_LOGLEVEL_FATAL, "Starting zarafa-gateway version " PROJECT_VERSION_GATEWAY_STR " (" PROJECT_SVN_REV_STR "), pid %d", getpid());
	// Mainloop
	while (!quit) {
		FD_ZERO(&readfds);
		if (bListenPOP3)
			FD_SET(ulListenPOP3, &readfds);
		if (bListenPOP3s)
			FD_SET(ulListenPOP3s, &readfds);
		if (bListenIMAP)
			FD_SET(ulListenIMAP, &readfds);
		if (bListenIMAPs)
			FD_SET(ulListenIMAPs, &readfds);

		struct timeval timeout;
		timeout.tv_sec = 10;
		timeout.tv_usec = 0;

		err = select(max(max(ulListenPOP3, ulListenIMAP), max(ulListenPOP3s, ulListenIMAPs)) + 1, &readfds, NULL, NULL, &timeout);
		if (err < 0) {
			if (errno != EINTR) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Socket error!");
				quit = 1;
				hr = MAPI_E_NETWORK_ERROR;
			}
			continue;
		} else if (err == 0) {
			continue;
		}

		// One socket has signalled a new incoming connection

		HandlerArgs *lpHandlerArgs = new HandlerArgs;

		lpHandlerArgs->lpLogger = lpLogger;
		lpHandlerArgs->lpConfig = lpConfig;
		lpHandlerArgs->tc = thread_counter++;

		if ((bListenPOP3 && FD_ISSET(ulListenPOP3, &readfds)) || (bListenPOP3s && FD_ISSET(ulListenPOP3s, &readfds))) {
			bool usessl;
			// Incoming POP3(s) connection
			if (bListenPOP3s && FD_ISSET(ulListenPOP3s, &readfds)) {
				usessl = true;
				hr = HrAccept(lpLogger, ulListenPOP3s, &lpHandlerArgs->lpChannel);
			} else {
				usessl = false;
				hr = HrAccept(lpLogger, ulListenPOP3, &lpHandlerArgs->lpChannel);
			}

			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to accept POP3 socket connection.");
				// just keep running
				delete lpHandlerArgs;
				hr = hrSuccess;
				continue;
			}

			lpHandlerArgs->bUseSSL = usessl;

			pthread_t POP3Thread;
			if (usessl)
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Starting worker thread [%08X] for POP3s request", thread_counter-1);
			else
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Starting worker thread [%08X] for POP3 request", thread_counter-1);
			if (pthread_create(&POP3Thread, &ThreadAttr, HandlerPOP3, lpHandlerArgs) != 0) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not create POP3 thread.");
				// just keep running
				delete lpHandlerArgs->lpChannel;
				delete lpHandlerArgs;
				hr = hrSuccess;
			}
			continue;
		}

		if ((bListenIMAP && FD_ISSET(ulListenIMAP, &readfds)) || (bListenIMAPs && FD_ISSET(ulListenIMAPs, &readfds))) {
			bool usessl;
			// Incoming IMAP(s) connection
			if (bListenIMAPs && FD_ISSET(ulListenIMAPs, &readfds)) {
				usessl = true;
				hr = HrAccept(lpLogger, ulListenIMAPs, &lpHandlerArgs->lpChannel);
			} else {
				usessl = false;
				hr = HrAccept(lpLogger, ulListenIMAP, &lpHandlerArgs->lpChannel);
			}
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to accept IMAP socket connection.");
				// just keep running
				hr = hrSuccess;
				delete lpHandlerArgs;
				continue;
			}

			lpHandlerArgs->bUseSSL = usessl;

			// Create pthread to handle request
			pthread_t IMAPThread;
			if (usessl)
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Starting worker thread [%08X] for IMAPs request", thread_counter-1);
			else
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Starting worker thread [%08X] for IMAP request", thread_counter-1);
			if (pthread_create(&IMAPThread, &ThreadAttr, HandlerIMAP, lpHandlerArgs) != 0) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not create IMAP thread.");
				// just keep running
				hr = hrSuccess;
				delete lpHandlerArgs->lpChannel;
				delete lpHandlerArgs;
			}
			continue;
		}

		// should not be able to get here because of continues
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Incoming traffic was not for me??");
	}
	lpLogger->Log(EC_LOGLEVEL_FATAL, "Gateway will now exit");

	MAPIUninitialize();

exit:
	ECChannel::HrFreeCtx();

	pthread_attr_destroy(&ThreadAttr);


	return hr;
}
