/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "icalrecurrence.h"
#include "vconverter.h"
#include "nameids.h"
#include "valarm.h"
#include <mapicode.h>
#include <mapiext.h>
#include <mapix.h>
#include <mapiutil.h>
#include <math.h>
#include <algorithm>

bool operator ==(SPropValue spv, ULONG ulPropTag) {
	return spv.ulPropTag == ulPropTag;
}

ICalRecurrence::ICalRecurrence() {
}

ICalRecurrence::~ICalRecurrence() {
}

// ical -> mapi
HRESULT ICalRecurrence::HrParseICalRecurrenceRule(TIMEZONE_STRUCT sTimeZone, icalcomponent *lpicRootEvent, icalcomponent *lpicEvent,
												  LPSPropTagArray lpNamedProps, icalitem *lpIcalItem, recurrence **lppRecurrence)
{
	HRESULT hr = hrSuccess;
	recurrence *lpRec = NULL;
	icalproperty *lpicProp = NULL;
	icalrecurrencetype icRRule;
	int i = 0;
	ULONG ulWeekDays = 0;
	time_t dtUTCStart = 0;
	time_t dtLocalStart = 0;
	time_t dtUTCEnd = 0;
	time_t dtUTCUntil = 0;
	time_t exUTCDate = 0;
	time_t exLocalDate = 0;
	SPropValue sPropVal = {0};
	struct tm tm = {0};
	bool bIsAllday = false;

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_RRULE_PROPERTY);
	if (lpicProp == NULL)
		goto exit;

	icRRule = icalproperty_get_rrule(lpicProp);

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_DTSTART_PROPERTY);
	if (!lpicProp) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	// when DTSTART is only a date, the item is an allday event
	if (icalproperty_get_dtstart(lpicProp).is_date == 1)
		bIsAllday = true;
	
	// use localtime for calculating weekday as in UTC time 
	// the weekday can change to previous day for time 00:00 am
	dtLocalStart = icaltime_as_timet(icalproperty_get_dtstart(lpicProp));	
	gmtime_safe(&dtLocalStart, &tm);
	
	if (bIsAllday)
		dtUTCStart = ICalTimeTypeToLocal(lpicRootEvent, lpicProp);
	else
		dtUTCStart = ICalTimeTypeToUTC(lpicRootEvent, lpicProp);

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_DTEND_PROPERTY);
	if (!lpicProp) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if (bIsAllday)
		dtUTCEnd = ICalTimeTypeToLocal(lpicRootEvent, lpicProp);
	else
		dtUTCEnd = ICalTimeTypeToUTC(lpicRootEvent, lpicProp);

	// Set 0x8235, also known as ClipStart in OutlookSpy
	UnixTimeToFileTime(recurrence::StartOfDay(dtUTCStart), &sPropVal.Value.ft);
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRENCE_START], PT_SYSTIME);
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	// set named prop 0x8510 to 353, needed for Outlook to ask for single or total recurrence when deleting
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_SIDEEFFECT], PT_LONG);
	sPropVal.Value.ul = 353;
	lpIcalItem->lstMsgProps.push_back(sPropVal);

	lpRec = new recurrence;

	// recurrence class contains LOCAL times only, so convert UTC -> LOCAL
	lpRec->setStartDateTime(UTCToLocal(dtUTCStart, sTimeZone));

	// default 1st day of week is sunday, except in weekly recurrences
	lpRec->setFirstDOW(0);

	switch (icRRule.freq) {
	case ICAL_DAILY_RECURRENCE:
		lpRec->setFrequency(recurrence::DAILY);
		break;

	case ICAL_WEEKLY_RECURRENCE:
		lpRec->setFrequency(recurrence::WEEKLY);
		// assume this weekly item is exactly on the start time day
		lpRec->setWeekDays(1 << tm.tm_wday);
		// Strange little thing for the recurrence type "every workday"
		lpRec->setFirstDOW(1);
		break;

	case ICAL_MONTHLY_RECURRENCE:
		lpRec->setFrequency(recurrence::MONTHLY);
		break;

	case ICAL_YEARLY_RECURRENCE:
		lpRec->setFrequency(recurrence::YEARLY);
		lpRec->setDayOfMonth(tm.tm_mday);
		break;
	default:
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	};

	// since we know the frequency, this value can be set correctly
	lpRec->setInterval(icRRule.interval);

	// ulWeekDays, ulWeekNumber (monthly)
	if (icRRule.by_day[i] != ICAL_RECURRENCE_ARRAY_MAX) {
		ulWeekDays = 0;

		// Handle BYSETPOS if it is present in the ical data.
		if(icRRule.by_set_pos[0] != ICAL_RECURRENCE_ARRAY_MAX)
		{
			icRRule.by_day[0] += (8 * abs(icRRule.by_set_pos[0]));
			icRRule.by_day[0] *= icRRule.by_set_pos[0]/abs(icRRule.by_set_pos[0]);
			
			if(icRRule.week_start == ICAL_SUNDAY_WEEKDAY)
				icRRule.week_start = ICAL_MONDAY_WEEKDAY;
		}

		if (icRRule.by_day[0] < 0) {
			// outlook can only have _one_ last day of the month/year (not daily or weekly here!)
			ulWeekDays |= (1 << (abs(icRRule.by_day[0]%8) - (icRRule.week_start-1)));
			// next call also changes pattern to 3!
			lpRec->setWeekNumber(5);
		} else if (icRRule.by_day[0] >= 1 && icRRule.by_day[0] <= 7) {
			// weekly, normal days
			i = 0;
			while (icRRule.by_day[i] != ICAL_RECURRENCE_ARRAY_MAX) {
				ulWeekDays |= (1 << (icRRule.by_day[i] - (icRRule.week_start-1)));
				i++;
			}
		} else {
			// monthly, first sunday: 9, monday: 10
			ulWeekDays |= (1 << ((icRRule.by_day[0]%8) - (icRRule.week_start-1)));
			lpRec->setWeekNumber((int)(icRRule.by_day[0]/8)); // 1..4
		}
		lpRec->setWeekDays(ulWeekDays);
	} else if (icRRule.by_month_day[0] != ICAL_RECURRENCE_ARRAY_MAX) {
		// outlook can only have _one_ day in the month/year, while ical can do multiple
		// do we need to create multiple items here ?? :(
		lpRec->setDayOfMonth(icRRule.by_month_day[0]);
	}else if (icRRule.freq == ICAL_MONTHLY_RECURRENCE || icRRule.freq == ICAL_YEARLY_RECURRENCE) {
		// When RRULE:FREQ=MONTHLY;INTERVAL=1 is set then day of the month is set from the start date.		
		lpRec->setDayOfMonth(tm.tm_mday);
	}

	if (icRRule.count != 0) {
		// count limit
		lpRec->setEndType(recurrence::NUMBER);
		lpRec->setCount(icRRule.count);

		// calculate end
		lpRec->setEndDate(lpRec->calcEndDate());
		dtUTCUntil = lpRec->getEndDate();
	} else if (icRRule.until.year != 0) {
		// date limit
		lpRec->setEndType(recurrence::DATE);
		// enddate is the date/time of the LAST occurrence, do not add duration
		dtUTCUntil = icaltime_as_timet_with_zone(icRRule.until, NULL);
		lpRec->setEndDate(UTCToLocal(dtUTCUntil, sTimeZone));

		// calculate number
		lpRec->setCount(lpRec->calcCount());
	} else {
		// never ending story
		lpRec->setEndType(recurrence::NEVER);

		// outlook also sets 10, so this get's displayed in the recurrence dialog window
		lpRec->setCount(10);
		// 1. dtUTCEnd -> end time offset (set later)
		// (date will overridden by recurrence.cpp when writing the blob)

		// 2. dtUTCUntil = clipend == start of month of item in 4500
		dtUTCUntil = 0x7FFFFFFF; // incorrect, but good enough
	}

	// offset in minutes after midnight of the start time
	lpRec->setEndTimeOffset((lpRec->getStartTimeOffset() + dtUTCEnd - dtUTCStart) / 60);

	// Set 0x8236, also known as ClipEnd in OutlookSpy
	UnixTimeToFileTime(dtUTCUntil, &sPropVal.Value.ft);
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRENCE_END], PT_SYSTIME);
	lpIcalItem->lstMsgProps.push_back(sPropVal);


	// find EXDATE properties, add to delete exceptions
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_EXDATE_PROPERTY);
	while (lpicProp != NULL)
	{
		exUTCDate = ICalTimeTypeToUTC(lpicRootEvent, lpicProp);
		exLocalDate = UTCToLocal(exUTCDate, sTimeZone);

		lpRec->addDeletedException(exLocalDate);

		lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_EXDATE_PROPERTY);
	}

	*lppRecurrence = lpRec;

exit:
	return hr;
}

HRESULT ICalRecurrence::HrMakeMAPIException(icalcomponent *lpEventRoot, icalcomponent *lpicEvent, icalitem *lpIcalItem, LPSPropTagArray lpNamedProps, ECIConv *lpIConv, icalitem::exception *lpEx)
{
	HRESULT hr = hrSuccess;
	icalproperty *lpicProp = NULL;
	icaltimetype icStart;
	icaltimetype icRecId;
	time_t ttStartLocalTime = 0;
	time_t ttEndLocalTime = 0;
	time_t ttStartUtcTime = 0;
	time_t ttEndUtcTime = 0;
	time_t ttOriginalTime = 0;
	ULONG ulId = 0;
	ULONG i = 0;
	SPropValue sPropVal;
	std::string strIcalProp;
	bool bXMS = false;
	icalcomponent *lpicAlarm = NULL;
	LONG ulRemindBefore = 0;
	time_t ttReminderTime = 0;
	bool bReminderSet = false;
	std::list<SPropValue>::iterator iProp;

	bool abOldPresent[8] = {false};
	bool abNewPresent[8] = {false};
	SizedSPropTagArray(8, sptaCopy) = { 8, {
			PR_SUBJECT,
			CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_LOCATION], PT_STRING8),
			PR_BODY,
			CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_LABEL], PT_LONG),
			CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_REMINDERSET], PT_BOOLEAN),
			CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_REMINDERMINUTESBEFORESTART], PT_LONG),
			CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_ALLDAYEVENT], PT_BOOLEAN),
			CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_BUSYSTATUS], PT_LONG)
	} };
	bool bOldIsAllDay = false;

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_RECURRENCEID_PROPERTY);
	if (!lpicProp) {
		// you tricked me! it's not an exception at all!
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	icRecId = icalvalue_get_datetime(icalproperty_get_value(lpicProp));
	if(icRecId.is_date)
		ttOriginalTime = icaltime_as_timet(icRecId);
	else
		ttOriginalTime = ICalTimeTypeToUTC(lpEventRoot, lpicProp);
	
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_DTSTART_PROPERTY);
	if (lpicProp == NULL) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	icStart = icalvalue_get_datetime(icalproperty_get_value(lpicProp));
	ttStartLocalTime = icaltime_as_timet(icStart);
	ttStartUtcTime = ICalTimeTypeToUTC(lpEventRoot, lpicProp);

	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_DTEND_PROPERTY);
	if (lpicProp == NULL) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	ttEndLocalTime = icaltime_as_timet(icalvalue_get_datetime(icalproperty_get_value(lpicProp)));
	ttEndUtcTime = ICalTimeTypeToUTC(lpEventRoot,lpicProp);
	
	hr = lpIcalItem->lpRecurrence->addModifiedException(ttStartLocalTime, ttEndLocalTime, ttOriginalTime, &ulId);
	if (hr != hrSuccess)
		goto exit;

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRINGBASE], PT_SYSTIME);
	UnixTimeToFileTime(ttOriginalTime , &sPropVal.Value.ft);
	lpEx->lstMsgProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_EXCEPTION_STARTTIME;
	UnixTimeToFileTime(ttStartLocalTime, &sPropVal.Value.ft);
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_EXCEPTION_ENDTIME;
	UnixTimeToFileTime(ttEndLocalTime, &sPropVal.Value.ft);
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_APPTSTARTWHOLE], PT_SYSTIME);
	UnixTimeToFileTime(ttStartUtcTime , &sPropVal.Value.ft);	
	lpEx->lstMsgProps.push_back(sPropVal);
	
	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_COMMONSTART], PT_SYSTIME);
	lpEx->lstMsgProps.push_back(sPropVal);

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRENCE_START], PT_SYSTIME);
	lpEx->lstMsgProps.push_back(sPropVal);

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_APPTENDWHOLE], PT_SYSTIME);
	UnixTimeToFileTime(ttEndUtcTime, &sPropVal.Value.ft);	
	lpEx->lstMsgProps.push_back(sPropVal);

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_COMMONEND], PT_SYSTIME);
	lpEx->lstMsgProps.push_back(sPropVal);

	sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRENCE_END], PT_SYSTIME);
	lpEx->lstMsgProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_EXCEPTION_ENDTIME;
	UnixTimeToFileTime(ttEndLocalTime, &sPropVal.Value.ft);
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_DISPLAY_NAME;
	HrCopyString(NULL, lpIcalItem->base, "Untitled", &sPropVal.Value.lpszA);
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_ATTACH_METHOD;
	sPropVal.Value.ul = ATTACH_EMBEDDED_MSG;
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_ATTACH_FLAGS;
	sPropVal.Value.ul = 0;
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_ATTACHMENT_LINKID;
	sPropVal.Value.ul = 0;
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_ATTACHMENT_HIDDEN;
	sPropVal.Value.b = TRUE;
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_ATTACHMENT_FLAGS;
	sPropVal.Value.ul = 2;
	lpEx->lstAttachProps.push_back(sPropVal);

	sPropVal.ulPropTag = PR_MESSAGE_CLASS;
	HrCopyString(NULL, lpIcalItem->base, "IPM.OLE.CLASS.{00061055-0000-0000-C000-000000000046}", &sPropVal.Value.lpszA);
	lpEx->lstMsgProps.push_back(sPropVal);

	// copy properties to exception and test if changed
	for (iProp = lpIcalItem->lstMsgProps.begin(); iProp != lpIcalItem->lstMsgProps.end(); iProp++) {
		for (i = 0; i < sptaCopy.cValues; i++) {
			if (sptaCopy.aulPropTag[i] == (*iProp).ulPropTag) {
				abOldPresent[i] = true;
				if (sptaCopy.aulPropTag[i] != PR_BODY) // no need to copy body
					lpEx->lstMsgProps.push_back(*iProp);
				if (sptaCopy.aulPropTag[i] == CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_ALLDAYEVENT], PT_BOOLEAN))
					bOldIsAllDay = iProp->Value.b; // remember allday event status
				break;
			}
		}
	}

	// find exceptional properties
	// TODO: should actually look at original message, and check for differences
	lpicProp = icalcomponent_get_first_property(lpicEvent, ICAL_ANY_PROPERTY);
	while (lpicProp) {

		switch (icalproperty_isa(lpicProp)) {
		case ICAL_SUMMARY_PROPERTY:
			strIcalProp = lpIConv->convert(icalproperty_get_summary(lpicProp));
			hr = lpIcalItem->lpRecurrence->setModifiedSubject(ulId, strIcalProp.c_str());
			if (hr != hrSuccess)
				goto exit;
			sPropVal.ulPropTag = PR_SUBJECT;
			HrCopyString (NULL, lpIcalItem->base, strIcalProp.c_str(), &sPropVal.Value.lpszA);
			lpEx->lstMsgProps.push_back(sPropVal);
			abNewPresent[0] = true;
			break;
		case ICAL_LOCATION_PROPERTY:
			strIcalProp = lpIConv->convert(icalproperty_get_location(lpicProp));
			hr = lpIcalItem->lpRecurrence->setModifiedLocation(ulId, strIcalProp.c_str());
			if (hr != hrSuccess)
				goto exit;
			sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_LOCATION], PT_STRING8);
			HrCopyString (NULL, lpIcalItem->base, strIcalProp.c_str(), &sPropVal.Value.lpszA);
			lpEx->lstMsgProps.push_back(sPropVal);
			abNewPresent[1] = true;
			break;
		case ICAL_TRANSP_PROPERTY:
			if (!bXMS) {
				ULONG ulBusyStatus = 2;	// default busy
				switch(icalproperty_get_transp(lpicProp)){
					case ICAL_TRANSP_TRANSPARENT: 
						ulBusyStatus = 0;
						break;
					case ICAL_TRANSP_OPAQUE:					
						ulBusyStatus = 2;
						break;
				}
				hr = lpIcalItem->lpRecurrence->setModifiedBusyStatus(ulId, ulBusyStatus);
				if (hr != hrSuccess)
					goto exit;

				sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_BUSYSTATUS], PT_LONG);
				sPropVal.Value.ul = ulBusyStatus;
				lpEx->lstMsgProps.push_back(sPropVal);
				abNewPresent[7] = true;
			}
			break;
		case ICAL_X_PROPERTY:
			if (strcmp(icalproperty_get_x_name(lpicProp), "X-MICROSOFT-CDO-BUSYSTATUS") == 0) {
				ULONG ulBusyStatus = 2;	// default busy
				const char *lpszIcalProp = icalproperty_get_x(lpicProp);
				if (strcmp(lpszIcalProp, "FREE") == 0)
					ulBusyStatus = 0;
				else if (strcmp(lpszIcalProp, "TENTATIVE") == 0)
					ulBusyStatus = 1;
				else if (strcmp(lpszIcalProp, "OOF") == 0)
					ulBusyStatus = 3;

				hr = lpIcalItem->lpRecurrence->setModifiedBusyStatus(ulId, ulBusyStatus);
				if (hr != hrSuccess)
					goto exit;

				sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_BUSYSTATUS], PT_LONG);
				sPropVal.Value.ul = ulBusyStatus;
				lpEx->lstMsgProps.push_back(sPropVal);

				bXMS = true;
				abNewPresent[7] = true;
			}
			break;
		case ICAL_DESCRIPTION_PROPERTY:
			strIcalProp = lpIConv->convert(icalproperty_get_description(lpicProp));
			hr = lpIcalItem->lpRecurrence->setModifiedBody(ulId);
			if (hr != hrSuccess)
				goto exit;
			sPropVal.ulPropTag = PR_BODY;
			HrCopyString (NULL, lpIcalItem->base, strIcalProp.c_str(), &sPropVal.Value.lpszA);
			lpEx->lstMsgProps.push_back(sPropVal);
			abNewPresent[2] = true;
			break;
		default:
			// ignore property
			break;
		};

		lpicProp = icalcomponent_get_next_property(lpicEvent, ICAL_ANY_PROPERTY);
	}

	// make sure these are not removed :| (body, label, reminderset, reminder minutes)
	abNewPresent[2] = abNewPresent[3] = abNewPresent[4] = abNewPresent[5] = true;
	// test if properties were just removed
	for (i = 0; i < sptaCopy.cValues; i++) {
		if (abOldPresent[i] == true && abNewPresent[i] == false) {
			iProp = find(lpEx->lstMsgProps.begin(), lpEx->lstMsgProps.end(), sptaCopy.aulPropTag[i]);
			if (iProp != lpEx->lstMsgProps.end()) {
				lpEx->lstMsgProps.erase(iProp);
				switch (i) {
				case 0:
					// subject
					hr = lpIcalItem->lpRecurrence->setModifiedSubject(ulId, std::string());
					if (hr != hrSuccess)
						goto exit;
					sPropVal.ulPropTag = PR_SUBJECT;
					sPropVal.Value.lpszA = "";
					lpEx->lstMsgProps.push_back(sPropVal);
					break;
				case 1:
					// location
					hr = lpIcalItem->lpRecurrence->setModifiedLocation(ulId, std::string());
					if (hr != hrSuccess)
						goto exit;
					sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_LOCATION], PT_STRING8);
					sPropVal.Value.lpszA = "";
					lpEx->lstMsgProps.push_back(sPropVal);
					break;
				case 2:
					// body, ignore!
					break;
				case 3:
					// label, ignore
					break;
				case 4:
					// reminder set, ignore
					break;
				case 5:
					// reminder minutes, ignore
					break;
				case 6:
					// allday event
					if (icStart.is_date != bOldIsAllDay) {
						// flip all day status
						hr = lpIcalItem->lpRecurrence->setModifiedSubType(ulId, 1);
						if (hr != hrSuccess)
							goto exit;

						sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_ALLDAYEVENT], PT_BOOLEAN);
						sPropVal.Value.ul = !bOldIsAllDay;
						lpEx->lstMsgProps.push_back(sPropVal);
					}
					break;
				case 7:
					// busy status, default: busy
					hr = lpIcalItem->lpRecurrence->setModifiedBusyStatus(ulId, 1);
					if (hr != hrSuccess)
						goto exit;

					sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_BUSYSTATUS], PT_LONG);
					sPropVal.Value.ul = 1;
					lpEx->lstMsgProps.push_back(sPropVal);
					break;
				};
			}
		}
	}
	

	// reminderset (flip reminder on/off) and reminderdelta (offset time change) in vevent component of this item
	lpicAlarm = icalcomponent_get_first_component(lpicEvent, ICAL_VALARM_COMPONENT);
	if (lpicAlarm) {
		hr = HrParseVAlarm(lpicAlarm, &ulRemindBefore, &ttReminderTime, &bReminderSet);
		if (hr == hrSuccess) {
			// abOldPresent[4] == reminderset marker in original message
			if (bReminderSet != abOldPresent[4]) {
				hr = lpIcalItem->lpRecurrence->setModifiedReminder(ulId, bReminderSet);
				if (hr != hrSuccess)
					goto exit;

				sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_REMINDERSET], PT_BOOLEAN);
				sPropVal.Value.b = bReminderSet;
				lpEx->lstMsgProps.push_back(sPropVal);

				if (ttReminderTime == 0)
					ttReminderTime = ttStartLocalTime;

				sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_REMINDERTIME], PT_SYSTIME);
				UnixTimeToFileTime(ttReminderTime, &sPropVal.Value.ft);
				lpEx->lstMsgProps.push_back(sPropVal);
			}

			hr = lpIcalItem->lpRecurrence->setModifiedReminderDelta(ulId, ulRemindBefore);
			if (hr != hrSuccess)
				goto exit;

			sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_REMINDERMINUTESBEFORESTART], PT_LONG);
			sPropVal.Value.ul = ulRemindBefore;
			lpEx->lstMsgProps.push_back(sPropVal);
		}
	} else {
		if (abOldPresent[4]) {
			// disable reminder in attachment
			hr = lpIcalItem->lpRecurrence->setModifiedReminder(ulId, 0);
			if (hr != hrSuccess)
				goto exit;

			sPropVal.ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_REMINDERSET], PT_BOOLEAN);
			sPropVal.Value.b = FALSE;
			lpEx->lstMsgProps.push_back(sPropVal);
		}
	}

exit:
	return hr;
}

// finalizes recurrence struct with exceptions into the mapi message
HRESULT ICalRecurrence::HrMakeMAPIRecurrence(recurrence *lpRecurrence, LPSPropTagArray lpNamedProps, LPMESSAGE lpMessage)
{
	HRESULT hr = hrSuccess;
	char *lpRecBlob = NULL;
	unsigned int ulRecBlob = 0;
	LPSPropValue lpPropVal = NULL;
	LPSPropValue lpsPropRecPattern = NULL;
	std::string strHRS;
	ULONG i = 0;

	hr = lpRecurrence->HrGetRecurrenceState(&lpRecBlob, &ulRecBlob);
	if (hr != hrSuccess)
		goto exit;

	hr = lpRecurrence->HrGetHumanReadableString(&strHRS);
	if (hr != hrSuccess)
		goto exit;

	// adjust number of props
	MAPIAllocateBuffer(sizeof(SPropValue)*4, (void**)&lpPropVal);

	lpPropVal[i].ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRING], PT_BOOLEAN);
	lpPropVal[i].Value.b = TRUE;
	i++;

	// TODO: combine with icon index in vevent .. the item may be a meeting request (meeting+recurring==1027)
	lpPropVal[i].ulPropTag = PR_ICON_INDEX;
	lpPropVal[i].Value.ul = 1025;
	i++;

	lpPropVal[i].ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRENCESTATE], PT_BINARY);
	lpPropVal[i].Value.bin.lpb = (BYTE*)lpRecBlob;
	lpPropVal[i].Value.bin.cb = ulRecBlob;
	i++;

	hr = HrGetOneProp(lpMessage, CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRENCEPATTERN], PT_STRING8), &lpsPropRecPattern);
	if(hr != hrSuccess)
	{
		lpPropVal[i].ulPropTag = CHANGE_PROP_TYPE(lpNamedProps->aulPropTag[PROP_RECURRENCEPATTERN], PT_STRING8);
		lpPropVal[i].Value.lpszA = (char*)strHRS.c_str();
		i++;
	}

	hr = lpMessage->SetProps(i, lpPropVal, NULL);
	if (FAILED(hr))
		goto exit;

exit:
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	if(lpRecBlob)
		MAPIFreeBuffer(lpRecBlob);

	if(lpsPropRecPattern)
		MAPIFreeBuffer(lpsPropRecPattern);
	return hr;
}

// mapi -> ical
HRESULT ICalRecurrence::HrCreateICalRecurrence(TIMEZONE_STRUCT sTimeZone, bool bIsAllDay, recurrence *lpRecurrence, icalcomponent *lpicEvent)
{
	HRESULT hr = hrSuccess;
	icalrecurrencetype icRRule;
	std::list<time_t> lstExceptions;
	std::list<time_t>::iterator iException;
	icaltimetype ittExDate;
	TIMEZONE_STRUCT sTZgmt = {0};

	hr = HrCreateICalRecurrenceType(sTimeZone, bIsAllDay, lpRecurrence, &icRRule);
	if (hr != hrSuccess)
		goto exit;

	icalcomponent_add_property(lpicEvent, icalproperty_new_rrule(icRRule));


	// all delete exceptions are in the delete list,
	lstExceptions = lpRecurrence->getDeletedExceptions();
	if (!lstExceptions.empty()) {
		// add EXDATE props
		for (iException = lstExceptions.begin(); iException != lstExceptions.end(); iException++) {
			if(bIsAllDay)
			{
				ittExDate = icaltime_from_timet(LocalToUTC(*iException, sTZgmt), bIsAllDay);
			}
			else
				ittExDate = icaltime_from_timet(LocalToUTC(*iException, sTimeZone), 0);

			ittExDate.is_utc = 1;

			icalcomponent_add_property(lpicEvent, icalproperty_new_exdate(ittExDate));
		}
	}

	// modified exceptions are done by the caller because of the attachments with info

exit:
	return hr;
}

HRESULT ICalRecurrence::HrCreateICalRecurrenceType(TIMEZONE_STRUCT sTimeZone, bool bIsAllday, recurrence *lpRecurrence, icalrecurrencetype *lpicRRule)
{
	HRESULT hr = hrSuccess;
	struct icalrecurrencetype icRec;

	icalrecurrencetype_clear(&icRec);

	switch (lpRecurrence->getFrequency()) {
	case recurrence::DAILY:
		icRec.freq = ICAL_DAILY_RECURRENCE;
		// only weekdays selected in outlook
		if (lpRecurrence->getWeekDays())
			WeekDaysToICalArray(lpRecurrence->getWeekDays(), &icRec);
		break;
	case recurrence::WEEKLY:
		icRec.freq = ICAL_WEEKLY_RECURRENCE;
		WeekDaysToICalArray(lpRecurrence->getWeekDays(), &icRec);
		break;
	case recurrence::MONTHLY:
		icRec.freq = ICAL_MONTHLY_RECURRENCE;
		if (lpRecurrence->getWeekNumber() != 0) {
			// mapi patterntype == 3
			// only 1 day should be set!
			if(lpRecurrence->getWeekDays() == 127)
			{
				// All Weekdays are set for recurrence type "second" "day" of month.
				// SU,MO,TU,WE,TH,FR,SA -> weekdays = 127.
				icRec.by_month_day[0] = lpRecurrence->getWeekNumber() == 5 ? -1: lpRecurrence->getWeekNumber(); // hack to handle nth day month type of rec.
				icRec.by_month_day[1] = ICAL_RECURRENCE_ARRAY_MAX;
				break;
			}
			else if (lpRecurrence->getWeekNumber() == 5)
				icRec.by_day[0] = ((log((double)lpRecurrence->getWeekDays())/log(2.0)) + 8 + 1 ) * -1;  // corrected last weekday
			else
				icRec.by_day[0] = (int)(log((double)lpRecurrence->getWeekDays())/log(2.0)) + (8 * lpRecurrence->getWeekNumber() ) +1; // +1 because outlook starts on sunday

			icRec.by_day[1] = ICAL_RECURRENCE_ARRAY_MAX;
		} else {
			// mapi patterntype == 2
			icRec.by_month_day[0] = lpRecurrence->getDayOfMonth();
			icRec.by_month_day[1] = ICAL_RECURRENCE_ARRAY_MAX;
		}
		break;
	case recurrence::YEARLY:
		icRec.freq = ICAL_YEARLY_RECURRENCE;
		if (lpRecurrence->getWeekNumber() != 0) {
			// mapi patterntype == 3
			// only 1 day should be set!
			if (lpRecurrence->getWeekNumber() == 5)
				icRec.by_day[0] = ((log((double)lpRecurrence->getWeekDays())/log(2.0)) + 8 + 1 ) * -1; 
			else
				icRec.by_day[0] = (int)(log((double)lpRecurrence->getWeekDays())/log(2.0)) + (8 * lpRecurrence->getWeekNumber() ) +1; // +1 because outlook starts on sunday

			icRec.by_day[1] = ICAL_RECURRENCE_ARRAY_MAX;

			icRec.by_month[0] = lpRecurrence->getMonth();
			icRec.by_month[1] = ICAL_RECURRENCE_ARRAY_MAX;
		}
		break;
	default:
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	icRec.interval = lpRecurrence->getInterval();

	switch (lpRecurrence->getEndType()) {
	case recurrence::DATE:
		/* From the RFC:
		   
		   The UNTIL rule part defines a date-time value which bounds the
		   recurrence rule in an inclusive manner. If the value specified by
		   UNTIL is synchronized with the specified recurrence, this date or
		   date-time becomes the last instance of the recurrence. If specified
		   as a date-time value, then it MUST be specified in an UTC time
		   format.
		*/
		icRec.count = 0;
		// if untiltime is saved as UTC it breaks last occurrence.
		icRec.until = icaltime_from_timet(lpRecurrence->getEndDate() + lpRecurrence->getStartTimeOffset(), bIsAllday);
		icRec.until.is_utc = 0;
		break;
	case recurrence::NUMBER:
		icRec.count = lpRecurrence->getCount();
		icRec.until = icaltime_null_time();
		break;
	case recurrence::NEVER:
		icRec.count = 0;
		icRec.until = icaltime_null_time();
		break;
	};

	*lpicRRule = icRec;

exit:
	return hr;
}

// input: 0x3E == 011 1110, output: MO,TU,WE,TH,FR
HRESULT ICalRecurrence::WeekDaysToICalArray(ULONG ulWeekDays, struct icalrecurrencetype *lpRec)
{
	int i = 0, j = 0;

	for (i = 0; i < 7; i++) {
		if ((ulWeekDays >> i) & 1) {
			lpRec->by_day[j++] = i+1;
		}
	}
	lpRec->by_day[j] = ICAL_RECURRENCE_ARRAY_MAX;
	return hrSuccess;
}

HRESULT ICalRecurrence::HrMakeICalException(icalcomponent *lpicEvent, icalcomponent **lppicException)
{
	HRESULT hr = hrSuccess;
	icalcomponent *lpicException = NULL;
	icalproperty *lpicProp = NULL;
	
	lpicException = icalcomponent_new_clone(lpicEvent);

	// these are always different in an exception
	lpicProp = icalcomponent_get_first_property(lpicException, ICAL_DTSTART_PROPERTY);
	if (lpicProp) {
		icalcomponent_remove_property(lpicException, lpicProp);
		icalproperty_free(lpicProp);
	}

	lpicProp = icalcomponent_get_first_property(lpicException, ICAL_DTEND_PROPERTY);
	if (lpicProp) {
		icalcomponent_remove_property(lpicException, lpicProp);
		icalproperty_free(lpicProp);
	}

	// exceptions don't have the rrule again
	lpicProp = icalcomponent_get_first_property(lpicException, ICAL_RRULE_PROPERTY);
	if (lpicProp) {
		icalcomponent_remove_property(lpicException, lpicProp);
		icalproperty_free(lpicProp);
	}
	// exceptions don't have the deleted exceptions anymore
	while ((lpicProp = icalcomponent_get_first_property(lpicException, ICAL_EXDATE_PROPERTY)) != NULL) {
		icalcomponent_remove_property(lpicException, lpicProp);
		icalproperty_free(lpicProp);
	}

	// exceptions can't be different in privacy
	lpicProp = icalcomponent_get_first_property(lpicException, ICAL_CLASS_PROPERTY);
	if (lpicProp) {
		icalcomponent_remove_property(lpicException, lpicProp);
		icalproperty_free(lpicProp);
	}

	*lppicException = lpicException;

	return hr;
}
