/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapidefs.h>
#include <mapicode.h>
#include <mapitags.h>
#include <mapiutil.h>

#include "ECIConv.h"
#include "WSTransport.h"
#include "ProviderUtil.h"
#include "SymmetricCrypt.h"
#include "soapH.h"
#include "ZarafaUtil.h"


// The header files we use for communication with the server
#include "ZarafaCode.h"
#include "soapZarafaCmdProxy.h"
#include "ZarafaCmd.nsmap"
#include "Mem.h"
#include "ECGuid.h"
#include "guid.h"

#include "SOAPUtils.h"
#include "WSUtil.h"
#include "Util.h"
#include "mapiext.h"

#include "WSABTableView.h"
#include "WSABPropStorage.h"
#include "ecversion.h"
#include "ClientUtil.h"
#include "ECSessionGroupManager.h"
#include "stringutil.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define	pbMUIDECSABGuid	"\xac\x21\xa9\x50\x40\xd3\xee\x48\xb3\x19\xfb\xa7\x53\x30\x44\x25"

/*
 *
 * This is the main WebServices transport object. All communications with the
 * web service server is done through this object. Also, this file is the 
 * coupling point between MAPI and our internal (network) formats, and
 * codes. This means that any classes communicating with this one either
 * use MAPI syntax (ie. MAPI_E_NOT_ENOUGH_MEMORY) OR use the EC syntax
 * (ie. EC_E_NOT_FOUND), but never BOTH.
 *
 */

#define START_SOAP_CALL retry:
#define END_SOAP_CALL 	\
	if(er == ZARAFA_E_END_OF_SESSION) { if(HrReLogon() == hrSuccess) goto retry; } \
	hr = ZarafaErrorToMAPIError(er, MAPI_E_NOT_FOUND); \
	if(hr != hrSuccess) \
		goto exit;

WSTransport::WSTransport()  : ECUnknown ("WSTransport")
{
    pthread_mutexattr_t attr;
    
	m_lpCmd = NULL;
	m_ecSessionGroupId = 0;
	m_ulReloadId = 1;
	memset(&m_sServerGuid, 0, sizeof(m_sServerGuid));
	pthread_mutexattr_init(&attr);
	pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE_NP);
	pthread_mutex_init(&m_hDataLock, &attr);
	pthread_mutex_init(&m_mutexSessionReload, &attr);
}

WSTransport::~WSTransport()
{
	if(m_lpCmd) {
		this->HrLogOff();
	}

	pthread_mutex_destroy(&m_hDataLock);
	pthread_mutex_destroy(&m_mutexSessionReload);
}

HRESULT WSTransport::QueryInterface(REFIID refiid, void **lpvoid)
{
	if(refiid == IID_ECTransport) {
		AddRef();
		*lpvoid = this;
		return hrSuccess;
	} else return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

HRESULT WSTransport::Create(WSTransport **lppTransport)
{
	HRESULT hr = hrSuccess;
	WSTransport *lpTransport = NULL;

	lpTransport = new WSTransport();

	hr = lpTransport->QueryInterface(IID_ECTransport, (void **) lppTransport);

	if(hr != hrSuccess)
		delete lpTransport;

	return hr;
}

/* Creates a transport working on the same session and session group as this transport */
HRESULT WSTransport::HrClone(WSTransport **lppTransport)
{
	HRESULT hr = hrSuccess;
	WSTransport *lpTransport = NULL;

	hr = WSTransport::Create(&lpTransport);
	if(hr != hrSuccess)
		goto exit;

	hr = CreateSoapTransport(m_sProfileProps, &lpTransport->m_lpCmd);
	if(hr != hrSuccess)
		goto exit;
	
	lpTransport->m_ecSessionId = this->m_ecSessionId;
	lpTransport->m_ecSessionGroupId = this->m_ecSessionGroupId;

	*lppTransport = lpTransport;

exit:
	return hr;
}

HRESULT WSTransport::HrOpenTransport(LPMAPISUP lpMAPISup, WSTransport **lppTransport, BOOL bOffline)
{
	HRESULT			hr = hrSuccess;
	WSTransport		*lpTransport = NULL;
	std::string		strServerPath;
	sGlobalProfileProps	sProfileProps;

		// Get the username and password from the profile settings
	hr = ClientUtil::GetGlobalProfileProperties(lpMAPISup, &sProfileProps);
	if(hr != hrSuccess)
		goto exit;

	// TODO: check usernameand serverpath
	
	// Create a transport for this user
	hr = WSTransport::Create(&lpTransport);
	if(hr != hrSuccess)
		goto exit;


	// Log on the transport to the server
	hr = lpTransport->HrLogon(sProfileProps);
	if(hr != hrSuccess) 
		goto exit;

	*lppTransport = lpTransport;

exit:

	if (hr != hrSuccess && lpTransport)
		lpTransport->Release();

	return hr;

}

HRESULT WSTransport::LockSoap()
{
	pthread_mutex_lock(&m_hDataLock);
	return erSuccess;
}

HRESULT WSTransport::UnLockSoap()
{
	//Clean up data create with soap_malloc
	if (m_lpCmd && m_lpCmd->soap)
		soap_end(m_lpCmd->soap);

	pthread_mutex_unlock(&m_hDataLock);
	return erSuccess;
}

HRESULT WSTransport::HrLogon(sGlobalProfileProps sProfileProps)
{
	HRESULT			hr = hrSuccess;
	ECRESULT		er = erSuccess;
	unsigned int	ulCapabilities = 0;
	unsigned int	ulServerCapabilities = 0;
	ECSESSIONID		ecSessionId = 0;
	ZarafaCmd*		lpCmd = NULL;
	bool			bPipeConnection = false;
	struct logonResponse sResponse;
	struct xsd__base64Binary sLicenseRequest = {0,0};
	

	if (strncmp("file:", sProfileProps.strServerPath.c_str(), 5) == 0)
		bPipeConnection = true;
	else
		bPipeConnection = false;

	if(m_lpCmd == NULL) {
		if (CreateSoapTransport(sProfileProps, &lpCmd) != hrSuccess)
		{
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
	} else {
		lpCmd = m_lpCmd;
	}

	/*
	 * Determine if we can attach this session to a SessionGroup
	 * FIXME: Should we do this here or in the ECMSProvider but that
	 * would mean we don't group AddressBook and Store Sessions together...
	 */
	if (!(sProfileProps.ulProfileFlags & EC_PROFILE_FLAGS_NO_NOTIFICATIONS))
		m_ecSessionGroupId = g_ecSessionManager.GetSessionGroupId(sProfileProps);

	ulCapabilities |= ZARAFA_CAP_MAILBOX_OWNER | ZARAFA_CAP_MULTI_SERVER;

	if (sizeof(ECSESSIONID) == 8)
		ulCapabilities |= ZARAFA_CAP_LARGE_SESSIONID;

	if (bPipeConnection == false) {
		// all connections except pipes use compression
		if(! (sProfileProps.ulProfileFlags & EC_PROFILE_FLAGS_NO_COMPRESSION))
			ulCapabilities |= ZARAFA_CAP_COMPRESSION; // only to remote server .. windows?

		// try single signon logon
		er = TrySSOLogon(lpCmd, sProfileProps.strUserName.c_str(), ulCapabilities, m_ecSessionGroupId, (char *)GetAppName().c_str(), &ecSessionId, &ulServerCapabilities, &m_llFlags, &m_sServerGuid);
		if (er == erSuccess)
			goto auth;
	}

	LockSoap();
	
	// Login with username and password
	if (SOAP_OK != lpCmd->ns__logon((char *)sProfileProps.strUserName.c_str(), (char *)sProfileProps.strPassword.c_str(), PROJECT_VERSION_CLIENT_STR, ulCapabilities, sLicenseRequest, m_ecSessionGroupId, (char *)GetAppName().c_str(), &sResponse))
		er = ZARAFA_E_SERVER_NOT_RESPONDING;
	else
		er = sResponse.er;

	// If the user was denied, and the server did not support encryption, and the password was encrypted, decrypt it now
	// so that we support older servers. If the password was not encrypted, then it was just wrong, and if the server supported encryption
	// then the password was also simply wrong.
	if(er == ZARAFA_E_LOGON_FAILED && SymmetricIsCrypted(sProfileProps.strPassword.c_str()) && !(sResponse.ulCapabilities & ZARAFA_CAP_CRYPT)) {
		// Login with username and password
		if (SOAP_OK != lpCmd->ns__logon((char *)sProfileProps.strUserName.c_str(), (char *)SymmetricDecrypt(sProfileProps.strPassword.c_str()).c_str(), PROJECT_VERSION_CLIENT_STR, ulCapabilities, sLicenseRequest, m_ecSessionGroupId, (char *)GetAppName().c_str(), &sResponse))
			er = ZARAFA_E_SERVER_NOT_RESPONDING;
		else
			er = sResponse.er;
	}

	hr = ZarafaErrorToMAPIError(er, MAPI_E_LOGON_FAILED);

	if (hr != hrSuccess) {
	    UnLockSoap();
		goto exit;
    }

	// PROJECT_VERSION_SERVER_STR == "0,6,00"
	if (!sResponse.lpszVersion || sResponse.lpszVersion[2] != '6') {
	    UnLockSoap();
		hr = MAPI_E_VERSION;
		goto exit;
	}


#if defined WIN32 && !defined _DEBUG
	// Check license info

	// For backward-compatibility: don't check license if server doesn't support licensing (pre 6.20), and don't check on the offline server
	if(sProfileProps.strServerPath.substr(0,7) != "file://" && sResponse.ulCapabilities & ZARAFA_CAP_LICENSE_SERVER) {
		hr = ProcessLicenseResponseEnc(ulTrackingId, sResponse.sLicenseResponse.__ptr, sResponse.sLicenseResponse.__size, &m_llFlags);

		if(hr != erSuccess) {
		    UnLockSoap();
			goto exit;
        }
	}

#endif

	ecSessionId = sResponse.ulSessionId;
	ulServerCapabilities = sResponse.ulCapabilities;

	if (sResponse.sServerGuid.__ptr != NULL && sResponse.sServerGuid.__size == sizeof(m_sServerGuid))
		memcpy(&m_sServerGuid, sResponse.sServerGuid.__ptr, sizeof(m_sServerGuid));

	UnLockSoap();
	// Login ok

auth: // User have a logon

	if (ulServerCapabilities & ZARAFA_CAP_COMPRESSION) {
		soap_set_imode(lpCmd->soap, SOAP_ENC_ZLIB);	// also autodetected
		soap_set_omode(lpCmd->soap, SOAP_ENC_ZLIB | SOAP_IO_CHUNK);
	}

	m_sProfileProps = sProfileProps;
	m_ulServerCapabilities = ulServerCapabilities;
	m_ecSessionId = ecSessionId;
	m_lpCmd = lpCmd;

exit:

	if (sLicenseRequest.__ptr)
		delete[] sLicenseRequest.__ptr;

	if(hr != hrSuccess) {
	    // UGLY FIX: due to the ugly code above that does lpCmd = m_lpCmd
	    // we need to check that we're not deleting our m_lpCmd. We also cannot
	    // set m_lpCmd to NULL since various functions in WSTransport rely on the
	    // fact that m_lpCmd is good after a successful HrLogon() call.
		if(lpCmd && lpCmd != m_lpCmd) {
			free((void*)lpCmd->endpoint);	// because of strdup()
		
			if (lpCmd->soap->proxy_host)
				free((void*)lpCmd->soap->proxy_host);

			if (lpCmd->soap->proxy_userid)
				free((void*)lpCmd->soap->proxy_userid);

			if (lpCmd->soap->proxy_passwd)
				free((void*)lpCmd->soap->proxy_passwd);

			delete lpCmd;
		}
	}

	return hr;
}

HRESULT WSTransport::HrSetRecvTimeout(unsigned int ulSeconds)
{
	HRESULT		hr = hrSuccess;

	if(this->m_lpCmd)
		this->m_lpCmd->soap->recv_timeout = ulSeconds;
	else
		hr = MAPI_E_NOT_INITIALIZED;

	return hr;
}


HRESULT WSTransport::CreateAndLogonAlternate(LPCTSTR szServer, WSTransport **lppTransport)
{
	HRESULT				hr = hrSuccess;
	WSTransport			*lpTransport = NULL;
	sGlobalProfileProps	sProfileProps = m_sProfileProps;

	if (!lppTransport)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = WSTransport::Create(&lpTransport);
	if (hr != hrSuccess)
		goto exit;

	sProfileProps.strServerPath = szServer;

	hr = lpTransport->HrLogon(sProfileProps);
	if (hr != hrSuccess)
		goto exit;

	*lppTransport = lpTransport;
	lpTransport = NULL;

	m_sProfileProps = sProfileProps;

exit:
	if (lpTransport)
		lpTransport->Release();

	return hr;
}

HRESULT WSTransport::HrReLogon()
{
	HRESULT hr = hrSuccess;
	SESSIONRELOADLIST::iterator iter;

	hr = HrLogon(m_sProfileProps);
	if(hr != hrSuccess)
		goto exit;

	// Notify new session to listeners
	pthread_mutex_lock(&m_mutexSessionReload);
	for(iter = m_mapSessionReload.begin(); iter != m_mapSessionReload.end(); iter++) {
		iter->second.second(iter->second.first, this->m_ecSessionId);
	}
	pthread_mutex_unlock(&m_mutexSessionReload);

exit:
	return hr;
}

ECRESULT WSTransport::TrySSOLogon(ZarafaCmd* lpCmd, LPCTSTR szUsername, unsigned int ulCapabilities, ECSESSIONGROUPID ecSessionGroupId, char *szAppName, ECSESSIONID* lpSessionId, unsigned int* lpulServerCapabilities, unsigned long long *lpllFlags, LPGUID lpsServerGuid)
{
	ECRESULT		er = ZARAFA_E_LOGON_FAILED;
	return er;
}

HRESULT WSTransport::HrGetPublicStore(ULONG* lpcbStoreID, LPENTRYID* lppStoreID, string *lpstrRedirServer)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct getStoreResponse sResponse;

	LockSoap();

	if(lppStoreID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getPublicStore(m_ecSessionId, &sResponse))
			er = ZARAFA_E_SERVER_NOT_RESPONDING;
		else
			er = sResponse.er;
	}
	//END_SOAP_CALL
	if(er == ZARAFA_E_END_OF_SESSION) { if(HrReLogon() == hrSuccess) goto retry; }
	hr = ZarafaErrorToMAPIError(er, MAPI_E_NOT_FOUND);
	if (hr == MAPI_E_UNABLE_TO_COMPLETE)
	{
		if (lpstrRedirServer)
			*lpstrRedirServer = sResponse.lpszServerPath;
		else
			hr = MAPI_E_NOT_FOUND;
	}
	if(hr != hrSuccess)
		goto exit;

	// Create a client store entry, add the servername
	hr = WrapServerClientStoreEntry(sResponse.lpszServerPath ? sResponse.lpszServerPath : m_sProfileProps.strServerPath.c_str(), &sResponse.sStoreId, lpcbStoreID, lppStoreID);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetStore(ULONG cbMasterID, LPENTRYID lpMasterID, ULONG* lpcbStoreID, LPENTRYID* lppStoreID, ULONG* lpcbRootID, LPENTRYID* lppRootID, string *lpstrRedirServer)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	
	entryId		sEntryId = {0,0}; // Do not free
	struct getStoreResponse sResponse;
	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	LockSoap();

	if(lpMasterID) {

		hr = UnWrapServerClientStoreEntry(cbMasterID, lpMasterID, &cbUnWrapStoreID, &lpUnWrapStoreID);
		if(hr != hrSuccess)
			goto exit;

		sEntryId.__ptr = (unsigned char*)lpUnWrapStoreID;
		sEntryId.__size = cbUnWrapStoreID;
	}

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getStore(m_ecSessionId, (lpMasterID)?&sEntryId:NULL, &sResponse))
			er = ZARAFA_E_SERVER_NOT_RESPONDING;
		else
			er = sResponse.er;
	}
	//END_SOAP_CALL
	if(er == ZARAFA_E_END_OF_SESSION) { if(HrReLogon() == hrSuccess) goto retry; }
	hr = ZarafaErrorToMAPIError(er, MAPI_E_NOT_FOUND);
	if (hr == MAPI_E_UNABLE_TO_COMPLETE)
	{
		if (lpstrRedirServer)
			*lpstrRedirServer = sResponse.lpszServerPath;
		else
			hr = MAPI_E_NOT_FOUND;
	}
	if(hr != hrSuccess)
		goto exit;

	if(lppRootID && lpcbRootID) {
		hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sRootId, lpcbRootID, lppRootID);
		if(hr != hrSuccess)
			goto exit;
	}

	if(lppStoreID && lpcbStoreID) {
		// Create a client store entry, add the servername
		hr = WrapServerClientStoreEntry(sResponse.lpszServerPath ? sResponse.lpszServerPath : m_sProfileProps.strServerPath.c_str(), &sResponse.sStoreId, lpcbStoreID, lppStoreID);
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	UnLockSoap();

	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT WSTransport::HrGetStoreName(ULONG cbStoreID, LPENTRYID lpStoreID, char **lppszStoreName)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId		sEntryId; // Do not free
	ULONG ulSize = 0;
	struct getStoreNameResponse sResponse;
	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	LockSoap();

	if(lpStoreID == NULL || lppszStoreName == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Remove the servername
	hr = UnWrapServerClientStoreEntry(cbStoreID, lpStoreID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	sEntryId.__ptr = (unsigned char*)lpUnWrapStoreID;
	sEntryId.__size = cbUnWrapStoreID;
	
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getStoreName(m_ecSessionId, sEntryId, &sResponse))
			er = ZARAFA_E_SERVER_NOT_RESPONDING;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	ulSize = strlen(sResponse.lpszStoreName)+1;
	hr = ECAllocateBuffer(ulSize, (void**)lppszStoreName);
	if(hr != hrSuccess)
		goto exit;

	memcpy(*lppszStoreName, sResponse.lpszStoreName, ulSize);

exit:
	UnLockSoap();

	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT WSTransport::HrLogOff()
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;

	if (m_lpCmd == NULL)
		return hrSuccess;

	LockSoap();

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__logoff(m_ecSessionId, &er) )
			er = ZARAFA_E_NETWORK_ERROR;

        er = erSuccess; // don't care
	}
	END_SOAP_CALL

	if (m_lpCmd != NULL) {
		if (m_lpCmd->endpoint != NULL)
			free((void*)m_lpCmd->endpoint);

		if (m_lpCmd->soap->proxy_host)
			free((void*)m_lpCmd->soap->proxy_host);

		if (m_lpCmd->soap->proxy_userid)
			free((void*)m_lpCmd->soap->proxy_userid);

		if (m_lpCmd->soap->proxy_passwd)
			free((void*)m_lpCmd->soap->proxy_passwd);

		delete m_lpCmd;
		m_lpCmd = NULL;
	}

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrCheckExistObject(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulFlags)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;
	entryId		sEntryId = {0}; // Do not free

	LockSoap();

	if(cbEntryID == 0 || lpEntryID == NULL) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryID, lpEntryID, &sEntryId, true);
	if(hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__checkExistObject(m_ecSessionId, sEntryId, ulFlags, &er))
			er = ZARAFA_E_SERVER_NOT_RESPONDING;

	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrOpenPropStorage(ULONG cbParentEntryID, LPENTRYID lpParentEntryID, ULONG cbEntryID, LPENTRYID lpEntryID, unsigned int ulFlags, IECPropStorage **lppPropStorage)
{
	HRESULT hr = hrSuccess;
	WSMAPIPropStorage *lpPropStorage = NULL;
	LPENTRYID	lpUnWrapParentID = NULL;
	ULONG		cbUnWrapParentID = 0;
	LPENTRYID	lpUnWrapEntryID = NULL;
	ULONG		cbUnWrapEntryID = 0;


	if (lpParentEntryID) {
		hr = UnWrapServerClientStoreEntry(cbParentEntryID, lpParentEntryID, &cbUnWrapParentID, &lpUnWrapParentID);
		if(hr != hrSuccess)
			goto exit;
	}

	hr = UnWrapServerClientStoreEntry(cbEntryID, lpEntryID, &cbUnWrapEntryID, &lpUnWrapEntryID);
	if(hr != hrSuccess)
		goto exit;

	hr = WSMAPIPropStorage::Create(cbUnWrapParentID, lpUnWrapParentID, cbUnWrapEntryID, lpUnWrapEntryID, ulFlags, m_lpCmd, m_hDataLock, m_ecSessionId, this->m_ulServerCapabilities, this, &lpPropStorage);
	if(hr != hrSuccess)
		goto exit;

	hr = lpPropStorage->QueryInterface(IID_IECPropStorage, (void **)lppPropStorage);

exit:
	if(lpPropStorage)
		lpPropStorage->Release();

	if(lpUnWrapEntryID)
		ECFreeBuffer(lpUnWrapEntryID);

	if(lpUnWrapParentID)
		ECFreeBuffer(lpUnWrapParentID);

	return hr;
}

HRESULT WSTransport::HrOpenParentStorage(ECGenericProp *lpParentObject, ULONG ulUniqueId, ULONG ulObjId, IECPropStorage *lpServerStorage, IECPropStorage **lppPropStorage)
{
	HRESULT hr = hrSuccess;
	ECParentStorage *lpPropStorage = NULL;

	hr = ECParentStorage::Create(lpParentObject, ulUniqueId, ulObjId, lpServerStorage, &lpPropStorage);
	if(hr != hrSuccess)
		goto exit;

	hr = lpPropStorage->QueryInterface(IID_IECPropStorage, (void **)lppPropStorage);

exit:
	if(lpPropStorage)
		lpPropStorage->Release();

	return hr;
}

HRESULT WSTransport::HrOpenABPropStorage(ULONG cbEntryID, LPENTRYID lpEntryID, IECPropStorage **lppPropStorage)
{
	HRESULT			hr = hrSuccess;
	WSABPropStorage *lpPropStorage = NULL;

	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;


	hr = UnWrapServerClientABEntry(cbEntryID, lpEntryID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	hr = WSABPropStorage::Create(cbUnWrapStoreID, lpUnWrapStoreID, m_lpCmd, m_hDataLock, m_ecSessionId, this, &lpPropStorage);

	if(hr != hrSuccess)
		goto exit;

	hr = lpPropStorage->QueryInterface(IID_IECPropStorage, (void **)lppPropStorage);

exit:
	if(lpPropStorage)
		lpPropStorage->Release();

	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT WSTransport::HrOpenFolderOps(ULONG cbEntryID, LPENTRYID lpEntryID, WSMAPIFolderOps **lppFolderOps)
{
	HRESULT hr = hrSuccess;

	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

//FIXME: create this function
//	hr = CheckEntryIDType(cbEntryID, lpEntryID, MAPI_FOLDER);
	if( hr != hrSuccess)
		goto exit;


	hr = UnWrapServerClientStoreEntry(cbEntryID, lpEntryID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	hr = WSMAPIFolderOps::Create(m_lpCmd, m_hDataLock, m_ecSessionId, cbUnWrapStoreID, lpUnWrapStoreID, this, lppFolderOps);

exit:
	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;

}

HRESULT WSTransport::HrOpenTableOps(ULONG ulType, ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID, ECMsgStore *lpMsgStore, WSTableView **lppTableOps)
{
	HRESULT hr = hrSuccess;
	
	/*
	FIXME: Do a check ?
	if(peid->ulType != MAPI_FOLDER && peid->ulType != MAPI_MESSAGE) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}*/

	hr = WSStoreTableView::Create(ulType, ulFlags, m_lpCmd, m_hDataLock, m_ecSessionId, cbEntryID, lpEntryID, lpMsgStore, this, lppTableOps);
	if(hr != hrSuccess)
		goto exit;

exit:
	return hr;
}

HRESULT WSTransport::HrOpenABTableOps(ULONG ulType, ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID, ECABLogon* lpABLogon, WSTableView **lppTableOps)
{
	HRESULT hr = hrSuccess;
	
	/*if(peid->ulType != MAPI_FOLDER && peid->ulType != MAPI_MESSAGE) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}*/

	hr = WSABTableView::Create(ulType, ulFlags, m_lpCmd, m_hDataLock, m_ecSessionId, cbEntryID, lpEntryID, lpABLogon, this, lppTableOps);

	return hr;
}

HRESULT WSTransport::HrOpenTableOutGoingQueueOps(ULONG cbStoreEntryID, LPENTRYID lpStoreEntryID, ECMsgStore *lpMsgStore, WSTableOutGoingQueue **lppTableOutGoingQueueOps)
{
	HRESULT hr = hrSuccess;
	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	// lpStoreEntryID == null for master queue
	if(lpStoreEntryID) {
		hr = UnWrapServerClientStoreEntry(cbStoreEntryID, lpStoreEntryID, &cbUnWrapStoreID, &lpUnWrapStoreID);
		if(hr != hrSuccess)
			goto exit;
	}

	hr = WSTableOutGoingQueue::Create(m_lpCmd, m_hDataLock, m_ecSessionId, cbUnWrapStoreID, lpUnWrapStoreID, lpMsgStore, this, lppTableOutGoingQueueOps);

exit:
	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT WSTransport::HrDeleteObjects(ULONG ulFlags, LPENTRYLIST lpMsgList, ULONG ulSyncId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct entryList sEntryList;

	LockSoap();
	memset(&sEntryList, 0, sizeof(struct entryList));

	if(lpMsgList->cValues == 0)
		goto exit;

	hr = CopyMAPIEntryListToSOAPEntryList(lpMsgList, &sEntryList);
	if(hr != hrSuccess)
		goto exit;


	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__deleteObjects(m_ecSessionId, ulFlags, &sEntryList, ulSyncId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL


exit:
	UnLockSoap();

	FreeEntryList(&sEntryList, false);

	return hr;
}

HRESULT WSTransport::HrNotify(LPNOTIFICATION lpNotification)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	struct notification	sNotification; 
	int ulSize = 0;

	memset(&sNotification, 0, sizeof(struct notification));

	LockSoap();

	//FIMXE: also notify other types ?
	if(lpNotification == NULL || lpNotification->ulEventType != fnevNewMail)
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	sNotification.ulConnection = 0;// The connection id should be calculate on the server side

	sNotification.ulEventType = lpNotification->ulEventType;
	sNotification.newmail = new notificationNewMail;

	hr = CopyMAPIEntryIdToSOAPEntryId(lpNotification->info.newmail.cbEntryID, (LPENTRYID)lpNotification->info.newmail.lpEntryID, &sNotification.newmail->pEntryId);
	if(hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(lpNotification->info.newmail.cbParentID, (LPENTRYID)lpNotification->info.newmail.lpParentID, &sNotification.newmail->pParentId);
	if(hr != hrSuccess)
		goto exit;
	
	if(lpNotification->info.newmail.lpszMessageClass){
		ulSize = strlen(lpNotification->info.newmail.lpszMessageClass)+1;
		sNotification.newmail->lpszMessageClass = new char[ulSize];
		memcpy(sNotification.newmail->lpszMessageClass , lpNotification->info.newmail.lpszMessageClass, ulSize);
	}
	sNotification.newmail->ulMessageFlags	= lpNotification->info.newmail.ulMessageFlags;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__notify(m_ecSessionId, sNotification, &er)) {
			er = ZARAFA_E_NETWORK_ERROR;
		}
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	FreeNotificationStruct(&sNotification, false);

	return hr;
}

HRESULT WSTransport::HrSubscribe(ULONG cbKey, LPBYTE lpKey, ULONG ulConnection, ULONG ulEventMask)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;
	notifySubscribe notSubscribe;

	LockSoap();

	notSubscribe.ulConnection = ulConnection;
	notSubscribe.sKey.__size = cbKey;
	notSubscribe.sKey.__ptr = lpKey;
	notSubscribe.ulEventMask = ulEventMask;			

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__notifySubscribe(m_ecSessionId, &notSubscribe, &er))
			er = ZARAFA_E_NETWORK_ERROR;

	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrUnSubscribe(ULONG ulConnection)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;

	LockSoap();

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__notifyUnSubscribe(m_ecSessionId, ulConnection, &er) )
			er = ZARAFA_E_NETWORK_ERROR;

	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetIDsFromNames(LPMAPINAMEID *lppPropNames, ULONG cNames, ULONG ulFlags, ULONG **lpServerIDs)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct namedPropArray sNamedProps;
	struct getIDsFromNamesResponse sResponse;
	unsigned int i=0;

	LockSoap();

	// Convert our data into a structure that the server can take
	sNamedProps.__size = cNames;
	ECAllocateBuffer(sizeof(struct namedProp) * cNames, (void **)&sNamedProps.__ptr);
	memset(sNamedProps.__ptr, 0 , sizeof(struct namedProp) * cNames);

	for(i=0;i<cNames;i++) {	
		switch(lppPropNames[i]->ulKind) {
		case MNID_ID:
			ECAllocateMore(sizeof(unsigned int), sNamedProps.__ptr,(void **)&sNamedProps.__ptr[i].lpId);
			*sNamedProps.__ptr[i].lpId = lppPropNames[i]->Kind.lID;
			break;
		case MNID_STRING: {
			// The string is actually utf-8, not windows-1252. This enables full support for wide char strings.
			ECIConv ic("utf-8", "UCS-2LE");
			std::string strNameW((char*)lppPropNames[i]->Kind.lpwstrName, unicodelen(lppPropNames[i]->Kind.lpwstrName)*sizeof(WCHAR));
			std::string strNameUTF8 = ic.convert(strNameW);

			ECAllocateMore(strNameUTF8.length()+1, sNamedProps.__ptr,(void **)&sNamedProps.__ptr[i].lpString);
			strcpy(sNamedProps.__ptr[i].lpString, strNameUTF8.c_str());
			break;
		}
		default:
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		if(lppPropNames[i]->lpguid) {
			ECAllocateMore(sizeof( xsd__base64Binary) , sNamedProps.__ptr, (void **) &sNamedProps.__ptr[i].lpguid);
			sNamedProps.__ptr[i].lpguid->__ptr = (unsigned char *)lppPropNames[i]->lpguid;
			sNamedProps.__ptr[i].lpguid->__size = sizeof(GUID);
		} else {
			sNamedProps.__ptr[i].lpguid = NULL;
		}
	}

	// Send the call off the the server
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getIDsFromNames(m_ecSessionId, &sNamedProps, ulFlags, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	// Make sure we response with the same amount of data that we requested
	if((ULONG)sResponse.lpsPropTags.__size != cNames) {
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	ECAllocateBuffer(sizeof(ULONG) * sResponse.lpsPropTags.__size, (void**)lpServerIDs);

	memcpy(*lpServerIDs, sResponse.lpsPropTags.__ptr, sizeof(ULONG) * sResponse.lpsPropTags.__size);

exit:
	UnLockSoap();

	if(sNamedProps.__ptr)
		ECFreeBuffer(sNamedProps.__ptr);

	return hr;
}

HRESULT WSTransport::HrGetNamesFromIDs(LPSPropTagArray lpsPropTags, LPMAPINAMEID **lpppNames, ULONG *lpcResolved)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	struct getNamesFromIDsResponse sResponse;
	struct propTagArray sPropTags;
	unsigned int i=0;
	LPMAPINAMEID *lppNames = NULL;

	sPropTags.__size = lpsPropTags->cValues;
	sPropTags.__ptr = (unsigned int *)&lpsPropTags->aulPropTag[0];

	LockSoap();

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getNamesFromIDs(m_ecSessionId, &sPropTags, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	ECAllocateBuffer(sizeof(LPMAPINAMEID) * sResponse.lpsNames.__size, (void **) &lppNames);

	// Loop through all the returned names, and put it into the return value
	for(i=0;i<sResponse.lpsNames.__size;i++) {
		// Each MAPINAMEID must be allocated
		ECAllocateMore(sizeof(MAPINAMEID), lppNames, (void **) &lppNames[i]);

		if(sResponse.lpsNames.__ptr[i].lpguid && sResponse.lpsNames.__ptr[i].lpguid->__ptr) {
			ECAllocateMore(sizeof(GUID), lppNames, (void **) &lppNames[i]->lpguid);
			memcpy(lppNames[i]->lpguid, sResponse.lpsNames.__ptr[i].lpguid->__ptr, sizeof(GUID));
		}
		if(sResponse.lpsNames.__ptr[i].lpId) {
			lppNames[i]->Kind.lID = *sResponse.lpsNames.__ptr[i].lpId;
			lppNames[i]->ulKind = MNID_ID;
		} else if(sResponse.lpsNames.__ptr[i].lpString) {
			ECIConv ic("UCS-2LE", "UTF-8");
			std::string strNameW = ic.convert(sResponse.lpsNames.__ptr[i].lpString);

			ECAllocateMore(strNameW.size() + sizeof(WCHAR), lppNames, (void **)&lppNames[i]->Kind.lpwstrName);
			memcpy(lppNames[i]->Kind.lpwstrName, (WCHAR *)strNameW.c_str(), strNameW.size());
			lppNames[i]->Kind.lpwstrName[strNameW.size()/sizeof(WCHAR)] = 0;
			lppNames[i]->ulKind = MNID_STRING;
		} else {
			// not found by server, we have actually allocated memory but it doesn't really matter
			lppNames[i] = NULL;
		}
	}

	*lpcResolved = sResponse.lpsNames.__size;
	*lpppNames = lppNames;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetReceiveFolderTable(ULONG cbStoreEntryID, LPENTRYID lpStoreEntryID, LPSRowSet* lppsRowSet)
{
	struct receiveFolderTableResponse sReceiveFolders;
	ECRESULT	er = erSuccess;
	HRESULT		hr = hrSuccess;
	LPSRowSet	lpsRowSet = NULL;
	ULONG		ulRowId = 0;
	unsigned int i = 0;
	int			nLen = 0;
	entryId		sEntryId = {0}; // Do not free

	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	LockSoap();

	hr = UnWrapServerClientStoreEntry(cbStoreEntryID, lpStoreEntryID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	sEntryId.__ptr = (unsigned char*)lpUnWrapStoreID;
	sEntryId.__size = cbUnWrapStoreID;



	// Get ReceiveFolder information from the server
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getReceiveFolderTable(m_ecSessionId, sEntryId, &sReceiveFolders))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sReceiveFolders.er;

	}
	END_SOAP_CALL

	ECAllocateBuffer(CbNewSRowSet(sReceiveFolders.sFolderArray.__size), (void**)&lpsRowSet);
	memset(lpsRowSet, 0, CbNewSRowSet(sReceiveFolders.sFolderArray.__size));
	lpsRowSet->cRows = sReceiveFolders.sFolderArray.__size;

	for(i=0; i < sReceiveFolders.sFolderArray.__size; i++)
	{
		ulRowId = i+1;

		lpsRowSet->aRow[i].cValues = NUM_RFT_PROPS;
		ECAllocateBuffer(sizeof(SPropValue) * NUM_RFT_PROPS, (void**)&lpsRowSet->aRow[i].lpProps);
		memset(lpsRowSet->aRow[i].lpProps, 0, sizeof(SPropValue)*NUM_RFT_PROPS);
		
		lpsRowSet->aRow[i].lpProps[RFT_ROWID].ulPropTag = PR_ROWID;
		lpsRowSet->aRow[i].lpProps[RFT_ROWID].Value.ul = ulRowId;

		lpsRowSet->aRow[i].lpProps[RFT_INST_KEY].ulPropTag = PR_INSTANCE_KEY;
		lpsRowSet->aRow[i].lpProps[RFT_INST_KEY].Value.bin.cb = 4; //fixme: maybe fix, normal 8 now
		ECAllocateMore(lpsRowSet->aRow[i].lpProps[RFT_INST_KEY].Value.bin.cb, lpsRowSet->aRow[i].lpProps, (void**)&lpsRowSet->aRow[i].lpProps[RFT_INST_KEY].Value.bin.lpb);
		memset(lpsRowSet->aRow[i].lpProps[RFT_INST_KEY].Value.bin.lpb, 0, lpsRowSet->aRow[i].lpProps[RFT_INST_KEY].Value.bin.cb);
		memcpy(lpsRowSet->aRow[i].lpProps[RFT_INST_KEY].Value.bin.lpb, &ulRowId, sizeof(ulRowId));
		
		lpsRowSet->aRow[i].lpProps[RFT_ENTRYID].ulPropTag = PR_ENTRYID;
		lpsRowSet->aRow[i].lpProps[RFT_ENTRYID].Value.bin.cb = sReceiveFolders.sFolderArray.__ptr[i].sEntryId.__size;
		ECAllocateMore(lpsRowSet->aRow[i].lpProps[RFT_ENTRYID].Value.bin.cb, lpsRowSet->aRow[i].lpProps, (void**)&lpsRowSet->aRow[i].lpProps[RFT_ENTRYID].Value.bin.lpb);
		memcpy(lpsRowSet->aRow[i].lpProps[RFT_ENTRYID].Value.bin.lpb, sReceiveFolders.sFolderArray.__ptr[i].sEntryId.__ptr, lpsRowSet->aRow[i].lpProps[RFT_ENTRYID].Value.bin.cb);

		// Use the entryid for record key
		lpsRowSet->aRow[i].lpProps[RFT_RECORD_KEY].ulPropTag = PR_RECORD_KEY;
		lpsRowSet->aRow[i].lpProps[RFT_RECORD_KEY].Value.bin.cb = sReceiveFolders.sFolderArray.__ptr[i].sEntryId.__size;
		ECAllocateMore(lpsRowSet->aRow[i].lpProps[RFT_RECORD_KEY].Value.bin.cb, lpsRowSet->aRow[i].lpProps, (void**)&lpsRowSet->aRow[i].lpProps[RFT_RECORD_KEY].Value.bin.lpb);
		memcpy(lpsRowSet->aRow[i].lpProps[RFT_RECORD_KEY].Value.bin.lpb, sReceiveFolders.sFolderArray.__ptr[i].sEntryId.__ptr, lpsRowSet->aRow[i].lpProps[RFT_RECORD_KEY].Value.bin.cb);

		lpsRowSet->aRow[i].lpProps[RFT_MSG_CLASS].ulPropTag = PR_MESSAGE_CLASS;
		nLen = strlen(sReceiveFolders.sFolderArray.__ptr[i].lpszAExplicitClass)+1;
		ECAllocateMore(nLen, lpsRowSet->aRow[i].lpProps, (void**)&lpsRowSet->aRow[i].lpProps[RFT_MSG_CLASS].Value.lpszA);
		memcpy(lpsRowSet->aRow[i].lpProps[RFT_MSG_CLASS].Value.lpszA, sReceiveFolders.sFolderArray.__ptr[i].lpszAExplicitClass, nLen);
	
	}

	*lppsRowSet = lpsRowSet;

exit:
	UnLockSoap();
	
	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT WSTransport::HrGetReceiveFolder(ULONG cbStoreEntryID, LPENTRYID lpStoreEntryID, LPTSTR lpszMessageClass, ULONG* lpcbEntryID, LPENTRYID* lppEntryID, LPTSTR* lppszExplicitClass)
{
	struct receiveFolderResponse sReceiveFolderTable;

	int			nLen = 0;
	ECRESULT	er = erSuccess;
	HRESULT		hr = hrSuccess;
	LPTSTR		lpszExplicitClass = NULL;
	entryId		sEntryId = {0}; // Do not free
	ULONG		cbEntryID = 0;
	LPENTRYID	lpEntryID = NULL;

	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	LockSoap();

	hr = UnWrapServerClientStoreEntry(cbStoreEntryID, lpStoreEntryID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	sEntryId.__ptr = (unsigned char*)lpUnWrapStoreID;
	sEntryId.__size = cbUnWrapStoreID;

	if(lppszExplicitClass)
		*lppszExplicitClass = NULL;


	// Get ReceiveFolder information from the server
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getReceiveFolder(m_ecSessionId, sEntryId, lpszMessageClass, &sReceiveFolderTable))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sReceiveFolderTable.er;
	}
	END_SOAP_CALL

	if(er == ZARAFA_E_NOT_FOUND && lppszExplicitClass)
	{
		// This is only by an empty message store ??
		hr = ECAllocateBuffer(1, (void **) &lpszExplicitClass);
		if(hr != hrSuccess)
			goto exit;

		*lpszExplicitClass = 0;

		*lppszExplicitClass = lpszExplicitClass;
		*lpcbEntryID = 0;
		*lppEntryID = NULL;

		hr = hrSuccess;
		goto exit;
	}

	if(hr != hrSuccess)
		goto exit;
	
	hr = CopySOAPEntryIdToMAPIEntryId(&sReceiveFolderTable.sReceiveFolder.sEntryId, &cbEntryID, &lpEntryID, NULL);
	if(hr != hrSuccess)
		goto exit;

	if(er != ZARAFA_E_NOT_FOUND && lppszExplicitClass != NULL) {

		nLen = strlen(sReceiveFolderTable.sReceiveFolder.lpszAExplicitClass)+1;

		hr = ECAllocateBuffer(nLen, (void **) &lpszExplicitClass);
		if(hr != hrSuccess)
			goto exit;

		memcpy(lpszExplicitClass, sReceiveFolderTable.sReceiveFolder.lpszAExplicitClass, nLen);

		*lppszExplicitClass = lpszExplicitClass;
	}

	*lppEntryID = lpEntryID;
	*lpcbEntryID = cbEntryID;

exit:
	if(hr != hrSuccess){
		if(lpEntryID)
			ECFreeBuffer(lpEntryID);

		if(lpszExplicitClass)
			ECFreeBuffer(lpszExplicitClass);
	}

	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrSetReceiveFolder(ULONG cbStoreID, LPENTRYID lpStoreID, LPTSTR lpszMessageClass, ULONG cbEntryID, LPENTRYID lpEntryID)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	unsigned int result;
	entryId		sStoreId = {0}; // Do not free
	entryId		sEntryId = {0}; // Do not free
	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	LockSoap();

	hr = UnWrapServerClientStoreEntry(cbStoreID, lpStoreID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	sStoreId.__ptr = (unsigned char*)lpUnWrapStoreID;
	sStoreId.__size = cbUnWrapStoreID;


	// Ignore error
	CopyMAPIEntryIdToSOAPEntryId(cbEntryID, lpEntryID, &sEntryId, true);


	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__setReceiveFolder(m_ecSessionId, sStoreId, (lpEntryID)?&sEntryId : NULL, lpszMessageClass, &result))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = result;
	}
	END_SOAP_CALL

exit:
    UnLockSoap();
    
	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT WSTransport::HrSetReadFlag(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulFlags, ULONG ulSyncId)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;

	struct entryList sEntryList;
	entryId			 sEntryId;

	sEntryId.__ptr = (unsigned char*)lpEntryID;
	sEntryId.__size = cbEntryID;

	sEntryList.__size = 1;
	sEntryList.__ptr = &sEntryId;

	LockSoap();

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__setReadFlags(m_ecSessionId, ulFlags, NULL, &sEntryList, ulSyncId, &er))
			er = ZARAFA_E_NETWORK_ERROR;

	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;

}

HRESULT WSTransport::HrSubmitMessage(ULONG cbMessageID, LPENTRYID lpMessageID, ULONG ulFlags)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;
	entryId		sEntryId = {0}; // Do not free

	LockSoap();

	hr = CopyMAPIEntryIdToSOAPEntryId(cbMessageID, lpMessageID, &sEntryId, true);
	if(hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__submitMessage(m_ecSessionId, sEntryId, ulFlags, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrFinishedMessage(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	entryId		sEntryId = {0}; // Do not free

	LockSoap();

	hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryID, lpEntryID, &sEntryId, true);
	if(hr != hrSuccess)
		goto exit;
	
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__finishedMessage(m_ecSessionId, sEntryId, ulFlags, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrAbortSubmit(ULONG cbEntryID, LPENTRYID lpEntryID)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	entryId		sEntryId = {0}; // Do not free

	LockSoap();

	hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryID, lpEntryID, &sEntryId, true);
	if(hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__abortSubmit(m_ecSessionId, sEntryId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrIsMessageInQueue(ULONG cbEntryID, LPENTRYID lpEntryID)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	entryId		sEntryId = {0}; // Do not free

	LockSoap();

	hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryID, lpEntryID, &sEntryId, true);
	if(hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__isMessageInQueue(m_ecSessionId, sEntryId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrResolveStore(ULONG ulStoreId, ULONG *lpulUserID, ULONG* lpcbStoreID, LPENTRYID* lppStoreID)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	struct resolveUserStoreResponse sResponse;

	LockSoap();

	if (!ulStoreId){
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__resolveStore(m_ecSessionId, ulStoreId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	if(lpulUserID)
		*lpulUserID = sResponse.ulUserId;

	if(lpcbStoreID && lppStoreID) {
		// Create a client store entry, add the servername
		hr = WrapServerClientStoreEntry(sResponse.lpszServerPath ? sResponse.lpszServerPath : m_sProfileProps.strServerPath.c_str(), &sResponse.sStoreId, lpcbStoreID, lppStoreID);
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrResolveUserStore(char *szUserName, ULONG *lpulUserID, ULONG* lpcbStoreID, LPENTRYID* lppStoreID, std::string *lpstrRedirServer)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;
	struct resolveUserStoreResponse sResponse;

	LockSoap();

	if(szUserName == NULL){
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}


	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__resolveUserStore(m_ecSessionId, szUserName, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	//END_SOAP_CALL
	if(er == ZARAFA_E_END_OF_SESSION) { if(HrReLogon() == hrSuccess) goto retry; }
	hr = ZarafaErrorToMAPIError(er, MAPI_E_NOT_FOUND);
	if (hr == MAPI_E_UNABLE_TO_COMPLETE)
	{
		if (lpstrRedirServer)
			*lpstrRedirServer = sResponse.lpszServerPath;
		else
			hr = MAPI_E_NOT_FOUND;
	}
	if(hr != hrSuccess)
		goto exit;

	if(lpulUserID) {
		*lpulUserID = sResponse.ulUserId;
	}

	if(lpcbStoreID && lppStoreID) {

		// Create a client store entry, add the servername
		hr = WrapServerClientStoreEntry(sResponse.lpszServerPath ? sResponse.lpszServerPath : m_sProfileProps.strServerPath.c_str(), &sResponse.sStoreId, lpcbStoreID, lppStoreID);
		if(hr != hrSuccess)
			goto exit;

	}

exit:
	UnLockSoap();

	return hr;
	
}

// FIXME naming
HRESULT WSTransport::HrGetUserProfile(char **lppProfileName, char **lppProfileAddress, ULONG* lpcbUserId, LPENTRYID* lppUserId)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;
	ULONG		ulSize = 0;
	PABEID		lpUserEID = NULL;
	entryId		sUserId = {0};

	struct getUserResponse sResponse;

	LockSoap();

	if(lppProfileName == NULL || lppProfileAddress == NULL || lpcbUserId == NULL  || lppUserId == NULL ) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getUser(m_ecSessionId, 0, sUserId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL
	
	ulSize = strlen(sResponse.lpsUser->lpszFullName)+1;
	ECAllocateBuffer(ulSize, (void**)lppProfileName);
	memcpy(*lppProfileName, sResponse.lpsUser->lpszFullName, ulSize);

	ulSize = strlen(sResponse.lpsUser->lpszMailAddress)+1;
	ECAllocateBuffer(ulSize, (void**)lppProfileAddress);
	memcpy(*lppProfileAddress, sResponse.lpsUser->lpszMailAddress, ulSize);
	
	// User entryid
	if (sResponse.lpsUser->sUserId.__size  < CbNewABEID("") || sResponse.lpsUser->sUserId.__ptr == NULL)
	{
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	hr = ECAllocateBuffer(sResponse.lpsUser->sUserId.__size, (void**)&lpUserEID);
	if (hr != hrSuccess)
		goto exit;

	memcpy(lpUserEID, sResponse.lpsUser->sUserId.__ptr, sResponse.lpsUser->sUserId.__size);

	*lpcbUserId = sResponse.lpsUser->sUserId.__size;
	*lppUserId = (LPENTRYID)lpUserEID;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrCreateUser(LPECUSER lpECUser, ULONG *lpcbUserId, LPENTRYID *lppUserId)
{
	HRESULT	hr = hrSuccess;
	ECRESULT er = erSuccess;
	struct user	sUser = {0};
	struct setUserResponse sResponse;

	LockSoap();

	if(lpECUser == NULL || lpcbUserId == NULL || lppUserId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	sUser.lpszUsername		= lpECUser->lpszUsername;
	sUser.lpszPassword		= lpECUser->lpszPassword;
	sUser.lpszMailAddress		= lpECUser->lpszFullEmail;
	sUser.ulUserId			= 0;
	sUser.ulIsNonActive		= lpECUser->ulIsNonActive;
	sUser.ulIsAdmin			= lpECUser->ulIsAdmin;
	sUser.lpszFullName		= lpECUser->lpszFullName;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__createUser(m_ecSessionId, &sUser, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sUserId, sResponse.ulUserId, lpcbUserId, lppUserId);

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetUser(ULONG cbUserID, LPENTRYID lpUserID, LPECUSER *lppECUser)
{
	HRESULT	hr = hrSuccess;
	ECRESULT er = erSuccess;
	struct getUserResponse	sResponse;
	LPECUSER lpECUser = NULL;
	entryId	sUserId = {0};
	
	LockSoap();

	if(lpUserID == NULL || lppECUser == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserID, lpUserID, &sUserId, true);
	if(hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getUser(m_ecSessionId, ABEID_ID(lpUserID), sUserId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapUserToUser(sResponse.lpsUser, &lpECUser);
	if(hr != hrSuccess)
		goto exit;

	*lppECUser = lpECUser;
	lpECUser = NULL;

exit:
	UnLockSoap();

	if (lpECUser != NULL)
		ECFreeBuffer(lpECUser);

	return hr;
}

HRESULT WSTransport::HrSetUser(LPECUSER lpECUser)
{
	HRESULT	hr = hrSuccess;
	ECRESULT er = erSuccess;
	struct user	sUser = {0};
	unsigned int result = 0;

	LockSoap();

	if(lpECUser == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	sUser.lpszUsername		= lpECUser->lpszUsername;
	sUser.lpszPassword		= lpECUser->lpszPassword;
	sUser.lpszMailAddress	= lpECUser->lpszFullEmail;
	sUser.ulUserId			= ABEID_ID(lpECUser->sUserId.lpb);
	sUser.ulIsNonActive		= lpECUser->ulIsNonActive;
	sUser.ulIsAdmin			= lpECUser->ulIsAdmin;
	sUser.lpszFullName		= lpECUser->lpszFullName;
	sUser.sUserId.__ptr		= lpECUser->sUserId.lpb;
	sUser.sUserId.__size	= lpECUser->sUserId.cb;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__setUser(m_ecSessionId, &sUser, &result))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = result;

	}
	END_SOAP_CALL


exit:
	UnLockSoap();

	return hr;
}

//TODO: change userId into entryid
HRESULT WSTransport::HrCreateStore(ULONG ulStoreType, ULONG cbUserID, LPENTRYID lpUserID, ULONG cbStoreID, LPENTRYID lpStoreID, ULONG cbRootID, LPENTRYID lpRootID)
{
	HRESULT hr = hrSuccess;
	ECRESULT er = erSuccess;

	entryId sUserId = {0}; // do not free
	entryId	sStoreId = {0}; // do not free
	entryId	sRootId = {0}; // do not free

	LockSoap();

	if(lpUserID == NULL || lpStoreID == NULL || lpRootID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserID, lpUserID, &sUserId, true);
	if(hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbStoreID, lpStoreID, &sStoreId, true);
	if(hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbRootID, lpRootID, &sRootId, true);
	if(hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__createStore(m_ecSessionId, ulStoreType, ABEID_ID(lpUserID), sUserId, sStoreId, sRootId, &er))
			er = ZARAFA_E_NETWORK_ERROR;

	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrHookStore(ULONG cbUserId, LPENTRYID lpUserId, LPGUID lpGuid, ULONG ulSyncId)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;
	entryId		sUserId = {0};
	struct xsd__base64Binary sStoreGuid = {0,0};

	if (cbUserId == 0 || lpUserId == NULL || lpGuid == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if(hr != hrSuccess)
		goto exit;

	sStoreGuid.__ptr = (unsigned char*)lpGuid;
	sStoreGuid.__size = sizeof(GUID);
	
	LockSoap();

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__hookStore(m_ecSessionId, sUserId, sStoreGuid, ulSyncId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrRemoveStore(LPGUID lpGuid, ULONG ulSyncId)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;
	struct xsd__base64Binary sStoreGuid = {0,0};

	if (lpGuid == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	sStoreGuid.__ptr = (unsigned char*)lpGuid;
	sStoreGuid.__size = sizeof(GUID);
	
	LockSoap();

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__removeStore(m_ecSessionId, sStoreGuid, ulSyncId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetStore(ULONG cbUserId, LPENTRYID lpUserId, ULONG* lpulStoreId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;

	LockSoap();

	struct getStoreByUserResponse sResponse;
	entryId sUserId = {0};

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(m_lpCmd->ns__getStoreByUser(m_ecSessionId, ABEID_ID(lpUserId), sUserId, &sResponse) != SOAP_OK)
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	*lpulStoreId = sResponse.ulStoreId;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrDeleteUser(ULONG cbUserId, LPENTRYID lpUserId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sUserId = {0};

	LockSoap();
	
	if(cbUserId < CbNewABEID("") || lpUserId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__deleteUser(m_ecSessionId, ABEID_ID(lpUserId), sUserId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetUserList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcUsers, LPECUSER* lppsUsers)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sCompanyId = {0};

	struct userListResponse sResponse;

	LockSoap();

	if(lpcUsers == NULL || lppsUsers == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if (cbCompanyId > 0 && lpCompanyId != NULL)
	{
		hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
		if (hr != hrSuccess)
			goto exit;
	}


	*lpcUsers = 0;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getUserList(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapUserArrayToUserArray(&sResponse.sUserArray, lpcUsers, lppsUsers);
	if(hr != hrSuccess)
		goto exit;

exit:	
	UnLockSoap();

	
	return hr;
}

/////////////////////////////////////////////////////////////////////
// IECServiceAdmin group functions
HRESULT WSTransport::HrCreateGroup(LPECGROUP lpECGroup, ULONG *lpcbGroupId, LPENTRYID *lppGroupId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct group sGroup = {0};
	struct setGroupResponse sResponse;

	LockSoap();

	if(lpECGroup == NULL || lpcbGroupId == NULL || lppGroupId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	sGroup.ulGroupId = 0;
	sGroup.lpszGroupname = lpECGroup->lpszGroupname;
	sGroup.lpszFullname = lpECGroup->lpszFullname;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__createGroup(m_ecSessionId, &sGroup, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sGroupId, sResponse.ulGroupId, lpcbGroupId, lppGroupId);

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrSetGroup(LPECGROUP lpECGroup)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;

	struct group	sGroup = {0};

	LockSoap();

	if(lpECGroup == NULL || lpECGroup->lpszGroupname == NULL || lpECGroup->lpszFullname == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	sGroup.lpszFullname = lpECGroup->lpszFullname;
	sGroup.lpszGroupname = lpECGroup->lpszGroupname;
	sGroup.sGroupId.__size = lpECGroup->sGroupId.cb;
	sGroup.sGroupId.__ptr = lpECGroup->sGroupId.lpb;
	sGroup.ulGroupId = ABEID_ID(lpECGroup->sGroupId.lpb);

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__setGroup(m_ecSessionId, &sGroup, &er))
			er = ZARAFA_E_NETWORK_ERROR;

	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetGroup(ULONG cbGroupID, LPENTRYID lpGroupID, LPECGROUP *lppECGroup)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	LPECGROUP lpGroup = NULL;
	entryId sGroupId = {0};

	struct getGroupResponse sResponse;
	
	LockSoap();

	if (lpGroupID == NULL || lppECGroup == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbGroupID, lpGroupID, &sGroupId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getGroup(m_ecSessionId, ABEID_ID(lpGroupID), sGroupId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapGroupToGroup(sResponse.lpsGroup, &lpGroup);
	if (hr != hrSuccess)
		goto exit;

	*lppECGroup = lpGroup;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrDeleteGroup(ULONG cbGroupId, LPENTRYID lpGroupId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sGroupId = {0};

	LockSoap();
	
	if(cbGroupId < CbNewABEID("") || lpGroupId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbGroupId, lpGroupId, &sGroupId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__groupDelete(m_ecSessionId, ABEID_ID(lpGroupId), sGroupId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetSendAsList(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcSenders, LPECUSER *lppSenders)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct userListResponse sResponse;
	entryId sUserId = {0};

	LockSoap();
	
	if(cbUserId < CbNewABEID("") || lpUserId == NULL || lpcSenders == NULL || lppSenders == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getSendAsList(m_ecSessionId, ABEID_ID(lpUserId), sUserId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapUserArrayToUserArray(&sResponse.sUserArray, lpcSenders, lppSenders);
	if(hr != hrSuccess)
		goto exit;

exit:	
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrAddSendAsUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbSenderId, LPENTRYID lpSenderId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sUserId = {0};
	entryId sSenderId = {0};

	LockSoap();

	if (cbUserId < CbNewABEID("") || lpUserId == NULL || cbSenderId < CbNewABEID("") || lpSenderId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbSenderId, lpSenderId, &sSenderId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__addSendAsUser(m_ecSessionId, ABEID_ID(lpUserId), sUserId, ABEID_ID(lpSenderId), sSenderId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	
	}
	END_SOAP_CALL

exit:	
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrDelSendAsUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbSenderId, LPENTRYID lpSenderId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sUserId = {0};
	entryId sSenderId = {0};

	LockSoap();

	if (cbUserId < CbNewABEID("") || lpUserId == NULL || cbSenderId < CbNewABEID("") || lpSenderId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbSenderId, lpSenderId, &sSenderId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__delSendAsUser(m_ecSessionId, ABEID_ID(lpUserId), sUserId, ABEID_ID(lpSenderId), sSenderId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:	
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrResolveUserName(char *lpszUserName, ULONG *lpcbUserId, LPENTRYID *lppUserId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct resolveUserResponse sResponse;

	LockSoap();

	if(lpszUserName == NULL || lpcbUserId == NULL || lppUserId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	//Resolve userid from username
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__resolveUsername(m_ecSessionId, lpszUserName, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	
	}
	END_SOAP_CALL

	hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sUserId, sResponse.ulUserId, lpcbUserId, lppUserId);

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrResolveGroupName(char *lpszGroupName, ULONG *lpcbGroupId, LPENTRYID *lppGroupId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct resolveGroupResponse sResponse;

	LockSoap();

	if(lpszGroupName == NULL || lpcbGroupId == NULL || lppGroupId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	//Resolve groupid from groupname
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__resolveGroupname(m_ecSessionId, lpszGroupName, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	
	}
	END_SOAP_CALL

	hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sGroupId, sResponse.ulGroupId, lpcbGroupId, lppGroupId);

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetGroupList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcGroups, LPECGROUP *lppsGroups)
{
	ECRESULT	er = erSuccess;
	HRESULT		hr = hrSuccess;

 	struct groupListResponse sResponse;
	entryId					sCompanyId = {0};

	LockSoap();

	if(lpcGroups == NULL || lppsGroups == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;
	
	*lpcGroups = 0;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__getGroupList(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapGroupArrayToGroupArray(&sResponse.sGroupArray, lpcGroups, lppsGroups);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrDeleteGroupUser(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG cbUserId, LPENTRYID lpUserId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sGroupId = {0};
	entryId sUserId = {0};

	LockSoap();

	hr = CopyMAPIEntryIdToSOAPEntryId(cbGroupId, lpGroupId, &sGroupId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	// Remove group
	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__deleteGroupUser(m_ecSessionId, ABEID_ID(lpGroupId), sGroupId, ABEID_ID(lpUserId), sUserId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrAddGroupUser(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG cbUserId, LPENTRYID lpUserId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;

	entryId sGroupId = {0};
	entryId sUserId = {0};

	LockSoap();

	hr = CopyMAPIEntryIdToSOAPEntryId(cbGroupId, lpGroupId, &sGroupId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	// Remove group
	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__addGroupUser(m_ecSessionId, ABEID_ID(lpGroupId), sGroupId, ABEID_ID(lpUserId), sUserId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetUserListOfGroup(ULONG cbGroupId, LPENTRYID lpGroupId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct userListResponse sResponse;
	entryId sGroupId = {0};

	LockSoap();

	if(lpGroupId == NULL || lpcUsers == NULL || lppsUsers == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbGroupId, lpGroupId, &sGroupId, true);
	if (hr != hrSuccess)
		goto exit;

	// Get an userlist of a group
	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__getUserListOfGroup(m_ecSessionId, ABEID_ID(lpGroupId), sGroupId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapUserArrayToUserArray(&sResponse.sUserArray, lpcUsers, lppsUsers);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetGroupListOfUser(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcGroup, LPECGROUP *lppsGroups)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct groupListResponse sResponse;
	entryId sUserId = {0};

	LockSoap();

	if(lpcGroup == NULL || lpUserId == NULL || lppsGroups == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	// Get a grouplist of an user
	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__getGroupListOfUser(m_ecSessionId, ABEID_ID(lpUserId), sUserId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapGroupArrayToGroupArray(&sResponse.sGroupArray, lpcGroup, lppsGroups);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrCreateCompany(LPECCOMPANY lpECCompany, ULONG *lpcbCompanyId, LPENTRYID *lppCompanyId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct company sCompany = {0};
	struct setCompanyResponse sResponse;

	LockSoap();

	if(lpECCompany == NULL || lpcbCompanyId == NULL || lppCompanyId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	sCompany.ulAdministrator = 0;
	sCompany.lpszCompanyname = lpECCompany->lpszCompanyname;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__createCompany(m_ecSessionId, &sCompany, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sCompanyId, sResponse.ulCompanyId, MAPI_ABCONT, lpcbCompanyId, lppCompanyId);

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrDeleteCompany(ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sCompanyId = {0};

	LockSoap();
	
	if(cbCompanyId < CbNewABEID("") || lpCompanyId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__deleteCompany(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrSetCompany(LPECCOMPANY lpECCompany)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct company sCompany = {0};

	LockSoap();

	if(lpECCompany == NULL || lpECCompany->lpszCompanyname == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	sCompany.lpszCompanyname = lpECCompany->lpszCompanyname;

	sCompany.ulCompanyId = ABEID_ID(lpECCompany->sCompanyId.lpb);
	sCompany.sCompanyId.__size = lpECCompany->sCompanyId.cb;
	sCompany.sCompanyId.__ptr = lpECCompany->sCompanyId.lpb;

	sCompany.ulAdministrator = ABEID_ID(lpECCompany->sAdministrator.lpb);
	sCompany.sAdministrator.__size = lpECCompany->sAdministrator.cb;
	sCompany.sAdministrator.__ptr = lpECCompany->sAdministrator.lpb;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__setCompany(m_ecSessionId, &sCompany, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetCompany(ULONG cbCompanyId, LPENTRYID lpCompanyId, LPECCOMPANY *lppECCompany)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	LPECCOMPANY lpCompany = NULL;
	struct getCompanyResponse sResponse;
	entryId sCompanyId = {0};

	LockSoap();

	if (lpCompanyId == NULL || lppECCompany == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__getCompany(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapCompanyToCompany(sResponse.lpsCompany, &lpCompany);
	if (hr != hrSuccess)
		goto exit;
	
	*lppECCompany = lpCompany;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrResolveCompanyName(char *lpszCompanyName, ULONG *lpcbCompanyId, LPENTRYID *lppCompanyId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct resolveCompanyResponse sResponse;

	LockSoap();

	if(lpszCompanyName == NULL || lpcbCompanyId == NULL || lppCompanyId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	//Resolve companyid from companyname
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__resolveCompanyname(m_ecSessionId, lpszCompanyName, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sCompanyId, sResponse.ulCompanyId, MAPI_ABCONT, lpcbCompanyId, lppCompanyId);

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetCompanyList(ULONG *lpcCompanies, LPECCOMPANY *lppsCompanies)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct companyListResponse sResponse;

	LockSoap();

	if(lpcCompanies == NULL || lppsCompanies == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	*lpcCompanies = 0;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getCompanyList(m_ecSessionId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	hr = SoapCompanyArrayToCompanyArray(&sResponse.sCompanyArray, lpcCompanies, lppsCompanies);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrAddCompanyToRemoteViewList(ULONG cbSetCompanyId, LPENTRYID lpSetCompanyId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;

	entryId sSetCompanyId = {0};
	entryId	sCompanyId = {0};

	LockSoap();

	if (lpSetCompanyId == NULL || lpCompanyId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbSetCompanyId, lpSetCompanyId, &sSetCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__addCompanyToRemoteViewList(m_ecSessionId, ABEID_ID(lpSetCompanyId), sSetCompanyId, ABEID_ID(lpCompanyId), sCompanyId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrDelCompanyFromRemoteViewList(ULONG cbSetCompanyId, LPENTRYID lpSetCompanyId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sSetCompanyId = {0};
	entryId sCompanyId = {0};

	LockSoap();

	if (lpSetCompanyId == NULL || lpCompanyId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbSetCompanyId, lpSetCompanyId, &sSetCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__delCompanyFromRemoteViewList(m_ecSessionId, ABEID_ID(lpSetCompanyId), sSetCompanyId, ABEID_ID(lpCompanyId), sCompanyId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetRemoteViewList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcCompanies, LPECCOMPANY *lppsCompanies)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct companyListResponse sResponse;
	entryId sCompanyId = {0};

	LockSoap();

	if(lpcCompanies == NULL || lpCompanyId == NULL || lppsCompanies == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	*lpcCompanies = 0;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__getRemoteViewList(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapCompanyArrayToCompanyArray(&sResponse.sCompanyArray, lpcCompanies, lppsCompanies);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrAddUserToRemoteAdminList(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId	sUserId = {0};
	entryId sCompanyId = {0};

	LockSoap();

	if (lpUserId == NULL || lpCompanyId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__addUserToRemoteAdminList(m_ecSessionId, ABEID_ID(lpUserId), sUserId, ABEID_ID(lpCompanyId), sCompanyId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrDelUserFromRemoteAdminList(ULONG cbUserId, LPENTRYID lpUserId, ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId	sUserId = {0};
	entryId sCompanyId = {0};

	LockSoap();

	if (lpUserId == NULL || lpCompanyId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__delUserFromRemoteAdminList(m_ecSessionId, ABEID_ID(lpUserId), sUserId, ABEID_ID(lpCompanyId), sCompanyId, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetRemoteAdminList(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct userListResponse sResponse;
	entryId sCompanyId = {0};

	LockSoap();

	if(lpcUsers == NULL || lpCompanyId == NULL || lppsUsers == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	*lpcUsers = 0;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__getRemoteAdminList(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapUserArrayToUserArray(&sResponse.sUserArray, lpcUsers, lppsUsers);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetPermissionRules(int ulType, ULONG cbEntryID, LPENTRYID lpEntryID, ULONG* lpcPermissions, LPECPERMISSION* lppECPermissions)
{
	ECRESULT		er = erSuccess;
	HRESULT			hr = hrSuccess;
	entryId			sEntryId = {0}; // Do not free
	LPECPERMISSION	lpECPermissions = NULL;
	
	LPENTRYID		lpUnWrapStoreID = NULL;
	ULONG			cbUnWrapStoreID = 0;

	struct rightsResponse sRightResponse;

	LockSoap();

	if(lpcPermissions == NULL || lppECPermissions == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Remove servername, always
	hr = UnWrapServerClientStoreEntry(cbEntryID, lpEntryID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	sEntryId.__ptr = (unsigned char*)lpUnWrapStoreID;
	sEntryId.__size = cbUnWrapStoreID;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getRights(m_ecSessionId, sEntryId, ulType, &sRightResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sRightResponse.er;

	}
	END_SOAP_CALL

	ECAllocateBuffer(sizeof(ECPERMISSION) * sRightResponse.pRightsArray->__size, (void**)&lpECPermissions);
	for(unsigned int i = 0; i < sRightResponse.pRightsArray->__size; i++)
	{
		lpECPermissions[i].ulRights	= sRightResponse.pRightsArray->__ptr[i].ulRights;
		lpECPermissions[i].ulState	= sRightResponse.pRightsArray->__ptr[i].ulState;
		lpECPermissions[i].ulType	= sRightResponse.pRightsArray->__ptr[i].ulType;

		hr = CopySOAPEntryIdToMAPIEntryId(&sRightResponse.pRightsArray->__ptr[i].sUserId, sRightResponse.pRightsArray->__ptr[i].ulUserid, MAPI_MAILUSER, (ULONG*)&lpECPermissions[i].sUserId.cb, (LPENTRYID*)&lpECPermissions[i].sUserId.lpb, lpECPermissions);
		if (hr != hrSuccess)
			goto exit;
	}

	*lppECPermissions = lpECPermissions;
	*lpcPermissions = sRightResponse.pRightsArray->__size;
	lpECPermissions = NULL;

exit:
	if (lpECPermissions != NULL)
		ECFreeBuffer(lpECPermissions);

	UnLockSoap();

	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT WSTransport::HrSetPermissionRules(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG cPermissions, LPECPERMISSION lpECPermissions)
{
	ECRESULT		er = erSuccess;
	HRESULT			hr = hrSuccess;
	entryId			sEntryId = {0}; // Do not free
	int				nChangedItems = 0;
	unsigned int	i,
					nItem;
	LPENTRYID		lpUnWrapStoreID = NULL;
	ULONG			cbUnWrapStoreID = 0;
	
	struct rightsArray rArray;
	
	LockSoap();

	if(cPermissions == 0 || lpECPermissions == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Remove servername, always
	hr = UnWrapServerClientStoreEntry(cbEntryID, lpEntryID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	sEntryId.__ptr = (unsigned char*)lpUnWrapStoreID;
	sEntryId.__size = cbUnWrapStoreID;

	// Count the updated items
	for(i=0; i < cPermissions; i++){
		if(lpECPermissions[i].ulState != RIGHT_NORMAL)
			nChangedItems++;
	}

	rArray.__ptr = s_alloc<rights>(m_lpCmd->soap, nChangedItems);
	rArray.__size = nChangedItems;

	nItem = 0;
	for(i=0; i < cPermissions; i++){
		if(lpECPermissions[i].ulState != RIGHT_NORMAL){
			rArray.__ptr[nItem].ulRights = lpECPermissions[i].ulRights;
			rArray.__ptr[nItem].ulState	 = lpECPermissions[i].ulState;
			rArray.__ptr[nItem].ulType	 = lpECPermissions[i].ulType;
			rArray.__ptr[nItem].ulUserid = ABEID_ID(lpECPermissions[i].sUserId.lpb);

			hr = CopyMAPIEntryIdToSOAPEntryId(lpECPermissions[i].sUserId.cb, (LPENTRYID)lpECPermissions[i].sUserId.lpb, &rArray.__ptr[nItem].sUserId, true);
			if (hr != hrSuccess)
				goto exit;

			nItem++;
		}
	}

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__setRights(m_ecSessionId, sEntryId, &rArray, &er))
			er = ZARAFA_E_NETWORK_ERROR;

	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

HRESULT WSTransport::HrGetUserObjectList(ULONG cbCompanyId, LPENTRYID lpCompanyId, int ulType, ULONG* lpcUserObjects, LPECUSEROBJECT* lppECUserobjects)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct userobjectResponse sResponse;
	LPECUSEROBJECT lpECUserobjects = NULL;
	int size;
	entryId sCompanyId = {0};

	LockSoap();
	
	if(lpcUserObjects == NULL || lppECUserobjects == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__getUserObjectList(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, ulType, &sResponse))
			er = ZARAFA_E_SERVER_NOT_RESPONDING;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL
	if (sResponse.pUserObjectArray->__size > 0)
	{
		hr = ECAllocateBuffer(sizeof(ECUSEROBJECT) * sResponse.pUserObjectArray->__size, (void**)&lpECUserobjects);
		if(hr != hrSuccess)
			goto exit;

		for(unsigned int i =0; i < sResponse.pUserObjectArray->__size; i++)
		{
			lpECUserobjects[i].type = (userobject_type_t)sResponse.pUserObjectArray->__ptr[i].ulType;

			if (sResponse.pUserObjectArray->__ptr[i].sId.__size < CbNewABEID("") || sResponse.pUserObjectArray->__ptr[i].sId.__ptr == NULL)
			{
				hr = MAPI_E_INVALID_ENTRYID;
				goto exit;
			}

			hr = ECAllocateMore(sResponse.pUserObjectArray->__ptr[i].sId.__size, lpECUserobjects, (void**)&lpECUserobjects[i].sId.lpb);
			if (hr != hrSuccess)
				goto exit;

			memcpy(lpECUserobjects[i].sId.lpb, sResponse.pUserObjectArray->__ptr[i].sId.__ptr, sResponse.pUserObjectArray->__ptr[i].sId.__size);
			lpECUserobjects[i].sId.cb = sResponse.pUserObjectArray->__ptr[i].sId.__size;

			if(sResponse.pUserObjectArray->__ptr[i].lpszName)
			{
				size = strlen(sResponse.pUserObjectArray->__ptr[i].lpszName) + 1;
				hr = ECAllocateMore(sizeof(char) * size, lpECUserobjects, (void**)&lpECUserobjects[i].lpszName);
				if(hr != hrSuccess)
					goto exit;
				
				memcpy(lpECUserobjects[i].lpszName, sResponse.pUserObjectArray->__ptr[i].lpszName, size);
			}else
				lpECUserobjects[i].lpszName = NULL;
		}
	}

	*lppECUserobjects = lpECUserobjects;
	*lpcUserObjects = sResponse.pUserObjectArray->__size;

exit:
	
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetOwner(ULONG cbEntryID, LPENTRYID lpEntryID, ULONG *lpcbOwnerId, LPENTRYID *lppOwnerId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId	sEntryId = {0}; // Do not free
	struct getOwnerResponse sResponse;
	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	LockSoap();

	if (lpcbOwnerId == NULL || lppOwnerId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Remove servername, always
	hr = UnWrapServerClientStoreEntry(cbEntryID, lpEntryID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	sEntryId.__ptr = (unsigned char*)lpUnWrapStoreID;
	sEntryId.__size = cbUnWrapStoreID;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getOwner(m_ecSessionId, sEntryId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL

	hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sOwner, sResponse.ulOwner, lpcbOwnerId, lppOwnerId);

exit:
	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrResolveNames(LPSPropTagArray lpPropTagArray, ULONG ulFlags, LPADRLIST lpAdrList, LPFlagList lpFlagList)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	struct propTagArray aPropTag;
	struct rowSet* lpsRowSet = NULL;
	struct flagArray aFlags;
	struct abResolveNamesResponse sResponse;
	unsigned int i;

	LockSoap();

	aPropTag.__ptr = (unsigned int *)&lpPropTagArray->aulPropTag; // just a referance
	aPropTag.__size = lpPropTagArray->cValues;

	aFlags.__ptr = (unsigned int *)&lpFlagList->ulFlag;
	aFlags.__size = lpFlagList->cFlags;

	hr = CopyMAPIRowSetToSOAPRowSet((LPSRowSet)lpAdrList, &lpsRowSet);
	if(hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__abResolveNames(m_ecSessionId, &aPropTag, lpsRowSet, &aFlags, ulFlags, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;

	}
	END_SOAP_CALL
	
	ASSERT(sResponse.aFlags.__size == lpFlagList->cFlags);
	ASSERT(sResponse.sRowSet.__size == lpAdrList->cEntries);

	for (i = 0; i < sResponse.aFlags.__size; i++)
	{
		// Set the resolved items
		if(lpFlagList->ulFlag[i] == MAPI_UNRESOLVED && sResponse.aFlags.__ptr[i] == MAPI_RESOLVED)
		{
			lpAdrList->aEntries[i].cValues = sResponse.sRowSet.__ptr[i].__size;
			ECFreeBuffer(lpAdrList->aEntries[i].rgPropVals);

			ECAllocateBuffer(sizeof(SPropValue)*lpAdrList->aEntries[i].cValues, (void**)&lpAdrList->aEntries[i].rgPropVals);

			hr = CopySOAPRowToMAPIRow(&sResponse.sRowSet.__ptr[i], lpAdrList->aEntries[i].rgPropVals, (void*)lpAdrList->aEntries[i].rgPropVals);
			if(hr != hrSuccess)
				goto exit;

			lpFlagList->ulFlag[i] = sResponse.aFlags.__ptr[i];
		}else { // MAPI_AMBIGUOUS or MAPI_UNRESOLVED
			// Set only the flag does notting with the row
			lpFlagList->ulFlag[i] = sResponse.aFlags.__ptr[i];
		}		
	}

exit:
	UnLockSoap();

	if(lpsRowSet)
		FreeRowSet(lpsRowSet, true);

	return hr;
}

HRESULT WSTransport::HrSyncUsers(ULONG cbCompanyId, LPENTRYID lpCompanyId)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	unsigned int sResponse;
	entryId sCompanyId = {0};
	ULONG ulCompanyId = 0;

	LockSoap();

	if (lpCompanyId)
	{
		hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
		if (hr != hrSuccess)
			goto exit;
		ulCompanyId = ABEID_ID(lpCompanyId);
	}

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__syncUsers(m_ecSessionId, ulCompanyId, sCompanyId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = (ECRESULT) sResponse;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::GetQuota(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTA* lppsQuota)
{
	ECRESULT				er = erSuccess;
	HRESULT					hr = hrSuccess;
	struct quotaResponse	sResponse;
	LPECQUOTA				lpsQuota =  NULL;
	entryId					sUserId = {0};
	
	LockSoap();
	
	if(lppsQuota == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__GetQuota(m_ecSessionId, ABEID_ID(lpUserId), sUserId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = (ECRESULT) sResponse.er;
	}
	END_SOAP_CALL

	ECAllocateBuffer(sizeof(ECQUOTA), (void**)&lpsQuota);

	lpsQuota->bUseDefaultQuota = sResponse.sQuota.bUseDefaultQuota;
	lpsQuota->bIsUserDefaultQuota = sResponse.sQuota.bIsUserDefaultQuota;
	lpsQuota->llHardSize = sResponse.sQuota.llHardSize;
	lpsQuota->llSoftSize = sResponse.sQuota.llSoftSize;
	lpsQuota->llWarnSize = sResponse.sQuota.llWarnSize;

	*lppsQuota = lpsQuota;

  exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::SetQuota(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTA lpsQuota)
{
	ECRESULT				er = erSuccess;
	HRESULT					hr = hrSuccess;
	unsigned int			sResponse;
	struct quota			sQuota;
	entryId					sUserId = {0};
	
	LockSoap();

	if(lpsQuota == NULL || lpUserId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	sQuota.bUseDefaultQuota = lpsQuota->bUseDefaultQuota;
	sQuota.bIsUserDefaultQuota = lpsQuota->bIsUserDefaultQuota;
	sQuota.llHardSize = lpsQuota->llHardSize;
	sQuota.llSoftSize = lpsQuota->llSoftSize;
	sQuota.llWarnSize = lpsQuota->llWarnSize;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__SetQuota(m_ecSessionId, ABEID_ID(lpUserId), sUserId, &sQuota, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = (ECRESULT) sResponse;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::AddQuotaRecipient(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG cbRecipientId, LPENTRYID lpRecipientId, ULONG ulType)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId	sCompanyId = {0};
	entryId	sRecipientId = {0};

	LockSoap();

	if (lpCompanyId == NULL || lpRecipientId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbRecipientId, lpRecipientId, &sRecipientId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__AddQuotaRecipient(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, ABEID_ID(lpRecipientId), sRecipientId, ulType, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();
	return hr;
}

HRESULT WSTransport::DeleteQuotaRecipient(ULONG cbCompanyId, LPENTRYID lpCompanyId, ULONG cbRecipientId, LPENTRYID lpRecipientId, ULONG ulType)
{
	ECRESULT er = erSuccess;
	HRESULT hr = hrSuccess;
	entryId sCompanyId = {0};
	entryId sRecipientId = {0};

	LockSoap();

	if (lpCompanyId == NULL || lpRecipientId == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbCompanyId, lpCompanyId, &sCompanyId, true);
	if (hr != hrSuccess)
		goto exit;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbRecipientId, lpRecipientId, &sRecipientId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__DeleteQuotaRecipient(m_ecSessionId, ABEID_ID(lpCompanyId), sCompanyId, ABEID_ID(lpRecipientId), sRecipientId, ulType, &er))
			er = ZARAFA_E_NETWORK_ERROR;
	}
	END_SOAP_CALL

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::GetQuotaRecipients(ULONG cbUserId, LPENTRYID lpUserId, ULONG *lpcUsers, LPECUSER *lppsUsers)
{
	ECRESULT	er = erSuccess;
	HRESULT		hr = hrSuccess;
	entryId		sUserId = {0};
	struct userListResponse sResponse;

	LockSoap();

	if(lpcUsers == NULL || lppsUsers == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	*lpcUsers = 0;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__GetQuotaRecipients(m_ecSessionId, ABEID_ID(lpUserId), sUserId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
	}
	END_SOAP_CALL

	hr = SoapUserArrayToUserArray(&sResponse.sUserArray, lpcUsers, lppsUsers);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::GetQuotaStatus(ULONG cbUserId, LPENTRYID lpUserId, LPECQUOTASTATUS* lppsQuotaStatus)
{
	ECRESULT				er = erSuccess;
	HRESULT					hr = hrSuccess;
	struct quotaStatus		sResponse;
	LPECQUOTASTATUS			lpsQuotaStatus =  NULL;
	entryId					sUserId = {0};
	
	LockSoap();

	if(lppsQuotaStatus == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CopyMAPIEntryIdToSOAPEntryId(cbUserId, lpUserId, &sUserId, true);
	if (hr != hrSuccess)
		goto exit;

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__GetQuotaStatus(m_ecSessionId, ABEID_ID(lpUserId), sUserId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = (ECRESULT) sResponse.er;
	}
	END_SOAP_CALL

	ECAllocateBuffer(sizeof(ECQUOTASTATUS), (void**)&lpsQuotaStatus);

	lpsQuotaStatus->llStoreSize = sResponse.llStoreSize;
	lpsQuotaStatus->quotaStatus = (eQuotaStatus)sResponse.ulQuotaStatus;

	*lppsQuotaStatus = lpsQuotaStatus;

  exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrPurgeSoftDelete(ULONG ulDays)
{
    HRESULT						hr = hrSuccess;
    ECRESULT					er = erSuccess;
    
    LockSoap();
    
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__purgeSoftDelete(m_ecSessionId, ulDays, &er))
			er = ZARAFA_E_NETWORK_ERROR;
        
	}
	END_SOAP_CALL
    
exit:
    UnLockSoap();
    
    return hr;
}

HRESULT WSTransport::HrResolvePseudoUrl(char *lpszPseudoUrl, char **lppszServerPath, bool *lpbIsPeer)
{
	ECRESULT						er = erSuccess;
	HRESULT							hr = hrSuccess;
	struct resolvePseudoUrlResponse	sResponse = {0};
	char							*lpszServerPath = NULL;
	unsigned int					ulLen = 0;

	LockSoap();

	if (lpszPseudoUrl == NULL || lppszServerPath == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	START_SOAP_CALL
	{
    	if(SOAP_OK != m_lpCmd->ns__resolvePseudoUrl(m_ecSessionId, lpszPseudoUrl, &sResponse))
    		er = ZARAFA_E_NETWORK_ERROR;
    	else
    		er = (ECRESULT)sResponse.er;
    }
	END_SOAP_CALL

	ulLen = strlen(sResponse.lpszServerPath) + 1;
	hr = ECAllocateBuffer(ulLen, (void**)&lpszServerPath);
	if (hr != hrSuccess)
		goto exit;

	memcpy(lpszServerPath, sResponse.lpszServerPath, ulLen);
	*lppszServerPath = lpszServerPath;
	*lpbIsPeer = sResponse.bIsPeer;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrGetServerDetails(LPECSVRNAMELIST lpServerNameList, ULONG ulFlags, LPECSERVERLIST* lppsServerList)
{
	ECRESULT						er = erSuccess;
	HRESULT							hr = hrSuccess;
	struct getServerDetailsResponse	sResponse = {{0}};
	struct mv_string8				*lpsSvrNameList = NULL;

	LockSoap();

	if (lpServerNameList == NULL || lppsServerList == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	hr = SvrNameListToSoapMvString8(lpServerNameList, &lpsSvrNameList);
	if (hr != hrSuccess)
		goto exit;


	START_SOAP_CALL
	{
    	if( SOAP_OK != m_lpCmd->ns__getServerDetails(m_ecSessionId, *lpsSvrNameList, ulFlags, &sResponse))
    		er = ZARAFA_E_NETWORK_ERROR;
    	else
    		er = (ECRESULT)sResponse.er;
    }
    END_SOAP_CALL

	hr = SoapServerListToServerList(&sResponse.sServerList, lppsServerList);
	if (hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();
	
	if (lpsSvrNameList)
		ECFreeBuffer(lpsSvrNameList);

	return hr;
}

HRESULT WSTransport::HrGetChanges(SBinary sSourceKeyFolder, ULONG ulSyncId, ULONG ulChangeId, ULONG ulSyncType, ULONG ulFlags, LPSRestriction lpsRestrict, ULONG *lpulMaxChangeId, ULONG* lpcChanges, ICSCHANGE **lppChanges){
	HRESULT						hr = hrSuccess;
	ECRESULT					er = erSuccess;
	struct icsChangeResponse	sResponse;
	ICSCHANGE *					lpChanges = NULL;
	unsigned int 				i = 0;
	struct xsd__base64Binary	sSourceKey;
	struct restrictTable		*lpsSoapRestrict = NULL;
	
	sSourceKey.__ptr = sSourceKeyFolder.lpb;
	sSourceKey.__size = sSourceKeyFolder.cb;

	LockSoap();

	if(lpsRestrict) {
    	hr = CopyMAPIRestrictionToSOAPRestriction(&lpsSoapRestrict, lpsRestrict);
    	if(hr != hrSuccess)
	        goto exit;
    }

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getChanges(m_ecSessionId, sSourceKey, ulSyncId, ulChangeId, ulSyncType, ulFlags, lpsSoapRestrict, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = (ECRESULT) sResponse.er;

	}
	END_SOAP_CALL

	ECAllocateBuffer(sResponse.sChangesArray.__size * sizeof(ICSCHANGE), (void**)&lpChanges);

	for(i=0;i<sResponse.sChangesArray.__size;i++) {
		lpChanges[i].ulChangeId = sResponse.sChangesArray.__ptr[i].ulChangeId;
		lpChanges[i].ulChangeType = sResponse.sChangesArray.__ptr[i].ulChangeType;
		lpChanges[i].ulFlags = sResponse.sChangesArray.__ptr[i].ulFlags;

		if(sResponse.sChangesArray.__ptr[i].sSourceKey.__size > 0) {
			ECAllocateMore( sResponse.sChangesArray.__ptr[i].sSourceKey.__size, lpChanges, (void **)&lpChanges[i].sSourceKey.lpb);
			lpChanges[i].sSourceKey.cb = sResponse.sChangesArray.__ptr[i].sSourceKey.__size;
			memcpy(lpChanges[i].sSourceKey.lpb, sResponse.sChangesArray.__ptr[i].sSourceKey.__ptr, sResponse.sChangesArray.__ptr[i].sSourceKey.__size);
		}
		
		if(sResponse.sChangesArray.__ptr[i].sParentSourceKey.__size > 0) {
			ECAllocateMore( sResponse.sChangesArray.__ptr[i].sParentSourceKey.__size, lpChanges, (void **)&lpChanges[i].sParentSourceKey.lpb);
			lpChanges[i].sParentSourceKey.cb = sResponse.sChangesArray.__ptr[i].sParentSourceKey.__size;
			memcpy(lpChanges[i].sParentSourceKey.lpb, sResponse.sChangesArray.__ptr[i].sParentSourceKey.__ptr, sResponse.sChangesArray.__ptr[i].sParentSourceKey.__size);
		}
		
		if(sResponse.sChangesArray.__ptr[i].sMovedFromSourceKey.__size > 0) {
			ECAllocateMore( sResponse.sChangesArray.__ptr[i].sMovedFromSourceKey.__size, lpChanges, (void **)&lpChanges[i].sMovedFromSourceKey.lpb);
			lpChanges[i].sMovedFromSourceKey.cb = sResponse.sChangesArray.__ptr[i].sMovedFromSourceKey.__size;
			memcpy(lpChanges[i].sMovedFromSourceKey.lpb, sResponse.sChangesArray.__ptr[i].sMovedFromSourceKey.__ptr, sResponse.sChangesArray.__ptr[i].sMovedFromSourceKey.__size);
		}
	}
	
	*lpulMaxChangeId = sResponse.ulMaxChangeId;
	*lpcChanges = sResponse.sChangesArray.__size;
	*lppChanges = lpChanges;

exit:
	UnLockSoap();
	
	if(lpsSoapRestrict)
	    FreeRestrictTable(lpsSoapRestrict);

	if(hr!=hrSuccess && lpChanges)
		ECFreeBuffer(lpChanges);

	return hr;
}

HRESULT WSTransport::HrSetSyncStatus(SBinary sSourceKeyFolder, ULONG ulSyncId, ULONG ulChangeId, ULONG ulSyncType, ULONG ulFlags, ULONG* lpulSyncId){
	HRESULT				hr = hrSuccess;
	ECRESULT			er = erSuccess;
	struct setSyncStatusResponse sResponse;
	struct xsd__base64Binary sSourceKey;
	
	sSourceKey.__size = sSourceKeyFolder.cb;
	sSourceKey.__ptr = sSourceKeyFolder.lpb;

	LockSoap();

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__setSyncStatus(m_ecSessionId, sSourceKey, ulSyncId, ulChangeId, ulSyncType, ulFlags, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = (ECRESULT) sResponse.er;

	}
	END_SOAP_CALL

	*lpulSyncId = sResponse.ulSyncId;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrEntryIDFromSourceKey(ULONG cbStoreID, LPENTRYID lpStoreID, ULONG ulFolderSourceKeySize, BYTE * lpFolderSourceKey, ULONG ulMessageSourceKeySize, BYTE * lpMessageSourceKey, ULONG * lpcbEntryID, LPENTRYID * lppEntryID)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;
	entryId		sStoreId;
	LPENTRYID	lpUnWrapStoreID = NULL;
	ULONG		cbUnWrapStoreID = 0;

	struct xsd__base64Binary	folderSourceKey;
	struct xsd__base64Binary	messageSourceKey;

	struct getEntryIDFromSourceKeyResponse sResponse;	
	

	LockSoap();

	if(ulFolderSourceKeySize == 0 || lpFolderSourceKey == NULL){
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	hr = UnWrapServerClientStoreEntry(cbStoreID, lpStoreID, &cbUnWrapStoreID, &lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	sStoreId.__ptr = (unsigned char*)lpUnWrapStoreID;
	sStoreId.__size = cbUnWrapStoreID;

	folderSourceKey.__size = ulFolderSourceKeySize;
	folderSourceKey.__ptr = lpFolderSourceKey;

	messageSourceKey.__size = ulMessageSourceKeySize; // can be zero
	messageSourceKey.__ptr = lpMessageSourceKey;

	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getEntryIDFromSourceKey(m_ecSessionId, sStoreId, folderSourceKey, messageSourceKey, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = (ECRESULT) sResponse.er;

	}
	END_SOAP_CALL

	hr = CopySOAPEntryIdToMAPIEntryId(&sResponse.sEntryId, lpcbEntryID, lppEntryID, NULL);
	if(hr != hrSuccess)
		goto exit;

exit:
	UnLockSoap();

	if(lpUnWrapStoreID)
		ECFreeBuffer(lpUnWrapStoreID);

	return hr;
}

const char* WSTransport::GetServerName()
{
	return m_sProfileProps.strServerPath.c_str();
}

bool WSTransport::IsConnected()
{
	return m_lpCmd != NULL;
}

HRESULT WSTransport::HrOpenMultiStoreTable(LPENTRYLIST lpMsgList, ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID, ECMsgStore *lpMsgStore, WSTableView **lppTableView)
{
	HRESULT hr = hrSuccess;
	WSTableMultiStore *lpMultiStoreTable = NULL;


	if (!lpMsgList || lpMsgList->cValues == 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}


	hr = WSTableMultiStore::Create(ulFlags, m_lpCmd, m_hDataLock, m_ecSessionId, cbEntryID, lpEntryID, lpMsgStore, this, &lpMultiStoreTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMultiStoreTable->HrSetEntryIDs(lpMsgList);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMultiStoreTable->QueryInterface(IID_ECTableView, (void **)lppTableView);

exit:
	if (lpMultiStoreTable)
		lpMultiStoreTable->Release();

	return hr;
}


HRESULT WSTransport::HrOpenMiscTable(ULONG ulTableType, ULONG ulFlags, ULONG cbEntryID, LPENTRYID lpEntryID, ECMsgStore *lpMsgStore, WSTableView **lppTableView)
{
	HRESULT hr = hrSuccess;
	WSTableMisc *lpMiscTable = NULL;

	if (ulTableType != TABLETYPE_STATS_SYSTEM && ulTableType != TABLETYPE_STATS_SESSIONS &&
		ulTableType != TABLETYPE_STATS_USERS && ulTableType != TABLETYPE_STATS_COMPANY  &&
		ulTableType != TABLETYPE_USERSTORES)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = WSTableMisc::Create(ulTableType, ulFlags, m_lpCmd, m_hDataLock, m_ecSessionId, cbEntryID, lpEntryID, lpMsgStore, this, &lpMiscTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMiscTable->QueryInterface(IID_ECTableView, (void **)lppTableView);

exit:
	if (lpMiscTable)
		lpMiscTable->Release();

	return hr;
}

HRESULT WSTransport::HrGetServerBehavior(ULONG *lpulBehavior)
{
	HRESULT								hr = hrSuccess;
	ULONG								ulBehavior = 0;
	struct getServerBehaviorResponse	sResponse = {0};
	
	LockSoap();

	int rc = m_lpCmd->ns__getServerBehavior(m_ecSessionId, &sResponse);
	TRACE_NOTIFY(TRACE_WARNING, "WSTransport::HrGetServerBehavior", "%d", rc);
	switch (rc)
	{
	case SOAP_OK:
		hr = ZarafaErrorToMAPIError(sResponse.er);
		if (hr != hrSuccess)
			goto exit;

		ulBehavior = sResponse.ulBehavior;
		break;

	case SOAP_CLI_FAULT:	// (old) Server doesn't know method.
		break;

	default:
		hr = MAPI_E_NETWORK_ERROR;
		goto exit;
	}

	*lpulBehavior = ulBehavior;

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrSetServerBehavior(ULONG ulBehavior)
{
	HRESULT			hr = hrSuccess;
	unsigned int	ulResponse = 0;
	
	LockSoap();

	int rc = m_lpCmd->ns__setServerBehavior(m_ecSessionId, ulBehavior, &ulResponse);
	TRACE_NOTIFY(TRACE_WARNING, "WSTransport::HrSetServerBehavior", "%d", rc);
	switch (rc)
	{
	case SOAP_OK:
		hr = ZarafaErrorToMAPIError(ulResponse);
		break;

	case SOAP_NO_METHOD:
		hr = MAPI_E_NOT_SUPPORTED;
		break;

	default:
		hr = MAPI_E_NETWORK_ERROR;
		break;
	}

exit:
	UnLockSoap();

	return hr;
}

HRESULT WSTransport::HrLicenseAuth(unsigned char *lpData, unsigned int ulSize, unsigned char **lppResponseData, unsigned int *lpulSize)
{
    HRESULT hr = hrSuccess;
    ECRESULT er = erSuccess;
    struct getLicenseAuthResponse sResponse;
    struct xsd__base64Binary sData;
    
    sData.__ptr = lpData;
    sData.__size = ulSize;
    
    LockSoap();
    
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getLicenseAuth(m_ecSessionId, sData, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
        
	}
	END_SOAP_CALL
        
    hr = MAPIAllocateBuffer(sResponse.sAuthResponse.__size, (void **) lppResponseData);
    if(hr != hrSuccess)
        goto exit;
        
    memcpy(*lppResponseData, sResponse.sAuthResponse.__ptr, sResponse.sAuthResponse.__size);
    *lpulSize = sResponse.sAuthResponse.__size;
    
exit:
    UnLockSoap();
    
    return hr;
}

HRESULT WSTransport::HrLicenseCapa(char ***lppszCapas, unsigned int * lpulSize)
{
    HRESULT hr = hrSuccess;
    ECRESULT er = erSuccess;
    struct getLicenseCapaResponse sResponse;
    
    char **lpszCapas = NULL;
    
    LockSoap();
    
	START_SOAP_CALL
	{
		if(SOAP_OK != m_lpCmd->ns__getLicenseCapa(m_ecSessionId, &sResponse))
			er = ZARAFA_E_NETWORK_ERROR;
		else
			er = sResponse.er;
        
	}
	END_SOAP_CALL
        
    hr = MAPIAllocateBuffer(sResponse.sCapabilities.__size * sizeof(char *), (void **)&lpszCapas);
    if(hr != hrSuccess)
        goto exit;

    for(int i=0; i < sResponse.sCapabilities.__size; i++) {
        MAPIAllocateMore(strlen(sResponse.sCapabilities.__ptr[i])+1, lpszCapas, (void **) &lpszCapas[i]);
        strcpy(lpszCapas[i], sResponse.sCapabilities.__ptr[i]);
    }
    
    *lppszCapas = lpszCapas;
    *lpulSize = sResponse.sCapabilities.__size;
    
exit:
    UnLockSoap();
    
    return hr;
}


HRESULT WSTransport::HrGetSessionId(ECSESSIONID *lpSessionId, ECSESSIONGROUPID *lpSessionGroupId)
{
	HRESULT hr = hrSuccess;

	if (lpSessionId)
		*lpSessionId = m_ecSessionId;
	if (lpSessionGroupId)
		*lpSessionGroupId = m_ecSessionGroupId;

    return hr;
}

sGlobalProfileProps WSTransport::GetProfileProps()
{
    return m_sProfileProps;
}

HRESULT WSTransport::GetLicenseFlags(unsigned long long *lpllFlags)
{
    *lpllFlags = m_llFlags;
    
    return hrSuccess;
}

HRESULT WSTransport::GetServerGUID(LPGUID lpsServerGuid)
{
	if (m_sServerGuid == GUID_NULL)
		return MAPI_E_NOT_FOUND;

	*lpsServerGuid = m_sServerGuid;
	return hrSuccess;
}

HRESULT WSTransport::AddSessionReloadCallback(void *lpParam, SESSIONRELOADCALLBACK callback, ULONG *lpulId)
{
	SESSIONRELOADLIST::mapped_type data(lpParam, callback);

	pthread_mutex_lock(&m_mutexSessionReload);
	m_mapSessionReload[m_ulReloadId] = data;
	if(lpulId)
		*lpulId = m_ulReloadId;
	m_ulReloadId++;
	pthread_mutex_unlock(&m_mutexSessionReload);

	return hrSuccess;
}

HRESULT WSTransport::RemoveSessionReloadCallback(ULONG ulId)
{
	HRESULT hr = hrSuccess;
	SESSIONRELOADLIST::iterator iter;
	
	pthread_mutex_lock(&m_mutexSessionReload);
	iter = m_mapSessionReload.find(ulId);
	if(iter != m_mapSessionReload.end()) {
		m_mapSessionReload.erase(iter);
	} else {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

exit:
	pthread_mutex_unlock(&m_mutexSessionReload);

	return hr;
}

int WSTransport::RefuseConnect(struct soap* soap, const char* endpoint, const char* host, int port)
{
	soap->error = SOAP_TCP_ERROR;
	return SOAP_ERR;
}

HRESULT WSTransport::HrCancelIO()
{
	HRESULT hr = hrSuccess;

	/* This looks like an ugly hack, but in practice it works very well: we want
	 * to make sure that no blocking call is running or will run in this class instance
	 * after HrCancelIO() returns.
	 *
	 * To do this, we first disable sending new requests by override the fsend() function
	 * in gSOAP. If a request was already started but still 'before' the fsend() call
	 * then the call will return directly after the fsend() because it will return an error.
	 * On the other hand, if fsend() has already been called, then the next blocking part
	 * will be the frecv() call. We therefore shutdown the socket here as well, so that
	 * any blocking recv() call will return with an error immediately.
	 *
	 * Please note though, that this last part (shutdown() on a socket will cause other threads
	 * with a blocking recv() on the socket to return with an error), is not 100% portable. 
	 * Apparently (and this is untested), some UNIX variants don't return until some data
	 * is received on the socket. Bummer. It works in Linux though.
	 *
	 * Win32 is a whole different story, we can summarize it to: this function does nothing on win32.
	 * The long story is this:
	 *		fshutdownsocket() on a named pipe (offline server) is not implemented and thus will not
	 *		work. Microsoft documentation doesn't give any hints about shutting down a named pipe
	 *		without calling fclose().
	 *		fshutdownsocket() on TCP socket does not break off any blocking TCP send/recv calls
	 *		which means the socket will only be closed _after_ the last blocking call has been
	 *		finalized.
	 *		fstop() on a named pipe will wait until all blocking send/recv calls to the server
	 *		have been finalized. It will ignore SO_LINGER/SO_DONTLINGER settings on the socket,
	 *		this means it should not be used in this function since we don't want to wait
	 *		on the send/recv calls but we want to STOP them.
	 *		fstop() on a TCP socket will work as expected and shutsdown all blocking send/recv calls.
	 */

	// Override the SOAP connect (fopen) so that all new connections will fail with a network error
	m_lpCmd->soap->fopen = WSTransport::RefuseConnect;

	// If there is a socket currently open, close it now
	if(m_lpCmd && m_lpCmd->soap) {
		int s = m_lpCmd->soap->socket;
		if (s != SOAP_INVALID_SOCKET) {
			m_lpCmd->soap->fshutdownsocket(m_lpCmd->soap, (SOAP_SOCKET)s, 2);
		}
	}

	return hr;
}

HRESULT WSTransport::HrGetNotify(struct notificationArray **lppsArrayNotifications)
{
	HRESULT		hr = hrSuccess;
	ECRESULT	er = erSuccess;

	struct notifyResponse sNotifications;
	
	LockSoap();

	if(SOAP_OK != m_lpCmd->ns__notifyGetItems(m_ecSessionId, &sNotifications))
		er = ZARAFA_E_NETWORK_ERROR;
	else
		er = sNotifications.er;	// hrSuccess or ZARAFA_W_KEEP_ALIVE only

	hr = ZarafaErrorToMAPIError(er);
	if(hr != erSuccess)
		goto exit;

	if(sNotifications.pNotificationArray != NULL) {
		*lppsArrayNotifications = new notificationArray;
		CopyNotificationArrayStruct(sNotifications.pNotificationArray, *lppsArrayNotifications);
	}else
		*lppsArrayNotifications = NULL;

exit:
	UnLockSoap();

	if(m_lpCmd->soap)
		soap_end(m_lpCmd->soap);

	return hr;
}

std::string WSTransport::GetAppName()
{
    if(!m_strAppName.empty())
        return m_strAppName;
    std::string procpath = "/proc/" + stringify(getpid()) + "/exe";
    char s[1024];
    memset(s, 0, sizeof(s));

    if(readlink(procpath.c_str(), s, sizeof(s)) == -1)
        m_strAppName = "<unknown>";
    else {
        s[sizeof(s)-1] = 0;

        m_strAppName = basename(s);    
    }
    return m_strAppName;
}
