/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapidefs.h>
#include <errno.h>

#include <algorithm>

#include <zlib.h>

#include "ECAttachmentStorage.h"
#include "ECDBDef.h"
#include "SOAPUtils.h"
#include "mapitags.h"
#include "stringutil.h"

// chunk size for attachment blobs, must be equal or larger than MAX, MAX may never shrink below 384*1024.
#define CHUNK_SIZE 384*1024

/*
 * Locking requirements of ECAttachmentStorage:
 * In the case of ECAttachmentStorage locking to protect against concurrent access is futile.
 * The theory is as follows:
 * If 2 users have a reference to the same attachment, neither can delete the mail causing
 * the other person to loose the attachment. This means that concurrent copy and delete actions
 * are possible on the same attachment. The only case where this does not hold is the case
 * when a user deletes the mail he is copying at the same time and that this is the last mail
 * which references that specific attachment.
 * The only race condition which might occur is that the dagent delivers a mail with attachment,
 * the server returns the attachment id back to the dagent, but the user for whom the message
 * was intended deletes the mail & attachment. In this case the dagent will no longer send the
 * attachment but only the attachment id. When that happens the server can return an error
 * and simply request the dagent to resend the attachment and to obtain a new attachment id.
 */

// Generic Attachment storage
ECAttachmentStorage::ECAttachmentStorage(ECDatabase *lpDatabase, bool bFileStorage, unsigned int ulCompressionLevel)
	: m_lpDatabase(lpDatabase), m_bFileStorage(bFileStorage)
{
	m_bFileCompression = ulCompressionLevel != 0;

	if (ulCompressionLevel > Z_BEST_COMPRESSION)
		ulCompressionLevel = Z_BEST_COMPRESSION;

	m_CompressionLevel = stringify(ulCompressionLevel);
}

ECAttachmentStorage::~ECAttachmentStorage()
{
}

bool ECAttachmentStorage::IsFileStorage()
{
	return m_bFileStorage;
}

ECRESULT ECAttachmentStorage::CreateAttachmentStorage(ECDatabase *lpDatabase, ECConfig *lpConfig, ECAttachmentStorage **lppAttachmentStorage)
{
	ECAttachmentStorage *lpAttachmentStorage = NULL;

	if (lpDatabase == NULL)
		return ZARAFA_E_DATABASE_ERROR; // somebody called this function too soon.

	if (strcmp(lpConfig->GetSetting("attachment_storage"), "files") == 0) {
		lpAttachmentStorage = new ECFileAttachment(lpDatabase, lpConfig->GetSetting("attachment_path"), atoi(lpConfig->GetSetting("attachment_compression")));
	} else {
		lpAttachmentStorage = new ECDatabaseAttachment(lpDatabase);
	}

	*lppAttachmentStorage = lpAttachmentStorage;

	return erSuccess;
}

ECRESULT ECAttachmentStorage::GetSingleInstanceId(ULONG ulObjId, ULONG *lpulAttachId, ULONG *lpulTag)
{	
	ECRESULT er = erSuccess;
	std::string strQuery;
	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;

	strQuery =
		"SELECT `instanceid`, `tag` "
		"FROM `singleinstances` "
		"WHERE `hierarchyid` = " + stringify(ulObjId) + " "
		"LIMIT 1";

	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpDBRow = m_lpDatabase->FetchRow(lpDBResult);

	if (!lpDBRow || !lpDBRow[0] || !lpDBRow[1]) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	if (lpulAttachId)
		*lpulAttachId = atoi(lpDBRow[0]);
	if (lpulTag)
		*lpulTag = atoi(lpDBRow[1]);

exit:
	if (lpDBResult)
		m_lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECAttachmentStorage::GetSingleInstanceIds(std::list<ULONG> &lstObjIds, std::list<ULONG> *lstAttachIds)
{
	ECRESULT er = erSuccess;
	std::string strQuery;
	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;
	ULONG ulRows = 0;

	strQuery =
		"SELECT DISTINCT `instanceid` "
		"FROM `singleinstances` "
		"WHERE `hierarchyid` IN (";
	for (std::list<ULONG>::iterator i = lstObjIds.begin(); i != lstObjIds.end(); i++) {
		if (i != lstObjIds.begin())
			strQuery += ",";
		strQuery += stringify(*i);
	}

	strQuery += ")";

	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	ulRows = m_lpDatabase->GetNumRows(lpDBResult);

	lstAttachIds->clear();

	for (ULONG i = 0; i < ulRows; i++) {
		lpDBRow = m_lpDatabase->FetchRow(lpDBResult);

		if (lpDBRow == NULL || lpDBRow[0] == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		lstAttachIds->push_back(atoi(lpDBRow[0]));
	}

exit:
	if (lpDBResult)
		m_lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECAttachmentStorage::SetSingleInstanceId(ULONG ulObjId, ULONG ulAttachId, ULONG ulTag)
{
	ECRESULT er = erSuccess;
	std::string strQuery;
	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;

	/*
	 * Check if attachment reference exists, if not return error
	 */
	strQuery =
		"SELECT COUNT(*) "
		"FROM `singleinstances` "
		"WHERE `instanceid` = " + stringify(ulAttachId);
	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpDBRow = m_lpDatabase->FetchRow(lpDBResult);

	/*
	 * No results, apparently the Instance ID no longer exists. 
	 */
	if (!lpDBRow || !lpDBRow[0] || atoi(lpDBRow[0]) == 0) {
		er = ZARAFA_E_UNABLE_TO_COMPLETE;
		goto exit;
	}

	/*
	 * Create Attachment reference, use provided attachment id
	 */
	strQuery =
		"REPLACE INTO `singleinstances` (`instanceid`, `hierarchyid`, `tag`) VALUES"
		"(" + stringify(ulAttachId) + ", " + stringify(ulObjId) + ", " +  stringify(ulTag) + ")";
	er = m_lpDatabase->DoInsert(strQuery, (unsigned int*)&ulAttachId);
	if (er != erSuccess)
		goto exit;

exit:
	if (lpDBResult)
		m_lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECAttachmentStorage::GetSingleInstanceParents(ULONG ulAttachId, std::list<ULONG> *lplstObjIds)
{
	ECRESULT er = erSuccess;
	std::string strQuery;
	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;
	ULONG ulRows = 0;

	strQuery =
		"SELECT DISTINCT `hierarchyid` "
		"FROM `singleinstances` "
		"WHERE `instanceid` = " + stringify(ulAttachId);
	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	ulRows = m_lpDatabase->GetNumRows(lpDBResult);

	lplstObjIds->clear();

	for (ULONG i = 0; i < ulRows; i++) {
		lpDBRow = m_lpDatabase->FetchRow(lpDBResult);

		if (!lpDBRow || !lpDBRow[0]) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		lplstObjIds->push_back(atoi(lpDBRow[0]));
	}

exit:
	if (lpDBResult)
		m_lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECAttachmentStorage::IsOrphanedSingleInstance(ULONG ulAttachId, bool *bOrphan)
{
	ECRESULT er = erSuccess;
	std::string strQuery;
	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;

	strQuery =
		"SELECT `instanceid` "
		"FROM `singleinstances` "
		"WHERE `instanceid` = " + stringify(ulAttachId) + " "
		"LIMIT 1";

	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpDBRow = m_lpDatabase->FetchRow(lpDBResult);

	/*
	 * No results: Single Instance ID has been cleared (refcount = 0)
	 */
	*bOrphan = (!lpDBRow || !lpDBRow[0]);

exit:
	if (lpDBResult)
		m_lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECAttachmentStorage::GetOrphanedSingleInstances(std::list<ULONG> &lstAttachments, std::list<ULONG> *lplstOrphanedAttachments)
{
	ECRESULT er = erSuccess;
	std::string strQuery;
	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;
	ULONG ulRows = 0;

	strQuery =
		"SELECT `instanceid` "
		"FROM `singleinstances` "
		"WHERE `instanceid` IN ( ";
	for (std::list<ULONG>::iterator i = lstAttachments.begin(); i != lstAttachments.end(); i++) {
		if (i != lstAttachments.begin())
			strQuery += ",";
		strQuery += stringify(*i);
	}
	strQuery +=
		") "
		"GROUP BY `instanceid`";

	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	ulRows = m_lpDatabase->GetNumRows(lpDBResult);

	/* First make a full copy of the list of Single Instance IDs */
	lplstOrphanedAttachments->assign(lstAttachments.begin(), lstAttachments.end());

	/*
	 * Now filter out any Single Instance IDs which were returned in the query,
	 * any results not returned by the query imply that the refcount is 0
	 */
	for (ULONG i = 0; i < ulRows; i++) {
		lpDBRow = m_lpDatabase->FetchRow(lpDBResult);

		if (!lpDBRow || !lpDBRow[0]) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		lplstOrphanedAttachments->remove(atoi(lpDBRow[0]));
	}

exit:
	if (lpDBResult)
		m_lpDatabase->FreeResult(lpDBResult);

	return er;
}

bool ECAttachmentStorage::ExistAttachment(ULONG ulObjId)
{
	ECRESULT er = erSuccess;
	ULONG ulAttachId = 0;

	/*
	 * Convert object id into attachment id
	 */
	er = GetSingleInstanceId(ulObjId, &ulAttachId);
	if (er != erSuccess)
		return false;

	return ExistAttachmentInstance(ulAttachId);
}

ECRESULT ECAttachmentStorage::LoadAttachment(struct soap *soap, ULONG ulObjId, int *lpiSize, unsigned char **lppData)
{
	ECRESULT er = erSuccess;
	ULONG ulAttachId = 0;

	/*
	 * Convert object id into attachment id
	 */
	er = GetSingleInstanceId(ulObjId, &ulAttachId);
	if (er != erSuccess)
		goto exit;

	er = LoadAttachmentInstance(soap, ulAttachId, lpiSize, lppData);
	if (er != erSuccess)
		goto exit;

exit:
	return er;
}

ECRESULT ECAttachmentStorage::SaveAttachment(ULONG ulObjId, bool bDeleteOld, int iSize, unsigned char *lpData, ULONG *lpulInstanceId)
{
	ECRESULT er = erSuccess;
	ULONG ulAttachId = 0;
	std::string strQuery;

	if (!lpData) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (bDeleteOld) {
		/*
		 * Call DeleteAttachment to decrease the refcount
		 * and optionally delete the original attachment.
		 */
		er = DeleteAttachment(ulObjId);
		if (er != erSuccess)
			goto exit;
	}

	/*
	 * Create Attachment reference, detect new attachment id.
	 */
	strQuery =
		"INSERT INTO `singleinstances` (`hierarchyid`, `tag`) VALUES"
		"(" + stringify(ulObjId) + ", " + stringify(PROP_ID(PR_ATTACH_DATA_BIN)) + ")";
	er = m_lpDatabase->DoInsert(strQuery, (unsigned int*)&ulAttachId);
	if (er != erSuccess)
		goto exit;

	er = SaveAttachmentInstance(ulAttachId, iSize, lpData);
	if (er != erSuccess)
		goto exit;

	if (lpulInstanceId)
		*lpulInstanceId = ulAttachId;

exit:
	return er;
}

ECRESULT ECAttachmentStorage::SaveAttachment(ULONG ulObjId, bool bDeleteOld, ULONG ulAttachId, ULONG *lpulInstanceId)
{
	ECRESULT er = erSuccess;
	ULONG ulOldAttachId = 0;

	if (bDeleteOld) {
		/*
		 * Call DeleteAttachment to decrease the refcount
		 * and optionally delete the original attachment.
		 */
		if(GetSingleInstanceId(ulObjId, &ulOldAttachId) == erSuccess && ulOldAttachId == ulAttachId) {
			// Nothing to do, we already have that instance ID
			goto exit;
		}
		 
		er = DeleteAttachment(ulObjId);
		if (er != erSuccess)
			goto exit;

	}

	er = SetSingleInstanceId(ulObjId, ulAttachId, PROP_ID(PR_ATTACH_DATA_BIN));
	if (er != erSuccess)
		goto exit;

	/* InstanceId is equal to provided AttachId */
	*lpulInstanceId = ulAttachId;

exit:
	return er;
}

ECRESULT ECAttachmentStorage::CopyAttachment(ULONG ulObjId, ULONG ulNewObjId)
{
	ECRESULT er = erSuccess;
	std::string strQuery;

	/*
	 * Only update the reference count in the `singleinstances` table,
	 * no need to really physically store the attachment twice.
	 */
	strQuery =
		"INSERT INTO `singleinstances` (`hierarchyid`, `instanceid`, `tag`) "
			"SELECT " + stringify(ulNewObjId) + ", `instanceid`, `tag` "
			"FROM `singleinstances` "
			"WHERE `hierarchyid` = " + stringify(ulObjId) + " "
			"LIMIT 1";

	er = m_lpDatabase->DoInsert(strQuery);
	if (er != erSuccess)
		goto exit;

exit:
	return er;
}

ECRESULT ECAttachmentStorage::DeleteAttachments(std::list<ULONG> lstDeleteObjects)
{
	ECRESULT er = erSuccess;
	std::list<ULONG> lstAttachments;
	std::list<ULONG> lstDeleteAttach;
	std::string strQuery;

	/*
	 * Convert object ids into attachment ids
	 * NOTE: lstDeleteObjects.size() >= lstAttachments.size()
	 * because the list does not 100% consists of attachments id and the
	 * duplicate attachment ids will be filtered out.
	 */
	er = GetSingleInstanceIds(lstDeleteObjects, &lstAttachments);
	if (er != erSuccess)
		goto exit;

	/* No attachments present, we're done */
	if (lstAttachments.empty())
		goto exit;

	/*
	 * Remove all objects from `singleinstances` table this will decrease the
	 * reference count for each attachment.
	 */
	strQuery =
		"DELETE FROM `singleinstances` "
		"WHERE `hierarchyid` IN (";
	for (std::list<ULONG>::iterator i = lstDeleteObjects.begin(); i != lstDeleteObjects.end(); i++) {
		if (i != lstDeleteObjects.begin())
			strQuery += ",";
		strQuery += stringify(*i);
	}
	strQuery += ")";

	er = m_lpDatabase->DoDelete(strQuery);
	if (er != erSuccess)
		goto exit;

	/*
	 * Get the list of orphaned Single Instance IDs which we can delete.
	 */
	er = GetOrphanedSingleInstances(lstAttachments, &lstDeleteAttach);
	if (er != erSuccess)
		goto exit;

	if (!lstDeleteAttach.empty()) {
		er = DeleteAttachmentInstances(lstDeleteAttach);
		if (er != erSuccess)
			goto exit;
	}

exit:
	return er;
}

ECRESULT ECAttachmentStorage::DeleteAttachment(ULONG ulObjId)
{
	ECRESULT er = erSuccess;
	ULONG ulAttachId = 0;
	bool bOrphan = false;
	std::string strQuery;

	/*
	 * Convert object id into attachment id
	 */
	er = GetSingleInstanceId(ulObjId, &ulAttachId);
	if (er != erSuccess)
		goto exit;

	/*
	 * Remove object from `singleinstances` table, this will decrease the
	 * reference count for the attachment.
	 */
	strQuery =
		"DELETE FROM `singleinstances` "
		"WHERE `hierarchyid` = " + stringify(ulObjId) + " "
		"LIMIT 1";

	er = m_lpDatabase->DoDelete(strQuery);
	if (er != erSuccess)
		goto exit;

	/*
	 * Check if the attachment can be permanently deleted.
	 */
	if ((IsOrphanedSingleInstance(ulAttachId, &bOrphan) == erSuccess) && (bOrphan == true)) {
		er = DeleteAttachmentInstance(ulAttachId);
		if (er != erSuccess)
			goto exit;
	}

exit:
	return er;
}

ECRESULT ECAttachmentStorage::GetSize(ULONG ulObjId, ULONG *lpulSize)
{
	ECRESULT er = erSuccess;
	ULONG ulAttachId = 0;

	*lpulSize = 0;

	/*
	 * Convert object id into attachment id
	 */
	er = GetSingleInstanceId(ulObjId, &ulAttachId);
	if (er == ZARAFA_E_NOT_FOUND) {
		er = erSuccess;
		goto exit;
	} else if (er != erSuccess) {
		goto exit;
	}

	er = GetSizeInstance(ulAttachId, lpulSize);
	if (er != erSuccess)
		goto exit;

exit:
	return er;
}

// Attachment storage is in database
ECDatabaseAttachment::ECDatabaseAttachment(ECDatabase *lpDatabase)
	: ECAttachmentStorage(lpDatabase, false, 0)
{
}

ECDatabaseAttachment::~ECDatabaseAttachment()
{
}

bool ECDatabaseAttachment::ExistAttachmentInstance(ULONG ulAttachId)
{
	ECRESULT er = erSuccess;
	std::string strQuery;
	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;

	strQuery = "SELECT instanceid FROM lob WHERE instanceid = " + stringify(ulAttachId) + " LIMIT 1";

	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpDBRow = m_lpDatabase->FetchRow(lpDBResult);

	if (!lpDBRow || !lpDBRow[0])
		er = ZARAFA_E_NOT_FOUND;

exit:
	if (lpDBResult)
		m_lpDatabase->FreeResult(lpDBResult);

	return er == erSuccess;
}

ECRESULT ECDatabaseAttachment::LoadAttachmentInstance(struct soap *soap, ULONG ulAttachId, int *lpiSize, unsigned char **lppData)
{
	ECRESULT er = erSuccess;
	int iSize = 0;
	int iReadSize = 0;
	unsigned char *lpData = NULL;
	std::string strQuery;
	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;
	DB_LENGTHS lpDBLen = NULL;


	// we first need to know the complete size of the attachment (some old databases don't have the correct chunk size)
	strQuery = "SELECT SUM(LENGTH(val_binary)) FROM lob WHERE instanceid = " + stringify(ulAttachId) + " AND tag = " + stringify(PROP_ID(PR_ATTACH_DATA_BIN));
	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpDBRow = m_lpDatabase->FetchRow(lpDBResult);
	if (lpDBRow == NULL || lpDBRow[0] == NULL) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	iSize = atoi(lpDBRow[0]);

	m_lpDatabase->FreeResult(lpDBResult);
	lpDBResult = NULL;

	// get all chunks
	strQuery = "SELECT val_binary FROM lob WHERE instanceid = " + stringify(ulAttachId) + " AND tag = " + stringify(PROP_ID(PR_ATTACH_DATA_BIN)) + " ORDER BY chunkid";

	er = m_lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpData = s_alloc<unsigned char>(soap, iSize);

	while ((lpDBRow = m_lpDatabase->FetchRow(lpDBResult))) {
		if (lpDBRow[0] == NULL) {
			// broken attachment !
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		lpDBLen = m_lpDatabase->FetchRowLengths(lpDBResult);

		memcpy(lpData + iReadSize, lpDBRow[0], lpDBLen[0]);
		iReadSize += lpDBLen[0];
	}

	*lpiSize = iReadSize;
	*lppData = lpData;

exit:
	if (er != erSuccess && lpData && !soap)
		delete lpData;

	if (lpDBResult)
		m_lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECDatabaseAttachment::SaveAttachmentInstance(ULONG ulAttachId, int iSize, unsigned char *lpData)
{
	ECRESULT er = erSuccess;
	std::string strQuery;
	int	iChunkSize;
	int	iSizeLeft;
	int	iPtr;
	unsigned int ulChunk;

	// make chunks of 393216 bytes (384*1024)
	iSizeLeft = iSize;
	iPtr = 0;
	ulChunk = 0;

	while (iSizeLeft > 0) {
		iChunkSize = iSizeLeft < CHUNK_SIZE ? iSizeLeft : CHUNK_SIZE;

		strQuery = (std::string)"INSERT INTO lob (instanceid, chunkid, tag, val_binary) VALUES (" +
			stringify(ulAttachId) + ", " + stringify(ulChunk) + ", " + stringify(PROP_ID(PR_ATTACH_DATA_BIN)) +
			", " + m_lpDatabase->EscapeBinary(lpData + iPtr, iChunkSize) + ")";

		er = m_lpDatabase->DoInsert(strQuery);
		if (er != erSuccess)
			goto exit;

		ulChunk++;
		iSizeLeft -= iChunkSize;
		iPtr += iChunkSize;
	}

exit:
	return er;
}

ECRESULT ECDatabaseAttachment::CopyAttachmentInstance(ULONG ulAttachId, ULONG ulNewAttachId)
{
	ECRESULT er = erSuccess;
	string strQuery;

	strQuery = "INSERT INTO lob (instanceid, tag, chunkid, val_binary) SELECT "+stringify(ulNewAttachId)+", tag, chunkid, val_binary FROM lob WHERE instanceid ="+stringify(ulAttachId);
	er = m_lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

exit:
	return er;
}

ECRESULT ECDatabaseAttachment::DeleteAttachmentInstances(std::list<ULONG> &lstDeleteAttachments)
{
	ECRESULT er = erSuccess;
	std::list<ULONG>::iterator iterDel;
	std::string strQuery;

	strQuery = (std::string)"DELETE FROM lob WHERE tag=" + stringify(PROP_ID(PR_ATTACH_DATA_BIN)) + " AND instanceid IN (";

	for (iterDel = lstDeleteAttachments.begin(); iterDel != lstDeleteAttachments.end(); iterDel++) {
		if (iterDel != lstDeleteAttachments.begin())
			strQuery += ",";
		strQuery += stringify(*iterDel);
	}

	strQuery += ")";

	er = m_lpDatabase->DoDelete(strQuery);

	return er;
}

ECRESULT ECDatabaseAttachment::DeleteAttachmentInstance(ULONG ulAttachId)
{
	ECRESULT er = erSuccess;
	std::string strQuery;

	strQuery = (std::string)"DELETE FROM lob WHERE instanceid=" + stringify(ulAttachId) + " AND tag=" + stringify(PROP_ID(PR_ATTACH_DATA_BIN));

	er = m_lpDatabase->DoDelete(strQuery);

	return er;
}

ECRESULT ECDatabaseAttachment::GetSizeInstance(ULONG ulAttachId, ULONG *lpulSize, bool *lpbCompressed)
{
	// this function is not called for database objects.
	// when we do, implement this correctly
	return ZARAFA_E_NOT_IMPLEMENTED;
}

// Attachment storage is in separate files
ECFileAttachment::ECFileAttachment(ECDatabase *lpDatabase, std::string basepath, unsigned int ulCompressionLevel)
	: ECAttachmentStorage(lpDatabase, true, ulCompressionLevel)
{
	m_basepath = basepath;
	if (m_basepath.empty())
		m_basepath = "/var/lib/zarafa";
}

ECFileAttachment::~ECFileAttachment()
{
}

bool ECFileAttachment::ExistAttachmentInstance(ULONG ulAttachId)
{
	ECRESULT er = erSuccess;
	string filename = CreateAttachmentFilename(ulAttachId, m_bFileCompression);
	FILE *f = NULL;

	f = fopen(filename.c_str(), "r");
	if (!f) {
		filename = CreateAttachmentFilename(ulAttachId, !m_bFileCompression);
		f = fopen(filename.c_str(), "r");
	}
	if (!f) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	fclose(f);

exit:
	return er == erSuccess;
}

// TODO: add zlib compression to files.
// best: make gzip compatible files!
ECRESULT ECFileAttachment::LoadAttachmentInstance(struct soap *soap, ULONG ulAttachId, int *lpiSize, unsigned char **lppData)
{
	ECRESULT er = erSuccess;
	string filename;
	FILE *fp = NULL;
	gzFile gzfp = NULL;
	unsigned char *lpData = NULL;
	ULONG ulSize = 0;
	ULONG ulReadSize = 0;
	bool bCompressed;

	/*
	 * First we need to determine the (uncompressed) filesize
	 */
	er = GetSizeInstance(ulAttachId, &ulSize, &bCompressed);
	if (er != erSuccess)
		goto exit;

	filename = CreateAttachmentFilename(ulAttachId, bCompressed);
	lpData = s_alloc<unsigned char>(soap, ulSize);

	/*
	 * CreateAttachmentFilename Already checked if we are working with an compressed or uncompressed file,
	 * no need to perform retries when our first guess (which is based on CreateAttachmentFilename) fails.
	 */
	if (bCompressed) {
		/* Compressed attachment */
		gzfp = gzopen(filename.c_str(), "rb");
		if (!gzfp) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		ulReadSize = gzread(gzfp, lpData, ulSize);
		if (ulSize != ulReadSize) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}
	} else {
		/* Uncompressed attachment */
		fp = fopen(filename.c_str(), "r");
		if (!fp) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		ulReadSize= fread(lpData, 1, ulSize, fp);
		if (ulSize != ulReadSize) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}
	}

	*lpiSize = ulSize;
	*lppData = lpData;

exit:
	if (fp)
		fclose(fp);
	if (gzfp)
		gzclose(gzfp);

	return er;
}

ECRESULT ECFileAttachment::SaveAttachmentInstance(ULONG ulAttachId, int iSize, unsigned char *lpData)
{
	ECRESULT er = erSuccess;
	string filename = CreateAttachmentFilename(ulAttachId, m_bFileCompression && iSize);

	// no need to remove the file, just overwrite it
	if (m_bFileCompression && iSize) {
		gzFile gzfp = gzopen(filename.c_str(), std::string("wb" + m_CompressionLevel).c_str());
		if (!gzfp) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}
		gzwrite(gzfp, lpData, iSize);
		gzclose(gzfp);
	} else {
		FILE *f = fopen(filename.c_str(), "w");
		if (!f) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}
		fwrite(lpData, 1, iSize, f);
		fclose(f);
	}

exit:
	return er;
}

#define BLOCK_SIZE 65536
ECRESULT ECFileAttachment::CopyAttachmentInstance(ULONG ulAttachId, ULONG ulNewAttachId)
{
	ECRESULT er = erSuccess;
	string strOldFilename = CreateAttachmentFilename(ulAttachId, m_bFileCompression);
	string strNewFilename;
	FILE *fOld = NULL;
	FILE *fNew = NULL;
	char *buffer = new char[BLOCK_SIZE];
	size_t ulRead, ulWrite;
	bool bCompressed = m_bFileCompression;

	/* Detect if the old attachment was compressed or not */
	fOld = fopen(strOldFilename.c_str(), "r");
	if (!fOld) {
		strOldFilename = CreateAttachmentFilename(ulAttachId, !m_bFileCompression);
		bCompressed = !m_bFileCompression;
		fOld = fopen(strOldFilename.c_str(), "r");
	}
	if (!fOld) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	strNewFilename = CreateAttachmentFilename(ulNewAttachId, bCompressed);
	fNew = fopen(strNewFilename.c_str(), "w");
	if (!fNew) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	/*
	 * Regardless of current compression type,
	 * we will copy the attachment *exactly* how we found it.
	 */
	while ((ulRead = fread(buffer, 1, BLOCK_SIZE, fOld)) ) {
		ulWrite = fwrite(buffer, 1, ulRead, fNew);
		if (ulRead != ulWrite) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}
	}

exit:
	if (er == ZARAFA_E_DATABASE_ERROR)
		unlink(strNewFilename.c_str()); // "rollback"

	if (buffer)
		delete [] buffer;

	if (fOld)
		fclose(fOld);

	if (fNew)
		fclose(fNew);

	return er;
}
#undef BLOCK_SIZE

ECRESULT ECFileAttachment::DeleteAttachmentInstances(std::list<ULONG> &lstDeleteAttachments)
{
	ECRESULT er = erSuccess;
	int errors = 0;
	std::list<ULONG>::iterator iterDel;

	for (iterDel = lstDeleteAttachments.begin(); iterDel != lstDeleteAttachments.end(); iterDel++) {
		er = this->DeleteAttachmentInstance(*iterDel);
		if (er != erSuccess)
			errors++;
	}

	return errors == 0 ? erSuccess : ZARAFA_E_DATABASE_ERROR;
}

ECRESULT ECFileAttachment::DeleteAttachmentInstance(ULONG ulAttachId)
{
	ECRESULT er = erSuccess;
	string filename = CreateAttachmentFilename(ulAttachId, m_bFileCompression);

	if (unlink(filename.c_str()) != 0) {
		if (errno == ENOENT){
			filename = CreateAttachmentFilename(ulAttachId, !m_bFileCompression);
			if (unlink(filename.c_str()) == 0)
				goto exit; /* Success! */
		}

		if (errno == EACCES || errno == EPERM)
			er = ZARAFA_E_NO_ACCESS;
		else if (errno != ENOENT) // ignore "file not found" error
			er = ZARAFA_E_DATABASE_ERROR;
	}

exit:
	return er;
}

std::string ECFileAttachment::CreateAttachmentFilename(ULONG ulAttachId, bool bCompressed)
{
	string filename;
	unsigned int l1, l2;

	l1 = ulAttachId % ATTACH_PATHDEPTH_LEVEL1;
	l2 = (ulAttachId / ATTACH_PATHDEPTH_LEVEL1) % ATTACH_PATHDEPTH_LEVEL2;

	filename = m_basepath + PATH_SEPARATOR + stringify(l1) + PATH_SEPARATOR + stringify(l2) + PATH_SEPARATOR + stringify(ulAttachId);

	if (bCompressed)
		filename += ".gz";

	return filename;
}

ECRESULT ECFileAttachment::GetSizeInstance(ULONG ulAttachId, ULONG *lpulSize, bool *lpbCompressed)
{
	ECRESULT er = erSuccess;
	string filename = CreateAttachmentFilename(ulAttachId, m_bFileCompression);
	FILE *f = NULL;
	ULONG ulSize = 0;
	bool bCompressed = m_bFileCompression;

	/*
	 * We are always going to use the normal FILE handler for determining the file size,
	 * the gzFile handler is broken since it doesn't support SEEK_END and gzseek itself
	 * is very slow. When the attachment has been zipped, we are going to read the
	 * last 4 bytes of the file, that contains the uncompressed filesize.
	 *
	 * For uncompressed files we can still use the FILE pointer and use SEEK_END
	 * which is the most simple solution.
	 */
	f = fopen(filename.c_str(), "r");
	if (!f) {
		filename = CreateAttachmentFilename(ulAttachId, !m_bFileCompression);
		bCompressed = !m_bFileCompression;
		f = fopen(filename.c_str(), "r");
	}
	if (!f) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	if (!bCompressed) {
		/* Uncompressed attachment */
		fseek(f, 0, SEEK_END);
		ulSize = ftell(f);
	} else {
		/* Compressed attachment */
		fseek(f, -4, SEEK_END);
		fread(&ulSize, 1, 4, f);
	}

	fclose(f);

	*lpulSize = ulSize;
	if (lpbCompressed)
		*lpbCompressed = bCompressed;

exit:
	return er;
}
