/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

//
//////////////////////////////////////////////////////////////////////
#include "platform.h"

#include <mapidefs.h>
#include <mapitags.h>

#include "ECDatabaseMySQL.h"
#include "ECSessionManager.h"
#include "ECDatabaseUtils.h"

#include "ECCacheManager.h"

#include "ECMAPI.h"
#include "stringutil.h"
#include "ECGenericObjectTable.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include <algorithm>

// Our cache-purging function designed to work with any std::map

template<typename Key>
	class KeyEntry {
	public:
		Key key;
		time_t ulLastAccess;
	};

template<typename Key>
	bool KeyEntryOrder(const KeyEntry<Key> &a, const KeyEntry<Key> &b) {
		return a.ulLastAccess < b.ulLastAccess;
	}

template<typename Value>
	unsigned int GetCacheAdditionalSize(const Value &val) {
		return 0;
	}

// Specialization for ECsACL
template<>
    unsigned int GetCacheAdditionalSize(const ECsACLs &val) {
        return val.ulACLs * sizeof(val.aACL[0]);
    }

template<>
    unsigned int GetCacheAdditionalSize(const ECsIndexProp &val) {
        return val.cbData;
    }

// Specialization for ECsCell
template<>
    unsigned int GetCacheAdditionalSize(const ECsCells &val) {
        return val.GetSize();
    }

// Removes an item from a cache
template<typename T>
ECRESULT RemoveCacheItem(T *lpMap, typename T::key_type key, unsigned long long *lpulAdditionalSize, pthread_mutex_t *hLock ) {
	ECRESULT er = erSuccess;
	typename T::iterator iter;

	pthread_mutex_lock(hLock);
	iter = lpMap->find(key);

	if(iter == lpMap->end()) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	*lpulAdditionalSize -= GetCacheAdditionalSize((*lpMap)[key]);
	lpMap->erase(iter);

exit:
	pthread_mutex_unlock(hLock);

	return er;
}


// Gets a cache item
// GetCacheItem(ECMapCells*, ECsSortKeyKey&, ECsCells*, int, pthread_mutex_t*)
template<typename T, typename Key, typename Data>
ECRESULT GetCacheItem(T *lpMap, Key key, Data **value, unsigned int ulMaxAge, pthread_mutex_t *hLock) {
	ECRESULT er = erSuccess;
	time_t	tNow;
	typename T::iterator iter;
	typename T::iterator iterDelete;

	iter = lpMap->find(key);

	if(iter != lpMap->end()) {
		tNow = GetProcessTime();
		// Cache age of the cached item, if expired remove the item from the cache
		if(ulMaxAge != 0 && (unsigned int)(tNow - iter->second.ulLastAccess) >= ulMaxAge) {

			// Loop through all items and check
			for(iter = lpMap->begin(); iter != lpMap->end();)
			{
				if((unsigned int)(tNow - iter->second.ulLastAccess) >= ulMaxAge)
				{
					iterDelete = iter;
					iter++;
					lpMap->erase(iterDelete);
				}else {
					iter++;
				}
			}
			er = ZARAFA_E_NOT_FOUND;
		}else {
			*value = &iter->second;
			// If we have an aging cache, we don't update the timestamp,
			// so we can't keep a value longer in the cache than the max age.
			// If we have a non-aging cache, we need to update it,
			// to see the oldest 5% to purge from the cache.
			if (ulMaxAge == 0)
				iter->second.ulLastAccess = tNow;
			er = erSuccess;
		}
	} else {
		er = ZARAFA_E_NOT_FOUND;
	}

	return er;
}

// Purges ratio percent of the given cache. Updates lpulAdditionalSize if more data was allocated for an entry
template<typename T>
ECRESULT PurgeCache(float ratio, T *lpMap, unsigned long long *lpulAdditionalSize, pthread_mutex_t *hLock) {
	std::vector<KeyEntry<typename T::key_type> > lstEntries;
	typename T::iterator iterMap;

	pthread_mutex_lock(hLock);

	for(iterMap = lpMap->begin(); iterMap != lpMap->end(); iterMap++) {
		KeyEntry<typename T::key_type> k;
		k.key = iterMap->first;
		k.ulLastAccess = iterMap->second.ulLastAccess;

		lstEntries.push_back(k);
	}

	sort(lstEntries.begin(), lstEntries.end(), KeyEntryOrder<typename T::key_type>);

	// We now have a list of all cache items, sorted by access time, (oldest first)
	unsigned int ulDelete = (unsigned int)(lpMap->size() * ratio);

	// Remove the oldest ulDelete entries from the cache, removing [ratio] % of all
	// cache entries.
	for(unsigned int i=0; i < ulDelete ; i++) {
		if(lpulAdditionalSize)
			*lpulAdditionalSize -= GetCacheAdditionalSize((*lpMap)[lstEntries[i].key]);

		lpMap->erase(lstEntries[i].key);
	}

	pthread_mutex_unlock(hLock);

	return erSuccess;
}

// Checks the size of the cache for over-allocation, removed 5% of items if over size. Constant
// time function, so can be called often.
template<typename T>
ECRESULT UpdateCache(float ratio, T *lpMap, unsigned long long * lpulAdditionalSize, unsigned long long ulMaxSize, pthread_mutex_t *hLock) {
	if((unsigned long long)lpMap->size() * sizeof(typename T::value_type) + *lpulAdditionalSize > ulMaxSize) {
		PurgeCache(ratio, lpMap, lpulAdditionalSize, hLock);
	}

	return erSuccess;
}

// Adds an item to a cache
template<typename Key, typename Value, typename LessFunc>
ECRESULT AddCacheItem(std::map<Key, Value, LessFunc> *lpMap, Key key, Value value, unsigned long long *lpulAdditionalSize, unsigned long long ulMaxSize, pthread_mutex_t *hLock) {
	typename std::map<Key, Value, LessFunc>::iterator iter;
	LessFunc less;

	pthread_mutex_lock(hLock);

	iter = lpMap->lower_bound(key);

	// lower_bound means *iter >= key
	if(iter != lpMap->end() && !less(key,iter->first)) {
	    // !(key < *iter) && (*iter >= key)
	    // (key >= *iter) && (*iter >= key)
	    // (*iter <= key) && (*iter >= key)
	    // key == *iter
	    *lpulAdditionalSize += (int)(GetCacheAdditionalSize(value) - GetCacheAdditionalSize(iter->second));
	    iter->second = value;
	    iter->second.ulLastAccess = GetProcessTime();
	    // Since there is a very small chance that we need to purge the cache, we're skipping that here.
	} else {
        value.ulLastAccess = GetProcessTime();

        lpMap->insert(iter, typename std::map<Key, Value, LessFunc>::value_type (key, value));
        
        *lpulAdditionalSize += GetCacheAdditionalSize(value);
        
        UpdateCache((float)0.05, lpMap, lpulAdditionalSize, ulMaxSize, hLock);
    }
    	
	pthread_mutex_unlock(hLock);	
	
	return erSuccess;
}


// Adds an item to a hash-based cache
template<typename Key, typename Value, typename HashFunc, typename EqualFunc>
ECRESULT AddCacheItem(sgi::hash_map<Key, Value, HashFunc, EqualFunc> *lpMap, Key key, Value value, unsigned long long *lpulAdditionalSize, unsigned long long ulMaxSize, pthread_mutex_t *hLock) {
	typename sgi::hash_map<Key, Value, HashFunc, EqualFunc>::iterator iter;

	pthread_mutex_lock(hLock);
	
	iter = lpMap->find(key);

	if(iter != lpMap->end()) {
	    *lpulAdditionalSize += (int)(GetCacheAdditionalSize(value) - GetCacheAdditionalSize(iter->second));
	    iter->second = value;
	    iter->second.ulLastAccess = GetProcessTime();
	    // Since there is a very small chance that we need to purge the cache, we're skipping that here.
	} else {
        value.ulLastAccess = GetProcessTime();

        lpMap->insert(typename sgi::hash_map<Key, Value, HashFunc, EqualFunc>::value_type (key, value));
        
        *lpulAdditionalSize += GetCacheAdditionalSize(value);
        
        UpdateCache((float)0.05, lpMap, lpulAdditionalSize, ulMaxSize, hLock);
    }
    	
	pthread_mutex_unlock(hLock);	
	
	return erSuccess;
}



ECCacheManager::ECCacheManager(ECConfig *lpConfig, ECDatabaseFactory *lpDatabaseFactory, ECLogger *lpLogger)
{
	pthread_mutexattr_t mattr;
	pthread_mutexattr_init(&mattr);
	pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	pthread_mutex_init(&m_hCacheMutex, &mattr);

	/* Initial cleaning/initialization of cache */
	PurgeCache();

	/* Initialization of constants */
	m_lpDatabaseFactory = lpDatabaseFactory;
	m_ulMaxQuotaAge = atoi(lpConfig->GetSetting("cache_quota_lifetime")) * 60;
	m_ulMaxUserDetailsAge = atoi(lpConfig->GetSetting("cache_userdetails_lifetime")) * 60;
	m_lpLogger = lpLogger;

	m_ulCacheObjectMaxSize = atoll(lpConfig->GetSetting("cache_object_size"));
	m_ulCacheIndexedObjectMaxSize = atoll(lpConfig->GetSetting("cache_indexedobject_size"));
	m_ulCacheCellMaxSize = atoll(lpConfig->GetSetting("cache_cell_size"));
}

ECCacheManager::~ECCacheManager()
{
}

ECRESULT ECCacheManager::PurgeCache()
{
	ECRESULT er = erSuccess;

	pthread_mutex_lock(&m_hCacheMutex);

	m_mapQuota.clear();
	m_ulCacheQuotaSize = 0;
	m_mapQuotaUserDefault.clear();
	m_ulCacheQuotaUserDefaultSize = 0;

	m_mapObjects.clear();
	m_ulCacheObjectSize = 0;

	m_mapStores.clear();
	m_ulCacheStoreSize = 0;

	m_mapACLs.clear();
	m_ulCacheACLSize = 0;

	m_mapCells.clear();
	m_ulCacheCellSize = 0;

	m_mapPropToObject.clear();
	m_mapObjectToProp.clear();

    m_ulCachePropToObjectSize = 0;
	m_ulCacheObjectToPropSize = 0;
	m_ulCellCacheHit = 0;
	m_ulCellCacheValid = 0;
	m_ulObjCacheHit = 0;
	m_ulObjCacheValid = 0;
	m_ulACLCacheHit = 0;
	m_ulACLCacheValid = 0;
	m_ulPropToObjectCacheHit = 0;
	m_ulPropToObjectCacheValid = 0;
	m_ulObjectToPropCacheHit = 0;
	m_ulObjectToPropCacheValid = 0;
	m_ulStoreCacheHit = 0;
	m_ulStoreCacheValid = 0;
	
	pthread_mutex_unlock(&m_hCacheMutex);

	er = PurgeUserCache();

	return er;
}

ECRESULT ECCacheManager::PurgeUserCache()
{
	ECRESULT er = erSuccess;

	pthread_mutex_lock(&m_hCacheMutex);

	m_mapUserObject.clear();
	m_ulCacheUserObjectSize = 0;

	m_mapUEIdObject.clear();
	m_ulCacheUEIdObjectSize = 0;

	m_mapUserObjectDetails.clear();
	m_ulCacheUserObjectDetailsSize = 0;

	m_mapServerDetails.clear();
	m_ulServerDetailsSize = 0;

	m_ulUserCacheHit = 0;
	m_ulUserCacheValid = 0;

	m_ulUserDetailsCacheHit = 0;
	m_ulUserDetailsCacheValid = 0;
	
	pthread_mutex_unlock(&m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::Update(unsigned int ulType, unsigned int ulObjId)
{
	ECRESULT		er = erSuccess;

	switch(ulType)
	{
		case fnevObjectModified:
			_DelACLs(ulObjId);
			_DelCell(ulObjId);
			_DelObject(ulObjId);
			break;
		case fnevObjectDeleted:
			_DelObject(ulObjId);
			_DelStore(ulObjId);
			_DelACLs(ulObjId);
			_DelCell(ulObjId);
			break;
		case fnevObjectMoved:
			_DelStore(ulObjId);
			_DelObject(ulObjId);
			_DelCell(ulObjId);
			break;
		default:
			//Do nothing
			break;
	}

	return er;
}

ECRESULT ECCacheManager::UpdateUser(unsigned int ulUserId)
{
	std::string strExternId;
	userobject_type_t ulType;

	if (_GetUserObject(ulUserId, &ulType, NULL, &strExternId, NULL) == erSuccess)
		_DelUEIdObject(strExternId, ulType);

	_DelUserObject(ulUserId);
	_DelUserObjectDetails(ulUserId);
	_DelQuota(ulUserId, false);
	_DelQuota(ulUserId, true);

	return erSuccess;
}

ECRESULT ECCacheManager::_GetObject(unsigned int ulObjId, unsigned int *ulParent, unsigned int *ulOwner, unsigned int *ulFlags, unsigned int *ulType)
{
	ECRESULT	er = erSuccess;
	ECsObjects	*sObject;

	m_ulObjCacheHit++;

	pthread_mutex_lock(&m_hCacheMutex);

	er = GetCacheItem(&m_mapObjects, ulObjId, &sObject, 0, &m_hCacheMutex);	
	
	if(er != erSuccess)
		goto exit;

	if(ulParent)
		*ulParent = sObject->ulParent;

	if(ulOwner)
		*ulOwner = sObject->ulOwner;

	if(ulFlags)
		*ulFlags = sObject->ulFlags;

	if(ulType)
		*ulType = sObject->ulType;

	m_ulObjCacheValid++;
exit:
	pthread_mutex_unlock(&m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::SetObject(unsigned int ulObjId, unsigned int ulParent, unsigned int ulOwner, unsigned int ulFlags, unsigned int ulType)
{
	ECRESULT		er = erSuccess;
	ECsObjects		sObjects;

	if(ulParent == 0 || ulObjId == 0 || ulOwner == 0)
		return 1;

	sObjects.ulParent	= ulParent;
	sObjects.ulOwner	= ulOwner;
	sObjects.ulFlags	= ulFlags;
	sObjects.ulType		= ulType;

	er = AddCacheItem(&m_mapObjects, ulObjId, sObjects, &m_ulCacheObjectSize, m_ulCacheObjectMaxSize, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_DelObject(unsigned int ulObjId)
{
	ECRESULT		er = erSuccess;

	er = RemoveCacheItem(&m_mapObjects, ulObjId, &m_ulCacheObjectSize, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_GetStore(unsigned int ulObjId, unsigned int *ulStore, GUID *lpGuid)
{
	ECRESULT	er = erSuccess;
	ECsStores	*sStores;

	m_ulStoreCacheHit++;
	
	pthread_mutex_lock(&m_hCacheMutex);

	er = GetCacheItem(&m_mapStores, ulObjId, &sStores, 0, &m_hCacheMutex);

	if(er != erSuccess)
		goto exit;

	if(ulStore)
		*ulStore = sStores->ulStore;

	if(lpGuid)
		memcpy(lpGuid, &sStores->guidStore, sizeof(GUID) );
		
    m_ulStoreCacheValid++;
    
exit:
	pthread_mutex_unlock(&m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::SetStore(unsigned int ulObjId, unsigned int ulStore, GUID *lpGuid)
{
	ECRESULT		er = erSuccess;
	ECsStores		sStores;
	sStores.ulStore = ulStore;
	sStores.guidStore = *lpGuid;

	er = AddCacheItem(&m_mapStores, ulObjId, sStores, &m_ulCacheStoreSize, STORE_CACHE_SIZE, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_DelStore(unsigned int ulObjId)
{
	ECRESULT		er = erSuccess;

	er = RemoveCacheItem(&m_mapStores, ulObjId, &m_ulCacheStoreSize, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::GetOwner(unsigned int ulObjId, unsigned int *ulOwner)
{
	ECRESULT	er = erSuccess;

	if(_GetObject(ulObjId, NULL, ulOwner, NULL, NULL) == erSuccess)
		goto exit;

	er = GetObject(ulObjId, NULL, ulOwner, NULL);

exit:
	return er;
}

ECRESULT ECCacheManager::GetParent(unsigned int ulObjId, unsigned int *lpulParent)
{
	ECRESULT er = erSuccess;
	unsigned int ulParent = 0;

	er = GetObject(ulObjId, &ulParent, NULL, NULL);

	if(er != erSuccess)
		goto exit;

	if(ulParent == CACHE_NO_PARENT) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	*lpulParent = ulParent;
exit:
	return er;
}

// Get the parent of the specified object
ECRESULT ECCacheManager::GetObject(unsigned int ulObjId, unsigned int *lpulParent, unsigned int *lpulOwner, unsigned int *lpulFlags, unsigned int *lpulType)
{
	ECRESULT	er = erSuccess;
	DB_RESULT	lpDBResult = NULL;
	DB_ROW		lpDBRow = NULL;
	std::string strQuery;
	ECDatabase	*lpDatabase = NULL;
	unsigned int	ulParent = 0, ulOwner = 0, ulFlags = 0, ulType = 0;

	er = GetThreadLocalDatabase(this->m_lpDatabaseFactory, &lpDatabase);

	if(er != erSuccess)
		goto exit;

	// first check the cache if the item exists
	if(_GetObject(ulObjId, &ulParent, &ulOwner, &ulFlags, &ulType) == erSuccess) {
		if(lpulParent)
			*lpulParent = ulParent;
		if(lpulOwner)
			*lpulOwner = ulOwner;
		if(lpulFlags)
			*lpulFlags = ulFlags;
		if(lpulType)
			*lpulType = ulType;
		goto exit;
	}

	strQuery = "SELECT hierarchy.parent, hierarchy.owner, hierarchy.flags, hierarchy.type FROM hierarchy WHERE hierarchy.id = " + stringify(ulObjId);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	lpDBRow = lpDatabase->FetchRow(lpDBResult);

	if(lpDBRow == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	if(lpDBRow[1] == NULL || lpDBRow[2] == NULL || lpDBRow[3] == NULL) {
		// owner or flags should not be NULL
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	ulParent = lpDBRow[0] == NULL ? CACHE_NO_PARENT : atoui(lpDBRow[0]);
	ulOwner = atoui(lpDBRow[1]);
	ulFlags = atoui(lpDBRow[2]);
	ulType = atoui(lpDBRow[3]);

	if(lpulParent)
		*lpulParent = ulParent;
	if(lpulOwner)
		*lpulOwner = ulOwner;
	if(lpulFlags)
		*lpulFlags = ulFlags;
	if(lpulType)
		*lpulType = ulType;

	SetObject(ulObjId, ulParent, ulOwner, ulFlags, ulType);

exit:
	if (lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;

}

// Get the store that the specified object belongs to
ECRESULT ECCacheManager::GetStore(unsigned int ulObjId, unsigned int *lpulStore, GUID *lpGuid, unsigned int maxdepth)
{
	ECRESULT	er = erSuccess;
	DB_RESULT	lpDBResult = NULL;
	DB_ROW		lpDBRow = NULL;
	std::string strQuery;
	ECDatabase	*lpDatabase = NULL;
	unsigned int ulSubObjId = 0;
	unsigned int ulStore = 0;
	GUID guid;
	
	if(maxdepth <= 0)
	    return ZARAFA_E_NOT_FOUND;

	er = GetThreadLocalDatabase(this->m_lpDatabaseFactory, &lpDatabase);

	if(er != erSuccess)
		goto exit;

	// first check the cache if we already know the store for this object
	if(_GetStore(ulObjId, &ulStore, &guid) == erSuccess)
		goto found;

    // Get our parent folder
	if(GetParent(ulObjId, &ulSubObjId) != erSuccess) {
	    // No parent, this must be the top-level item, get the store data from here
    	strQuery = "SELECT hierarchy_id, guid FROM stores WHERE hierarchy_id = " + stringify(ulObjId);
    	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
    	if(er != erSuccess)
	    	goto exit;

    	if(lpDatabase->GetNumRows(lpDBResult) < 1) {
    		er = ZARAFA_E_NOT_FOUND;
    		goto exit;
    	}

    	lpDBRow = lpDatabase->FetchRow(lpDBResult);

    	if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL) {
    		er = ZARAFA_E_DATABASE_ERROR;
    		goto exit;
    	}
    	ulStore = atoi(lpDBRow[0]);
        memcpy(&guid, lpDBRow[1], sizeof(GUID));


	} else {
	    // We have a parent, get the store for our parent by recursively calling ourselves
	    er = GetStore(ulSubObjId, &ulStore, &guid, maxdepth-1);
	    if(er != erSuccess)
	        goto exit;
    }

    // insert the item into the cache
    SetStore(ulObjId, ulStore, &guid);

found:    
    if(lpulStore)
        *lpulStore = ulStore;
    if(lpGuid)
        *lpGuid = guid;

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECCacheManager::GetUserObject(unsigned int ulUserId, userobject_type_t* lpulType, unsigned int *lpulCompanyId,
									   std::string *lpstrExtern, std::string *lpstrSignature)
{
	ECRESULT	er = erSuccess;
	DB_RESULT	lpDBResult = NULL;
	DB_ROW		lpDBRow = NULL;
	DB_LENGTHS	lpDBLen = NULL;
	std::string	strQuery;
	ECDatabase	*lpDatabase = NULL;
	userobject_type_t ulType;
	unsigned int ulCompanyId;
	std::string externid;
	std::string signature;

	// first check the cache if we already know the external id for this user
	if (_GetUserObject(ulUserId, lpulType, lpulCompanyId, lpstrExtern, lpstrSignature) == erSuccess) {
		goto exit;
	}

	er = GetThreadLocalDatabase(this->m_lpDatabaseFactory, &lpDatabase);
	if (er != erSuccess)
		goto exit;

	er = lpDatabase->DoSelect("SELECT externid, object_type, signature, company FROM users "
							  "WHERE id=" + stringify(ulUserId), &lpDBResult);
	if (er != erSuccess) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL || lpDBRow[3] == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	ulType = (userobject_type_t)atoui(lpDBRow[1]);
	ulCompanyId = atoui(lpDBRow[3]);

	externid.assign(lpDBRow[0], lpDBLen[0]);
	signature.assign(lpDBRow[2], lpDBLen[2]);

	// insert the item into the cache
	_AddUserObject(ulUserId, ulType, ulCompanyId, externid, signature);

	if(lpulType)
		*lpulType = ulType;

	if(lpulCompanyId)
		*lpulCompanyId = ulCompanyId;

	if(lpstrExtern)
		*lpstrExtern = externid;

	if(lpstrSignature)
		*lpstrSignature = signature;

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECCacheManager::GetUserExternId(unsigned int ulUserId, userobject_type_t *lpsType, std::string *lpstrExternId, std::string *lpstrSignature)
{
	ECRESULT		er = erSuccess;
	userobject_type_t sType;
	std::string		strExternId;
	std::string		strSignature;

	er = GetUserObject(ulUserId, &sType, NULL, &strExternId, &strSignature);
	if(er != erSuccess)
		goto exit;

	if (lpsType)
		*lpsType = sType;
	if (lpstrExternId)
		*lpstrExternId = strExternId;
	if (lpstrSignature)
		*lpstrSignature = strSignature;

exit:

	return er;
}

ECRESULT ECCacheManager::GetUserCompany(unsigned int ulObjectId, userobject_type_t ulType, unsigned int *lpulCompanyId)
{
	ECRESULT		er = erSuccess;
	userobject_type_t ulTmpType;
	unsigned int	ulCompanyId;

	er = GetUserObject(ulObjectId, &ulTmpType, &ulCompanyId, NULL, NULL);
	if(er != erSuccess)
		goto exit;

	if(ulTmpType != ulType) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	if (lpulCompanyId)
		*lpulCompanyId = ulCompanyId;

exit:
	return er;
}

ECRESULT ECCacheManager::GetUserDetails(unsigned int ulUserId, userobject_type_t ulType, objectdetails_t *details)
{
	ECRESULT er = erSuccess;

	er = _GetUserObjectDetails(ulUserId, ulType, details);

	// on error, ECUserManagement will update the cache

	return er;
}

ECRESULT ECCacheManager::SetUserDetails(unsigned int ulUserId, userobject_type_t ulType, objectdetails_t *details)
{
	ECRESULT er = erSuccess;

	er = _AddUserObjectDetails(ulUserId, ulType, details);

	return er;
}

ECRESULT ECCacheManager::GetUserObject(std::string strExternId, userobject_type_t ulType, unsigned int *lpulCompanyId,
									   unsigned int *lpulUserId, std::string *lpstrSignature)
{
	ECRESULT	er = erSuccess;
	DB_RESULT	lpDBResult = NULL;
	DB_ROW		lpDBRow = NULL;
	DB_LENGTHS	lpDBLen = NULL;
	std::string	strQuery;
	ECDatabase	*lpDatabase = NULL;
	unsigned int ulCompanyId;
	unsigned int ulUserId;
	std::string signature;

	// first check the cache if we already know the external id for this user
	if (_GetUEIdObject(strExternId, ulType, lpulCompanyId, lpulUserId, lpstrSignature) == erSuccess) {
		goto exit;
	}

	er = GetThreadLocalDatabase(this->m_lpDatabaseFactory, &lpDatabase);
	if (er != erSuccess)
		goto exit;

	er = lpDatabase->DoSelect("SELECT id, signature, company FROM users "
							  "WHERE externid='" + lpDatabase->Escape(strExternId) + "' AND object_type=" + stringify(ulType), &lpDBResult);
	if (er != erSuccess) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	ulUserId = atoui(lpDBRow[0]);
	signature.assign(lpDBRow[1], lpDBLen[1]);
	ulCompanyId = atoui(lpDBRow[2]);

	// insert the item into the cache
	_AddUEIdObject(strExternId, ulType, ulCompanyId, ulUserId, signature);

	if(lpulCompanyId)
		*lpulCompanyId = ulCompanyId;

	if(lpulUserId)
		*lpulUserId = ulUserId;

	if(lpstrSignature)
		*lpstrSignature = signature;

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECCacheManager::_AddUserObject(unsigned int ulUserId, userobject_type_t ulType, unsigned int ulCompanyId,
										std::string strExternId, std::string strSignature)
{
	ECRESULT	er = erSuccess;

	ECsUserObject sData;

	sData.ulType = ulType;
	sData.ulCompanyId = ulCompanyId;
	sData.strExternId = strExternId;
	sData.strSignature = strSignature;

	er = AddCacheItem(&m_mapUserObject, ulUserId, sData, &m_ulCacheUserObjectSize, USER_CACHE_SIZE, &m_hCacheMutex);
	if(er == erSuccess)
		goto exit;


exit:
	return er;
}


ECRESULT ECCacheManager::_GetUserObject(unsigned int ulUserId, userobject_type_t* lpulType, unsigned int *lpulCompanyId,
										std::string* lpstrExternId, std::string* lpstrSignature)
{
	ECRESULT		er = erSuccess;
	ECsUserObject	*sData;

	m_ulUserCacheHit++;

	pthread_mutex_lock(&m_hCacheMutex);

	er = GetCacheItem(&m_mapUserObject, ulUserId, &sData, m_ulMaxUserDetailsAge, &m_hCacheMutex);
	if(er != erSuccess)
		goto exit;

	if(lpulType)
		*lpulType = sData->ulType;

	if(lpulCompanyId)
		*lpulCompanyId = sData->ulCompanyId;

	if(lpstrExternId)
		*lpstrExternId = sData->strExternId;

	if(lpstrSignature)
		*lpstrSignature = sData->strSignature;

	m_ulUserCacheValid++;

exit:
	pthread_mutex_unlock(&m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_DelUserObject(unsigned int ulUserId)
{
	ECRESULT			er = erSuccess;

	// Remove the user
	RemoveCacheItem(&m_mapUserObject, ulUserId, &m_ulCacheUserObjectSize, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_AddUserObjectDetails(unsigned int ulUserId, userobject_type_t ulType, objectdetails_t *details)
{
	ECRESULT 			er = erSuccess;
	ECsUserObjectDetails sObjectDetails;

	if (!details) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	sObjectDetails.ulType = ulType;
	sObjectDetails.sDetails = *details;

	er = AddCacheItem(&m_mapUserObjectDetails, ulUserId, sObjectDetails, &m_ulCacheUserObjectDetailsSize, USER_CACHE_SIZE, &m_hCacheMutex);
	if (er != erSuccess)
		goto exit;

exit:
	return er;
}

ECRESULT ECCacheManager::_GetUserObjectDetails(unsigned int ulUserId, userobject_type_t ulType, objectdetails_t *details)
{
	ECRESULT		er = erSuccess;
	ECsUserObjectDetails *sObjectDetails;

	pthread_mutex_lock(&m_hCacheMutex);

	if (!details) { 
		er = ZARAFA_E_INVALID_PARAMETER; 
		goto exit; 
	}

	m_ulUserDetailsCacheHit++;

	er = GetCacheItem(&m_mapUserObjectDetails, ulUserId, &sObjectDetails, m_ulMaxUserDetailsAge, &m_hCacheMutex);
	if (er != erSuccess)
		goto exit;

	if (sObjectDetails->ulType != ulType) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	*details = sObjectDetails->sDetails; 

	m_ulUserDetailsCacheValid++;

exit:
	pthread_mutex_unlock(&m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_DelUserObjectDetails(unsigned int ulUserId)
{
	ECRESULT			er = erSuccess;

	// Remove the user details
	RemoveCacheItem(&m_mapUserObjectDetails, ulUserId, &m_ulCacheUserObjectDetailsSize, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_AddUEIdObject(std::string strExternId, userobject_type_t ulType, unsigned int ulCompanyId,
						unsigned int ulUserId, std::string strSignature)
{
	ECRESULT	er = erSuccess;
	ECsUEIdKey sKey;
	ECsUEIdObject sData;

	pthread_mutex_lock(&m_hCacheMutex);

	sData.ulCompanyId = ulCompanyId;
	sData.ulUserId = ulUserId;
	sData.strSignature = strSignature;

	sKey.ulType = ulType;
	sKey.strExternId = strExternId;

	er = AddCacheItem(&m_mapUEIdObject, sKey, sData, &m_ulCacheUserObjectSize, USER_CACHE_SIZE, &m_hCacheMutex);
	if(er == erSuccess)
		goto exit;


exit:
	pthread_mutex_unlock(&m_hCacheMutex);
	return er;
}

ECRESULT ECCacheManager::_GetUEIdObject(std::string strExternId, userobject_type_t ulType, unsigned int *lpulCompanyId,
						unsigned int* lpulUserId, std::string* lpstrSignature)
{
	ECRESULT		er = erSuccess;
	ECsUEIdKey		sKey;
	ECsUEIdObject	*sData;

	m_ulUserCacheHit++;

	pthread_mutex_lock(&m_hCacheMutex);

	sKey.ulType = ulType;
	sKey.strExternId = strExternId;

	er = GetCacheItem(&m_mapUEIdObject, sKey, &sData, m_ulMaxUserDetailsAge, &m_hCacheMutex);
	if(er != erSuccess)
		goto exit;

	if(lpulCompanyId)
		*lpulCompanyId = sData->ulCompanyId;

	if(lpulUserId)
		*lpulUserId = sData->ulUserId;

	if(lpstrSignature)
		*lpstrSignature = sData->strSignature;

	m_ulUserCacheValid++;

exit:
	pthread_mutex_unlock(&m_hCacheMutex);
	return er;
}

ECRESULT ECCacheManager::_DelUEIdObject(std::string strExternId, userobject_type_t ulType)
{
	ECRESULT	er = erSuccess;
	ECsUEIdKey	sKey;

	// Remove the user
	sKey.strExternId = strExternId;
	sKey.ulType = ulType;

	RemoveCacheItem(&m_mapUEIdObject, sKey, &m_ulCacheUserObjectSize, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::GetACLs(unsigned int ulObjId, struct rightsArray **lppRights)
{
    ECRESULT er = erSuccess;
    DB_RESULT	lpResult = NULL;
    DB_ROW		lpRow = NULL;
    ECDatabase *lpDatabase = NULL;
    std::string strQuery;
    struct rightsArray *lpRights = NULL;
    unsigned int ulRows = 0;
    unsigned int i = 0;

    // Try cache first
    if(_GetACLs(ulObjId, lppRights) == erSuccess)
        goto exit;

    // Failed, get it from the cache
	er = GetThreadLocalDatabase(this->m_lpDatabaseFactory, &lpDatabase);

	if(er != erSuccess)
		goto exit;

    strQuery = "SELECT id, type, rights FROM acl WHERE hierarchy_id=" + stringify(ulObjId);

    er = lpDatabase->DoSelect(strQuery, &lpResult);
    if(er != erSuccess)
        goto exit;

    ulRows = lpDatabase->GetNumRows(lpResult);

    lpRights = new struct rightsArray;
    if (ulRows > 0)
    {
	    lpRights->__size = ulRows;
		lpRights->__ptr = new struct rights [ulRows];
		memset(lpRights->__ptr, 0, sizeof(struct rights) * ulRows);

		for(i=0;i<ulRows;i++) {
			lpRow = lpDatabase->FetchRow(lpResult);

			if(lpRow == NULL || lpRow[0] == NULL || lpRow[1] == NULL || lpRow[2] == NULL) {
				er = ZARAFA_E_DATABASE_ERROR;
				goto exit;
			}

			lpRights->__ptr[i].ulUserid = atoi(lpRow[0]);
			lpRights->__ptr[i].ulType = atoi(lpRow[1]);
			lpRights->__ptr[i].ulRights = atoi(lpRow[2]);
		}
	}
	else
	    memset(lpRights, 0, sizeof *lpRights);

    SetACLs(ulObjId, lpRights);

    *lppRights = lpRights;

exit:
    if(lpResult)
        lpDatabase->FreeResult(lpResult);

    return er;
}

ECRESULT ECCacheManager::_GetACLs(unsigned int ulObjId, struct rightsArray **lppRights)
{
    ECRESULT er = erSuccess;
    ECsACLs *sACL;
    struct rightsArray *lpRights = NULL;
    unsigned int i;

	m_ulACLCacheHit++;
	
	pthread_mutex_lock(&m_hCacheMutex);
	er = GetCacheItem(&m_mapACLs, ulObjId, &sACL, 0, &m_hCacheMutex);	
	
	if(er != erSuccess)
		goto exit;

    lpRights = new struct rightsArray;
    if (sACL->ulACLs > 0)
    {
        lpRights->__size = sACL->ulACLs;
        lpRights->__ptr = new struct rights [sACL->ulACLs];
        memset(lpRights->__ptr, 0, sizeof(struct rights) * sACL->ulACLs);

        for(i=0;i<sACL->ulACLs;i++) {
            lpRights->__ptr[i].ulType = sACL->aACL[i].ulType;
            lpRights->__ptr[i].ulRights = sACL->aACL[i].ulMask;
            lpRights->__ptr[i].ulUserid = sACL->aACL[i].ulUserId;
        }
    }
    else
        memset(lpRights, 0, sizeof *lpRights);

    *lppRights = lpRights;

	m_ulACLCacheValid++;
exit:
	pthread_mutex_unlock(&m_hCacheMutex);

    return er;
}

ECRESULT ECCacheManager::SetACLs(unsigned int ulObjId, struct rightsArray *lpRights)
{
    ECsACLs sACLs;
    unsigned int i;
    ECRESULT er = erSuccess;

    sACLs.ulACLs = lpRights->__size;
    sACLs.aACL = new ECsACLs::ACL [lpRights->__size];

    for(i=0;i<lpRights->__size;i++) {
        sACLs.aACL[i].ulType = lpRights->__ptr[i].ulType;
        sACLs.aACL[i].ulMask = lpRights->__ptr[i].ulRights;
        sACLs.aACL[i].ulUserId = lpRights->__ptr[i].ulUserid;
    }

	er = AddCacheItem(&m_mapACLs, ulObjId, sACLs, &m_ulCacheACLSize, ACL_CACHE_SIZE, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_DelACLs(unsigned int ulObjId)
{
    ECRESULT er = erSuccess;

    er = RemoveCacheItem(&m_mapACLs, ulObjId, &m_ulCacheACLSize, &m_hCacheMutex);

    return er;
}

ECRESULT ECCacheManager::GetQuota(unsigned int ulUserId, bool bIsDefaultQuota, quotadetails_t *quota)
{
	ECRESULT er = erSuccess;

	// Try cache first
	er = _GetQuota(ulUserId, bIsDefaultQuota, quota);

	// on error, ECSecurity will update the cache

	return er;
}

ECRESULT ECCacheManager::SetQuota(unsigned int ulUserId, bool bIsDefaultQuota, quotadetails_t quota)
{
	ECRESULT er = erSuccess;
	ECsQuota	sQuota;

	sQuota.quota = quota;

	if (bIsDefaultQuota)
		er = AddCacheItem(&m_mapQuotaUserDefault, ulUserId, sQuota, &m_ulCacheQuotaUserDefaultSize, QUOTA_CACHE_SIZE, &m_hCacheMutex);
	else
		er = AddCacheItem(&m_mapQuota, ulUserId, sQuota, &m_ulCacheQuotaSize, QUOTA_CACHE_SIZE, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_GetQuota(unsigned int ulUserId, bool bIsDefaultQuota, quotadetails_t *quota)
{
	ECRESULT er = erSuccess;
	ECsQuota	*sQuota;

	pthread_mutex_lock(&m_hCacheMutex);

	if (!quota) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (bIsDefaultQuota)
		er = GetCacheItem(&m_mapQuotaUserDefault, ulUserId, &sQuota, m_ulMaxQuotaAge, &m_hCacheMutex);
	else
		er = GetCacheItem(&m_mapQuota, ulUserId, &sQuota, m_ulMaxQuotaAge, &m_hCacheMutex);

	if(er != erSuccess)
		goto exit;

	*quota = sQuota->quota;

exit:
	pthread_mutex_unlock(&m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::_DelQuota(unsigned int ulUserId, bool bIsDefaultQuota)
{
	ECRESULT er = erSuccess;

	if (bIsDefaultQuota)
		er = RemoveCacheItem(&m_mapQuotaUserDefault, ulUserId, &m_ulCacheQuotaUserDefaultSize, &m_hCacheMutex);
	else
		er = RemoveCacheItem(&m_mapQuota, ulUserId, &m_ulCacheQuotaSize, &m_hCacheMutex);

	return er;
}

void ECCacheManager::ForEachCacheItem(void(callback)(const std::string &, const std::string &, const std::string &, void*), void *obj)
{
	string value;

	pthread_mutex_lock(&m_hCacheMutex);

	callback("Object cache size (items)", "", stringify(this->m_mapObjects.size()), obj);
	callback("Object cache size (bytes)", "", stringify(this->m_mapObjects.size() * sizeof(ECMapObjects::value_type)), obj);
	callback("Object cache queries", "", stringify(m_ulObjCacheHit), obj);
	callback("Object cache hits", "", stringify(m_ulObjCacheValid), obj);

	callback("User id cache size (items)", "", stringify(this->m_mapUserObject.size()), obj);
	callback("User id cache size (bytes)", "", stringify(this->m_mapUserObject.size() * sizeof(ECMapUserObject::value_type)), obj);
	callback("User id cache queries", "", stringify(m_ulUserCacheHit), obj);
	callback("User id cache hits", "", stringify(m_ulUserCacheValid), obj);

	callback("User details cache size (items)", "", stringify(this->m_mapUserObjectDetails.size()), obj);
	callback("User details cache size (bytes)", "", stringify(this->m_mapUserObjectDetails.size() * sizeof(ECMapUserObjectDetails::value_type) + m_ulCacheUserObjectDetailsSize), obj);
	callback("User details cache queries", "", stringify(m_ulUserDetailsCacheHit), obj);
	callback("User details cache hits", "", stringify(m_ulUserDetailsCacheValid), obj);

	callback("Store cache size (items)", "", stringify(this->m_mapStores.size()), obj);
	callback("Store cache size (bytes)", "", stringify(this->m_mapStores.size() * sizeof(ECMapStores::value_type)), obj);
	callback("Store cache queries", "", stringify(m_ulStoreCacheHit), obj);
	callback("Store cache hits", "", stringify(m_ulStoreCacheValid), obj);

	callback("ACL cache size (items)", "", stringify(this->m_mapACLs.size()), obj);
	callback("ACL cache size (bytes)", "", stringify(this->m_mapACLs.size() * sizeof(ECMapACLs::value_type)), obj);
	callback("ACL cache queries", "", stringify(m_ulACLCacheHit), obj);
	callback("ACL cache hits", "", stringify(m_ulACLCacheValid), obj);

	callback("Cell cache size (items)", "", stringify(this->m_mapCells.size()), obj);
	callback("Cell cache size (bytes)", "", stringify(this->m_mapCells.size() * sizeof(ECMapCells::value_type) + m_ulCacheCellSize), obj);
	callback("Cell cache queries", "", stringify(m_ulCellCacheHit), obj);
	callback("Cell cache hits", "", stringify(m_ulCellCacheValid), obj);

	callback("Index1 cache size (items)", "", stringify(this->m_mapPropToObject.size()), obj);
	callback("Index1 cache size (bytes)", "", stringify(this->m_mapPropToObject.size() * sizeof(ECMapPropToObject::value_type) + m_ulCachePropToObjectSize), obj);
	callback("Index1 cache queries", "", stringify(m_ulPropToObjectCacheHit), obj);
	callback("Index1 cache hits", "", stringify(m_ulPropToObjectCacheValid), obj);

	callback("Index2 cache size (items)", "", stringify(this->m_mapObjectToProp.size()), obj);
	callback("Index2 cache size (bytes)", "", stringify(this->m_mapObjectToProp.size() * sizeof(ECMapObjectToProp::value_type) + m_ulCacheObjectToPropSize), obj);
	callback("Index2 cache queries", "", stringify(m_ulObjectToPropCacheHit), obj);
	callback("Index2 cache hits", "", stringify(m_ulObjectToPropCacheValid), obj);

	callback("Quota cache size (items)", "", stringify(this->m_mapQuota.size() + this->m_mapQuotaUserDefault.size()), obj);
	callback("Quota cache size (bytes)", "", stringify((this->m_mapQuota.size() + this->m_mapQuotaUserDefault.size()) * sizeof(ECMapQuota::value_type)), obj);

	pthread_mutex_unlock(&m_hCacheMutex);
}

ECRESULT ECCacheManager::DumpStats()
{
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Dumping cache stats:");
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Object cache size:  %8d (%8d bytes)", this->m_mapObjects.size(), this->m_mapObjects.size() * sizeof(ECMapObjects::value_type));
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Object hits:        %d / %d (%.02f)", m_ulObjCacheHit, m_ulObjCacheValid, (double)m_ulObjCacheValid/(double)m_ulObjCacheHit);

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  User cache size:    %8d (%8d bytes)", this->m_mapUserObject.size(), this->m_mapUserObject.size() * sizeof(ECMapUserObject::value_type));
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  User Details size:  %8d (%8d bytes)", this->m_mapUserObjectDetails.size(), this->m_mapUserObjectDetails.size() * sizeof(ECMapUserObjectDetails::value_type));
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  User hits:          %d / %d (%.02f)", m_ulUserCacheHit, m_ulUserCacheValid, (double)m_ulUserCacheValid/(double)m_ulUserCacheHit);

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Store cache size:   %8d (%8d bytes)", this->m_mapStores.size(), this->m_mapStores.size() * sizeof(ECMapStores::value_type));

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  ACL cache size:     %8d (%8d bytes)", this->m_mapACLs.size(), this->m_mapACLs.size() * sizeof(ECMapACLs::value_type));
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  ACL cache hits:     %d / %d (%.02f)", m_ulACLCacheHit, m_ulACLCacheValid, (double)m_ulACLCacheValid/(double)m_ulACLCacheHit);

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Cell cache size:    %8d (%8llu bytes)", this->m_mapCells.size(), (unsigned long long)this->m_mapCells.size() * sizeof(ECMapCells::value_type) + m_ulCacheCellSize);
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Cell cache hits:    %d / %d (%.02f)", m_ulCellCacheHit, m_ulCellCacheValid, (double)m_ulCellCacheValid/(double)m_ulCellCacheHit);

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Index1 cache size:    %8d (%8d bytes)", this->m_mapPropToObject.size(), this->m_mapPropToObject.size() * sizeof(ECMapPropToObject::value_type) + m_ulCachePropToObjectSize);
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Index1 cache hits:    %d / %d (%.02f)", m_ulPropToObjectCacheHit, m_ulPropToObjectCacheValid, (double)m_ulPropToObjectCacheValid/(double)m_ulPropToObjectCacheHit);

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Index2 cache size:    %8d (%8d bytes)", this->m_mapObjectToProp.size(), this->m_mapObjectToProp.size() * sizeof(ECMapObjectToProp::value_type) + m_ulCacheObjectToPropSize);
	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Index2 cache hits:    %d / %d (%.02f)", m_ulObjectToPropCacheHit, m_ulObjectToPropCacheValid, (double)m_ulObjectToPropCacheValid/(double)m_ulObjectToPropCacheHit);

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "  Quota cache size:     %8d (%8d bytes)", this->m_mapQuota.size()+this->m_mapQuotaUserDefault.size(), (this->m_mapQuota.size()+this->m_mapQuotaUserDefault.size()) * sizeof(ECMapQuota::value_type) + m_ulCacheQuotaSize + m_ulCacheQuotaUserDefaultSize);

	return erSuccess;
}

ECRESULT ECCacheManager::GetObjectFlags(unsigned int ulObjId, unsigned int *ulFlags)
{
	return GetObject(ulObjId, NULL, NULL, ulFlags);
}

ECRESULT ECCacheManager::GetCell(sObjectTableKey* lpsRowItem, unsigned int ulPropTag, struct propVal *lpDest, struct soap *soap)
{
    ECRESULT er = erSuccess;
    ECsCells *sCell;

	m_ulCellCacheHit++;

    // only support caching order id 0 (non-multi-valued)
    if(lpsRowItem->ulOrderId != 0)
        return ZARAFA_E_NOT_FOUND;

	pthread_mutex_lock(&m_hCacheMutex);
	
	er = GetCacheItem(&m_mapCells, lpsRowItem->ulObjId, &sCell, 0, &m_hCacheMutex);	
	if(er != erSuccess)
	    goto exit;

    if(!sCell->GetPropVal(ulPropTag, lpDest, soap)) {
        // Object is not complete, and item is not in cache. We simply don't know anything about
        // the item, so return NOT_FOUND
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    }

	m_ulCellCacheValid++;    
exit:
	pthread_mutex_unlock(&m_hCacheMutex);
	
    return er;
}

ECRESULT ECCacheManager::SetCell(sObjectTableKey* lpsRowItem, unsigned int ulPropTag, struct propVal *lpSrc)
{
    ECRESULT er = erSuccess;
    ECsCells *sCell;

    if(lpsRowItem->ulOrderId != 0)
        return ZARAFA_E_NOT_FOUND;

	pthread_mutex_lock(&m_hCacheMutex);

    if(GetCacheItem(&m_mapCells, lpsRowItem->ulObjId, &sCell, 0, &m_hCacheMutex) == erSuccess) {
        long long ulSize = sCell->GetSize();
        sCell->AddPropVal(ulPropTag, lpSrc);
        ulSize -= sCell->GetSize();
        
        // ulSize is positive if the cache shrank
        m_ulCacheCellSize -= ulSize;
    } else {
        ECsCells sNewCell;
        
        sNewCell.AddPropVal(ulPropTag, lpSrc);
        
    	er = AddCacheItem(&m_mapCells, lpsRowItem->ulObjId, sNewCell, &m_ulCacheCellSize, m_ulCacheCellMaxSize, &m_hCacheMutex);
    	if(er != erSuccess)
    	    goto exit;
    }

exit:
    pthread_mutex_unlock(&m_hCacheMutex);

    return er;
}

ECRESULT ECCacheManager::UpdateCell(unsigned int ulObjId, unsigned int ulPropTag, int lDelta)
{
    ECRESULT er = erSuccess;
    ECsCells *sCell;
    
	pthread_mutex_lock(&m_hCacheMutex);

    if(GetCacheItem(&m_mapCells, ulObjId, &sCell, 0, &m_hCacheMutex) == erSuccess) {
        sCell->UpdatePropVal(ulPropTag, lDelta);
    } else {
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    }

exit:
    pthread_mutex_unlock(&m_hCacheMutex);

    return er;
}

ECRESULT ECCacheManager::_DelCell(unsigned int ulObjId)
{
    ECRESULT er = erSuccess;
    
	er = RemoveCacheItem(&m_mapCells, ulObjId, &m_ulCacheCellSize, &m_hCacheMutex);

	return er;
}

ECRESULT ECCacheManager::GetServerDetails(const std::string &strServerId, serverdetails_t *lpsDetails)
{
	ECRESULT			er = erSuccess;
	ECsServerDetails	*sEntry;

	pthread_mutex_lock(&m_hCacheMutex);
	er = GetCacheItem(&m_mapServerDetails, strServerId, &sEntry, SERVER_CACHE_AGE, &m_hCacheMutex);
	if (er != erSuccess)
		goto exit;

	if (lpsDetails)
		*lpsDetails = sEntry->sDetails;

exit:
	pthread_mutex_unlock(&m_hCacheMutex);
	
	return er;
}

ECRESULT ECCacheManager::SetServerDetails(const std::string &strServerId, const serverdetails_t &sDetails)
{
	ECsServerDetails	sEntry;

	sEntry.sDetails = sDetails;
	return AddCacheItem(&m_mapServerDetails, strServerId, sEntry, &m_ulServerDetailsSize, SERVER_CACHE_SIZE, &m_hCacheMutex);
}

ECRESULT ECCacheManager::RemoveIndexData(unsigned int ulObjId)
{
	ECRESULT				er = erSuccess;
	ECsIndexObject	sObjectKey;
	ECMapObjectToProp::iterator	iter, iDel;
	ECMapPropToObject::iterator	iterPropToObj;

	sObjectKey.ulObjId = ulObjId;
	sObjectKey.ulTag = 0;

	pthread_mutex_lock(&m_hCacheMutex);

	iter = m_mapObjectToProp.lower_bound(sObjectKey);
	while(iter != m_mapObjectToProp.end() && iter->first.ulObjId == ulObjId)
	{
		iDel = iter++;

		iterPropToObj = m_mapPropToObject.find(iDel->second);
		if(iterPropToObj != m_mapPropToObject.end())
			RemoveCacheItem(&m_mapPropToObject, iDel->second, &m_ulCachePropToObjectSize, &m_hCacheMutex);


		RemoveCacheItem(&m_mapObjectToProp, iDel->first, &m_ulCacheObjectToPropSize, &m_hCacheMutex);


	}

	pthread_mutex_unlock(&m_hCacheMutex);
	return er;
}

ECRESULT ECCacheManager::_AddIndexData(ECsIndexObject* lpObject, ECsIndexProp* lpProp)
{
	ECRESULT	er = erSuccess;

	er = AddCacheItem(&m_mapPropToObject, *lpProp, *lpObject, &m_ulCachePropToObjectSize, m_ulCacheIndexedObjectMaxSize, &m_hCacheMutex);
	if(er != erSuccess)
		goto exit;

	er = AddCacheItem(&m_mapObjectToProp, *lpObject, *lpProp, &m_ulCacheObjectToPropSize, m_ulCacheIndexedObjectMaxSize, &m_hCacheMutex);
	if(er != erSuccess)
		goto exit;

exit:
	return er;
}

ECRESULT ECCacheManager::GetPropFromObject(unsigned int ulTag, unsigned int ulObjId, struct soap *soap, unsigned int* lpcbData, unsigned char** lppData)
{
	ECRESULT		er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	DB_LENGTHS		lpDBLenths = NULL;
	std::string		strQuery;
	ECDatabase*		lpDatabase = NULL;
	ECsIndexProp	*sObject;
	ECsIndexObject	sObjectKey;
    ECsIndexProp sNewObject;

	sObjectKey.ulObjId = ulObjId;
	sObjectKey.ulTag = ulTag;

	m_ulObjectToPropCacheHit++;

	pthread_mutex_lock(&m_hCacheMutex);
	
	er = GetCacheItem(&m_mapObjectToProp, sObjectKey, &sObject, 0, &m_hCacheMutex);

	if(er != erSuccess) {
		er = GetThreadLocalDatabase(this->m_lpDatabaseFactory, &lpDatabase);
		if(er != erSuccess)
			goto exit;

		// Get them from the database
		strQuery = "SELECT val_binary FROM indexedproperties WHERE tag="+stringify(ulTag)+" AND hierarchyid="+stringify(ulObjId);
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		lpDBLenths = lpDatabase->FetchRowLengths(lpDBResult);
		if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBLenths == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		sNewObject.SetValue(ulTag, (unsigned char*) lpDBRow[0], (unsigned int) lpDBLenths[0]);

		er = _AddIndexData(&sObjectKey, &sNewObject);
		if(er != erSuccess)
			goto exit;

        sObject = &sNewObject;
	}else {
		m_ulObjectToPropCacheValid++;
	}

	*lppData = s_alloc<unsigned char>(soap, sObject->cbData);
	*lpcbData = sObject->cbData;

	memcpy(*lppData, sObject->lpData, sObject->cbData);

exit:
	pthread_mutex_unlock(&m_hCacheMutex);

	if (lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECCacheManager::GetObjectFromProp(unsigned int ulTag, unsigned int cbData, unsigned char* lpData, unsigned int* lpulObjId)
{
	ECRESULT		er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	std::string		strQuery;
	ECDatabase*		lpDatabase = NULL;
	ECsIndexProp	sObject;
	ECsIndexObject	*sIndexObject;
    ECsIndexObject sNewIndexObject;

	pthread_mutex_lock(&m_hCacheMutex);

	if(lpData == NULL || lpulObjId == NULL || cbData == 0) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	m_ulPropToObjectCacheHit++;

	sObject.ulTag = ulTag;
	sObject.cbData = cbData;
	sObject.lpData = lpData; // Cheap copy, Set this item on NULL before you exit

	er = GetCacheItem(&m_mapPropToObject, sObject, &sIndexObject, 0, &m_hCacheMutex);
	if(er != erSuccess) {
		er = GetThreadLocalDatabase(this->m_lpDatabaseFactory, &lpDatabase);
		if(er != erSuccess)
			goto exit;

		// Get them from the database
		strQuery = "SELECT hierarchyid FROM indexedproperties WHERE tag="+stringify(ulTag)+" AND val_binary="+ lpDatabase->EscapeBinary(lpData, cbData);
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		if(lpDBRow == NULL || lpDBRow[0] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		sNewIndexObject.ulTag = ulTag;
		sNewIndexObject.ulObjId = atoui(lpDBRow[0]);

		er = _AddIndexData(&sNewIndexObject, &sObject);
		if(er != erSuccess)
			goto exit;
			
        sIndexObject = &sNewIndexObject;
	}else {
		m_ulPropToObjectCacheValid++;
	}

	*lpulObjId = sIndexObject->ulObjId;

exit:
	pthread_mutex_unlock(&m_hCacheMutex);

	if (lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	sObject.lpData = NULL; // Remove reference

	return er;
}

ECRESULT ECCacheManager::SetObjectProp(unsigned int ulTag, unsigned int cbData, unsigned char *lpData, unsigned int ulObjId)
{
    ECRESULT er = erSuccess;

    ECsIndexObject sObject;
    ECsIndexProp sProp;

    sObject.ulTag = ulTag;
    sObject.ulObjId = ulObjId;

    sProp.SetValue(ulTag, lpData, cbData);

    er = _AddIndexData(&sObject, &sProp);

    return er;
}

ECRESULT ECCacheManager::GetEntryIdFromObject(unsigned int ulObjId, struct soap *soap, entryId** lppEntryId)
{
	ECRESULT	er = erSuccess;
	entryId*	lpEntryId = s_alloc<entryId>(soap);

	er = GetEntryIdFromObject(ulObjId, soap, lpEntryId);
	if (er != erSuccess)
		goto exit;

	*lppEntryId = lpEntryId;
exit:
	if (er != erSuccess && lpEntryId)
		delete lpEntryId;

	return er;
}

ECRESULT ECCacheManager::GetEntryIdFromObject(unsigned int ulObjId, struct soap *soap, entryId* lpEntryId)
{
	ECRESULT	er = erSuccess;

	er = GetPropFromObject( PROP_ID(PR_ENTRYID), ulObjId, soap, (unsigned int*)&lpEntryId->__size, &lpEntryId->__ptr);

	return er;
}

ECRESULT ECCacheManager::GetObjectFromEntryId(entryId* lpEntryId, unsigned int* lpulObjId)
{
	ECRESULT	er = erSuccess;

	er = GetObjectFromProp( PROP_ID(PR_ENTRYID), lpEntryId->__size, lpEntryId->__ptr, lpulObjId);

	return er;
}

ECRESULT ECCacheManager::SetObjectEntryId(entryId* lpEntryId, unsigned int ulObjId)
{
    ECRESULT 	er = erSuccess;

    er = SetObjectProp( PROP_ID(PR_ENTRYID), lpEntryId->__size, lpEntryId->__ptr, ulObjId);

    return er;
}

//Convert entryid to database object id
ECRESULT ECCacheManager::GetEntryListToObjectList(struct entryList *lpEntryList, ECListInt* lplObjectList)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulId = 0;
	bool			bPartialCompletion = false;

	if(lpEntryList == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	for(unsigned int i=0; i < lpEntryList->__size; i++)
	{
		if(GetObjectFromEntryId(&lpEntryList->__ptr[i], &ulId) != erSuccess) {
			bPartialCompletion = true;
			continue; // Unknown entryid, next item
		}

		lplObjectList->push_back(ulId);
	}

exit:
	if(bPartialCompletion)
		er = ZARAFA_W_PARTIAL_COMPLETION;

	return er;
}
