/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef ECCACHEMANAGER
#define ECCACHEMANAGER

#include <map>
#include <pthread.h>

#include "ECDatabaseFactory.h"
#include "ECConfig.h"
#include "ECLogger.h"
#include "SOAPUtils.h"
#include <mapidefs.h>

#include "../common/ECKeyTable.h"

#ifdef __GNUC__
#if __GNUC__ < 3
    #include <hash_map.h>
    namespace sgi { using ::hash_map; }; // inherit globals
#else
	#if __GNUC__ == 4 && __GNUC_MINOR__ >= 3 && defined(__DEPRECATED)
		#undef __DEPRECATED
	#endif

    #include <ext/hash_map>

    #if __GNUC__ == 3 && __GNUC_MINOR__ == 0
        namespace sgi = std;               // GCC 3.0
    #else
        namespace sgi = ::__gnu_cxx;       // GCC 3.1 and later
    #endif

	#if __GNUC__ == 4 && __GNUC_MINOR__ >= 3
		#define __DEPRECATED 1
	#endif
#endif
#else
	// Win32
	#include <hash_map>

	// The win32 standard hash_map is different to sgi's hash_map, so we have a workaround here
	// We basically map the standard SGI interface to the win32 interface, BUT the EqualFunc
	// function must be a less<> function in win32 (!!), This is automatically fixed when you
	// use sgi::equal_to as we define this here, but when you define your own EqualFunc, this must
	// be a less<> function in win32, and equal_to<> in sgi-compatible environments
	namespace sgi {

		template <typename Key, typename HashFunc, typename EqualFunc>
			class Traits {
			private:
				const HashFunc hasher;
				const EqualFunc comparer;
			public:
				Traits() {};
				~Traits() {};
				enum
					{	// parameters for hash table
					bucket_size = 4,	// 0 < bucket_size
					min_buckets = 8};	// min_buckets = 2 ^^ N, 0 < N

				size_t operator() (const Key &a) const { return hasher(a); };
				bool operator() (const Key &a, const Key &b) const { return comparer(a,b); };
			};

		template <typename Key, typename Value, typename HashFunc, typename EqualFunc>
			class hash_map : public stdext::hash_map<Key, Value, Traits<Key, HashFunc, EqualFunc> > { };

		template <typename Key>
			class hash {
			public:
				hash() {};
				~hash() {};

				size_t operator() (const Key value) const { return stdext::hash_value<Key>(value); }
			};

		template <typename Key>
			class equal_to {
			private:
				const std::less<Key> comparer;
			public:
				equal_to() {};
				~equal_to() {};

				bool operator() (const Key &a, const Key &b) const { return comparer(a,b); }
			};
	}
#endif

class ECSessionManager;

class ECsCacheEntry {
public:
	ECsCacheEntry() { ulLastAccess = 0; }

	time_t 	ulLastAccess;
};

class ECsStores : public ECsCacheEntry {
public:
	unsigned int	ulStore;
	GUID			guidStore;
};

class ECsUserObject : public ECsCacheEntry {
public:
	userobject_type_t	ulType;
	unsigned int		ulCompanyId;
	std::string			strExternId;
	std::string			strSignature;
};

typedef struct {
	userobject_type_t	ulType;
	std::string			strExternId;
} ECsUEIdKey;

inline bool operator < (ECsUEIdKey a, ECsUEIdKey b) {
	if (a.ulType < b.ulType)
		return true;
	if ((a.ulType == b.ulType) && a.strExternId < b.strExternId)
		return true;
	return false;
}

class ECsUEIdObject : public ECsCacheEntry {
public:
	unsigned int		ulCompanyId;
	unsigned int		ulUserId;
	std::string			strSignature;
};

class ECsUserObjectDetails : public ECsCacheEntry {
public:
	userobject_type_t		ulType;
	objectdetails_t			sDetails;
};

class ECsServerDetails : public ECsCacheEntry {
public:
	serverdetails_t			sDetails;
};

class ECsObjects : public ECsCacheEntry {
public:
	unsigned int	ulParent;
	unsigned int	ulOwner;
	unsigned int	ulFlags;
	unsigned int	ulType;
};

class ECsQuota : public ECsCacheEntry {
public:
	quotadetails_t	quota;
};

class ECsIndexObject : public ECsCacheEntry {
public:
	unsigned int ulObjId;
	unsigned int ulTag;
};

class ECsIndexProp : public ECsCacheEntry {
public:
    ECsIndexProp() : ECsCacheEntry() { 
		lpData = NULL; 
		ulTag = 0; 
		cbData = 0;
	}

	~ECsIndexProp() {
		if(lpData) {
			delete[] lpData;
		}
	}
    
    ECsIndexProp(const ECsIndexProp &src) {
        if(this == &src)
            return;
        Copy(&src, this);
    }
    
    void operator=(const ECsIndexProp &src) {
        if(this == &src)
            return;
		Free();
        Copy(&src, this);
    }

	void SetValue(unsigned int ulTag, unsigned char* lpData, unsigned int cbData) {

		if(lpData == NULL|| cbData == 0)
			return;

		Free();

		this->lpData = new unsigned char[cbData];
		this->cbData = cbData;
		this->ulTag = ulTag;

		memcpy(this->lpData, lpData, (size_t)cbData);

	}
protected:
	void Free() {
		if(lpData) {
			delete[] lpData;
		}
		ulTag = 0; 
		cbData = 0;
	}

	void Copy(const ECsIndexProp* src, ECsIndexProp* dst) {
		if(src->lpData != NULL && src->cbData>0) {
			dst->lpData = new unsigned char[src->cbData];
			memcpy(dst->lpData, src->lpData, (size_t)src->cbData);
			dst->cbData = src->cbData;
		}

		dst->ulTag = src->ulTag;
	}
public:
	unsigned int	ulTag;
	unsigned char*	lpData;
	unsigned int	cbData;
};

class ECsCells : public ECsCacheEntry {
public:
    ECsCells() : ECsCacheEntry() { bComplete = false; };
    ~ECsCells() {
        std::map<unsigned int, struct propVal>::iterator i;
        for(i=mapPropVals.begin(); i!=mapPropVals.end(); i++) {
            FreePropVal(&i->second, false); 
        }
    };
    
    ECsCells(const ECsCells &src) {
        struct propVal val;
        std::map<unsigned int, struct propVal>::const_iterator i;
        for(i=src.mapPropVals.begin(); i!=src.mapPropVals.end(); i++) {
            CopyPropVal((struct propVal *)&i->second, &val);
            mapPropVals[i->first] = val;
        }
        bComplete = src.bComplete;
    }
    
    void operator=(const ECsCells &src) {
        struct propVal val;
        std::map<unsigned int, struct propVal>::iterator i;
        for(i=mapPropVals.begin(); i!=mapPropVals.end(); i++) {
            FreePropVal(&i->second, false); 
        }
        mapPropVals.clear();
        
        for(i=((ECsCells &)src).mapPropVals.begin(); i!=src.mapPropVals.end(); i++) {
            CopyPropVal((struct propVal *)&i->second, &val);
            mapPropVals[i->first] = val;
        }
        bComplete = src.bComplete;
    }
    
    // Add a property value for this object
    void AddPropVal(unsigned int ulPropTag, struct propVal *lpPropVal) {
        struct propVal val;
        std::map<unsigned int, struct propVal>::iterator i;
        CopyPropVal(lpPropVal, &val);
        i = mapPropVals.find(ulPropTag);
        if(i != mapPropVals.end()) {
            FreePropVal(&i->second, false); 
            mapPropVals.erase(i);
        }
        mapPropVals[ulPropTag] = val;
    }
    
    // get a property value for this object
    bool GetPropVal(unsigned int ulPropTag, struct propVal *lpPropVal, struct soap *soap) {
        std::map<unsigned int, struct propVal>::iterator i;
        i = mapPropVals.find(ulPropTag);
        if(i == mapPropVals.end())
            return false;
        CopyPropVal(&i->second, lpPropVal, soap);
        return true;
    }
    
    // Updates a LONG type property
    void UpdatePropVal(unsigned int ulPropTag, int lDelta) {
        std::map<unsigned int, struct propVal>::iterator i;
        if(PROP_TYPE(ulPropTag) != PT_LONG)
            return;
        i = mapPropVals.find(ulPropTag);
        if(i == mapPropVals.end())
            return;
        i->second.Value.ul += lDelta;
    }

    // Gets the amount of memory used by this object    
    size_t GetSize() const {
        size_t ulSize = 0;
        
        std::map<unsigned int, struct propVal>::const_iterator i;
        for(i=mapPropVals.begin(); i!=mapPropVals.end(); i++) {
            switch(i->second.__union) {
                case SOAP_UNION_propValData_lpszA:
                    ulSize += i->second.Value.lpszA ? (unsigned int)strlen(i->second.Value.lpszA) : 0;
					break;
                case SOAP_UNION_propValData_bin:
                    ulSize += i->second.Value.bin ? i->second.Value.bin->__size + sizeof(i->second.Value.bin[0]) : 0;
					break;
                case SOAP_UNION_propValData_hilo:
                    ulSize += sizeof(i->second.Value.hilo[0]);
					break;
                default:
                    break;
            }
            
            ulSize += sizeof(std::map<unsigned int, struct propVal>::value_type);
        }
        ulSize += sizeof(this);
        
        return ulSize;
    }
    
    // All properties for this object; propTag => propVal
    std::map<unsigned int, struct propVal> mapPropVals;
    
    bool bComplete;
};

class ECsACLs : public ECsCacheEntry {
public:
	ECsACLs() : ECsCacheEntry() { ulACLs = 0; aACL = NULL; }
    ECsACLs(const ECsACLs &src) {
        ulACLs = src.ulACLs;
        aACL = new ACL[src.ulACLs];
        memcpy(aACL, src.aACL, sizeof(ACL) * src.ulACLs);
    };
    void operator=(const ECsACLs &src) {
        if(aACL) delete [] aACL;
        ulACLs = src.ulACLs;
        aACL = new ACL[src.ulACLs];
        memcpy(aACL, src.aACL, sizeof(ACL) * src.ulACLs);
    };
    ~ECsACLs() { if(aACL) delete [] aACL; }
    unsigned int 	ulACLs;
    struct ACL {
        unsigned int ulType;
        unsigned int ulMask;
        unsigned int ulUserId;
    } *aACL;
};

typedef struct {
	sObjectTableKey	sKey;
	unsigned int	ulPropTag;
}ECsSortKeyKey;


struct lessindexobjectkey {
	bool operator()(const ECsIndexObject& a, const ECsIndexObject& b) const
	{
		if(a.ulObjId < b.ulObjId)
			return true;
		else if(a.ulObjId == b.ulObjId && a.ulTag < b.ulTag)
			return true;

		return false;
	}

};

struct lessindexpropkey {
	bool operator()(const ECsIndexProp& a, const ECsIndexProp& b) const
	{
		
		if(a.cbData < b.cbData)
			return true;

		if(a.cbData == b.cbData) {
			int c = memcmp(a.lpData, b.lpData, a.cbData);
			if(c < 0)
				return true;
			else if(c == 0 && a.ulTag < b.ulTag)
				return true;
		}

		return false;
	}

};

struct lesskey {
  bool operator()(const ECsSortKeyKey& a, const ECsSortKeyKey& b) const
  {
	  return (a.sKey.ulObjId < b.sKey.ulObjId) || (a.sKey.ulObjId == b.sKey.ulObjId && a.sKey.ulOrderId < b.sKey.ulOrderId) || (a.sKey.ulObjId == b.sKey.ulObjId && a.sKey.ulOrderId == b.sKey.ulOrderId && a.ulPropTag < b.ulPropTag);

  }
};

struct equalkey {
    bool operator()(const ECsSortKeyKey& a, const ECsSortKeyKey& b) const
    {
        return a.sKey.ulObjId == b.sKey.ulObjId && a.sKey.ulOrderId == b.sKey.ulOrderId && a.ulPropTag == b.ulPropTag;
    }
};


struct keyhash {
    size_t operator () (const ECsSortKeyKey &s) const {
	sgi::hash<unsigned int> hasher;

	return hasher(s.sKey.ulObjId) + hasher(s.sKey.ulOrderId) + hasher(s.ulPropTag);
    }
};

typedef sgi::hash_map<unsigned int, ECsObjects, sgi::hash<unsigned int>, sgi::equal_to<unsigned int> > ECMapObjects;
typedef sgi::hash_map<unsigned int, ECsStores, sgi::hash<unsigned int>, sgi::equal_to<unsigned int> > ECMapStores;
typedef sgi::hash_map<unsigned int, ECsACLs, sgi::hash<unsigned int>, sgi::equal_to<unsigned int> > ECMapACLs;
typedef sgi::hash_map<unsigned int, ECsQuota, sgi::hash<unsigned int>, sgi::equal_to<unsigned int> > ECMapQuota;
typedef sgi::hash_map<unsigned int, ECsUserObject, sgi::hash<unsigned int>, sgi::equal_to<unsigned int> > ECMapUserObject; // userid to user object
typedef std::map<ECsUEIdKey, ECsUEIdObject> ECMapUEIdObject; // user type + externid to user object
typedef sgi::hash_map<unsigned int, ECsUserObjectDetails, sgi::hash<unsigned int>, sgi::equal_to<unsigned int> > ECMapUserObjectDetails; // userid to user object data
typedef std::map<std::string, ECsServerDetails> ECMapServerDetails;
typedef sgi::hash_map<unsigned int, ECsCells, sgi::hash<unsigned int>, sgi::equal_to<unsigned int> > ECMapCells;

// Index properties
typedef std::map<ECsIndexObject, ECsIndexProp, lessindexobjectkey > ECMapObjectToProp;
typedef std::map<ECsIndexProp, ECsIndexObject, lessindexpropkey > ECMapPropToObject;

#define CACHE_NO_PARENT 0xFFFFFFFF

class ECCacheManager  
{
public:
	ECCacheManager(ECConfig *lpConfig, ECDatabaseFactory *lpDatabase, ECLogger *lpLogger);
	virtual ~ECCacheManager();

	ECRESULT PurgeCache();
	ECRESULT PurgeUserCache();

	// These are read-through (ie they access the DB if they can't find the data)
	ECRESULT GetParent(unsigned int ulObjId, unsigned int *ulParent);
	ECRESULT GetOwner(unsigned int ulObjId, unsigned int *ulOwner);
	ECRESULT GetObject(unsigned int ulObjId, unsigned int *ulParent, unsigned int *ulOwner, unsigned int *ulFlags, unsigned int *ulType = NULL);
	ECRESULT SetObject(unsigned int ulObjId, unsigned int ulParent, unsigned int ulOwner, unsigned int ulFlags, unsigned int ulType);

	ECRESULT GetStore(unsigned int ulObjId, unsigned int *ulStore, GUID *lpGuid, unsigned int maxdepth = 100);
	ECRESULT GetObjectFlags(unsigned int ulObjId, unsigned int *ulFlags);
	ECRESULT SetStore(unsigned int ulObjId, unsigned int ulStore, GUID *lpGuid);

	ECRESULT GetServerDetails(const std::string &strServerId, serverdetails_t *lpsDetails);
	ECRESULT SetServerDetails(const std::string &strServerId, const serverdetails_t &sDetails);

	// Cache user table
	ECRESULT GetUserObject(unsigned int ulUserId, userobject_type_t* lpulType, unsigned int *lpulCompanyId,
						   std::string *lpstrExtern, std::string *lpstrSignature);
	ECRESULT GetUserCompany(unsigned int ulUserId, userobject_type_t ulType, unsigned int *lpulCompanyId);
	ECRESULT GetUserExternId(unsigned int ulUserId, userobject_type_t *lpsType, std::string* lpstrExtern, std::string* lpstrSignature);

	ECRESULT GetUserObject(std::string strExternId, userobject_type_t ulType, unsigned int *lpulCompanyId,
						   unsigned int *lpulUserId, std::string *lpstrSignature);

	// Cache user information
	ECRESULT GetUserDetails(unsigned int ulUserId, userobject_type_t ulType, objectdetails_t *details);
	ECRESULT SetUserDetails(unsigned int ulUserId, userobject_type_t ulType, objectdetails_t *details);

	ECRESULT GetACLs(unsigned int ulObjId, struct rightsArray **lppRights);
	ECRESULT SetACLs(unsigned int ulObjId, struct rightsArray *lpRights);

	ECRESULT GetQuota(unsigned int ulUserId, bool bIsDefaultQuota, quotadetails_t *quota);
	ECRESULT SetQuota(unsigned int ulUserId, bool bIsDefaultQuota, quotadetails_t quota);

	ECRESULT Update(unsigned int ulType, unsigned int ulObjId);
	ECRESULT UpdateUser(unsigned int ulUserId);

	ECRESULT GetEntryIdFromObject(unsigned int ulObjId, struct soap *soap, entryId* lpEntrId);
	ECRESULT GetEntryIdFromObject(unsigned int ulObjId, struct soap *soap, entryId** lppEntryId);
	ECRESULT GetObjectFromEntryId(entryId* lpEntrId, unsigned int* lpulObjId);
	ECRESULT SetObjectEntryId(entryId *lpEntryId, unsigned int ulObjId);
	ECRESULT GetEntryListToObjectList(struct entryList *lpEntryList, ECListInt* lplObjectList);

	// Table data functions (pure cache functions, they will never access the DB themselves. Data must be provided through Set functions)
	ECRESULT GetCell(sObjectTableKey* lpsRowItem, unsigned int ulPropTag, struct propVal *lpDest, struct soap *soap);
	ECRESULT SetCell(sObjectTableKey* lpsRowItem, unsigned int ulPropTag, struct propVal *lpSrc);
	ECRESULT UpdateCell(unsigned int ulObjId, unsigned int ulPropTag, int lDelta);
	
	// Cache Index properties
	ECRESULT GetPropFromObject(unsigned int ulTag, unsigned int ulObjId, struct soap *soap, unsigned int* lpcbData, unsigned char** lppData);
	ECRESULT GetObjectFromProp(unsigned int ulTag, unsigned int cbData, unsigned char* lpData, unsigned int* lpulObjId);
	ECRESULT RemoveIndexData(unsigned int ulObjId);

	ECRESULT SetObjectProp(unsigned int ulTag, unsigned int cbData, unsigned char* lpData, unsigned int ulObjId);

	void ForEachCacheItem(void(callback)(const std::string &, const std::string &, const std::string &, void*), void *obj);
	ECRESULT DumpStats();
private:
	// cache functions
	ECRESULT _GetACLs(unsigned int ulObjId, struct rightsArray **lppRights);
	ECRESULT _DelACLs(unsigned int ulObjId);
	
	ECRESULT _GetObject(unsigned int ulObjId, unsigned int *ulParent, unsigned int *ulOwner, unsigned int *ulFlags, unsigned int *ulType);
	ECRESULT _DelObject(unsigned int ulObjId);

	ECRESULT _GetStore(unsigned int ulObjId, unsigned int *ulStore, GUID *lpGuid);
	ECRESULT _DelStore(unsigned int ulObjId);

	ECRESULT _AddUserObject(unsigned int ulUserId, userobject_type_t ulType, unsigned int ulCompanyId,
							std::string strExternId, std::string strSignature);
	ECRESULT _GetUserObject(unsigned int ulUserId, userobject_type_t* lpulType, unsigned int *lpulCompanyId,
							std::string* lpstrExternId, std::string* lpstrSignature);
	ECRESULT _DelUserObject(unsigned int ulUserId);

	ECRESULT _AddUEIdObject(std::string strExternId, userobject_type_t ulType, unsigned int ulCompanyId,
							unsigned int ulUserId, std::string strSignature);
	ECRESULT _GetUEIdObject(std::string strExternId, userobject_type_t ulType, unsigned int *lpulCompanyId,
							unsigned int* lpulUserId, std::string* lpstrSignature);
	ECRESULT _DelUEIdObject(std::string strExternId, userobject_type_t ulType);

	ECRESULT _AddUserObjectDetails(unsigned int ulUserId, userobject_type_t ulType, objectdetails_t *details);
	ECRESULT _GetUserObjectDetails(unsigned int ulUserId, userobject_type_t ulType, objectdetails_t *details);
	ECRESULT _DelUserObjectDetails(unsigned int ulUserId);

	ECRESULT _DelCell(unsigned int ulObjId);

	ECRESULT _GetQuota(unsigned int ulUserId, bool bIsDefaultQuota, quotadetails_t *quota);
	ECRESULT _DelQuota(unsigned int ulUserId, bool bIsDefaultQuota);

	// Cache Index properties
	ECRESULT _AddIndexData(ECsIndexObject* lpObject, ECsIndexProp* lpProp);	

private:
	ECDatabaseFactory*	m_lpDatabaseFactory;
	ECLogger*			m_lpLogger;

	pthread_mutex_t		m_hCacheMutex;			// Currently, all caches are locked simultaneously
	
	// Quota cache, to reduce the impact of the user plugin
	// m_mapQuota contains user and company cache, except when it's the company user default quota
	// m_mapQuotaUserDefault contains company user default quota
	// this can't be in the same map, since the id is the same for "company" and "company user default"
	ECMapQuota			m_mapQuota, m_mapQuotaUserDefault;
	unsigned long long	m_ulCacheQuotaSize, m_ulCacheQuotaUserDefaultSize;
	unsigned int		m_ulMaxQuotaAge;		// Age in seconds

	// Object cache, (hierarchy table)
	ECMapObjects		m_mapObjects;
	unsigned long long	m_ulCacheObjectSize;

	// Store cache
	ECMapStores			m_mapStores;
	unsigned long long	m_ulCacheStoreSize;

	// User cache
	ECMapUserObject		m_mapUserObject;
	unsigned long long	m_ulCacheUserObjectSize;

	ECMapUEIdObject		m_mapUEIdObject;
	unsigned long long	m_ulCacheUEIdObjectSize;

	ECMapUserObjectDetails	m_mapUserObjectDetails;
	unsigned long long	m_ulCacheUserObjectDetailsSize;

	unsigned int		m_ulMaxUserDetailsAge;		// Age in seconds

	// ACL cache
	ECMapACLs			m_mapACLs;
	unsigned long long	m_ulCacheACLSize;

	// Cell cache, include the column data of a loaded table
	ECMapCells			m_mapCells;
	unsigned long long	m_ulCacheCellSize;
	
	unsigned long long	m_ulCacheObjectMaxSize;
	unsigned long long	m_ulCacheIndexedObjectMaxSize;
	unsigned long long	m_ulCacheCellMaxSize;

	// Server cache
	ECMapServerDetails	m_mapServerDetails;
	unsigned long long	m_ulServerDetailsSize;
	
	//Index properties
	ECMapPropToObject	m_mapPropToObject;
	ECMapObjectToProp	m_mapObjectToProp;

	unsigned long long	m_ulCachePropToObjectSize;
	unsigned long long	m_ulCacheObjectToPropSize;

	//Cache hits
	unsigned int		m_ulCellCacheHit;
	unsigned int		m_ulCellCacheValid;
	unsigned int		m_ulObjCacheHit;
	unsigned int		m_ulObjCacheValid;
	unsigned int		m_ulACLCacheHit;
	unsigned int		m_ulACLCacheValid;
	unsigned int		m_ulPropToObjectCacheHit;
	unsigned int		m_ulPropToObjectCacheValid;
	unsigned int		m_ulObjectToPropCacheHit;
	unsigned int		m_ulObjectToPropCacheValid;
	unsigned int		m_ulUserCacheHit;
	unsigned int		m_ulUserCacheValid;
	unsigned int		m_ulUserDetailsCacheHit;
	unsigned int		m_ulUserDetailsCacheValid;
	unsigned int		m_ulStoreCacheHit;
	unsigned int		m_ulStoreCacheValid;
	
};

#define ACL_CACHE_SIZE		1048576
#define STORE_CACHE_SIZE	1048576
#define QUOTA_CACHE_SIZE	1048576
#define USER_CACHE_SIZE		1048576
#define SERVER_CACHE_SIZE	1048576
#define SERVER_CACHE_AGE	1800		//(30 min)

#endif // #ifndef ECCACHEMANAGER
