/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef ECSEARCHFOLDERS_H
#define ECSEARCHFOLDERS_H

#include "ECLogger.h"
#include "ECDatabaseFactory.h"
#include "ECKeyTable.h"

#include "soapH.h"

#include <map>
#include <set>

class ECSessionManager;

typedef struct {
    struct searchCriteria 	*lpSearchCriteria;
    pthread_t 				sThreadId;
    pthread_mutex_t			mMutexThreadFree;
    bool 					bThreadFree;
    bool					bThreadExit;
    unsigned int			ulStoreId;
    unsigned int			ulFolderId;
} SEARCHFOLDER;

typedef struct EVENT {
    unsigned int			ulStoreId;
    unsigned int			ulFolderId;
    unsigned int			ulObjectId;
    ECKeyTable::UpdateType  ulType;
    
    bool operator< (const struct EVENT &b) { return ulFolderId < b.ulFolderId ? true : (ulType < b.ulType ? true : ( ulObjectId < b.ulObjectId ? true : false ) ); }
    bool operator== (const struct EVENT &b) { return ulFolderId == b.ulFolderId && ulType == b.ulType && ulObjectId ==  b.ulObjectId; }
} EVENT;

typedef std::map<unsigned int, SEARCHFOLDER *> FOLDERIDSEARCH;
typedef std::map<unsigned int, FOLDERIDSEARCH> STOREFOLDERIDSEARCH;
typedef std::map<unsigned int, pthread_t> SEARCHTHREADMAP;

class ECSearchFolders {
public:
    ECSearchFolders(ECSessionManager *lpSessionManager, ECDatabaseFactory *lpFactory, ECLogger *lpLogger);
    virtual ~ECSearchFolders();

    // Initial load
    virtual ECRESULT LoadSearchFolders();

    // Set search criteria for a new or existing search folder
    virtual ECRESULT SetSearchCriteria(unsigned int ulStoreId, unsigned int ulFolderId, struct searchCriteria *lpSearchCriteria);

    // Get search criteria for an existing search folder
    virtual ECRESULT GetSearchCriteria(unsigned int ulStoreId, unsigned int ulFolderId, struct searchCriteria **lppSearchCriteria, unsigned int *lpulSearchFlags);

    // Get current search results for a folder. Simply a database query, nothing more.
    virtual ECRESULT GetSearchResults(unsigned int ulStoreId, unsigned int ulFolderId, unsigned int *lpulCount, unsigned int **lppObjIdList);

    // Queue a messages change that should be processed to update the search folders
    virtual ECRESULT UpdateSearchFolders(unsigned int ulStoreId, unsigned int ulFolderId, unsigned int ulObjId, ECKeyTable::UpdateType ulType);
    
    // Returns erSuccess if the folder is a search folder
    virtual ECRESULT IsSearchFolder(unsigned int ulStoreId, unsigned int ulFolderId);

    // Remove a search folder because it has been deleted. Cancels the search before removing the information. It will
    // remove all results from the database. The folder may not be in mapSearchFolders because it may have been previously
    // cancelled.
    virtual ECRESULT RemoveSearchFolder(unsigned int ulStoreId, unsigned int ulFolderId);

    // Restart all searches. This is a rather heavy operation, and runs synchronously. You have to wait until it
    // has finished.
    virtual ECRESULT RestartSearches();
    
	// Save search criteria to the database
	static ECRESULT SaveSearchCriteria(ECDatabase *lpDatabase, unsigned int ulStoreId, unsigned int ulFolderId, struct searchCriteria *lpSearchCriteria);

private:
    // Process all events in the queue and remove them from the queue.
    virtual ECRESULT FlushEvents();

    // Processes a message change that should be sent to any search folders that are important. This in turn
    // will update the search results views through the Table Manager to update the actual user views.
    virtual ECRESULT ProcessMessageChange(unsigned int ulStoreId, unsigned int ulFolderId, ECObjectTableList *lstObjectIDs, ECKeyTable::UpdateType ulType);

    // Add or update a new search folder that should be monitored. If the search criteria has not changed, this
    // is a restart. This is the only function that starts a new thread. It will cancel any running searches on this
    // folder first.
    virtual ECRESULT AddSearchFolder(unsigned int ulStoreId, unsigned int ulFolderId, bool fStartSearch, struct searchCriteria *lpSearchCriteria);
    
    // Cancel a search. This means that the search results are 'frozen'. If a search thread is running, it is cancelled.
    // After a search has been cancelled, we can ignore any updates for that folder, so it is removed from the list
    // of searches. (but the results remain in the database). We also have to remember this fact in the database because
    // after a server restart, the search should still be 'stopped'.
    virtual ECRESULT CancelSearchFolder(unsigned int ulStoreID, unsigned int ulFolderId);
    
    // Does the actual search. Adds information in the database, and sends updates through the table manager to
    // previously opened tables. This is called only from the search thread and from RestartSearches().
    virtual ECRESULT Search(unsigned int ulStoreId, unsigned int ulFolderId, struct searchCriteria *lpSearchCriteria, bool *lpbCancel, bool bNotify = true);

    // Get the state of a search folder; it may be rebuilding (thread running), running (no thread) or stopped (no entry in mapSearchFolders)
    virtual ECRESULT GetState(unsigned int ulStoreId, unsigned int ulFolderId, unsigned int *lpulState);

    // Search thread entrypoint. Simply a wrapper for Search(), and has code to do thread deregistration.
    static void* SearchThread(void *lpParam);

    // Functions to do things in the database
    virtual ECRESULT ResetResults(unsigned int ulStoreId, unsigned int ulFolderId);
    virtual ECRESULT AddResults(unsigned int ulStoreId, unsigned int ulFolderId, unsigned int ulObjId, unsigned int ulFlags);
    virtual ECRESULT DeleteResults(unsigned int ulStoreId, unsigned int ulFolderId, unsigned int ulObjId);
    virtual ECRESULT GetResults(unsigned int ulStoreId, unsigned int ulFolderId, unsigned int *lpulCount, unsigned int **lppulObjIdList);
    virtual ECRESULT SetStatus(unsigned int ulStoreId, unsigned int ulFolderId, unsigned int ulStatus);

    // Functions to load/save search criteria to the database
    virtual ECRESULT LoadSearchCriteria(unsigned int ulStoreId, unsigned int ulFolderId, struct searchCriteria **lppSearchCriteria);
    virtual ECRESULT SaveSearchCriteria(unsigned int ulStoreId, unsigned int ulFolderId, struct searchCriteria *lpSearchCriteria);

    // Main processing thread
    static void * ProcessThread(void *lpSearchFolders);

    // Map StoreID -> SearchFolderId -> SearchCriteria
    // Because searchfolders only work within a store, this allows us to skip 99% of all
    // search folders during UpdateSearchFolders (depending on how many users you have)
    pthread_mutex_t m_mutexMapSearchFolders;
    STOREFOLDERIDSEARCH m_mapSearchFolders;

    // Pthread condition to signal a thread exit
    pthread_cond_t m_condThreadExited;

    ECDatabaseFactory *m_lpDatabaseFactory;
    ECSessionManager *m_lpSessionManager;
    ECLogger *m_lpLogger;

    // List of change events
    std::list<EVENT> m_lstEvents;
    pthread_mutex_t m_mutexEvents;
    pthread_cond_t m_condEvents;
    
    // Change processing thread
    pthread_t m_threadProcess;
    
    // Exit request for processing thread
    bool m_bExitThread;
};

#endif
