/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

// ECSecurity.cpp: implementation of the ECSecurity class.
//
//////////////////////////////////////////////////////////////////////

#include "platform.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <dirent.h>

#include "ECDatabaseUtils.h"
#include "ECDatabase.h"
#include "ECSessionManager.h"
#include "ECSession.h"

#include "ECDefs.h"
#include "ECSecurity.h"

#include "stringutil.h"
#include "Trace.h"
#include "Zarafa.h"
#include "md5.h"

#include "ECDefs.h"
#include <mapidefs.h>
#include <mapicode.h>
#include <mapitags.h>

#include <mapiext.h>

#include <stdarg.h>

#include <openssl/ssl.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/bio.h>

#include <algorithm>

#include "ECUserManagement.h"
#include "SOAPUtils.h"
#include "edkmdb.h"
#include "ECDBDef.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
/*
 *  -+[Store]
 *   |-+[Inbox]
 *   | |--[mailfolder1]
 *   |
 *   |--[Agenda]
 *   |--[Tasks]
 *   |
 *  -+[Store_michel]
 *   |-+[Inbox]
 *   | |--[mailfolder1]
 *   |
 *   |--[Agenda]
 *   |--[Tasks]
 *
 */

ECSecurity::ECSecurity(ECSession *lpSession, ECConfig *lpConfig, ECLogger *lpLogger)
{
	m_lpSession = lpSession;
	m_lpConfig = lpConfig;
	m_lpLogger = lpLogger;
	m_lpGroups = NULL;
	m_lpViewCompanies = NULL;
	m_lpAdminCompanies = NULL;
	m_ulUserID = 0;
	m_ulCompanyID = 0;
}

ECSecurity::~ECSecurity()
{
    if(m_lpGroups)
        delete m_lpGroups;
	if(m_lpViewCompanies)
		delete m_lpViewCompanies;
	if(m_lpAdminCompanies)
		delete m_lpAdminCompanies;
}

ECRESULT ECSecurity::SetUserContext(unsigned int ulUserId)
{
	ECRESULT er = erSuccess;
	ECUserManagement *lpUserManagement = m_lpSession->GetUserManagement();

	m_ulUserID = ulUserId;

	er = lpUserManagement->GetObjectDetailsAndSync(m_ulUserID, &m_details);
	if(er != erSuccess)
		goto exit;

    // Get the company we're assigned to
    if(m_lpSession->GetSessionManager()->IsHostedSupported()) {
		string company = m_details.GetPropString(OB_PROP_S_COMPANYNAME);
		/* FIXME: When offline server supports companies correctly,
		 * we should remove the ifdef and fully block logging in
		 * without a valid company. */
		if (company.empty() && ulUserId != ZARAFA_UID_SYSTEM) {
			er = ZARAFA_E_NO_ACCESS;
			goto exit;
		}

    	er = lpUserManagement->ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY, (char *)company.c_str(),
		    										&m_ulCompanyID);
        if (er != erSuccess)
		    goto exit;
    } else {
        m_ulCompanyID = 0;
    }

	/*
	 * Don't initialize m_lpGroups, m_lpViewCompanies and m_lpAdminCompanies
	 * We should wait with that until the first time we actually use it,
	 * this will save quite a lot of LDAP queries since often we don't
	 * even need the list at all.
	 */

exit:
	return er;
}

ECRESULT ECSecurity::GetGroupsForUser(unsigned int ulUserId, std::list<localobjectdetails_t> **lppGroups)
{
	ECRESULT er = erSuccess;
	std::list<localobjectdetails_t> *lpGroups = NULL;
	std::list<localobjectdetails_t>::iterator iterGroups;
	std::list<localobjectdetails_t>::iterator iterGroupsRemove;

	/* Gets the current user's membership information.
	 * This means you will be in the same groups until you login again */
	er = m_lpSession->GetUserManagement()->GetParentObjectsOfObjectAndSync(OBJECTRELATION_GROUP_MEMBER, USEROBJECT_TYPE_GROUP,
																		   ulUserId, &lpGroups, USERMANAGEMENT_IDS_ONLY);
	if (er != erSuccess)
		goto exit;

	/* A user is only member of a group when he can also view the group */
	for (iterGroups = lpGroups->begin(); iterGroups != lpGroups->end(); ) {
		if (IsUserObjectVisible(iterGroups->ulId) != erSuccess) {
			iterGroupsRemove = iterGroups;
			iterGroups++;

			lpGroups->erase(iterGroupsRemove);
		} else {
			iterGroups++;
		}
	}

	*lppGroups = lpGroups;

exit:
	return er;
}

// Return the bitmask of permissions for an object
ECRESULT ECSecurity::GetObjectPermission(unsigned int ulObjId, unsigned int* lpulRights)
{
	ECRESULT		er = erSuccess;
	unsigned int	i = 0;
	list<localobjectdetails_t>::iterator iterGroups;
	struct rightsArray *lpRights = NULL;
	unsigned int ulCompanyId;
    unsigned		ulCurObj = ulObjId;
	bool 			bFoundACL = false;
	unsigned int	ulDepth = 0;

	if(lpulRights)
    	*lpulRights = 0;

	// Get the deepest GRANT ACL that applies to this user or groups that this user is in
	// WARNING we totally ignore DENY ACL's here. This means that the deepest GRANT counts. In practice
	// this doesn't matter because GRANTmask = ~DENYmask.
	while(true)
	{
		if(m_lpSession->GetSessionManager()->GetCacheManager()->GetACLs(ulCurObj, &lpRights) == erSuccess) {
		    // This object has ACL's, check if any of them are for this user

		    for(i=0;i<lpRights->__size;i++) {
		        if(lpRights->__ptr[i].ulType == ACCESS_TYPE_GRANT && lpRights->__ptr[i].ulUserid == m_ulUserID) {
		            *lpulRights |= lpRights->__ptr[i].ulRights;
		            bFoundACL = true;
                }
            }

            // Check for the company we are in and add the permissions
            er = GetUserCompany(&ulCompanyId);
            if (er == erSuccess) {
                for (i = 0; i < lpRights->__size; i++) {
                    if (lpRights->__ptr[i].ulType == ACCESS_TYPE_GRANT && lpRights->__ptr[i].ulUserid == ulCompanyId) {
                        *lpulRights |= lpRights->__ptr[i].ulRights;
                        bFoundACL = true;
                    }
				}
            }

            // Also check for groups that we are in, and add those permissions
            if(m_lpGroups || GetGroupsForUser(m_ulUserID, &m_lpGroups) == erSuccess) {
    		    for(iterGroups = m_lpGroups->begin(); iterGroups != m_lpGroups->end(); iterGroups++) {
	    	        for(i=0;i<lpRights->__size;i++) {
		                if(lpRights->__ptr[i].ulType == ACCESS_TYPE_GRANT && lpRights->__ptr[i].ulUserid == iterGroups->ulId) {
		                    *lpulRights |= lpRights->__ptr[i].ulRights;
		                    bFoundACL = true;
                        }
                    }
		        }
            }
		}

		if(lpRights)
		{
			FreeRightsArray(lpRights);
			lpRights = NULL;
		}

        if(bFoundACL) {
            // If any of the ACLs at this level were for us, then use these ACLs.
            break;
        }

		// There were no ACLs or no ACLs for us, go to the parent and try there
		er = m_lpSession->GetSessionManager()->GetCacheManager()->GetParent(ulCurObj, &ulCurObj);
		if(er != erSuccess) {
		    // No more parents, break (with ulRights = 0)
		    er = erSuccess;
			goto exit;
		}

		if(ulCurObj == 0) { // Unsaved attachment
			er = erSuccess;
			goto exit;
		}
		
		// This can really only happen if you have a broken tree in the database, eg a record which has
		// parent == id. To break out of the loop we limit the depth to 64 which is very deep in practice. This means
		// that you never have any rights for folders that are more than 64 levels of folders away from their ACL ..
		ulDepth++;
		
		if(ulDepth > 64) {
			er = erSuccess;
			goto exit;
		}
	}

exit:

    if(lpRights)
        FreeRightsArray(lpRights);

	return er;
}

ECRESULT ECSecurity::HaveObjectPermission(unsigned int ulObjId, unsigned int ulACLMask)
{
	ECRESULT		er = ZARAFA_E_NO_ACCESS;
	unsigned int	ulRights = 0;

	er = GetObjectPermission(ulObjId, &ulRights);

	if(er != erSuccess)
	    goto exit;

    if((ulRights & ulACLMask) > 0) {
        er = erSuccess;
    } else {
        er = ZARAFA_E_NO_ACCESS;
    }

exit:
	return er;
}

ECRESULT ECSecurity::IsOwner(unsigned int ulObjId)
{
	ECRESULT		er;
	unsigned int	ulOwner = 0;

	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetOwner(ulObjId, &ulOwner);

	if(er != erSuccess || ulOwner != m_ulUserID)
	{
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	er = erSuccess;

exit:
	return er;
}

ECRESULT ECSecurity::GetOwner(unsigned int ulObjId, unsigned int *lpulOwnerId)
{
	ECRESULT		er = erSuccess;

	// Default setting
	*lpulOwnerId = 0;

	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetOwner(ulObjId, lpulOwnerId);
	if(er != erSuccess)
	{
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

exit:

	return er;
}

ECRESULT ECSecurity::CheckPermission(unsigned int ulObjId, unsigned int ulecRights)
{
	ECRESULT		er = ZARAFA_E_NO_ACCESS;
	bool			bOwnerFound = false;
	unsigned int	ulACL = 0;
	int				nCheckType = 0;

	// Is the current user the owner of the store
	if(IsStoreOwner(ulObjId) == erSuccess) {
		er = erSuccess;
		goto exit;
	}

	// is current user the owner of the object
	if(IsOwner(ulObjId) == erSuccess)
	{
		bOwnerFound = true;
		if(ulecRights == ecSecurityFolderVisible || ulecRights == ecSecurityRead || ulecRights == ecSecurityCreate) {
			er = erSuccess;
			goto exit;
		}
	}

	// Since this is the most complicated check, do this one last
	if(ulObjId == 0 || IsObjectAdmin(ulObjId) == erSuccess) {
		er = erSuccess;
		goto exit;
	}

	ulACL = 0;
	switch(ulecRights){
		case ecSecurityRead:// 1
			ulACL |= ecRightsReadAny;
			nCheckType = 1;
			break;
		case ecSecurityCreate:// 2
			ulACL |= ecRightsCreate;
			nCheckType = 1;
			break;
		case ecSecurityEdit:// 3
			ulACL |= ecRightsEditAny;
			if(bOwnerFound)
				ulACL |= ecRightsEditOwned;
			nCheckType = 1;
			break;
		case ecSecurityDelete:// 4
			ulACL |= ecRightsDeleteAny;
			if(bOwnerFound)
				ulACL |= ecRightsDeleteOwned;
			nCheckType = 1;
			break;
		case ecSecurityCreateFolder:// 5
			ulACL |= ecRightsCreateSubfolder;
			nCheckType = 1;
			break;
		case ecSecurityFolderVisible:// 6
			ulACL |= ecRightsFolderVisible;
			nCheckType = 1;
			break;
		case ecSecurityFolderAccess: // 7
			if(bOwnerFound == false)
				ulACL |= ecRightsFolderAccess;
			nCheckType = 1;
			break;
		case ecSecurityOwner:// 8
			nCheckType = 2;
			break;
		case ecSecurityAdmin:// 9
			nCheckType = 3;
 			break;
		default:
			nCheckType = 0;//  No rights
		break;
	}

	if(nCheckType == 1) { //Get the acl of the object
		if(ulACL == 0) {
			// No ACLs required, so grant access
			er = erSuccess;
			goto exit;
		}
		er = HaveObjectPermission(ulObjId, ulACL);
	} else if(nCheckType == 2) {// Is owner ?
		if(bOwnerFound){ er = erSuccess; }
	} else if(nCheckType == 3) { // Is admin?
	    // FIXME no point in doing this, due to IsObjectAdmin above
		if(IsObjectAdmin(ulObjId) != erSuccess)
			er = ZARAFA_E_NO_ACCESS;
		else
			er = erSuccess;
	}

exit:
	if(er != erSuccess)
		TRACE_INTERNAL(TRACE_ENTRY,"Security","ECSecurity::CheckPermission","object=%d, rights=%d",ulObjId, ulecRights );

	return er;
}

ECRESULT ECSecurity::GetRights(unsigned int objid, int ulType, struct rightsArray *lpsRightArray)
{
	ECRESULT			er = ZARAFA_E_NO_ACCESS;
	DB_RESULT			lpDBResult = NULL;
	DB_ROW				lpDBRow = NULL;
	ECDatabase			*lpDatabase = m_lpSession->GetDatabase();
	std::string			strQuery;
	unsigned int		ulCount = 0;
	unsigned int		i=0;
	std::string			strExternId;

	unsigned int		ulMapiType = 0;
	userobject_type_t	sObjType;

	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if (lpsRightArray == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	//FIXME: Not secure
	strQuery = "SELECT a.id, a.type, a.rights FROM acl AS a WHERE a.hierarchy_id="+stringify(objid);

	if(ulType == ACCESS_TYPE_DENIED)
		strQuery += " AND a.type="+stringify(ACCESS_TYPE_DENIED);
	else if(ulType == ACCESS_TYPE_GRANT)
		strQuery += " AND a.type="+stringify(ACCESS_TYPE_GRANT);
	// else ACCESS_TYPE_DENIED and ACCESS_TYPE_GRANT

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	ulCount = lpDatabase->GetNumRows(lpDBResult);
	if(ulCount > 0)
	{
		lpsRightArray->__ptr = new struct rights[ulCount];
		lpsRightArray->__size = ulCount;

		memset(lpsRightArray->__ptr, 0, sizeof(struct rights) * ulCount);

		for(i=0; i < ulCount; i++){
			lpDBRow = lpDatabase->FetchRow(lpDBResult);

			if(lpDBRow == NULL) {
				er = ZARAFA_E_DATABASE_ERROR;
				goto exit;
			}

			lpsRightArray->__ptr[i].ulUserid= atoi(lpDBRow[0]);
			lpsRightArray->__ptr[i].ulType	= atoi(lpDBRow[1]);
			lpsRightArray->__ptr[i].ulRights= atoi(lpDBRow[2]);
			lpsRightArray->__ptr[i].ulState	= RIGHT_NORMAL;

			// do not use internal id's with the cache
			if (lpsRightArray->__ptr[i].ulUserid == ZARAFA_UID_SYSTEM) {
				strExternId.clear();
				sObjType = USEROBJECT_TYPE_USER;
			} else if (lpsRightArray->__ptr[i].ulUserid == ZARAFA_UID_EVERYONE) {
				strExternId.clear();
				sObjType = USEROBJECT_TYPE_GROUP;
			} else {
				er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserExternId(lpsRightArray->__ptr[i].ulUserid, &sObjType, &strExternId, NULL);
				if (er != erSuccess)
					goto exit;
			}

			er = TypeToMAPIType(sObjType, &ulMapiType);
			if (er != erSuccess)
				goto exit;

			er = ABIDToEntryID(NULL, lpsRightArray->__ptr[i].ulUserid, strExternId, ulMapiType, &lpsRightArray->__ptr[i].sUserId);
			if (er != erSuccess)
				goto exit;
		}
	}else{
		lpsRightArray->__ptr = NULL;
		lpsRightArray->__size = 0;
	}

	er = erSuccess;

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECSecurity::SetRights(unsigned int objid, struct rightsArray *lpsRightArray){
	ECRESULT			er = erSuccess;
	unsigned int		i;
	std::string			strQueryNew, strQueryDeniedNew;
	std::string			strQueryModify, strQueryDeniedModify;
	std::string			strQueryDelete, strQueryDeniedDelete;
	unsigned int		ulDeniedRights=0;
	ECDatabase			*lpDatabase = m_lpSession->GetDatabase();
	unsigned int		ulUserId = 0;
	std::string			strExternId;

	unsigned int      ulMapiType;
	userobject_type_t sObjType;

	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if (lpsRightArray == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// Invalidate cache for this object
	m_lpSession->GetSessionManager()->GetCacheManager()->Update(fnevObjectModified, objid);

	for(i=0; i< lpsRightArray->__size; i++)
	{
		// FIXME: check for each object if it belongs to the store we're logged into (except for admin)

		// Get the correct local id
		if (lpsRightArray->__ptr[i].sUserId.__size > 0 && lpsRightArray->__ptr[i].sUserId.__ptr != NULL)
		{
			er = ABEntryIDToID(&lpsRightArray->__ptr[i].sUserId, &ulUserId, &strExternId, &ulMapiType);
			if (er != erSuccess)
				goto exit;

			// internal user/group doesn't have an externid
			if (!strExternId.empty())
			{
				er = MAPITypeToType(ulMapiType, &sObjType);
				if (er != erSuccess)
					goto exit;

				er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserObject(strExternId, sObjType, NULL, &ulUserId, NULL);
				if (er != erSuccess && sObjType == USEROBJECT_TYPE_USER) {
    				er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserObject(strExternId, USEROBJECT_TYPE_NONACTIVE, NULL, &ulUserId, NULL);
				}
				if (er != erSuccess)
					goto exit;
			}
		}
		else
			ulUserId = lpsRightArray->__ptr[i].ulUserid;

		// Auto create denied rules
		ulDeniedRights = lpsRightArray->__ptr[i].ulRights^ecRightsAllMask;
		if(lpsRightArray->__ptr[i].ulRights & ecRightsEditAny)
			ulDeniedRights&=~ecRightsEditOwned;
		else if(lpsRightArray->__ptr[i].ulRights & ecRightsEditOwned){
			ulDeniedRights&=~ecRightsEditOwned;
			ulDeniedRights|=ecRightsEditAny;
		}

		if(lpsRightArray->__ptr[i].ulRights & ecRightsDeleteAny)
			ulDeniedRights&=~ecRightsDeleteOwned;
		else if(lpsRightArray->__ptr[i].ulRights & ecRightsDeleteOwned)
		{
			ulDeniedRights&=~ecRightsDeleteOwned;
			ulDeniedRights|=ecRightsDeleteAny;
		}

		if(lpsRightArray->__ptr[i].ulState == RIGHT_NORMAL)
		{
			// Do nothing...
		}
		else if((lpsRightArray->__ptr[i].ulState & RIGHT_NEW) || (lpsRightArray->__ptr[i].ulState & RIGHT_MODIFY))
		{
			strQueryNew = "REPLACE INTO acl (id, hierarchy_id, type, rights) VALUES ";
			strQueryNew+="("+stringify(ulUserId)+","+stringify(objid)+","+stringify(lpsRightArray->__ptr[i].ulType)+","+stringify(lpsRightArray->__ptr[i].ulRights)+")";

			er = lpDatabase->DoInsert(strQueryNew);
			if(er != erSuccess)
			    goto exit;

			if(lpsRightArray->__ptr[i].ulState & RIGHT_AUTOUPDATE_DENIED){
				strQueryNew = "REPLACE INTO acl (id, hierarchy_id, type, rights) VALUES ";
				strQueryNew+=" ("+stringify(ulUserId)+","+stringify(objid)+","+stringify(ACCESS_TYPE_DENIED)+","+stringify(ulDeniedRights)+")";

				er = lpDatabase->DoInsert(strQueryNew);
				if(er != erSuccess)
				   	goto exit;

			}

		}
		else if(lpsRightArray->__ptr[i].ulState & RIGHT_DELETED)
		{
			strQueryDelete = "DELETE FROM acl WHERE ";
			strQueryDelete+="(hierarchy_id="+stringify(objid)+" AND id="+stringify(ulUserId)+" AND type="+stringify(lpsRightArray->__ptr[i].ulType)+")";

			er = lpDatabase->DoDelete(strQueryDelete);
			if(er != erSuccess)
				goto exit;

			if(lpsRightArray->__ptr[i].ulState & RIGHT_AUTOUPDATE_DENIED) {
				strQueryDelete = "DELETE FROM acl WHERE ";
				strQueryDelete+="(hierarchy_id="+stringify(objid)+" AND id="+stringify(ulUserId)+" AND type="+stringify(ACCESS_TYPE_DENIED)+")";

				er = lpDatabase->DoDelete(strQueryDelete);
				if(er != erSuccess)
					goto exit;
			}

		}else{
			// a Hacker ?
		}

	}


	er = erSuccess;

exit:
	return er;
}

ECRESULT ECSecurity::GetUserCompany(unsigned int *lpulCompanyId)
{
	*lpulCompanyId = m_ulCompanyID;
	return erSuccess;
}

ECRESULT ECSecurity::GetViewableCompanyIds(list<localobjectdetails_t> **lppObjects, unsigned int ulFlags)
{
	ECRESULT er = erSuccess;

	/*
	 * We have the viewable companies stored in our cache,
	 * if it is present use that, otherwise just create a
	 * new one.
	 * NOTE: We always request GetViewableCompanies with 0 as ulFlags,
	 * this because we are caching the list here and some callers might
	 * want all details while others will only want the id's.
	 */
	if (!m_lpViewCompanies) {
		er = GetViewableCompanies(&m_lpViewCompanies, 0);
		if (er != erSuccess)
			goto exit;
	}

	*lppObjects = new list<localobjectdetails_t>(*m_lpViewCompanies);

exit:
	return er;
}

ECRESULT ECSecurity::IsUserObjectVisible(unsigned int ulUserObjectId)
{
	ECRESULT er = erSuccess;
	list<localobjectdetails_t>::iterator iterCompany;
	userobject_type_t ulUserType;
	unsigned int ulCompanyId;

	if (ulUserObjectId == 0 ||
		ulUserObjectId == m_ulUserID ||
		ulUserObjectId == m_ulCompanyID ||
		ulUserObjectId == ZARAFA_UID_SYSTEM ||
		ulUserObjectId == ZARAFA_UID_EVERYONE ||
		m_details.GetPropInt(OB_PROP_I_ADMINLEVEL) == ADMIN_LEVEL_SYSADMIN ||
		!m_lpSession->GetSessionManager()->IsHostedSupported()) {
		goto exit;
	}

	er = m_lpSession->GetUserManagement()->GetItemObjectType(ulUserObjectId, &ulUserType);
	if (er != erSuccess)
		goto exit;

	switch (ulUserType) {
	case USEROBJECT_TYPE_COMPANY:
		ulCompanyId = ulUserObjectId;
		break;
	case USEROBJECT_TYPE_GROUP:
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
	case USEROBJECT_TYPE_ADDRESSLIST:
		er = m_lpSession->GetUserManagement()->GetItemCompany(ulUserObjectId, ulUserType, &ulCompanyId);
		if (er != erSuccess)
			goto exit;
		break;
	default:
		er = ZARAFA_E_UNKNOWN_OBJECT;
		goto exit;
	}

	if (!m_lpViewCompanies) {
		er = GetViewableCompanies(&m_lpViewCompanies);
		if (er != erSuccess)
			goto exit;
	}

	for (iterCompany = m_lpViewCompanies->begin(); iterCompany != m_lpViewCompanies->end(); iterCompany++) {
		if (iterCompany->ulId == ulCompanyId) {
			er = erSuccess;
			goto exit;
		}
	}

	/* Item was not found */
	er = ZARAFA_E_NOT_FOUND;

exit:
	return er;
}

ECRESULT ECSecurity::GetViewableCompanies(list<localobjectdetails_t> **lppObjects, unsigned int ulFlags)
{
	ECRESULT er = erSuccess;
	list<localobjectdetails_t> *lpObjects = NULL;
	objectdetails_t details;

	if (m_details.GetPropInt(OB_PROP_I_ADMINLEVEL) == ADMIN_LEVEL_SYSADMIN)
		er = m_lpSession->GetUserManagement()->GetCompanyObjectListAndSync(USEROBJECT_TYPE_COMPANY, 0, &lpObjects, ulFlags);
	else
		er = m_lpSession->GetUserManagement()->GetParentObjectsOfObjectAndSync(OBJECTRELATION_COMPANY_VIEW,
																			   USEROBJECT_TYPE_COMPANY,
																			   m_ulCompanyID, &lpObjects,
																			   ulFlags);
	if (er != erSuccess) {
		/* Whatever the error might be, it only indicates we
		 * are not allowed to view _other_ companyspaces.
		 * It doesn't restrict us from viewing our own... */
		lpObjects = new list<localobjectdetails_t>();
		er = erSuccess;
	}

	/* We are going to insert the requested companyID to the list as well,
	 * this way we guarentee that _all_ viewable companies are in the list.
	 * And we can use sort() and unique() to prevent duplicate entries while
	 * making sure the we can savely handle things when the current company
	 * is either added or not present in the RemoteViewableCompanies. */
	if (m_ulCompanyID != 0) {
		if (!(ulFlags & USERMANAGEMENT_IDS_ONLY)) {
			er = m_lpSession->GetUserManagement()->GetObjectDetailsAndSync(m_ulCompanyID, &details);
			if (er != erSuccess)
				goto exit;
		} else {
				details = objectdetails_t(USEROBJECT_TYPE_COMPANY);
		}

		lpObjects->push_back(localobjectdetails_t(m_ulCompanyID, details));
	}

	lpObjects->sort();
	lpObjects->unique();

	*lppObjects = lpObjects;

exit:
	if (er != erSuccess && lpObjects)
		delete lpObjects;

	return er;
}

ECRESULT ECSecurity::GetAdminCompanies(list<localobjectdetails_t> **lppObjects, unsigned int ulFlags)
{
	ECRESULT er = erSuccess;
	list<localobjectdetails_t> *lpObjects = NULL;
	list<localobjectdetails_t>::iterator iterObjects;
	list<localobjectdetails_t>::iterator iterObjectsRemove;

	if (m_details.GetPropInt(OB_PROP_I_ADMINLEVEL) == ADMIN_LEVEL_SYSADMIN)
		er = m_lpSession->GetUserManagement()->GetCompanyObjectListAndSync(USEROBJECT_TYPE_COMPANY, 0, &lpObjects, ulFlags);
	else
		er = m_lpSession->GetUserManagement()->GetParentObjectsOfObjectAndSync(OBJECTRELATION_COMPANY_ADMIN,
																			   USEROBJECT_TYPE_COMPANY,
																			   m_ulUserID, &lpObjects,
																			   ulFlags);
	if (er != erSuccess)
		goto exit;

	/* A user is only admin over an company when he has privileges to view the company */
	for (iterObjects = lpObjects->begin(); iterObjects != lpObjects->end(); ) {
		if (IsUserObjectVisible(iterObjects->ulId) != erSuccess) {
			iterObjectsRemove = iterObjects;
			iterObjects++;
			lpObjects->erase(iterObjectsRemove);
		} else {
			iterObjects++;
		}
	}

	*lppObjects = lpObjects;

exit:
	if (er != erSuccess && lpObjects)
		delete lpObjects;

	return er;
}

unsigned int ECSecurity::GetUserId(unsigned int ulObjId)
{
	ECRESULT er = erSuccess;

	unsigned int ulUserId = this->m_ulUserID;

	if(ulObjId != 0 && IsObjectAdmin(ulObjId) == erSuccess){
		er = GetOwner(ulObjId, &ulUserId);
		if(er != erSuccess)
			ulUserId = this->m_ulUserID;
	}

	return ulUserId;
}

ECRESULT ECSecurity::IsStoreOwner(unsigned int ulObjId)
{
	ECRESULT er = erSuccess;
	unsigned int ulStoreId = 0;

	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetStore(ulObjId, &ulStoreId, NULL);
	if(er != erSuccess)
		goto exit;

	er = IsOwner(ulStoreId);
	if(er != erSuccess)
		goto exit;

exit:
	return er;
}

int ECSecurity::GetAdminLevel()
{
	return m_details.GetPropInt(OB_PROP_I_ADMINLEVEL);
}

ECRESULT ECSecurity::GetStoreOwner(unsigned int ulObjId, unsigned int* lpulOwnerId)
{
	ECRESULT er = erSuccess;
	unsigned int ulStoreId = 0;

	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetStore(ulObjId, &ulStoreId, NULL);
	if(er != erSuccess)
		goto exit;

	er = GetOwner(ulStoreId, lpulOwnerId);
	if(er != erSuccess)
		goto exit;

exit:
	return er;
}

// TODO: this function should be renamed to IsAdminOfUserObject(id) or something like that
ECRESULT ECSecurity::IsUserObjectAdmin(unsigned int ulUserObjectId)
{
	ECRESULT er = ZARAFA_E_NO_ACCESS;
	list<localobjectdetails_t>::iterator objectIter;
	unsigned int ulObjectCompanyId;
	objectdetails_t objectdetails;
	userobject_type_t ulObjectType;

	/* Hosted disabled: When admin level is not zero, then the user
	 * is the administrator, otherwise the user isn't and we don't need
	 * to look any further. */
	if (!m_lpSession->GetSessionManager()->IsHostedSupported()) {
		if (m_details.GetPropInt(OB_PROP_I_ADMINLEVEL) != 0)
			er = erSuccess;
		goto exit;
	}

	/* If hosted is enabled, system administrators are administrator over all users. */
	if (m_details.GetPropInt(OB_PROP_I_ADMINLEVEL) == ADMIN_LEVEL_SYSADMIN) {
		er = erSuccess;
		goto exit;
	}

	/*
	 * Determine to which company the user belongs
	 */
	er = m_lpSession->GetUserManagement()->GetItemObjectType(ulUserObjectId, &ulObjectType);
	if (er != erSuccess)
		goto exit;

	switch (ulObjectType) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
	case USEROBJECT_TYPE_GROUP:
		er = m_lpSession->GetUserManagement()->GetItemCompany(ulUserObjectId, ulObjectType, &ulObjectCompanyId);
		if (er != erSuccess)
			goto exit;

		break;
	case USEROBJECT_TYPE_COMPANY:
		ulObjectCompanyId = ulUserObjectId; // Easy, no need to perform further requests. :)
		break;
	default:
		er = ZARAFA_E_UNKNOWN_OBJECT;
		goto exit;
	}

	/*
	 * If ulCompanyId is the company where the logged in user belongs to,
	 * then the only thing we need to check is the "isadmin" boolean.
	 */
	if (m_ulCompanyID == ulObjectCompanyId) {
		if (m_details.GetPropInt(OB_PROP_I_ADMINLEVEL) != 0)
			er = erSuccess;
		else
			er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	if (!m_lpAdminCompanies) {
		er = GetAdminCompanies(&m_lpAdminCompanies, USERMANAGEMENT_IDS_ONLY);
		if (er != erSuccess)
			goto exit;
	}

	for (objectIter = m_lpAdminCompanies->begin(); objectIter != m_lpAdminCompanies->end(); objectIter++) {
		if (objectIter->ulId == ulObjectCompanyId) {
			er = erSuccess;
			goto exit;
		}
	}

	/* Item was not found, so no access */
	er = ZARAFA_E_NO_ACCESS;

exit:
	return er;
}

ECRESULT  ECSecurity::IsObjectAdmin(unsigned int ulObjectId)
{
	ECRESULT er = ZARAFA_E_NO_ACCESS;
	unsigned int ulOwner;

	/*
	 * Request the ownership if the object.
	 */
	er = GetStoreOwner(ulObjectId, &ulOwner);
	if (er != erSuccess)
		goto exit;

	er = IsUserObjectAdmin(ulOwner);

exit:
	return er;
}

ECRESULT ECSecurity::GetStoreSize(unsigned int ulObjId, long long* lpllStoreSize)
{
	ECRESULT		er = erSuccess;
	ECDatabase		*lpDatabase = NULL;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;

	std::string		strQuery;
	unsigned int	ulStore;
	long long		llStoreSize = 0;

	lpDatabase = m_lpSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetStore(ulObjId, &ulStore, NULL);
	if(er != erSuccess)
		goto exit;

	strQuery = "SELECT val_longint FROM properties WHERE storeid="+stringify(ulStore)+" AND tag="+stringify(PROP_ID(PR_MESSAGE_SIZE_EXTENDED))+" AND type="+stringify(PROP_TYPE(PR_MESSAGE_SIZE_EXTENDED)) + " AND hierarchyid="+stringify(ulStore);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) != 1) {
		er = erSuccess;//ZARAFA_E_NOT_FOUND; // ignore (when creating a new store)
		*lpllStoreSize = 0;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if(lpDBRow == NULL || lpDBRow[0] == NULL) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	llStoreSize = _atoi64(lpDBRow[0]);

	// Free stuff
	if(lpDBResult){ lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

	*lpllStoreSize = llStoreSize;

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECSecurity::GetUserSize(unsigned int ulUserId, long long* lpllUserSize)
{
	ECRESULT		er = erSuccess;
	ECDatabase		*lpDatabase = NULL;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;

	std::string		strQuery;
	long long		llUserSize = 0;

	lpDatabase = m_lpSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	strQuery =
		"SELECT p.val_longint "
		"FROM properties AS p "
		"JOIN stores AS s "
			"ON s.hierarchy_id = p.hierarchyid "
		"WHERE "
			"s.user_id = " + stringify(ulUserId) + " " +
			"AND p.storeid = p.hierarchyid " +
			"AND p.tag = " + stringify(PROP_ID(PR_MESSAGE_SIZE_EXTENDED)) + " "
			"AND p.type = " + stringify(PROP_TYPE(PR_MESSAGE_SIZE_EXTENDED));
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

	if (lpDatabase->GetNumRows(lpDBResult) != 1) {
		*lpllUserSize = 0;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if (lpDBRow == NULL) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if (lpDBRow[0] == NULL)
		llUserSize = 0;
	else
		llUserSize = _atoi64(lpDBRow[0]);

	*lpllUserSize = llUserSize;

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECSecurity::GetCompanySize(unsigned int ulCompanyId, long long* lpllCompanySize)
{
	ECRESULT		er = erSuccess;
	ECDatabase		*lpDatabase = NULL;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;

	std::string		strQuery;
	long long		llCompanySize = 0;

	lpDatabase = m_lpSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	strQuery =
		"SELECT SUM(p.val_longint) "
		"FROM users AS u "
		"JOIN hierarchy AS h "
			"ON u.id = h.owner "
		"JOIN properties AS p "
			"ON h.id = p.hierarchyid "
		"WHERE "
			"(u.company = " + stringify(ulCompanyId) + " "  /* Member of company */
			 "OR u.id = " + stringify(ulCompanyId) + ") "   /* Public store of company */
			"AND p.storeid = p.hierarchyid "				/* Store root entry */
			"AND p.tag = " + stringify(PROP_ID(PR_MESSAGE_SIZE_EXTENDED)) + " "
			"AND p.type = " + stringify(PROP_TYPE(PR_MESSAGE_SIZE_EXTENDED));
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) != 1) {
		*lpllCompanySize = 0;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if(lpDBRow == NULL) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if (lpDBRow[0] == NULL)
		llCompanySize = 0;
	else
		llCompanySize = _atoi64(lpDBRow[0]);

	*lpllCompanySize = llCompanySize;

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECSecurity::CheckQuota(unsigned int ulStoreId, long long llStoreSize, eQuotaStatus* lpQuotaStatus)
{
	ECRESULT er = erSuccess;
	unsigned int ulOwnerId;

	// Get the store owner
	er = GetStoreOwner(ulStoreId, &ulOwnerId);
	if(er != erSuccess)
		goto exit;

	er = CheckUserQuota(ulOwnerId, llStoreSize, lpQuotaStatus);

exit:
	return er;
}

ECRESULT ECSecurity::CheckUserQuota(unsigned int ulUserId, long long llStoreSize, eQuotaStatus *lpQuotaStatus)
{
	ECRESULT		er = erSuccess;
	quotadetails_t	quotadetails;

	if (ulUserId == ZARAFA_UID_EVERYONE) {
		/* Publicly owned stores are never over quota.
		 * But do publicly owned stores actually exist since the owner is either a user or company */
		*lpQuotaStatus = QUOTA_OK;
		goto exit;
	}

	er = GetUserQuota(ulUserId, &quotadetails);
	if (er != erSuccess)
		goto exit;

	// check the options
	if(quotadetails.llHardSize > 0 && llStoreSize >= quotadetails.llHardSize)
		*lpQuotaStatus = QUOTA_HARDLIMIT;
	else if(quotadetails.llSoftSize > 0 && llStoreSize >= quotadetails.llSoftSize)
		*lpQuotaStatus = QUOTA_SOFTLIMIT;
	else if(quotadetails.llWarnSize > 0 && llStoreSize >= quotadetails.llWarnSize)
		*lpQuotaStatus = QUOTA_WARN;
	else
		*lpQuotaStatus = QUOTA_OK;

exit:
	return er;
}

ECRESULT ECSecurity::GetUserQuota(unsigned int ulUserId, quotadetails_t *lpDetails)
{
	ECRESULT er = erSuccess;
	char* lpszWarnQuota = NULL;
	char* lpszSoftQuota = NULL;
	char* lpszHardQuota = NULL;
	quotadetails_t quotadetails;
	userobject_type_t ulType;
	unsigned int ulCompanyId;

	if (!lpDetails) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = m_lpSession->GetUserManagement()->GetItemObjectType(ulUserId, &ulType);
	if (er != erSuccess)
		goto exit;

	er = m_lpSession->GetUserManagement()->GetQuotaDetailsAndSync(ulUserId, &quotadetails);
	if (er != erSuccess)
		goto exit;

	/* When the default quota boolean is set, we need to look at the next quota level */
	if (!quotadetails.bUseDefaultQuota)
		goto exit;

	/* Request default quota values from company level if enabled */
	if (ulType == USEROBJECT_TYPE_USER || ulType == USEROBJECT_TYPE_NONACTIVE) {
		er = m_lpSession->GetUserManagement()->GetItemCompany(ulUserId, ulType, &ulCompanyId);
		if (er != erSuccess)
			goto exit;

		if(ulCompanyId) {
			er = m_lpSession->GetUserManagement()->GetQuotaDetailsAndSync(ulCompanyId, &quotadetails, true);
			if (er == erSuccess && !quotadetails.bUseDefaultQuota)
				goto exit; /* On failure, or when we should use the default, we're done */
			er = erSuccess;
		}
	}

	/* No information from company, the last level we can check is the configuration file */
	if (ulType == USEROBJECT_TYPE_USER || ulType == USEROBJECT_TYPE_NONACTIVE) {
		lpszWarnQuota = m_lpSession->GetSessionManager()->GetConfig()->GetSetting("quota_warn");
		lpszSoftQuota = m_lpSession->GetSessionManager()->GetConfig()->GetSetting("quota_soft");
		lpszHardQuota = m_lpSession->GetSessionManager()->GetConfig()->GetSetting("quota_hard");
	} else if (ulType == USEROBJECT_TYPE_COMPANY) {
		lpszWarnQuota = m_lpSession->GetSessionManager()->GetConfig()->GetSetting("companyquota_warn");
		lpszSoftQuota = m_lpSession->GetSessionManager()->GetConfig()->GetSetting("companyquota_soft");
		lpszHardQuota = m_lpSession->GetSessionManager()->GetConfig()->GetSetting("companyquota_hard");
	}

	quotadetails.bUseDefaultQuota = true;
	quotadetails.bIsUserDefaultQuota = false;
	if (lpszWarnQuota)
		quotadetails.llWarnSize = _atoi64(lpszWarnQuota) * 1024 * 1024;
	if (lpszSoftQuota)
		quotadetails.llSoftSize = _atoi64(lpszSoftQuota) * 1024 * 1024;
	if (lpszHardQuota)
		quotadetails.llHardSize = _atoi64(lpszHardQuota) * 1024 * 1024;

exit:
	if (er == erSuccess)
		*lpDetails = quotadetails;

	return er;
}

ECRESULT ECSecurity::GetUsername(std::string *lpstrUsername)
{
	if (m_ulUserID)
		*lpstrUsername = m_details.GetPropString(OB_PROP_S_LOGIN);
	else
		*lpstrUsername = ZARAFA_SYSTEM_USER;
	return erSuccess;
}
