/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

/* Returns the rows for a contents- or hierarchytable
 *
 * objtype == MAPI_MESSAGE, then contents table
 * objtype == MAPI_MESSAGE, flags == MAPI_ASSOCIATED, then associated contents table
 * objtype == MAPI_FOLDER, then hierarchy table
 *
 * Tables are generated from SQL in the following way:
 *
 * Tables are constructed by joining the hierarchy table with the property table multiple
 * times, once for each requested property (column). Because each column of each row can always have
 * only one or zero records, a unique index is created on the property table, indexing (hierarchyid, type, tag).
 *
 * This means that for each cell that we request, the index needs to be accessed by the SQL
 * engine only once, which makes the actual query extremely fast.
 *
 * In tests, this has shown to required around 60ms for 30 rows and 10 columns from a table of 10000
 * rows. Also, this is a O(n log n) operation and therefore not prone to large scaling problems. (Yay!)
 * (with respect to the amount of columns, it is O(n), but that's quite constant, and so is the
 * actual amount of rows requested per query (also O(n)).
 *
 */

#include "soapH.h"
#include "ZarafaCode.h"

#include <mapidefs.h>
#include <mapitags.h>

#include "Zarafa.h"
#include "ECDatabaseUtils.h"
#include "ECKeyTable.h"
#include "ECGenProps.h"
#include "ECStoreObjectTable.h"
#include "ECSecurity.h"
#include "SOAPUtils.h"
#include "stringutil.h"
#include "ECIConv.h"

#include "Trace.h"
#include "ECSessionManager.h"

#include "ECSession.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

bool IsTruncatableType(unsigned int ulTag)
{
    switch(PROP_TYPE(ulTag)) {
        case PT_STRING8:
        case PT_UNICODE:
        case PT_BINARY:
            return true;
        default:
            return false;
    }
    
    return false;
}

ECStoreObjectTable::ECStoreObjectTable(ECSession *lpSession, unsigned int ulStoreId, GUID *lpGuid, unsigned int ulFolderId,unsigned int ulObjType, unsigned int ulFlags) : ECGenericObjectTable(lpSession, ulObjType)
{
	ECODStore* lpODStore = new ECODStore;

	lpODStore->ulStoreId = ulStoreId;
	lpODStore->ulFolderId = ulFolderId;
	lpODStore->ulObjType = ulObjType;
	lpODStore->ulFlags = ulFlags;

	if(lpGuid) {
		lpODStore->lpGuid = new GUID;
		*lpODStore->lpGuid = *lpGuid;
	} else {
		lpODStore->lpGuid = NULL; // When NULL is specified, we get the store ID & guid for each seperate row
	}

	// Set dataobject
	m_lpObjectData = lpODStore;

	// Set callback function for queryrowdata
	m_lpfnQueryRowData = QueryRowData;
	
	ulPermission = 0;
	fPermissionRead = false;
}

ECStoreObjectTable::~ECStoreObjectTable()
{
	if(m_lpObjectData) {
		ECODStore* lpODStore = (ECODStore*)m_lpObjectData;
		if(lpODStore->lpGuid)
			delete lpODStore->lpGuid;

		delete lpODStore;
	}
}

ECRESULT ECStoreObjectTable::GetColumnsAll(ECListInt* lplstProps)
{
	ECRESULT		er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	std::string		strQuery;
	ECDatabase*		lpDatabase = lpSession->GetDatabase();
	ECODStore*		lpODStore = (ECODStore*)m_lpObjectData;
	ULONG			ulPropID = 0;
	ECObjectTableMap::iterator		iterObjects;

	pthread_mutex_lock(&m_hLock);

	ASSERT(lplstProps != NULL);
	
	//List always emtpy
	lplstProps->clear();

	if(!mapObjects.empty())
	{
		if (!lpDatabase) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		// Properties
		strQuery = "SELECT DISTINCT properties.tag, properties.type FROM properties JOIN hierarchy ON properties.hierarchyid = hierarchy.id WHERE hierarchy.id IN (";

		for(iterObjects = mapObjects.begin(); iterObjects != mapObjects.end(); iterObjects++) {
			strQuery += stringify(iterObjects->first.ulObjId);
			strQuery += ",";
		}

		// Remove trailing comma
		strQuery.resize(strQuery.size()-1);

		strQuery += ")";

		// Large object properties
		strQuery += " UNION SELECT DISTINCT singleinstances.tag, " + stringify(PT_BINARY) + " FROM `singleinstances` WHERE singleinstances.hierarchyid IN (";

		for(iterObjects = mapObjects.begin(); iterObjects != mapObjects.end(); iterObjects++) {
			strQuery += stringify(iterObjects->first.ulObjId);
			strQuery += ",";
		}

		strQuery.resize(strQuery.size()-1);

		strQuery += ")";

		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;
		
		// Put the results into a STL list
		while((lpDBRow = lpDatabase->FetchRow(lpDBResult)) != NULL) {

			if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL)
				continue;

			ulPropID = atoi(lpDBRow[0]);
			// Exclude properties in a table
			if (ulPropID == PROP_ID(0x10130102/*PR_HTML*/) || ulPropID == PROP_ID(PR_RTF_COMPRESSED) )
				continue;

			lplstProps->push_back(PROP_TAG(atoi(lpDBRow[1]), ulPropID));
		}
	}// if(!mapObjects.empty())

	// Add some generated and standard properties
	lplstProps->push_back(PR_ENTRYID);
	lplstProps->push_back(PR_INSTANCE_KEY);
	lplstProps->push_back(PR_RECORD_KEY);
	lplstProps->push_back(PR_OBJECT_TYPE);
	lplstProps->push_back(PR_LAST_MODIFICATION_TIME);
	lplstProps->push_back(PR_CREATION_TIME);
	lplstProps->push_back(PR_PARENT_ENTRYID);
	lplstProps->push_back(PR_MAPPING_SIGNATURE);

	// Add properties only in the contents tables
	if(lpODStore->ulObjType == MAPI_MESSAGE && (lpODStore->ulFlags&MSGFLAG_ASSOCIATED) == 0)
	{
		lplstProps->push_back(PR_MSG_STATUS);
	}
	
	lplstProps->push_back(PR_ACCESS_LEVEL);

	//FIXME: only in folder or message table	
	lplstProps->push_back(PR_ACCESS);

exit:
	pthread_mutex_unlock(&m_hLock);

	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECStoreObjectTable::ReloadTableMVData(ECObjectTableList* lplistRows, ECListInt* lplistMVPropTag)
{
	ECRESULT			er = erSuccess;
	DB_RESULT			lpDBResult = NULL;
	DB_ROW				lpDBRow = NULL;
	std::string			strQuery;
	std::string			strColName;
	ECDatabase*			lpDatabase = lpSession->GetDatabase();	
	int					j;
	ECListIntIterator	iterListMVPropTag;
	sObjectTableKey		sRowItem;
	size_t				cListRows = 0;

	ECObjectTableList::iterator	iterListRows;

	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	ASSERT(lplistMVPropTag->size() <2); //FIXME: Limit of one 1 MV column

	// scan for MV-props and add rows
	strQuery = "SELECT h.id, orderid FROM hierarchy as h";
	j = 0;
	for(iterListMVPropTag = lplistMVPropTag->begin(); iterListMVPropTag != lplistMVPropTag->end(); iterListMVPropTag++)
	{
		strColName = "col"+stringify(j);
		strQuery += " LEFT JOIN mvproperties as "+strColName+" ON h.id="+strColName+".hierarchyid AND "+strColName+".tag="+stringify(PROP_ID(*iterListMVPropTag))+" AND "+strColName+".type="+stringify(PROP_TYPE(*iterListMVPropTag &~MV_INSTANCE));
		j++;
		break; //FIXME: limit 1 column, multi MV cols
	}//for(iterListMVPropTag
		
	strQuery += " WHERE h.id IN(";

	j = 0;
	cListRows = lplistRows->size();
	for(iterListRows = lplistRows->begin(); iterListRows != lplistRows->end(); iterListRows++)
	{
		strQuery += stringify(iterListRows->ulObjId);

		if(j != (int)(cListRows-1))
			strQuery += ",";
		j++;
	}
	strQuery += ") ORDER by id, orderid";

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	while(1)
	{
		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		if(lpDBRow == NULL)
			break;

		sRowItem.ulObjId = atoui(lpDBRow[0]);
		sRowItem.ulOrderId = (lpDBRow[1] == NULL)? 0 : atoi(lpDBRow[1]);

		if(sRowItem.ulOrderId > 0)
			lplistRows->push_back(sRowItem);
	}

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

//
// This function queries the database for value of the specified rows and columns
//
ECRESULT ECStoreObjectTable::QueryRowData(ECGenericObjectTable *lpThis, struct soap *soap, ECSession *lpSession, ECObjectTableList* lpRowList, struct propTagArray *lpsPropTagArray, void* lpObjectData, struct rowSet **lppRowSet, bool bCacheTableData, bool bTableLimit)
{
	ECRESULT		er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	DB_LENGTHS		lpDBLen = NULL;
	std::string		strQuery;
	std::string		strSubquery;
	int				i = 0;
	int				k = 0;

	unsigned int	ulPropTag = 0;
	unsigned int	ulRowStoreId = 0;

	struct rowSet	*lpsRowSet = NULL;
	GUID			sRowGuid;

	ECODStore*		lpODStore = (ECODStore*)lpObjectData;

	ECDatabase		*lpDatabase = lpSession->GetDatabase();	

	std::map<unsigned int, ECObjectTableMap> mapStoreIdObjIds;
	std::map<unsigned int, ECObjectTableMap>::iterator iterStoreIdObjIds;
	ECObjectTableMap::iterator		iterIDs, iterIDsErase;
	ECObjectTableList::iterator		iterRowList;
	sObjectTableKey					sMapKey;

	ECListInt			listMVSortCols;//Other mvprops then normal column set
	ECListInt::iterator	iterListInt;
	ECListInt			listMVIColIds;
	string				strCol;
	
	std::map<unsigned int,unsigned int> mapRowsToCache;

	std::list<unsigned int> propList;

	ASSERT(lpRowList != NULL);

	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	lpsRowSet = s_alloc<rowSet>(soap);
	lpsRowSet->__size = 0;
	lpsRowSet->__ptr = NULL;

	if(lpRowList->empty()) {
		*lppRowSet = lpsRowSet;
		goto exit; // success
	}

	if (lpODStore->ulFlags & EC_TABLE_NOCAP)
		bTableLimit = false;
	
	// We return a square array with all the values
	lpsRowSet->__size = lpRowList->size();
	lpsRowSet->__ptr = s_alloc<propValArray>(soap, lpsRowSet->__size);
	memset(lpsRowSet->__ptr, 0, sizeof(propValArray) * lpsRowSet->__size);

	// Allocate memory for all rows
	for(i=0; i<lpsRowSet->__size; i++) {
		lpsRowSet->__ptr[i].__size = lpsPropTagArray->__size;
		lpsRowSet->__ptr[i].__ptr = s_alloc<propVal>(soap, lpsPropTagArray->__size);
		memset(lpsRowSet->__ptr[i].__ptr, 0, sizeof(propVal) * lpsPropTagArray->__size);
	}


	for(k=0; k < lpsPropTagArray->__size; k++)
	{
		mapStoreIdObjIds.clear();

		if(ECGenProps::GetPropSubstitute(lpsPropTagArray->__ptr[k], &ulPropTag) != hrSuccess)
			ulPropTag = lpsPropTagArray->__ptr[k];
        
        // Fake unicode support by retrieving PT_STRING8 instead of PT_UNICODE
        if(PROP_TYPE(ulPropTag) == PT_UNICODE)
            ulPropTag = PROP_TAG(PT_STRING8, PROP_ID(ulPropTag));

		// Loop through the cache items 
		for(i=0, iterRowList = lpRowList->begin(); iterRowList != lpRowList->end(); iterRowList++, i++)
		{
			// Get StoreId if needed
			if(lpODStore->lpGuid == NULL) {
				// No store specified, so determine the store ID & guid from the object id
				lpSession->GetSessionManager()->GetCacheManager()->GetStore(iterRowList->ulObjId, &ulRowStoreId, &sRowGuid);
			} else {
				ulRowStoreId = lpODStore->ulStoreId;
			}
			
			if(iterRowList->ulObjId == 0) {
                if(lpThis->GetPropCategory(soap, lpsPropTagArray->__ptr[k], *iterRowList, &lpsRowSet->__ptr[i].__ptr[k]) == erSuccess)
    			    continue;
                else {
                    // Other properties are not found
                    lpsRowSet->__ptr[i].__ptr[k].__union = SOAP_UNION_propValData_ul;
                    lpsRowSet->__ptr[i].__ptr[k].Value.ul = ZARAFA_E_NOT_FOUND;
                    lpsRowSet->__ptr[i].__ptr[k].ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpsPropTagArray->__ptr[k]));
                    continue;
                }
                    
            }
            
			
			// Handle PR_DEPTH
			if(lpODStore->ulObjType == MAPI_FOLDER && lpsPropTagArray->__ptr[k] == PR_DEPTH && ECGenProps::GetComputedDepth(soap, lpSession,lpODStore->ulFolderId,iterRowList->ulObjId, &lpsRowSet->__ptr[i].__ptr[k]) == erSuccess)
				continue;
				
            if(lpODStore->ulObjType == MAPI_MESSAGE && lpsPropTagArray->__ptr[k] == PR_DEPTH) {
                lpsRowSet->__ptr[i].__ptr[k].__union = SOAP_UNION_propValData_ul;
                lpsRowSet->__ptr[i].__ptr[k].ulPropTag = PR_DEPTH;
                if(lpThis)
                    lpsRowSet->__ptr[i].__ptr[k].Value.ul = lpThis->GetCategories();
                else
                    lpsRowSet->__ptr[i].__ptr[k].Value.ul = 0;
                continue;
            }

			if(ECGenProps::GetPropComputedUncached(soap, lpSession, lpsPropTagArray->__ptr[k], iterRowList->ulObjId, iterRowList->ulOrderId, (lpODStore->lpGuid)?lpODStore->ulStoreId:ulRowStoreId, &lpsRowSet->__ptr[i].__ptr[k]) ==erSuccess)
				continue;

			if((bTableLimit == false && IsTruncatableType(lpsPropTagArray->__ptr[k])) || lpSession->GetSessionManager()->GetCacheManager()->GetCell(&(*iterRowList), lpsPropTagArray->__ptr[k], &lpsRowSet->__ptr[i].__ptr[k], soap) != erSuccess) 
			{
				// We don't support the same row being requested twice in the same call. WARNING this assertion has a sideeffect
				// due to the use of the [] operator; the value will be initialized. This shouldn't be a problem though.
				ASSERT(mapStoreIdObjIds[ulRowStoreId].find(*iterRowList) == mapStoreIdObjIds[ulRowStoreId].end());

				// item is not in the cache, or the complete data is requested
				// remember this id and the rownr it should be placed at.
				// thus, for each store: Map ID -> rownr
				
				mapStoreIdObjIds[ulRowStoreId].insert(ECObjectTableMap::value_type(*iterRowList, i));
			}
		}

		// Get column from database for the remaining non-cached items
		for (iterStoreIdObjIds = mapStoreIdObjIds.begin(); iterStoreIdObjIds != mapStoreIdObjIds.end(); iterStoreIdObjIds++)
		{

			if((ulPropTag&MV_FLAG) == MV_FLAG) {
				// MV(i) properties
				if((ulPropTag&MV_INSTANCE) == MV_INSTANCE) {
					strQuery = "SELECT hierarchyid,orderid,"+GetColName(PROP_TYPE(ulPropTag), "mvproperties", bTableLimit)+" FROM mvproperties WHERE ";
					strQuery+= "tag="+stringify(PROP_ID(ulPropTag))+" AND type="+stringify(PROP_TYPE(ulPropTag)&~MV_INSTANCE)+" AND (";
				}else if(ECGenProps::GetMVPropSubquery(ulPropTag, strCol) == erSuccess) {
					strQuery = "SELECT id,("+strCol+") FROM hierarchy WHERE id IN (";
				}else {
					// should never happen
					ASSERT(FALSE);
				}

				for(iterIDs = (*iterStoreIdObjIds).second.begin(); iterIDs != (*iterStoreIdObjIds).second.end(); iterIDs++) {
					if(iterIDs != (*iterStoreIdObjIds).second.begin()) {
						if(ulPropTag&MV_INSTANCE)
							strQuery+=" OR ";
						else
							strQuery+=",";
					}

					if(ulPropTag&MV_INSTANCE)
						strQuery+="(hierarchyid=" + stringify(iterIDs->first.ulObjId)+ " AND orderid="+ stringify(iterIDs->first.ulOrderId)+")";
					else
						strQuery+=stringify(iterIDs->first.ulObjId);
				}
				strQuery+= ")";
			}else {
				// normal property
				if(ECGenProps::GetPropSubquery(ulPropTag, strCol) == erSuccess) {
					strQuery = "SELECT id,("+strCol+") FROM hierarchy WHERE id IN (";
				}else {
					strQuery = "SELECT hierarchyid,"+GetColName(PROP_TYPE(ulPropTag), "properties", bTableLimit)+" FROM properties IGNORE INDEX(hi) WHERE storeid="+stringify((*iterStoreIdObjIds).first)+" AND tag="+stringify(PROP_ID(ulPropTag))+" AND type="+stringify(PROP_TYPE(ulPropTag))+" AND hierarchyid IN (";
				}

				for (iterIDs = (*iterStoreIdObjIds).second.begin(); iterIDs != (*iterStoreIdObjIds).second.end(); iterIDs++) {
					if(iterIDs != (*iterStoreIdObjIds).second.begin())
						strQuery+=",";
					strQuery+=stringify(iterIDs->first.ulObjId);
				}				
				strQuery+= ")";
			}


			// we can only use streaming of no other SQL commands are mixed with the current query
			// cache may run a query, so if we need the cache, we disable streaming, otherwise enable it.
			er = lpDatabase->DoSelect(strQuery, &lpDBResult, lpODStore->lpGuid != NULL && lpODStore->ulFolderId != 0);
			if(er != erSuccess)
				goto exit;

			while(1) {
				lpDBRow = lpDatabase->FetchRow(lpDBResult);
				lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);
				if(lpDBRow == NULL) {
					break;
				}
				// lpDBRow[0] = hierarchyid
				//Normal property and MV property
				// lpDBRow[1] = column data
				//MVI property
				// lpDBRow[1] = orderid
				// lpDBRow[2] = column data

				if(lpDBRow[0] == NULL || ( (ulPropTag&MV_INSTANCE) == MV_INSTANCE && lpDBRow[1] == NULL) ) {
					// should never happen
					ASSERT(FALSE);
					continue;
				}

				sMapKey.ulObjId = atoui(lpDBRow[0]);
				if((ulPropTag&MV_INSTANCE) == MV_INSTANCE)
					sMapKey.ulOrderId = atoui(lpDBRow[1]); //MVI properties
				else
					sMapKey.ulOrderId = 0;

				// Find out which row number the row goes into ..
				//iterIDs = mapIDs.find(sMapKey);
				iterIDs = (*iterStoreIdObjIds).second.lower_bound(sMapKey);
				if(iterIDs == (*iterStoreIdObjIds).second.end() || iterIDs->first.ulObjId != sMapKey.ulObjId ||
					((ulPropTag&MV_INSTANCE) == MV_INSTANCE && iterIDs->first.ulOrderId!=sMapKey.ulOrderId))
				{
					// should never happen
					ASSERT(FALSE);
					continue;
				}

				while(iterIDs != (*iterStoreIdObjIds).second.end() && iterIDs->first.ulObjId == sMapKey.ulObjId && ((ulPropTag&MV_INSTANCE) != MV_INSTANCE || ((ulPropTag&MV_INSTANCE) == MV_INSTANCE && iterIDs->first.ulOrderId == sMapKey.ulOrderId)) )
				{
					// Get StoreId ( this if for the master outgoing queue )
					if(lpODStore->lpGuid == NULL) {
						// No store specified, so determine the store ID & guid from the object id
						lpSession->GetSessionManager()->GetCacheManager()->GetStore(sMapKey.ulObjId, &ulRowStoreId, &sRowGuid);
					}

					//
					er = CopyDatabaseRowToSOAPPropVal(lpODStore, soap, lpSession, (sObjectTableKey*)&iterIDs->first, lpsPropTagArray->__ptr[k], lpDBRow, lpDBLen, (lpODStore->lpGuid)?lpODStore->ulStoreId:ulRowStoreId, (lpODStore->lpGuid)?lpODStore->lpGuid:&sRowGuid, &lpsRowSet->__ptr[iterIDs->second].__ptr[k]);
					if(er == erSuccess)
					{
						// Cell ok, add to cache
						if(bCacheTableData)
							lpSession->GetSessionManager()->GetCacheManager()->SetCell((sObjectTableKey*)&iterIDs->first, lpsPropTagArray->__ptr[k], &lpsRowSet->__ptr[iterIDs->second].__ptr[k]);

						// Remove it from the list, we no longer need it
						iterIDsErase = iterIDs++;
						(*iterStoreIdObjIds).second.erase(iterIDsErase);
					}else {
						ASSERT(FALSE);
						goto exit;
					}
					
				}
			} //while(1)

			// Set not-found rows to MAPI_E_NOT_FOUND (all rows left in mapIDs)
			for(iterIDs = (*iterStoreIdObjIds).second.begin(); iterIDs != (*iterStoreIdObjIds).second.end(); iterIDs++)
			{
				// Get StoreId ( this if for the master outgoing queue )
				if(lpODStore->lpGuid == NULL) {
					// No store specified, so determine the store ID & guid from the object id
					lpSession->GetSessionManager()->GetCacheManager()->GetStore(iterIDs->first.ulObjId, &ulRowStoreId, &sRowGuid);
				}

				if(ECGenProps::IsPropComputed(ulPropTag, lpODStore->ulObjType) == erSuccess)
					CopyDatabaseRowToSOAPPropVal(lpODStore, soap, lpSession,(sObjectTableKey*)&iterIDs->first, ulPropTag, NULL, NULL, (lpODStore->lpGuid)?lpODStore->ulStoreId:ulRowStoreId, (lpODStore->lpGuid)?lpODStore->lpGuid:&sRowGuid, &lpsRowSet->__ptr[iterIDs->second].__ptr[k]);
				else
					CopyEmptyCellToSOAPPropVal(soap, ulPropTag, (lpODStore->lpGuid)?lpODStore->ulStoreId:ulRowStoreId, (lpODStore->lpGuid)?lpODStore->lpGuid:&sRowGuid, iterIDs->first.ulObjId, MAPI_MESSAGE, &lpsRowSet->__ptr[iterIDs->second].__ptr[k]);

				// Cell ok, add to cache
				if(bCacheTableData)
					lpSession->GetSessionManager()->GetCacheManager()->SetCell((sObjectTableKey*)&iterIDs->first, ulPropTag, &lpsRowSet->__ptr[iterIDs->second].__ptr[k]);
					
			}

			// We now have filled in all rows with data

			if(lpDBResult) {
				lpDatabase->FreeResult(lpDBResult);
				lpDBResult = NULL;
			}
		}// for (each store)

	}

	er = erSuccess;
	*lppRowSet = lpsRowSet;

exit:

	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT ECStoreObjectTable::CopyEmptyCellToSOAPPropVal(struct soap *soap, unsigned int ulPropTag, unsigned int ulStoreId, GUID *lpGuid, unsigned int ulObjId, unsigned int ulObjType, struct propVal *lpPropVal)
{
	ECRESULT	er = erSuccess;

	lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(ulPropTag));
	lpPropVal->__union = SOAP_UNION_propValData_ul;
	lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;

	return er;
}

//
// Copies one database row/col to a propval/cell
//
ECRESULT ECStoreObjectTable::CopyDatabaseRowToSOAPPropVal(ECODStore *lpODStore, struct soap *soap, ECSession* lpSession, sObjectTableKey*	lpsMapKey, unsigned int ulPropTag, DB_ROW lpDBRow, DB_LENGTHS lpDBLen, unsigned int ulStoreId, GUID *lpGuid, struct propVal *lpPropVal)
{
	ECRESULT	er = erSuccess;
	int			j;

	unsigned int ulParentId = 0;
	unsigned int ulObjType = 0;
	unsigned int ulFlags = 0;
	unsigned int ulOwner = 0;
	unsigned int ulLastPos;
	unsigned int ulColId = 0;
	std::string strData;
	
	if(lpSession == NULL || lpsMapKey == NULL)
	{
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}
	
	if(lpODStore->ulFolderId) {
	    ulObjType = lpODStore->ulObjType;
	    ulFlags = lpODStore->ulFlags;
	    ulParentId = lpODStore->ulFolderId;
	    ulOwner = 0; // unused here
	} else {
    	er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(lpsMapKey->ulObjId, &ulParentId, &ulOwner, &ulFlags, &ulObjType);
    	if(er != erSuccess)
		    goto exit;
    }

	if((ulPropTag&MV_INSTANCE) == MV_INSTANCE)
		ulColId = 2;
	else
		ulColId = 1;

	// Try to retrieve the data through the computed properties generator. If this
	// doesn't work, just get the value from the data
	if(ECGenProps::GetPropComputed(soap, lpSession, ulPropTag, lpsMapKey->ulObjId, lpsMapKey->ulOrderId, ulParentId, ulObjType, ulFlags, ulStoreId, (lpDBRow)?lpDBRow[ulColId]:NULL, lpPropVal) != erSuccess) {

		lpPropVal->ulPropTag = ulPropTag;
		if((PROP_TYPE(lpPropVal->ulPropTag)&MVI_FLAG) == MVI_FLAG)
			lpPropVal->ulPropTag&=~MVI_FLAG;

		if(lpDBRow == NULL || lpDBRow[ulColId] == NULL) {
			// A column was not found
			lpPropVal->ulPropTag = PROP_TAG(PT_ERROR,PROP_ID(lpPropVal->ulPropTag));
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;

			goto exit;
		}
		
		switch(PROP_TYPE(lpPropVal->ulPropTag)) {
		case PT_I2:
 			lpPropVal->__union = SOAP_UNION_propValData_i;
			lpPropVal->Value.i = (short)atoi(lpDBRow[ulColId]);
			break;
		case PT_LONG: // and PT_ULONG
 			lpPropVal->__union = SOAP_UNION_propValData_ul;
			lpPropVal->Value.ul = atoui(lpDBRow[ulColId]);
			break;
		case PT_R4:
 			lpPropVal->__union = SOAP_UNION_propValData_flt;
			lpPropVal->Value.flt = (float)atof(lpDBRow[ulColId]);
			break;
		case PT_BOOLEAN:
 			lpPropVal->__union = SOAP_UNION_propValData_b;
			lpPropVal->Value.b = atoi(lpDBRow[ulColId]) ? true : false;
			break;
		case PT_DOUBLE:
 			lpPropVal->__union = SOAP_UNION_propValData_dbl;
			lpPropVal->Value.dbl = atof(lpDBRow[ulColId]);
			break;
		case PT_CURRENCY:
			lpPropVal->Value.hilo = s_alloc<hiloLong>(soap);
 			lpPropVal->__union = SOAP_UNION_propValData_hilo;
			lpPropVal->Value.hilo->hi = atoi(lpDBRow[ulColId]);
			lpPropVal->Value.hilo->lo = atoui(lpDBRow[ulColId+1]);
			break;
		case PT_APPTIME:
 			lpPropVal->__union = SOAP_UNION_propValData_dbl;
			lpPropVal->Value.dbl = atof(lpDBRow[ulColId]);
			break;
		case PT_SYSTIME:
			lpPropVal->Value.hilo = s_alloc<hiloLong>(soap);
 			lpPropVal->__union = SOAP_UNION_propValData_hilo;
			lpPropVal->Value.hilo->hi = atoi(lpDBRow[ulColId]);
			lpPropVal->Value.hilo->lo = atoui(lpDBRow[ulColId+1]);
			break;
		case PT_I8:
 			lpPropVal->__union = SOAP_UNION_propValData_li;
			lpPropVal->Value.li = _atoi64(lpDBRow[ulColId]);
			break;
		case PT_UNICODE: {
		    ECIConv iconv("utf-8", "windows-1252");
		    std::string utf8 = iconv.convert(std::string(lpDBRow[ulColId], lpDBLen[ulColId]));
			lpPropVal->Value.lpszA = s_alloc<char>(soap, utf8.size()+1);
 			lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			lstrcpy(lpPropVal->Value.lpszA, utf8.c_str());
			break;
        }
		case PT_STRING8:
			lpPropVal->Value.lpszA = s_alloc<char>(soap, lpDBLen[ulColId]+1);
 			lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			lstrcpy(lpPropVal->Value.lpszA, (lpDBRow[ulColId])?lpDBRow[ulColId]:"");
			break;
		case PT_CLSID:
		case PT_BINARY:
            lpPropVal->Value.bin = s_alloc<xsd__base64Binary>(soap);
            if(lpDBLen[ulColId] > 0) {
                lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, lpDBLen[ulColId]);
    			lpPropVal->Value.bin->__size = lpDBLen[ulColId];
				memcpy(lpPropVal->Value.bin->__ptr, lpDBRow[ulColId], lpDBLen[ulColId]);
            } else {
                lpPropVal->Value.bin->__ptr = NULL;
                lpPropVal->Value.bin->__size = 0;
            }

 			lpPropVal->__union = SOAP_UNION_propValData_bin;
			break;
		case PT_MV_I2:
			lpPropVal->__union = SOAP_UNION_propValData_mvi;
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvi.__size);

			if(lpPropVal->Value.mvi.__size == 0) {
				lpPropVal->Value.mvi.__ptr = NULL;
				break;
			}

			lpPropVal->Value.mvi.__ptr = s_alloc<short int>(soap, lpPropVal->Value.mvi.__size);
			
			for(j=0; j < lpPropVal->Value.mvi.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);

				lpPropVal->Value.mvi.__ptr[j] = (short)atoi(strData.c_str());
			}
			break;
		case PT_MV_LONG:
			lpPropVal->__union = SOAP_UNION_propValData_mvl;
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvl.__size);

			if(lpPropVal->Value.mvl.__size == 0) {
				lpPropVal->Value.mvl.__ptr = NULL;
				break;
			}

			lpPropVal->Value.mvl.__ptr = s_alloc<unsigned int>(soap, lpPropVal->Value.mvl.__size);
			
			for(j=0; j < lpPropVal->Value.mvl.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);

				lpPropVal->Value.mvl.__ptr[j] = atoui((char*)strData.c_str());
			}
			break;
		case PT_MV_R4:
			lpPropVal->__union = SOAP_UNION_propValData_mvflt;
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvflt.__size);

			if(lpPropVal->Value.mvflt.__size == 0) {
				lpPropVal->Value.mvflt.__ptr = NULL;
				break;
			}

			lpPropVal->Value.mvflt.__ptr = s_alloc<float>(soap, lpPropVal->Value.mvflt.__size);
			
			for(j=0; j < lpPropVal->Value.mvflt.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);

				lpPropVal->Value.mvflt.__ptr[j] = (float)atof(strData.c_str());
			}
			break;
		case PT_MV_APPTIME:
		case PT_MV_DOUBLE:
			lpPropVal->__union = SOAP_UNION_propValData_mvdbl;
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvdbl.__size);

			if(lpPropVal->Value.mvdbl.__size == 0) {
				lpPropVal->Value.mvdbl.__ptr = NULL;
				break;
			}

			lpPropVal->Value.mvdbl.__ptr = s_alloc<double>(soap, lpPropVal->Value.mvdbl.__size);
			
			for(j=0; j < lpPropVal->Value.mvdbl.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);

				lpPropVal->Value.mvdbl.__ptr[j] = atof(strData.c_str());
			}
			break;
		case PT_MV_CURRENCY:
		case PT_MV_SYSTIME:
			lpPropVal->__union = SOAP_UNION_propValData_mvhilo;
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvhilo.__size);

			if(lpPropVal->Value.mvhilo.__size == 0) {
				lpPropVal->Value.mvhilo.__ptr = NULL;
				break;
			}

			lpPropVal->Value.mvhilo.__ptr = s_alloc<hiloLong>(soap, lpPropVal->Value.mvhilo.__size);

			//hi
			for(j=0; j < lpPropVal->Value.mvhilo.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);

				lpPropVal->Value.mvhilo.__ptr[j].hi = atoui((char*)strData.c_str());
			}
			//Lo
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvhilo.__size);
			for(j=0; j < lpPropVal->Value.mvhilo.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);

				lpPropVal->Value.mvhilo.__ptr[j].lo = atoi(strData.c_str());
			}
			break;
		case PT_MV_STRING8:
		case PT_MV_UNICODE:
			lpPropVal->__union = SOAP_UNION_propValData_mvszA;
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvszA.__size);

			if(lpPropVal->Value.mvszA.__size == 0) {
				lpPropVal->Value.mvszA.__ptr = NULL;
				break;
			}

			lpPropVal->Value.mvszA.__ptr = s_alloc<char *>(soap, lpPropVal->Value.mvszA.__size);
			
			for(j=0; j < lpPropVal->Value.mvszA.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);
				lpPropVal->Value.mvszA.__ptr[j] = s_alloc<char>(soap, strData.size()+1);
				lstrcpy(lpPropVal->Value.mvszA.__ptr[j], strData.c_str());
			}
			break;
		case PT_MV_I8:
			lpPropVal->__union = SOAP_UNION_propValData_mvli;
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvli.__size);

			if(lpPropVal->Value.mvli.__size == 0) {
				lpPropVal->Value.mvli.__ptr = NULL;
				break;
			}

			lpPropVal->Value.mvli.__ptr = s_alloc<long long>(soap, lpPropVal->Value.mvli.__size);
			
			for(j=0; j < lpPropVal->Value.mvli.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);

				lpPropVal->Value.mvli.__ptr[j] = _atoi64(strData.c_str());
			}
			break;
		case PT_MV_CLSID:
		case PT_MV_BINARY:
			lpPropVal->__union = SOAP_UNION_propValData_mvbin;
			ulLastPos = 0;
			ParseMVPropCount(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &lpPropVal->Value.mvbin.__size);

			if(lpPropVal->Value.mvbin.__size == 0) {
				lpPropVal->Value.mvbin.__ptr = NULL;
				break;
			}

			lpPropVal->Value.mvbin.__ptr = s_alloc<xsd__base64Binary>(soap, lpPropVal->Value.mvbin.__size);
			
			for(j=0; j < lpPropVal->Value.mvbin.__size; j++)
			{
				ParseMVProp(lpDBRow[ulColId], lpDBLen[ulColId], &ulLastPos, &strData);
				lpPropVal->Value.mvbin.__ptr[j].__size = strData.size();
				if(lpPropVal->Value.mvbin.__ptr[j].__size > 0)
				{
					lpPropVal->Value.mvbin.__ptr[j].__ptr = s_alloc<unsigned char>(soap, lpPropVal->Value.mvbin.__ptr[j].__size);
					memcpy(lpPropVal->Value.mvbin.__ptr[j].__ptr, strData.c_str(), sizeof(unsigned char) * lpPropVal->Value.mvbin.__ptr[j].__size);
				}
				else
					lpPropVal->Value.mvbin.__ptr[j].__ptr = NULL;
			}
			break;
		default: // ignore column
			ASSERT(FALSE);
			lpPropVal->ulPropTag = PROP_TAG(PT_ERROR,PROP_ID(lpPropVal->ulPropTag));
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
			break;
		}
	} 

exit:
	return er;
}


ECRESULT ECStoreObjectTable::GetMVRowCount(unsigned int ulObjId, unsigned int *lpulCount)
{
	ECRESULT er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpRow = NULL;
	std::string		strQuery, strColName;
	int j;
    ECObjectTableList			listRows;
	ECObjectTableList::iterator iterListRows;
	ECObjectTableMap::iterator		iterIDs;
	ECDatabase *lpDatabase = lpSession->GetDatabase();
	ECListInt::iterator	iterListMVPropTag;

	pthread_mutex_lock(&m_hLock);

	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	// scan for MV-props and add rows
	strQuery = "SELECT count(h.id) FROM hierarchy as h";
	j=0;
	for(iterListMVPropTag = m_listMVSortCols.begin(); iterListMVPropTag != m_listMVSortCols.end(); iterListMVPropTag++)
	{
		strColName = "col"+stringify(j);
		strQuery += " LEFT JOIN mvproperties as "+strColName+" ON h.id="+strColName+".hierarchyid AND "+strColName+".tag="+stringify(PROP_ID(*iterListMVPropTag))+" AND "+strColName+".type="+stringify(PROP_TYPE(*iterListMVPropTag &~MV_INSTANCE));
		j++;
	}
	
	strQuery += " WHERE h.id="+stringify(ulObjId)+" ORDER by h.id, orderid";
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;
		
    lpRow = lpDatabase->FetchRow(lpDBResult);
    
    if(lpRow == NULL || lpRow[0] == NULL) {
        er = ZARAFA_E_DATABASE_ERROR;
        goto exit;
    }
	
	*lpulCount = atoi(lpRow[0]);

exit:
	pthread_mutex_unlock(&m_hLock);

	if(lpDBResult)
           lpDatabase->FreeResult(lpDBResult);
	return er;
}

ECRESULT ECStoreObjectTable::Load()
{
    ECRESULT er = erSuccess;
    ECDatabase *lpDatabase = lpSession->GetDatabase();;
    DB_RESULT 	lpDBResult = NULL;
    DB_ROW		lpDBRow = NULL;
    std::string	strQuery;
    ECODStore	*lpData = (ECODStore *)m_lpObjectData;
    sObjectTableKey		sRowItem;
    struct sortOrderArray*      lpsPrevSortOrderArray = NULL;
    
    unsigned int ulFlags = lpData->ulFlags;
    unsigned int ulFolderId = lpData->ulFolderId;
    unsigned int ulObjType = lpData->ulObjType;
    
    std::list<unsigned int> lstObjIds;
    
    if(ulFolderId) {
        // Clear old entries
        mapObjects.clear();
        lpKeyTable->Clear();

        // Load the table with all the objects of type ulObjType and flags ulFlags in container ulParent
        
		strQuery = "SELECT hierarchy.id FROM hierarchy WHERE hierarchy.parent=" + stringify(ulFolderId);

        if(ulObjType == MAPI_MESSAGE)
        {
			strQuery += " AND hierarchy.type = " +  stringify(ulObjType);

            if((ulFlags&MSGFLAG_DELETED) == 0)// Normal message and associated message
                strQuery += " AND hierarchy.flags & "+stringify(MSGFLAG_ASSOCIATED)+" = " + stringify(ulFlags&MSGFLAG_ASSOCIATED) + " AND hierarchy.flags & "+stringify(MSGFLAG_DELETED)+" = 0";
            else
                strQuery += " AND hierarchy.flags & "+stringify(MSGFLAG_ASSOCIATED)+" = " + stringify(ulFlags&MSGFLAG_ASSOCIATED) + " AND hierarchy.flags & "+stringify(MSGFLAG_DELETED)+" = " + stringify(MSGFLAG_DELETED);
            
        }
		else if(ulObjType == MAPI_FOLDER) {
            strQuery += " AND hierarchy.type = " +  stringify(ulObjType);
			strQuery += " AND hierarchy.flags & "+stringify(MSGFLAG_DELETED)+" = " + stringify(ulFlags&MSGFLAG_DELETED);
		}else if(ulObjType == MAPI_MAILUSER) { //Read MAPI_MAILUSER and MAPI_DISTLIST
			strQuery += " AND (hierarchy.type = " +  stringify(ulObjType) + " OR hierarchy.type = " +  stringify(MAPI_DISTLIST) + ")";
		}else {
			 strQuery += " AND hierarchy.type = " +  stringify(ulObjType);
		}

        er = lpDatabase->DoSelect(strQuery, &lpDBResult);
        if(er != erSuccess)
            goto exit;

        while(1) {
            lpDBRow = lpDatabase->FetchRow(lpDBResult);

            if(lpDBRow == NULL)
                break;

            if(lpDBRow[0] == NULL)
                continue;
                
			lstObjIds.push_back(atoi(lpDBRow[0]));
        }

        UpdateRows(ECKeyTable::TABLE_ROW_ADD, &lstObjIds, 0);
        
    }
    
exit:
    if(lpsPrevSortOrderArray)
        FreeSortOrderArray(lpsPrevSortOrderArray);
        
    if(lpDBResult)
        lpDatabase->FreeResult(lpDBResult);
    
    return er;
}

ECRESULT ECStoreObjectTable::CheckPermissions(unsigned int ulObjId)
{
    ECRESULT er = erSuccess;
    unsigned int ulParent = 0;
    unsigned int ulFolderFlags = 0;
    ECODStore	*lpData = (ECODStore *)m_lpObjectData;

    if(m_ulObjType == MAPI_MESSAGE) {
        if(lpData->ulFolderId) {
            // We can either see all messages or none at all. Do this check once only as an optimisation.
            if(!this->fPermissionRead) {
                this->ulPermission = lpSession->GetSecurity()->CheckPermission(lpData->ulFolderId, ecSecurityRead);
                this->fPermissionRead = true;
            }

            er = this->ulPermission;                    
        } else {
            // Get the parent id of the row we're inserting (this is very probably cached from Load())
            er = lpSession->GetSessionManager()->GetCacheManager()->GetParent(ulObjId, &ulParent);
            if(er != erSuccess)
                goto exit;
            
            // This will be cached after the first row in the table is added
            er = lpSession->GetSecurity()->CheckPermission(ulParent, ecSecurityRead);
            if(er != erSuccess)
                goto exit;
        }
    } else if(m_ulObjType == MAPI_FOLDER) {
	    // Check the folder type for searchfolders
		er = lpSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, NULL, NULL, &ulFolderFlags, NULL);
		if(er != erSuccess)
		    goto exit;
		
		if(ulFolderFlags == FOLDER_SEARCH) {
		    // Searchfolders are only visible in the home store
		    if(lpSession->GetSecurity()->IsObjectAdmin(ulObjId) != erSuccess && lpSession->GetSecurity()->IsStoreOwner(ulObjId) != erSuccess) {
				er = ZARAFA_E_NO_ACCESS;
		        goto exit;
            }
		}

        er = lpSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityFolderVisible);
        if(er != erSuccess)
            goto exit;
    }
            
exit:
    return er;

}
