/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapidefs.h>
#include <mapitags.h>
#include <mapiext.h>

#include <map>
#include <algorithm>

#include "Zarafa.h"
#include "stringutil.h"
#include "ECUserManagement.h"
#include "ECSessionManager.h"
#include "ECPluginFactory.h"
#include "ECSecurity.h"
#include "ECIConv.h"
#include "SymmetricCrypt.h"
#include "ECPamAuth.h"
#include "ECKrbAuth.h"

#include "../common/base64.h"

#include "ECICS.h"
#include "ZarafaICS.h"
#include "ECGuid.h"
#include "ECDefs.h"
#include <EMSAbTag.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern pthread_key_t plugin_key;
extern ECSessionManager*	g_lpSessionManager;


static bool execute_script(const char *scriptname, ECLogger *lpLogger, ...) {
	int status;
	va_list v;
	char *envname, *envval;
	char buffer[1024];
	bool rv = true;

	va_start(v, lpLogger);
	/* Set environment */
	for (;;) {
		envname = va_arg(v, char *);
		if (!envname)
			break;
		envval = va_arg(v, char *);
		// SuSE 9.1 segfaults when putenv() is used in a detached thread on the next getenv() call.
		setenv(envname, envval, 1);
	}
	va_end(v);


	// get output from script
	FILE *p = popen(scriptname, "r");
	if (p) {
		while (fgets(buffer, 1024, p)) {
			buffer[strlen(buffer) - 1] = '\0'; // strip enter
			lpLogger->Log(EC_LOGLEVEL_ERROR, "%s: %s", scriptname, buffer);
		}
		status = pclose(p);
	} else {
		status = -1;
	}

	/* Unset environment */
	va_start(v, lpLogger);
	for (;;) {
		envname = va_arg(v, char *);
		if (!envname)
			break;
		envval = va_arg(v, char *);
		// SuSE 9.1 segfaults when putenv() is used in a detached thread on the next getenv() call.
		unsetenv(envname);
	}
	va_end(v);

	if(status != -1) {
#ifdef WEXITSTATUS
		if(WIFEXITED(status)) { /* Child exited by itself */
			if(WEXITSTATUS(status)) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, "Script %s exited with non-zero status %d",
							  scriptname, WEXITSTATUS(status));
				rv = false;
			}
		} else if(WIFSIGNALED(status)) {        /* Child was killed by a signal */
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Script %s was killed by signal %d",
				   scriptname, WTERMSIG(status));
			rv = false;
		} else {                        /* Something strange happened */
			lpLogger->Log(EC_LOGLEVEL_ERROR, string("Script ") + scriptname + " terminated abnormally");
			rv = false;
		}
#else
		if (status)
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Script %s exited with status %d",
						  scriptname, status);
#endif
	} else {
		lpLogger->Log(EC_LOGLEVEL_ERROR, string("System call `system' failed: ") + strerror(errno));
		rv = false;
	}

	if (rv == false) {
		va_start(v, lpLogger);
		for (;;) {
			envname = va_arg(v, char *);
			if (!envname)
				break;
			envval = va_arg(v, char *);
			lpLogger->Log(EC_LOGLEVEL_ERROR, "%s had value %s ", envname, envval);
		}
		va_end(v);
	}

	return rv;
}

// Returns a plugin local to this thread. Works the same as GetThreadLocalDatabase
ECRESULT GetThreadLocalPlugin(ECPluginFactory *lpPluginFactory, UserPlugin **lppPlugin, ECLogger *lpLogger)
{
	ECRESULT er = erSuccess;
	UserPlugin *lpPlugin = NULL;
	std::string error;

	lpPlugin = (UserPlugin *)pthread_getspecific(plugin_key);

	if(lpPlugin == NULL) {
		er = lpPluginFactory->CreateUserPlugin(&lpPlugin);

		if(er != erSuccess) {
		    lpPlugin = NULL;
			lpLogger->Log(EC_LOGLEVEL_FATAL,"Unable to instantiate user plugin");
			goto exit;
		}

		pthread_setspecific(plugin_key, (void *)lpPlugin);
	}

	*lppPlugin = lpPlugin;

exit:
	return er;
}

std::string HostnameFromSoap(struct soap *soap)
{
	/*
	 * Is the client connecting using an IP address or fully qualified hostname?
	 * SOAP knows... SOAP is all-knowing.
	 */
	if (soap->host[0] == '\0') {
		/* Use the endpoint which is something like: http://<ip>:<port>/<path>*/
		std::string ip = soap->endpoint;
		size_t pos;

		pos = ip.find("://");
		if (pos != std::string::npos)
			ip.erase(0, ip.find_first_not_of(":/", pos));

		pos = ip.find_first_of(":/");
		if (pos != std::string::npos)
			ip.erase(pos, std::string::npos);

		return ip;
	}
	return std::string(soap->host);
}																		

char* ObjectTypeToName(userobject_type_t type)
{
	switch (type) {
	case USEROBJECT_TYPE_USER:
		return "user";
	case USEROBJECT_TYPE_GROUP:
		return "group";
	case USEROBJECT_TYPE_CONTACT:
		return "contact";
	case USEROBJECT_TYPE_COMPANY:
		return "company";
	case USEROBJECT_TYPE_NONACTIVE:
		return "non-active user";
	case USEROBJECT_TYPE_ADDRESSLIST:
		return "address list";
	};

	return "Unknown object";
}

char* RelationTypeToName(userobject_relation_t type)
{
	switch(type) {
	case OBJECTRELATION_GROUP_MEMBER:
		return "groupmember";
	case OBJECTRELATION_COMPANY_VIEW:
		return "company-view";
	case OBJECTRELATION_COMPANY_ADMIN:
		return "company-admin";
	case OBJECTRELATION_QUOTA_USERRECIPIENT:
		return "userquota-recipient";
	case OBJECTRELATION_QUOTA_COMPANYRECIPIENT:
		return "companyquota-recipient";
	case OBJECTRELATION_USER_SENDAS:
		return "sendas-user-allowed";
    case OBJECTRELATION_ADDRESSLIST_MEMBER:
        return "addressbook-member";
	};

	return "Unknown relation type";
}

ECUserManagement::ECUserManagement(BTSession *lpSession, ECPluginFactory *lpPluginFactory, ECConfig *lpConfig, ECLogger *lpLogger) {
    this->m_lpSession = lpSession;
    this->m_lpPluginFactory = lpPluginFactory;
    this->m_lpConfig = lpConfig;
    this->m_lpLogger = lpLogger;
}

ECUserManagement::~ECUserManagement() {


}

// Authenticate a user (NOTE: ECUserManagement will never authenticate SYSTEM unless it is actually
// authenticated by the remote server, which normally never happens)
ECRESULT ECUserManagement::AuthUserAndSync(char *szLoginname, char *szPassword, unsigned int *lpulUserId) {
    ECRESULT er = erSuccess;
    objectsignature_t external;
    UserPlugin *lpPlugin = NULL;
    string username;
    string companyname;
    string password;
	string error;
	char *szAuthMethod = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	/*
	 * GetUserAndCompanyFromLoginName() will return ZARAFA_W_PARTIAL_COMPLETION
	 * when only the username was resolved and not the companyname.
	 * When that happens continue like nothing happens because perhaps the
	 * LDAP plugin is used and the login attribute is the email address.
	 * We will resolve the company later when the session is created and the
	 * company is requested through getDetails().
	 */
    er = GetUserAndCompanyFromLoginName(string(szLoginname), &username, &companyname);
	if (er != erSuccess && er != ZARAFA_W_PARTIAL_COMPLETION)
		goto exit;

    if(SymmetricIsCrypted(szPassword)) {
        password = SymmetricDecrypt(szPassword);
    } else {
        password = szPassword;
    }

	szAuthMethod = m_lpConfig->GetSetting("auth_method");
	if (szAuthMethod && strcmp(szAuthMethod, "pam") == 0) {
		// authenticate through pam
		er = ECPAMAuthenticateUser(m_lpConfig->GetSetting("pam_service"), username, password, &error);
		if (er != erSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication through pam failed for user %s: %s", szLoginname, error.c_str());
			goto exit;
		}
		try {
			external = lpPlugin->resolveName(USEROBJECT_TYPE_USER, username, companyname);
		}
		catch (objectnotfound &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication through pam succeeded for %s, but not found by plugin: %s", szLoginname, e.what());
			er = ZARAFA_E_LOGON_FAILED;
			goto exit;
		}
		catch (std::exception &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication through pam succeeded for %s, but plugin returns error: %s", szLoginname, e.what());
			er = ZARAFA_E_LOGON_FAILED;
			goto exit;
		}
	} else if (szAuthMethod && strcmp(szAuthMethod, "kerberos") == 0) {
		// authenticate through kerberos
		er = ECKrb5AuthenticateUser(username, password, &error);
		if (er != erSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication through kerberos failed for user %s: %s", szLoginname, error.c_str());
			goto exit;
		}
		try {
			external = lpPlugin->resolveName(USEROBJECT_TYPE_USER, username, companyname);
		}
		catch (objectnotfound &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication through kerberos succeeded for %s, but not found by plugin: %s", szLoginname, e.what());
			er = ZARAFA_E_LOGON_FAILED;
			goto exit;
		}
		catch (std::exception &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication through kerberos succeeded for %s, but plugin returns error: %s", szLoginname, e.what());
			er = ZARAFA_E_LOGON_FAILED;
			goto exit;
		}
	} else
	{
		// default method is plugin
		try {
			external = lpPlugin->authenticateUser(username, companyname, password.c_str());
		} catch (login_error &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication by plugin failed for user %s: %s", szLoginname, e.what());
			er = ZARAFA_E_LOGON_FAILED;
			goto exit;
		} catch (objectnotfound &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication by plugin failed for user %s: %s", szLoginname, e.what());
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		} catch (std::exception &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Authentication error by plugin for user %s: %s", szLoginname, e.what());
			er = ZARAFA_E_PLUGIN_ERROR;
			goto exit;
		}
	}
		
    er = GetLocalObjectIdOrCreate(USEROBJECT_TYPE_USER, external, lpulUserId);
    if(er != erSuccess)
        goto exit;

exit:
    return er;
}

// Get data for an object, with on-the-fly delete of the specified object id
ECRESULT ECUserManagement::GetObjectDetailsAndSync(unsigned int ulObjectId, objectdetails_t *lpDetails) {
    ECRESULT er = erSuccess;

    if(IsLocalItem(ulObjectId)) {
        er = GetLocalObjectDetails(ulObjectId, lpDetails);
    } else {
		er = GetExternalObjectDetails(ulObjectId, lpDetails);
    }

    return er;
}

// Get a list of objects in a company space
ECRESULT ECUserManagement::GetCompanyObjectListAndSync(userobject_type_t type, unsigned int ulCompanyId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags)
{
    ECRESULT er = erSuccess;

	std::list<localobjectdetails_t> *lpObjects = new std::list<localobjectdetails_t>;
    std::list<unsigned int> *lpLocalObjects = NULL;
    std::list<unsigned int>::iterator iterLocalObjects;

    std::list<objectsignature_t> lstExternSignature;
    std::list<objectsignature_t>::iterator iterExternSignature;
    std::list<objectid_t> lstExternId;

    std::map<objectid_t, std::pair<unsigned int, std::string> > mapExternIdToLocal;
    std::map<objectid_t, std::pair<unsigned int, std::string> >::iterator iterExternIdToLocal;

    auto_ptr<map<objectid_t, objectdetails_t> > lpExternDetails;
    map<objectid_t, objectdetails_t>::iterator iterDetails;

    objectdetails_t details(type);
    objectid_t externid;
	std::string signature;
    objectid_t cidExtern;

    unsigned int ulObjectId = 0;

    UserPlugin *lpPlugin = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	if (m_lpSession->GetSessionManager()->IsHostedSupported()) {
		/* When hosted is enabled, the companyid _must_ have an external id,
		 * unless we are requesting the companylist in which case the companyid is 0 and doesn't
		 * need to be resolved at all.*/
		if (type != USEROBJECT_TYPE_COMPANY && !IsLocalItem(ulCompanyId) ){
			er = GetExternalId(ulCompanyId, NULL, &cidExtern);
			if (er != erSuccess)
				goto exit;
		}
	} else {
		if (type == USEROBJECT_TYPE_COMPANY) {
			er = ZARAFA_E_NO_SUPPORT;
			goto exit;
		}

		/* When hosted is disabled, use company id 0 */
		ulCompanyId = 0;
	}

    // Get all the items of the requested type
    er = GetLocalItemIdList(ulCompanyId, &lpLocalObjects, type);
    if(er != erSuccess)
        goto exit;

	for(iterLocalObjects = lpLocalObjects->begin(); iterLocalObjects != lpLocalObjects->end(); iterLocalObjects++) {
		if(IsLocalItem(*iterLocalObjects)) {
			// Local user, add it to the result array directly
 			details = objectdetails_t(type);

			if(! (ulFlags & USERMANAGEMENT_IDS_ONLY)) {
				er = GetLocalObjectDetails(*iterLocalObjects, &details);
				if(er != erSuccess)
					goto exit;
			}

			lpObjects->push_back(localobjectdetails_t(*iterLocalObjects,details));
		} else if(GetExternalId(*iterLocalObjects, NULL, &externid, &signature) == erSuccess) {
			mapExternIdToLocal.insert(
				std::map<objectid_t, std::pair<unsigned int, std::string> >::value_type(externid, std::pair<unsigned int, std::string>(*iterLocalObjects, signature)));
		} else {
			// There seems to be an item in the users table which has no extern ID and is not a local
 			// user. This is very strange, but we'll just ignore it.
		}
	}

    // We now have a map, mapping external id's to local user id's (and their signatures)
    try {
        // Get full user list
        lstExternSignature = lpPlugin->getAllObjects(type, cidExtern);

        // Get details for all users we just retrieved
        if(! (ulFlags & USERMANAGEMENT_IDS_ONLY)) {
			for(iterExternSignature = lstExternSignature.begin(); iterExternSignature != lstExternSignature.end(); iterExternSignature++) {
				lstExternId.push_back(iterExternSignature->id);
			}

            lpExternDetails = lpPlugin->getObjectDetails(type, lstExternId);
			/* We convert user and companyname to loginname later in this function */
		}
	} catch (notsupported &) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	} catch (objectnotfound &) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to retrieve %s list from external user source: %s", ObjectTypeToName(type), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

    // Loop through all the external signatures, adding them to the lpUsers list which we're going to be returning
    for(iterExternSignature = lstExternSignature.begin(); iterExternSignature != lstExternSignature.end(); iterExternSignature++) {
        iterExternIdToLocal = mapExternIdToLocal.find(iterExternSignature->id);

        if(iterExternIdToLocal == mapExternIdToLocal.end()) {
            // User is in external user database, but not in local, so add
            er = MoveOrCreateLocalObject(type, *iterExternSignature, &ulObjectId);

            // Create failed, so skip this entry
            if(er != erSuccess) {
                er = erSuccess;
                continue;
            }
        } else {
            // The object was found in both local and external user databases. Check the signature for changes
            std::string localsignature = iterExternIdToLocal->second.second;
            std::string remotesignature = iterExternSignature->signature;
            int localChange = 0, remoteChange = 0;
            char *end = NULL;

			// second == map value, first == id
			ulObjectId = iterExternIdToLocal->second.first; 

			remoteChange = strtoul(remotesignature.c_str(), &end, 10);
			if (end && *end == '\0') {
				// ADS uses uSNChanged incrementing attribute
				// whenChanged and modifyTimestamp attributes are not replicated, giving too many changes
				localChange = strtoul(localsignature.c_str(), &end, 10);
				if (!end || *end != '\0' || localChange < remoteChange)
					ProcessModification(ulObjectId, type, remotesignature);
			} else if (localsignature != remotesignature) {
				ProcessModification(ulObjectId, type, remotesignature);
			}
		}

        // ulObjectId is the local ID for the user, iterExternSignature->id is the external ID
		if(! (ulFlags & USERMANAGEMENT_IDS_ONLY)) {
		    iterDetails = lpExternDetails->find(iterExternSignature->id);

		    if(iterDetails == lpExternDetails->end()) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Details missing for %s", iterExternSignature->id.c_str());
				er = ZARAFA_E_PLUGIN_ERROR;
				goto exit;
			}

            details = iterDetails->second;

			er = UpdateUserDetails(&details);
			if (er != erSuccess)
				goto exit;
		}

        // Add the details to the final list
        lpObjects->push_back(localobjectdetails_t(ulObjectId,details));

        // Remove the external user from the list of internal ID's
        if(iterExternIdToLocal != mapExternIdToLocal.end()) {
            mapExternIdToLocal.erase(iterExternIdToLocal);
        }
    }

    // mapExternIdToLocal is now a map of objects that were NOT in the external user database

    for(iterExternIdToLocal = mapExternIdToLocal.begin(); iterExternIdToLocal != mapExternIdToLocal.end(); iterExternIdToLocal++) {
		MoveOrDeleteLocalObject(iterExternIdToLocal->second.first, type); // second == map value, first == id
    }

	lpObjects->sort();
    *lppObjects = lpObjects;
	lpObjects = NULL;

exit:
	if (lpLocalObjects)
		delete lpLocalObjects;

	if (lpObjects)
		delete lpObjects;

    return er;
}

ECRESULT ECUserManagement::GetObjectListAndSync(userobject_type_t type, unsigned int ulCompanyId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags)
{
	ECRESULT er = erSuccess;
	ECSession *lpecSession = NULL;

	std::list<localobjectdetails_t> *lpCompanies = NULL;
	std::list<localobjectdetails_t> *lpObjects = NULL;
	std::list<localobjectdetails_t> *lpObjectsTmp = NULL;
	std::list<localobjectdetails_t>::iterator iterCompanies;

	if (type == USEROBJECT_TYPE_COMPANY && !m_lpSession->GetSessionManager()->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	if (type == USEROBJECT_TYPE_COMPANY) {
		lpecSession = dynamic_cast<ECSession*>(m_lpSession);
		if (!lpecSession) {
			er = ZARAFA_E_INVALID_PARAMETER;
			goto exit;
		}

		er = lpecSession->GetSecurity()->GetViewableCompanyIds(&lpObjects, ulFlags);
		if (er != erSuccess)
			goto exit;
	} else {
		er = GetCompanyObjectListAndSync(type, ulCompanyId, &lpObjects, ulFlags);
		if (er != erSuccess)
			goto exit;
	}

	lpObjects->sort();
	*lppObjects = lpObjects;

exit:
	if (lpCompanies)
		delete lpCompanies;

	if (lpObjectsTmp)
		delete lpObjectsTmp;

	if (er != erSuccess && lpObjects)
		delete lpObjects;

	return er;
}

// Get merged object list for entire server, with on-the-fly delete/create of users and groups
ECRESULT ECUserManagement::GetMergedObjectListAndSync(unsigned int ulListTypes, unsigned int ulCompanyId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags)
{
    ECRESULT er = erSuccess;

	std::list<localobjectdetails_t> *lpObjects = NULL;
	std::list<localobjectdetails_t> *lpMergedObjects = new std::list<localobjectdetails_t>;
	localobjectdetails_t userobject;

	struct { unsigned int ulMask; userobject_type_t type; } sTypeMap[] = {
		{ ECUSEROBJECT_LIST_USERS_ACTIVE, USEROBJECT_TYPE_USER },
		{ ECUSEROBJECT_LIST_USERS_NONACTIVE, USEROBJECT_TYPE_NONACTIVE },
		{ ECUSEROBJECT_LIST_GROUP, USEROBJECT_TYPE_GROUP },
		{ ECUSEROBJECT_LIST_COMPANY, USEROBJECT_TYPE_COMPANY },
	};

	/* If hosted isn't enabled, we shouldn't check for company requests */
	if (!m_lpSession->GetSessionManager()->IsHostedSupported())
		ulListTypes &= ~ECUSEROBJECT_LIST_COMPANY;

    for (unsigned int i = 0; i < (sizeof(sTypeMap) / sizeof(sTypeMap[0])); i++) {
        // Merge userlist and grouplist
        if(ulListTypes & sTypeMap[i].ulMask) {
            er = GetObjectListAndSync(sTypeMap[i].type, ulCompanyId, &lpObjects, ulFlags);
            if(er != erSuccess)
                goto exit;

			lpMergedObjects->merge(*lpObjects);

            if(lpObjects) {
                delete lpObjects;
                lpObjects = NULL;
            }
        }
    }

	lpMergedObjects->sort();
	lpMergedObjects->unique();

    *lppObjects = lpMergedObjects;

exit:
    if(lpObjects)
        delete lpObjects;

    return er;
}

// Get 'live' members in a group, with on-the-fly delete of the specified group id
ECRESULT ECUserManagement::GetSubObjectsOfObjectAndSync(userobject_relation_t relation, userobject_type_t childtype, unsigned int ulParentId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags)
{
    ECRESULT er = erSuccess;
    objectid_t objectid;
	std::list<objectsignature_t> lstObjects;
	std::list<objectsignature_t>::iterator iterObjects;
    std::list<objectid_t> lstObjectIDs;
    objectdetails_t details;
    unsigned int ulObjectId = 0;
	userobject_type_t ulParentType;
    std::list<localobjectdetails_t> *lpObjects = NULL;
    std::list<localobjectdetails_t> *lpObjectsTmp = NULL;
    std::list<localobjectdetails_t> *lpCompanies = NULL;
	std::list<localobjectdetails_t>::iterator iterCompany;
    auto_ptr<map<objectid_t, objectdetails_t> > lpExternDetails;
    map<objectid_t, objectdetails_t>::iterator iterDetails;

	UserPlugin *lpPlugin = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	if (!m_lpSession->GetSessionManager()->IsHostedSupported() &&
		(relation == OBJECTRELATION_COMPANY_VIEW ||
		 relation == OBJECTRELATION_COMPANY_ADMIN)) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	// The 'everyone' group contains all visible users for the currently logged in user.
	if (relation == OBJECTRELATION_GROUP_MEMBER && ulParentId == ZARAFA_UID_EVERYONE) {
		ECSession *lpecSession = NULL;

		lpecSession = dynamic_cast<ECSession*>(m_lpSession);
		if (!lpecSession) {
			er = ZARAFA_E_INVALID_PARAMETER;
			goto exit;
		}

		er = lpecSession->GetSecurity()->GetViewableCompanyIds(&lpCompanies, ulFlags);
		if (er != erSuccess)
			goto exit;

		/* Fallback in case hosted is not supported */
		if (lpCompanies->empty())
			lpCompanies->push_back(localobjectdetails_t(0, USEROBJECT_TYPE_COMPANY));

		lpObjects = new std::list<localobjectdetails_t>();

		for (iterCompany = lpCompanies->begin(); iterCompany != lpCompanies->end(); iterCompany++) {
			er = GetObjectListAndSync(childtype, iterCompany->ulId, &lpObjectsTmp, ulFlags);
			if (er != erSuccess)
				goto exit;

			lpObjects->merge(*lpObjectsTmp);

			delete lpObjectsTmp;
			lpObjectsTmp = NULL;
		}
    } else {
		lpObjects = new std::list<localobjectdetails_t>();

		er = GetItemInfo(ulParentId, &ulParentType, &objectid);
        if(er != erSuccess)
            goto exit;

        try {
            lstObjects = lpPlugin->getSubObjectsForObject(relation, ulParentType, objectid, childtype);
            if(! (ulFlags & USERMANAGEMENT_IDS_ONLY)) {
                for(iterObjects = lstObjects.begin(); iterObjects != lstObjects.end(); iterObjects++) {
                    lstObjectIDs.push_back(iterObjects->id);
                }
                lpExternDetails = lpPlugin->getObjectDetails(childtype, lstObjectIDs);
				/* We convert user and companyname to loginname later in this function */
            }
        } catch(objectnotfound &) {
            MoveOrDeleteLocalObject(ulParentId, ulParentType);
            er = ZARAFA_E_NOT_FOUND;
            goto exit;
		} catch (notsupported &) {
			er = ZARAFA_E_NO_SUPPORT;
			goto exit;
        } catch(std::exception &e) {
            m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to retrieve members for relation %s: %s.", RelationTypeToName(relation), e.what());
            er = ZARAFA_E_PLUGIN_ERROR;
            goto exit;
        }

		for(iterObjects = lstObjects.begin(); iterObjects != lstObjects.end(); iterObjects++) {
			if(GetLocalObjectIdOrCreate(childtype, *iterObjects, &ulObjectId) != erSuccess)
				continue;

			if(! (ulFlags & USERMANAGEMENT_IDS_ONLY)) {
				iterDetails = lpExternDetails->find(iterObjects->id);

				if(iterDetails == lpExternDetails->end())
					continue;

				details = iterDetails->second;

				er = UpdateUserDetails(&details);
				if (er != erSuccess)
					goto exit;
			}

			lpObjects->push_back(localobjectdetails_t(ulObjectId,details));
		}
    }

	/*
	 * We now have the full list of objects coming from the plugin.
	 * NOTE: This list is not fool proof, the caller should check
	 * which entries are visible to the user and which ones are not.
	 * We cannot do that here since ECSecurity calls this function
	 * when determining if an object is visible. And endless recursive
	 * loops are not really a pretty sight.
	 */
	lpObjects->sort();
	lpObjects->unique();
    *lppObjects = lpObjects;

exit:
	if (lpCompanies)
		delete lpCompanies;

	if (lpObjectsTmp)
		delete lpObjectsTmp;

	if (er != erSuccess && lpObjects)
		delete lpObjects;

    return er;
}

// Get parent objects which an object belongs, with on-the-fly delete of the specified parent object
ECRESULT ECUserManagement::GetParentObjectsOfObjectAndSync(userobject_relation_t relation, userobject_type_t parenttype, unsigned int ulChildId, std::list<localobjectdetails_t> **lppObjects, unsigned int ulFlags) {
    ECRESULT er = erSuccess;
    objectid_t objectid;
    std::list<objectsignature_t> lstObjects;
    std::list<objectsignature_t>::iterator iterObjects;
    std::list<objectid_t> lstObjectIDs;
    objectdetails_t details;
    unsigned int ulObjectId = 0;
    std::list<localobjectdetails_t> *lpObjects = new std::list<localobjectdetails_t>;
    auto_ptr<map<objectid_t, objectdetails_t> > lpExternDetails;
    map<objectid_t, objectdetails_t>::iterator iterDetails;
    userobject_type_t ulType;

    UserPlugin *lpPlugin = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	if (!m_lpSession->GetSessionManager()->IsHostedSupported() &&
		(relation == OBJECTRELATION_COMPANY_VIEW ||
		 relation == OBJECTRELATION_COMPANY_ADMIN)) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	if (relation == OBJECTRELATION_GROUP_MEMBER && ulChildId == ZARAFA_UID_SYSTEM) {
        // System has no objects
	} else {
		er = GetItemInfo(ulChildId, &ulType, &objectid);
		if (er != erSuccess)
			goto exit;

		try {
			lstObjects = lpPlugin->getParentObjectsForObject(relation, ulType, objectid, parenttype);
			if(! (ulFlags & USERMANAGEMENT_IDS_ONLY)) {
				for(iterObjects = lstObjects.begin(); iterObjects != lstObjects.end(); iterObjects++) {
					lstObjectIDs.push_back(iterObjects->id);
				}
				lpExternDetails = lpPlugin->getObjectDetails(parenttype, lstObjectIDs);
				/* We convert user and companyname to loginname later in this function */
			}
		} catch(objectnotfound &) {
			MoveOrDeleteLocalObject(ulChildId, ulType);
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		} catch (notsupported &) {
			er = ZARAFA_E_NO_SUPPORT;
			goto exit;
		} catch(std::exception &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to retrieve parents for relation %s: %s.", RelationTypeToName(relation), e.what());
			er = ZARAFA_E_PLUGIN_ERROR;
			goto exit;
		}

		for(iterObjects = lstObjects.begin(); iterObjects != lstObjects.end(); iterObjects++) {
			if(GetLocalObjectIdOrCreate(parenttype, *iterObjects, &ulObjectId) != erSuccess)
				continue;

			if(! (ulFlags & USERMANAGEMENT_IDS_ONLY)) {
				iterDetails = lpExternDetails->find(iterObjects->id);

				if(iterDetails == lpExternDetails->end())
					continue;

				details = iterDetails->second;

				er = UpdateUserDetails(&details);
				if (er != erSuccess)
					goto exit;
			}

			lpObjects->push_back(localobjectdetails_t(ulObjectId,details));
		}
	}

	/* If we are requesting group membership we should always insert the everyone, since everyone is member of EVERYONE except for SYSTEM */
	if (relation == OBJECTRELATION_GROUP_MEMBER && ulObjectId != ZARAFA_UID_SYSTEM) {
		if (!(ulFlags & USERMANAGEMENT_IDS_ONLY)) {
			er = GetObjectDetailsAndSync(ZARAFA_UID_EVERYONE, &details);
			if(er != erSuccess)
				goto exit;
		} else
			details = objectdetails_t(USEROBJECT_TYPE_GROUP);

		lpObjects->push_back(localobjectdetails_t(ZARAFA_UID_EVERYONE, details));
	}

	/*
	 * We now have the full list of objects coming from the plugin.
	 * NOTE: This list is not fool proof, the caller should check
	 * which entries are visible to the user and which ones are not.
	 * We cannot do that here since ECSecurity calls this function
	 * when determining if an object is visible. And endless recursive
	 * loops are not really a pretty sight.
	 */
	*lppObjects = lpObjects;

exit:
	if (er != erSuccess && lpObjects)
		delete lpObjects;

	return er;
}

// Set data for a single object, with on-the-fly delete of the specified object id
ECRESULT ECUserManagement::SetObjectDetailsAndSync(unsigned int ulObjectId, objectdetails_t sDetails, std::list<std::string> *lpRemoveProps) {
    ECRESULT er = erSuccess;
    objectid_t objectid;
	userobject_type_t type;

    UserPlugin *lpPlugin = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    // Resolve the objectname and sync the object into local database
	er = GetItemInfo(ulObjectId, &type, &objectid);
	if (er != erSuccess)
		goto exit;

    try {
        lpPlugin->changeObject(type, objectid, sDetails, lpRemoveProps);
    } catch (objectnotfound &) {
        MoveOrDeleteLocalObject(ulObjectId, type);
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
	} catch (notimplemented &e) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to set details for external %s: %s", ObjectTypeToName(type), e.what());
		er = ZARAFA_E_NOT_IMPLEMENTED;
		goto exit;
    } catch (std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to set details for external %s: %s", ObjectTypeToName(type), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

    // Purge cache
    m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulObjectId);


exit:
    return er;
}

// Create an object with specific ID or modify an existing object. Used for sync purposes
ECRESULT ECUserManagement::CreateOrModifyObject(const string &strExternId, userobject_type_t type, objectdetails_t sDetails, unsigned int ulPreferredId, std::list<std::string> *lpRemoveProps) {
    ECRESULT er = erSuccess;
    objectsignature_t signature;
    UserPlugin *lpPlugin = NULL;
	unsigned int ulObjectId = 0;

	// Local items have no external id
	if (strExternId.empty())
		goto exit;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    // Resolve the objectname and sync the object into local database
    if(GetLocalId(strExternId, type, &ulObjectId, &signature.signature) == erSuccess) {
        try {
			signature.id = strExternId;
            lpPlugin->changeObject(type, signature.id, sDetails, lpRemoveProps);
        } catch (objectnotfound &) {
            MoveOrDeleteLocalObject(ulObjectId, type);
            er = ZARAFA_E_NOT_FOUND;
            goto exit;
		} catch (notimplemented &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to set details for external %s: %s", ObjectTypeToName(type), e.what());
			er = ZARAFA_E_NOT_IMPLEMENTED;
			goto exit;
        } catch (std::exception &e) {
            m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to set details for external %s: %s", ObjectTypeToName(type), e.what());
            er = ZARAFA_E_PLUGIN_ERROR;
            goto exit;
        }
    } else {
        // Object does not exist yet, create in external database

        try {
            signature = lpPlugin->createObject(type, sDetails);
		} catch (notimplemented &e) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create %s in external database: %s", ObjectTypeToName(type), e.what());
			er = ZARAFA_E_NOT_IMPLEMENTED;
			goto exit;
        } catch (std::exception &e) {
            m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create %s in external database: %s", ObjectTypeToName(type), e.what());
            er = ZARAFA_E_PLUGIN_ERROR;
            goto exit;
        }

        er = CreateLocalObjectSimple(signature, type, ulPreferredId);
        if(er != erSuccess) {
			/*
			 * Failed to create object locally, it should be removed from external DB.
			 * Otherwise it will remain present in the external DB and synced later while
			 * ECSync will also perform a retry to create the object. Obviously this
			 * can lead to unexpected behavior because the Offline server should never(!)
			 * sync between plugin and ECUsermanagement somewhere else other than this
			 * function
			 */
			try {
				lpPlugin->deleteObject(type, signature.id);
			} catch (std::exception &e) {
				m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to delete %s from external database after failed (partial) creation: %s",
								ObjectTypeToName(type), e.what());
			}

            goto exit;
		}
    }

exit:
    return er;
}

ECRESULT ECUserManagement::ModifyExternId(unsigned int ulObjectId, const string &strExternId)
{
	ECRESULT er = erSuccess;
	userobject_type_t objecttype;
	std::string strOldExternId;
	std::auto_ptr<objectdetails_t> ptrOldDetails;
	objectdetails_t sNewDetails;
	std::string strQuery;
	unsigned int ulRows = 0;
	bool bRevert = false;

    ECDatabase *lpDatabase = m_lpSession->GetDatabase();
	UserPlugin *lpPlugin = NULL;

	if (IsLocalItem(ulObjectId) || strExternId.empty()) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    er = GetItemInfo(ulObjectId, &objecttype, &strOldExternId);
    if(er != erSuccess)
        goto exit;

	try {
		ptrOldDetails = lpPlugin->getObjectDetails(objecttype, strOldExternId);
		sNewDetails = *ptrOldDetails;
	} catch(objectnotfound &) {
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
	}


	// Delete the current user as we can't add the new one otherwise.
	try {
		lpPlugin->deleteObject(objecttype, strOldExternId);
		bRevert = true;
	}catch (std::runtime_error &) {
		er = ZARAFA_E_CALL_FAILED;
		goto exit;
	}

	// Create a new user in the plugin, delete the old one if everything else succeeds.
	try {
		sNewDetails.SetPropString(OB_PROP_S_EXTERNID, strExternId);
		lpPlugin->createObject(objecttype, sNewDetails);
	} catch (std::runtime_error &) {
		er = ZARAFA_E_CALL_FAILED;
		goto exit;
	}

	// Update the users table.
	er = lpDatabase->Begin();
	if (er != erSuccess)
		goto exit;

	strQuery = "UPDATE users SET externid='" + lpDatabase->Escape(strExternId) + "' WHERE id=" + stringify(ulObjectId) + " AND externid='" + strOldExternId + "'";
	er = lpDatabase->DoUpdate(strQuery, &ulRows);
	if (er != erSuccess)
		goto exit;

	if (ulRows != 1) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	er = m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulObjectId);
	if (er != erSuccess)
		goto exit;

	er = lpDatabase->Commit();
	if (er != erSuccess)
		goto exit;

	bRevert = false;

exit:
	if (er != erSuccess && bRevert) {
		try {
			lpPlugin->deleteObject(objecttype, strExternId);
		} catch (std::exception &) {
			// Ignore error.
		}

		try {
			// Restore the original object.
			lpPlugin->createObject(objecttype, *ptrOldDetails);
		} catch (std::exception &) {
			// This would be problematic...
		}
	}

	return er;
}

// Add a member to a group, with on-the-fly delete of the specified group id
ECRESULT ECUserManagement::AddSubObjectToObjectAndSync(userobject_relation_t relation, unsigned int ulParentId, unsigned int ulChildId) {
    ECRESULT er = erSuccess;
	ABEID eid(MAPI_ABCONT, *(GUID*)&MUIDECSAB_SERVER, 1);
    objectid_t parentid;
	userobject_type_t parenttype;
    objectid_t childid;
	userobject_type_t childtype;
	SOURCEKEY sSourceKey;

    UserPlugin *lpPlugin = NULL;

	/* We don't support operations on local items */
	if (IsLocalItem(ulParentId) || IsLocalItem(ulChildId)) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    er = GetItemInfo(ulParentId, &parenttype, &parentid);
    if(er != erSuccess)
        goto exit;

    er = GetItemInfo(ulChildId, &childtype, &childid);
    if(er != erSuccess)
	    goto exit;

    try {
        lpPlugin->addSubObjectRelation(relation, parenttype, parentid, childtype, childid);
    } catch(objectnotfound &) {
        MoveOrDeleteLocalObject(ulParentId, parenttype);
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
	} catch (notimplemented &) {
		er = ZARAFA_E_NOT_IMPLEMENTED;
		goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to add relation %s to external user database: %s", RelationTypeToName(relation), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

	/*
	 * Add addressbook change, note that setting the objectrelation won't update the signature
	 * which means we have to run this update manually. Also note that when adding a child to a
	 * group the _group_ has changed, but when adding a child to a company the _child_ has changed.
	 * This will cause the sync to recreate the correct object with the updated rights.
	 */
	if (relation == OBJECTRELATION_GROUP_MEMBER)
		er = GetABSourceKeyV1(ulParentId, &sSourceKey);
	else if (relation == OBJECTRELATION_COMPANY_ADMIN || relation == OBJECTRELATION_COMPANY_VIEW)
		er = GetABSourceKeyV1(ulChildId, &sSourceKey);

	if (relation == OBJECTRELATION_GROUP_MEMBER || relation == OBJECTRELATION_COMPANY_ADMIN || relation == OBJECTRELATION_COMPANY_VIEW)
	{
		if (er != erSuccess)
			goto exit;

		AddABChange(m_lpSession, ICS_AB_CHANGE, sSourceKey, SOURCEKEY(CbABEID(&eid), (char *)&eid));
	}

    m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulParentId);

exit:
    return er;
}

ECRESULT ECUserManagement::DeleteSubObjectFromObjectAndSync(userobject_relation_t relation, unsigned int ulParentId, unsigned int ulChildId) {
    ECRESULT er = erSuccess;
	ABEID eid(MAPI_ABCONT, *(GUID*)&MUIDECSAB_SERVER, 1);
    objectid_t parentid;
	userobject_type_t parenttype;
    objectid_t childid;
	userobject_type_t childtype;
	SOURCEKEY sSourceKey;

    UserPlugin *lpPlugin = NULL;

	/* We don't support operations on local items */
	if (IsLocalItem(ulParentId) || IsLocalItem(ulChildId)) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    er = GetItemInfo(ulParentId, &parenttype, &parentid);
    if(er != erSuccess)
        goto exit;

    er = GetItemInfo(ulChildId, &childtype, &childid);
    if(er != erSuccess)
	    goto exit;

    try {
        lpPlugin->deleteSubObjectRelation(relation, parenttype, parentid, childtype, childid);
    } catch(objectnotfound &) {
        // We can't delete the parent because we don't know whether the child or the entire parent was not found
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    } catch (notimplemented &) {
		er = ZARAFA_E_NOT_IMPLEMENTED;
		goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to remove relation %s from external user database: %s.", RelationTypeToName(relation), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

	/*
	 * Add addressbook change, note that setting the objectrelation won't update the signature
	 * which means we have to run this update manually. Also note that when deleting a child from a
	 * group the _group_ has changed, but when deleting a child from a company the _child_ has changed.
	 * This will cause the sync to recreate the correct object with the updated rights.
	 */
	if (relation == OBJECTRELATION_GROUP_MEMBER)
		er = GetABSourceKeyV1(ulParentId, &sSourceKey);
	else if (relation == OBJECTRELATION_COMPANY_ADMIN || relation == OBJECTRELATION_COMPANY_VIEW)
		er = GetABSourceKeyV1(ulChildId, &sSourceKey);

	if (relation == OBJECTRELATION_GROUP_MEMBER || relation == OBJECTRELATION_COMPANY_ADMIN || relation == OBJECTRELATION_COMPANY_VIEW)
	{
		if (er != erSuccess)
			goto exit;

		AddABChange(m_lpSession, ICS_AB_CHANGE, sSourceKey, SOURCEKEY(CbABEID(&eid), (char *)&eid));
	}

    m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulParentId);

exit:
    return er;
}

// Resolve an object name to an object id, with on-the-fly create of the specified object type
ECRESULT ECUserManagement::ResolveObjectAndSync(userobject_type_t type, char *szName, unsigned int *lpulObjectId) {
    ECRESULT er = erSuccess;
    objectsignature_t objectsignature;
	string username;
	string companyid;
    UserPlugin *lpPlugin = NULL;
	ECSession *lpecSession = dynamic_cast<ECSession*>(m_lpSession);
	char *szHideEveryone = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	if (type == USEROBJECT_TYPE_USER && stricmp(szName, ZARAFA_ACCOUNT_SYSTEM) == 0) {
		*lpulObjectId = ZARAFA_UID_SYSTEM;
		goto exit;
	} else if (type == USEROBJECT_TYPE_GROUP && stricmp(szName, ZARAFA_FULLNAME_EVERYONE) == 0) {
		// Hide group everyone when requested
		if (lpecSession && lpecSession->GetSecurity()->GetAdminLevel() == 0) {
			szHideEveryone = lpecSession->GetSessionManager()->GetConfig()->GetSetting("hide_everyone");
			if (szHideEveryone && parseBool(szHideEveryone)) {
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}
		}
		
		*lpulObjectId = ZARAFA_UID_EVERYONE;
		goto exit;
	} else if (type == USEROBJECT_TYPE_COMPANY && stricmp(szName, "") == 0) {
		*lpulObjectId = 0;
		goto exit;
	}

	if ((type == USEROBJECT_TYPE_USER || type == USEROBJECT_TYPE_NONACTIVE || type == USEROBJECT_TYPE_GROUP) &&
		m_lpSession->GetSessionManager()->IsHostedSupported()) {
			/* GetUserAndCompanyFromLoginName() will return ZARAFA_W_PARTIAL_COMPLETION
			 * when the companyname could not be determined from the username. This isn't
			 * something bad since perhaps the username is unique between all companies.
			 */
			er = GetUserAndCompanyFromLoginName(szName, &username, &companyid);
			if (er == ZARAFA_W_PARTIAL_COMPLETION)
				er = erSuccess;
			else if (er != erSuccess)
				goto exit;
	} else {
		username = szName;
		companyid = "";
	}

    try {
        objectsignature = lpPlugin->resolveName(type, username, companyid);
	} catch (notsupported &) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	} catch (objectnotfound &e) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to resolve %s '%s': %s", ObjectTypeToName(type), szName, e.what());
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to resolve %s '%s': %s", ObjectTypeToName(type), szName, e.what());
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    }

    er = GetLocalObjectIdOrCreate(type, objectsignature, lpulObjectId);

exit:
    return er;
}

// Get MAPI property data for a group or user/group id, with on-the-fly delete of the specified user/group
ECRESULT ECUserManagement::GetProps(struct soap *soap, unsigned int ulId, struct propTagArray *lpPropTagArray, struct propValArray *lpPropValArray) {
    ECRESULT er = erSuccess;
    objectdetails_t objectdetails;

    er = GetObjectDetailsAndSync(ulId, &objectdetails);
	if(er != erSuccess) {
		er = ZARAFA_E_NOT_FOUND;
        goto exit;
	}

	er = ConvertObjectDetailsToProps(soap, ulId, objectdetails.GetType(), &objectdetails, lpPropTagArray, lpPropValArray);

exit:
	return er;
}

ECRESULT ECUserManagement::GetContainerProps(struct soap *soap, unsigned int ulObjectId, struct propTagArray *lpPropTagArray, struct propValArray *lpPropValArray)
{
	ECRESULT er = erSuccess;
	objectdetails_t objectdetails;

	if (ulObjectId == 0 || ulObjectId == 1) /* Zarafa Address Book || Global Address Book */
		er = ConvertABContainerToProps(soap, ulObjectId, lpPropTagArray, lpPropValArray);
	else {
        er = GetObjectDetailsAndSync(ulObjectId, &objectdetails);
        if(er != erSuccess) {
            er = ZARAFA_E_NOT_FOUND;
            goto exit;
        }
    	er = ConvertContainerObjectDetailsToProps(soap, ulObjectId, objectdetails.GetType(), &objectdetails, lpPropTagArray, lpPropValArray);
    }
    
exit:
	return er;
}

// Get local details
ECRESULT ECUserManagement::GetLocalObjectDetails(unsigned int ulId, objectdetails_t *lpDetails) {
    ECRESULT er = erSuccess;
    objectdetails_t sDetails;
	objectdetails_t	sPublicStoreDetails;

    if(ulId == ZARAFA_UID_SYSTEM) {
        sDetails = objectdetails_t(USEROBJECT_TYPE_USER);
        sDetails.SetPropString(OB_PROP_S_LOGIN, ZARAFA_ACCOUNT_SYSTEM);
        sDetails.SetPropString(OB_PROP_S_PASSWORD, "");
        sDetails.SetPropString(OB_PROP_S_FULLNAME, ZARAFA_FULLNAME_SYSTEM);
        sDetails.SetPropString(OB_PROP_S_EMAIL, m_lpConfig->GetSetting("system_email_address"));
		sDetails.SetPropString(OB_PROP_S_COMPANYNAME, "");
        sDetails.SetPropInt(OB_PROP_I_ADMINLEVEL, ADMIN_LEVEL_SYSADMIN);
    } else if (ulId == ZARAFA_UID_EVERYONE) {
        sDetails = objectdetails_t(USEROBJECT_TYPE_GROUP);
		sDetails.SetPropString(OB_PROP_S_LOGIN, ZARAFA_ACCOUNT_EVERYONE);
        sDetails.SetPropString(OB_PROP_S_FULLNAME, ZARAFA_FULLNAME_EVERYONE);
		sDetails.SetPropString(OB_PROP_S_COMPANYNAME, "");

		if (m_lpSession->GetSessionManager()->IsDistributedSupported()) {
			if (GetPublicStoreDetails(&sPublicStoreDetails) == erSuccess)
				sDetails.SetPropString(OB_PROP_S_SERVERNAME, sPublicStoreDetails.GetPropString(OB_PROP_S_SERVERNAME));
		}
	}

    *lpDetails = sDetails;

    return er;
}

// Get remote details
ECRESULT ECUserManagement::GetExternalObjectDetails(unsigned int ulId, objectdetails_t *lpDetails)
{
	ECRESULT er = erSuccess;
	auto_ptr<objectdetails_t> details;
	objectdetails_t detailscached;
	objectid_t externid;
	userobject_type_t ulType;
    UserPlugin *lpPlugin = NULL;

	er = GetItemInfo(ulId, &ulType, &externid);
	if (er != erSuccess)
		goto exit;

	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserDetails(ulId, ulType, &detailscached);
	if (er == erSuccess) {
        er = UpdateUserDetails(&detailscached);
        if (er != erSuccess)
            goto exit;

        *lpDetails = detailscached;
        
        goto exit;
    }
    
    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    try {
        details = lpPlugin->getObjectDetails(ulType, externid);
    } catch (objectnotfound &) {
		/*
		 * MoveOrDeleteLocalObject() will call this function to determine if the
		 * object has been moved. So do _not_ call MoveOrDeleteLocalObject() from
		 * this location!.
		 */
        DeleteLocalObject(ulId, ulType);
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
	} catch (notsupported &) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
    } catch (std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to get %s details for object id %d: %s", ObjectTypeToName(ulType), ulId, e.what());
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    }

	/* Update cache so we don't have to bug the plugin until the data has changed.
	 * Note that we don't care if the update succeeded, if it fails we will retry
	 * when the user details are requested for a second time. */
	m_lpSession->GetSessionManager()->GetCacheManager()->SetUserDetails(ulId, ulType, details.get());

    er = UpdateUserDetails(details.get());
    if (er != erSuccess)
        goto exit;

    *lpDetails = *details;

exit:
    return er;
}

// **************************************************************************************************************************************
//
// Functions that actually do the SQL
//
// **************************************************************************************************************************************

ECRESULT ECUserManagement::GetItemInfo(unsigned int ulId, userobject_type_t *lpulType, std::string *lpStr, std::string *lpSignature)
{
	ECRESULT er = erSuccess;

	if (IsLocalItem(ulId))
		er = ZARAFA_E_INVALID_PARAMETER;
	else
		er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserObject(ulId, lpulType, NULL, lpStr, lpSignature);

	return er;
}

ECRESULT ECUserManagement::GetExternalId(unsigned int ulId, userobject_type_t *lpulType, std::string *lpExternId, std::string *lpSignature)
{
	ECRESULT er = erSuccess;

	if (IsLocalItem(ulId))
		er = ZARAFA_E_INVALID_PARAMETER;
	else
		er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserExternId(ulId, lpulType, lpExternId, lpSignature);

	return er;
}

ECRESULT ECUserManagement::GetLocalId(const string &strExternId, userobject_type_t ulType, unsigned int *lpulId, std::string *lpSignature)
{
	return m_lpSession->GetSessionManager()->GetCacheManager()->GetUserObject(strExternId, ulType, NULL, lpulId, lpSignature);	
}

ECRESULT ECUserManagement::GetItemObjectType(unsigned int ulId, userobject_type_t *lpulType)
{
	ECRESULT er = erSuccess;

	if (ulId == ZARAFA_UID_EVERYONE)
		*lpulType = USEROBJECT_TYPE_GROUP;
	else if (ulId == ZARAFA_UID_SYSTEM)
		*lpulType = USEROBJECT_TYPE_USER;
	else
		er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserObject(ulId, lpulType, NULL, NULL, NULL);

	return er;
}

ECRESULT ECUserManagement::GetItemCompany(unsigned int ulObjectId, userobject_type_t ulType, unsigned int *lpulCompanyId)
{
	ECRESULT er = erSuccess;

	if (ulType == USEROBJECT_TYPE_COMPANY)
		*lpulCompanyId = ulObjectId;
	else if (IsLocalItem(ulObjectId) || !m_lpSession->GetSessionManager()->IsHostedSupported())
		*lpulCompanyId = 0;
	else
		er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserCompany(ulObjectId, ulType, lpulCompanyId);

	return er;
}

ECRESULT ECUserManagement::GetLocalObjectIdOrCreate(userobject_type_t type, objectsignature_t signature, unsigned int *lpulObjectId)
{
    ECRESULT er = erSuccess;
    ECDatabase *lpDatabase = m_lpSession->GetDatabase();
    DB_RESULT lpResult = NULL;
    DB_ROW lpRow = NULL;

    if(lpDatabase == NULL) {
        er = ZARAFA_E_DATABASE_ERROR;
        goto exit;
    }

    er = lpDatabase->DoSelect(
	"SELECT id, object_type FROM users "
	"WHERE externid=" + lpDatabase->EscapeBinary((unsigned char *)signature.id.c_str(), signature.id.size()) + " "
		"AND object_type = " + stringify(type), &lpResult);

    if(er != erSuccess) {
        er = ZARAFA_E_DATABASE_ERROR;
        goto exit;
    }

    lpRow = lpDatabase->FetchRow(lpResult);

    if(lpRow == NULL || lpRow[0] == NULL) {
        // Object is not in local database
        er = MoveOrCreateLocalObject(type, signature, lpulObjectId);
        goto exit;
    }

    if(lpulObjectId)
        *lpulObjectId = atoi(lpRow[0]);

exit:
    if(lpResult)
        lpDatabase->FreeResult(lpResult);

    return er;
}

ECRESULT ECUserManagement::GetLocalItemIdList(unsigned int ulCompanyId, std::list<unsigned int> **lppObjects, userobject_type_t ulType)
{
    ECRESULT er = erSuccess;
    ECDatabase *lpDatabase = NULL;
    DB_RESULT lpResult = NULL;
    DB_ROW lpRow = NULL;
	std::list<unsigned int> *lpObjects = new std::list<unsigned int>();
	string strQuery;

	lpDatabase = m_lpSession->GetDatabase();
	if (lpDatabase == NULL) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	strQuery =
		"SELECT id "
		"FROM users "
		"WHERE object_type = " + stringify(ulType);
	/* As long as the Offline server has partial hosted support,
	 * we must comment out this additional where statement... */
	if (m_lpSession->GetSessionManager()->IsHostedSupported())
		strQuery += " AND company = " + stringify(ulCompanyId);

	er = lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	while(1) {
		lpRow = lpDatabase->FetchRow(lpResult);

		if(lpRow == NULL)
			break;

		if(lpRow[0] == NULL)
			continue;

		lpObjects->push_back(atoi(lpRow[0]));
	}

	if(ulType == USEROBJECT_TYPE_GROUP)
		lpObjects->push_back(ZARAFA_UID_EVERYONE);
	else if (ulType == USEROBJECT_TYPE_USER)
		lpObjects->push_back(ZARAFA_UID_SYSTEM);

	lpObjects->sort();
	lpObjects->unique();

	*lppObjects = lpObjects;

exit:
	if(lpResult)
		lpDatabase->FreeResult(lpResult);

	if (er != erSuccess && lpObjects)
		delete lpObjects;

	return er;
}

ECRESULT ECUserManagement::CreateObjectAndSync(userobject_type_t type, objectdetails_t details, unsigned int *lpulId)
{
    ECRESULT er = erSuccess;
    objectsignature_t objectsignature;

    UserPlugin *lpPlugin = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    try {
        objectsignature = lpPlugin->createObject(type, details);
	} catch (notimplemented &e) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create %s in user database: %s", ObjectTypeToName(type), e.what());
		er = ZARAFA_E_NOT_IMPLEMENTED;
		goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create %s in user database: %s", ObjectTypeToName(type), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

    er = MoveOrCreateLocalObject(type, objectsignature, lpulId);

    if(er != erSuccess) {
        // We could not create the object in the local database. This means that we have to rollback
        // our object creation in the plugin, because otherwise the database and the plugin would stay
        // out-of-sync until you add new licenses. Also, it would be impossible for the user to manually
        // rollback the object creation, as you need a Zarafa object id to delete an object, and we don't have
        // one of those, because CreateLocalObject() failed.

        try {
            lpPlugin->deleteObject(type, objectsignature.id);
        } catch(std::exception &e) {
            m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to delete %s from plugin database after failed creation: %s", ObjectTypeToName(type), e.what());
            er = ZARAFA_E_PLUGIN_ERROR;
            goto exit;
        }
    }

exit:
    return er;
}

ECRESULT ECUserManagement::DeleteObjectAndSync(unsigned int ulId)
{
	ECRESULT er = erSuccess;
	objectid_t objectid;
	userobject_type_t type;

    UserPlugin *lpPlugin = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	er = GetItemInfo(ulId, &type, &objectid);
	if (er != erSuccess)
		goto exit;

    try {
        lpPlugin->deleteObject(type, objectid);
	} catch (objectnotfound &) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	} catch (notimplemented &e) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to delete %s in user database: %s", ObjectTypeToName(type), e.what());
		er = ZARAFA_E_NOT_IMPLEMENTED;
		goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to delete %s in user database: %s", ObjectTypeToName(type), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

	er = DeleteLocalObject(ulId, type);

exit:
	return er;
}

ECRESULT ECUserManagement::SetQuotaDetailsAndSync(unsigned int ulId, quotadetails_t details)
{
    ECRESULT er = erSuccess;
    userobject_type_t ulType;
	string externid;

    UserPlugin *lpPlugin = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	er = GetItemInfo(ulId, &ulType, &externid);
	if (er != erSuccess)
		goto exit;

	try {
        lpPlugin->setQuota(ulType, externid, details);
    } catch(objectnotfound &) {
		MoveOrDeleteLocalObject(ulId, ulType);
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
	} catch (notimplemented &e) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to set quota for %s: %s", ObjectTypeToName(ulType), e.what());
		er = ZARAFA_E_NOT_IMPLEMENTED;
		goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to set quota for %s: %s", ObjectTypeToName(ulType), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

    m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulId);

exit:
    return er;
}

ECRESULT ECUserManagement::GetQuotaDetailsAndSync(unsigned int ulId, quotadetails_t *lpDetails, bool bGetUserDefault)
{
    ECRESULT er = erSuccess;
    objectid_t userid;
    quotadetails_t details;
    userobject_type_t ulType;

    UserPlugin *lpPlugin = NULL;

	if(IsLocalItem(ulId)) {
		if (bGetUserDefault)
			er = ZARAFA_E_NO_SUPPORT;
		else {
			lpDetails->bUseDefaultQuota = false;
			lpDetails->bIsUserDefaultQuota = false;
			lpDetails->llWarnSize = 0;
			lpDetails->llSoftSize = 0;
			lpDetails->llHardSize = 0;
		}
		goto exit;
	}

	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetQuota(ulId, bGetUserDefault, lpDetails);
	if (er == erSuccess)
		goto exit; /* Cache contained requested information, we're done. */

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	er = GetItemInfo(ulId, &ulType, &userid);
	if (er != erSuccess)
		goto exit;

	if ((ulType == USEROBJECT_TYPE_COMPANY && !m_lpSession->GetSessionManager()->IsHostedSupported()) ||
		(ulType != USEROBJECT_TYPE_COMPANY && bGetUserDefault)) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

    try {
        details = *lpPlugin->getQuota(ulType, userid, bGetUserDefault);
    } catch(objectnotfound &) {
        MoveOrDeleteLocalObject(ulId, ulType);
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to get quota for %s: %s", ObjectTypeToName(ulType), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

	/* Update cache so we don't have to bug the plugin until the data has changed.
	 * Note that we don't care if the update succeeded, if it fails we will retry
	 * when the quota is requested for a second time. */
	m_lpSession->GetSessionManager()->GetCacheManager()->SetQuota(ulId, bGetUserDefault, details);

    *lpDetails = details;

exit:
    return er;
}

ECRESULT ECUserManagement::SearchObjectAndSync(userobject_type_t type, char *szSearchString, unsigned int ulFlags, unsigned int *lpulID)
{
	ECRESULT er = erSuccess;
	list<objectsignature_t> lObjectsignature;
	list<objectsignature_t>::iterator iterSignature;
	unsigned int ulId = 0;
	string strUsername;
	string strCompanyname;
	ECSession *lpecSession = dynamic_cast<ECSession*>(m_lpSession);
	UserPlugin *lpPlugin = NULL;
	char *szHideEveryone = NULL;

    // Special case: SYSTEM
    if(type == USEROBJECT_TYPE_USER && stricmp(szSearchString, ZARAFA_ACCOUNT_SYSTEM) == 0) {
        *lpulID = ZARAFA_UID_SYSTEM;
        goto exit;
    } else if(type == USEROBJECT_TYPE_GROUP && stricmp(szSearchString, ZARAFA_ACCOUNT_EVERYONE) == 0) {
		// Hide group everyone when requested
		if (lpecSession && lpecSession->GetSecurity()->GetAdminLevel() == 0) {
			szHideEveryone = lpecSession->GetSessionManager()->GetConfig()->GetSetting("hide_everyone");
			if (szHideEveryone && parseBool(szHideEveryone)) {
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}
		}
		
		*lpulID = ZARAFA_UID_EVERYONE;
		goto exit;
	}

	if (!lpecSession) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	/*
	 * The search string might be the following:
	 *  1) Portion of the fullname
	 *  2) Portion of the username
	 *  3) Portion of the username including company name
	 *  4) Portion of the email address
	 *
	 *  Regular search in the plugin will correctly handle (1), (2) and (4),
	 *  but not (3) since the username, company name combination depends on the
	 *  server.cfg settings. Because of that we need to check if we can convert
	 *  the search string into a username & companyname which will tell us if
	 *  we are indeed received a string of type (3) and should use resolveName().
	 *  However it is common that '@' is used as seperation character, which could
	 *  mean that email addresses are also identified as (3). So as fallback we
	 *  should still call searchObject() if the search string contains an '@'.
	 */
	if (type == USEROBJECT_TYPE_USER ||
		type == USEROBJECT_TYPE_NONACTIVE ||
		type == USEROBJECT_TYPE_GROUP) {

		/*
		 * We don't care about the return argument, when the call failed
		 * strCompanyname will remain empty and we will catch that when
		 * deciding if we are going to call resolveName() or searchObject().
		 */
		GetUserAndCompanyFromLoginName(szSearchString, &strUsername, &strCompanyname);
	}

	try {
		if (!strCompanyname.empty()) {
			objectsignature_t resolved;

			try {
				resolved = lpPlugin->resolveName(type, strUsername, strCompanyname);

				/* Loginname correctly resolved, this is an exact match so we don't
				 * have to do anything else for this object type.
				 * TODO: Optimize this further by sending message to caller that this
				 * entry is a 100% match and doesn't need to try to resolve any other
				 * object type. (IMPORTANT: when doing this, make sure we still check
				 * if the returned object is actually visible to the user or not!) */
				lObjectsignature.push_back(resolved);
			} catch (exception &) { /* We should_ catch all exceptions in one go */
				/* Apparently the searchkey wasn't a loginname after all, or it was invalid */
				lObjectsignature = lpPlugin->searchObject(type, szSearchString, ulFlags);
			}
		} else
			lObjectsignature = lpPlugin->searchObject(type, szSearchString, ulFlags);
	} catch(notimplemented &) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	} catch(objectnotfound &) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	} catch(notsupported &) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	} catch(std::exception &e) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to perform %s search for string '%s' on user database: %s", ObjectTypeToName(type), szSearchString, e.what());
		er = ZARAFA_E_PLUGIN_ERROR;
		goto exit;
	}

	if (lObjectsignature.empty()) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lObjectsignature.sort();
	lObjectsignature.unique();

	/* Check each returned entry to see which one we are allowed to view
	 * TODO: check with a point system,
	 * if you have 2 objects, one have a match of 99% and one 50%
	 * use the one with 99% */
	for (iterSignature = lObjectsignature.begin(); iterSignature != lObjectsignature.end(); iterSignature++) {
		unsigned int ulIdTmp = 0;

		er = GetLocalObjectIdOrCreate(type, *iterSignature, &ulIdTmp);
		if (er != erSuccess)
			goto exit;

		if (lpecSession->GetSecurity()->IsUserObjectVisible(ulIdTmp) != erSuccess)
			continue;

		/* We can only report a single (visible) entry, when we find multiple entries report a collision */
		if (ulId == 0) {
			ulId = ulIdTmp;
		} else if (ulId != ulIdTmp) {
			er = ZARAFA_E_COLLISION;
			goto exit;
		}
	}

	*lpulID = ulId;

exit:
	return er;
}

ECRESULT ECUserManagement::QueryContentsRowData(struct soap *soap, ECObjectTableList *lpRowList, struct propTagArray *lpPropTagArray, struct rowSet **lppRowSet)
{
	ECRESULT er = erSuccess;
	int i = 0, j = 0;
	map<userobject_type_t, list<objectid_t>* > lstObjects;
	map<userobject_type_t, list<objectid_t>* >::iterator iterObjects;

	struct rowSet *lpsRowSet = NULL;
	objectid_t externid;
	map<objectid_t, objectdetails_t> * lpExternObjectDetails;
	map<userobject_type_t, map<objectid_t, objectdetails_t> * > mapAllObjectDetails;
	map<userobject_type_t, map<objectid_t, objectdetails_t> * >::iterator iterAllObjectDetails;

	map<objectid_t, objectdetails_t>::iterator iterExternObjectDetails;
	map<pair<objectid_t, userobject_type_t>, unsigned int> mapExternObjectIdToRowId;
	map<pair<objectid_t, userobject_type_t>, unsigned int>::iterator iterExternIdToRowId;
	map<pair<objectid_t, userobject_type_t>, unsigned int> mapExternIdToObjectId;
	
	objectdetails_t details;

	unsigned int ulRowId = 0;
	ECObjectTableList::iterator iterRowList;
	UserPlugin *lpPlugin = NULL;
	userobject_type_t ulType;
	string signature;

	er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
	if (er != erSuccess)
		goto exit;

	ASSERT(lpRowList != NULL);

	lpsRowSet = s_alloc<struct rowSet>(soap);
	lpsRowSet->__size = 0;
	lpsRowSet->__ptr = NULL;

	if (lpRowList->empty()) {
		*lppRowSet = lpsRowSet;
		goto exit; // success
	}

	// We return a square array with all the values
	lpsRowSet->__size = lpRowList->size();
	lpsRowSet->__ptr = s_alloc<propValArray>(soap, lpRowList->size());
	memset(lpsRowSet->__ptr, 0, sizeof(propValArray) * lpRowList->size());

	// Get Extern ID and Types for all items
	i = 0;
	for(iterRowList = lpRowList->begin(); iterRowList != lpRowList->end(); iterRowList++) {
		er = GetItemInfo(iterRowList->ulObjId, &ulType, &externid);
		if (er != erSuccess) {
			er = erSuccess; /* Skip entry, but don't complain */
			goto skip;
		}
		
		// See if the item data is cached
		if(m_lpSession->GetSessionManager()->GetCacheManager()->GetUserDetails(iterRowList->ulObjId, ulType, &details) == erSuccess)
		{
		    // Item was in the cache, no need to query the backend. We directly put the data in mapAllObjectDetails
		    iterAllObjectDetails = mapAllObjectDetails.find(ulType);
		    
		    if(iterAllObjectDetails == mapAllObjectDetails.end()) {
		        mapAllObjectDetails.insert(map<userobject_type_t, map<objectid_t, objectdetails_t> * >::value_type(ulType, new map<objectid_t, objectdetails_t>));
    		    iterAllObjectDetails = mapAllObjectDetails.find(ulType);
		    }
		    
		    iterAllObjectDetails->second->insert(map<objectid_t, objectdetails_t>::value_type(externid, details));
		} else {
		    // Item needs to be retrieved from the plugin
		    
            iterObjects = lstObjects.find(ulType);
            if(iterObjects == lstObjects.end()) {
                lstObjects.insert(std::pair<userobject_type_t, std::list<objectid_t>* > (ulType, new std::list<objectid_t>));
                iterObjects = lstObjects.find(ulType);
            }
            iterObjects->second->push_back(externid);

        }

        mapExternObjectIdToRowId[std::pair<objectid_t, userobject_type_t>(externid, ulType)] = i;
        mapExternIdToObjectId[std::pair<objectid_t, userobject_type_t>(externid,ulType)] = iterRowList->ulObjId;
skip:
        i++;
    }

    // Do one request to the plugin for each type of object requested
    try {
        for(iterObjects = lstObjects.begin(); iterObjects != lstObjects.end(); iterObjects++) {
            lpExternObjectDetails = lpPlugin->getObjectDetails(iterObjects->first, *iterObjects->second).release();
            
            // Add all the results to mapAllObjectDetails
            if(!lpExternObjectDetails->empty()) {
                iterAllObjectDetails = mapAllObjectDetails.find(iterObjects->first);
                if(iterAllObjectDetails == mapAllObjectDetails.end()) {
                    mapAllObjectDetails.insert(map<userobject_type_t, map<objectid_t, objectdetails_t> * >::value_type(iterObjects->first, new map<objectid_t, objectdetails_t>));
                    iterAllObjectDetails = mapAllObjectDetails.find(iterObjects->first);
                }
                 
                // Copy each item over
                for(iterExternObjectDetails = lpExternObjectDetails->begin(); iterExternObjectDetails != lpExternObjectDetails->end(); iterExternObjectDetails++) {
                    iterAllObjectDetails->second->insert(*iterExternObjectDetails);
                }                   
            }

            delete lpExternObjectDetails; // No longer required.
			/* We convert user and companyname to loginname later this function */
        }
	} catch (objectnotfound &) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	} catch (notsupported &) {
        er = ZARAFA_E_NO_SUPPORT;
        goto exit;
    } catch(std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to retrieve %s list from external user source: %s", ObjectTypeToName(iterObjects->first), e.what());
        er = ZARAFA_E_PLUGIN_ERROR;
        goto exit;
    }

    // mapAllObjectDetails now contains the details per type of all objects that we need.

    // Loop through user data and fill in data in rowset
    for(iterAllObjectDetails = mapAllObjectDetails.begin(); iterAllObjectDetails != mapAllObjectDetails.end(); iterAllObjectDetails++) {
        ulType = iterAllObjectDetails->first;
        lpExternObjectDetails = iterAllObjectDetails->second;

        for(iterExternObjectDetails = lpExternObjectDetails->begin(); iterExternObjectDetails != lpExternObjectDetails->end(); iterExternObjectDetails++)
        {
            objectid_t objectid = iterExternObjectDetails->first;
            objectdetails_t objectdetails = iterExternObjectDetails->second;

            // Get the row id the information should be in
            ulRowId = mapExternObjectIdToRowId[std::pair<objectid_t, userobject_type_t>(objectid, ulType) ];

			// Get the local object id for the item
			unsigned int ulObjectId = mapExternIdToObjectId[std::pair<objectid_t, userobject_type_t>(objectid, ulType)];

			// Add data to the cache before updating it in UpdateUserDetails()
			m_lpSession->GetSessionManager()->GetCacheManager()->SetUserDetails(ulObjectId, ulType, &objectdetails);

			er = UpdateUserDetails(&objectdetails);
			if (er != erSuccess)
				goto exit;

			ConvertObjectDetailsToProps(soap, ulObjectId, ulType, &objectdetails, lpPropTagArray, &lpsRowSet->__ptr[ulRowId]);

			// Ignore the error, missing rows will be filled in automatically later (assumes lpsRowSet is untouched on error!!)
			
		}
	}

    // Free details
	for (iterAllObjectDetails = mapAllObjectDetails.begin();
		 iterAllObjectDetails != mapAllObjectDetails.end();
		 iterAllObjectDetails++)
		delete iterAllObjectDetails->second;

	// Loop through items, set local data and set other non-found data to NOT FOUND
	i = 0;
	for (iterRowList = lpRowList->begin(); iterRowList != lpRowList->end(); iterRowList++) {
		// Fill in any missing data with ZARAFA_E_NOT_FOUND
		if(IsLocalItem(iterRowList->ulObjId)) {
			objectdetails_t details;

			er = GetLocalObjectDetails(iterRowList->ulObjId, &details);
			if (er != erSuccess)
				goto exit;

			er = ConvertObjectDetailsToProps(soap, iterRowList->ulObjId, details.GetType(), &details, lpPropTagArray, &lpsRowSet->__ptr[i]);
			if(er != erSuccess)
				goto exit;
		} else {
			if (lpsRowSet->__ptr[i].__ptr == NULL) {
				lpsRowSet->__ptr[i].__ptr = s_alloc<propVal>(soap, lpPropTagArray->__size);
				lpsRowSet->__ptr[i].__size = lpPropTagArray->__size;

				for (j = 0; j < lpPropTagArray->__size; j++) {
					lpsRowSet->__ptr[i].__ptr[j].ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTagArray->__ptr[j]));
					lpsRowSet->__ptr[i].__ptr[j].Value.ul = ZARAFA_E_NOT_FOUND;
					lpsRowSet->__ptr[i].__ptr[j].__union = SOAP_UNION_propValData_ul;
				}
			}
		}
		i++;
	}

    // All done
	*lppRowSet = lpsRowSet;

exit:
    for (iterObjects = lstObjects.begin(); iterObjects != lstObjects.end(); iterObjects++) {
        delete iterObjects->second;
   }

	return er;
}

ECRESULT ECUserManagement::QueryHierarchyRowData(struct soap *soap, ECObjectTableList *lpRowList, struct propTagArray *lpPropTagArray, struct rowSet **lppRowSet)
{
	ECRESULT er = erSuccess;
	struct rowSet *lpsRowSet = NULL;
	ECObjectTableList::iterator iterRowList;
	unsigned int i = 0;
	userobject_type_t ulType;

	ASSERT(lpRowList != NULL);

	lpsRowSet = s_alloc<struct rowSet>(soap);
	lpsRowSet->__size = 0;
	lpsRowSet->__ptr = NULL;

	if (lpRowList->empty()) {
		*lppRowSet = lpsRowSet;
		goto exit; // success
	}

	// We return a square array with all the values
	lpsRowSet->__size = lpRowList->size();
	lpsRowSet->__ptr = s_alloc<propValArray>(soap, lpRowList->size());
	memset(lpsRowSet->__ptr, 0, sizeof(propValArray) * lpRowList->size());

	for (iterRowList = lpRowList->begin(); iterRowList != lpRowList->end(); iterRowList++) {
		/* Although it propbably doesn't make a lot of sense, we need to check for company containers here.
		 * this is because with convenient depth tables, the children of the Zarafa Address Book will not
		 * only be the Global Address Book, but also the company containers. */
		if (iterRowList->ulObjId == 0 || iterRowList->ulObjId == 1) {
			er = ConvertABContainerToProps(soap, iterRowList->ulObjId, lpPropTagArray, &lpsRowSet->__ptr[i]);
			if (er != erSuccess)
				goto exit;
		} else {
			objectdetails_t details;

			er = GetExternalObjectDetails(iterRowList->ulObjId, &details);
			if (er != erSuccess)
				goto exit;
				
            er = GetItemObjectType(iterRowList->ulObjId, &ulType);
            if (er != erSuccess)
                goto exit;

			er = ConvertContainerObjectDetailsToProps(soap, iterRowList->ulObjId, ulType, &details, lpPropTagArray, &lpsRowSet->__ptr[i]);
			if(er != erSuccess)
				goto exit;
		}
		i++;
	}

    // All done
	*lppRowSet = lpsRowSet;

exit:
	return er;
}

ECRESULT ECUserManagement::GetUserAndCompanyFromLoginName(const string &strLoginName, string *lpstrUserName, string *lpstrCompanyName)
{
	ECRESULT er = erSuccess;
	string format = m_lpSession->GetSessionManager()->GetConfig()->GetSetting("loginname_format");
	bool bHosted = m_lpSession->GetSessionManager()->IsHostedSupported();
	size_t pos_u = format.find("%u");
	size_t pos_c = format.find("%c");
	string username, companyname;
	string start, middle, end;
	size_t pos_s, pos_m, pos_e;
	size_t pos_a, pos_b;

	if (!bHosted || pos_u == string::npos || pos_c == string::npos) {
		/* When hosted is enabled, return a warning. Otherwise
		 * this call was successfull */
		if (bHosted)
			er = ZARAFA_W_PARTIAL_COMPLETION;
		username = strLoginName;
		companyname = "";
		goto exit;
	}

	pos_a = (pos_u < pos_c) ? pos_u : pos_c;
	pos_b = (pos_u < pos_c) ? pos_c : pos_u;

	/*
	 * Read strLoginName to determine the fields, this check should
	 * keep in mind that there can be characters before, inbetween and
	 * after the different fields.
	 */
	start = format.substr(0, pos_a);
	middle = format.substr(pos_a + 2, pos_b - pos_a - 2);
	end = format.substr(pos_b + 2, string::npos);

	/*
	 * There must be some sort of seperator between username and companyname.
	 */
	if (middle.empty()) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	pos_s = !start.empty() ? strLoginName.find(start, 0) : 0;
	pos_m = strLoginName.find(middle, pos_s + start.size());
	pos_e = !end.empty() ? strLoginName.find(end, pos_m + middle.size()) : string::npos;

	if ((!start.empty() && pos_s == string::npos) ||
		(!middle.empty() && pos_m == string::npos) ||
		(!end.empty() && pos_e == string::npos)) {
		/* When hosted is enabled, return a warning. Otherwise
		 * this call was successfull */
		if (strLoginName != ZARAFA_ACCOUNT_SYSTEM &&
			strLoginName != ZARAFA_ACCOUNT_EVERYONE)
				er = ZARAFA_E_INVALID_PARAMETER;
		username = strLoginName;
		companyname = "";
		goto exit;
	}

	if (pos_s == string::npos)
		pos_s = 0;
	if (pos_m == string::npos)
		pos_m = pos_b;

	if (pos_u < pos_c) {
		username = strLoginName.substr(pos_a, pos_m - pos_a);
		companyname = strLoginName.substr(pos_m + middle.size(), pos_e - pos_m - middle.size());
	} else {
		username = strLoginName.substr(pos_m + middle.size(), pos_e - pos_m - middle.size());
		companyname = strLoginName.substr(pos_a, pos_m - pos_a);
	}

	/* Neither username or companyname are allowed to be empty */
	if (username.empty() || companyname.empty()) {
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

exit:
	*lpstrUserName = username;
	*lpstrCompanyName = companyname;
	return er;
}

ECRESULT ECUserManagement::ConvertLoginToUserAndCompany(objectdetails_t *lpDetails)
{
	ECRESULT er = erSuccess;
	string loginname;
	string companyname;

	if (lpDetails->GetType() != USEROBJECT_TYPE_USER &&
		lpDetails->GetType() != USEROBJECT_TYPE_NONACTIVE &&
		lpDetails->GetType() != USEROBJECT_TYPE_GROUP)
			goto exit;

	er = GetUserAndCompanyFromLoginName(lpDetails->GetPropString(OB_PROP_S_LOGIN), &loginname, &companyname);
	/* GetUserAndCompanyFromLoginName() uses ZARAFA_W_PARTIAL_COMPLETION to indicate
	 * it could not fully convert the loginname. This means the user provided an
	 * invalid parameter and we should give the proper code back to the user */
	if (er == ZARAFA_W_PARTIAL_COMPLETION)
		er = ZARAFA_E_INVALID_PARAMETER;
	if (er != erSuccess)
		goto exit;

	lpDetails->SetPropString(OB_PROP_S_LOGIN, loginname);
	lpDetails->SetPropString(OB_PROP_S_COMPANYNAME, companyname);
	/* Groups require fullname to be updated as well */
	if (lpDetails->GetType() == USEROBJECT_TYPE_GROUP)
		lpDetails->SetPropString(OB_PROP_S_FULLNAME, loginname);

exit:
	return er;
}

ECRESULT ECUserManagement::ConvertUserAndCompanyToLogin(objectdetails_t *lpDetails)
{
	ECRESULT er = erSuccess;
	string format = m_lpSession->GetSessionManager()->GetConfig()->GetSetting("loginname_format");
	bool bHosted = m_lpSession->GetSessionManager()->IsHostedSupported();
	string login = lpDetails->GetPropString(OB_PROP_S_LOGIN);
	string company = lpDetails->GetPropString(OB_PROP_S_COMPANYNAME);
	size_t pos = 0;


	/*
	 * We don't have to do anything when hosted is disabled,
	 * when we perform this operation on SYSTEM or EVERYONE (since they don't belong to a company),
	 * or when the user objecttype does not need any conversions.
	 */
	if (!bHosted ||
		login == ZARAFA_ACCOUNT_SYSTEM ||
		login == ZARAFA_ACCOUNT_EVERYONE ||
		(lpDetails->GetType() != USEROBJECT_TYPE_USER &&
		 lpDetails->GetType() != USEROBJECT_TYPE_NONACTIVE &&
		 lpDetails->GetType() != USEROBJECT_TYPE_GROUP))
			goto exit;

	/*
	 * This really shouldn't happen, the loginname must contain *something* under any circumstance,
	 * company must contain *something* when hosted is enabled, and we already confirmed that this
	 * is the case.
	 */
	if (login.empty() || company.empty()) {
		er = ZARAFA_E_UNABLE_TO_COMPLETE;
		goto exit;
	}

	pos = format.find("%u");
	if (pos != string::npos)
		format.replace(pos, 2, login);

	pos = format.find("%c");
	if (pos != string::npos)
		format.replace(pos, 2, company);

	lpDetails->SetPropString(OB_PROP_S_LOGIN, format);
	/* Groups require fullname to be updated as well */
	if (lpDetails->GetType() == USEROBJECT_TYPE_GROUP)
		lpDetails->SetPropString(OB_PROP_S_FULLNAME, format);

exit:
	return er;
}

ECRESULT ECUserManagement::ConvertExternIDsToLocalIDs(objectdetails_t *lpDetails)
{
	ECRESULT er = erSuccess;
	list<string> lstExternIDs;
	list<string>::iterator iterExternIDs;
	string strQuery;
    ECDatabase *lpDatabase = m_lpSession->GetDatabase();
    DB_RESULT lpResult = NULL;
    DB_ROW lpRow = NULL;



    if(lpDatabase == NULL) {
        er = ZARAFA_E_DATABASE_ERROR;
        goto exit;
    }

	// type must be USER, since NONACTIVE users cannot login to press a send button.
	if (lpDetails->GetType() != USEROBJECT_TYPE_USER) {
		lpDetails->ClearPropList(OB_PROP_L_SENDAS);
		goto exit;
	}

	lstExternIDs = lpDetails->GetPropList(OB_PROP_L_SENDAS);
	if (lstExternIDs.empty())
		goto exit;

	lpDetails->ClearPropList(OB_PROP_L_SENDAS);

	strQuery = "SELECT id FROM users WHERE (object_type = "+
		stringify(USEROBJECT_TYPE_USER)+" OR object_type = "+
		stringify(USEROBJECT_TYPE_NONACTIVE)+") AND externid in (";

	iterExternIDs = lstExternIDs.begin();	
	while (true) {
		strQuery += lpDatabase->EscapeBinary((unsigned char*)iterExternIDs->data(), iterExternIDs->size());
		iterExternIDs++;
		if (iterExternIDs == lstExternIDs.end())
			break;
		strQuery += ",";
	}

	strQuery += ")";

    er = lpDatabase->DoSelect(strQuery, &lpResult);
    if(er != erSuccess)
        goto exit;

	while (true) {
		lpRow = lpDatabase->FetchRow(lpResult);
		if(lpRow == NULL || lpRow[0] == NULL)
			break;

		lpDetails->AddPropString(OB_PROP_L_SENDAS, lpRow[0]);
    }
	

exit:
    if(lpResult)
        lpDatabase->FreeResult(lpResult);

	return er;
}

ECRESULT ECUserManagement::UpdateUserDetails(objectdetails_t *lpDetails)
{
	ECRESULT er = erSuccess;

	er = ConvertUserAndCompanyToLogin(lpDetails);
	if (er != erSuccess)
		goto exit;

	er = ConvertExternIDsToLocalIDs(lpDetails);
	if (er != erSuccess)
		goto exit;

exit:
	return er;
}

// ******************************************************************************************************
//
// Create/Delete routines
//
// ******************************************************************************************************

// Perform a license check
ECRESULT ECUserManagement::CheckUserLicense(unsigned int *lpulLicenseStatus)
{
	ECRESULT er = erSuccess;
    std::string strQuery;
	unsigned int ulTotalUsers = 0;
    unsigned int ulActive = 0;
    unsigned int ulNonActive = 0;
    unsigned int ulLicensedUsers = 0;
	unsigned int ulActiveLimit = 0;
	unsigned int ulNonActiveLimit = 0;

	// NOTE: this function is only a precaution; if you circumvent this function, and create 'too many'
	// users, the zarafa-licensed server will complain and disable all outlook logins.

	*lpulLicenseStatus = 0;

	er = GetUserCount(&ulActive, &ulNonActive);
	if (er != erSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to query user count");
		goto exit;
	}
	ulTotalUsers = ulActive + ulNonActive;

	er = m_lpSession->GetSessionManager()->GetLicensedUsers(&ulLicensedUsers);
	if (er != erSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to query license user count");
		goto exit;
	}

	/* Active limit is always licensed users limit when the limit is 0 we have unlimited users,
	 * Inactive limit is minimally the licensed user limit + 20 and maximally the licensed user limit + 1/2 licensed user limit */
	ulActiveLimit = ulLicensedUsers;
	ulNonActiveLimit = ulLicensedUsers ? ulLicensedUsers + ((ulLicensedUsers >= 40) ? (ulLicensedUsers / 2) : 20) : 0;

	if (ulActiveLimit) {
		if (ulActive == ulActiveLimit)
			*lpulLicenseStatus |= USERMANAGEMENT_LIMIT_ACTIVE_USERS;
		if (ulActive > ulActiveLimit)
			*lpulLicenseStatus |= USERMANAGEMENT_EXCEED_ACTIVE_USERS;
	}

	if (ulNonActiveLimit) {
		if (ulTotalUsers == ulNonActiveLimit)
			*lpulLicenseStatus |= USERMANAGEMENT_LIMIT_NONACTIVE_USERS;
		if (ulTotalUsers > ulNonActiveLimit)
			*lpulLicenseStatus |= USERMANAGEMENT_EXCEED_NONACTIVE_USERS;
	}

exit:
	return er;
}

// Create a local user corresponding to the given userid on the external database
ECRESULT ECUserManagement::CreateLocalObject(userobject_type_t type, objectsignature_t signature, unsigned int *lpulId) {
    ECRESULT er = erSuccess;
    ECDatabase *lpDatabase = m_lpSession->GetDatabase();
    std::string strQuery;
    objectdetails_t details;
    unsigned int ulId;
    unsigned int ulCompanyId;
	ABEID eid(MAPI_ABCONT, *(GUID*)&MUIDECSAB_SERVER, 1);
	unsigned int ulLicenseStatus;
	SOURCEKEY sSourceKey;

    UserPlugin *lpPlugin = NULL;

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

	if(lpDatabase == NULL) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if(type == USEROBJECT_TYPE_USER || type == USEROBJECT_TYPE_NONACTIVE) {
		er = CheckUserLicense(&ulLicenseStatus);
		if (er != erSuccess)
			goto exit;

		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Auto-creating user from external source");

		if ((type == USEROBJECT_TYPE_USER) && (ulLicenseStatus & USERMANAGEMENT_BLOCK_CREATE_ACTIVE_USER)) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create active user: Your license doesn't permit this amount of users.");
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
		}
		if ((type == USEROBJECT_TYPE_NONACTIVE) && (ulLicenseStatus & USERMANAGEMENT_BLOCK_CREATE_NONACTIVE_USER)) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create non-active user: Your license doesn't permit this amount of users.");
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
		}
	}

	try {
		details = *lpPlugin->getObjectDetails(type, signature.id);

		/*
		 * The property OB_PROP_S_LOGIN is mandatory, when that property is an empty string
		 * somebody (aka: system administrator) has messed up its LDAP tree or messed around
		 * with the Database.
		 */
		if (details.GetPropString(OB_PROP_S_LOGIN).empty()) {
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create object in local database: %s has no name", ObjectTypeToName(type));
			er = ZARAFA_E_UNKNOWN_OBJECT;
			goto exit;
		}

		er = UpdateUserDetails(&details);
		if (er != erSuccess)
			goto exit;
	} catch (objectnotfound &) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	} catch (notsupported &) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	} catch(std::exception &e) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to get object data while adding new %s: %s", ObjectTypeToName(type), e.what());
		er = ZARAFA_E_PLUGIN_ERROR;
		goto exit;
	}

	/*
	 * 1) we create the object in the database, for users/groups we set the company to 0
	 * 2) Resolve companyid for users/groups to a companyid
	 * 3) Update object in the database to correctly set the company column
	 *
	 * Now why do we use this approach:
	 * Suppose we have user A who is member of Company B and is the sysadmin for B.
	 * The following will happen when we insert all information in a single shot:
	 * 1) GetLocalObjectIdOrCreate(A) is called
	 * 2) Database does not contain A, so CreateLocalObject(A) is called
	 * 3) CreateLocalObject(A) calls ResolveObjectAndSync(B) for COMPANYID
	 *   3.1) B is resolved, GetLocalObjectIdOrCreate(B) is called
	 *   3.2) Database does not contain B, so CreateLocalObject(B) is called
	 *   3.3) B is inserted into the database, createcompany script is called
	 *   3.4) createcompany script calls getCompany(B)
	 *   3.5) Server obtains details for B, finds A as sysadmin for B
	 *   3.6) Server calls GetLocalObjectIdOrCreate(A)
	 *   3.7) Database does not contain A, so CreateLocalObject(A) is called
	 *   3.8) CreateLocalObject(A) inserts A into the database
	 *   3.9) createcompany script does its work and completes
	 * 4) CreateLocalObject(A) inserts A into the database
	 * 5) BOOM! Primary key violation since A is already present in the database
	 */
	strQuery =
		"INSERT INTO users (externid, object_type, signature) "
		"VALUES("
			"'" + lpDatabase->Escape(signature.id) + "', " +
			stringify(type) + ", " +
			"'" + lpDatabase->Escape(signature.signature) + "')";
	er = lpDatabase->DoInsert(strQuery, &ulId);
	if(er != erSuccess)
		goto exit;

	if (type == USEROBJECT_TYPE_COMPANY)
		ulCompanyId = 0;
	else {
	    er = ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY,
								  (char *)details.GetPropString(OB_PROP_S_COMPANYNAME).c_str(),
								  &ulCompanyId);
	    if (er != erSuccess)
	        goto exit;
	}

	if (ulCompanyId) {
		strQuery =
			"UPDATE users "
			"SET company = " + stringify(ulCompanyId) + " "
			"WHERE id = " + stringify(ulId);
		er = lpDatabase->DoUpdate(strQuery);
		if (er != erSuccess)
			goto exit;
	}

	switch(type) {
	case USEROBJECT_TYPE_NONACTIVE:
	case USEROBJECT_TYPE_USER: {
			string strUserServer = details.GetPropString(OB_PROP_S_SERVERNAME);
			string strThisServer = g_lpSessionManager->GetConfig()->GetSetting("server_name");
			bool bDistributed = m_lpSession->GetSessionManager()->IsDistributedSupported();
			if (!bDistributed || strUserServer.compare(strThisServer) == 0) {
				execute_script(m_lpSession->GetSessionManager()->GetConfig()->GetSetting("createuser_script"),
							   m_lpLogger,
							   "ZARAFA_USER", details.GetPropString(OB_PROP_S_LOGIN).c_str(),
							   NULL);
			}
		}
        break;
	case USEROBJECT_TYPE_GROUP:
		execute_script(m_lpSession->GetSessionManager()->GetConfig()->GetSetting("creategroup_script"),
					   m_lpLogger,
					   "ZARAFA_GROUP", details.GetPropString(OB_PROP_S_LOGIN).c_str(),
					   NULL);
		break;
	case USEROBJECT_TYPE_COMPANY: {
			string strUserServer = details.GetPropString(OB_PROP_S_SERVERNAME);
			string strThisServer = g_lpSessionManager->GetConfig()->GetSetting("server_name");
			bool bDistributed = m_lpSession->GetSessionManager()->IsDistributedSupported();
			if (!bDistributed || strUserServer.compare(strThisServer) == 0) {
				execute_script(m_lpSession->GetSessionManager()->GetConfig()->GetSetting("createcompany_script"),
							   m_lpLogger,
							   "ZARAFA_COMPANY", details.GetPropString(OB_PROP_S_FULLNAME).c_str(),
							   NULL);
			}
		}
		break;
	default:
		break;
	}

	// Log the change to ICS
	er = GetABSourceKeyV1(ulId, &sSourceKey);
	if (er != erSuccess)
		goto exit;

	AddABChange(m_lpSession, ICS_AB_NEW, sSourceKey, SOURCEKEY(CbABEID(&eid), (char *)&eid));

	*lpulId = ulId;

exit:
    return er;
}

// Creates a local object under a specific object ID
ECRESULT ECUserManagement::CreateLocalObjectSimple(objectsignature_t signature, userobject_type_t type, unsigned int ulPreferredId) {
    ECRESULT er = erSuccess;
    ECDatabase *lpDatabase = m_lpSession->GetDatabase();
	objectdetails_t details;
    std::string strQuery;
	unsigned int ulCompanyId;
	UserPlugin *lpPlugin = NULL;
    DB_RESULT lpResult = NULL;

    // No user count checking or script starting in this function; it is only used in for addressbook synchronization in the offline server

	er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
	if(er != erSuccess)
		goto exit;

	try {
		details = *lpPlugin->getObjectDetails(type, signature.id);
		/* No need to convert the user and company name to login, since we are not using
		 * the loginname in this function. */
	} catch (objectnotfound &) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	} catch (notsupported &) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	} catch(std::exception &e) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to get details while adding new %s: %s", ObjectTypeToName(type), e.what());
		er = ZARAFA_E_PLUGIN_ERROR;
		goto exit;
	}

	if (type == USEROBJECT_TYPE_COMPANY)
		ulCompanyId = 0;
	else {
		er = ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY,
								  (char *)details.GetPropString(OB_PROP_S_COMPANYNAME).c_str(),
								  &ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	strQuery = "LOCK TABLES users WRITE";
	er = lpDatabase->DoInsert(strQuery);
	if (er != erSuccess)
		goto exit;

	// Check if we can use the preferred id.
	strQuery = "SELECT 0 FROM users WHERE id=" + stringify(ulPreferredId);
	er = lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		goto exit;

	if (lpDatabase->GetNumRows(lpResult) == 1) {
		strQuery =
			"INSERT INTO users (externid, object_type, company, signature) "
			"VALUES ("
				"'" + lpDatabase->Escape(signature.id) + "', " +
				stringify(type) + ", " +
				stringify(ulCompanyId) + ", " +
				"'" + lpDatabase->Escape(signature.signature) + "')";
	} else {
		strQuery =
			"INSERT INTO users (id, externid, object_type, company, signature) "
			"VALUES ("
				+ stringify(ulPreferredId) + ", " +
				"'" + lpDatabase->Escape(signature.id) + "', " +
				stringify(type) + ", " +
				stringify(ulCompanyId) + ", " +
				"'" + lpDatabase->Escape(signature.signature) + "')";
	}

    er = lpDatabase->DoInsert(strQuery);
    if(er != erSuccess)
        goto exit;

exit:
	lpDatabase->DoInsert("UNLOCK TABLES");

    return er;
}

// Check if an object has moved to a new company, or if it was created as new 
ECRESULT ECUserManagement::MoveOrCreateLocalObject(userobject_type_t ulType, objectsignature_t signature, unsigned int *lpulObjectId)
{
	ECRESULT er = erSuccess;
	objectdetails_t details;
	unsigned int ulObjectId = 0;
	unsigned int ulNewCompanyId = 0;

	/*
	 * This function is called when an object with an external id has appeared
	 * there are various reasons why this might happen:
	 * 1) Object was created, and we should create the local item.
	 * 2) Object was moved to different company, we should move the local item
	 * 3) Object hasn't changed, but multicompany support was toggled, we should move the local item
	 * 
	 * The obvious benefit from moving an user will be that it preserves the store, when
	 * deleting and recreating a user when it has moved the same will happen for the store
	 * belonging to that user. When we perform a move operation we will only update some
	 * columns in the database and the entire store is moved along with the user, preserving
	 * all messages in that store.
	 *
	 * How to determine if an object was created or moved:
	 * We arrived here when a search for an object within its currently known company succeeded,
	 * and we found a (until then unknown) objectid. What we need now is to perform a search
	 * in the database without announcing the company.
	 * If that does produce a result, then the user has moved, and we need to determine
	 * the new company.
	 */

	/* We don't support moving entire companies, create it. */
	if (ulType == USEROBJECT_TYPE_COMPANY) {
		er = CreateLocalObject(ulType, signature, lpulObjectId);
		goto exit;
	}

	/* If object doesn't have a local id, we should create it. */
	er = GetLocalId(signature.id, ulType, &ulObjectId);
	if (er != erSuccess) {
		er = CreateLocalObject(ulType, signature, lpulObjectId);
		goto exit;
	}

	/* Delete cache, to force call to plugin */
	er = m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulObjectId);
    if(er != erSuccess)
        goto exit;

	/* Result new object details */
	er = GetObjectDetailsAndSync(ulObjectId, &details);
	if (er != erSuccess)
		goto exit;

	er = ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY,
							  (char *)details.GetPropString(OB_PROP_S_COMPANYNAME).c_str(),
							  &ulNewCompanyId);
	if (er != erSuccess)
		goto exit;

	er = MoveLocalObject(ulObjectId, ulType, ulNewCompanyId, details.GetPropString(OB_PROP_S_LOGIN));
	if (er != erSuccess)
			goto exit;

	*lpulObjectId = ulObjectId;

exit:
	return er;
}

// Check if an object has moved to a new company, or if it was deleted complete
ECRESULT ECUserManagement::MoveOrDeleteLocalObject(unsigned int ulObjectId, userobject_type_t ulType)
{
	ECRESULT er = erSuccess;
	objectdetails_t details;
	unsigned int ulOldCompanyId = 0;
	unsigned int ulNewCompanyId = 0;

    if(IsLocalItem(ulObjectId)) {
        er = ZARAFA_E_NO_ACCESS;
        goto exit;
    }

	/*
	 * This function is called when an object with an external id has disappeared,
	 * there are various reasons why this might happen:
	 * 1) Object was deleted, and we should delete the local item.
	 * 2) Object was moved to different company, we should move the local item
	 * 3) Object hasn't changed, but multicompany support was toggled, we should move the local item
	 * 
	 * The obvious benefit from moving an user will be that it preserves the store, when
	 * deleting and recreating a user when it has moved the same will happen for the store
	 * belonging to that user. When we perform a move operation we will only update some
	 * columns in the database and the entire store is moved along with the user, preserving
	 * all messages in that store.
	 *
	 * How to determine if an object was deleted or moved:
	 * We arrived here when a search for an object within its currently known company failed,
	 * what we need now is to perform a search in the plugin without announcing the company.
	 * If that does produce a result, then the user has moved, and we need to determine
	 * the new company. Fortunately, the best call to determine if a user still exists
	 * is GetObjectDetailsAndSync() which automatically gives us the new company.
	 */

	/* We don't support moving entire companies, delete it. */
	if (ulType == USEROBJECT_TYPE_COMPANY) {
		er = DeleteLocalObject(ulObjectId, ulType);
		goto exit;
	}

	/* Request old company */
	er = GetItemCompany(ulObjectId, ulType, &ulOldCompanyId);
	if (er != erSuccess)
		goto exit;

	/* Delete cache, to force call to plugin */
	er = m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulObjectId);
    if(er != erSuccess)
        goto exit;

	/* Result new object details */
	er = GetObjectDetailsAndSync(ulObjectId, &details);
	if (er == ZARAFA_E_NOT_FOUND) {
		/* Object was indeed deleted, GetObjectDetailsAndSync() has called DeleteLocalObject() */
		er = erSuccess;
		goto exit;
	} else if (er != erSuccess) {
		goto exit;
	}

	er = ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY,
							  (char *)details.GetPropString(OB_PROP_S_COMPANYNAME).c_str(),
							  &ulNewCompanyId);
	if (er != erSuccess)
		goto exit;

	/* We have the object details, check if object was moved or not */
	if (ulOldCompanyId == ulNewCompanyId)
		er = DeleteLocalObject(ulObjectId, ulType);
	else
		er = MoveLocalObject(ulObjectId, ulType, ulNewCompanyId, details.GetPropString(OB_PROP_S_LOGIN));
	if (er != erSuccess)
			goto exit;

exit:
	return er;
}

// Move a local user with the specified id to a new company
ECRESULT ECUserManagement::MoveLocalObject(unsigned int ulObjectId, userobject_type_t ulType, unsigned int ulCompanyId, const string &strNewUserName)
{
	ECRESULT er = erSuccess;
	ECDatabase *lpDatabase = NULL;
	std::string strQuery;
	ABEID eid(MAPI_ABCONT, *(GUID*)&MUIDECSAB_SERVER, 1);
	bool bTransaction = false;
	SOURCEKEY sSourceKey;

    if(IsLocalItem(ulObjectId)) {
        er = ZARAFA_E_NO_ACCESS;
        goto exit;
    }

	if (ulType == USEROBJECT_TYPE_COMPANY) {
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	m_lpLogger->Log(EC_LOGLEVEL_INFO, "Auto-moving %s to different company from external source", ObjectTypeToName(ulType));

	lpDatabase = m_lpSession->GetDatabase();

	bTransaction = true;

	er = lpDatabase->Begin();
	if(er != erSuccess)
		goto exit;

	/*
	 * Moving a user to a different company consists of the following tasks:
	 * 1) Change 'company' column in 'users' table
	 * 2) Change 'company' column in 'stores' table
	 * 3) Change 'user_name' column in 'stores table
	 */
    strQuery =
		"UPDATE users "
		"SET company=" + stringify(ulCompanyId) + " "
		"WHERE id=" + stringify(ulObjectId);
	er = lpDatabase->DoUpdate(strQuery);
	if(er != erSuccess)
		goto exit;

	strQuery =
		"UPDATE stores "
		"SET company=" + stringify(ulCompanyId) + ", "
			"user_name='" + lpDatabase->Escape(strNewUserName) + "' "
		"WHERE user_id=" + stringify(ulObjectId);
	er = lpDatabase->DoUpdate(strQuery);
	if(er != erSuccess)
		goto exit;

	/* Log the change to ICS */
	er = GetABSourceKeyV1(ulObjectId, &sSourceKey);
	if (er != erSuccess)
		goto exit;

	er = AddABChange(m_lpSession, ICS_AB_CHANGE, sSourceKey, SOURCEKEY(CbABEID(&eid), (char *)&eid));
	if(er != erSuccess)
		goto exit;

	er = lpDatabase->Commit();
	if(er != erSuccess)
		goto exit;

	bTransaction = false;

	er = m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulObjectId);
    if(er != erSuccess)
        goto exit;

exit:
	if (bTransaction && er != erSuccess)
		lpDatabase->Rollback();

    return er;
}

// Delete a local user with the specified id
ECRESULT ECUserManagement::DeleteLocalObject(unsigned int ulObjectId, userobject_type_t ulType) {
    ECRESULT er = erSuccess;
    ECDatabase *lpDatabase = NULL;
    DB_RESULT lpResult = NULL;
    DB_ROW lpRow = NULL;
    unsigned int ulStoreId = 0;
    unsigned int ulCompanyId;
    std::string strQuery;
	ABEID eid(MAPI_ABCONT, *(GUID*)&MUIDECSAB_SERVER, 1);
	bool bTransaction = false;
	SOURCEKEY sSourceKey;

    if(IsLocalItem(ulObjectId)) {
        er = ZARAFA_E_NO_ACCESS;
        goto exit;
    }

	er = GetItemCompany(ulObjectId, ulType, &ulCompanyId);
	if (er != erSuccess)
		goto exit;

    m_lpLogger->Log(EC_LOGLEVEL_INFO, "Auto-deleting %s from external source", ObjectTypeToName(ulType));

    lpDatabase = m_lpSession->GetDatabase();

	if (ulType == USEROBJECT_TYPE_COMPANY) {
		/* We are removing a company, delete all company members as well since
		 * those members no longer exist either and we won't get a delete event
		 * for those users since we already get a ZARAFA_E_NOT_FOUND for the
		 * company... */
		strQuery =
			"SELECT id, object_type FROM users "
			"WHERE company = " + stringify(ulObjectId);
		er = lpDatabase->DoSelect(strQuery, &lpResult);
		if (er != erSuccess)
			goto exit;

    	m_lpLogger->Log(EC_LOGLEVEL_INFO, "Start auto-deleting %s members", ObjectTypeToName(ulType));

		while (1) {
			lpRow = lpDatabase->FetchRow(lpResult);
			if (lpRow == NULL || lpRow[0] == NULL || lpRow[1] == NULL)
				break;

			/* Perhaps the user was moved to a different company */
			er = MoveOrDeleteLocalObject(atoui(lpRow[0]), (userobject_type_t)atoui(lpRow[1]));
			if (er != erSuccess)
				goto exit;
		}

		lpDatabase->FreeResult(lpResult);
		lpResult = NULL;
		lpRow = NULL;

    	m_lpLogger->Log(EC_LOGLEVEL_INFO, "Done auto-deleting %s members", ObjectTypeToName(ulType));
	}

	bTransaction = true;

	er = lpDatabase->Begin();
	if(er != erSuccess)
		goto exit;

    strQuery =
		"DELETE FROM users "
		"WHERE id=" + stringify(ulObjectId);
    er = lpDatabase->DoDelete(strQuery);
	if(er != erSuccess)
		goto exit;

	// Delete the user ACL's
	strQuery =
		"DELETE FROM acl "
		"WHERE id=" + stringify(ulObjectId);
	er = lpDatabase->DoDelete(strQuery);
	if(er != erSuccess)
		goto exit;

	// Log the change to ICS
	er = GetABSourceKeyV1(ulObjectId, &sSourceKey);
	if (er != erSuccess)
		goto exit;

	er = AddABChange(m_lpSession, ICS_AB_DELETE, sSourceKey, SOURCEKEY(CbABEID(&eid), (char *)&eid));
	if(er != erSuccess)
		goto exit;

	er = lpDatabase->Commit();
	if(er != erSuccess)
		goto exit;

	bTransaction = false;

	switch (ulType) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		strQuery = "SELECT id FROM stores WHERE user_id=" + stringify(ulObjectId);
		er = lpDatabase->DoSelect(strQuery, &lpResult);

		if(er != erSuccess)
			goto exit;

		lpRow = lpDatabase->FetchRow(lpResult);

		if(lpRow == NULL || lpRow[0] == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		// just to make sure it's an integer ..
		ulStoreId = atoi(lpRow[0]);

		execute_script(m_lpSession->GetSessionManager()->GetConfig()->GetSetting("deleteuser_script"),
					   m_lpLogger,
					   "ZARAFA_STOREID", stringify(ulStoreId).c_str(),
					   NULL);
		break;
	case USEROBJECT_TYPE_GROUP:
		execute_script(m_lpSession->GetSessionManager()->GetConfig()->GetSetting("deletegroup_script"),
					   m_lpLogger,
					   "ZARAFA_GROUPID", stringify(ulObjectId).c_str(),
					   NULL);
		break;
	case USEROBJECT_TYPE_COMPANY:
		execute_script(m_lpSession->GetSessionManager()->GetConfig()->GetSetting("deletecompany_script"),
					   m_lpLogger,
					   "ZARAFA_COMPANYID", stringify(ulObjectId).c_str(),
					   NULL);
		break;
	default:
		break;
	}

    er = m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulObjectId);
    if(er != erSuccess)
        goto exit;

exit:
	if (bTransaction && er != erSuccess)
		lpDatabase->Rollback();

    if (lpResult)
        lpDatabase->FreeResult(lpResult);

    return er;
}

bool ECUserManagement::IsLocalItem(unsigned int ulUserId)
{
    return ulUserId == ZARAFA_UID_SYSTEM || ulUserId == ZARAFA_UID_EVERYONE;
}

// Convert a userdetails_t to a set of MAPI properties
ECRESULT ECUserManagement::ConvertObjectDetailsToProps(struct soap *soap, unsigned int ulId, userobject_type_t type, objectdetails_t *lpDetails, struct propTagArray *lpPropTags, struct propValArray *lpPropVals)
{
	ECRESULT er = erSuccess;
	int i = 0;
	struct propVal *lpPropVal;
	unsigned int ulOrder = 0;
	ECSession *lpecSession = NULL;
	std::string strEncExId;

	lpecSession = dynamic_cast<ECSession*>(m_lpSession);
	if (!lpecSession) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulId);
	if (er != erSuccess)
		goto exit;

	lpPropVals->__ptr = s_alloc<struct propVal>(soap, lpPropTags->__size);
	lpPropVals->__size = lpPropTags->__size;

	for(i=0;i<lpPropTags->__size;i++) {
	    unsigned int ulPropTag = lpPropTags->__ptr[i];
		lpPropVal = &lpPropVals->__ptr[i];

		// Handle PT_UNICODE as PT_STRING8 and convert later
		if(PROP_TYPE(ulPropTag) == PT_UNICODE)
		    ulPropTag = CHANGE_PROP_TYPE(ulPropTag, PT_STRING8);

		lpPropVal->ulPropTag = ulPropTag;

		switch(type) {
		default:
			er = ZARAFA_E_NOT_FOUND;
			break;
		case USEROBJECT_TYPE_NONACTIVE:
		case USEROBJECT_TYPE_USER: {
			switch(ulPropTag) {
            case PR_ENTRYID: {
				er = CreateABEntryID(soap, ulId, MAPI_MAILUSER, lpPropVal);
				if (er != erSuccess)
					goto exit;
                break;
            }
			case PR_EC_NONACTIVE:
				lpPropVal->Value.b = (type == USEROBJECT_TYPE_NONACTIVE);
				lpPropVal->__union = SOAP_UNION_propValData_b;
				break;
			case PR_EC_ADMINISTRATOR:
				lpPropVal->Value.ul = lpDetails->GetPropInt(OB_PROP_I_ADMINLEVEL);
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_COMPANY_NAME: {
				string strCompany = lpDetails->GetPropString(OB_PROP_S_COMPANYNAME);
				lpPropVal->Value.lpszA = s_strcpy(soap, strCompany.c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			}
			case PR_SMTP_ADDRESS:
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_EMAIL).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_MESSAGE_CLASS:
				lpPropVal->Value.lpszA = s_strcpy(soap, "IPM.Contact");
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_NORMALIZED_SUBJECT:
			case PR_7BIT_DISPLAY_NAME:
			case PR_DISPLAY_NAME:
			case 0x3A20001E:// PR_TRANSMITABLE_DISPLAY_NAME
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_FULLNAME).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_INSTANCE_KEY:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, 2 * sizeof(ULONG));
				lpPropVal->Value.bin->__size = 2*sizeof(ULONG);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
				memcpy(lpPropVal->Value.bin->__ptr+sizeof(ULONG), &ulOrder, sizeof(ULONG));
				break;
			case PR_OBJECT_TYPE:
				lpPropVal->Value.ul = MAPI_MAILUSER;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_DISPLAY_TYPE:
				lpPropVal->Value.ul = DT_MAILUSER;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_SEARCH_KEY: {
				std::string strSearchKey = (std::string)"ZARAFA:" + strToUpper(lpDetails->GetPropString(OB_PROP_S_EMAIL));
				lpPropVal->Value.bin =  s_alloc<xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, strSearchKey.size()+1);
				lpPropVal->Value.bin->__size = strSearchKey.size()+1;
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				strcpy((char *)lpPropVal->Value.bin->__ptr, strSearchKey.c_str());
				break;
			}
			case PR_ADDRTYPE:
				lpPropVal->Value.lpszA = s_strcpy(soap, "ZARAFA");
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_RECORD_KEY:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ULONG));
				lpPropVal->Value.bin->__size = sizeof(ULONG);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
				break;
			case 0x8006001E: { //PR_EMS_AB_HOME_MDB:
				/* Make BlackBerry happy */
				std::string hostname =
					"/o=Domain/ou=Location/cn=Configuration/cn=Servers/cn=" + HostnameFromSoap(soap) + "/cn=Microsoft Private MDB";
			    lpPropVal->Value.lpszA = s_strcpy(soap, hostname.c_str());
			    lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			    break;
			}
			case 0x8007001E: {//PR_EMS_AB_HOME_MTA:
				/* Make BlackBerry happy */
				std::string hostname =
					"/o=ZARAFA/ou=First Administrative Group/cn=Configuration/cn=Servers/cn=" + HostnameFromSoap(soap) + "/cn=Microsoft MTA";
				lpPropVal->Value.lpszA = s_strcpy(soap, hostname.c_str());
			    lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			    break;
			}
			case PR_ACCOUNT:
			case PR_EMAIL_ADDRESS:
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_LOGIN).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_SEND_INTERNET_ENCODING:
				lpPropVal->Value.ul = 0;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_SEND_RICH_INFO:
				lpPropVal->Value.b = true;
				lpPropVal->__union = SOAP_UNION_propValData_b;
				break;
			case PR_AB_PROVIDER_ID:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(GUID));
				lpPropVal->Value.bin->__size = sizeof(GUID);

				lpPropVal->__union = SOAP_UNION_propValData_bin;
				memcpy(lpPropVal->Value.bin->__ptr, MUIDECSAB_SERVER, sizeof(GUID));
				break;
			case PR_EMS_AB_X509_CERT: {
				list<string> strCerts = lpDetails->GetPropList(OB_PROP_L_CERTIFICATE);
				list<string>::iterator iCert;

				if (!strCerts.empty()) {
					unsigned int i = 0;

					lpPropVal->__union = SOAP_UNION_propValData_mvbin;
					lpPropVal->Value.mvbin.__size = strCerts.size();
					lpPropVal->Value.mvbin.__ptr = s_alloc<struct xsd__base64Binary>(soap, strCerts.size());

					for (i = 0, iCert = strCerts.begin(); iCert != strCerts.end(); iCert++, i++) {
						lpPropVal->Value.mvbin.__ptr[i].__size = iCert->size();
						lpPropVal->Value.mvbin.__ptr[i].__ptr = s_alloc<unsigned char>(soap, iCert->size());

						memcpy(lpPropVal->Value.mvbin.__ptr[i].__ptr, iCert->data(), iCert->size());
					}
				} else {
					lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
					lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
					lpPropVal->__union = SOAP_UNION_propValData_ul;
				}
				break;
			}
			case PR_EC_SENDAS_USER_ENTRYIDS: {
				list<string> userIds = lpDetails->GetPropList(OB_PROP_L_SENDAS);
				list<string>::iterator iterUserIds;

				if (!userIds.empty()) {
					unsigned int i;
					struct propVal sPropVal;

					lpPropVal->__union = SOAP_UNION_propValData_mvbin;
					lpPropVal->Value.mvbin.__size = userIds.size();
					lpPropVal->Value.mvbin.__ptr = s_alloc<struct xsd__base64Binary>(soap, userIds.size());

					for (i = 0, iterUserIds = userIds.begin(); iterUserIds != userIds.end(); iterUserIds++, i++) {
	
						er = CreateABEntryID(soap, atoui((char *)iterUserIds->c_str()), MAPI_MAILUSER, &sPropVal);
						if (er != erSuccess)
							goto exit;
						
						lpPropVal->Value.mvbin.__ptr[i].__ptr = sPropVal.Value.bin->__ptr;
						lpPropVal->Value.mvbin.__ptr[i].__size = sPropVal.Value.bin->__size;
					}
				} else {
					lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
					lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
					lpPropVal->__union = SOAP_UNION_propValData_ul;
				}
				break;
			}
			case 0x800F101E: { //PR_EMS_AB_PROXY_ADDRESSES:
				/* Make BlackBerry happy */
				std::string address = (std::string)"SMTP:" + lpDetails->GetPropString(OB_PROP_S_EMAIL);

				lpPropVal->__union = SOAP_UNION_propValData_mvszA;
				lpPropVal->Value.mvszA.__size = 1;
				lpPropVal->Value.mvszA.__ptr = s_alloc<char *>(soap, 1);
				lpPropVal->Value.mvszA.__ptr[0] = s_strcpy(soap, address.c_str());
				break;
			}
			case PR_EC_EXCHANGE_DN: {
				std::string exchangeDN = lpDetails->GetPropString(OB_PROP_S_EXCH_DN);
				if (!exchangeDN.empty()) {
					lpPropVal->Value.lpszA = s_strcpy(soap, exchangeDN.c_str());
					lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				} else {
					lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
					lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
					lpPropVal->__union = SOAP_UNION_propValData_ul;
				}
				break;
			}
			case PR_EC_HOMESERVER_NAME: {
				std::string serverName = lpDetails->GetPropString(OB_PROP_S_SERVERNAME);
				if (!serverName.empty()) {
					lpPropVal->Value.lpszA = s_strcpy(soap, serverName.c_str());
					lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				} else {
					lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
					lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
					lpPropVal->__union = SOAP_UNION_propValData_ul;
				}
				break;
			}
			case PR_OFFICE_TELEPHONE_NUMBER:
			    lpPropVal->ulPropTag = PR_OFFICE_TELEPHONE_NUMBER;
			    lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_TELEPHONE).c_str());
			    lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			    break;
			case PR_DEPARTMENT_NAME:
			    lpPropVal->ulPropTag = PR_DEPARTMENT_NAME;
			    lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_DEPARTMENT).c_str());
			    lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			    break;
			case PR_OFFICE_LOCATION:
			    lpPropVal->ulPropTag = PR_OFFICE_LOCATION;
			    lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_LOCATION).c_str());
			    lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			    break;
            case PR_PRIMARY_FAX_NUMBER:
                lpPropVal->ulPropTag = PR_PRIMARY_FAX_NUMBER;
                lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_FAX).c_str());
                lpPropVal->__union = SOAP_UNION_propValData_lpszA;
                break;
			default:
				lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
				lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			}
			break;
		} // end case USEROBJECT_TYPE_USER_*

		case USEROBJECT_TYPE_COMPANY:
		case USEROBJECT_TYPE_GROUP: {
			switch(lpPropTags->__ptr[i]) {
			case PR_COMPANY_NAME: {
				string strCompany = lpDetails->GetPropString(OB_PROP_S_COMPANYNAME);
				lpPropVal->Value.lpszA = s_strcpy(soap, strCompany.c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			}
			case PR_SEARCH_KEY: {
				std::string strSearchKey = (std::string)"ZARAFA:" + strToUpper(lpDetails->GetPropString(OB_PROP_S_FULLNAME));
				lpPropVal->Value.bin =  s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, strSearchKey.size()+1);
				lpPropVal->Value.bin->__size = strSearchKey.size()+1;
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				strcpy((char *)lpPropVal->Value.bin->__ptr, strSearchKey.c_str());
				break;
			}
			case PR_ADDRTYPE:
				lpPropVal->Value.lpszA = s_strcpy(soap, "ZARAFA");
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_EMAIL_ADDRESS:
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_LOGIN).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_MESSAGE_CLASS:
				lpPropVal->Value.lpszA = s_strcpy(soap, "IPM.DistList");
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_ENTRYID: {
				er = CreateABEntryID(soap, ulId, MAPI_DISTLIST, lpPropVal);
				if (er != erSuccess)
					goto exit;
                break;
			}
			case PR_NORMALIZED_SUBJECT:
			case PR_DISPLAY_NAME:
			case 0x3A20001E:// PR_TRANSMITABLE_DISPLAY_NAME
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_FULLNAME).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_INSTANCE_KEY:
				lpPropVal->Value.bin = s_alloc<xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, 2 * sizeof(ULONG));
				lpPropVal->Value.bin->__size = 2*sizeof(ULONG);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
				memcpy(lpPropVal->Value.bin->__ptr+sizeof(ULONG), &ulOrder, sizeof(ULONG));
				break;
			case PR_OBJECT_TYPE:
				lpPropVal->Value.ul = MAPI_DISTLIST;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_DISPLAY_TYPE:
				lpPropVal->Value.ul = DT_DISTLIST;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_RECORD_KEY:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ULONG));
				lpPropVal->Value.bin->__size = sizeof(ULONG);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
				break;
			case PR_ACCOUNT:
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_LOGIN).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_AB_PROVIDER_ID:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(GUID));
				lpPropVal->Value.bin->__size = sizeof(GUID);

				lpPropVal->__union = SOAP_UNION_propValData_bin;
				memcpy(lpPropVal->Value.bin->__ptr, MUIDECSAB_SERVER, sizeof(GUID));
				break;
			case PR_EC_HOMESERVER_NAME: {
				std::string serverName = lpDetails->GetPropString(OB_PROP_S_SERVERNAME);
				if (!serverName.empty()) {
					lpPropVal->Value.lpszA = s_strcpy(soap, serverName.c_str());
					lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				} else {
					lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
					lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
					lpPropVal->__union = SOAP_UNION_propValData_ul;
				}
				break;
			}
			default:
				lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
				lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			}
			break;
		} // end case USEROBJECT_TYPE_GROUP

		} // end switch(type)
		
		// Output unicode if requested
		if(PROP_TYPE(lpPropVal->ulPropTag) == PT_STRING8 && PROP_TYPE(lpPropTags->__ptr[i]) == PT_UNICODE) {
		    ECIConv iconv("utf-8", "windows-1252");
		    std::string utf8 = iconv.convert(lpPropVal->Value.lpszA);
		    
		    lpPropVal->Value.lpszA = s_strcpy(soap, utf8.c_str());
		    lpPropVal->ulPropTag = CHANGE_PROP_TYPE(lpPropVal->ulPropTag, PT_UNICODE);
		}
	}

exit:
	return er;
}

// Convert a userdetails_t to a set of MAPI properties
ECRESULT ECUserManagement::ConvertContainerObjectDetailsToProps(struct soap *soap, unsigned int ulId, userobject_type_t type, objectdetails_t *lpDetails, struct propTagArray *lpPropTags, struct propValArray *lpPropVals)
{
	ECRESULT er = erSuccess;
	int i = 0;
	struct propVal *lpPropVal;
	unsigned int ulOrder = 0;
	ECSession *lpecSession = NULL;
	std::string strEncExId;

	lpecSession = dynamic_cast<ECSession*>(m_lpSession);
	if (!lpecSession) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulId);
	if (er != erSuccess)
		goto exit;

	lpPropVals->__ptr = s_alloc<struct propVal>(soap, lpPropTags->__size);
	lpPropVals->__size = lpPropTags->__size;

	for(i=0;i<lpPropTags->__size;i++) {
		lpPropVal = &lpPropVals->__ptr[i];
		lpPropVal->ulPropTag = lpPropTags->__ptr[i];

		switch(type) {
		default:
			er = ZARAFA_E_NOT_FOUND;
			break;

		case USEROBJECT_TYPE_ADDRESSLIST: {
			switch(lpPropTags->__ptr[i]) {
			case PR_ENTRYID: {
				er = CreateABEntryID(soap, ulId, MAPI_ABCONT, lpPropVal);
				if (er != erSuccess)
					goto exit;
                break;
			}
			case PR_PARENT_ENTRYID: {
				ABEID abeid;
				abeid.ulType = MAPI_ABCONT;
				abeid.ulId = 1;
				memcpy(&abeid.guid, &MUIDECSAB_SERVER, sizeof(GUID));

				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ABEID));
				lpPropVal->Value.bin->__size = sizeof(ABEID);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				*(ABEID *)lpPropVal->Value.bin->__ptr = abeid;
				break;
			}
			case PR_NORMALIZED_SUBJECT:
			case PR_DISPLAY_NAME:
			case 0x3A20001E:// PR_TRANSMITABLE_DISPLAY_NAME
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_FULLNAME).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_INSTANCE_KEY:
				lpPropVal->Value.bin = s_alloc<xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, 2 * sizeof(ULONG));
				lpPropVal->Value.bin->__size = 2*sizeof(ULONG);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
				memcpy(lpPropVal->Value.bin->__ptr+sizeof(ULONG), &ulOrder, sizeof(ULONG));
				break;
			case PR_OBJECT_TYPE:
				lpPropVal->Value.ul = MAPI_ABCONT;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_CONTAINER_FLAGS:
				lpPropVal->Value.ul = AB_RECIPIENTS | AB_UNMODIFIABLE;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_DISPLAY_TYPE:
				lpPropVal->Value.ul = DT_FOLDER;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_RECORD_KEY:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ULONG));
				lpPropVal->Value.bin->__size = sizeof(ULONG);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
				break;
			case PR_AB_PROVIDER_ID:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(GUID));
				lpPropVal->Value.bin->__size = sizeof(GUID);

				lpPropVal->__union = SOAP_UNION_propValData_bin;
				memcpy(lpPropVal->Value.bin->__ptr, MUIDECSAB_SERVER, sizeof(GUID));
				break;
			default:
				lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
				lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			}
			break;
		} // end case USEROBJECT_TYPE_ADDRESSLIST
		
		case USEROBJECT_TYPE_COMPANY: {
			switch (lpPropTags->__ptr[i]) {
			case PR_CONTAINER_CLASS:
				lpPropVal->Value.lpszA = s_strcpy(soap, "IPM.Contact");
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_ENTRYID: {
				er = CreateABEntryID(soap, ulId, MAPI_ABCONT, lpPropVal);
				if (er != erSuccess)
					goto exit;
                break;
			}
			case PR_EMS_AB_PARENT_ENTRYID:
			case PR_PARENT_ENTRYID: {
				ABEID abeid;
				abeid.ulType = MAPI_ABCONT;
				abeid.ulId = 1;
				memcpy(&abeid.guid, &MUIDECSAB_SERVER, sizeof(GUID));

				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ABEID));
				lpPropVal->Value.bin->__size = sizeof(ABEID);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				*(ABEID *)lpPropVal->Value.bin->__ptr = abeid;
				break;
			}
			case PR_ACCOUNT:
			case PR_NORMALIZED_SUBJECT:
			case PR_DISPLAY_NAME:
			case 0x3A20001E:// PR_TRANSMITABLE_DISPLAY_NAME
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_FULLNAME).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_COMPANY_NAME:
				lpPropVal->Value.lpszA = s_strcpy(soap, lpDetails->GetPropString(OB_PROP_S_COMPANYNAME).c_str());
				lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				break;
			case PR_INSTANCE_KEY:
				lpPropVal->Value.bin = s_alloc<xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, 2 * sizeof(ULONG));
				lpPropVal->Value.bin->__size = 2 * sizeof(ULONG);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
				memcpy(lpPropVal->Value.bin->__ptr + sizeof(ULONG), &ulOrder, sizeof(ULONG));
				break;
			case PR_OBJECT_TYPE:
				lpPropVal->Value.ul = MAPI_ABCONT;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_DISPLAY_TYPE:
				lpPropVal->Value.ul = DT_ORGANIZATION;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_RECORD_KEY:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ULONG));
				lpPropVal->Value.bin->__size = sizeof(ULONG);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
				break;
			case PR_CONTAINER_FLAGS:
				lpPropVal->Value.ul = AB_RECIPIENTS | AB_UNMODIFIABLE;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_AB_PROVIDER_ID:
				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(GUID));
				lpPropVal->Value.bin->__size = sizeof(GUID);

				lpPropVal->__union = SOAP_UNION_propValData_bin;
				memcpy(lpPropVal->Value.bin->__ptr, MUIDECSAB_SERVER, sizeof(GUID));
				break;
			case PR_EMS_AB_IS_MASTER:
				lpPropVal->Value.b = false;
				lpPropVal->__union = SOAP_UNION_propValData_b;
				break;
			case PR_EMS_AB_CONTAINERID:
				lpPropVal->Value.ul = ulId;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			case PR_EC_HOMESERVER_NAME: {
				std::string serverName = lpDetails->GetPropString(OB_PROP_S_SERVERNAME);
				if (!serverName.empty()) {
					lpPropVal->Value.lpszA = s_strcpy(soap, serverName.c_str());
					lpPropVal->__union = SOAP_UNION_propValData_lpszA;
				} else {
					lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
					lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
					lpPropVal->__union = SOAP_UNION_propValData_ul;
				}
				break;
			}
			default:
				lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTags->__ptr[i]));
				lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
				break;
			}
		} // end USEROBJECT_TYPE_COMPANY
		} // end switch(type)
	}

exit:
	return er;
}

ECRESULT ECUserManagement::ConvertABContainerToProps(struct soap *soap, unsigned int ulId, struct propTagArray *lpPropTagArray, struct propValArray *lpPropValArray)
{
	ECRESULT er = erSuccess;
	int i = 0;
	struct propVal *lpPropVal;
	std::string strName;
	ABEID abeid;
	char MUIDEMSAB[] = "\xDC\xA7\x40\xC8\xC0\x42\x10\x1A\xB4\xB9\x08\x00\x2B\x2F\xE1\x82";

	lpPropValArray->__ptr = s_alloc<struct propVal>(soap, lpPropTagArray->__size);
	lpPropValArray->__size = lpPropTagArray->__size;

	abeid.ulType = MAPI_ABCONT;
	memcpy(&abeid.guid, &MUIDECSAB_SERVER, sizeof(GUID));
	abeid.ulId = ulId;

	// FIXME: Should this name be hardcoded like this?
	// Are there any other values that might be passed as name?
	if (ulId == 0)
		strName = ZARAFA_ADDRESS_BOOK;
	else if (ulId == 1)
		strName = ZARAFA_GLOBAL_ADDRESS_BOOK;
	else {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	for (i = 0; i < lpPropTagArray->__size; i++) {
		lpPropVal = &lpPropValArray->__ptr[i];
		lpPropVal->ulPropTag = lpPropTagArray->__ptr[i];

		switch (lpPropTagArray->__ptr[i]) {
		case PR_SEARCH_KEY:
			lpPropVal->Value.bin =  s_alloc<struct xsd__base64Binary>(soap);
			lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ABEID));
			lpPropVal->Value.bin->__size = sizeof(ABEID);
			lpPropVal->__union = SOAP_UNION_propValData_bin;
			memcpy(lpPropVal->Value.bin->__ptr, &abeid, sizeof(ABEID));
			break;
		case PR_CONTAINER_CLASS:
			lpPropVal->Value.lpszA = s_strcpy(soap, "IPM.Contact");
			lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			break;
		case PR_ENTRYID:
			lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
			lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ABEID));
			lpPropVal->Value.bin->__size = sizeof(ABEID);
			lpPropVal->__union = SOAP_UNION_propValData_bin;
			memcpy(lpPropVal->Value.bin->__ptr, &abeid, sizeof(ABEID));
			break;
		case PR_ACCOUNT:
		case PR_NORMALIZED_SUBJECT:
		case PR_DISPLAY_NAME:
		case 0x3A20001E:// PR_TRANSMITABLE_DISPLAY_NAME
			lpPropVal->Value.lpszA = s_strcpy(soap, strName.c_str());
			lpPropVal->__union = SOAP_UNION_propValData_lpszA;
			break;
		case PR_INSTANCE_KEY:
			lpPropVal->Value.bin = s_alloc<xsd__base64Binary>(soap);
			lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, 2 * sizeof(ULONG));
			lpPropVal->Value.bin->__size = 2 * sizeof(ULONG);
			lpPropVal->__union = SOAP_UNION_propValData_bin;

			memcpy(lpPropVal->Value.bin->__ptr, &ulId, sizeof(ULONG));
			memcpy(lpPropVal->Value.bin->__ptr + sizeof(ULONG), &ulId, sizeof(ULONG));
			break;
		case PR_OBJECT_TYPE:
			lpPropVal->Value.ul = MAPI_ABCONT;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			break;
		case PR_DISPLAY_TYPE:
			/*
			 * Outlook 2000 requires DT_GLOBAL
			 * Outlook 2003 requires DT_GLOBAL
			 * Outlook 2007 requires something which is not DT_GLOBAL,
			 * allowed values is DT_CONTAINER (0x00000100), DT_LOCAL or
			 * DT_NOT_SPECIFIC
			 *
			 * Only Outlook 2007 really complains when it gets a value
			 * which is different then what it epxects.
			 */
			if (ulId == 0)
				lpPropVal->Value.ul = DT_GLOBAL;
			else
				lpPropVal->Value.ul = DT_NOT_SPECIFIC;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			break;
		case PR_RECORD_KEY:
			lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
			lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ABEID));
			lpPropVal->Value.bin->__size = sizeof(ABEID);
			lpPropVal->__union = SOAP_UNION_propValData_bin;
			memcpy(lpPropVal->Value.bin->__ptr, &abeid, sizeof(ABEID));
			break;
		case PR_CONTAINER_FLAGS:
			/*
			 * We support subcontainers, but don't even consider setting the
			 * flag AB_SUBCONTAINERS for the Global Address Book since that
			 * will cause Outlook 2007 to bug() which then doesn't place
			 * the Global Address Book in the SearchPath.
			 */
			lpPropVal->Value.ul = AB_RECIPIENTS | AB_UNMODIFIABLE;
			if (ulId == 0)
				lpPropVal->Value.ul |= AB_SUBCONTAINERS;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			break;
		case PR_AB_PROVIDER_ID: {
		    std::string strApp;
		    ECSession *lpSession = NULL;
		    
			lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
			lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(GUID));
			lpPropVal->Value.bin->__size = sizeof(GUID);

			lpPropVal->__union = SOAP_UNION_propValData_bin;
		
			lpSession = dynamic_cast<ECSession *>(m_lpSession);
			if(lpSession)
    			lpSession->GetClientApp(&strApp);
			
			if(stricmp(strApp.c_str(), "blackberryagent.exe") == 0) {
			    // For blackberry, we pose as being the Exchange AddressList. We have to do this
			    // since it searches for the GAB by restricting by this GUID, otherwise the Lookup
			    // function will not function properly.
    			memcpy(lpPropVal->Value.bin->__ptr, MUIDEMSAB, sizeof(GUID));
            } else {
    			memcpy(lpPropVal->Value.bin->__ptr, MUIDECSAB_SERVER, sizeof(GUID));
            }
			break;
        }
		case PR_EMS_AB_IS_MASTER:
			lpPropVal->Value.b = (ulId == 0);
			lpPropVal->__union = SOAP_UNION_propValData_b;
			break;
		case PR_EMS_AB_CONTAINERID:
		    // 'Global Address Book' should be container ID 0, rest follows
		    if(ulId > 0) {
       			lpPropVal->Value.ul = ulId == 1 ? 0 : ulId;
        		lpPropVal->__union = SOAP_UNION_propValData_ul;
            } else {
                // id 0 (zarafa address book) has no container id
                lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTagArray->__ptr[i]));
                lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
                lpPropVal->__union = SOAP_UNION_propValData_ul;
            }
			break;
		case PR_EMS_AB_PARENT_ENTRYID:
		case PR_PARENT_ENTRYID:
			if (ulId == 1) { /* Global Address Book */
				ABEID abeid;
				abeid.ulType = MAPI_ABCONT;
				abeid.ulId = 0;
				memcpy(&abeid.guid, &MUIDECSAB_SERVER, sizeof(GUID));

				lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
				lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, sizeof(ABEID));
				lpPropVal->Value.bin->__size = sizeof(ABEID);
				lpPropVal->__union = SOAP_UNION_propValData_bin;

				*(ABEID *)lpPropVal->Value.bin->__ptr = abeid;
			} else { /* Zarafa Address Book */
				lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTagArray->__ptr[i]));
				lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
				lpPropVal->__union = SOAP_UNION_propValData_ul;
			}
			break;
		default:
			lpPropVal->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropTagArray->__ptr[i]));
			lpPropVal->Value.ul = ZARAFA_E_NOT_FOUND;
			lpPropVal->__union = SOAP_UNION_propValData_ul;
			break;
		}
	}

exit:
	return er;
}

ECRESULT ECUserManagement::GetUserCount(unsigned int *lpulActive, unsigned int *lpulNonActive) 
{
	ECRESULT er;
	
	if (m_lpSession->GetSessionManager()->IsDistributedSupported())
		er = GetExternUserCount(lpulActive, lpulNonActive);
	else
		er = GetLocalUserCount(lpulActive, lpulNonActive);
	
	if (er == erSuccess)
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Users Active: %u, NonActive: %u", *lpulActive, *lpulNonActive);
		
	return er;
}

ECRESULT ECUserManagement::GetLocalUserCount(unsigned int *lpulActive, unsigned int *lpulNonActive)
{
    ECRESULT er = erSuccess;
    ECDatabase *lpDatabase = NULL;
    DB_RESULT lpResult = NULL;
    DB_ROW lpRow = NULL;
    std::string strQuery;
    unsigned int ulActive = 0;
    unsigned int ulNonActive = 0;

    lpDatabase = m_lpSession->GetDatabase();
    if(lpDatabase == NULL) {
        er = ZARAFA_E_DATABASE_ERROR;
        goto exit;
    }

	strQuery =
		"SELECT COUNT(*), object_type "
		"FROM users "
		"WHERE externid IS NOT NULL " // Keep local entries outside of COUNT()
			"AND (object_type = " + stringify(USEROBJECT_TYPE_USER) + " "
				 "OR object_type = " + stringify(USEROBJECT_TYPE_NONACTIVE) + ") "
		"GROUP BY object_type";
    er = lpDatabase->DoSelect(strQuery, &lpResult);
    if(er != erSuccess)
        goto exit;

	while((lpRow = lpDatabase->FetchRow(lpResult)) != NULL) {
		if(lpRow[0] == NULL || lpRow[1] == NULL)
			continue;

		switch (atoi(lpRow[1])) {
		case USEROBJECT_TYPE_USER:
			ulActive = atoi(lpRow[0]);
			break;
		case USEROBJECT_TYPE_NONACTIVE:
			ulNonActive = atoi(lpRow[0]);
			break;
		}
	}

    if(lpulActive)
        *lpulActive = ulActive;

    if(lpulNonActive)
        *lpulNonActive = ulNonActive;

exit:
    if(lpResult)
        lpDatabase->FreeResult(lpResult);

    return er;
}

ECRESULT ECUserManagement::GetExternUserCount(unsigned int *lpulActive, unsigned int *lpulNonActive)
{
	ECRESULT er = erSuccess;
	auto_ptr<objectdetails_t> details;
    UserPlugin *lpPlugin = NULL;

	er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    try {
        lpPlugin->getUserCount(lpulActive, lpulNonActive);
    } catch (objectnotfound &) {
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
	} catch (notsupported &) {
		er = GetLocalUserCount(lpulActive, lpulNonActive);
    } catch (std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to get user count: %s", e.what());
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    }

exit:
	return er;
}

ECRESULT ECUserManagement::GetPublicStoreDetails(objectdetails_t *lpDetails)
{
	ECRESULT er = erSuccess;
	auto_ptr<objectdetails_t> details;
    UserPlugin *lpPlugin = NULL;

	/* We pretend that the Public store is a company. So request (and later store) it as such. */
	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetUserDetails(ZARAFA_UID_EVERYONE, USEROBJECT_TYPE_COMPANY, lpDetails);
	if (er == erSuccess)
		goto exit; /* Cache contained requested information, we're done.*/

    er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    try {
        details = lpPlugin->getPublicStoreDetails();
    } catch (objectnotfound &) {
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
	} catch (notsupported &) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
    } catch (std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to get %s details for object id %d: %s", ObjectTypeToName(USEROBJECT_TYPE_COMPANY), ZARAFA_UID_EVERYONE, e.what());
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    }

	/* Update cache so we don't have to bug the plugin until the data has changed.
	 * Note that we don't care if the update succeeded, if it fails we will retry
	 * when the user details are requested for a second time. */
	m_lpSession->GetSessionManager()->GetCacheManager()->SetUserDetails(ZARAFA_UID_EVERYONE, USEROBJECT_TYPE_COMPANY, details.get());

    *lpDetails = *details;

exit:
    return er;
}

ECRESULT ECUserManagement::GetServerDetails(const std::string &strServer, serverdetails_t *lpDetails)
{
	ECRESULT er = erSuccess;
	auto_ptr<serverdetails_t> details;
    UserPlugin *lpPlugin = NULL;

	// Try the cache first
	er = m_lpSession->GetSessionManager()->GetCacheManager()->GetServerDetails(strServer, lpDetails);
	if (er == erSuccess)
		goto exit;


	er = GetThreadLocalPlugin(m_lpPluginFactory, &lpPlugin, m_lpLogger);
    if(er != erSuccess)
        goto exit;

    try {
        details = lpPlugin->getServerDetails(strServer);
		m_lpSession->GetSessionManager()->GetCacheManager()->SetServerDetails(strServer, *details);
    } catch (objectnotfound &) {
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
	} catch (notsupported &) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
    } catch (std::exception &e) {
        m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to get server details for %s: %s", strServer.c_str(), e.what());
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    }

    *lpDetails = *details;

exit:
	return er;
}

ECRESULT ECUserManagement::ProcessModification(unsigned int ulId, userobject_type_t /*ulType*/, const std::string &newsignature)
{
    ECRESULT er = erSuccess;
    std::string strQuery;
    ECDatabase *lpDatabase = NULL;
	ABEID eid(MAPI_ABCONT, *(GUID*)&MUIDECSAB_SERVER, 1);
	SOURCEKEY sSourceKey;

    // Log the change to ICS
    er = GetABSourceKeyV1(ulId, &sSourceKey);
    if (er != erSuccess)
    	goto exit;

	AddABChange(m_lpSession, ICS_AB_CHANGE, sSourceKey, SOURCEKEY(CbABEID(&eid), (char *)&eid));

    // Ignore ICS error

    // Save the new signature
    lpDatabase = m_lpSession->GetDatabase();

    er = lpDatabase->DoUpdate("UPDATE users SET signature=" + lpDatabase->EscapeBinary((unsigned char *)newsignature.c_str(), newsignature.size()) + " WHERE id=" + stringify(ulId));
    if(er != erSuccess)
        goto exit;

	er = m_lpSession->GetSessionManager()->GetCacheManager()->UpdateUser(ulId);
	if (er != erSuccess)
		goto exit;

exit:
    return er;
}

ECRESULT ECUserManagement::GetABSourceKeyV1(unsigned int ulUserId, SOURCEKEY *lpsSourceKey)
{
	ECRESULT 			er = erSuccess;
	std::string			strExternId;
	std::string			strEncExId;
	PABEID				lpAbeid = NULL;
	unsigned int		ulLen = 0;
	userobject_type_t	sType;
	unsigned int		ulType = 0;

	er = GetExternalId(ulUserId, &sType, &strExternId);
	if (er != erSuccess)
		goto exit;

	strEncExId = base64_encode((unsigned char*)strExternId.c_str(), strExternId.size());

	er = TypeToMAPIType(sType, &ulType);
	if (er != erSuccess)
		goto exit;

	ulLen = CbNewABEID(strEncExId.c_str());
	lpAbeid = (PABEID)new char[ulLen];
	memset(lpAbeid, 0, ulLen);
	lpAbeid->ulId = ulUserId;
	lpAbeid->ulType = ulType;
	memcpy(&lpAbeid->guid, MUIDECSAB_SERVER, sizeof(GUID));
	if (!strEncExId.empty())
	{
		lpAbeid->ulVersion = 1;
		// avoid FORTIFY_SOURCE checks in strcpy to an address that the compiler thinks is 1 size large
		memcpy(lpAbeid->szExId, strEncExId.c_str(), strEncExId.length()+1);
	}

	*lpsSourceKey = SOURCEKEY(ulLen, (char*)lpAbeid);

exit:
	if (lpAbeid)
		delete[] lpAbeid;

	return er;
}

ECRESULT ECUserManagement::CreateABEntryID(struct soap *soap, unsigned int ulObjId, unsigned int ulType, struct propVal *lpPropVal)
{
	ECRESULT	er = erSuccess;
	PABEID		lpEid = NULL;
	ABEID		localEid;
	std::string strExternId;
	std::string	strEncExId;

	ASSERT(soap != NULL);
	ASSERT(lpPropVal != NULL);

	if (IsLocalItem(ulObjId))
		lpEid = &localEid;

	else {

		er = GetExternalId(ulObjId, NULL, &strExternId);
		if (er != erSuccess)
			goto exit;

		strEncExId = base64_encode((unsigned char*)strExternId.c_str(), strExternId.size());

		lpEid = (PABEID)new char[CbNewABEID(strEncExId.c_str())];
		memset(lpEid, 0, CbNewABEID(strEncExId.c_str()));

		lpEid->ulVersion = 1;
		// avoid FORTIFY_SOURCE checks in strcpy to an address that the compiler thinks is 1 size large
		memcpy(lpEid->szExId, strEncExId.c_str(), strEncExId.length()+1);

	}

	lpEid->ulType = ulType;
	lpEid->ulId = ulObjId;
	memcpy(&lpEid->guid,MUIDECSAB_SERVER,sizeof(lpEid->guid));

	lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
	lpPropVal->Value.bin->__ptr = s_alloc<unsigned char>(soap, CbABEID(lpEid));
	lpPropVal->Value.bin->__size = CbABEID(lpEid);
	lpPropVal->__union = SOAP_UNION_propValData_bin;

	memcpy(lpPropVal->Value.bin->__ptr, lpEid, CbABEID(lpEid));

exit:
	if (lpEid != &localEid)
		delete[] (char*)lpEid;

	return er;
}
