/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "DBBase.h"
#include "ECDefs.h"
#include "EMSAbTag.h"
#include "stringutil.h"
#include "md5.h"

ZARAFA_API ECRESULT GetDatabaseObject(ECDatabase **lppDatabase);

DBPlugin::DBPlugin(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed) :
	UserPlugin(pluginlock, config, logger, lpStatsCollector, bHosted, bDistributed), m_lpDatabase(NULL) {
}

DBPlugin::~DBPlugin() {
    // Do not delete m_lpDatabase as it is freed when the thread exits
}

void DBPlugin::InitPlugin() throw(std::exception) {

	if(GetDatabaseObject(&m_lpDatabase) != erSuccess)
	    throw runtime_error(string("db_init: cannot get handle to database"));
}

list<objectsignature_t> DBPlugin::getAllObjects(userobject_type_t type, const objectid_t &company) throw(std::exception)
{
	list<objectsignature_t> lObjects;
	ECRESULT er = erSuccess;
	string strQuery;
	string id;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;
	DB_LENGTHS lpDBLen = NULL;

	strQuery =
		"SELECT om.externid, op.value "
		"FROM " + (string)DB_OBJECT_TABLE + " AS om "
		"LEFT JOIN " + (string)DB_OBJECTPROPERTY_TABLE " AS op "
			"ON op.objectid = om.id "
			"AND op.propname = '" + OP_MODTIME + "' ";
	if (m_bHosted && type != USEROBJECT_TYPE_COMPANY && !company.empty()) {
		strQuery +=
			"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS opc "
				"ON opc.objectid = om.id "
				"AND opc.propname = '" + OP_COMPANYID + "' "
			"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS oc "
				"ON opc.value = oc.value "
				"AND oc.propname = '" + OP_COMPANYNAME + "' "
			"JOIN " + (string)DB_OBJECT_TABLE + " AS oceid "
				" ON oceid.id = oc.objectid ";
	}
	strQuery += "WHERE om.objecttype = " + stringify(type);
	if (m_bHosted && type != USEROBJECT_TYPE_COMPANY && !company.empty())
		 strQuery += " AND oceid.externid = '" + m_lpDatabase->Escape(company) + "'";

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	while ((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if(lpDBRow[0] == NULL)
		    continue;

		string signature;
		if (lpDBRow[1] != NULL)
			signature = lpDBRow[1];

		lpDBLen = m_lpDatabase->FetchRowLengths(lpResult);
		ASSERT(lpDBLen != NULL);
		if (lpDBLen[0] == 0)
			throw runtime_error(string("db_row_failed: object empty"));

		id.assign(lpDBRow[0], lpDBLen[0]);
		lObjects.push_back(objectsignature_t(id, signature));
	}
	return lObjects;
}

auto_ptr<objectdetails_t> DBPlugin::getObjectDetails(userobject_type_t type, const objectid_t &objectid) throw(std::exception)
{
	ECRESULT er = erSuccess;
	auto_ptr<objectdetails_t> lpDetails = auto_ptr<objectdetails_t>(new objectdetails_t(type));
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;
	string strQuery;

	strQuery =
		"SELECT op.propname, op.value "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"RIGHT JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS op "
			"ON op.objectid = o.id "
		"WHERE o.objecttype = " + stringify(type) + " "
			"AND o.externid = '" +  m_lpDatabase->Escape(objectid) + "'";

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	if (m_lpDatabase->GetNumRows(lpResult) == 0)
		throw objectnotfound(objectid);

	/* By default the sysadmin is SYSTEM, will be overwritten later */
	lpDetails->SetPropString(OB_PROP_S_SYSADMIN, "SYSTEM");

	if(type == USEROBJECT_TYPE_NONACTIVE)
		lpDetails->SetPropBool(OB_PROP_B_NONACTIVE, true);
	else if (type == USEROBJECT_TYPE_USER)
		lpDetails->SetPropBool(OB_PROP_B_NONACTIVE, false);

	while((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if(lpDBRow[0] == NULL || lpDBRow[1] == NULL)
			continue;

		if (strcmp(lpDBRow[0], OP_LOGINNAME) == 0)
			lpDetails->SetPropString(OB_PROP_S_LOGIN, lpDBRow[1]);
		else if (strcmp(lpDBRow[0], OP_FULLNAME) == 0)
			lpDetails->SetPropString(OB_PROP_S_FULLNAME, lpDBRow[1]);
		else if (strcmp(lpDBRow[0], OP_EMAILADDRESS) == 0)
			lpDetails->SetPropString(OB_PROP_S_EMAIL, lpDBRow[1]);
		else if(strcmp(lpDBRow[0], OP_ISADMIN) == 0)
			lpDetails->SetPropInt(OB_PROP_I_ADMINLEVEL, atoi(lpDBRow[1]));
		else if(strcmp(lpDBRow[0], OP_GROUPNAME) == 0) {
			lpDetails->SetPropString(OB_PROP_S_LOGIN, lpDBRow[1]);
			lpDetails->SetPropString(OB_PROP_S_FULLNAME, lpDBRow[1]);
		} else if (strcmp(lpDBRow[0], OP_COMPANYNAME) == 0) {
			lpDetails->SetPropString(OB_PROP_S_LOGIN, lpDBRow[1]);
			lpDetails->SetPropString(OB_PROP_S_FULLNAME, lpDBRow[1]);
		} else if (strcmp(lpDBRow[0], OP_COMPANYID) == 0)
			lpDetails->SetPropString(OB_PROP_S_COMPANYNAME, lpDBRow[1]);
		else if (strcmp(lpDBRow[0], OP_COMPANYADMIN) == 0)
			lpDetails->SetPropString(OB_PROP_S_SYSADMIN, lpDBRow[1]);
	}

	return lpDetails;
}

list<objectsignature_t> DBPlugin::getSubObjectsForObject(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype) throw(std::exception)
{
	list<objectsignature_t> lObjects;
	ECRESULT er = erSuccess;
	string strQuery;
	string signature;
	string id;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;
	DB_LENGTHS lpDBLen = NULL;

	strQuery =
		"SELECT o.externid, modtime.value "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"JOIN " + (string)DB_OBJECT_RELATION_TABLE + " AS ort "
			"ON o.id = ort.objectid "
		"JOIN " + (string)DB_OBJECT_TABLE + " AS p "
			"ON p.id = ort.parentobjectid "
		"LEFT JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS modtime "
			"ON modtime.objectid=o.id "
			"AND modtime.propname = '" + OP_MODTIME + "' "
		"WHERE p.externid = '" +  m_lpDatabase->Escape(parentobject) + "' "
		    "AND p.objecttype = " + stringify(parenttype) + " "
			"AND ort.relationtype = " + stringify(relation) + " "
			"AND o.objecttype = " + stringify(childtype);

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	while ((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if(lpDBRow[0] == NULL)
			throw runtime_error(string("db_row_failed: object null"));

		if(lpDBRow[1] != NULL)
			signature = lpDBRow[1];
		else
			signature.clear();

		lpDBLen = m_lpDatabase->FetchRowLengths(lpResult);
		ASSERT(lpDBLen != NULL);
		if (lpDBLen[0] == 0)
			throw runtime_error(string("db_row_failed: object empty"));

		id.assign(lpDBRow[0], lpDBLen[0]);
		lObjects.push_back(objectsignature_t(id, signature));
	}

	return lObjects;
}

list<objectsignature_t> DBPlugin::getParentObjectsForObject(userobject_relation_t relation, userobject_type_t childtype, const objectid_t &childobject, userobject_type_t parenttype) throw(std::exception)
{
	list<objectsignature_t> lGroups;
	ECRESULT er;
	string strQuery;
	string signature;
	string id;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;
	DB_LENGTHS lpDBLen = NULL;

	strQuery =
		"SELECT o.externid, modtime.value "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"JOIN " + (string)DB_OBJECT_RELATION_TABLE + " AS ort "
			"ON o.id = ort.parentobjectid "
		"JOIN " + (string)DB_OBJECT_TABLE + " AS c "
			"ON ort.objectid = c.id "
		"LEFT JOIN " +(string)DB_OBJECTPROPERTY_TABLE + " AS modtime "
			"ON modtime.objectid = o.id "
			"AND modtime.propname = '" + OP_MODTIME + "' "
		"WHERE c.externid = '" +  m_lpDatabase->Escape(childobject) + "' "
		    "AND c.objecttype = " + stringify(childtype) + " "
			"AND ort.relationtype = " + stringify(relation) + " "
			"AND o.objecttype = " + stringify(parenttype);

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	while ((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if(lpDBRow[0] == NULL)
			throw runtime_error(string("db_row_failed: object null"));

		if(lpDBRow[1] != NULL)
			signature = lpDBRow[1];
		else
			signature.clear();

		lpDBLen = m_lpDatabase->FetchRowLengths(lpResult);
		ASSERT(lpDBLen != NULL);
		if (lpDBLen[0] == 0)
			throw runtime_error(string("db_row_failed: object empty"));

		id.assign(lpDBRow[0], lpDBLen[0]);
		lGroups.push_back(objectsignature_t(id, signature));
	}

	return lGroups;
}

int DBPlugin::changeObject(userobject_type_t type, const objectid_t &objectid, objectdetails_t details, std::list<std::string> *lpDeleteProps) throw(std::exception)
{
	ECRESULT er;
	string strQuery;
	string strSubQuery;
	bool bFirstOne = true;
	string strMD5Pw;
	std::list<std::string>::iterator iterDeleteProps;
	char *sUserValidProps[] = {
		OB_PROP_S_LOGIN, OP_LOGINNAME,
		OB_PROP_S_PASSWORD, OP_PASSWORD,
		OB_PROP_S_EMAIL, OP_EMAILADDRESS,
		OB_PROP_I_ADMINLEVEL, OP_ISADMIN,
		OB_PROP_S_FULLNAME, OP_FULLNAME,
		OB_PROP_S_COMPANYNAME, OP_COMPANYID,
		NULL };
	char *sGroupValidProps[] = {
		OB_PROP_S_FULLNAME, OP_GROUPNAME,
		OB_PROP_S_COMPANYNAME, OP_COMPANYID,
		NULL };
	char *sCompanyValidProps[] = {
		OB_PROP_S_FULLNAME, OP_COMPANYNAME,
		OB_PROP_S_SYSADMIN, OP_COMPANYADMIN,
		NULL };
	char **sValidProps;


	strSubQuery = 
		"SELECT id FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(objectid) + "'" +
		" AND objecttype = " + stringify(type);

	if (lpDeleteProps) {
		// delete properties
		strQuery =
			"DELETE FROM " + (string)DB_OBJECTPROPERTY_TABLE + " "
			"WHERE objectid = (" + strSubQuery + ") " +
			" AND propname IN (";

		for (iterDeleteProps = lpDeleteProps->begin();
			 iterDeleteProps != lpDeleteProps->end();
			 iterDeleteProps++) {
			strQuery += *iterDeleteProps + ", ";
		}

		// Replace trailing comma with ')'
		strQuery[strQuery.size() - 1] = ')';

		er = m_lpDatabase->DoDelete(strQuery);
		if(er != erSuccess)
			throw runtime_error(string("db_query: ") + strerror(er));

	}

	strQuery = "REPLACE INTO " + (string)DB_OBJECTPROPERTY_TABLE + "(objectid, propname, value) VALUES ";

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		sValidProps = sUserValidProps;
		break;
	case USEROBJECT_TYPE_GROUP:
		sValidProps = sGroupValidProps;
		break;
	case USEROBJECT_TYPE_COMPANY:
		sValidProps = sCompanyValidProps;
		break;
	default:
		throw runtime_error("Object is wrong type");
	}

	unsigned int i = 0;
	while (sValidProps[i] != NULL) {
		string propvalue = details.GetPropString(sValidProps[i]);
		string propname = string(sValidProps[i + 1]);

		if (propname == OP_PASSWORD && !propvalue.empty()) {
			// Password value has special treatment
			if (CreateMD5Hash(propvalue, &propvalue) != erSuccess) // WARNING input and output point to the same data
				throw runtime_error(string("db_changeUser: create md5"));
		}

		if (!propvalue.empty()) {
			if (!bFirstOne)
				strQuery += ",";
			strQuery += "((" + strSubQuery + "),'" + m_lpDatabase->Escape(propname) + "','" +  m_lpDatabase->Escape(propvalue) + "')";
			bFirstOne = false;
		}

		i += 2;
	}

	/* Only update when there were actually properties provided. */
	if (!bFirstOne) {
		er = m_lpDatabase->DoInsert(strQuery);
		if (er != erSuccess)
			throw runtime_error(string("db_query: ") + strerror(er));
	}

	// Remember modtime for this object
	strQuery = "REPLACE INTO " +  (string)DB_OBJECTPROPERTY_TABLE + "(objectid, propname, value) VALUES ((" + strSubQuery + "),'" + OP_MODTIME + "',now())";
	er = m_lpDatabase->DoInsert(strQuery);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	return 0;
}

objectsignature_t DBPlugin::createObject(userobject_type_t type, objectdetails_t details) throw(std::exception)
{
	ECRESULT er;
	string strQuery;
	string strObjectId;
	string strExternId;
	string strPropName;
	string strPropValue;
	unsigned int ulObjectId = 0;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	GUID guidExternId = {0};

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		strPropName = OP_LOGINNAME;
		strPropValue = details.GetPropString(OB_PROP_S_LOGIN);
		break;
	case USEROBJECT_TYPE_GROUP:
		strPropName = OP_GROUPNAME;
		strPropValue = details.GetPropString(OB_PROP_S_FULLNAME);
		break;
	case USEROBJECT_TYPE_COMPANY:	
		strPropName = OP_COMPANYNAME;
		strPropValue = details.GetPropString(OB_PROP_S_FULLNAME);
		break;
	default:
		throw runtime_error("Object is wrong type");
	}

	// check if object already exists
	strQuery =
		"SELECT o.id "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS op "
			"ON op.objectid = o.id AND op.propname = '" + strPropName + "' "
		"LEFT JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS oc "
			"ON oc.objectid = o.id AND oc.propname = '" + (string)OP_COMPANYID  +"' ";
	if (type == USEROBJECT_TYPE_USER || type == USEROBJECT_TYPE_NONACTIVE)
		// check all user types
		strQuery += "WHERE (o.objecttype = " + stringify(USEROBJECT_TYPE_USER) + " OR o.objecttype = " + stringify(USEROBJECT_TYPE_NONACTIVE) + ") ";
	else
		strQuery += "WHERE o.objecttype = " + stringify(type) + " ";
	strQuery += "AND op.value = '" + m_lpDatabase->Escape(strPropValue) + "' ";

		if (m_bHosted && type != USEROBJECT_TYPE_COMPANY)
			strQuery += "AND (oc.value IS NULL OR oc.value = '" + m_lpDatabase->Escape(details.GetPropString(OB_PROP_S_COMPANYNAME)) + "')";

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	if (m_lpDatabase->GetNumRows(lpResult) != 0)
		throw runtime_error(string("Object exist: ") + details.GetPropString(OB_PROP_S_LOGIN));

	strExternId = details.GetPropString(OB_PROP_S_EXTERNID);
	if (!strExternId.empty()) {
		// Offline "force" create object
		strQuery =
			"INSERT INTO " + (string)DB_OBJECT_TABLE + "(externid, objecttype) "
			"VALUES('" + m_lpDatabase->Escape(strExternId) + "'," + stringify(type) + ")";

		er = m_lpDatabase->DoInsert(strQuery, &ulObjectId);
		if (er != erSuccess)
			throw runtime_error(string("db_query: ") + strerror(er));

		strObjectId = strExternId;
	} else {
		// zarafa-admin online create object
		if (CoCreateGuid(&guidExternId) != S_OK)
			throw runtime_error("failed to generate extern id");

		strExternId.assign((const char*)&guidExternId, sizeof(guidExternId));

		strQuery =
			"INSERT INTO " + (string)DB_OBJECT_TABLE + "(objecttype) "
			"VALUES(" + stringify(type) + ")";

		er = m_lpDatabase->DoInsert(strQuery, &ulObjectId);
		if (er != erSuccess)
			throw runtime_error(string("db_query: ") + strerror(er));

		strObjectId = stringify(ulObjectId);

		strQuery =
			"UPDATE " + (string)DB_OBJECT_TABLE + " "
			"SET externid = '" + m_lpDatabase->Escape(strExternId) + "' WHERE id = " + strObjectId;

		er = m_lpDatabase->DoUpdate(strQuery);
		if (er != erSuccess)
			throw runtime_error(string("db_query: ") + strerror(er));

		strObjectId = strExternId;
	}

	// Insert all properties into the database
	er = changeObject(type, strObjectId, details, NULL);
	if (er != erSuccess)
		throw runtime_error(string("Object properties could not be saved"));

	// signature is empty on first create. This is OK because it doesn't matter what's in it, as long as it changes when the object is modified
	return objectsignature_t(strObjectId, string());
}

void DBPlugin::deleteObject(userobject_type_t type, const objectid_t &id) throw(std::exception)
{
	ECRESULT er;
	string strQuery;
	string strSubQuery;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;
	unsigned int ulAffRows = 0;

	strSubQuery = 
		"SELECT id FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(id) + "'"
		" AND objecttype = " + stringify(type);

	/* First delete company children */
	if (type == USEROBJECT_TYPE_COMPANY) {
	    strQuery =
			"SELECT op.objectid "
			"FROM " + (string)DB_OBJECTPROPERTY_TABLE + " AS op "
			"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS oc "
				"ON op.propname = '" + OP_COMPANYID + "' "
				"AND oc.propname = '" + OP_COMPANYNAME + "' "
				"AND oc.value = op.value "
			"WHERE oc.objectid = (" + strSubQuery + ")";

		er = m_lpDatabase->DoSelect(strQuery, &lpResult);
		if (er != erSuccess)
			throw runtime_error(string("db_query: ") + strerror(er));

		string children;

		while ((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
			if(lpDBRow[0] == NULL)
				throw runtime_error(string("db_row_failed: object null"));

			if (!children.empty())
				children += ",";
			children += lpDBRow[0];
		}

		if (!children.empty()) {
			strQuery =
				"DELETE FROM " + (string)DB_OBJECTPROPERTY_TABLE + " "
				"WHERE objectid IN (" + children + ")";
			er = m_lpDatabase->DoDelete(strQuery);
			if (er != erSuccess){
				//ignore error
			}

			strQuery =
				"DELETE FROM " + (string)DB_OBJECT_TABLE + " "
				"WHERE id IN (" + children + ")";
			er = m_lpDatabase->DoDelete(strQuery);
			if (er != erSuccess){
				//ignore error
			}
		}
	}

	// first delete details of user, since we need the id from the sub query, which is removed next
	strQuery = "DELETE FROM "+(string)DB_OBJECTPROPERTY_TABLE+" WHERE objectid=("+strSubQuery+")";
	er = m_lpDatabase->DoDelete(strQuery);
	if (er != erSuccess){
		// ignore error
	}

	// delete user from object table .. we now have no reference to the user anymore.
	strQuery =
		"DELETE FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(id) + "' "
		"AND objecttype = " + stringify(type);

	er = m_lpDatabase->DoDelete(strQuery, &ulAffRows);
	if(er != erSuccess){
		//FIXME: ....
	}
	
	if (ulAffRows != 1)
		throw objectnotfound("db_user: "+id);
}

void DBPlugin::addSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception)
{
	ECRESULT er = erSuccess;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	string strQuery;
	string strParentSubQuery;
	string strChildSubQuery;

	if (relation == OBJECTRELATION_USER_SENDAS && parenttype != USEROBJECT_TYPE_USER)
		throw notsupported("only users can send mail");

	strParentSubQuery = 
		"SELECT id FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(parentobject) + "'" 
		" AND objecttype = " + stringify(parenttype);

	strChildSubQuery = 
		"SELECT id FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(childobject) + "'"
		" AND objecttype = " + stringify(childtype);

	/* Check if relation already exists */ 
	strQuery = 
		"SELECT objectid FROM " + (string)DB_OBJECT_RELATION_TABLE + " " 
		"WHERE objectid = (" + strChildSubQuery + ") " 
		"AND parentobjectid = (" + strParentSubQuery + ") " 
		"AND relationtype = " + stringify(relation); 
	er = m_lpDatabase->DoSelect(strQuery, &lpResult); 
	if (er != erSuccess) 
		throw runtime_error(string("db_query: ") + strerror(er)); 

	if (m_lpDatabase->GetNumRows(lpResult) != 0) 
		throw runtime_error(string("Relation exist: ") + stringify(relation)); 

	/* Insert new relation */  
	strQuery =
		"INSERT INTO " + (string)DB_OBJECT_RELATION_TABLE + " (objectid, parentobjectid, relationtype) "
		"VALUES ((" + strChildSubQuery + "),(" + strParentSubQuery + ")," + stringify(relation) + ")";

	er = m_lpDatabase->DoInsert(strQuery);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));
}

void DBPlugin::deleteSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception)
{
	ECRESULT er = erSuccess;
	string strQuery;
	unsigned int ulAffRows = 0;
	string strParentSubQuery;
	string strChildSubQuery;

	strParentSubQuery = 
		"SELECT id FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(parentobject) + "'" 
		" AND objecttype = " + stringify(parenttype);

	strChildSubQuery = 
		"SELECT id FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(childobject) + "'"
		" AND objecttype = " + stringify(childtype);

	strQuery =
		"DELETE FROM " + (string)DB_OBJECT_RELATION_TABLE + " "
		"WHERE objectid = (" + strChildSubQuery + ") "
			"AND parentobjectid = (" + strParentSubQuery + ") "
			"AND relationtype = " + stringify(relation);

	er = m_lpDatabase->DoDelete(strQuery, &ulAffRows);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	if (ulAffRows != 1)
		throw objectnotfound("db_user: relation " + parentobject);
}

list<objectsignature_t> DBPlugin::searchObjects(string match, userobject_type_t type, char *search_props[], char *return_prop, unsigned int ulFlags) throw(std::exception) {
	ECRESULT er = erSuccess;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;
	DB_LENGTHS lpDBLen = NULL;
	string signature;
	string id;
	list<objectsignature_t> lSignatures;

	string strQuery =
		"SELECT DISTINCT ";
	if (return_prop)
		strQuery += "opret.value, modtime.value ";
	else
		strQuery += "o.externid, modtime.value ";
	strQuery +=
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS op "
			"ON op.objectid=o.id ";
    
	if (return_prop) {
		strQuery +=
			"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS opret "
				"ON opret.objectid=o.id ";
	}

	strQuery +=
		"LEFT JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS modtime "
			"ON modtime.objectid=o.id "
			"AND modtime.propname = '" + OP_MODTIME + "' "
		"WHERE o.objecttype=" + stringify(type) + " AND (";

	string strMatch = m_lpDatabase->Escape(match);
	string strMatchPrefix;
	
	if (!(ulFlags & EMS_AB_ADDRESS_LOOKUP)) {
		strMatch = "%" + strMatch + "%";
		strMatchPrefix = " LIKE ";
	} else {
		strMatchPrefix = " = ";
	}

	for (unsigned int i = 0; search_props[i] != NULL; i++) {
		strQuery += "(op.propname='" + (string)search_props[i] + "' AND op.value " + strMatchPrefix + " '" + strMatch + "')";
		if (search_props[i + 1] != NULL)
			strQuery += " OR ";
	}
	
	strQuery += ")";

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	/* TODO: check with a point system,
	 * if you have 2 objects, one have a match of 99% and one 50%
	 * use the one with 99% */
	if(m_lpDatabase->GetNumRows(lpResult) == 0)
		throw objectnotfound(string("db_user: no match: ") + match);

	while ((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if(lpDBRow[0] == NULL)
			throw runtime_error(string("db_row_failed: object null"));

		if(lpDBRow[1] != NULL)
			signature = lpDBRow[1];
		else
			signature.clear();

		lpDBLen = m_lpDatabase->FetchRowLengths(lpResult);
		ASSERT(lpDBLen != NULL);
		if (lpDBLen[0] == 0)
			throw runtime_error(string("db_row_failed: object empty"));

		id.assign(lpDBRow[0], lpDBLen[0]);
		lSignatures.push_back(objectsignature_t(id, signature));
	}

	return lSignatures;
}

auto_ptr<quotadetails_t> DBPlugin::getQuota(userobject_type_t type, const objectid_t &id, bool bGetUserDefault) throw(std::exception)
{
	auto_ptr<quotadetails_t> lpDetails;
	ECRESULT er = erSuccess;
	string strQuery;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;

	strQuery =
		"SELECT op.propname, op.value "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS op "
			"ON op.objectid = o.id "
		"WHERE o.externid = '" +  m_lpDatabase->Escape(id) + "' "
			"AND o.objecttype = " + stringify(type);

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	lpDetails = auto_ptr<quotadetails_t>(new quotadetails_t());
	lpDetails->bIsUserDefaultQuota = bGetUserDefault;

	while ((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if(lpDBRow[0] == NULL || lpDBRow[1] == NULL)
			continue;

		if (bGetUserDefault) {
			if (strcmp(lpDBRow[0], OP_UD_HARDQUOTA) == 0)
				lpDetails->llHardSize = _atoi64(lpDBRow[1]);
			else if(strcmp(lpDBRow[0], OP_UD_SOFTQUOTA) == 0)
				lpDetails->llSoftSize = _atoi64(lpDBRow[1]);
			else if(strcmp(lpDBRow[0], OP_UD_WARNQUOTA) == 0)
				lpDetails->llWarnSize = _atoi64(lpDBRow[1]);
			else if(strcmp(lpDBRow[0], OP_UD_USEDEFAULTQUOTA) == 0)
				lpDetails->bUseDefaultQuota = !!atoi(lpDBRow[1]);
		} else {
			if (strcmp(lpDBRow[0], OP_HARDQUOTA) == 0)
				lpDetails->llHardSize = _atoi64(lpDBRow[1]);
			else if(strcmp(lpDBRow[0], OP_SOFTQUOTA) == 0)
				lpDetails->llSoftSize = _atoi64(lpDBRow[1]);
			else if(strcmp(lpDBRow[0], OP_WARNQUOTA) == 0)
				lpDetails->llWarnSize = _atoi64(lpDBRow[1]);
			else if(strcmp(lpDBRow[0], OP_USEDEFAULTQUOTA) == 0)
				lpDetails->bUseDefaultQuota = !!atoi(lpDBRow[1]);
		}
	}

	return lpDetails;
}

int DBPlugin::setQuota(userobject_type_t type, const objectid_t &id, quotadetails_t quotadetails) throw(std::exception)
{
	ECRESULT er = erSuccess;
	string strQuery;
	string strSubQuery;
	string op_default;
	string op_hard;
	string op_soft;
	string op_warn;

	if (quotadetails.bIsUserDefaultQuota) {
		op_default = OP_UD_USEDEFAULTQUOTA;
		op_hard = OP_UD_HARDQUOTA;
		op_soft = OP_UD_SOFTQUOTA;
		op_warn = OP_UD_WARNQUOTA;
	} else {
		op_default = OP_USEDEFAULTQUOTA;
		op_hard = OP_HARDQUOTA;
		op_soft = OP_SOFTQUOTA;
		op_warn = OP_WARNQUOTA;
	}

	strSubQuery = 
		"SELECT id FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(id) + "'"
		" AND objecttype = " + stringify(type);

	// Update new quota settings
	strQuery =
		"REPLACE INTO " + (string)DB_OBJECTPROPERTY_TABLE + "(objectid, propname, value) VALUES"
			"((" + strSubQuery + "), '" + op_default + "','" + stringify(quotadetails.bUseDefaultQuota) + "'),"
			"((" + strSubQuery + "), '" + op_hard + "','" + stringify_int64(quotadetails.llHardSize) + "'),"
			"((" + strSubQuery + "), '" + op_soft + "','" + stringify_int64(quotadetails.llSoftSize) + "'),"
			"((" + strSubQuery + "), '" + op_warn + "','" + stringify_int64(quotadetails.llWarnSize) + "')";

	er = m_lpDatabase->DoInsert(strQuery);
	if (er != erSuccess) // Maybe on this point the user is broken.
		throw runtime_error(string("db_query: ") + strerror(er));

	return 0;
}

ECRESULT DBPlugin::CreateMD5Hash(const std::string &strData, std::string* lpstrResult)
{
	ECRESULT er = erSuccess;
	MD5 *crypt = NULL;
	char *hex = NULL;
	std::string salt;
	std::ostringstream s;

	if (strData.empty() || lpstrResult == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	crypt = new MD5();

	s.setf(ios::hex, ios::basefield);
	s.fill('0');
	s.width(8);
	s << rand_mt();
	salt = s.str();

	crypt->update((unsigned char*)salt.c_str(), (unsigned int)salt.size());
	crypt->update((unsigned char*)strData.c_str(), (unsigned int)strData.size());
	crypt->finalize();

	hex = crypt->hex_digest();

	(*lpstrResult) = salt+hex;

exit:
	if(hex)
		delete [] hex;

	if(crypt)
		delete crypt;

	return er;
}

void DBPlugin::addSendAsToDetails(const objectid_t &objectid, objectdetails_t *lpDetails)
{
	ECRESULT er = erSuccess;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;
	DB_LENGTHS lpDBLen = NULL;
	string strQuery;

	// add sendas relations
	strQuery = 
		"SELECT o.externid "
		"FROM " + (string)DB_OBJECT_RELATION_TABLE + " AS ort "
		"JOIN " + (string)DB_OBJECT_TABLE + " AS p "
			"ON p.id = ort.parentobjectid "
		"JOIN " + (string)DB_OBJECT_TABLE + " AS o "
			"ON o.id = ort.objectid "
		"WHERE p.externid = '" + m_lpDatabase->Escape(objectid) + "' "
			"AND p.objecttype = " + stringify(USEROBJECT_TYPE_USER) + " "
			"AND ort.relationtype = " + stringify(OBJECTRELATION_USER_SENDAS);

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	while((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if(lpDBRow[0] == NULL)
			continue;
		lpDBLen = m_lpDatabase->FetchRowLengths(lpResult);
		ASSERT(lpDBLen != NULL);
		if (lpDBLen[0] == 0)
			continue;
		lpDetails->AddPropString(OB_PROP_L_SENDAS, string(lpDBRow[0], lpDBLen[0]));
	}
}
