/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <iostream>
#include <string>
#include <stdexcept>
#include <algorithm>

#include <errno.h>
#include <assert.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "EMSAbTag.h"
#include "ECConfig.h"
#include "ECDefs.h"
#include "ECLogger.h"

#include "stringutil.h"
#include "md5.h"

using namespace std;
#include "ECDatabaseFactory.h"
#include "DBUserPlugin.h"


extern "C" {
	UserPlugin* getUserPluginInstance(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed) {
		return new DBUserPlugin(pluginlock, config, logger, lpStatsCollector, bHosted, bDistributed);
	}

	void deleteUserPluginInstance(UserPlugin *up) {
		delete up;
	}
}

DBUserPlugin::DBUserPlugin(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed) : DBPlugin(pluginlock, config, logger, lpStatsCollector, bHosted, bDistributed)
{
	if (bDistributed)
		throw notsupported("Distributed Zarafa not supported when using the Database Plugin");
}

DBUserPlugin::~DBUserPlugin()
{
}

void DBUserPlugin::InitPlugin() throw(exception)
{
	DBPlugin::InitPlugin();
}

objectsignature_t DBUserPlugin::resolveName(userobject_type_t type, const string &name, const string &company) throw(exception)
{
	objectid_t	id;
	ECRESULT	er;
	string		strQuery;
	DB_RESULT_AUTOFREE	lpResult(m_lpDatabase);
	DB_ROW		lpDBRow = NULL;
	DB_LENGTHS	lpDBLen = NULL;
	string username;
	string companyname;
	string signature;
	char *lpszSearchProperty;

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
	    lpszSearchProperty = OP_LOGINNAME;
		break;
    case USEROBJECT_TYPE_GROUP:
        lpszSearchProperty = OP_GROUPNAME;
		break;
	case USEROBJECT_TYPE_COMPANY:
		lpszSearchProperty = OP_COMPANYNAME;
		break;
	default:
		throw runtime_error("Object is wrong type");
	}

	/*
	 * Join the DB_OBJECT_TABLE together twice. This is done because
	 * we need to link the entry for the user and the entry for the
	 * company into a single line. Once those are all linked we can
	 * set the WHERE restrictions to get the correct line.
	 */
	strQuery =
		"SELECT DISTINCT o.externid, modtime.value "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS user "
			"ON user.objectid = o.id "
			"AND user.propname = '" + (string)lpszSearchProperty + "' "
			"AND upper(user.value) = upper('" + m_lpDatabase->Escape(name) + "') ";
	if (m_bHosted && !company.empty()) {
		strQuery +=
			"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS oc "
				"ON oc.objectid = o.id "
				"AND oc.propname = '" + OP_COMPANYID + "' "
				"AND oc.value = '" + m_lpDatabase->Escape(company) + "' ";
	}
	strQuery +=
        "LEFT JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS modtime "
			"ON modtime.propname = '" + OP_MODTIME + "' "
			"AND modtime.objectid = o.id "
		"WHERE o.objecttype = " + stringify(type);

    er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if(er != erSuccess) {
        throw runtime_error(string("db_query: ") + strerror(er));
	}
    
	if(m_lpDatabase->GetNumRows(lpResult) != 1) {
        throw objectnotfound(username);
	}

    lpDBRow = m_lpDatabase->FetchRow(lpResult);
	if(lpDBRow == NULL || lpDBRow[0] == NULL) {
        throw runtime_error(string("db_row_failed: object null"));
	}
    
	lpDBLen = m_lpDatabase->FetchRowLengths(lpResult);
	if (lpDBLen == NULL || lpDBLen[0] == 0) {
		throw runtime_error(string("db_row_failed: object empty"));
	}

    if(lpDBRow[1] != NULL)
        signature = lpDBRow[1];

	id.assign(lpDBRow[0], lpDBLen[0]);

    return objectsignature_t(id, signature);
}


objectsignature_t DBUserPlugin::authenticateUser(const string &username, const string &companyname, const string &password) throw(exception)
{
	objectid_t	id;
	std::string signature;
	ECRESULT	er;
	string		strQuery;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW		lpDBRow = NULL;
	DB_LENGTHS	lpDBLen = NULL;
	
	std::string salt; 
	MD5*		crypt; 
	char*		hex; 
	std::string strMD5;

	/*
	 * Join the DB_OBJECT_TABLE together twice. This is done because
	 * we need to link the entry for the user and the entry for the
	 * company into a single line. Once those are all linked we can
	 * set the WHERE restrictions to get the correct line.
	 */
	strQuery =
		"SELECT pass.propname, pass.value, o.externid, modtime.value "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS op "
			"ON o.id = op.objectid "
		"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS pass "
			"ON pass.objectid = o.id ";
	if (m_bHosted) {
		strQuery +=
		"JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS oc "
			"ON oc.objectid = o.id ";
	}
	strQuery +=
        "LEFT JOIN " + (string)DB_OBJECTPROPERTY_TABLE + " AS modtime "
			"ON modtime.objectid = o.id "
			"AND modtime.propname = '" + OP_MODTIME + "' "
		"WHERE o.objecttype = " + stringify(USEROBJECT_TYPE_USER) + " "
			"AND op.propname = '" + (string)OP_LOGINNAME + "' "
			"AND op.value = '" + m_lpDatabase->Escape(username) + "' "
			"AND pass.propname = '" + (string)OP_PASSWORD "'";
	if (m_bHosted) {
		strQuery +=
			"AND oc.propname = '" + (string)OP_COMPANYID + "' "
			"AND oc.value = '" + m_lpDatabase->Escape(companyname) + "' ";
	}
    
    er = m_lpDatabase->DoSelect(strQuery, &lpResult);
    if(er != erSuccess) {
        throw runtime_error(string("db_query: ") + strerror(er));
    }

    if(m_lpDatabase->GetNumRows(lpResult) != 1) {
        throw login_error("Trying to authenticate failed: wrong username or password");
    }
    
    lpDBRow = m_lpDatabase->FetchRow(lpResult);

    if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL){
        throw runtime_error("Trying to authenticate failed: database error");
    }			

	lpDBLen = m_lpDatabase->FetchRowLengths(lpResult);
	if (lpDBLen == NULL || lpDBLen[2] == 0) {
        throw runtime_error("Trying to authenticate failed: database error");
	}

    if(strcmp(lpDBRow[0], OP_PASSWORD) == 0)
    {
        // Check Password
        crypt = new MD5(); 
        salt = lpDBRow[1]; 
        salt.resize(8); 
        
        crypt->update((unsigned char*)salt.c_str(), (unsigned int)salt.length()); 
        crypt->update((unsigned char*)password.c_str(), (unsigned int)password.size()); 
        crypt->finalize(); 
        
        hex = crypt->hex_digest(); 
        strMD5 = salt+hex; 
        delete [] hex; 
        delete crypt;

        if(strMD5.compare((string)lpDBRow[1]) == 0) {
			id.assign(lpDBRow[2], lpDBLen[2]);	// Password is oke
        }else{
            throw login_error("Trying to authenticate failed: wrong username or password");
        }
    }else{
        throw login_error("Trying to authenticate failed: wrong username or password");
    }
    
    if(lpDBRow[3] != NULL)
        signature = lpDBRow[3];
		
	return objectsignature_t(id,signature);
}

// User management
auto_ptr<objectdetails_t> DBUserPlugin::getObjectDetails(userobject_type_t type, const objectid_t &userid) throw(std::exception)
{
	auto_ptr<objectdetails_t> lpDetails = DBPlugin::getObjectDetails(type, userid);
	if (type == USEROBJECT_TYPE_USER)
		addSendAsToDetails(userid, lpDetails.get());
	return lpDetails;
}

auto_ptr<map<objectid_t, objectdetails_t> > DBUserPlugin::getObjectDetails(userobject_type_t type, list<objectid_t> &userids) throw(std::exception)
{
    map<objectid_t,objectdetails_t> *mapdetails = new map<objectid_t,objectdetails_t>;
    ECRESULT er;
	string strIds;
    string strQuery;
    DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
    DB_ROW lpDBRow = NULL;
	DB_LENGTHS lpDBLen = NULL;
    bool bFirst = true;
    objectdetails_t details;
    objectid_t lastid;
    
    list<objectid_t>::iterator iterID;
    
    if(userids.empty())
        return auto_ptr<map<objectid_t, objectdetails_t> >(mapdetails);

	for(iterID = userids.begin(); iterID != userids.end(); iterID++) { 
		if(!bFirst) 
			strIds += ","; 

		strIds += "'" + m_lpDatabase->Escape(*iterID) + "'"; 
		bFirst = false; 
	} 
        
    strQuery =
	"SELECT op.propname, op.value, o.externid, o.objecttype "
	"FROM " + (string)DB_OBJECT_TABLE + " AS o "
	"JOIN "+(string)DB_OBJECTPROPERTY_TABLE+" AS op "
		"ON op.objectid=o.id "
	"WHERE o.objecttype = " + stringify(type) +
		" AND o.externid IN (" + strIds + ") ORDER BY o.externid";
    
    er = m_lpDatabase->DoSelect(strQuery, &lpResult);
    if(er != erSuccess) 
        throw runtime_error(string("db_query: ") + strerror(er));

	details = objectdetails_t(type);
	details.SetPropString(OB_PROP_S_SYSADMIN, "SYSTEM");

    while((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL)
    {
        if(lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL)
            continue;

		lpDBLen = m_lpDatabase->FetchRowLengths(lpResult);
		if (lpDBLen == NULL || lpDBLen[2] == 0)
			continue;

		objectid_t curid(lpDBRow[2], lpDBLen[2]);
            
        if (lastid != curid && lastid != "") {
			if (type == USEROBJECT_TYPE_USER)
				addSendAsToDetails(lastid, &details);
            (*mapdetails)[lastid] = details;

			details = objectdetails_t(type);
			/* By default the sysadmin is SYSTEM, will be overwritten later */
			details.SetPropString(OB_PROP_S_SYSADMIN, "SYSTEM");
        }
        
        lastid = curid;
    	details.SetPropBool(OB_PROP_B_NONACTIVE, !!(atoi(lpDBRow[3]) == USEROBJECT_TYPE_NONACTIVE));

        if(strcmp(lpDBRow[0], OP_LOGINNAME) == 0)
            details.SetPropString(OB_PROP_S_LOGIN, lpDBRow[1]);
        else if(strcmp(lpDBRow[0], OP_FULLNAME) == 0)
            details.SetPropString(OB_PROP_S_FULLNAME, lpDBRow[1]);
        else if(strcmp(lpDBRow[0], OP_EMAILADDRESS) == 0)
            details.SetPropString(OB_PROP_S_EMAIL, lpDBRow[1]);
        else if(strcmp(lpDBRow[0], OP_ISADMIN) == 0)
            details.SetPropInt(OB_PROP_I_ADMINLEVEL, atoi(lpDBRow[1]));
        else if(strcmp(lpDBRow[0], OP_GROUPNAME) == 0) {
			details.SetPropString(OB_PROP_S_LOGIN, lpDBRow[1]);
            details.SetPropString(OB_PROP_S_FULLNAME, lpDBRow[1]);
		} else if (strcmp(lpDBRow[0], OP_COMPANYNAME) == 0) {
			details.SetPropString(OB_PROP_S_LOGIN, lpDBRow[1]);
			details.SetPropString(OB_PROP_S_FULLNAME, lpDBRow[1]);
			details.SetPropString(OB_PROP_S_COMPANYNAME, lpDBRow[1]);
		} else if (strcmp(lpDBRow[0], OP_COMPANYID) == 0)
			details.SetPropString(OB_PROP_S_COMPANYNAME, lpDBRow[1]);
    }
    
    if(lastid != "") {
		if (type == USEROBJECT_TYPE_USER)
			addSendAsToDetails(lastid, &details);
        (*mapdetails)[lastid] = details;
	}
    
    return auto_ptr<map<objectid_t, objectdetails_t> >(mapdetails);
}

list<objectsignature_t> DBUserPlugin::searchObject(userobject_type_t type, const string &match, unsigned int ulFlags) throw(std::exception)
{
    char *search_props_user[] = { OP_LOGINNAME, OP_FULLNAME, OP_EMAILADDRESS, NULL };
    char *search_props_group[] = { OP_GROUPNAME, NULL };
	char *search_props_company[] = { OP_COMPANYNAME, NULL };
    char **search_props;

	switch (type) {
    case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
        search_props = search_props_user;
		break;
    case USEROBJECT_TYPE_GROUP:
        search_props = search_props_group;
		break;
    case USEROBJECT_TYPE_COMPANY:
		search_props = search_props_company;
		break;
	default:
		throw runtime_error("Object is wrong type");
	}
    
    return searchObjects(match.c_str(), type, search_props, NULL, ulFlags);
}

int DBUserPlugin::setQuota(userobject_type_t type, const objectid_t &id, quotadetails_t quotadetails) throw(std::exception)
{
	string strQuery;
	ECRESULT er = erSuccess;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;

	// check if user exist
	strQuery =
		"SELECT o.externid "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"WHERE o.externid='" + m_lpDatabase->Escape(id) + "' "
			"AND objecttype = " + stringify(type);

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if(er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	if(m_lpDatabase->GetNumRows(lpResult) != 1)
		throw objectnotfound(id);

	lpDBRow = m_lpDatabase->FetchRow(lpResult);
	if(lpDBRow == NULL || lpDBRow[0] == NULL)
		throw runtime_error(string("db_row_failed: object null"));

	return DBPlugin::setQuota(type, id, quotadetails);
}

auto_ptr<objectdetails_t> DBUserPlugin::getPublicStoreDetails() throw(std::exception)
{
	throw notsupported("public store details");
}

auto_ptr<serverdetails_t> DBUserPlugin::getServerDetails(const string &server) throw(std::exception)
{
	throw notsupported("server details");
}

void DBUserPlugin::getUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives) throw (std::exception)
{
	throw notsupported("user count");
}

void DBUserPlugin::addSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception)
{
	ECRESULT er = erSuccess;
	string strQuery;
	DB_RESULT_AUTOFREE  lpResult(m_lpDatabase);

	// Check if parent exist
	strQuery =
		"SELECT o.externid "
		"FROM " + (string)DB_OBJECT_TABLE + " AS o "
		"WHERE o.externid='" + m_lpDatabase->Escape(parentobject) + "'"
		" AND o.objecttype=" + stringify(parenttype);
	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(string("db_query: ") + strerror(er));

	if(m_lpDatabase->GetNumRows(lpResult) != 1)
		throw objectnotfound(string("db_user: Relation does not exist, id:") + parentobject);

	DBPlugin::addSubObjectRelation(relation, parenttype, parentobject, childtype, childobject);
}
