/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <pthread.h>
#include "LDAPCache.h"
#include "LDAPUserPlugin.h"
#include "stringutil.h"

LDAPCache::LDAPCache()
{
	pthread_mutexattr_init(&m_hMutexAttrib);
	pthread_mutexattr_settype(&m_hMutexAttrib, PTHREAD_MUTEX_RECURSIVE);

	pthread_mutex_init(&m_hMutex, &m_hMutexAttrib);

	m_lpCompanyCache = std::auto_ptr<dn_cache_t>(new dn_cache_t());
	m_lpGroupCache = std::auto_ptr<dn_cache_t>(new dn_cache_t());
	m_lpNonactiveCache = std::auto_ptr<dn_cache_t>(new dn_cache_t());
	m_lpUserCache = std::auto_ptr<dn_cache_t>(new dn_cache_t());
	m_lpAddressListCache = std::auto_ptr<dn_cache_t>(new dn_cache_t());
}

LDAPCache::~LDAPCache()
{
	pthread_mutex_destroy(&m_hMutex);
	pthread_mutexattr_destroy(&m_hMutexAttrib);
}

bool LDAPCache::isObjectTypeCached(userobject_type_t ulType)
{
	bool bCached = false;

	pthread_mutex_lock(&m_hMutex);

	switch (ulType) {
	case USEROBJECT_TYPE_USER:
		bCached = !m_lpUserCache->empty();
		break;
	case USEROBJECT_TYPE_NONACTIVE:
		bCached = !m_lpNonactiveCache->empty();
		break;
	case USEROBJECT_TYPE_GROUP:
		bCached = !m_lpGroupCache->empty();
		break;
	case USEROBJECT_TYPE_COMPANY:
		bCached = !m_lpCompanyCache->empty();
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
		bCached = !m_lpAddressListCache->empty();
		break;
	default:
		break;
	}

	pthread_mutex_unlock(&m_hMutex);

	return bCached;
}

void LDAPCache::setObjectDNCache(userobject_type_t ulType, std::auto_ptr<dn_cache_t> lpCache)
{
	
	/*
	 * users/groups caches must be merged instead of overwritten.
	 */
	if (ulType != USEROBJECT_TYPE_COMPANY) {
		/*
		 * Since the key on this map is the DN of a user, we also have to make sure that the 
		 * dn_cache_t part is unique (essentially creating a bimap)
		 */
		 
		// Create a reverse map, making both key and value unique
		std::map<cache_entry_t, std::string> mapReverse;
		std::map<cache_entry_t, std::string>::iterator iterReverse;
		std::auto_ptr<dn_cache_t> lpTmp = getObjectDNCache(NULL, ulType);
		
		// Put in old entries
		for(dn_cache_t::iterator iter = lpTmp->begin(); iter != lpTmp->end(); iter++)
			mapReverse[iter->second] = iter->first;

		// Add new entries
		for (dn_cache_t::iterator iter = lpCache->begin(); iter != lpCache->end(); iter++)
			mapReverse[iter->second] = iter->first;

		// Convert back to normal map
		lpTmp->clear();
		
		for(iterReverse = mapReverse.begin(); iterReverse != mapReverse.end(); iterReverse++)
			(*lpTmp)[iterReverse->second] = iterReverse->first;

		lpCache = lpTmp;
	}

	pthread_mutex_lock(&m_hMutex);

	switch (ulType) {
	case USEROBJECT_TYPE_USER:
		m_lpUserCache = lpCache;
		break;
	case USEROBJECT_TYPE_NONACTIVE:
		m_lpNonactiveCache = lpCache;
		break;
	case USEROBJECT_TYPE_GROUP:
		m_lpGroupCache = lpCache;
		break;
	case USEROBJECT_TYPE_COMPANY:
		m_lpCompanyCache = lpCache;
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
		m_lpAddressListCache = lpCache;
		break;
	default:
		break;
	}

	pthread_mutex_unlock(&m_hMutex);
}

std::auto_ptr<dn_cache_t> LDAPCache::getObjectDNCache(LDAPUserPlugin *lpPlugin, userobject_type_t ulType)
{
	std::auto_ptr<dn_cache_t> cache;

	pthread_mutex_lock(&m_hMutex);

	/* If item was not yet cached, make sure it is done now. */
	if (!isObjectTypeCached(ulType) && lpPlugin)
		lpPlugin->getAllObjects(ulType, string());

	switch (ulType) {
	case USEROBJECT_TYPE_USER:
		cache.reset(new dn_cache_t(*m_lpUserCache.get()));
		break;
	case USEROBJECT_TYPE_NONACTIVE:
		cache.reset(new dn_cache_t(*m_lpNonactiveCache.get()));
		break;
	case USEROBJECT_TYPE_GROUP:
		cache.reset(new dn_cache_t(*m_lpGroupCache.get()));
		break;
	case USEROBJECT_TYPE_COMPANY:
		cache.reset(new dn_cache_t(*m_lpCompanyCache.get()));
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
		cache.reset(new dn_cache_t(*m_lpAddressListCache.get()));
		break;
	default:
		break;
	}

	pthread_mutex_unlock(&m_hMutex);

	return cache;
}

cache_entry_t LDAPCache::getParentForDN(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &dn)
{
	cache_entry_t entry;
	std::string parent_dn;

	if (lpCache->empty())
		goto exit;

	for (dn_cache_t::iterator it = lpCache->begin(); it != lpCache->end(); it++) {
		/* Key should be larger then current guess, but has to be smaller then the userobject dn */
		if (it->first.size() > parent_dn.size() && it->first.size() < dn.size()) {
			/* If key matches the end of the userobject dn, we have a positive match */
			if (stricmp(dn.c_str() + (dn.size() - it->first.size()), it->first.c_str()) == 0) {
				parent_dn = it->first;
				entry = it->second;
			}
		}
	}

exit:
	/* Either empty, or the correct result */
	return entry;
}

std::auto_ptr<dn_list_t> LDAPCache::getChildrenForDN(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &dn)
{
	std::auto_ptr<dn_list_t> list = std::auto_ptr<dn_list_t>(new dn_list_t());

	/* Find al DN's which are hierarchically below the given dn */
	for (dn_cache_t::iterator iter = lpCache->begin(); iter != lpCache->end(); iter++) {
		/* Key should be larger then root DN */
		if (iter->first.size() > dn.size()) {
			/* If key matches the end of the root dn, we have a positive match */
			if (stricmp(iter->first.c_str() + (iter->first.size() - dn.size()), dn.c_str()) == 0)
				list->push_back(iter->first);
		}
	}

	return list;
}

std::auto_ptr<dn_list_t> LDAPCache::GetChildrenForObject(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &name)
{
	return getChildrenForDN(lpCache, getDNForObject(lpCache, name));
}

std::string LDAPCache::getDNForObject(const std::auto_ptr<dn_cache_t> &lpCache, const objectid_t &externid)
{
	for (dn_cache_t::iterator it = lpCache->begin(); it != lpCache->end(); it++) {
		if ((it->second.object.size() == externid.size()) &&
			memcmp(it->second.object.c_str(), externid.c_str(), externid.size()) == 0)
				return it->first;
	}

	return std::string();
}

std::string LDAPCache::getDNForObjectName(const std::auto_ptr<dn_cache_t> &lpCache, const std::string &name)
{
	for (dn_cache_t::iterator it = lpCache->begin(); it != lpCache->end(); it++) {
		if (stricmp(it->second.name.c_str(), name.c_str()) == 0)
				return it->first;
	}

	return std::string();
}

bool LDAPCache::isDNInList(const std::auto_ptr<dn_list_t> &lpList, const std::string &dn)
{
	/* We were given an DN, check if a parent of that dn is listed as filterd */
	for (dn_list_t::iterator iter = lpList->begin(); iter != lpList->end(); iter++) {
		/* Key should be larger then user DN */
		if (iter->size() < dn.size()) {
			/* If key matches the end of the user dn, we have a positive match */
			if (stricmp(dn.c_str() + (dn.size() - iter->size()), iter->c_str()) == 0)
				return true;
		}
	}

	return false;
}
