/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <iostream>
#include <string>
#include <stdexcept>
#include <algorithm>
#include <map>

#include <errno.h>
#include <assert.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "EMSAbTag.h"
#include "ECConfig.h"
#include "ECLogger.h"
#include "stringutil.h"

using namespace std;

#include "LDAPUserPlugin.h"
#include "ldappasswords.h"



#define LDAP_DATA_TYPE_DN			"dn"	// data in attribute like cn=piet,cn=user,dc=localhost,dc=com
#define LDAP_DATA_TYPE_ATTRIBUTE	"attribute"	// data in attribute like piet
#define LDAP_DATA_TYPE_BINARY		"binary"
#define LDAP_DATA_TYPE_TEXT			"text"

#define FETCH_ATTR_VALS 0
#define DONT_FETCH_ATTR_VALS 1

auto_ptr<LDAPCache> LDAPUserPlugin::m_lpCache = auto_ptr<LDAPCache>(new LDAPCache());

LDAPUserPlugin::LDAPUserPlugin(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed)
	: UserPlugin(pluginlock, config, logger, lpStatsCollector, bHosted, bDistributed), m_ldap(NULL), m_iconv(NULL), m_iconvrev(NULL)
{
	// LDAP Defaults
	const configsetting_t lpDefaults[] = {
		{ "ldap_host","localhost" },
		{ "ldap_port","389" },
		{ "ldap_bind_user","" },
		{ "ldap_bind_passwd","" },
		{ "ldap_user_search_base","cn=users,dc=zarafa,dc=com" },
		{ "ldap_user_search_filter","(objectClass=posixAccount)" },
		{ "ldap_user_unique_attribute","cn" },
		{ "ldap_user_unique_attribute_type","text" },
		{ "ldap_group_search_base","cn=groups,dc=zarafa,dc=com" },
		{ "ldap_group_search_filter","(objectClass=posixGroup)" },
		{ "ldap_group_unique_attribute","cn" },
		{ "ldap_group_unique_attribute_type","text" },
		{ "ldap_company_search_base","dc=zarafa,dc=com" },
		{ "ldap_company_search_filter","(objectClass=organizationalUnit)" },
		{ "ldap_company_unique_attribute","ou" },
		{ "ldap_company_unique_attribute_type","text" },
		{ "ldap_fullname_attribute","cn" },
		{ "ldap_loginname_attribute","uid" },
		{ "ldap_password_attribute","userPassword" },
		{ "ldap_nonactive_attribute","zarafaSharedStoreOnly" },
		{ "ldap_user_certificate_attribute", "userCertificate"},
		{ "ldap_emailaddress_attribute","mail" },
		{ "ldap_groupname_attribute","cn" },
		{ "ldap_groupmembers_attribute","member" },
		{ "ldap_groupmembers_attribute_type","text" },
		{ "ldap_companyname_attribute","ou" },
		{ "ldap_isadmin_attribute","zarafaAdmin" },
		{ "ldap_user_sendas_attribute","zarafaSendAsPrivilege" },
		{ "ldap_user_sendas_attribute_type","text" },
		{ "ldap_user_sendas_relation_attribute","" },
		{ "ldap_user_exchange_dn_attribute","" },
		{ "ldap_user_telephone_attribute", "telephoneNumber" },
		{ "ldap_user_department_attribute", "departmentNumber" },
		{ "ldap_user_location_attribute", "physicalDeliveryOfficeName" },
		{ "ldap_user_fax_attribute", "facsimileTelephoneNumber" },
		{ "ldap_company_view_attribute","zarafaViewPrivilege" },
		{ "ldap_company_view_attribute_type","text" },
		{ "ldap_company_view_relation_attribute","" },
		{ "ldap_company_admin_attribute","zarafaAdminPrivilege" },
		{ "ldap_company_admin_attribute_type","text" },
		{ "ldap_company_admin_relation_attribute","" },
		{ "ldap_company_system_admin_attribute","zarafaSystemAdmin" },
		{ "ldap_company_system_admin_attribute_type","text" },
		{ "ldap_company_system_admin_relation_attribute","" },
		{ "ldap_authentication_method","bind" },
		{ "ldap_quotaoverride_attribute","zarafaQuotaOverride" },
		{ "ldap_warnquota_attribute","zarafaQuotaWarn" },
		{ "ldap_softquota_attribute","zarafaQuotaSoft" },
		{ "ldap_hardquota_attribute","zarafaQuotaHard" },
		{ "ldap_userdefault_quotaoverride_attribute","zarafaUserDefaultQuotaOverride" },
		{ "ldap_userdefault_warnquota_attribute","zarafaUserDefaultQuotaWarn" },
		{ "ldap_userdefault_softquota_attribute","zarafaUserDefaultQuotaSoft" },
		{ "ldap_userdefault_hardquota_attribute","zarafaUserDefaultQuotaHard" },
		{ "ldap_quota_userwarning_recipients_attribute","zarafaQuotaUserWarningRecipients" },
		{ "ldap_quota_userwarning_recipients_attribute_type","text" },
		{ "ldap_quota_userwarning_recipients_relation_attribute","" },
		{ "ldap_quota_companywarning_recipients_attribute","zarafaQuotaCompanyWarningRecipients" },
		{ "ldap_quota_companywarning_recipients_attribute_type","text" },
		{ "ldap_quota_companywarning_recipients_relation_attribute","" },
		{ "ldap_quota_multiplier","1" },
		{ "ldap_user_scope","sub" },
		{ "ldap_group_scope","sub" },
		{ "ldap_company_scope","sub" },
		{ "ldap_groupmembers_relation_attribute", "" },
		{ "ldap_protocol", "ldap" },
		{ "ldap_server_charset", "UTF-8" },
		{ "ldap_last_modification_attribute", "modifyTimestamp" },
		{ "ldap_addresslist_search_base","cn=addresslists,dc=zarafa,dc=com" },
		{ "ldap_addresslist_scope","sub" },
		{ "ldap_addresslist_search_filter","(objectClass=zarafaAddressList)" },
		{ "ldap_addresslist_unique_attribute","cn" },
		{ "ldap_addresslist_unique_attribute_type","text" },
		{ "ldap_addresslist_filter_attribute","zarafaFilter" },
		{ "ldap_addresslist_name_attribute","cn" },
		{ "ldap_network_timeout", "30" },
		{ "ldap_object_search_filter", "" },
		{ "ldap_filter_cutoff_elements", "1000" },
		{ NULL, NULL },
	};

	m_config = new ECConfig(lpDefaults);
	if (!m_config->LoadSettings(config->GetSetting("user_plugin_config")))
		logger->Log(EC_LOGLEVEL_ERROR, "Failed to open LDAP plugin configuration file, using defaults.");
	if (m_config->HasErrors()) {
		LogConfigErrors(m_config, logger);
		delete m_config;
		throw runtime_error(string("not a valid configuration file."));
	}
}

void LDAPUserPlugin::InitPlugin() throw(exception)
{
	char *ldap_binddn = m_config->GetSetting("ldap_bind_user");
	char *ldap_bindpw = m_config->GetSetting("ldap_bind_passwd");

	// FIXME: encode the user and password, now it's depended in which charset the config is saved
	m_ldap = ConnectLDAP(ldap_binddn, ldap_bindpw);

	m_iconv =    new ECIConv("windows-1252", m_config->GetSetting("ldap_server_charset"));
	m_iconvrev = new ECIConv(m_config->GetSetting("ldap_server_charset"), "windows-1252");
}

LDAP *LDAPUserPlugin::ConnectLDAP(const char *bind_dn, const char *bind_pw) throw(exception) {
	int rc;
	
	LDAP *ld = NULL;
	struct timeval tstart, tend;
	LONGLONG llelapsedtime;
	struct timeval timeout;

	gettimeofday(&tstart, NULL);

	if((bind_dn && bind_dn[0] != 0) && (bind_pw == NULL || bind_pw[0] == 0)) {
	    // Username specified, but no password. Apparently, OpenLDAP will attempt
	    // an anonymous bind when this is attempted. We therefore disallow this
	    // to make sure you can authenticate a user's password with this function
		throw runtime_error(string("Disallowing NULL password for user ") + bind_dn);
    }

	// Initialize LDAP struct
	char *ldap_host = m_config->GetSetting("ldap_host");
	char *ldap_port = m_config->GetSetting("ldap_port");

	int port = strtoul(ldap_port, NULL, 10);

	ld = ldap_init(ldap_host, port);
	if (ld == NULL) {
		m_lpStatsCollector->Increment(SCN_LDAP_CONNECT_FAILED);
		throw runtime_error(string("ldap_init: ") + strerror(errno));
	}

	// Go to SSL if required
	int tls = LDAP_OPT_X_TLS_HARD;
	if(strcmp(m_config->GetSetting("ldap_protocol"), "ldaps") == 0) {
	    if((rc = ldap_set_option(ld, LDAP_OPT_X_TLS, &tls)) != LDAP_SUCCESS) {
	        m_logger->Log(EC_LOGLEVEL_WARNING, "Failed to initiate SSL for ldap: %s", ldap_err2string(rc));
        }
    }

	int version = LDAP_VERSION3;
	ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &version);
	// Disable response message size restrictions (but the server's
	// restrictions still apply)
	int limit = 0;
	ldap_set_option(ld, LDAP_OPT_SIZELIMIT, &limit);

	// Search referrals are never accepted  - FIXME maybe needs config option
	ldap_set_option(ld, LDAP_OPT_REFERRALS, LDAP_OPT_OFF);

	timeout.tv_sec = atoui(m_config->GetSetting("ldap_network_timeout"));
	timeout.tv_usec = 0;
	
	// Set network timeout (for connect)
	ldap_set_option(ld, LDAP_OPT_NETWORK_TIMEOUT, &timeout);

#if 0		// OpenLDAP stupidly closes the connection when TLS is not configured on the server...
#ifdef LINUX // Only available in Windows XP, so we can't use this on windows platform ..
#ifdef HAVE_LDAP_START_TLS_S
	// Initialize TLS-secured connection - this is the first command
	// after ldap_init, so it will be the call that actually connects
	// to the server.
	if ((rc = ldap_start_tls_s(ld, NULL, NULL)) != LDAP_SUCCESS) {
		m_logger->Log(EC_LOGLEVEL_WARNING, "Failed to secure ldap connection: %s", ldap_err2string(rc));
	}
#endif
#endif
#endif

	// Bind
	// For these two values: if they are both NULL, anonymous bind
	// will be used (ldap_binddn, ldap_bindpw)
	if ((rc = ldap_simple_bind_s(
			 ld, (char *)bind_dn, (char *)bind_pw)) != LDAP_SUCCESS)
	{
	    ldap_unbind_s(ld);

		m_lpStatsCollector->Increment(SCN_LDAP_CONNECT_FAILED);

		throw runtime_error(string("ldap_bind_s: ") + ldap_err2string(rc));
	}

	gettimeofday(&tend, NULL);

	llelapsedtime = difftimeval(&tstart,&tend);

	m_lpStatsCollector->Increment(SCN_LDAP_CONNECTS);
	m_lpStatsCollector->Increment(SCN_LDAP_CONNECT_TIME, llelapsedtime);
	m_lpStatsCollector->Max(SCN_LDAP_CONNECT_TIME_MAX, llelapsedtime);
	
	return ld;
}

LDAPUserPlugin::~LDAPUserPlugin() {
	// Disconnect from the LDAP server

	if(m_ldap)
		ldap_unbind_s(m_ldap);

	if (m_config)
		delete m_config;

    if (m_iconv)
        delete m_iconv;

    if (m_iconvrev)
        delete m_iconvrev;
}

int LDAPUserPlugin::my_ldap_search_s(char *base, int scope, char *filter, char *attrs[], int attrsonly, LDAPMessage **lppres)
{
	int result=LDAP_SUCCESS;
	string req;
	struct timeval tstart, tend;
	LONGLONG llelapsedtime;
	LDAPMessage *res = NULL;

	gettimeofday(&tstart, NULL);

	if (attrs) {
		for (unsigned int i = 0; attrs[i] != NULL; i++)
			req += string(attrs[i]) + " ";
	}
	
	m_logger->Log(EC_LOGLEVEL_DEBUG, "ldapsearch(\"%s\" \"%s\" %s)", base, filter, req.c_str());

	if (m_ldap != NULL)
		result = ldap_search_s(m_ldap,base,scope,filter,attrs,attrsonly,&res);

    if (m_ldap == NULL || result == LDAP_SERVER_DOWN) {
        // server is 'down'. try 1 reconnect and retry, and if that fails, just completely fail
        // We need this because LDAP server connections can timeout
    	char *ldap_binddn = m_config->GetSetting("ldap_bind_user");
    	char *ldap_bindpw = m_config->GetSetting("ldap_bind_passwd");

		if(res) {
			ldap_msgfree(res);
			res = NULL;
		}

		if (m_ldap != NULL) {
			ldap_unbind_s(m_ldap);
			m_ldap = NULL;
		}
// FIXME: encode the user and password, now it's depended in which charset the config is saved
    	m_ldap = ConnectLDAP(ldap_binddn, ldap_bindpw);

		m_lpStatsCollector->Increment(SCN_LDAP_RECONNECTS);

        result = ldap_search_s(m_ldap,base,scope,filter,attrs,attrsonly,&res);
    }
	if(result != LDAP_SUCCESS)
		goto exit;

	*lppres = res;
	res = NULL;

	gettimeofday(&tend, NULL);
	
	llelapsedtime = difftimeval(&tstart,&tend);
	
	m_logger->Log(EC_LOGLEVEL_INFO, "ldaptiming[%08.2f] (\"%s\" \"%s\" %s)", llelapsedtime/1000000., base, filter, req.c_str());

	m_lpStatsCollector->Increment(SCN_LDAP_SEARCH);
	m_lpStatsCollector->Increment(SCN_LDAP_SEARCH_TIME, llelapsedtime);
	m_lpStatsCollector->Max(SCN_LDAP_SEARCH_TIME_MAX, llelapsedtime);

exit:
	if (result != LDAP_SUCCESS)
		m_lpStatsCollector->Increment(SCN_LDAP_SEARCH_FAILED);

	if (res)
		ldap_msgfree(res);

    return result;
}

string LDAPUserPlugin::getSearchBase(userobject_type_t type, const string &companyid) throw(std::exception)
{
	string search_base;

	/*
	 * When using hosted the company search base is the search base for
	 * all types. This is mandated because sometimes we query a username
	 * without knowing the company, to perform such a search we should
	 * search on the lowest possible position. _however_ forcing users
	 * to always set the ldap_user_search_base and ldap_group_search_base
	 * to the same value of ldap_company_search_base will guarantee that
	 * a lot of users will run into issues because they didn't change the
	 * settings. Because of that, we perform the following actions when
	 * determing the search base:
	 *  1: Is type USEROBJECT_TYPE_COMPANY
	 *  	- Return ldap_company_search_base
	 *  2: Is companyid non-empty resolve companyid to DN
	 *  	- On Success: Use this as search_base
	 *  3: Is ldap_company_search_base non-null
	 *  	- Use this as search_base
	 *  4: Use type to get the matching {user,group}_search_base
	 *  	- Use this as search base
	 *  5: Panic and start screaming while running in circles.
	 */
	if (m_bHosted) {
		if (!companyid.empty()) {
			std::auto_ptr<dn_cache_t> lpCache = m_lpCache->getObjectDNCache(this, USEROBJECT_TYPE_COMPANY);

			search_base = m_lpCache->getDNForObjectName(lpCache, companyid);
			if (!search_base.empty())
				return search_base;

			search_base = m_lpCache->getDNForObject(lpCache, companyid);
			if (!search_base.empty())
				return search_base;

			// If the company was not found, and we're not searching for companies, throw search error.
			// this fixes searching for email addresses on hosted installations,
			// where the domain of the email address does not match with the company name.
			if (type != USEROBJECT_TYPE_COMPANY)
				throw objectnotfound(companyid);
		}

		search_base = m_config->GetSetting("ldap_company_search_base");
		if (!search_base.empty())
			return search_base;

		throw runtime_error("no search base available");
	}

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
			return m_config->GetSetting("ldap_user_search_base");
		break;
	case USEROBJECT_TYPE_GROUP:
			return m_config->GetSetting("ldap_group_search_base");
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
			return m_config->GetSetting("ldap_addresslist_search_base");
		break;
	default:
		throw runtime_error("Object is wrong type");
	}

	throw runtime_error("no search base available");
}

string LDAPUserPlugin::getSearchFilter(userobject_type_t type)
{
	string filter;
	char *nonactive = NULL;

	/*
	 * The ldap_nonactive_attribute is an optional configuration option.
	 * By default the option is set to an empty string which means that by
	 * default there is no support for nonactive users. It can be enabled
	 * by the user by providing a sensible value to the ldap_nonactive_attribute.
	 *
	 * Why is this nescessery:
	 * We are not forcing our users to install the zarafa.schema but all our
	 * default configuration values make use of the zarafa Attributes from that
	 * schema. LDAP/Active Directory behave rather odd to queries containing
	 * attributes which are not defined in any of the installed schemas.
	 * For example:
	 * 		(|(!(zarafaSharedStore=1))(zarafaSharedStore=1))
	 * this query will return FALSE when zarafaSharedStore is unknown.
	 * On some Active Directory implementations this could be bypassed by
	 * providing the query:
	 * 		(|(!(zarafaSharedStore=*))(zarafaSharedStore=0))
	 * but unfortunately this is not true for all Active Directory versions
	 * and certainly not for OpenLDAP (The query is correct when you use it
	 * on an object which does not have the atribute, although it really is
	 * defined in the server).
	 *
	 * On all previous Zarafa versions we accepted users as Active users
	 * when the schema was not installed. To prevent accidental user deletions
	 * this behavior must be maintained. In addition it makes it easier for
	 * new installations as well since new users are often confronted with
	 * empty userlists when they try to list all users when they haven't
	 * installed the zarafa.schema yet.
	 */
	switch (type) {
	case USEROBJECT_TYPE_USER:
		nonactive = m_config->GetSetting("ldap_nonactive_attribute");

		/* No nonactive attribute: All detected users will be active */
		filter = string(m_config->GetSetting("ldap_user_search_filter"));
		if (nonactive && (strlen(nonactive) > 0))
			filter = "(&" + filter + "(|(!(" + string(nonactive) + "=*))(" + string(nonactive) + "=0)))";

		break;
	case USEROBJECT_TYPE_NONACTIVE:
		nonactive = m_config->GetSetting("ldap_nonactive_attribute");

		/* No nonactive attribute: nonactive users don't exist */
		if (nonactive && (strlen(nonactive) > 0))
			filter = "(&" + string(m_config->GetSetting("ldap_user_search_filter")) +
					  "(" + string(m_config->GetSetting("ldap_nonactive_attribute")) + "=1))";

		break;
	case USEROBJECT_TYPE_GROUP:
		filter = string(m_config->GetSetting("ldap_group_search_filter"));
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
		filter = string(m_config->GetSetting("ldap_addresslist_search_filter"));
		break;
	case USEROBJECT_TYPE_COMPANY:
		if (m_bHosted)
			filter = string(m_config->GetSetting("ldap_company_search_filter"));
		/* Hosted disabled, fallback to default... */
	default:
		break;
	}

	return filter;
}

string LDAPUserPlugin::getObjectSearchFilter(userobject_type_t type, const objectid_t &id, const string &attr, const string &attr_type)
{
	string search_filter = getSearchFilter(type);
	string search_data;

	/* No generic search filter, we shouldn't provide object search filter */
	if (search_filter.empty())
		return search_filter;

	// Set binary uniqueid to escaped string
	if(attr_type == LDAP_DATA_TYPE_BINARY)
		BintoEscapeSequence(id.c_str(), id.size(), &search_data);
	else
		search_data = StringEscapeSequence(id);

	return "(&" + search_filter + "(" + attr + "=" + search_data + "))";
}

ULONG LDAPUserPlugin::ScopetoScope(const char* lpScope, ULONG ulDefaultScope)
{
	if(lpScope == NULL)
		return ulDefaultScope;

	if(stricmp(lpScope, "sub") == 0)
		return LDAP_SCOPE_SUBTREE;
	else if(stricmp(lpScope, "one") == 0)
		return LDAP_SCOPE_ONELEVEL;
	else if(stricmp(lpScope, "base") == 0)
		return LDAP_SCOPE_BASE;
	else
		return ulDefaultScope;
}

string LDAPUserPlugin::objectUniqueIDtoAttributeData(userobject_type_t type, objectid_t uniqueid, const char* lpAttr) throw(std::exception)
{
	string			ldap_filter;
	LDAPMessage*	res = NULL;
	int				rc;
	string			strData;
	LDAPMessage*	entry;
	char*			att;
	BerElement*		ber = NULL;
	bool			bDataAttrFound = false;
	ULONG			ulScope;

	string ldap_basedn;
	string active_filter;
	char *unique_attr;
	char *unique_attr_type;

	char *request_attrs[] = { (char*)lpAttr,
							  NULL };

	switch (type) {
    case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
    	ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));
    	unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
    	unique_attr_type = m_config->GetSetting("ldap_user_unique_attribute_type");
		break;
    case USEROBJECT_TYPE_GROUP:
    	ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
    	unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
    	unique_attr_type = m_config->GetSetting("ldap_group_unique_attribute_type");
		break;
    case USEROBJECT_TYPE_COMPANY:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_company_unique_attribute_type");
		break;
    case USEROBJECT_TYPE_ADDRESSLIST:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_addresslist_scope"));
		unique_attr = m_config->GetSetting("ldap_addresslist_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_addresslist_unique_attribute_type");
		break;
	default:
		throw runtime_error("Object is wrong type");
	}

	ldap_basedn = getSearchBase(type);
	ldap_filter = getObjectSearchFilter(type, uniqueid, unique_attr, unique_attr_type);
	
	/* No LDAP filter, object does not exist */
	if (ldap_filter.empty())
		throw objectnotfound(uniqueid);

	if ((rc = my_ldap_search_s(
			 (char *)ldap_basedn.c_str(), ulScope,
			 (char *)ldap_filter.c_str(),
			 request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	switch(ldap_count_entries(m_ldap, res)) {
	case 0:
	    ldap_msgfree(res);
		throw objectnotfound(ldap_filter);
	case 1:
		break;
	default:
	    ldap_msgfree(res);
		throw toomanyobjects(string("More than one object returned in search ") + ldap_filter);
	}

	entry = ldap_first_entry(m_ldap, res);
	if(entry == NULL) {
	    ldap_msgfree(res);
		throw runtime_error("ldap_dn: broken.");
	}

	for (att = ldap_first_attribute(m_ldap, entry, &ber);
		 att != NULL;
		 att = ldap_next_attribute(m_ldap, entry, ber))
	{
		if (stricmp(att, lpAttr) == 0) {
			strData = getLDAPAttributeValue(att, entry);
			bDataAttrFound = true;
			ldap_memfree(att);
			break;
		}
		ldap_memfree(att);
	}

	if(ber)
		ber_free(ber, 0);

	if(res)
		ldap_msgfree(res);

	if(bDataAttrFound == false)
		throw runtime_error(string(lpAttr)+" attribute not found");

	return strData;
}

string LDAPUserPlugin::objectDNtoAttributeData(userobject_type_t type, string dn, const char *lpAttr) throw(std::exception)
{
	string			ldap_filter;
	LDAPMessage*	res = NULL;
	int				rc;
	LDAPMessage*	entry = NULL;
	char*			att = NULL;
	BerElement*		ber = NULL;
	bool			bAttrFound = false;

	string search_filter;
	string attr_data;

	search_filter = getSearchFilter(type);
	if (search_filter.empty())
		throw objectnotfound(dn);

	char *request_attrs[] = { (char *)lpAttr,
							  NULL };

	if ((rc = my_ldap_search_s(
			(char*)dn.c_str(), LDAP_SCOPE_BASE,
			(char*)search_filter.c_str(),
			 request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	switch(ldap_count_entries(m_ldap, res)) {
	case 0:
	    ldap_msgfree(res);
		throw objectnotfound(dn);
	case 1:
		break;
	default:
	    ldap_msgfree(res);
		throw toomanyobjects(string("More than one object returned in search ") + dn);
	}

	entry = ldap_first_entry(m_ldap, res);
	if(entry == NULL) {
	    ldap_msgfree(res);
		throw runtime_error("ldap_dn: group broken.");
	}

	for (att = ldap_first_attribute(m_ldap, entry, &ber);
		 att != NULL;
		 att = ldap_next_attribute(m_ldap, entry, ber))
	{
		if (stricmp(att, lpAttr) == 0) {
			attr_data = getLDAPAttributeValue(att, entry);
			bAttrFound = true;
		}

		ldap_memfree(att);
	}

	if(ber)
		ber_free(ber, 0);

	if(res)
		ldap_msgfree(res);

	if(bAttrFound == false)
		throw runtime_error("attribute not found: " + dn);

	return attr_data;
}

string LDAPUserPlugin::objectUniqueIDtoObjectDN(userobject_type_t type, objectid_t uniqueid) throw(std::exception)
{
	std::auto_ptr<dn_cache_t> lpCache = m_lpCache->getObjectDNCache(this, type);
	string			ldap_filter;
	LDAPMessage*	res = NULL;
	int				rc;
	string			strDN;
	LDAPMessage*	entry = NULL;
	char*			dn = NULL;
	ULONG			ulScope;
	char*			unique_attr;
	char*			unique_attr_type;

	string			ldap_basedn;

	/* 
	 * The cache should actually contain this entry, search for the uniqueid in there first.
	 * In the rare case that the cache didn't contain the entry, check LDAP.
	 */
	strDN = m_lpCache->getDNForObject(lpCache, uniqueid);
	if (!strDN.empty())
		return strDN;

	/*
	 * That's odd, the cache didn't contain the uniqueid. Perform a LDAP query
	 * to search for the object, but chances are high the object doesn't exist at all.
	 */
	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
    	ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));
		unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_user_unique_attribute_type");
		break;
   	case USEROBJECT_TYPE_GROUP:
    	ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_group_unique_attribute_type");
		break;
    case USEROBJECT_TYPE_COMPANY:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_company_unique_attribute_type");
    	break;
    case USEROBJECT_TYPE_ADDRESSLIST:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_addresslist_scope"));
		unique_attr = m_config->GetSetting("ldap_addresslist_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_addresslist_unique_attribute_type");
    	break;
	default:
		throw runtime_error("Object is wrong type");
	}

	ldap_basedn = getSearchBase(type);
	ldap_filter = getObjectSearchFilter(type, uniqueid, unique_attr, unique_attr_type);

	/* No LDAP filter, object does not exist */
	if (ldap_filter.empty())
		throw objectnotfound(uniqueid);

	if ((rc = my_ldap_search_s(
			 (char *)ldap_basedn.c_str(), LDAP_SCOPE_SUBTREE,
			 (char *)ldap_filter.c_str(),
			 NULL, DONT_FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	switch(ldap_count_entries(m_ldap, res)) {
	case 0:
	    ldap_msgfree(res);
		throw objectnotfound(ldap_filter);
	case 1:
		break;
	default:
	    ldap_msgfree(res);
		throw toomanyobjects(string("More than one object returned in search ") + ldap_filter);
	}

	entry = ldap_first_entry(m_ldap, res);
	if(entry == NULL) {
	    ldap_msgfree(res);
		throw runtime_error("ldap_dn: broken.");
	}

	dn = ldap_get_dn(m_ldap, entry);

	strDN = dn;

	if(dn)
		ldap_memfree(dn);

	if(res)
		ldap_msgfree(res);

	return strDN;
}

objectsignature_t LDAPUserPlugin::objectDNtoObjectSignature(userobject_type_t type, const string &dn) throw(std::exception)
{
	objectid_t		id;
	std::string	    signature;
	string			ldap_filter;
	LDAPMessage*	res = NULL;
	int				rc;
	LDAPMessage*	entry = NULL;
	char*			att = NULL;
	BerElement*		ber = NULL;
	bool			bUniqueAttrFound = false;

	char *unique_attr;
	string search_filter;

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		break;
    case USEROBJECT_TYPE_GROUP:
	    unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		break;
	case USEROBJECT_TYPE_COMPANY:
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
		unique_attr = m_config->GetSetting("ldap_addresslist_unique_attribute");
		break;
	default:
		throw runtime_error("Object is wrong type");
	}

	search_filter = getSearchFilter(type);
	if (search_filter.empty())
		throw objectnotfound(dn);

    char *modify_attr = m_config->GetSetting("ldap_last_modification_attribute");

	char *request_attrs[] = { unique_attr,
	                          modify_attr,
							  NULL };

	if ((rc = my_ldap_search_s(
			(char*)dn.c_str(), LDAP_SCOPE_BASE,
			(char*)search_filter.c_str(),
			 request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	switch(ldap_count_entries(m_ldap, res)) {
	case 0:
	    ldap_msgfree(res);
		throw objectnotfound(dn);
	case 1:
		break;
	default:
	    ldap_msgfree(res);
		throw toomanyobjects(string("More than one object returned in search ") + dn);
	}

	entry = ldap_first_entry(m_ldap, res);
	if(entry == NULL) {
	    ldap_msgfree(res);
		throw runtime_error("ldap_dn: group broken.");
	}

	for (att = ldap_first_attribute(m_ldap, entry, &ber);
		 att != NULL;
		 att = ldap_next_attribute(m_ldap, entry, ber))
	{
		if (stricmp(att, unique_attr) == 0) {
			id = getLDAPAttributeValue(att, entry);
			bUniqueAttrFound = true;
		}
		if (stricmp(att, modify_attr) == 0) {
		    signature = getLDAPAttributeValue(att, entry);
		}
		ldap_memfree(att);
	}

	if(ber)
		ber_free(ber, 0);

	if(res)
		ldap_msgfree(res);

	if(bUniqueAttrFound == false)
		throw runtime_error("Unique id not found: " + dn);

	return objectsignature_t(id, signature);
}

// lpAttrData must be in ldap_server_charset
objectsignature_t LDAPUserPlugin::resolveObjectFromAttribute(userobject_type_t type, string company,
															 const char *lpAttrData, const char* lpAttr) throw(std::exception)
{
	string			ldap_filter;
	LDAPMessage*	res = NULL;
	int				rc;
	objectid_t		id;
	std::string		signature;
	char*			att = NULL;
	BerElement*		ber = NULL;
	LDAPMessage*	entry = NULL;
	ULONG			ulScope;

	if(lpAttr == NULL || lpAttrData == NULL)
		throw runtime_error("ldap: missing or wrong ldap setting");

	string ldap_basedn;
	char *unique_attr;
	char *modify_attr = m_config->GetSetting("ldap_last_modification_attribute");

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));
		break;
	case USEROBJECT_TYPE_GROUP:
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		break;
	case USEROBJECT_TYPE_COMPANY:
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		break;
	default:
		throw runtime_error(string("resolveObjectFromAttribute: request for unknown object type"));
	}

	char *request_attrs[] = { unique_attr,
                              modify_attr,
							  NULL };

	ldap_basedn = getSearchBase(type, company);
	ldap_filter = getObjectSearchFilter(type, lpAttrData, lpAttr);

	if (ldap_filter.empty())
		throw objectnotfound(string(lpAttrData) + "=" + string(lpAttr));

	if ((rc = my_ldap_search_s(
			 (char *)ldap_basedn.c_str(), ulScope,
			 (char *)ldap_filter.c_str(),
			 request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc) + " " + ldap_filter);
	}

	switch(ldap_count_entries(m_ldap, res)) {
	case 0:
    	ldap_msgfree(res);
		throw objectnotfound(ldap_filter);
	case 1:
		break;
	default:
    	ldap_msgfree(res);
		throw toomanyobjects(string("More than one object returned in search ") + ldap_filter);
	}

	entry = ldap_first_entry(m_ldap, res);

	for (att = ldap_first_attribute(m_ldap, entry, &ber);
		 att != NULL;
		 att = ldap_next_attribute(m_ldap, entry, ber))
	{
		if (stricmp(att, unique_attr) == 0) {
			id = getLDAPAttributeValue(att, entry);
		}
		if (stricmp(att, modify_attr) == 0) {
		    signature = getLDAPAttributeValue(att, entry);
        }
		ldap_memfree(att);
	}

	if(ber)
		ber_free(ber, 0);

	if(res)
		ldap_msgfree(res);

	if(id.empty())
		throw runtime_error("ldap: missing or wrong unique_attr");

	return objectsignature_t(id, signature);
}

list<objectsignature_t> LDAPUserPlugin::resolveObjectsFromAttribute(userobject_type_t type, list<string> objects, const char* lpAttr) throw (std::exception)
{
	list<objectsignature_t> signatures;
	LDAPMessage *res = NULL;
	int rc;

	string ldap_basedn;
	string search_filter;
	char *unique_attr;
	char *modify_attr = m_config->GetSetting("ldap_last_modification_attribute");
	ULONG ulScope;

	LDAPMessage *entry = NULL;
	BerElement *ber = NULL;
	char *att = NULL;

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));
		break;
	case USEROBJECT_TYPE_GROUP:
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		break;
	case USEROBJECT_TYPE_COMPANY:
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
		unique_attr = m_config->GetSetting("ldap_addresslist_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_addresslist_scope"));
		break;
	default:
		throw runtime_error("Object is wrong type");
		break;
	}

	char *request_attrs[] = {
		unique_attr,
		modify_attr,
		NULL
	};

	ldap_basedn = getSearchBase(type);
	search_filter = getSearchFilter(type);

	/* No searchfilter: none of the objects can be resolved */
	if (search_filter.empty())
		goto exit;

	search_filter = "(&" + search_filter + "(|";
	for (list<string>::iterator si = objects.begin(); si != objects.end(); si++)
		search_filter += "(" + string(lpAttr) + "=" + StringEscapeSequence(*si) + ")";
	search_filter += "))";

	if ((rc = my_ldap_search_s(
			(char *)ldap_basedn.c_str(), ulScope,
			(char *)search_filter.c_str(), request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	for (entry = ldap_first_entry(m_ldap, res);
		 entry != NULL;
		 entry = ldap_next_entry(m_ldap, entry)) {

		objectid_t objectid;
		string signature;

		for (att = ldap_first_attribute(m_ldap, entry, &ber);
			 att != NULL;
			 att = ldap_next_attribute(m_ldap, entry, ber)) {

			if(stricmp(att, unique_attr) == 0)
				objectid = getLDAPAttributeValue(att, entry);
			if(stricmp(att, modify_attr) == 0)
				signature = getLDAPAttributeValue(att, entry);

			ldap_memfree(att);
		}

		signatures.push_back(objectsignature_t(objectid, signature));

		if(ber)
			ber_free(ber, 0);
	}

exit:
	if(res)
		ldap_msgfree(res);

	return signatures;
}

list<objectsignature_t> LDAPUserPlugin::resolveObjectsFromAttributeType(userobject_type_t type, list<string> &objects, const char* lpAttr, const char* lpAttrType) throw (std::exception)
{
	list<objectsignature_t> members;

	/* When the relation attribute the is the DN, we cannot perform any optimizations
	 * and we must incur the penalty of having to resolve each entry one by one.
	 * When the relation attribute is not the DN, we can optimize the lookup
	 * by creating a single query that obtains all the required data in a single query. */
	if (lpAttrType && stricmp(lpAttrType, LDAP_DATA_TYPE_DN) == 0) {
		for (list<string>::iterator si = objects.begin(); si != objects.end(); si++) {
			try {
				members.push_back(objectDNtoObjectSignature(type, *si));
			} catch (objectnotfound &e) {
				// resolve failed, drop entry
				continue;
			} catch (std::exception &e) {
				// query failed, drop entry
				continue;
			}
		}
	} else {
		/* We have the full member list, create a new query that
		 * will request the unique modification attributes for all
		 * members in a single shot. With this data we can construct
		 * the list of object signatures */
		try {
			members = resolveObjectsFromAttribute(type, objects, lpAttr);
		} catch (objectnotfound &e) {
			// resolve failed, drop entry
			goto exit;
		} catch (std::exception &e) {
			// query failed, drop entry
			goto exit;
		}
	}

exit:
	return members;
}

// username is in windows-1252
objectsignature_t LDAPUserPlugin::resolveName(userobject_type_t type, const string &name, const string &company) throw(std::exception)
{
	char *attr = NULL;

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		attr = m_config->GetSetting("ldap_loginname_attribute");
		break;
	case USEROBJECT_TYPE_GROUP:
		attr = m_config->GetSetting("ldap_groupname_attribute");
		break;
	case USEROBJECT_TYPE_COMPANY:
		attr = m_config->GetSetting("ldap_companyname_attribute");
		break;
	default:
		throw runtime_error(string("resolveName: request for unknown object type"));
	}

	return resolveObjectFromAttribute(type, company, m_iconvrev->convert(name).c_str(), attr);
}

// username is in windows-1252
// password is in windows-1252
objectsignature_t LDAPUserPlugin::authenticateUser(const string &username, const string &companyname, const string &password) throw(std::exception)
{
	struct timeval tstart, tend;
	char *authmethod = m_config->GetSetting("ldap_authentication_method");
	objectsignature_t id;
	LONGLONG	llelapsedtime;
	
	gettimeofday(&tstart, NULL);

	try {
		if (!stricmp(authmethod, "password")) {
			id = authenticateUserPassword(username, companyname, password);
		} else {
			id = authenticateUserBind(username, companyname, password);
		}
	} catch (...) {
		m_lpStatsCollector->Increment(SCN_LDAP_AUTH_DENIED);
		throw;
	}

	gettimeofday(&tend, NULL);
	llelapsedtime = difftimeval(&tstart,&tend);

	m_lpStatsCollector->Increment(SCN_LDAP_AUTH_LOGINS);
	m_lpStatsCollector->Increment(SCN_LDAP_AUTH_TIME, llelapsedtime);
	m_lpStatsCollector->Max(SCN_LDAP_AUTH_TIME_MAX, llelapsedtime);

	return id;
}

// username is in windows-1252
// password is in windows-1252
objectsignature_t LDAPUserPlugin::authenticateUserBind(const string &username, const string &companyname, const string &password) throw(std::exception) {
	LDAP*		ld = NULL;
	string		dn;
	objectsignature_t	id;

	try {
		id = resolveName(USEROBJECT_TYPE_USER, username, companyname);
		dn = objectUniqueIDtoObjectDN(USEROBJECT_TYPE_USER, id.id);

		ld = ConnectLDAP(dn.c_str(), m_iconvrev->convert(password).c_str());
	} catch (exception &e) {
		throw login_error((string)"Trying to authenticate failed: " + e.what() + (string)"; username = " + username);
	}

	if(ld == NULL) {
		throw runtime_error("Trying to authenticate failed: connection failed");
	}

	ldap_unbind_s(ld);

	return id;
}

// username is in windows-1252
// password is in windows-1252
objectsignature_t LDAPUserPlugin::authenticateUserPassword(const string &username, const string &companyname, const string &password) throw(std::exception) {
	LDAPMessage*	res = NULL;
	int				rc;
	BerElement*		ber = NULL;
	char*			att = NULL;
	objectdetails_t	d;
	LDAPMessage*	entry = NULL;
	ULONG			ulScope;
	string			ldap_basedn;
	string			ldap_filter;
	string			strCryptedpw;
	string			strPasswordConverted;
	objectsignature_t signature;

	char *loginname_attr = m_config->GetSetting("ldap_loginname_attribute");
	char *password_attr = m_config->GetSetting("ldap_password_attribute");
	char *unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
	char *request_attrs[] = { loginname_attr,
							  password_attr,
							  unique_attr,
							  NULL };

	ldap_basedn = getSearchBase(USEROBJECT_TYPE_USER, companyname);
	ldap_filter = getObjectSearchFilter(USEROBJECT_TYPE_USER, m_iconvrev->convert(username), loginname_attr);
	ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));

	/* LDAP filter does not exist, user does not exist, user cannot login */
	if (ldap_filter.empty())
		throw objectnotfound(username);

	if ((rc = my_ldap_search_s(
			 (char *)ldap_basedn.c_str(), ulScope,
			 (char *)ldap_filter.c_str(),
			 request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	switch(ldap_count_entries(m_ldap, res)) {
	case 0:
	    ldap_msgfree(res);
		throw login_error("Trying to authenticate failed: wrong username or password");
	case 1:
		break;
	default:
	    ldap_msgfree(res);
		throw login_error("Trying to authenticate failed: unknown error");
	}

	entry = ldap_first_entry(m_ldap, res);

	for (att = ldap_first_attribute(m_ldap, entry, &ber);
		 att != NULL;
		 att = ldap_next_attribute(m_ldap, entry, ber)) {
		if (loginname_attr && !stricmp(att, loginname_attr)) {
			d.SetPropString(OB_PROP_S_LOGIN, m_iconv->convert(getLDAPAttributeValue(att, entry)));
		} else if (password_attr && !stricmp(att, password_attr)) {
			d.SetPropString(OB_PROP_S_PASSWORD, getLDAPAttributeValue(att, entry));
		}

		if (unique_attr && !stricmp(att, unique_attr)) {
			signature.id = getLDAPAttributeValue(att, entry);
		}
		ldap_memfree(att);
	}

	if(ber)
		ber_free(ber, 0);

	if (res)
		ldap_msgfree(res);

	strCryptedpw = d.GetPropString(OB_PROP_S_PASSWORD);
	
	strPasswordConverted = m_iconvrev->convert(password).c_str();

	if (strCryptedpw.empty() || signature.id.empty()) {
		throw login_error("Trying to authenticate failed: Password field or/and uniqueid are empty or unreadable");
	}

	if (!strncmp("{CRYPT}", strCryptedpw.c_str(), 7)) {
		if(checkPassword(PASSWORD_CRYPT, strPasswordConverted.c_str(), &(strCryptedpw.c_str()[7])) != 0)
			throw login_error("Trying to authenticate failed: wrong username or password");
	} else if (!strncmp("{MD5}", strCryptedpw.c_str(), 5)) {
		if(checkPassword(PASSWORD_MD5, strPasswordConverted.c_str(), &(strCryptedpw.c_str()[5])) != 0)
			throw login_error("Trying to authenticate failed: wrong username or password");
	} else if (!strncmp("{SMD5}", strCryptedpw.c_str(), 6)) {
		if(checkPassword(PASSWORD_SMD5, strPasswordConverted.c_str(), &(strCryptedpw.c_str()[6])) != 0)
			throw login_error("Trying to authenticate failed: wrong username or password");
	} else if (!strncmp("{SSHA}", strCryptedpw.c_str(), 6)) {
		if(checkPassword(PASSWORD_SSHA, strPasswordConverted.c_str(), &(strCryptedpw.c_str()[6])) != 0)
			throw login_error("Trying to authenticate failed: wrong username or password");
	} else if(!strncmp("{SHA}", strCryptedpw.c_str(), 5) ||
			!strncmp("{MD5CRYPT}", strCryptedpw.c_str(), 10) )
	{
		throw login_error("Trying to authenticate failed: unsupported encryption scheme");
	}else {
		if(strcmp(strCryptedpw.c_str(), strPasswordConverted.c_str()) != 0) { //Plain password
			throw login_error("Trying to authenticate failed: wrong username or password");
		}
	}

	return signature;
}

list<objectsignature_t> LDAPUserPlugin::getAllObjects(userobject_type_t type, const objectid_t &companyid) throw(exception) {
	return getAllObjectsByFilter(type, companyid, "");
}

list<objectsignature_t> LDAPUserPlugin::getAllObjectsByFilter(userobject_type_t type, const objectid_t &companyid, const std::string &additional_search_filter) throw(exception)
{
	list<objectsignature_t>	l;
	auto_ptr<dn_cache_t> lpCache = auto_ptr<dn_cache_t>(new dn_cache_t());
	LDAPMessage*	res = NULL;
	int				rc;
	LDAPMessage*	entry = NULL;
    BerElement*		ber = NULL; 
    char*			att = NULL;
	ULONG			ulScope;
	std::string		strFilter;
	std::string		strDN;
	char*			dn = NULL;

	string ldap_basedn;
	std::string search_filter = getSearchFilter(type);
	char *unique_attr;
	char *name_attr;
	char *modify_attr = m_config->GetSetting("ldap_last_modification_attribute");

	auto_ptr<dn_list_t> dnFilter;

	/*
	 * When working in multi-company mode we need to determine if the found object
	 * is really a member of the requested company and not turned up in the
	 * result because he is member of a subcompany.
	 * Create a filter by requesting all subcompanies for he given company,
	 * and remove any returned objects which are member of these subcompanies.
	 */
	if (m_bHosted && type != USEROBJECT_TYPE_COMPANY) {
		std::auto_ptr<dn_cache_t> lpCache = m_lpCache->getObjectDNCache(this, USEROBJECT_TYPE_COMPANY);
		dnFilter = m_lpCache->GetChildrenForObject(lpCache, companyid);
	}

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		name_attr = m_config->GetSetting("ldap_loginname_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));
		break;
    case USEROBJECT_TYPE_GROUP:
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		name_attr = m_config->GetSetting("ldap_groupname_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		break;
	case USEROBJECT_TYPE_COMPANY:
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		name_attr = m_config->GetSetting("ldap_companyname_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
		unique_attr = m_config->GetSetting("ldap_addresslist_unique_attribute");
		name_attr = m_config->GetSetting("ldap_addresslist_name_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_addresslist_scope"));
		break;
	default:
		throw runtime_error(string("getAllObjects: request for unknown object type"));
	}

   	char *request_attrs[] = {
		unique_attr,
		name_attr,
		modify_attr,
		NULL
	};
	
	ldap_basedn = getSearchBase(type, companyid);

	/* No search filter! we can't execute the query because this would return _all_ objects */
	if (search_filter.empty())
		goto exit;

    /* Do an additional search filter if required */
    if (!additional_search_filter.empty()) {
        search_filter = "(&" + search_filter + additional_search_filter + ")";
    }

	if ((rc = my_ldap_search_s(
			 (char*)ldap_basedn.c_str(), ulScope,
			 (char*)search_filter.c_str(), request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("getAllObjects: ldap_search_s: ") + ldap_err2string(rc));
	}

	// Get the unique id for each of the returned entries
	for (entry = ldap_first_entry(m_ldap, res);
		 entry != NULL;
		 entry = ldap_next_entry(m_ldap, entry)) {

		objectid_t objectid;
		std::string name;
		std::string signature;

		dn = ldap_get_dn(m_ldap, entry);
		if(dn) {
			strDN = dn;
			ldap_memfree(dn);
		}

		/* Make sure the DN isn't filtered */
		if (m_bHosted && type != USEROBJECT_TYPE_COMPANY) {
			if (m_lpCache->isDNInList(dnFilter, strDN))
				continue;
		}

		for (att = ldap_first_attribute(m_ldap, entry, &ber);
			 att != NULL;
			 att = ldap_next_attribute(m_ldap, entry, ber)) {

            if(stricmp(att, unique_attr) == 0)
                objectid = getLDAPAttributeValue(att, entry);
            if(stricmp(att, modify_attr) == 0)
                signature = getLDAPAttributeValue(att, entry);
			if(stricmp(att, name_attr) == 0)
				name = getLDAPAttributeValue(att, entry);

            ldap_memfree(att);
        }
		
        if(ber)
            ber_free(ber, 0);

		if (objectid.empty()) {
			throw runtime_error(string("getAllObjects: Object with a empty unique attribute. object dn=") + strDN );
		}
		
		l.push_back(objectsignature_t(objectid, signature));
		lpCache->insert(dn_cache_t::value_type(strDN, cache_entry_t(objectid, name)));
	}

	/* Update cache */
	m_lpCache->setObjectDNCache(type, lpCache);

exit:
	if(res)
		ldap_msgfree(res);

	return l;
}


string LDAPUserPlugin::getLDAPAttributeValue(char *attribute, LDAPMessage *entry) {
	list<string> l = getLDAPAttributeValues(attribute, entry);
	if (!l.empty())
		return *(l.begin());
	else
		return string();
}


list<string> LDAPUserPlugin::getLDAPAttributeValues(char *attribute, LDAPMessage *entry) {
	list<string> r;
	string s;
	struct berval** berval;

	berval = ldap_get_values_len(m_ldap, entry, attribute);

	if (berval != NULL) {
		for (int i = 0; berval[i] != NULL; i++) {
			s.assign(berval[i]->bv_val, berval[i]->bv_len);
			r.push_back(s);
		}

		ldap_value_free_len(berval);
	}
	return r;
}

// Returns a whole set of objectdetails. User may be missing if the user details
// cannot be retrieved for some reason.
auto_ptr<map<objectid_t, objectdetails_t> > LDAPUserPlugin::getObjectDetails(userobject_type_t type, list<objectid_t> &objectids) throw(std::exception)
{
    list<objectid_t>::iterator iterId;
    string filter;
	std::auto_ptr<dn_cache_t> lpCache;
    auto_ptr<map<objectid_t, objectdetails_t> > mapdetails = auto_ptr<map<objectid_t, objectdetails_t> >(new map<objectid_t, objectdetails_t>);
    unsigned int rc;
	LDAPMessage *res = NULL;
	std::string uniqueIDEscaped;
	objectid_t id;

	string ldap_filter;
	string ldap_basedn;
	unsigned int ulScope = 0;
	char *unique_attr = NULL;
	char *unique_attr_type = NULL;
	char *sysadmin_attr = NULL;
	char *sysadmin_attr_type = NULL;
	char *sysadmin_attr_rel = NULL;

	char *loginname_attr = NULL;
	char *fullname_attr = NULL;
	char *password_attr = NULL;
	char *emailaddress_attr = NULL;
	char *isadmin_attr = NULL;
	char *nonactive_attr = NULL;
	char *groupname_attr = NULL;
	char *companyname_attr = NULL;
	char *usercert_attr = NULL;
	char *sendas_attr = NULL;
	char *exchangedn_attr = NULL;
	char *addresslistname_attr = NULL;
	char *location_attr = NULL;
	char *telephone_attr = NULL;
	char *department_attr = NULL;
	char *fax_attr = NULL;

	char *request_attrs_user[16];
	char *request_attrs_group[3];
	char *request_attrs_company[6];
	char *request_attrs_addresslist[3];
	char **request_attrs;
	unsigned int attrs = 0;
	unsigned int ulCutoff = atoui(m_config->GetSetting("ldap_filter_cutoff_elements"));
	
	/*
	 * When working in multi-company mode we need to determine to which company
	 * this object belongs. To do this efficiently we are using the cache for
	 * resolving the company based on the DN.
	 */
	if (m_bHosted && type != USEROBJECT_TYPE_COMPANY)
		lpCache = m_lpCache->getObjectDNCache(this, USEROBJECT_TYPE_COMPANY);

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		department_attr = m_config->GetSetting("ldap_user_department_attribute");
		telephone_attr = m_config->GetSetting("ldap_user_telephone_attribute");
		location_attr = m_config->GetSetting("ldap_user_location_attribute");
		fax_attr = m_config->GetSetting("ldap_user_fax_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));
		unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_user_unique_attribute_type");
		loginname_attr = m_config->GetSetting("ldap_loginname_attribute");
		fullname_attr = m_config->GetSetting("ldap_fullname_attribute");
		password_attr = m_config->GetSetting("ldap_password_attribute");
		emailaddress_attr = m_config->GetSetting("ldap_emailaddress_attribute");
		isadmin_attr = m_config->GetSetting("ldap_isadmin_attribute");
		nonactive_attr = m_config->GetSetting("ldap_nonactive_attribute");
		usercert_attr = m_config->GetSetting("ldap_user_certificate_attribute");
		sendas_attr = m_config->GetSetting("ldap_user_sendas_attribute");
		exchangedn_attr = m_config->GetSetting("ldap_user_exchange_dn_attribute", "", NULL);

		if (loginname_attr)
			request_attrs_user[attrs++] = loginname_attr;
		if (fullname_attr)
			request_attrs_user[attrs++] = fullname_attr;
		if (password_attr)
			request_attrs_user[attrs++] = password_attr;
		if (emailaddress_attr)
			request_attrs_user[attrs++] = emailaddress_attr;
		if (isadmin_attr)
			request_attrs_user[attrs++] = isadmin_attr;
		if (telephone_attr)
			request_attrs_user[attrs++] = telephone_attr;
		if (department_attr)
			request_attrs_user[attrs++] = department_attr;
		if (location_attr)
			request_attrs_user[attrs++] = location_attr;
		if (fax_attr)
			request_attrs_user[attrs++] = fax_attr;
		if (nonactive_attr)
			request_attrs_user[attrs++] = nonactive_attr;
		if (unique_attr)
			request_attrs_user[attrs++] = unique_attr;
		if (usercert_attr)
			request_attrs_user[attrs++] = usercert_attr;
		if (sendas_attr)
			request_attrs_user[attrs++] = sendas_attr;
		if (exchangedn_attr)
			request_attrs_user[attrs++] = exchangedn_attr;
		request_attrs_user[attrs] = NULL;

		ASSERT(attrs <= (sizeof(request_attrs_user)/sizeof(request_attrs_user[0])));
		request_attrs = request_attrs_user;
		break;
	case USEROBJECT_TYPE_GROUP:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_group_unique_attribute_type");
		groupname_attr = m_config->GetSetting("ldap_groupname_attribute");

		if (groupname_attr)
			request_attrs_group[attrs++] = groupname_attr;
		if (unique_attr)
			request_attrs_group[attrs++] = unique_attr;
		request_attrs_group[attrs] = NULL; 

		ASSERT(attrs <= (sizeof(request_attrs_group)/sizeof(request_attrs_group[0])));
		request_attrs = request_attrs_group;
		break;
	case USEROBJECT_TYPE_ADDRESSLIST:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_addresslist_scope"));
		unique_attr = m_config->GetSetting("ldap_addresslist_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_addresslist_unique_attribute_type");
		addresslistname_attr = m_config->GetSetting("ldap_addresslist_name_attribute");

		if (addresslistname_attr)
			request_attrs_addresslist[attrs++] = addresslistname_attr;
		if (unique_attr)
			request_attrs_addresslist[attrs++] = unique_attr;
		request_attrs_addresslist[attrs] = NULL; 

		ASSERT(attrs <= (sizeof(request_attrs_addresslist)/sizeof(request_attrs_addresslist[0])));
		request_attrs = request_attrs_addresslist;
		break;
	case USEROBJECT_TYPE_COMPANY:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_company_unique_attribute_type");
		companyname_attr = m_config->GetSetting("ldap_companyname_attribute");
		sysadmin_attr = m_config->GetSetting("ldap_company_system_admin_attribute");
		sysadmin_attr_type = m_config->GetSetting("ldap_company_system_admin_attribute_type");
		sysadmin_attr_rel = m_config->GetSetting("ldap_company_system_admin_relation_attribute");

		if (companyname_attr)
			request_attrs_company[attrs++] = companyname_attr;
		if (sysadmin_attr)
			request_attrs_company[attrs++] = sysadmin_attr;
		if (sysadmin_attr_type)
			request_attrs_company[attrs++] = sysadmin_attr_type;
		if (unique_attr)
			request_attrs_company[attrs++] = unique_attr;
		request_attrs_company[attrs] = NULL;

		ASSERT(attrs <= (sizeof(request_attrs_company)/sizeof(request_attrs_company[0])));
		request_attrs = request_attrs_company;
		break;
	default:
		throw runtime_error("Object is wrong type");
	}

	if(objectids.empty())
        return mapdetails;

	ldap_basedn = getSearchBase(type);
	ldap_filter = getSearchFilter(type);

	/* No LDAP filter, just assume the object does not exist. */
	if (ldap_filter.empty())
		throw objectnotfound("userobject_type " + stringify(type));

    /*
     * This is purely an optimization: to make sure we don't send huge 5000-item filters
     * to the LDAP Server (which may be slow), we have a cutoff point at which we simply
     * retrieve all LDAP data. This is sometimes (depending on the type of LDAP Server)
     * much, much faster (factor of 40 has been seen), with the minor expense of receiving
     * more data than actually needed. This means we're trading minor local processing time
     * and network activity for a much lower LDAP-server processing overhead.
     */
    
    if (ulCutoff != 0 && objectids.size() >= ulCutoff) {
        filter = ldap_filter;
    } else {
        filter = "(&" + ldap_filter + "(|";

        for(iterId = objectids.begin(); iterId != objectids.end(); iterId++) {
            filter += "(";

            if(unique_attr_type && stricmp(unique_attr_type, LDAP_DATA_TYPE_BINARY) == 0) {
                // Set binary uniqueid to escaped string
                BintoEscapeSequence(iterId->c_str(), iterId->size(), &uniqueIDEscaped);
            }else {
                uniqueIDEscaped = StringEscapeSequence((*iterId));
            }

            filter += unique_attr;
            filter += "=";
            filter += uniqueIDEscaped;
            filter += ")";
        }
        filter += "))";
    }
    
	if ((rc = my_ldap_search_s(
			 (char*)ldap_basedn.c_str(), ulScope,
			 (char*)filter.c_str(), request_attrs,
			 FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	LDAPMessage *entry;

	for (entry = ldap_first_entry(m_ldap, res);
		 entry != NULL;
		 entry = ldap_next_entry(m_ldap, entry)) {

    	objectdetails_t d(type);

		if (m_bHosted && type != USEROBJECT_TYPE_COMPANY) {
			/* OB_PROP_S_COMPANYNAME is only required for non-company userobjects */
			char *dn = ldap_get_dn(m_ldap, entry);
			d.SetPropString(OB_PROP_S_COMPANYNAME, m_lpCache->getParentForDN(lpCache, dn).name);
			ldap_memfree(dn);
		}

		if (type == USEROBJECT_TYPE_COMPANY) {
			/* By default the sysadmin is SYSTEM, will be overwritten later */
			d.SetPropString(OB_PROP_S_SYSADMIN, "SYSTEM");
		}

		BerElement *ber;
		char *att;
		for (att = ldap_first_attribute(m_ldap, entry, &ber);
			 att != NULL;
			 att = ldap_next_attribute(m_ldap, entry, ber)) {

			/*
			 * Don't put loginname in the if-else-if statement series.
			 * It is possible (and valid) for the loginname_attr to be equal
			 * to a different attribute like emailaddress_attr.
			 */
			if (loginname_attr && !stricmp(att, loginname_attr))
				d.SetPropString(OB_PROP_S_LOGIN, m_iconv->convert(getLDAPAttributeValue(att, entry)));

			if (fullname_attr && !stricmp(att, fullname_attr)) {
				d.SetPropString(OB_PROP_S_FULLNAME, m_iconv->convert(getLDAPAttributeValue(att, entry)));
			}
			if (password_attr && !stricmp(att, password_attr)) {
				d.SetPropString(OB_PROP_S_PASSWORD, getLDAPAttributeValue(att, entry));
			}
			if (emailaddress_attr && !stricmp(att, emailaddress_attr)) {
				d.SetPropString(OB_PROP_S_EMAIL, m_iconv->convert(getLDAPAttributeValue(att, entry)));
			}
			if (isadmin_attr && !stricmp(att, isadmin_attr)) {
				d.SetPropInt(OB_PROP_I_ADMINLEVEL, atoi(getLDAPAttributeValue(att, entry).c_str()));
			}
			if (nonactive_attr && !stricmp(att, nonactive_attr)) {
			    d.SetPropBool(OB_PROP_B_NONACTIVE, parseBool(getLDAPAttributeValue(att, entry)));
            }
			if (usercert_attr && !stricmp(att, usercert_attr)) {
				list<string> certs = getLDAPAttributeValues(att, entry);
				list<string>::iterator i;
				for (i = certs.begin(); i != certs.end(); i++)
					d.AddPropString(OB_PROP_L_CERTIFICATE, *i);
            }
			if (sendas_attr && !stricmp(att, sendas_attr)) {
				list<string> saUsers = getLDAPAttributeValues(att, entry);
				list<string>::iterator i;
				char *relAttr = m_config->GetSetting("ldap_user_sendas_relation_attribute");
				char *relAttrType = m_config->GetSetting("ldap_user_sendas_attribute_type");
				char *uniqueAttr = m_config->GetSetting("ldap_user_unique_attribute");

				if (!uniqueAttr || uniqueAttr[0] == '\0')
					uniqueAttr = "cn"; // default

				if (!relAttrType || relAttrType[0] == '\0')
					relAttrType = LDAP_DATA_TYPE_TEXT;

				if (!relAttr || relAttr[0] == '\0')
					relAttr = uniqueAttr;

				if (strcmp(relAttr, uniqueAttr) == 0) {
					for (i = saUsers.begin(); i != saUsers.end(); i++) {
						d.AddPropString(OB_PROP_L_SENDAS, *i);
					}
				} else {
					// lookup user when ldap_user_sendas_relation_attribute != ldap_user_unique_attribute
					list<objectsignature_t> lstSenders;
					list<objectsignature_t>::iterator iSenders;
					try {
						lstSenders = resolveObjectsFromAttributeType(USEROBJECT_TYPE_USER, saUsers, relAttr, relAttrType);
					}
					catch(...) { }
					for (iSenders = lstSenders.begin(); iSenders != lstSenders.end(); iSenders++) {
						d.AddPropString(OB_PROP_L_SENDAS, iSenders->id);
					}
					try {
						lstSenders = resolveObjectsFromAttributeType(USEROBJECT_TYPE_NONACTIVE, saUsers, relAttr, relAttrType);
					}
					catch(...) { }
					for (iSenders = lstSenders.begin(); iSenders != lstSenders.end(); iSenders++) {
						d.AddPropString(OB_PROP_L_SENDAS, iSenders->id);
					}
				}
            }
			if (groupname_attr && !stricmp(att, groupname_attr)) {
				string groupname = m_iconv->convert(getLDAPAttributeValue(att, entry));
				d.SetPropString(OB_PROP_S_LOGIN, groupname);
				d.SetPropString(OB_PROP_S_FULLNAME, groupname);
			}
			if (companyname_attr && !stricmp(att, companyname_attr)) {
				string companyname = m_iconv->convert(getLDAPAttributeValue(att, entry));
				d.SetPropString(OB_PROP_S_LOGIN, companyname);
				d.SetPropString(OB_PROP_S_FULLNAME, companyname);
				d.SetPropString(OB_PROP_S_COMPANYNAME, companyname);
			}
			if (sysadmin_attr && !stricmp(att, sysadmin_attr)) {
				/* Resolve sysadmin to unique id for user, this allows for a nice
				 * optimization: When the relation attribute is the unique id, we
				 * don't need to resolve anything... */
				string sysadmin = getLDAPAttributeValue(att, entry);
				char *user_unique_attr = m_config->GetSetting("ldap_user_unique_attribute");

				if (user_unique_attr && sysadmin_attr_rel && stricmp(user_unique_attr, sysadmin_attr_rel) == 0) {
					d.SetPropString(OB_PROP_S_SYSADMIN, sysadmin);
				} else if (sysadmin_attr_type && stricmp(sysadmin_attr_type, LDAP_DATA_TYPE_DN) == 0) {
					try {
						d.SetPropString(OB_PROP_S_SYSADMIN, objectDNtoObjectSignature(USEROBJECT_TYPE_USER, sysadmin).id);
					} catch (exception &e) {
						/* Ignore, keep administrator field empty */
					}
				} else {
					try {
						d.SetPropString(OB_PROP_S_SYSADMIN,
										objectUniqueIDtoAttributeData(USEROBJECT_TYPE_USER,
																	  resolveObjectFromAttribute(USEROBJECT_TYPE_USER, "",
																	  							 sysadmin.c_str(),
																								 sysadmin_attr_rel).id,
																	  user_unique_attr));
					} catch (exception &e) {
						/* Ignore, keep administrator field empty */
					}
				}
			}
			if (exchangedn_attr && !stricmp(att, exchangedn_attr)) {
				string exchangedn = m_iconv->convert(getLDAPAttributeValue(att, entry));
				d.SetPropString(OB_PROP_S_EXCH_DN, exchangedn);
			}
			if (addresslistname_attr && !stricmp(att, addresslistname_attr)) {
				string addresslistname = m_iconv->convert(getLDAPAttributeValue(att, entry));
				d.SetPropString(OB_PROP_S_LOGIN, addresslistname);
				d.SetPropString(OB_PROP_S_FULLNAME, addresslistname);
			}
			if (department_attr && !stricmp(att, department_attr)) {
			    d.SetPropString(OB_PROP_S_DEPARTMENT, m_iconv->convert(getLDAPAttributeValue(att, entry)));
            }
			if (telephone_attr && !stricmp(att, telephone_attr)) {
			    d.SetPropString(OB_PROP_S_TELEPHONE, m_iconv->convert(getLDAPAttributeValue(att, entry)));
            }
			if (location_attr && !stricmp(att, location_attr)) {
			    d.SetPropString(OB_PROP_S_LOCATION, m_iconv->convert(getLDAPAttributeValue(att, entry)));
            }
			if (fax_attr && !stricmp(att, fax_attr)) {
			    d.SetPropString(OB_PROP_S_FAX, m_iconv->convert(getLDAPAttributeValue(att, entry)));
            }

            // not else-if because the unique attribute could be the same as another field
            if (unique_attr && !stricmp(att, unique_attr)) {
                id = getLDAPAttributeValue(att, entry);
            }
			ldap_memfree(att);
		}
		ber_free(ber, 0);

		if(id != "")
		    (*mapdetails)[id] = d;
	}

	if (res)
		ldap_msgfree(res);

	return mapdetails;
}

auto_ptr<objectdetails_t> LDAPUserPlugin::getObjectDetails(userobject_type_t type, const objectid_t &id) throw(exception) {
    auto_ptr<map<objectid_t, objectdetails_t> > mapDetails;
    list<objectid_t> objectids;
    
    objectids.push_back(id);
    
    mapDetails = getObjectDetails(type, objectids);
    
    if(mapDetails->size() == 0)
        throw objectnotfound(id);
        
    return auto_ptr<objectdetails_t>(new objectdetails_t((*mapDetails)[id]));
}

static LDAPMod *newLDAPModification(char *attribute, const list<string> &values) {
	LDAPMod *mod = (LDAPMod*) calloc(1, sizeof(LDAPMod));

	// The only type of modification allowed here is replace.  It
	// should be enough for our needs, but if an addition or removal
	// is needed, this value has to be changed.
	mod->mod_op = LDAP_MOD_REPLACE;
	mod->mod_type = attribute;
	mod->mod_vals.modv_strvals = (char**) calloc(values.size() + 1, sizeof(char*));
	int idx = 0;
	for (list<string>::const_iterator i = values.begin(); i != values.end(); i++) {
		// A strdup is necessary to be able to call free on it in the
		// method changeAttribute, below.
		mod->mod_vals.modv_strvals[idx++] = strdup((*i).c_str());
	}
	mod->mod_vals.modv_strvals[idx] = NULL;
	return mod;
}

static LDAPMod *newLDAPModification(char *attribute, const char *value) {
	// Pretty lame function, all it does is use newLDAPModification
	// with a list with only one element.
	list<string> values;
	values.push_back(value);
	return newLDAPModification(attribute, values);
}


/** Change one attribute for this DN
 *
 */
int LDAPUserPlugin::changeAttribute(const char *dn, char *attribute, const char *value) {
	LDAPMod *mods[2];

	mods[0] = newLDAPModification(attribute, value);
	mods[1] = NULL;

	// Actual LDAP call
	int rc;
	if ((rc = ldap_modify_s(m_ldap, (char *)dn, mods)) != LDAP_SUCCESS) {
		return 1;
	}

	// Free all calloced memory
	free(mods[0]->mod_vals.modv_strvals[0]);
	free(mods[0]->mod_vals.modv_strvals);
	free(mods[0]);

	return 0;
}

int LDAPUserPlugin::changeAttribute(const char *dn, char *attribute, const list<string> &values) {
	LDAPMod *mods[2];

	mods[0] = newLDAPModification(attribute, values);
	mods[1] = NULL;

	// Actual LDAP call
	int rc;
	if ((rc = ldap_modify_s(m_ldap, (char *)dn, mods)) != LDAP_SUCCESS) {
		return 1;
	}

	// Free all calloced / strduped memory
	for (int i = 0; mods[0]->mod_vals.modv_strvals[i] != NULL; i++) {
		free(mods[0]->mod_vals.modv_strvals[i]);
	}
	free(mods[0]->mod_vals.modv_strvals);
	free(mods[0]);

	return 0;
}

int LDAPUserPlugin::changeObject(userobject_type_t type, const objectid_t &id, objectdetails_t details, std::list<std::string> *lpDelProps) throw(std::exception) {
	throw notimplemented("Change object is not supported when using the LDAP user plugin.");
}

objectsignature_t LDAPUserPlugin::createObject(userobject_type_t type, objectdetails_t details) throw(std::exception) {
	throw notimplemented("Creating objects is not supported when using the LDAP user plugin.");
}

void LDAPUserPlugin::deleteObject(userobject_type_t type, const objectid_t &id) throw(std::exception) {
	throw notimplemented("Deleting users is not supported when using the LDAP user plugin.");
}

// FIXME:
// speedup: read group info on the user (with ADS you know the groups if you have the user dn)
// Missing one group: with linux you have one group id on the user and other objectid on every possible group
// NOTE:
// this function checks all groups from searchfilter for an existing objectid.
list<objectsignature_t> LDAPUserPlugin::getParentObjectsForObject(userobject_relation_t relation, userobject_type_t childtype, const objectid_t &id, userobject_type_t parenttype) throw(std::exception)
{
	list<objectsignature_t> parents;
	LDAPMessage*	res = NULL;
	LDAPMessage*	entry = NULL;
	BerElement*		ber = NULL;
	char*			att = NULL;
	int				rc;
	string			memberSearchData;
	string			ldap_filter;
	string			search_filter;
	ULONG			ulScope;

	string ldap_basedn;
	char *unique_attr;
	char *member_attr;
	char *member_attr_type;
	char *member_attr_rel;
	char *child_unique_attr;
	char *modify_attr = m_config->GetSetting("ldap_last_modification_attribute");

	switch (relation) {
	case OBJECTRELATION_GROUP_MEMBER:
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		member_attr = m_config->GetSetting("ldap_groupmembers_attribute");
		member_attr_type = m_config->GetSetting("ldap_groupmembers_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_groupmembers_relation_attribute");
		child_unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		break;
	case OBJECTRELATION_COMPANY_VIEW:
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		member_attr = m_config->GetSetting("ldap_company_view_attribute");
		member_attr_type = m_config->GetSetting("ldap_company_view_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_company_view_relation_attribute");
		child_unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		break;
	case OBJECTRELATION_COMPANY_ADMIN:
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		member_attr = m_config->GetSetting("ldap_company_admin_attribute");
		member_attr_type = m_config->GetSetting("ldap_company_admin_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_company_admin_relation_attribute");
		child_unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		break;
	case OBJECTRELATION_QUOTA_USERRECIPIENT:
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		member_attr = m_config->GetSetting("ldap_quota_userwarning_recipients_attribute");
		member_attr_type = m_config->GetSetting("ldap_quota_userwarning_recipients_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_quota_userwarning_recipients_relation_attribute");
		child_unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		break;
	case OBJECTRELATION_QUOTA_COMPANYRECIPIENT:
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		member_attr = m_config->GetSetting("ldap_quota_companywarning_recipients_attribute");
		member_attr_type = m_config->GetSetting("ldap_quota_companywarning_recipients_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_quota_companywarning_recipients_relation_attribute");
		child_unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		break;
	default:
		throw runtime_error("Cannot obtain parents for relation " + stringify(relation));
		break;
	}

	char *request_attrs[] = {
		unique_attr,
		modify_attr,
		NULL
	};

	ldap_basedn = getSearchBase(parenttype);
	search_filter = getSearchFilter(parenttype);

	/* Search filter is empty, so there are no parents */
	if (search_filter.empty())
		goto exit;

	if(member_attr_rel == NULL  || strlen(member_attr_rel) == 0)
		member_attr_rel = child_unique_attr;

	if (member_attr_type && stricmp(member_attr_type, LDAP_DATA_TYPE_DN) == 0) {
		memberSearchData = objectUniqueIDtoObjectDN(childtype, id); //ads
	} else { //LDAP_DATA_TYPE_ATTRIBUTE
		// FIXME: No binary support for member_attr_rel
		memberSearchData = objectUniqueIDtoAttributeData(childtype, id, member_attr_rel); //posix ldap
	}

	ldap_filter = "(&" + search_filter + "(" + member_attr + "=" + StringEscapeSequence(memberSearchData) + "))";

	// Do a search request to get all attributes for this user. The
	// attributes requested should be passed as the fifth argument, if
	// necessary
	if ((rc = my_ldap_search_s(
			 (char *)ldap_basedn.c_str(), ulScope,
			 (char *)ldap_filter.c_str(), request_attrs,
			 FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	// Get the DN for each of the returned entries
	// ldap_first_message would work too, but that needs much more
	// work parsing the results from the ber structs.
	for (entry = ldap_first_entry(m_ldap, res);
		 entry != NULL;
		 entry = ldap_next_entry(m_ldap, entry)) {

		objectid_t objectid;
		string signature;

		for (att = ldap_first_attribute(m_ldap, entry, &ber);
			 att != NULL;
			 att = ldap_next_attribute(m_ldap, entry, ber)) {

			if (stricmp(att, unique_attr) == 0)
				objectid = getLDAPAttributeValue(att, entry);

			if (stricmp(att, modify_attr) == 0)
				signature = getLDAPAttributeValue(att, entry);

			ldap_memfree(att);
		}

		parents.push_back(objectsignature_t(objectid, signature));

		if(ber)
			ber_free(ber, 0);
	}

exit:
	if (res)
		ldap_msgfree(res);
	return parents;
}

list<objectsignature_t> LDAPUserPlugin::getSubObjectsForObject(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &id, userobject_type_t childtype) throw(std::exception) {
	LDAPMessage*	res = NULL;
	int				rc;
	string			ldap_basedn;

	ULONG ulScope;
	string ldap_filter;
	char *unique_attr = NULL;
	char *unique_attr_type = NULL;
	char *member_attr = NULL;
	char *member_attr_type = NULL;
	char *member_attr_rel = NULL;
	char *child_unique_attr = NULL;
	enum LISTTYPE { MEMBERS, FILTER } ulType = MEMBERS;
	string ldap_member_filter;
	string dn;
	objectid_t companyid;

	switch (childtype) {
	    case USEROBJECT_TYPE_USER:
	    case USEROBJECT_TYPE_NONACTIVE:
	        child_unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
	        break;
	    case USEROBJECT_TYPE_GROUP:
	        child_unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
	        break;
	    case USEROBJECT_TYPE_COMPANY:
	        child_unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
	        break;
	    case USEROBJECT_TYPE_ADDRESSLIST:
	        child_unique_attr = m_config->GetSetting("ldap_addresslist_unique_attribute");
	        break;
	}

	switch (relation) {
	case OBJECTRELATION_GROUP_MEMBER:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_group_unique_attribute_type");
		member_attr = m_config->GetSetting("ldap_groupmembers_attribute");
		member_attr_type = m_config->GetSetting("ldap_groupmembers_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_groupmembers_relation_attribute");
		break;
	case OBJECTRELATION_COMPANY_VIEW:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_company_unique_attribute_type");
		member_attr = m_config->GetSetting("ldap_company_view_attribute");
		member_attr_type = m_config->GetSetting("ldap_company_view_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_company_view_relation_attribute");
		break;
	case OBJECTRELATION_COMPANY_ADMIN:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_company_unique_attribute_type");
		member_attr = m_config->GetSetting("ldap_company_admin_attribute");
		member_attr_type = m_config->GetSetting("ldap_company_admin_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_company_admin_relation_attribute");
		break;
	case OBJECTRELATION_QUOTA_USERRECIPIENT:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_company_unique_attribute_type");
		member_attr = m_config->GetSetting("ldap_quota_userwarning_recipients_attribute");
		member_attr_type = m_config->GetSetting("ldap_quota_userwarning_recipients_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_quota_userwarning_recipients_relation_attribute");
		break;
	case OBJECTRELATION_QUOTA_COMPANYRECIPIENT:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_company_unique_attribute_type");
		member_attr = m_config->GetSetting("ldap_quota_companywarning_recipients_attribute");
		member_attr_type = m_config->GetSetting("ldap_quota_companywarning_recipients_attribute_type");
		member_attr_rel = m_config->GetSetting("ldap_quota_companywarning_recipients_relation_attribute");
		break;
    case OBJECTRELATION_ADDRESSLIST_MEMBER:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_addresslist_scope"));
		unique_attr = m_config->GetSetting("ldap_addresslist_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_addresslist_unique_attribute_type");
		member_attr = m_config->GetSetting("ldap_addresslist_filter_attribute");
		ulType = FILTER;
        break;        
	default:
		throw runtime_error("Cannot obtain children for relation " + stringify(relation));
		break;
	}

	if(member_attr_rel == NULL  || strlen(member_attr_rel) == 0)
		member_attr_rel = child_unique_attr;

	char *request_attrs[] = { member_attr,
							  NULL };

	//Get group DN from unique id
	ldap_basedn = getSearchBase(parenttype);
	ldap_filter = getObjectSearchFilter(parenttype, id, unique_attr, unique_attr_type);

	/* LDAP filter empty, parent does not exist */
	if (ldap_filter.empty())
		throw objectnotfound(id);

	if ((rc = my_ldap_search_s(
			 (char *)ldap_basedn.c_str(), ulScope,
			 (char *)ldap_filter.c_str(),
			 request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	LDAPMessage *entry = NULL;
	list<objectsignature_t> members;
	list<string> memberlist;

	if(ulType == MEMBERS) {
        // Get the DN for each of the returned entries
        // ldap_first_message would work too, but that needs much more
        // work parsing the results from the ber structs.
        for (entry = ldap_first_entry(m_ldap, res);
             entry != NULL;
             entry = ldap_next_entry(m_ldap, entry)) {

            BerElement *ber;
            char *att;
            for (att = ldap_first_attribute(m_ldap, entry, &ber);
                 att != NULL;
                 att = ldap_next_attribute(m_ldap, entry, ber)) {
                if (member_attr && !stricmp(att, member_attr)) {
                    memberlist = getLDAPAttributeValues(att, entry);

                    if (memberlist.empty())
                        continue;

                    members = resolveObjectsFromAttributeType(childtype, memberlist, member_attr_rel, member_attr_type);
                }
                ldap_memfree(att);
            }
            ber_free(ber, 0);

        }
        if (res)
            ldap_msgfree(res);
    } else {
        // Members are specified by a filter
        for (entry = ldap_first_entry(m_ldap, res);
             entry != NULL;
             entry = ldap_next_entry(m_ldap, entry)) {

            // We need the DN of the addresslist so that we can later find out which company it is in
            dn = ldap_get_dn(m_ldap, entry);
            
            BerElement *ber;
            char *att;
            for (att = ldap_first_attribute(m_ldap, entry, &ber);
                 att != NULL;
                 att = ldap_next_attribute(m_ldap, entry, ber)) {
                if (member_attr && !stricmp(att, member_attr)) {
                    ldap_member_filter = getLDAPAttributeValue(att, entry);
                }
                ldap_memfree(att);
            }
            ber_free(ber, 0);
        }
        
        if (res)
            ldap_msgfree(res);

        if(!ldap_member_filter.empty()) {
            // Use the filter to get all members matching the specified search filter
            ldap_filter = "(&" + getSearchFilter(childtype) + ldap_member_filter + ")";
            
            if(m_bHosted) {
                std::auto_ptr<dn_cache_t> lpCache = m_lpCache->getObjectDNCache(this, USEROBJECT_TYPE_COMPANY);
                companyid = m_lpCache->getParentForDN(lpCache, dn).object;
            }
      
            members = getAllObjectsByFilter(childtype, companyid, ldap_filter);
        }
    }

	return members;
}

void LDAPUserPlugin::addSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &id, userobject_type_t childtype, const objectid_t &member) throw(std::exception) {
	throw notimplemented("add object relations is not supported when using the LDAP user plugin.");
}

void LDAPUserPlugin::deleteSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &id, userobject_type_t childtype, const objectid_t &member) throw(std::exception) {
	throw notimplemented("Delete object relations is not supported when using the LDAP user plugin.");
}

// match is in windows-1252
list<objectsignature_t> LDAPUserPlugin::searchObject(userobject_type_t type, const string &match, unsigned int ulFlags) throw(std::exception) {
	LDAPMessage*	res = NULL;
	LDAPMessage*	entry = NULL;
	BerElement*		ber = NULL;
	int				rc;
	string			ldap_filter;
	ULONG			ulScope;
	string			escMatch;
	size_t			pos;

	string ldap_basedn;
 	string search_filter;
	char *modify_attr = m_config->GetSetting("ldap_last_modification_attribute");
	char *unique_attr;
 	char *loginname_attr;
	char *fullname_attr;
	char *emailaddress_attr = m_config->GetSetting("ldap_emailaddress_attribute");

	list<objectsignature_t> lSignatures;

	switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));
		unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		loginname_attr = m_config->GetSetting("ldap_loginname_attribute");
		fullname_attr = m_config->GetSetting("ldap_fullname_attribute");
		break;
	case USEROBJECT_TYPE_GROUP:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		loginname_attr = m_config->GetSetting("ldap_groupname_attribute");
		fullname_attr = m_config->GetSetting("ldap_groupname_attribute"); // HACK to make the code easier
		break;
	case USEROBJECT_TYPE_COMPANY:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		loginname_attr = m_config->GetSetting("ldap_companyname_attribute");
		fullname_attr = m_config->GetSetting("ldap_companyname_attribute");
		break;
	default:
		throw runtime_error(string("Unknown type passed to searchObject"));
	}

	char *request_attrs[] = {
		unique_attr,
		modify_attr,
		NULL
	};

	ldap_basedn = getSearchBase(type);

	search_filter = getSearchFilter(type);

	/* We can't search for objects when no filter was specified */
	if (search_filter.empty())
		goto exit;

	// Escape match string
	escMatch = StringEscapeSequence(m_iconvrev->convert(match));

	if (! (ulFlags & EMS_AB_ADDRESS_LOOKUP)) {
		try {
			// the admin should place %s* in the search filter
			ldap_filter = m_config->GetSetting("ldap_object_search_filter");
			// search/replace %s -> escMatch
			while ((pos = ldap_filter.find("%s")) != string::npos)
				ldap_filter.replace(pos, 2, escMatch);
		} catch (...) {};
		// custom filter was empty, add * for a full search
		if (ldap_filter.empty())
			escMatch += "*";
	}

	if (ldap_filter.empty()) {
		// if building the filter failed, use our search filter
		ldap_filter =
			"(|(" + string(loginname_attr) + "=" + escMatch + ")"
			   "(" + string(fullname_attr) + "=" + escMatch + ")"
			   "(" + string(emailaddress_attr) + "=" + escMatch + "))";
	}
	ldap_filter = "(&" + search_filter + ldap_filter + ")";

	if ((rc = my_ldap_search_s(
			 (char *)ldap_basedn.c_str(), ulScope,
			 (char *)ldap_filter.c_str(),
			 request_attrs, FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	if (ldap_count_entries(m_ldap, res) == 0) {
		ldap_msgfree(res);
		throw objectnotfound(ldap_filter);
	}

	for (entry = ldap_first_entry(m_ldap, res);
		 entry != NULL;
		 entry = ldap_next_entry(m_ldap, entry)) {

		char *att;
		objectid_t objectid;
		std::string signature;

		for (att = ldap_first_attribute(m_ldap, entry, &ber);
			 att != NULL;
			 att = ldap_next_attribute(m_ldap, entry, ber)) {

			if (stricmp(att, unique_attr) == 0)
				objectid = getLDAPAttributeValue(att, entry);

			if (stricmp(att, modify_attr) == 0)
				signature = getLDAPAttributeValue(att, entry);

			ldap_memfree(att);
		}

		lSignatures.push_back(objectsignature_t(objectid, signature));
	}

exit:
	if (ber)
		ber_free(ber, 0);

	if(res)
		ldap_msgfree(res);

	return lSignatures;
}

auto_ptr<objectdetails_t> LDAPUserPlugin::getPublicStoreDetails() throw(std::exception)
{
	throw notimplemented("distributed");
}

auto_ptr<serverdetails_t> LDAPUserPlugin::getServerDetails(const string &server) throw(std::exception)
{
	throw notimplemented("distributed");
}

extern "C" {
	UserPlugin* getUserPluginInstance(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed) {
		return new LDAPUserPlugin(pluginlock, config, logger, lpStatsCollector, bHosted, bDistributed);
	}

	void deleteUserPluginInstance(UserPlugin *up) {
		delete up;
	}
}

// Convert unsigned char to 2-char hex string
std::string toHex(unsigned char n) {
	std::string s;
	static char digits[] = "0123456789ABCDEF";

	s += digits[n >> 4];
	s += digits[n & 0xf];

	return s;
}

HRESULT LDAPUserPlugin::BintoEscapeSequence(const char* lpdata, size_t size, string* lpEscaped)
{
	HRESULT hr = 0;
	lpEscaped->clear();

	for(size_t t=0; t < size; t++) {
		lpEscaped->append("\\"+toHex(lpdata[t]));
	}

	return hr;
}
std::string LDAPUserPlugin::StringEscapeSequence(const string &strData)
{
	return StringEscapeSequence(strData.c_str(), strData.size());
}

std::string LDAPUserPlugin::StringEscapeSequence(const char* lpdata, size_t size)
{
	std::string strEscaped;

	for(size_t t=0; t < size; t++) {

		if( lpdata[t] != ' ' &&
			(lpdata[t] < '0' || lpdata[t] > 122 || (lpdata[t] >= 58/*:*/ && lpdata[t] <= 64 /*:;<=>?@*/) || (lpdata[t] >= 91 && lpdata[t] <= 96)/* [\]^_`*/) )
		{
			strEscaped.append("\\"+toHex(lpdata[t]));
		}else{
			strEscaped.append((char*)&lpdata[t], 1);
		}
	}

	return strEscaped;
}

auto_ptr<quotadetails_t> LDAPUserPlugin::getQuota(userobject_type_t type, const objectid_t &id, bool bGetUserDefault) throw(std::exception)
{
	LDAPMessage *res = NULL;
	int rc;
	string dn;
	ULONG ulScope;
	char *unique_attr;
	char *unique_attr_type;

	char *multiplier_s = m_config->GetSetting("ldap_quota_multiplier");
	long long multiplier = 1L;
	if (multiplier_s) {
		multiplier = fromstring<const char *, long long>(multiplier_s);
	}

	char *usedefaults_attr = NULL;
	char *warnquota_attr = NULL;
	char *softquota_attr = NULL;
	char *hardquota_attr = NULL;

	if (bGetUserDefault) {
		usedefaults_attr = m_config->GetSetting("ldap_userdefault_quotaoverride_attribute");
		warnquota_attr = m_config->GetSetting("ldap_userdefault_warnquota_attribute");
		softquota_attr = m_config->GetSetting("ldap_userdefault_softquota_attribute");
		hardquota_attr = m_config->GetSetting("ldap_userdefault_hardquota_attribute");
	} else {
		usedefaults_attr = m_config->GetSetting("ldap_quotaoverride_attribute");
		warnquota_attr = m_config->GetSetting("ldap_warnquota_attribute");
		softquota_attr = m_config->GetSetting("ldap_softquota_attribute");
		hardquota_attr = m_config->GetSetting("ldap_hardquota_attribute");
	}

    switch (type) {
	case USEROBJECT_TYPE_USER:
	case USEROBJECT_TYPE_NONACTIVE:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_user_scope"));
		unique_attr = m_config->GetSetting("ldap_user_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_user_unique_attribute_type");
		break;
	case USEROBJECT_TYPE_GROUP:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_group_scope"));
		unique_attr = m_config->GetSetting("ldap_group_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_group_unique_attribute_type");
		break;
	case USEROBJECT_TYPE_COMPANY:
		ulScope = ScopetoScope(m_config->GetSetting("ldap_company_scope"));
		unique_attr = m_config->GetSetting("ldap_company_unique_attribute");
		unique_attr_type = m_config->GetSetting("ldap_company_unique_attribute_type");
		break;
	default:
		throw runtime_error(string("Unknown object type"));
	}

	char *request_attrs[] = { usedefaults_attr,
							  warnquota_attr,
							  softquota_attr,
							  hardquota_attr,
							  NULL };

	string ldap_basedn = getSearchBase(type);
	string ldap_filter = getObjectSearchFilter(type, id, unique_attr, unique_attr_type);

	/* LDAP filter empty, object does not exist */
	if (ldap_filter.empty())
		throw objectnotfound(id);

	// Do a search request to get all attributes for this user. The
	// attributes requested should be passed as the fifth argument, if
	// necessary
	if ((rc = my_ldap_search_s(
			 (char *)ldap_basedn.c_str(), ulScope,
			 (char *)ldap_filter.c_str(), request_attrs,
			 FETCH_ATTR_VALS, &res)) != LDAP_SUCCESS) {
		throw runtime_error(string("ldap_search_s: ") + ldap_err2string(rc));
	}

	// cerr << ldap_count_entries(m_ldap, res) << " entries" << endl;

	LDAPMessage *entry = NULL;
	quotadetails_t *d = new quotadetails_t();
	d->bIsUserDefaultQuota = bGetUserDefault;

	// Get the DN for each of the returned entries
	// ldap_first_message would work too, but that needs much more
	// work parsing the results from the ber structs.
	for (entry = ldap_first_entry(m_ldap, res);
		 entry != NULL;
		 entry = ldap_next_entry(m_ldap, entry)) {

		BerElement *ber;
		char *att;
		for (att = ldap_first_attribute(m_ldap, entry, &ber);
			 att != NULL;
			 att = ldap_next_attribute(m_ldap, entry, ber)) {

			if (usedefaults_attr && !stricmp(att, usedefaults_attr)) {
				// Workarround quotaoverride == !usedefaultquota
				d->bUseDefaultQuota = !parseBool(getLDAPAttributeValue(att, entry));
			} else if (warnquota_attr && !stricmp(att, warnquota_attr)) {
				d->llWarnSize = fromstring<string, long long>(getLDAPAttributeValue(att, entry)) * multiplier;
			} else if (softquota_attr && !stricmp(att, softquota_attr)) {
				d->llSoftSize = fromstring<string, long long>(getLDAPAttributeValue(att, entry)) * multiplier;
			} else if (hardquota_attr && !stricmp(att, hardquota_attr)) {
				d->llHardSize = fromstring<string, long long>(getLDAPAttributeValue(att, entry)) * multiplier;
			}
			ldap_memfree(att);
		}
		ber_free(ber, 0);
	}

	if (res)
		ldap_msgfree(res);

	return auto_ptr<quotadetails_t>(d);
}

int LDAPUserPlugin::setQuota(userobject_type_t type, const objectid_t &id, quotadetails_t quotadetails) throw(std::exception)
{
	throw notimplemented("set quota is not supported when using the LDAP user plugin.");
}

void LDAPUserPlugin::getUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives) throw (std::exception)
{
	throw notimplemented("distributed");
}
