/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

// -*- Mode: c++ -*-
#ifndef LDAPUSERPLUGIN_H
#define LDAPUSERPLUGIN_H

#include <stdexcept>
#include <string>
#include <ECIConv.h>

// OpenLDAP
#include <ldap.h>

#include "plugin.h"
#include "LDAPCache.h"

/** LDAP user plugin
 *
 * User management based on LDAP.
 */
class LDAPUserPlugin: public UserPlugin {
public:
	/** \name Constructors/destructors */
	/*@{*/
	/** Constructor
	 *
	 * Create a connection to the LDAP server and do some
	 * initialization.
	 *
	 * Configuration parameters: \c ldap_url
	 *
	 * Possible configuration parameters: \c ldap_use_tls, \c
	 * ldap_bind_user, \c ldap_bind_passwd
	 *
	 * \throws std::runtime_error on failure, such as being unable to
	 * connect to the LDAP server, or when the bind credentials are
	 * incorrect.
	 *
	 * \param config Used to get the plugin's configuration values.
	 *
	 * \todo The constructor sets the maximum size of query results to
	 * infinite, instead it'd be better to use ldap_search instead of
	 * ldap_search_s.
	 */
	LDAPUserPlugin(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed);
	virtual ~LDAPUserPlugin();

	virtual void InitPlugin() throw(std::exception);

	// Reading information
	virtual objectsignature_t 			resolveName(userobject_type_t type, const string &name, const string &company) throw(std::exception);
	virtual objectsignature_t 			authenticateUser(const string &username, const string &companyname, const string &password) throw(std::exception);
	virtual list<objectsignature_t> 	getAllObjects(userobject_type_t type, const objectid_t &company) throw(std::exception);
	virtual auto_ptr<objectdetails_t> 	getObjectDetails(userobject_type_t type, const objectid_t &objectid) throw(std::exception);
	virtual auto_ptr<map<objectid_t, objectdetails_t> > getObjectDetails(userobject_type_t type, list<objectid_t> &objectids) throw(std::exception);
	virtual list<objectsignature_t> 	getSubObjectsForObject(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype) throw(std::exception);
	virtual list<objectsignature_t>		getParentObjectsForObject(userobject_relation_t relation, userobject_type_t childtype, const objectid_t &childobject, userobject_type_t parenttype) throw(std::exception);
	virtual list<objectsignature_t>		searchObject(userobject_type_t, const string &match, unsigned int ulFlags) throw(std::exception);
	virtual auto_ptr<objectdetails_t>	getPublicStoreDetails() throw(std::exception);
	virtual auto_ptr<serverdetails_t>	getServerDetails(const string &server) throw(std::exception);
	virtual void						getUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives) throw (std::exception);
	
	// Writing information (optional)
	virtual int 						changeObject(userobject_type_t type, const objectid_t &id, objectdetails_t details, std::list<std::string> *lpRemove) throw(std::exception);
	virtual objectsignature_t 			createObject(userobject_type_t type, objectdetails_t details) throw(std::exception);
	virtual void 						deleteObject(userobject_type_t type, const objectid_t &id) throw(std::exception);
	
	virtual void 						addSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception);
	virtual void 						deleteSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t& parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception);
	
	// Quota 
	virtual auto_ptr<quotadetails_t> 	getQuota(userobject_type_t type, const objectid_t &id, bool bGetUserDefault) throw(std::exception);
	virtual int 						setQuota(userobject_type_t type, const objectid_t &id, quotadetails_t quotadetails) throw(std::exception);

protected:
	LDAP *m_ldap; 		///< Pointer to the LDAP state struct.
	ECIConv *m_iconv; 	// converter FROM ldap TO zarafa-server
	ECIConv *m_iconvrev; 	// converter FROM zarafa-server TO ldap

	static auto_ptr<LDAPCache> m_lpCache;

private:
	/** Get the value of the given attribute from the search results.
	 *
	 * This is a convenience function that uses
	 * getLDAPAttributeValues.  Parse the result of an LDAP search
	 * query, and retrieve exactly one string.  If the attribute does
	 * not occur in the search results, or if there are no values for
	 * the attribute, returns an empty string.
	 *
	 * \param attribute The name of the attribute.
	 * \param entry The entry result from \c ldap_first_entry &c.
	 * \returns The first value of the attribute, or an empty string
	 * if the attribute or a value was not found.
	 */
	std::string getLDAPAttributeValue(char *attribute, LDAPMessage *entry);

	/** Get multiple values of the given attribute from the search
	 * results.
	 *
	 * Parse the result of an LDAP search query, and retrieve all the
	 * values for the given attribute.
	 *
	 * \param attribute The name of the attribute.
	 * \param entry The entry result from \c ldap_first_entry &c.
	 */
	std::list<std::string> getLDAPAttributeValues(char *attribute, LDAPMessage *entry);

	/** Set one attribute of \c dn to \c value
	 *
	 * \param dn The DN to modify.
	 * \param attribute The name of the attribute to modify.
	 * \param value The new value of the attribute.
	 * \retval 0 When the modification was successful.
	 * \retval 1 Otherwise (error).
	 */
	int changeAttribute(const char *dn, char *attribute, const char *value);

	/** Set one attribute of \c dn to the values in \c values
	 *
	 * \param dn The DN to modify.
	 * \param attribute The name of the attribute to modify.
	 * \param values A list of new values.  Any old values are
	 * overwritten or removed.
	 * \retval 0 When the modification was successful.
	 * \retval 1 Otherwise (error).
	 */
	int changeAttribute(const char *dn, char *attribute, const std::list<std::string> &values);

	LDAP *ConnectLDAP(const char *bind_dn, const char *bind_pw) throw(std::exception);
	objectsignature_t authenticateUserBind(const string &username, const string &companyname, const string &password) throw(std::exception);
	objectsignature_t authenticateUserPassword(const string &username, const string &companyname, const string &password) throw(std::exception);

	string objectUniqueIDtoObjectDN(userobject_type_t type, objectid_t uniqueid) throw(std::exception);
	objectsignature_t objectDNtoObjectSignature(userobject_type_t type, const string &dn) throw(std::exception);

	HRESULT BintoEscapeSequence(const char* lpdata, size_t size, string* lpEscaped);
	std::string StringEscapeSequence(const char* lpdata, size_t size);
	std::string StringEscapeSequence(const string &strData);

	string getSearchBase(userobject_type_t type, const string &companyname = "") throw(std::exception);
	string getSearchFilter(userobject_type_t type);
	string getObjectSearchFilter(userobject_type_t type, const objectid_t &id, const string &attr, const string &attr_type = "");
	ULONG ScopetoScope(const char* lpScope, ULONG ulDefaultScope = LDAP_SCOPE_SUBTREE);
	objectsignature_t resolveObjectFromAttribute(userobject_type_t type, string company, const char* lpAttrData, const char* lpAttr) throw(std::exception);
	list<objectsignature_t> resolveObjectsFromAttribute(userobject_type_t type, list<string> objects, const char* lpAttr) throw(std::exception);
	list<objectsignature_t> resolveObjectsFromAttributeType(userobject_type_t type, list<string> &objects, const char* lpAttr, const char* lpAttrType) throw(std::exception);
	string objectUniqueIDtoAttributeData(userobject_type_t type, objectid_t uniqueid, const char* lpAttr) throw(std::exception);
	string objectDNtoAttributeData(userobject_type_t type, string dn, const char *lpAttr) throw(std::exception);
    list<objectsignature_t> getAllObjectsByFilter(userobject_type_t type, const objectid_t &company, const std::string &filter) throw(std::exception);

	// Wrapper function for ldap_search_s which has reconnect
	int my_ldap_search_s(char *base, int scope, char *filter, char *attrs[], int attrsonly, LDAPMessage **lppres);
};

extern "C" {
	extern UserPlugin* getUserPluginInstance(pthread_mutex_t*, ECConfig*, ECLogger*, IECStatsCollector *,bool, bool);
	extern void deleteUserPluginInstance(UserPlugin*);
}

#endif
