/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <iostream>
#include <string>
#include <stdexcept>
#include <sys/types.h>
#include <pwd.h>
#include <sstream>
#include <assert.h>
#include <sys/stat.h>
#include <grp.h>
#include <crypt.h>

#ifdef HAVE_SHADOW_H
#include <shadow.h>
#include <errno.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "EMSAbTag.h"
#include "ECConfig.h"
#include "ECDefs.h"
#include "ECLogger.h"
#include "stringutil.h"

using namespace std;

#include "UnixUserPlugin.h"

// static buffer size for getpwnam_r() calls etc.
// needs to be big enough for all strings in the passwd/group/spwd struct:
#define PWBUFSIZE 16384

enum {
	ADD_ACTION,
	REMOVE_ACTION,
	UPDATE_ACTION,
	ADD_MEMBER_ACTION,
	DEL_MEMBER_ACTION
};

extern "C" {
	UserPlugin* getUserPluginInstance(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed) {
		return new UnixUserPlugin(pluginlock, config, logger, lpStatsCollector, bHosted, bDistributed);
	}

	void deleteUserPluginInstance(UserPlugin *up) {
		delete up;
	}
}


UnixUserPlugin::UnixUserPlugin(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed) throw(exception)
	: DBPlugin(pluginlock, config, logger, lpStatsCollector, bHosted, bDistributed), m_iconv(NULL)
{
	const configsetting_t lpDefaults [] = {
		{ "fullname_charset", "iso-8859-15" },		// us-ascii compatible with support for high characters
		{ "default_domain", "localhost" },			// no sane default
		{ "non_login_shell", "/bin/false" },		// create a non-login box when a user has this shell
		{ "min_user_uid", "1000", CONFIGSETTING_RELOADABLE },
		{ "max_user_uid", "10000", CONFIGSETTING_RELOADABLE },
		{ "min_group_gid", "1000", CONFIGSETTING_RELOADABLE },
		{ "max_group_gid", "10000", CONFIGSETTING_RELOADABLE },
		{ NULL, NULL }
	};

	m_config = new ECConfig(lpDefaults);
	if (!m_config->LoadSettings(config->GetSetting("user_plugin_config")))
		logger->Log(EC_LOGLEVEL_ERROR, "Failed to open Unix plugin configuration file, using defaults.");
	if (m_config->HasErrors()) {
		LogConfigErrors(m_config, logger);
		delete m_config;
		throw runtime_error(string("not a valid configuration file."));
	}

	if (bHosted)
		throw notsupported("Hosted Zarafa not supported when using the Unix Plugin");
	if (bDistributed)
		throw notsupported("Distributed Zarafa not supported when using the Unix Plugin");
}

UnixUserPlugin::~UnixUserPlugin() {
	if (m_config)
		delete m_config;

	if (m_iconv)
		delete m_iconv;
}

void UnixUserPlugin::InitPlugin() throw(std::exception) {
	DBPlugin::InitPlugin();

	// we only need unix_charset -> zarafa charset
	m_iconv = new ECIConv("windows-1252", m_config->GetSetting("fullname_charset"));
}

objectsignature_t UnixUserPlugin::resolveName(userobject_type_t type, const string &name, const string &company) throw(std::exception) {
    char buffer[PWBUFSIZE];
    errno = 0;

    // This function is so different when doing groups and users that we're just splitting the two
    // cases entirely.
    if(type == USEROBJECT_TYPE_GROUP) {
        struct group grs, *gr = NULL;
        gid_t mingid = fromstring<const char *, gid_t>(m_config->GetSetting("min_group_gid"));
        gid_t maxgid = fromstring<const char *, gid_t>(m_config->GetSetting("max_group_gid"));

        getgrnam_r(name.c_str(), &grs, buffer, PWBUFSIZE, &gr);
        errnoCheck(name);

        if (gr == NULL) {
            throw objectnotfound(name);
        }

        if (gr->gr_gid < mingid || gr->gr_gid >= maxgid)
            throw objectnotfound(name);

        return objectsignature_t(tostring(gr->gr_gid), gr->gr_name);
    } else if(type == USEROBJECT_TYPE_USER || type == USEROBJECT_TYPE_NONACTIVE) {
        struct passwd pws, *pw = NULL;
        uid_t minuid = fromstring<const char *, uid_t>(m_config->GetSetting("min_user_uid"));
        uid_t maxuid = fromstring<const char *, uid_t>(m_config->GetSetting("max_user_uid"));

        getpwnam_r(name.c_str(), &pws, buffer, PWBUFSIZE, &pw);
        errnoCheck(name);

        if (pw == NULL) {
            throw objectnotfound(name);
        }

        if (pw->pw_uid < minuid || pw->pw_uid >= maxuid)
            throw objectnotfound(name);

		if(type == USEROBJECT_TYPE_USER && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) == 0)
			throw objectnotfound(name);	// user requested but type is nonactive

		if(type == USEROBJECT_TYPE_NONACTIVE && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) != 0)
			throw objectnotfound(name); // nonactive requested but type is user

        return objectsignature_t(tostring(pw->pw_uid), getDBSignature(tostring(pw->pw_uid), type) + pw->pw_gecos + pw->pw_name);
    }
    
    throw(runtime_error("Bad userobject type"));
}

objectsignature_t UnixUserPlugin::authenticateUser(const string &username, const string &companyname, const string &password) throw(std::exception) {
	struct passwd pws, *pw = NULL;
	char buffer[PWBUFSIZE];
	uid_t minuid = fromstring<const char *, uid_t>(m_config->GetSetting("min_user_uid"));
	uid_t maxuid = fromstring<const char *, uid_t>(m_config->GetSetting("max_user_uid"));

	auto_ptr<struct crypt_data> cryptdata = auto_ptr<struct crypt_data>(new struct crypt_data); // malloc because it is > 128K !
	memset(cryptdata.get(), 0, sizeof(struct crypt_data));

	errno = 0;
	getpwnam_r(username.c_str(), &pws, buffer, PWBUFSIZE, &pw);
	errnoCheck(username);

	if (pw == NULL) {
		throw objectnotfound(username);
	}

	if (pw->pw_uid < minuid || pw->pw_uid >= maxuid)
		throw objectnotfound(username);

	if (strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) == 0)
		throw login_error("Non-active user disallowed to login");

	auto_ptr<objectdetails_t> ud = objectdetailsFromPwent(pw);
	char *crpw = crypt_r(password.c_str(), ud->GetPropString(OB_PROP_S_PASSWORD).c_str(), cryptdata.get());
	if (!crpw || strcmp(crpw, ud->GetPropString(OB_PROP_S_PASSWORD).c_str())) {
		throw login_error("Trying to authenticate failed: wrong username or password");
	} else {
		return objectsignature_t(tostring(pw->pw_uid), getDBSignature(tostring(pw->pw_uid), USEROBJECT_TYPE_USER) + pw->pw_gecos + pw->pw_name);
	}
}

list<objectsignature_t> UnixUserPlugin::getAllObjects(userobject_type_t type, const objectid_t &companyid) throw(exception) {
	ECRESULT er = erSuccess;
	list<objectsignature_t> objectlist;
	list<int> objectIds;
	list<int>::iterator oidIter;
	string strQuery, strIds;
	unsigned int ulAffRows = 0;
	char buffer[PWBUFSIZE];
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;

	// use mutex to protect thread-unsafe setpwent()/setgrent() calls
	pthread_mutex_lock(m_plugin_lock);
	if(type == USEROBJECT_TYPE_USER || type == USEROBJECT_TYPE_NONACTIVE) {
		uid_t minuid = fromstring<const char *, uid_t>(m_config->GetSetting("min_user_uid"));
		uid_t maxuid = fromstring<const char *, uid_t>(m_config->GetSetting("max_user_uid"));
		struct passwd pws, *pw = NULL;
		setpwent();
		while (1) {
			getpwent_r(&pws, buffer, PWBUFSIZE, &pw);
			if (pw == NULL)
				break;

			// system users don't have zarafa accounts
			if (pw->pw_uid < minuid || pw->pw_uid >= maxuid)
				continue;
				
            if(type == USEROBJECT_TYPE_USER && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) == 0)
                continue; // user requested but type is nonactive
                
            if(type == USEROBJECT_TYPE_NONACTIVE && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) != 0)
                continue; // nonactive requested but type is user

			objectIds.push_back(pw->pw_uid);
			objectlist.push_back(objectsignature_t(tostring(pw->pw_uid), getDBSignature(tostring(pw->pw_uid), type) + pw->pw_gecos + pw->pw_name));
		}
		endpwent();
	} else if(type == USEROBJECT_TYPE_GROUP) {
		gid_t mingid = fromstring<const char *, gid_t>(m_config->GetSetting("min_group_gid"));
		gid_t maxgid = fromstring<const char *, gid_t>(m_config->GetSetting("max_group_gid"));
		struct group grs, *gr = NULL;

		setgrent();
		while (1) {
			getgrent_r(&grs, buffer, PWBUFSIZE, &gr);
			if (gr == NULL)
				break;

			if (gr->gr_gid < mingid || gr->gr_gid >= maxgid)
				continue; // system groups don't have zarafa accounts

			objectIds.push_back(gr->gr_gid);
			objectlist.push_back(objectsignature_t(tostring(gr->gr_gid),gr->gr_name));
		}
		endgrent();
	}
	pthread_mutex_unlock(m_plugin_lock);

	// Cleanup old entries from deleted users/groups
	if (!objectIds.empty()) {
		for (oidIter = objectIds.begin(); oidIter != objectIds.end(); ) {
			strIds += stringify(*oidIter);
			oidIter++;
			if (oidIter != objectIds.end())
				strIds += ", ";
		}

		// make list of obsolete objects
		strQuery = "SELECT id FROM object WHERE externid NOT IN (" + strIds + ") AND objecttype=" + stringify(type);
		er = m_lpDatabase->DoSelect(strQuery, &lpResult);
		if (er != erSuccess) {
			m_logger->Log(EC_LOGLEVEL_ERROR, "Unix plugin: Unable to cleanup old entries");
			goto exit;
		}

		ulAffRows = m_lpDatabase->GetNumRows(lpResult);
		objectIds.clear();
		while ((lpDBRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
			// WARNING: objectIds is now object.id, not externid !!
			objectIds.push_back(atoi(lpDBRow[0]));
		}

		// remove obsolete objects
		strQuery = "DELETE FROM object WHERE externid NOT IN (" + strIds + ") AND objecttype=" + stringify(type);
		er = m_lpDatabase->DoDelete(strQuery, &ulAffRows);
		if (er != erSuccess) {
			m_logger->Log(EC_LOGLEVEL_ERROR, "Unix plugin: Unable to cleanup old entries in object table");
			goto exit;
		} else if (ulAffRows > 0) {
			m_logger->Log(EC_LOGLEVEL_INFO, "Unix plugin: Cleaned-up %d old entries from object table", ulAffRows);
		}

		// make list of removed objects, and remove those properties
		if (!objectIds.empty()) {
			strIds.clear();
			for (oidIter = objectIds.begin(); oidIter != objectIds.end(); ) {
				strIds += stringify(*oidIter);
				oidIter++;
				if (oidIter != objectIds.end())
					strIds += ", ";
			}

			strQuery = "DELETE FROM objectproperty WHERE objectid IN (" + strIds + ")";
			er = m_lpDatabase->DoDelete(strQuery, &ulAffRows);
			if (er != erSuccess) {
				m_logger->Log(EC_LOGLEVEL_ERROR, "Unix plugin: Unable to cleanup old entries in objectproperty table");
				goto exit;
			} else if (ulAffRows > 0) {
				m_logger->Log(EC_LOGLEVEL_INFO, "Unix plugin: Cleaned-up %d old entries from objectproperty table", ulAffRows);
			}

			strQuery = "DELETE FROM objectrelation WHERE objectid IN (" + strIds + ") OR parentobjectid in (" + strIds + ")";
			er = m_lpDatabase->DoDelete(strQuery, &ulAffRows);
			if (er != erSuccess) {
				m_logger->Log(EC_LOGLEVEL_ERROR, "Unix plugin: Unable to cleanup old entries in objectrelation table");
				goto exit;
			} else if (ulAffRows > 0) {
				m_logger->Log(EC_LOGLEVEL_INFO, "Unix plugin: Cleaned-up %d old entries from objectrelation table", ulAffRows);
			}
		}
	}

exit:
	return objectlist;
}

auto_ptr<objectdetails_t> UnixUserPlugin::getObjectDetails(userobject_type_t type, const objectid_t &externid) throw(exception) {
   	char buffer[PWBUFSIZE];
	auto_ptr<objectdetails_t> ud;
	string strQuery;
	ECRESULT er = erSuccess;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpRow = NULL;

   	errno = 0;

	if(type == USEROBJECT_TYPE_USER || type == USEROBJECT_TYPE_NONACTIVE) {
		uid_t uid = fromstring<objectid_t, uid_t>(externid);
		struct passwd pws, *pw = NULL;
		uid_t minuid = fromstring<const char *, uid_t>(m_config->GetSetting("min_user_uid"));
		uid_t maxuid = fromstring<const char *, uid_t>(m_config->GetSetting("max_user_uid"));

		if (uid == 0) {  // unparseable objectid
			throw objectnotfound(externid);
		}

		getpwuid_r(uid, &pws, buffer, PWBUFSIZE, &pw);
		errnoCheck(externid);

		if (pw == NULL) { // user does not exist in /etc/passwd
			throw objectnotfound(externid);
		}

		if (pw->pw_uid < minuid || pw->pw_uid >= maxuid)
			throw objectnotfound(externid);

		if(type == USEROBJECT_TYPE_USER && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) == 0)
			throw objectnotfound(externid); // user requested but type is nonactive
                
		if(type == USEROBJECT_TYPE_NONACTIVE && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) != 0)
			throw objectnotfound(externid); // nonactive requested but type is user

		ud = objectdetailsFromPwent(pw);

	} else if (type == USEROBJECT_TYPE_GROUP) {
		gid_t gid = fromstring<objectid_t, gid_t>(externid);
		struct group grs, *gr = NULL;
		gid_t mingid = fromstring<const char *, gid_t>(m_config->GetSetting("min_group_gid"));
		gid_t maxgid = fromstring<const char *, gid_t>(m_config->GetSetting("max_group_gid"));

		if (gid == 0) {  // unparseable objectid
			throw objectnotfound(externid);
		}

		getgrgid_r(gid, &grs, buffer, PWBUFSIZE, &gr);
		errnoCheck(externid);

		if (gr == NULL) { // group does not exist in /etc/group
			throw objectnotfound(externid);
		}

		if (gr->gr_gid < mingid || gr->gr_gid >= maxgid)
			throw objectnotfound(externid);

		ud = objectdetailsFromGrent(gr);
	} else
		throw runtime_error("Object is wrong type");

	strQuery = "SELECT id FROM " + (string)DB_OBJECT_TABLE + " WHERE externid = '"+externid+"' AND objecttype = " + stringify(type);
	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		throw runtime_error(externid);

	lpRow = m_lpDatabase->FetchRow(lpResult);

	if (lpRow && lpRow[0]) {
		strQuery = "UPDATE " + (string)DB_OBJECT_TABLE + " SET externid='" + externid + "',objecttype=" + stringify(type) + " WHERE id=" + lpRow[0];
		er = m_lpDatabase->DoUpdate(strQuery);
	} else {
		strQuery = "INSERT INTO " + (string)DB_OBJECT_TABLE + " (externid, objecttype) VALUES ('" + externid + "', " + stringify(type) + ")";
		er = m_lpDatabase->DoInsert(strQuery);
	}
	if (er != erSuccess)
		throw runtime_error(externid);

	if (type == USEROBJECT_TYPE_USER)
		addSendAsToDetails(externid, ud.get());

	try {
		ud->CopyFrom(*DBPlugin::getObjectDetails(type, externid));
	} catch (...) { } // ignore errors; we'll try with just the information we have from Pwent

	return auto_ptr<objectdetails_t>(ud);
}

int UnixUserPlugin::changeObject(userobject_type_t type, const objectid_t &objectid, objectdetails_t details, std::list<std::string> *lpRemoveProps) throw(std::exception) {
	// explicitly deny pam updates
	if (!details.GetPropString(OB_PROP_S_PASSWORD).empty())
		throw runtime_error("Updating the password is not allowed with the Unix plugin.");
	if (!details.GetPropString(OB_PROP_S_LOGIN).empty())
		throw runtime_error("Updating the username is not allowed with the Unix plugin.");
	if (!details.GetPropString(OB_PROP_S_FULLNAME).empty())
		throw runtime_error("Updating the fullname is not allowed with the Unix plugin.");

	return DBPlugin::changeObject(type, objectid, details, lpRemoveProps);
}

objectsignature_t UnixUserPlugin::createObject(userobject_type_t type, objectdetails_t details) throw(std::exception) {
	throw notimplemented("Creating objects is not supported when using the Unix user plugin.");
}

void UnixUserPlugin::deleteObject(userobject_type_t type, const objectid_t &id) throw(std::exception) {
	throw notimplemented("Deleting objects is not supported when using the Unix user plugin.");
}

list<objectsignature_t> UnixUserPlugin::getParentObjectsForObject(userobject_relation_t relation, userobject_type_t childtype, const objectid_t &childid, userobject_type_t parenttype) throw(std::exception)
{
	if (relation != OBJECTRELATION_GROUP_MEMBER)
		return DBPlugin::getParentObjectsForObject(relation, childtype, childid, parenttype);

	uid_t uid = fromstring<const objectid_t &, uid_t>(childid);
	gid_t mingid = fromstring<const char *, gid_t>(m_config->GetSetting("min_group_gid"));
	gid_t maxgid = fromstring<const char *, gid_t>(m_config->GetSetting("max_group_gid"));
	struct passwd pws, *pw = NULL;
	char buffer[PWBUFSIZE];

	if (uid == 0) {  // unparseable objectid
		throw objectnotfound(childid);
	}

	errno = 0;
	getpwuid_r(uid, &pws, buffer, PWBUFSIZE, &pw);
	errnoCheck(childid);

	if (pw == NULL) {
		throw objectnotfound(childid);
	}

	string user(pw->pw_name);	// make sure we have a _copy_ of the string, not just another pointer
	list<objectsignature_t> grouplist;

	// add user default group
	if (pw->pw_gid >= mingid || pw->pw_gid < maxgid)
		grouplist.push_back(objectsignature_t(tostring(pw->pw_gid),""));

	// This is a rather expensive operation: loop through all the
	// groups, and check each member for each group.
	pthread_mutex_lock(m_plugin_lock);
	setgrent();
	while (1) {
		struct group grs, *gr = NULL;

		getgrent_r(&grs, buffer, PWBUFSIZE, &gr);
		if (gr == NULL)
			break;

		if (gr->gr_gid < mingid || gr->gr_gid >= maxgid)
			continue;

		for (int i = 0; gr->gr_mem[i] != NULL; i++) {
			if (strcmp(user.c_str(), gr->gr_mem[i]) == 0) {
				grouplist.push_back(objectsignature_t(tostring(gr->gr_gid),gr->gr_name));
				break;
			}
		}
	}
	endgrent();
	pthread_mutex_unlock(m_plugin_lock);

	// because users can be explicitly listed in their default group
	grouplist.sort();
	grouplist.unique();

	return grouplist;
}

list<objectsignature_t> UnixUserPlugin::getSubObjectsForObject(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentid, userobject_type_t childtype) throw(std::exception)
{
	if (relation != OBJECTRELATION_GROUP_MEMBER)
		return DBPlugin::getSubObjectsForObject(relation, parenttype, parentid, childtype);

	// unix plugin does not support group-in-group
	if (childtype == USEROBJECT_TYPE_GROUP)
		return list<objectsignature_t>();

	gid_t gid = fromstring<const objectid_t &, gid_t>(parentid);
	gid_t mingid = fromstring<const char *, gid_t>(m_config->GetSetting("min_group_gid"));
	gid_t maxgid = fromstring<const char *, gid_t>(m_config->GetSetting("max_group_gid"));
	struct group grs, *gr = NULL;
	char buffer[PWBUFSIZE];

	if (gid == 0) {  // unparseable objectid
		throw objectnotfound(parentid);
	}

	errno = 0;
	getgrgid_r(gid, &grs, buffer, PWBUFSIZE, &gr);
	errnoCheck(parentid);

	if (gr == NULL) { // group does not exist in /etc/passwd
		throw objectnotfound(parentid);
	}

	if (gr->gr_gid < mingid || gr->gr_gid >= maxgid)
		throw objectnotfound(parentid);


	list<objectsignature_t> userlist;
	uid_t minuid = fromstring<const char *, uid_t>(m_config->GetSetting("min_user_uid"));
	uid_t maxuid = fromstring<const char *, uid_t>(m_config->GetSetting("max_user_uid"));

	// find all group members
	for (int i = 0; gr->gr_mem[i] != NULL; i++) {
		struct passwd pws, *pw = NULL;
		char pbuffer[PWBUFSIZE];

		errno = 0;
		getpwnam_r(gr->gr_mem[i], &pws, pbuffer, PWBUFSIZE, &pw);
		errnoCheck(gr->gr_mem[i]);

		if (!pw || pw->pw_uid < minuid || pw->pw_uid >= maxuid)
			continue;
			
        if(childtype == USEROBJECT_TYPE_USER && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) == 0)
            continue; // user requested but type is nonactive
            
        if(childtype == USEROBJECT_TYPE_NONACTIVE && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) != 0)
            continue; // nonactive requested but type is user

		userlist.push_back(objectsignature_t(tostring(pw->pw_uid),getDBSignature(tostring(pw->pw_uid), childtype) + pw->pw_gecos + pw->pw_name));
	}

	// iterate through /etc/passwd users to find default group (eg. users) and add it to the list
	pthread_mutex_lock(m_plugin_lock);
	setpwent();
	while (1) {
		struct passwd pws, *pw = NULL;
		char pbuffer[PWBUFSIZE];

		getpwent_r(&pws, pbuffer, PWBUFSIZE, &pw);
		if (pw == NULL)
			break;

		// system users don't have zarafa accounts
		if (pw->pw_uid < minuid || pw->pw_uid >= maxuid)
			continue;

        if(childtype == USEROBJECT_TYPE_USER && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) == 0)
            continue; // user requested but type is nonactive
            
        if(childtype == USEROBJECT_TYPE_NONACTIVE && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) != 0)
            continue; // nonactive requested but type is user

		// is it a member, and fits the default group in the range?
		if (pw->pw_gid == gr->gr_gid && pw->pw_gid >= mingid && pw->pw_gid < maxgid)
			userlist.push_back(objectsignature_t(tostring(pw->pw_uid),getDBSignature(tostring(pw->pw_uid), childtype) + pw->pw_gecos + pw->pw_name));
	}
	endpwent();
	pthread_mutex_unlock(m_plugin_lock);

	// because users can be explicitly listed in their default group
	userlist.sort();
	userlist.unique();

	return userlist;
}

void UnixUserPlugin::addSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &id, userobject_type_t childtype, const objectid_t &member) throw(std::exception)
{
	if (relation != OBJECTRELATION_QUOTA_USERRECIPIENT && relation != OBJECTRELATION_USER_SENDAS)
		throw notimplemented("Adding object relations is not supported when using the Unix user plugin.");
	DBPlugin::addSubObjectRelation(relation, parenttype, id, childtype, member);
}

void UnixUserPlugin::deleteSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &id, userobject_type_t childtype, const objectid_t &member) throw(std::exception)
{
	if (relation != OBJECTRELATION_QUOTA_USERRECIPIENT && relation != OBJECTRELATION_USER_SENDAS)
		throw notimplemented("Deleting object relations is not supported when using the Unix user plugin.");
	DBPlugin::deleteSubObjectRelation(relation, parenttype, id, childtype, member);
}

list<objectsignature_t> UnixUserPlugin::searchObject(userobject_type_t type, const string &match, unsigned int ulFlags) throw(std::exception)
{
	bool matched = false;
	char buffer[PWBUFSIZE];
	list<objectsignature_t> lObjects;
	struct passwd pws, *pw = NULL;

	if(type == USEROBJECT_TYPE_USER || type == USEROBJECT_TYPE_NONACTIVE) {
		uid_t minuid = fromstring<const char *, uid_t>(m_config->GetSetting("min_user_uid"));
		uid_t maxuid = fromstring<const char *, uid_t>(m_config->GetSetting("max_user_uid"));

		pthread_mutex_lock(m_plugin_lock);
		setpwent();
		while (1) {
			getpwent_r(&pws, buffer, PWBUFSIZE, &pw);
			if (pw == NULL)
				break;

			// system users don't have zarafa accounts
			if (pw->pw_uid < minuid || pw->pw_uid >= maxuid)
				continue;

            if(type == USEROBJECT_TYPE_USER && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) == 0)
                continue; // user requested but type is nonactive
                
            if(type == USEROBJECT_TYPE_NONACTIVE && strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) != 0)
                continue; // nonactive requested but type is user

			// username or fullname
			if(ulFlags & EMS_AB_ADDRESS_LOOKUP)
				matched = stricmp(pw->pw_name, (char*)match.c_str()) == 0 || stricmp((char*)m_iconv->convert(pw->pw_gecos).c_str(), (char*)match.c_str()) == 0;
			else
				matched = strnicmp(pw->pw_name, (char*)match.c_str(), match.size()) == 0 || strnicmp((char*)m_iconv->convert(pw->pw_gecos).c_str(), (char*)match.c_str(), match.size()) == 0;

			if (!matched) {
				// check "generated" email address
				string email = string(pw->pw_name) + "@" + m_config->GetSetting("default_domain");
				if(ulFlags & EMS_AB_ADDRESS_LOOKUP)
					matched = (email == match);
				else
					matched = stristr((char*)email.c_str(), (char*)match.c_str());
			}
		
			if (matched)
				lObjects.push_back(objectsignature_t(tostring(pw->pw_uid),getDBSignature(tostring(pw->pw_uid), type) + pw->pw_gecos + pw->pw_name));
		}
		endpwent();
		pthread_mutex_unlock(m_plugin_lock);

		// find emailaddress in database, since this prop is either in the db or "generated" <username@domain>
		try {
			char *search_props[] = { OP_EMAILADDRESS, NULL };
			list<objectsignature_t> lObjectsTmp = DBPlugin::searchObjects(match, type, search_props, NULL, ulFlags);
			list<objectsignature_t>::iterator iterObject;

			for (iterObject = lObjectsTmp.begin(); iterObject != lObjectsTmp.end(); iterObject++) {
				// the DBPlugin returned the DB signature, so we need to prepend this with the gecos signature
				getpwuid_r(atoi(iterObject->id.c_str()), &pws, buffer, PWBUFSIZE, &pw);

				errnoCheck(iterObject->id);

				if (pw == NULL)	// object not found anymore
					continue;

				lObjects.push_back(objectsignature_t(iterObject->id, iterObject->signature + pw->pw_gecos + pw->pw_name)); // set unix signature
			}
		}
		catch (objectnotfound &e) {
			// Ignore exception, we will check lObjects.empty() later.
		} // All other exceptions should be thrown further up the chain.

		lObjects.sort();
		lObjects.unique();

	} else if(type == USEROBJECT_TYPE_GROUP) {
		gid_t mingid = fromstring<const char *, gid_t>(m_config->GetSetting("min_group_gid"));
		gid_t maxgid = fromstring<const char *, gid_t>(m_config->GetSetting("max_group_gid"));
		struct group grs, *gr = NULL;

		pthread_mutex_lock(m_plugin_lock);
		setgrent();
		while (1) {
			getgrent_r(&grs, buffer, PWBUFSIZE, &gr);
			if (gr == NULL)
				break;

			if (gr->gr_gid < mingid || gr->gr_gid >= maxgid)
				continue;

			if(ulFlags & EMS_AB_ADDRESS_LOOKUP)
				matched = stricmp(gr->gr_name, (char*)match.c_str()) == 0;
			else
				matched = stristr(gr->gr_name, (char*)match.c_str());
		
			if (matched)
				lObjects.push_back(objectsignature_t(tostring(gr->gr_gid),gr->gr_name));
		}
		endgrent();
		pthread_mutex_unlock(m_plugin_lock);
	}

	if (lObjects.empty())
		throw objectnotfound(string("unix_plugin: no match: ") + match);

	return lObjects;
}

auto_ptr<objectdetails_t> UnixUserPlugin::getPublicStoreDetails() throw(std::exception)
{
	throw notsupported("public store details");
}

auto_ptr<serverdetails_t> UnixUserPlugin::getServerDetails(const string &server) throw(std::exception)
{
	throw notsupported("server details");
}
void UnixUserPlugin::getUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives) throw (std::exception)
{
	throw notsupported("user count");
}

auto_ptr<map<objectid_t, objectdetails_t> > UnixUserPlugin::getObjectDetails(userobject_type_t type, list<objectid_t> &objectids) throw (std::exception) {
	map<objectid_t,objectdetails_t> *mapdetails = new map<objectid_t,objectdetails_t>;

	if (objectids.empty())
		return auto_ptr<map<objectid_t, objectdetails_t> >(mapdetails);

	auto_ptr<objectdetails_t> uDetails;
	objectdetails_t details;

	list<objectid_t>::iterator iterID;
	for (iterID = objectids.begin(); iterID != objectids.end(); iterID++) {
		try {
			uDetails = this->getObjectDetails(type, *iterID);
		}
		catch (objectnotfound &e) {
			// ignore not found error
			continue;
		}

		(*mapdetails)[*iterID] = (*uDetails.get());
	}
    
    return auto_ptr<map<objectid_t, objectdetails_t> >(mapdetails);
}


// -------------
// private
// -------------

auto_ptr<objectdetails_t> UnixUserPlugin::objectdetailsFromPwent(struct passwd *pw) {
	auto_ptr<objectdetails_t> ud;
	userobject_type_t type;
	string gecos;
	size_t comma;

	if (strcmp(pw->pw_shell, m_config->GetSetting("non_login_shell")) != 0)
		type = USEROBJECT_TYPE_USER;
	else
		type = USEROBJECT_TYPE_NONACTIVE;

	ud = auto_ptr<objectdetails_t>(new objectdetails_t(type));

	ud->SetPropString(OB_PROP_S_LOGIN, string(pw->pw_name));
	ud->SetPropBool(OB_PROP_B_NONACTIVE, type == USEROBJECT_TYPE_NONACTIVE);

	gecos = m_iconv->convert(pw->pw_gecos);

	// gecos may contain room/phone number etc. too
	comma = gecos.find(",");
	if (comma != string::npos) {
		ud->SetPropString(OB_PROP_S_FULLNAME, gecos.substr(0,comma));
	} else {
		ud->SetPropString(OB_PROP_S_FULLNAME, gecos);
	}

	if (!strcmp(pw->pw_passwd, "x")) {
		// shadow password entry
		struct spwd spws, *spw = NULL;
		char sbuffer[PWBUFSIZE];

		getspnam_r(pw->pw_name, &spws, sbuffer, PWBUFSIZE, &spw);
		if (spw == NULL) {
			// invalid entry, must have a shadow password set in this case
			// throw objectnotfound(ud->id);
			// too bad that the password couldn't be found, but it's not that critical
			m_logger->Log(EC_LOGLEVEL_WARNING, "Warning: unable to find password for user '%s', errno: %s", pw->pw_name, strerror(errno));
			ud->SetPropString(OB_PROP_S_PASSWORD, string("x"));	// set invalid password entry, cannot login without a password
		} else {
			ud->SetPropString(OB_PROP_S_PASSWORD, string(spw->sp_pwdp));
		}
	} else if (!strcmp(pw->pw_passwd, "*") || !strcmp(pw->pw_passwd, "!")){
		throw objectnotfound("");
	} else {
		ud->SetPropString(OB_PROP_S_PASSWORD, string(pw->pw_passwd));
	}
	
	// This may be overridden by settings in the database
	ud->SetPropString(OB_PROP_S_EMAIL, string(pw->pw_name) + "@" + m_config->GetSetting("default_domain"));

	return ud;
}

auto_ptr<objectdetails_t> UnixUserPlugin::objectdetailsFromGrent(struct group *gr) {
	auto_ptr<objectdetails_t> gd(new objectdetails_t(USEROBJECT_TYPE_GROUP));

	gd->SetPropString(OB_PROP_S_LOGIN, string(gr->gr_name));
	gd->SetPropString(OB_PROP_S_FULLNAME, string(gr->gr_name));

	return gd;
}

std::string UnixUserPlugin::getDBSignature(const objectid_t &id, userobject_type_t type)
{
	string strQuery;
	string strSubQuery;
	DB_RESULT_AUTOFREE lpResult(m_lpDatabase);
	DB_ROW lpDBRow = NULL;
	ECRESULT er = erSuccess;

	strSubQuery = 
		"SELECT id FROM " + (string)DB_OBJECT_TABLE + " "
		"WHERE externid = '" + m_lpDatabase->Escape(id) + "'"
		" AND objecttype = " + stringify(type);

	strQuery =
		"SELECT op.value "
		"FROM " + (string)DB_OBJECTPROPERTY_TABLE + " as op "
		"WHERE op.propname = '" + OP_MODTIME + "' "
			"AND op.objectid=("+ strSubQuery +")";

	// TODO: rewrite with JOIN ?

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		return string();

	lpDBRow = m_lpDatabase->FetchRow(lpResult);
	if (lpDBRow == NULL || lpDBRow[0] == NULL)
		return string();

	return lpDBRow[0];
}

void UnixUserPlugin::errnoCheck(string user) {
	if (errno) {
		char buffer[256];
		char *retbuf;
		retbuf = strerror_r(errno, buffer, 256);

		// from the getpwnam() man page: (notice the last or...)
		//  ERRORS
		//    0 or ENOENT or ESRCH or EBADF or EPERM or ...
		//    The given name or uid was not found.

		switch (errno) {
			// 0 is handled in top if()
		case ENOENT:
		case ESRCH:
		case EBADF:
		case EPERM:
			// calling function must check pw == NULL to throw objectnotfound()
			break;

		default:
			// broken system .. do not delete user from database
			throw runtime_error(string("unable to query for user ")+user+string(". Error: ")+retbuf);
		};

	}
}

