/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

// -*- Mode: c++ -*-
#ifndef PLUGIN_H
#define PLUGIN_H

// define to see which exception is thrown from a plugin
//#define EXCEPTION_DEBUG

#include "ECDefs.h"
#include "ZarafaUser.h"

#include <list>
#include <map>
#include <string>
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <memory>
#include <pthread.h>
#include "IECStatsCollector.h"

using namespace std;

class objectsignature_t {
public:
    objectsignature_t(objectid_t i, const std::string &s) : id(i), signature(s) {};
    objectsignature_t() : id(""), signature("") {};
    bool operator== (const objectsignature_t &sig) { return id == sig.id; };
    bool operator< (const objectsignature_t &sig) { return id < sig.id; } ;
    objectid_t id;
    std::string signature;
};

class ECConfig;
class ECLogger;

/** Main user plugin interface
 *
 * The user plugin interface defines methods for user management.
 */
class UserPlugin {
public:
	UserPlugin(pthread_mutex_t *pluginlock, ECConfig *config, ECLogger *logger, IECStatsCollector *lpStatsCollector, bool bHosted, bool bDistributed) : m_plugin_lock(pluginlock), p_config(config), m_config(NULL), m_logger(logger), m_lpStatsCollector(lpStatsCollector), m_bHosted(bHosted), m_bDistributed(bDistributed) { };

	virtual ~UserPlugin() {};

	virtual void InitPlugin() throw(std::exception) = 0;

	// Reading information
	virtual objectsignature_t 			resolveName(userobject_type_t type, const string &name, const string &company) throw(std::exception) = 0;
	virtual objectsignature_t 			authenticateUser(const string &username, const string &companyname, const string &password) throw(std::exception) = 0;
	virtual list<objectsignature_t> 	getAllObjects(userobject_type_t type, const objectid_t &company) throw(std::exception) = 0;
	virtual auto_ptr<objectdetails_t> 	getObjectDetails(userobject_type_t type, const objectid_t &objectid) throw(std::exception) = 0;
	virtual auto_ptr<map<objectid_t, objectdetails_t> > getObjectDetails(userobject_type_t type, list<objectid_t> &objectids) throw(std::exception) = 0;
	virtual list<objectsignature_t> 	getSubObjectsForObject(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype) throw(std::exception) = 0;
	virtual list<objectsignature_t>		getParentObjectsForObject(userobject_relation_t relation, userobject_type_t childtype, const objectid_t &childobject, userobject_type_t parenttype) throw(std::exception) = 0;
	virtual list<objectsignature_t> 	searchObject(userobject_type_t type, const string &match, unsigned int ulFlags) throw(std::exception) = 0;
	virtual auto_ptr<objectdetails_t>	getPublicStoreDetails() throw(std::exception) = 0;
	virtual auto_ptr<serverdetails_t>	getServerDetails(const string &server) throw(std::exception) = 0;
	virtual void						getUserCount(unsigned int *lpulUsers, unsigned int *lpulNonActives) throw (std::exception) = 0;
	
	// Writing information (optional)
	virtual int 						changeObject(userobject_type_t type, const objectid_t &id, objectdetails_t details, std::list<std::string> *lpRemove) throw(std::exception) = 0;
	virtual objectsignature_t 			createObject(userobject_type_t type, objectdetails_t details) throw(std::exception) = 0;	
	virtual void 						deleteObject(userobject_type_t type, const objectid_t &id) throw(std::exception) = 0;
	
	virtual void 						addSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception) = 0;
	virtual void 						deleteSubObjectRelation(userobject_relation_t relation, userobject_type_t parenttype, const objectid_t &parentobject, userobject_type_t childtype, const objectid_t &childobject) throw(std::exception) = 0;
	
	// Quota 
	virtual auto_ptr<quotadetails_t> 	getQuota(userobject_type_t type, const objectid_t &id, bool bGetUserDefault) throw(std::exception) = 0;
	virtual int 						setQuota(userobject_type_t type, const objectid_t &id, quotadetails_t quotadetails) throw(std::exception) = 0;

protected:
	pthread_mutex_t *m_plugin_lock; ///< Pointer to pthread mutex
	ECConfig *p_config; ///< Pointer to parent (server) configuration manager.
	ECConfig *m_config; ///< Pointer to local configuration manager.
	ECLogger *m_logger; ///< Pointer to logger
	IECStatsCollector *m_lpStatsCollector; ///< Pointer to statscollector

	bool m_bHosted;
	bool m_bDistributed;
};

class objectnotfound: public runtime_error {
public:
	objectnotfound(const string &arg) : runtime_error(arg) {
#ifdef EXCEPTION_DEBUG
		cerr << "objectnotfound exception: " << arg << endl;
#endif
	}
};

class toomanyobjects: public runtime_error {
public:
	toomanyobjects(const string &arg) : runtime_error(arg) {
#ifdef EXCEPTION_DEBUG
		cerr << "toomanyobjects exception: " << arg << endl;
#endif
	}
};

class data_error: public runtime_error {
public:
	data_error(const string &arg) : runtime_error(arg) {
#ifdef EXCEPTION_DEBUG
		cerr << "data_error exception: " << arg << endl;
#endif
	}
};

class notimplemented: public runtime_error {
public:
	notimplemented(const string &arg) : runtime_error(arg) {
#ifdef EXCEPTION_DEBUG
		cerr << "notimplemented exception: " << arg << endl;
#endif
	}
};

class notsupported : public runtime_error {
public:
	notsupported(const string &arg) : runtime_error(arg) {
#ifdef EXCEPTION_DEBUG
		cerr << "notsupported exception: " << arg << endl;
#endif
	}
};

class login_error: public runtime_error {
public:
	login_error(const string &arg) : runtime_error(arg) {
#ifdef EXCEPTION_DEBUG
		cerr << "login_error exception: " << arg << endl;
#endif
	}
};

template <class Tin, class Tout>
static inline Tout fromstring(Tin s) {
	istringstream i(s);
	Tout res;
	i >> res;
	return res;
}

template <class Tin>
static inline string tostring(Tin i) {
	ostringstream o;
	o << i;
	return o.str();
}

#endif
