/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "ECLogger.h"

#include <sys/stat.h>
#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif

#include "ECSoapServerConnection.h"
#include "ECServerEntrypoint.h"
#include "ECStatsCollector.h"
#include "ECClientUpdate.h"

struct soap_connection_thread {
	ECSoapServerConnection*	lpSoapConnClass;
	struct soap*			lpSoap;
};


int create_pipe_socket(const char *unix_socket, ECLogger *lpLogger, bool bInit) {
	int s;
	struct sockaddr_un saddr;
	memset(&saddr, 0, sizeof(struct sockaddr_un));

	s = socket(PF_UNIX, SOCK_STREAM, 0);
	if(s < 0) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create AF_UNIX socket.");
		return -1;
	}
	
	memset(&saddr,0,sizeof(saddr));
	saddr.sun_family = AF_UNIX;
	memcpy(saddr.sun_path,unix_socket,strlen(unix_socket)+1);
	
	unlink(unix_socket);

	if(bind(s,(struct sockaddr*)&saddr,2 + strlen(unix_socket) ) < 0) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to bind unix socket %s", unix_socket);
		close(s);
		return -1;
	}
	
	chmod(unix_socket,0666);
	
	if(listen(s,8) == -1) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Can't listen on unix socket %s", unix_socket);
		close(s);
		return -1;
	}

	return s;
}

/* ALERT! Big hack!
 *
 * This function relocates an open file descriptor to a new file descriptor above 1024. The
 * reason we do this, is because although we support many open fd's up to FD_SETSIZE, libraries
 * that we use may not (most notably libldap). This means that if a new socket is allocated within
 * libldap as socket 1025, libldap will fail because it was compiled with FD_SETSIZE=1024. To fix
 * this problem, we make sure that most FD's under 1024 are free for use by external libraries, while
 * we use the range 1024 -> 8192
 */
int relocate_fd(int fd, ECLogger *lpLogger) {
	// If we only have a 1024-fd limit, just return the original fd
	if(getdtablesize() <= 1024)
		return fd;

	int relocated = fcntl(fd, F_DUPFD, 1024);
	
	if(relocated < 0) {
		// OMG we have more than FD_SETSIZE-1024 sockets in use ?? Expect problems!
		lpLogger->Log(EC_LOGLEVEL_FATAL, "WARNING: Out of file descriptors, more than %d sockets in use. You cannot increase this value, so you must decrease socket usage.", getdtablesize()-1024);
		relocated = fd; // might as well try using FD's under 1024 ....
	} else {
		close(fd);
	}
	
	return relocated;
}

// Soap call fserveloop
// Called after successful invocation of a server operation in the server loop, 
// immediately after sending the response to a client.
// can be used for all connections
int gsoap_fserveloop(struct soap *soap) {
	// cleanup memory on pipe socket which lives forever
	soap_destroy(soap);
	soap_end(soap);

	return SOAP_OK;
}

/*
 * Handles the HTTP GET command from soap, only the client update install may be downloaded.
 *
 * This function can only be called when client_update_enabled is set to yes.
 */
static int http_get(struct soap *soap) 
{
	int nRet = 404;

	if (!soap || !soap->path)
		goto exit;

	if (strncmp(soap->path, "/autoupdate/", strlen("/autoupdate/")) == 0) {
		nRet = HandleClientUpdate(soap);
	}

exit:
	soap_end_send(soap); 

	return nRet;
}

ECSoapServerConnection::ECSoapServerConnection(ECConfig* lpConfig, ECLogger* lpLogger, int* lpQuit)
{
	unsigned int ulStackSize;

	// Default socket settings
	m_nMaxKeepAlive = 100;
	m_nRecvTimeout = 5;
	m_nSendTimeout = 60;

	// TCP
	memset(&m_sSoapTCP, 0, sizeof(m_sSoapTCP));
	m_bUseTCP = false;
	m_socketTCP = SOAP_INVALID_SOCKET;

	// SSL
	memset(&m_sSoapSSL, 0, sizeof(m_sSoapSSL));
	m_bUseSSL = false;
	m_socketSSL = SOAP_INVALID_SOCKET;
	
	// Pipe
	memset(&m_sSoapPipe, 0, sizeof(m_sSoapPipe));
	m_socketPipe = SOAP_INVALID_SOCKET;
	m_bUsePipe = false;

	//Init mutex for connection list
	pthread_mutex_init(&m_hMutexConnThread, NULL);

	// Logger
	m_lpLogger = lpLogger;

	// Fast quit possibility
	m_lpQuit = lpQuit;


	// Connection attribute and stacksize
	// NOTE:
	// We're going to create detached threads that are going to handle the incomming soap call
	// Otherwise, we would have to pthread_join() the threads again to clean the used memory
	pthread_attr_init(&m_thread_attrib);
	pthread_attr_setdetachstate(&m_thread_attrib, PTHREAD_CREATE_DETACHED);
	ulStackSize = atoi(lpConfig->GetSetting("thread_stacksize"));
	if (ulStackSize == 0 || ulStackSize == INT_MAX) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Invalid thread_stacksize found: %s. Reverting to 512 Kb.", lpConfig->GetSetting("thread_stacksize"));
		ulStackSize = 512;		// default 512 KB of stack space per thread
	}
	pthread_attr_setstacksize(&m_thread_attrib, ulStackSize * 1024);

}

ECSoapServerConnection::~ECSoapServerConnection(void)
{
	Destroy();
	
	pthread_attr_destroy(&m_thread_attrib);

	// remove mutex for connection list
	pthread_mutex_destroy(&m_hMutexConnThread);
}

ECRESULT ECSoapServerConnection::ListenTCP(const char* lpServerName, int nServerPort, bool bEnableGET)
{
	ECRESULT	er = erSuccess;
	int			socket = SOAP_INVALID_SOCKET;
	struct soap	sSoap = {0};
	

	if(lpServerName == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	//init soap
	soap_init2(&sSoap, SOAP_IO_KEEPALIVE | SOAP_XML_TREE, SOAP_IO_KEEPALIVE | SOAP_XML_TREE);

	if (bEnableGET)
		sSoap.fget = http_get;

	sSoap.bind_flags = SO_REUSEADDR;

	socket = soap_bind(&sSoap, lpServerName, nServerPort, 100);
	if (socket < 0) {
		soap_set_fault(&sSoap);
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to start server on port %d: %s", nServerPort, sSoap.fault->faultstring);
		er = ZARAFA_E_CALL_FAILED;
		goto exit;
	}

	m_sSoapTCP = sSoap;
	m_bUseTCP = true;
	m_socketTCP = socket;

exit:
	//TODO: Free sSoap on error?

	return er;
}

ECRESULT ECSoapServerConnection::ListenSSL(const char* lpServerName, int nServerPort, bool bEnableGET,
										   const char* lpszKeyFile, const char* lpszKeyPass,
										   const char* lpszCAFile, const char* lpszCAPath)
{
	ECRESULT	er = erSuccess;
	int			socket = SOAP_INVALID_SOCKET;
	struct soap	sSoap = {0};

	if(lpServerName == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	soap_init2(&sSoap, SOAP_IO_KEEPALIVE | SOAP_XML_TREE, SOAP_IO_KEEPALIVE | SOAP_XML_TREE);

	if (bEnableGET)
		sSoap.fget = http_get;

	if (soap_ssl_server_context(
			&sSoap,
			SOAP_SSL_DEFAULT,	// we set SSL_VERIFY_PEER and more soon ourselves
			lpszKeyFile,		// key file
			lpszKeyPass,		// key password
			lpszCAFile,			// CA certificate file which signed clients
			lpszCAPath,			// CA certificate path of thrusted sources
			NULL,				// dh file, null == rsa
			NULL,				// create random data on the fly (/dev/urandom is slow .. create file?)
			"zarafa")			// unique name for ssl session cache
		)
	{
		soap_set_fault(&sSoap);
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to setup ssl context: %s", *soap_faultdetail(&sSoap));
		er = ZARAFA_E_CALL_FAILED;
		goto exit;
	}
	
	// request certificate from client, is OK if not present.
	SSL_CTX_set_verify(sSoap.ctx, SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE, NULL);

	sSoap.bind_flags = SO_REUSEADDR;

	socket = soap_bind(&sSoap, lpServerName, nServerPort, 100);
	if (socket < 0) {
		soap_set_fault(&sSoap);
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to start SSL server on port %d: %s", nServerPort, sSoap.fault->faultstring);
		er = ZARAFA_E_CALL_FAILED;
		goto exit;
	}

	m_sSoapSSL = sSoap;
	m_bUseSSL = true;
	m_socketSSL = socket;

exit:
	//TODO: Free sSoap on error?

	return er;
}

ECRESULT ECSoapServerConnection::ListenPipe(const char* lpPipeName)
{
	ECRESULT	er = erSuccess;
	int			sPipe = -1;
	struct soap	sSoap = {0};

	if(lpPipeName == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	//init soap
	soap_init2(&sSoap, SOAP_IO_KEEPALIVE | SOAP_XML_TREE | SOAP_ENC_XML, SOAP_IO_KEEPALIVE | SOAP_XML_TREE | SOAP_ENC_XML);
	
	// Create a unix or windows pipe
	m_strPipeName = lpPipeName;
	sPipe = create_pipe_socket(m_strPipeName.c_str(), m_lpLogger, true);

	if (sPipe == -1) {
		er = ZARAFA_E_CALL_FAILED;
		goto exit;
	}

	sSoap.master = sPipe;
	m_sSoapPipe = sSoap;
	m_socketPipe = sPipe;
	m_bUsePipe = true;

exit:
	return er;
}

void ECSoapServerConnection::Destroy()
{

	if (m_bUseTCP && m_socketTCP != SOAP_INVALID_SOCKET) {
		m_bUseTCP = false;
		m_socketTCP = SOAP_INVALID_SOCKET;

		soap_destroy(&m_sSoapTCP);	// dealloc C++ data
		soap_end(&m_sSoapTCP);		// dealloc data and clean up
		soap_done(&m_sSoapTCP);		// Reset and detach context: close master/slave sockets and remove callbacks 
	}

	if (m_bUseSSL && m_socketSSL != SOAP_INVALID_SOCKET) {
		m_bUseSSL = false;
		m_socketSSL = SOAP_INVALID_SOCKET;

		soap_destroy(&m_sSoapSSL);	// dealloc C++ data
		soap_end(&m_sSoapSSL);		// dealloc data and clean up
		soap_done(&m_sSoapSSL);		// Reset and detach context: close master/slave sockets and remove callbacks
	}

	// Close pipe socket
	if (m_bUsePipe && m_socketPipe != SOAP_INVALID_SOCKET) {
		m_bUsePipe = false;
		m_socketPipe = SOAP_INVALID_SOCKET;

		soap_destroy(&m_sSoapPipe);	// dealloc C++ data
		soap_end(&m_sSoapPipe);		// dealloc data and clean up
		soap_done(&m_sSoapPipe);	// Reset and detach context: close master/slave sockets and remove callbacks
	}
}

// Set socket time out for TCP and SSL, call this before ListenTCP or ListenSSL
ECRESULT ECSoapServerConnection::SetSocketTimeout(int nRecvTimeout, int nSendTimeout)
{
	ECRESULT er = erSuccess;

	m_nRecvTimeout = nRecvTimeout;
	m_nSendTimeout = nSendTimeout;

	return er;
}

// Set keep alive for TCP and SSL, call this before ListenTCP or ListenSSL
ECRESULT ECSoapServerConnection::SetMaxKeepAliveRequests(int nMaxKeepAlive)
{
	ECRESULT er = erSuccess;

	m_nMaxKeepAlive = nMaxKeepAlive;

	return er;
}

ECRESULT ECSoapServerConnection::Select()
{
	ECRESULT		er = erSuccess;
	int				max = 0;
	int				ret;
	struct timeval	tv;

	//  init fd
	FD_ZERO(&m_fds);

	// Set interval
	tv.tv_sec = 10;
	tv.tv_usec = 0;

	// Get the max socket value
	if(m_bUsePipe)
		max = m_socketPipe;

	if(m_bUseSSL)
		max = max > m_socketSSL ? max : m_socketSSL;
	
	if(m_bUseTCP)
		max = max > m_socketTCP ? max : m_socketTCP;

	// Add sockets
	if (m_bUseTCP && m_socketTCP != SOAP_INVALID_SOCKET)
		FD_SET(m_socketTCP, &m_fds);

	if (m_bUseSSL && m_socketSSL != SOAP_INVALID_SOCKET)
		FD_SET(m_socketSSL, &m_fds);
	
	if (m_bUsePipe && m_socketPipe != SOAP_INVALID_SOCKET)
		FD_SET(m_socketPipe, &m_fds);

	ret = select(max+1, &m_fds, NULL, NULL, &tv);

	if (ret < 0) {
		if (errno != EINTR) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Socket error!");
			er = ZARAFA_E_NETWORK_ERROR;
		} else {
			// Time out limit
			er = ZARAFA_E_TIMEOUT;
		}
		goto exit;
	} else if (ret == 0) {
		// zero handles, next time better
		er = ZARAFA_E_TIMEOUT;
		goto exit;
	}

exit:
	return er;
}

ECRESULT ECSoapServerConnection::Accept()
{
	ECRESULT			er = erSuccess;
	struct soap*		lpSoap = NULL;
	struct soap*		lptsoap = NULL;
	int					slave = SOAP_INVALID_SOCKET;
	const char*			lpszType = NULL;
	CONNECTION_TYPE		ulType;
	
	int err = 0;
	pthread_t threadid;
	
	soap_connection_thread*	lpsSoapConnThread = NULL;


// seems we have data waiting
	if (m_bUseTCP && m_socketTCP != SOAP_INVALID_SOCKET && FD_ISSET(m_socketTCP, &m_fds)) {
		lpSoap = &m_sSoapTCP;

		slave = soap_accept(lpSoap);
		lpSoap->socket = slave = relocate_fd(slave, m_lpLogger);
		lpSoap->max_keep_alive = m_nMaxKeepAlive;
		lpSoap->recv_timeout = m_nRecvTimeout;
		lpSoap->send_timeout = m_nSendTimeout;  // Timeout after x seconds stall on send 
		lpSoap->fserveloop = gsoap_fserveloop;
	}
	else if (m_bUseSSL && m_socketSSL != SOAP_INVALID_SOCKET && FD_ISSET(m_socketSSL, &m_fds)) {
		lpSoap = &m_sSoapSSL;

		slave = soap_accept(lpSoap);
		lpSoap->socket = slave = relocate_fd(slave, m_lpLogger);
		lpSoap->max_keep_alive = m_nMaxKeepAlive;
		lpSoap->recv_timeout = m_nRecvTimeout;
		lpSoap->send_timeout = m_nSendTimeout;  // Timeout after x seconds stall on send 
		lpSoap->fserveloop = gsoap_fserveloop;
	}
	else if (m_bUsePipe && m_socketPipe != SOAP_INVALID_SOCKET && FD_ISSET(m_socketPipe, &m_fds)) {
		struct sockaddr_un	peer;
		socklen_t			n;

		memset(&peer, 0, sizeof(struct sockaddr_un));
		n = sizeof(struct sockaddr_un);

		lpSoap = &m_sSoapPipe;
		slave = accept(m_socketPipe, (struct sockaddr*)&peer, &n);
		slave = relocate_fd(slave, m_lpLogger);
		
		lpSoap->socket = slave;
		lpSoap->ip = 0;
		lpSoap->max_keep_alive = 0;
		lpSoap->recv_timeout = 0;
		lpSoap->fserveloop = gsoap_fserveloop;
	}
	else {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Had socket data, but disappeared.");

		er = ZARAFA_E_NETWORK_ERROR;
		goto exit;
	}

	if (slave == SOAP_INVALID_SOCKET) {
		soap_set_fault(lpSoap);
#ifdef DEBUG
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "%s: %s", *soap_faultstring(lpSoap), *soap_faultdetail(lpSoap));
#endif
		// Ignore the error, it is most probably a timeout
		// In debug mode, we dump some stats about the server
		er = erSuccess;
		goto exit;
	}

	// abort new incoming connection when we actually want to stop the server
	if (m_lpQuit && *m_lpQuit) {
		er = ZARAFA_E_END_OF_SESSION;
		goto exit;
	}

	// Connection type
	if(m_bUseTCP && m_socketTCP == lpSoap->master) {
		ulType = CONNECTION_TYPE_TCP;
		lpszType = "TCP";
	} else if(m_bUseSSL && m_socketSSL == lpSoap->master) {
		ulType = CONNECTION_TYPE_SSL;
		lpszType = "SSL";
	} else if(m_bUsePipe && m_socketPipe == lpSoap->master) {
		ulType = CONNECTION_TYPE_NAMED_PIPE;
		lpszType = "pipe";
	} else {
		er = ZARAFA_E_NETWORK_ERROR;
		goto exit;
	}

	m_lpLogger->Log(EC_LOGLEVEL_INFO, "Starting %s worker thread for request from IP %d.%d.%d.%d", lpszType,
				(lpSoap->ip >> 24)&0xFF, (lpSoap->ip >> 16)&0xFF, (lpSoap->ip >> 8)&0xFF, lpSoap->ip&0xFF);

	lptsoap = soap_copy(lpSoap); // make a safe copy
	
	if (!lptsoap) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to copy incoming traffic");
		er = ZARAFA_E_NETWORK_ERROR;
		goto exit;
	}

	zarafa_new_soap_connection(ulType, lptsoap, &ECSoapServerConnection::ThreadStatusCallback, (void*)this);

	// ssl accept must be on the tsoap structure
	if (m_bUseSSL && m_socketSSL == lpSoap->master && soap_ssl_accept(lptsoap)) {
		// - plain text on the SSL socket
		// - client did not send certificate when required
		// - ... ?
		soap_set_fault(lptsoap);
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to accept SSL data from IP %d.%d.%d.%d", 
				(lpSoap->ip >> 24)&0xFF, (lpSoap->ip >> 16)&0xFF, (lpSoap->ip >> 8)&0xFF, lpSoap->ip&0xFF);
#ifdef DEBUG
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "%s: %s", *soap_faultstring(lptsoap), *soap_faultdetail(lptsoap));
#endif

		zarafa_end_soap_connection(lptsoap);

		soap_destroy(lptsoap); // dealloc C++ data 
		soap_end(lptsoap); // dealloc data and clean up 
		soap_done(lptsoap); // detach soap struct 
		free(lptsoap); 
		lptsoap = NULL;
		
		//Ignore, waiting for the next one
		er = erSuccess;
		goto exit;
	}

	lpsSoapConnThread = new soap_connection_thread;
	lpsSoapConnThread->lpSoapConnClass = this;
	lpsSoapConnThread->lpSoap = lptsoap;

	g_lpStatsCollector->Max(SCN_MAX_SOCKET_NUMBER, (LONGLONG)lptsoap->socket);

	// Note: The thread will start mostly after exit this function, on a bad moment when
	// the server exits and a thread will start and not added with AddConnThread(...).
	// the server crashed. In a normal situation it isn't a problem.
	err = pthread_create(&threadid, &m_thread_attrib, process_request, (void*)lpsSoapConnThread);
	if(err != 0) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create thread to handle request: %s", strerror(err));

		zarafa_end_soap_connection(lptsoap);

		soap_destroy(lptsoap); // dealloc C++ data 
		soap_end(lptsoap); // dealloc data and clean up 
		soap_done(lptsoap); // detach soap struct 
		free(lptsoap); 
		lptsoap = NULL;

		if(lpsSoapConnThread)
			delete lpsSoapConnThread;

		// Ignore, waiting for the next one
		er = erSuccess;
		goto exit;
	}

	g_lpStatsCollector->Increment(SCN_SERVER_CONNECTIONS);
	er = erSuccess;

exit:
	return er;
}

void ECSoapServerConnection::AddConnThread(pthread_t threadid)
{
	pthread_mutex_lock(&m_hMutexConnThread);

	m_ListConnThread.insert(std::map<pthread_t, std::string>::value_type(threadid, ""));

	pthread_mutex_unlock(&m_hMutexConnThread);
}

void ECSoapServerConnection::RemoveConnThread(pthread_t threadid)
{
	pthread_mutex_lock(&m_hMutexConnThread);

	m_ListConnThread.erase(std::set<pthread_t>::key_type(threadid));

	pthread_mutex_unlock(&m_hMutexConnThread);
}

unsigned int ECSoapServerConnection::GetConnThreadCount()
{
	unsigned int ulThreadCount = 0;

	pthread_mutex_lock(&m_hMutexConnThread);

	ulThreadCount = (unsigned int)m_ListConnThread.size();

	pthread_mutex_unlock(&m_hMutexConnThread);

	return ulThreadCount;
}

void ECSoapServerConnection::WaitAndCloseThreads()
{
	unsigned int ulCount = 0;
	unsigned ulThreads = GetConnThreadCount();

	while(ulThreads > 0 && ulCount < 60) {
		if ((ulCount % 5) == 0 && m_lpLogger)
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Still waiting for %d thread%c to exit.", ulThreads, ulThreads!=1?'s':' ');

		Sleep(1000);
		ulCount++;
		ulThreads = GetConnThreadCount();
	}

	if (m_lpLogger) {
		if (ulCount == 60)
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "%d threads did not exit within timeout.", ulThreads);
		else
			m_lpLogger->Log(EC_LOGLEVEL_WARNING, "All threads exited.");
	}
}

void* ECSoapServerConnection::process_request(void *param) 
{
	int	ret = 0;
	ECSoapServerConnection*	lpSoapConnClass;
	struct soap_connection_thread* lpsct = (struct soap_connection_thread*)param;
	struct soap*		lpSoap;
	
	if(param == NULL)
		goto exit;

	lpSoapConnClass = lpsct->lpSoapConnClass;
	lpSoap = lpsct->lpSoap;
	
	if(lpSoapConnClass == NULL || lpSoap == NULL)
		goto exit;

	// Add Thread for list
	lpSoapConnClass->AddConnThread( pthread_self() );
	g_lpStatsCollector->Increment(SCN_THREAD_COUNTER);
	
	// We have a receive timeout which keeps the amount of 'sleeping' threads
	// on the server down to a minimum, while also keeping the connection open long enough
	// for performance.

	if ((ret = soap_serve(lpSoap)) != SOAP_OK) {
		soap_set_fault(lpSoap);
		if (ret != SOAP_EOF) {
			lpSoapConnClass->GetLogger()->Log(EC_LOGLEVEL_ERROR, "Unable to handle incoming request from %d.%d.%d.%d. Error %d",
						  (lpSoap->ip >> 24)&0xFF, (lpSoap->ip >> 16)&0xFF, 
						  (lpSoap->ip >> 8)&0xFF, lpSoap->ip&0xFF, ret);
#ifdef DEBUG
			// this message is quite useless to end-users.
			lpSoapConnClass->GetLogger()->Log(EC_LOGLEVEL_ERROR, "%s. Details: %s", *soap_faultstring(lpSoap), *soap_faultdetail(lpSoap));
#endif
		}
	}

	lpSoapConnClass->GetLogger()->Log(EC_LOGLEVEL_INFO, "Thread exiting due to timeout or connection close");

	zarafa_end_soap_connection(lpSoap);

	soap_destroy(lpSoap); // dealloc C++ data 
	soap_end(lpSoap); // dealloc data and clean up 
	soap_done(lpSoap); // detach soap struct 
	free(lpSoap); 

	// Remove Thread for list
	lpSoapConnClass->RemoveConnThread( pthread_self() );

exit:
	if(lpsct)
		delete lpsct;

	g_lpStatsCollector->Increment(SCN_THREAD_COUNTER, -1);

	// Do not pthread_exit() because linuxthreads is broken and will not free any objects
	// pthread_exit(0);

	return NULL;
}

void ECSoapServerConnection::SetThreadStatus(pthread_t threadid, std::string strStatus) 
{
    std::map<pthread_t, std::string>::iterator iterThreads;
    
	pthread_mutex_lock(&m_hMutexConnThread);

	iterThreads = m_ListConnThread.find(std::set<pthread_t>::key_type(threadid));
	if(iterThreads != m_ListConnThread.end())
	    iterThreads->second = strStatus;

	pthread_mutex_unlock(&m_hMutexConnThread);

}

void ECSoapServerConnection::ThreadStatusCallback(pthread_t thread, std::string status, void* ulParam)
{
    ECSoapServerConnection *lpSSC = NULL;

    if(!ulParam)
        return;

    lpSSC = (ECSoapServerConnection *)ulParam;
    
    lpSSC->SetThreadStatus(thread, status);
}

void ECSoapServerConnection::DumpThreads() {
    std::map<pthread_t, std::string>::iterator iterThreads;
    
    m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "%d active threads:", m_ListConnThread.size());
	pthread_mutex_lock(&m_hMutexConnThread);

	for(iterThreads = m_ListConnThread.begin(); iterThreads != m_ListConnThread.end(); iterThreads++) {
	    m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "  Thread %u: %s", iterThreads->first, iterThreads->second.c_str());
	}
	pthread_mutex_unlock(&m_hMutexConnThread);
}
