/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * This is the Zarafa spooler.
 *
 *
 * The actual encoding is done by the inetmapi library.
 *
 * The spooler starts up, runs the queue once, and then
 * waits for changes in the outgoing queue. If any changes
 * occur, the whole queue is run again. This is done by having
 * an advise sink which is called when a table change is detected.
 * This advise sink unblocks the main (waiting) thread.
 */

#include "platform.h"

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <algorithm>
#include <signal.h>

#define USES_IID_IMAPIFolder
#define USES_IID_IMessage
#define USES_IID_IMsgStore

#include <inetmapi.h>

#include <mapi.h>
#include <mapix.h>
#include <mapiutil.h>
#include <mapidefs.h>

#include <mapiguid.h>
#include <ctype.h>

#include "IECUnknown.h"
#include "IECSpooler.h"
#include "IECServiceAdmin.h"
#include "IECSecurity.h"
#include "ECGuid.h"
#include "EMSAbTag.h"
#include "ECTags.h"
#include "ECABEntryID.h"
#include "CommonUtil.h"
#include "ECLogger.h"
#include "ECConfig.h"
#include "UnixUtil.h"
#include "my_getopt.h"
#include "ecversion.h"
#include "Util.h"
#include "stringutil.h"

#include "mapiext.h"
#include "edkmdb.h"
#include "edkguid.h"
#include "mapiguidext.h"
#include "mapicontact.h"

#define SMTP_HOST	"127.0.0.1"
#define MAX_DISCONNECTS 5
#define THREAD_DBG		"0x%08x: "
#define WHITESPACE	" \t\n\r"

using namespace std;

ECLogger *lpLogger;
ECConfig *lpConfig;

int				nQuit;
bool			bMessagesWaiting;
pthread_mutex_t	hMutexMessagesWaiting;
pthread_cond_t	hCondMessagesWaiting;
int				disconnects;

/* Threading */
pthread_t				mainthread;
pthread_mutex_t			threadmap_lock;
pthread_attr_t			thread_attrib;
map<SBinary, pthread_t>	threadmap;
typedef pair<SBinary, pthread_t> threadpair_t;
typedef map<SBinary, pthread_t>::iterator threaditer_t;


// WARNING only compares for min(len(a), len(b)) chars!!
bool operator <(SBinary left, SBinary right)
{
	return memcmp(left.lpb, right.lpb, (left.cb < right.cb) ? left.cb : right.cb) < 0;
}

/*
 * Threading:
 * The Outgoing queue check will be performed inside the main thread,
 * each outgoing message will be send out inside a new thread.
 * In each thread we will open a seperate connection to the server.
 */
struct thread_data {
	/* Thread to which this data belongs to */
	pthread_t		thread;

	/* Initialization data.
	 * Initialized before thread has started. */
	char			*lpszPath;
	IMAPISession	*lpAdminSession;
	IMAPISession	*lpMAPISession;
	char			*lpszSMTP;

	/* Data from outgoing queue.
	 * Initialized before thread has started. */
	char			*lpszSender;
	SBinary			sStore;
	SBinary			sMessage;
	ULONG			ulFlags;

	/* Data strutures.
	 * Initialized when the thread has been started. */
	LPMDB			lpMDB;
	IECSpooler		*lpSpooler;
	IECServiceAdmin	*lpServiceAdmin;
	LPADRBOOK		lpAddrBook;
};

void FreeThreadData(struct thread_data *lpData)
{
	if (!lpData)
		return;

	if (lpData->lpServiceAdmin)
		lpData->lpServiceAdmin->Release();
	if (lpData->lpSpooler)
		lpData->lpSpooler->Release();
	if (lpData->lpMDB)
		lpData->lpMDB->Release();
	if (lpData->lpMAPISession)
		lpData->lpMAPISession->Release();
	if (lpData->lpAdminSession)
		lpData->lpAdminSession->Release();
	if (lpData->lpAddrBook)
		lpData->lpAddrBook->Release();
	MAPIFreeBuffer(lpData);
}

HRESULT AllocThreadData(struct thread_data **lppData, char *lpszPath, LPMAPISESSION lpAdminSession = NULL,
						char *lpszSMTP = NULL, char *lpszSender = NULL,
						LPSBinary lpStore = NULL, LPSBinary lpMessage = NULL,
						ULONG ulFlags = 0)
{
	HRESULT hr = hrSuccess;
	struct thread_data *lpData = NULL;

	hr = MAPIAllocateBuffer(sizeof(struct thread_data), (void **)&lpData);
	if (hr != hrSuccess)
		goto exit;

	memset(lpData, 0, sizeof(struct thread_data));

	if (lpData) {
		hr = MAPIAllocateMore(strlen(lpszPath)+1, lpData, (void **)&lpData->lpszPath);
		if (hr != hrSuccess)
			goto exit;
		strcpy(lpData->lpszPath, lpszPath);
	}

    if (lpszSMTP) {
		hr = MAPIAllocateMore(strlen(lpszSMTP)+1, lpData, (void **)&lpData->lpszSMTP);
		if (hr != hrSuccess)
			goto exit;
		strcpy(lpData->lpszSMTP, lpszSMTP);
	}

    if (lpszSender) {
		hr = MAPIAllocateMore(strlen(lpszSender)+1, lpData, (void **)&lpData->lpszSender);
		if (hr != hrSuccess)
			goto exit;
		strcpy(lpData->lpszSender, lpszSender);
	}

	if (lpAdminSession) {
		hr = lpAdminSession->QueryInterface(IID_IMAPISession, (void**)&lpData->lpAdminSession);
		if (hr != hrSuccess)
			goto exit;
	}

    if (lpStore) {
		hr = MAPIAllocateMore(lpStore->cb, lpData, (void **)&lpData->sStore.lpb);
		if (hr != hrSuccess)
			goto exit;
		memcpy(lpData->sStore.lpb, lpStore->lpb, lpStore->cb);
		lpData->sStore.cb = lpStore->cb;
	}

	if (lpMessage) {
		hr = MAPIAllocateMore(lpMessage->cb, lpData, (void **)&lpData->sMessage.lpb);
		if (hr != hrSuccess)
			goto exit;
		memcpy(lpData->sMessage.lpb, lpMessage->lpb, lpMessage->cb);
		lpData->sMessage.cb = lpMessage->cb;
	}

	lpData->ulFlags = ulFlags;

	/*
	 * Initialization complete, the remaining variables should be initialized
	 * when the thread has started.
	 */
	*lppData = lpData;

exit:
	if (hr != hrSuccess && lpData)
		FreeThreadData(lpData);

	return hr;
}

HRESULT InitThreadData(struct thread_data *lpData)
{
	HRESULT		hr = hrSuccess;
	SPropValue	*lpsProp = NULL;
	IECUnknown	*lpMsgStore = NULL;

	/* When lpAdminSession is NULL we should create an admin session, otherwise
	 * we should read the message to determine the sender and create
	 * a session with that user.
	 * This is required for proper group expansion, when sending an email
	 * to group 'Everyone' it should not be possible to send the email to
	 * users that cannot be viewed (because they are in a different company.
	 * By using a usersession for email sending we will let the server
	 * handle all permissions and can correctly resolve everything. */
	if (lpData->lpAdminSession == NULL) {
		hr = HrOpenECAdminSession(&lpData->lpMAPISession, lpData->lpszPath, 0,
								  lpConfig->GetSetting("sslkey_file", "", NULL),
								  lpConfig->GetSetting("sslkey_pass", "", NULL));
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to open admin session. Error 0x%X", lpData->thread, hr);
			goto exit;
		}
		lpData->lpMAPISession->QueryInterface(IID_IMAPISession, (void**)&lpData->lpAdminSession);
	} else {
		hr = HrOpenECSession(&lpData->lpMAPISession, lpData->lpszSender, "", lpData->lpszPath, 0,
							 lpConfig->GetSetting("sslkey_file","",NULL),
							 lpConfig->GetSetting("sslkey_pass","",NULL));
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to open session for user %s. Error 0x%X",
						  lpData->thread, lpData->lpszSender, hr);
			goto exit;
		}
	}

	hr = lpData->lpMAPISession->OpenAddressBook(0, NULL, AB_NO_DIALOG, &lpData->lpAddrBook);
	if (hr != hrSuccess)
		goto exit;

	hr = HrOpenDefaultStore(lpData->lpMAPISession, &lpData->lpMDB);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to open default store for system account. Error 0x%X", lpData->thread, hr);
		goto exit;
	}

	hr = HrGetOneProp(lpData->lpMDB, PR_EC_OBJECT, &lpsProp);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to get Zarafa internal object", lpData->thread);
		goto exit;
	}

	lpMsgStore = (IECUnknown *)lpsProp->Value.lpszA;

	hr = lpMsgStore->QueryInterface(IID_IECSpooler, (void **)&lpData->lpSpooler);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Spooler interface not supported", lpData->thread);
		goto exit;
	}

	hr = lpMsgStore->QueryInterface(IID_IECServiceAdmin, (void **)&lpData->lpServiceAdmin);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "ServiceAdmin interface not supported", lpData->thread);
		goto exit;
	}

exit:
	if (lpsProp)
		MAPIFreeBuffer(lpsProp);

	return hr;
}

HRESULT running_server(char* szSMTP, char* szPath);

void sighandle(int sig) {
	// Win32 has unix semantics and therefore requires us to reset the signal handler.
	signal(SIGTERM , sighandle);
	signal(SIGINT  , sighandle);	// CTRL+C

	if (!nQuit)						// do not log multimple shutdown messages
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Termination requested, shutting down.");

	nQuit = true;

	// There is a small chance that you may have to wait 60 seconds here if we trigger
	// the signal just before pthread_cond_wait() and just after while(nQuit). We could
	// put a mutex in here just for that....

	// Trigger condition as it may be waiting for messages
	pthread_cond_signal(&hCondMessagesWaiting);
}


void print_help(char *name) {
	cout << "Usage:\n" << endl;
	cout << name << " [-F] [-h|--host <serverpath>] [-c|--config <configfile>] [smtp server]" << endl;
	cout << "  -F\t\tDo not run in the background" << endl;
	cout << "  -h path\tUse alternate connect path (e.g. file:///var/run/socket).\n\t\tDefault: file:///var/run/zarafa" << endl;
	cout << "  -V Print version info." << endl;
	cout << "  -c filename\tUse alternate config file (e.g. /etc/zarafa-spooler.cfg)\n\t\tDefault: /etc/zarafa/spooler.cfg" << endl;
	cout << "  smtp server: The name or IP-address of the SMTP server, overriding the configuration" << endl;
	cout << endl;
}

LONG __stdcall AdviseCallback(void *lpContext, ULONG cNotif, LPNOTIFICATION lpNotif)
{
	pthread_mutex_lock(&hMutexMessagesWaiting);
	for (ULONG i = 0; i < cNotif; i++) {
		if (lpNotif[i].info.tab.ulTableEvent != TABLE_ROW_DELETED) {
			bMessagesWaiting = true;
			pthread_cond_signal(&hCondMessagesWaiting);
			break;
		}
	}
	pthread_mutex_unlock(&hMutexMessagesWaiting);

	return 0;
}

HRESULT ExpandRecipientsRecursive(struct thread_data *lpData, IMessage *lpMessage,
								  IMAPITable *lpTable, LPSRestriction lpEntryRestriction, ULONG ulRecipType,
								  list<SBinary> *lpExpandedGroups, bool recurrence = true)
{
	HRESULT			hr = hrSuccess;
	IMAPITable		*lpContentsTable = NULL;
    LPSRowSet		lpsRowSet = NULL;
	LPDISTLIST		lpDistlist = NULL;
	ULONG			ulObj = 0;
	bool			bExpandSub = recurrence;

	SizedSPropTagArray(7, sptaColumns) = {
		7, {
			PR_ROWID,
			PR_DISPLAY_NAME,
			PR_SMTP_ADDRESS,
			PR_RECIPIENT_TYPE,
			PR_OBJECT_TYPE,
			PR_DISPLAY_TYPE,
			PR_ENTRYID,
		}
	};

	hr = lpTable->SetColumns((LPSPropTagArray)&sptaColumns, 0);
	if (hr != hrSuccess)
		goto exit;

	while (true) {
		LPSPropValue lpRowId;
		LPSPropValue lpEntryId;
		LPSPropValue lpDisplayType;
		LPSPropValue lpObjectType;
		LPSPropValue lpRecipType;
		LPSPropValue lpDisplayName;
		LPSPropValue lpEmailAddress;

		/* Perform cleanup first. */
		if (lpsRowSet) {
			FreeProws(lpsRowSet);
			lpsRowSet = NULL;
		}

		if (lpDistlist) {
			lpDistlist->Release();
			lpDistlist = NULL;
		}

		if (lpContentsTable) {
			lpContentsTable->Release();
			lpContentsTable = NULL;
		}

		/* Request group from table */
		hr = lpTable->QueryRows(1, 0, &lpsRowSet);
		if (hr != hrSuccess)
			goto exit;

		if (lpsRowSet->cRows != 1)
			break;

		/* From this point on we use 'continue' when something fails,
		 * since all errors are related to the current entry and we should
		 * make sure we resolve as many recipients as possible. */

		lpRowId = PpropFindProp(lpsRowSet->aRow[0].lpProps, lpsRowSet->aRow[0].cValues, PR_ROWID);
		lpEntryId = PpropFindProp(lpsRowSet->aRow[0].lpProps, lpsRowSet->aRow[0].cValues, PR_ENTRYID);
		lpDisplayType = PpropFindProp(lpsRowSet->aRow[0].lpProps, lpsRowSet->aRow[0].cValues, PR_DISPLAY_TYPE);
		lpObjectType = PpropFindProp(lpsRowSet->aRow[0].lpProps, lpsRowSet->aRow[0].cValues, PR_OBJECT_TYPE);
		lpRecipType = PpropFindProp(lpsRowSet->aRow[0].lpProps, lpsRowSet->aRow[0].cValues, PR_RECIPIENT_TYPE);
		lpDisplayName = PpropFindProp(lpsRowSet->aRow[0].lpProps, lpsRowSet->aRow[0].cValues, PR_DISPLAY_NAME);
		lpEmailAddress = PpropFindProp(lpsRowSet->aRow[0].lpProps, lpsRowSet->aRow[0].cValues, PR_SMTP_ADDRESS);

		/* lpRowId, lpRecipType, and lpDisplayType are optional.
		 * lpEmailAddress is only mandatory for MAPI_MAILUSER */
		if (!lpEntryId || !lpObjectType || !lpDisplayName)
			continue;

		/* By default we inherit the recipient type from parent */
		if (lpRecipType)
			ulRecipType = lpRecipType->Value.ul;

		if (lpObjectType->Value.ul == MAPI_MAILUSER) {
			if (!lpEmailAddress)
				continue;

			SRowSet sRowSMTProwSet;
			SPropValue sPropSMTPProperty[4];

			sRowSMTProwSet.cRows = 1;
			sRowSMTProwSet.aRow[0].cValues = 4;
			sRowSMTProwSet.aRow[0].lpProps = sPropSMTPProperty;

			sRowSMTProwSet.aRow[0].lpProps[0].ulPropTag = PR_EMAIL_ADDRESS;
			sRowSMTProwSet.aRow[0].lpProps[0].Value.lpszA = lpEmailAddress->Value.lpszA;

			sRowSMTProwSet.aRow[0].lpProps[1].ulPropTag = PR_SMTP_ADDRESS;
			sRowSMTProwSet.aRow[0].lpProps[1].Value.lpszA = lpEmailAddress->Value.lpszA;

			sRowSMTProwSet.aRow[0].lpProps[2].ulPropTag = PR_RECIPIENT_TYPE;
			sRowSMTProwSet.aRow[0].lpProps[2].Value.ul = ulRecipType; /* Inherit from parent group */

			sRowSMTProwSet.aRow[0].lpProps[3].ulPropTag = PR_DISPLAY_NAME;
			sRowSMTProwSet.aRow[0].lpProps[3].Value.lpszA = lpDisplayName->Value.lpszA;

			hr = lpMessage->ModifyRecipients(MODRECIP_ADD, (LPADRLIST)&sRowSMTProwSet);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to add e-mail address of %s from group",
							  lpData->thread, lpEmailAddress->Value.lpszA);
				continue;
			}
		} else {
			SBinary sEntryId;

			/* If we should recur further, just remove the group from the recipients list */
			if (!recurrence)
				goto remove_group;

			/* Only continue when this group has not yet been expanded previously */
			if (find(lpExpandedGroups->begin(), lpExpandedGroups->end(), lpEntryId->Value.bin) != lpExpandedGroups->end())
				goto remove_group;

			hr = lpData->lpAddrBook->OpenEntry(lpEntryId->Value.bin.cb, (LPENTRYID)lpEntryId->Value.bin.lpb,
											   NULL, 0, &ulObj, (LPUNKNOWN*)&lpDistlist);
			if (hr != hrSuccess)
				continue;
				
			if (ulObj != MAPI_DISTLIST) {
				hr = MAPI_E_NOT_FOUND; // Hard error - user is trying to send to someone who has probably been deleted ..
				lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to expand group %s from recipient list: wrong type, possibly deleted?",
				                              lpData->thread, lpDisplayName->Value.lpszA);
				goto exit;
			}

			hr = lpDistlist->GetContentsTable(0, &lpContentsTable);
			if (hr != hrSuccess)
				continue;

			hr = lpContentsTable->Restrict(lpEntryRestriction, 0);
			if (hr != hrSuccess)
				continue;

			/* Group has been expanded (because we successfully have the contents table) time
			 * to add it to our expanded group list. This has to be done or at least before the
			 * recursive call to ExpandRecipientsRecursive().*/
			hr = Util::HrCopyEntryId(lpEntryId->Value.bin.cb, (LPENTRYID)lpEntryId->Value.bin.lpb,
									 &sEntryId.cb, (LPENTRYID*)&sEntryId.lpb);
			lpExpandedGroups->push_back(sEntryId);

			/* Don't expand group Everyone or companies since both already contain all users
			 * which should be put in the recipient list. */
			bExpandSub = !(((lpDisplayType) ? lpDisplayType->Value.ul == DT_ORGANIZATION : false) ||
						   ((lpDisplayName) ? stricmp(lpDisplayName->Value.lpszA, "Everyone") == 0 : false));

			/* Start/Continue recursion */
			hr = ExpandRecipientsRecursive(lpData, lpMessage, lpContentsTable,
										   lpEntryRestriction, ulRecipType, lpExpandedGroups, bExpandSub);
			/* Ignore errors */

remove_group:
			/* Only delete row when the rowid is present */
			if (!lpRowId)
				continue;

			hr = lpMessage->ModifyRecipients(MODRECIP_REMOVE, (LPADRLIST)lpsRowSet);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to remove group %s from recipient list.",
							  lpData->thread, lpDisplayName->Value.lpszA);
				continue;
			}
		}
	}

exit:
	if (lpDistlist)
		lpDistlist->Release();

	if (lpsRowSet)
		FreeProws(lpsRowSet);

	if (lpContentsTable)
		lpContentsTable->Release();

    return hr;
}

HRESULT ExpandRecipients(struct thread_data *lpData, IMessage *lpMessage)
{
	HRESULT hr = hrSuccess;
	list<SBinary> lExpandedGroups;
	list<SBinary>::iterator iterGroups;
	IMAPITable *lpTable = NULL;
	LPSRestriction lpRestriction = NULL;
	LPSRestriction lpEntryRestriction = NULL;

	/*
	 * Setup group restriction:
	 * PR_OBJECT_TYPE == MAPI_DISTLIST && PR_ADDR_TYPE == "ZARAFA"
	 */
	hr = MAPIAllocateBuffer(sizeof(SRestriction), (LPVOID*)&lpRestriction);
	if(hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRestriction, (LPVOID*)&lpRestriction->res.resAnd.lpRes);
	if (hr != hrSuccess)
		goto exit;

	lpRestriction->rt = RES_AND;
	lpRestriction->res.resAnd.cRes = 2;

	hr = MAPIAllocateMore(sizeof(SPropValue), lpRestriction, (LPVOID*)&lpRestriction->res.resAnd.lpRes[0].res.resProperty.lpProp);
	if (hr != hrSuccess)
		goto exit;

	lpRestriction->res.resAnd.lpRes[0].rt = RES_PROPERTY;
	lpRestriction->res.resAnd.lpRes[0].res.resProperty.relop = RELOP_EQ;
	lpRestriction->res.resAnd.lpRes[0].res.resProperty.ulPropTag = PR_OBJECT_TYPE;
	lpRestriction->res.resAnd.lpRes[0].res.resProperty.lpProp->ulPropTag = PR_OBJECT_TYPE;
	lpRestriction->res.resAnd.lpRes[0].res.resProperty.lpProp->Value.ul = MAPI_DISTLIST;

	hr = MAPIAllocateMore(sizeof(SPropValue), lpRestriction, (LPVOID*)&lpRestriction->res.resAnd.lpRes[1].res.resProperty.lpProp);
	if (hr != hrSuccess)
		goto exit;

	lpRestriction->res.resAnd.lpRes[1].rt = RES_PROPERTY;
	lpRestriction->res.resAnd.lpRes[1].res.resProperty.relop = RELOP_EQ;
	lpRestriction->res.resAnd.lpRes[1].res.resProperty.ulPropTag = PR_ADDRTYPE;
	lpRestriction->res.resAnd.lpRes[1].res.resProperty.lpProp->ulPropTag = PR_ADDRTYPE;
	lpRestriction->res.resAnd.lpRes[1].res.resProperty.lpProp->Value.lpszA = "ZARAFA";

	/*
	 * Setup entry restriction:
	 * PR_ADDR_TYPE == "ZARAFA"
	 */
	hr = MAPIAllocateBuffer(sizeof(SRestriction), (LPVOID*)&lpEntryRestriction);
	if(hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateMore(sizeof(SPropValue), lpEntryRestriction, (LPVOID*)&lpEntryRestriction->res.resProperty.lpProp);
	if (hr != hrSuccess)
		goto exit;

	lpEntryRestriction->rt = RES_PROPERTY;
	lpEntryRestriction->res.resProperty.relop = RELOP_EQ;
	lpEntryRestriction->res.resProperty.ulPropTag = PR_ADDRTYPE;
	lpEntryRestriction->res.resProperty.lpProp->ulPropTag = PR_ADDRTYPE;
	lpEntryRestriction->res.resProperty.lpProp->Value.lpszA = "ZARAFA";

    hr = lpMessage->GetRecipientTable(0, &lpTable);
    if(hr != hrSuccess)
        goto exit;

	/* The first table we send with ExpandRecipientsRecursive() is the RecipientTable itself,
	 * we need to put a restriction on this table since the first time only the groups
	 * should be added to the recipients list. Subsequent calls to ExpandRecipientsRecursive()
	 * will send the group member table and will correct add the members to the recipients
	 * table. */
	hr = lpTable->Restrict(lpRestriction, 0);
	if (hr != hrSuccess)
		goto exit;

	/* ExpandRecipientsRecursive() will run recursively expanding each group
	 * it finds including all subgroups. It will use the lExpandedGroups list
	 * to protect itself for circular subgroup membership */
	hr = ExpandRecipientsRecursive(lpData, lpMessage, lpTable, lpEntryRestriction, MAPI_TO, &lExpandedGroups); 
	if (hr != hrSuccess)
		goto exit;

exit:
	for (iterGroups = lExpandedGroups.begin(); iterGroups != lExpandedGroups.end(); iterGroups++)
		MAPIFreeBuffer(iterGroups->lpb);

	if (lpTable)
		lpTable->Release();

	if (lpRestriction)
		MAPIFreeBuffer(lpRestriction);

	if (lpEntryRestriction)
		MAPIFreeBuffer(lpEntryRestriction);

	return hr;
}

HRESULT RewriteRecipients(struct thread_data *lpData, IMessage *lpMessage)
{
	HRESULT			hr = hrSuccess;
	IMAPITable		*lpTable = NULL;
	LPSRowSet		lpRowSet = NULL;
	LPSPropTagArray	lpRecipColumns = NULL;
	LPSPropValue	lpEmailAddress = NULL;
	LPSPropValue	lpEmailName = NULL;
	LPSPropValue	lpAddrType = NULL;
	LPSPropValue	lpEntryID = NULL;
	ULONG			cbNewEntryID;
	LPENTRYID		lpNewEntryID = NULL;

	char			*lpszFaxDomain = lpConfig->GetSetting("fax_domain");
	char			*lpszFaxInternational = lpConfig->GetSetting("fax_international");
	string			strFaxMail;
	LPMAILUSER		lpFaxMailuser = NULL;
	LPSPropValue	lpFaxNumbers = NULL;
	ULONG			ulObjType;
	ULONG			cValues;

	// contab email_offset: 0: business, 1: home, 2: primary (outlook uses string 'other')
	SizedSPropTagArray(3, sptaFaxNumbers) = { 3, {PR_BUSINESS_FAX_NUMBER, PR_HOME_FAX_NUMBER, PR_PRIMARY_FAX_NUMBER } };

	if (!lpszFaxDomain || strcmp(lpszFaxDomain, "") == 0)
		goto exit;

	hr = lpMessage->GetRecipientTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	// we need all columns when rewriting FAX to SMTP
	hr = lpTable->QueryColumns(TBL_ALL_COLUMNS, &lpRecipColumns);
	if (hr != hrSuccess)
		goto exit;
	
	hr = lpTable->SetColumns(lpRecipColumns, 0);
	if (hr != hrSuccess)
		goto exit;

	while (TRUE) {
		if (lpRowSet) {
			FreeProws(lpRowSet);
			lpRowSet = NULL;
		}

		hr = lpTable->QueryRows(1, 0, &lpRowSet);
		if (hr != hrSuccess)
			goto exit;

		if (lpRowSet->cRows == 0)
			break;

		lpEmailAddress = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_EMAIL_ADDRESS);
		lpEmailName = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_DISPLAY_NAME);
		lpAddrType = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_ADDRTYPE);
		lpEntryID = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_ENTRYID);

		if (!(lpEmailAddress && lpAddrType && lpEntryID && lpEmailName))
			continue;

		if (strcmp(lpAddrType->Value.lpszA, "FAX") == 0)
		{
			// rewrite FAX address to <number>@<faxdomain>
			string strName, strType, strEmailAddress;

			if (ECParseOneOff((LPENTRYID)lpEntryID->Value.bin.lpb, lpEntryID->Value.bin.cb, strName, strType, strEmailAddress) == hrSuccess) {
				// user entered manual fax address
				strFaxMail = strEmailAddress;
			} else {
				// check if entry is in contacts folder
				LPCONTAB_ENTRYID lpContabEntryID = (LPCONTAB_ENTRYID)lpEntryID->Value.bin.lpb;
				GUID* guid = (GUID*)&lpContabEntryID->muid;

				// check validity of lpContabEntryID
				if (sizeof(CONTAB_ENTRYID) > lpEntryID->Value.bin.cb ||
					*guid != PSETID_CONTACT_FOLDER_RECIPIENT ||
					lpContabEntryID->email_offset < 3 ||
					lpContabEntryID->email_offset > 5)
				{
					hr = MAPI_E_INVALID_PARAMETER;
					lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to convert FAX recipient, using %s", lpData->thread, lpEmailAddress->Value.lpszA);
					goto nextfax;
				}

				// 0..2 == reply to email offsets
				// 3..5 == fax email offsets
				lpContabEntryID->email_offset -= 3;

				hr = lpData->lpMAPISession->OpenEntry(lpContabEntryID->cbeid, (LPENTRYID)lpContabEntryID->abeid, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpFaxMailuser);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to convert FAX recipient, using %s", lpData->thread, lpEmailAddress->Value.lpszA);
					goto nextfax;
				}

				hr = lpFaxMailuser->GetProps((LPSPropTagArray)&sptaFaxNumbers, 0, &cValues, &lpFaxNumbers);
				if (FAILED(hr)) {
					lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to convert FAX recipient, using %s", lpData->thread, lpEmailAddress->Value.lpszA);
					goto nextfax;
				}

				if (lpFaxNumbers[lpContabEntryID->email_offset].ulPropTag != sptaFaxNumbers.aulPropTag[lpContabEntryID->email_offset]) {
					lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "No suitable FAX number found, using %s", lpData->thread, lpEmailAddress->Value.lpszA);
					goto nextfax;
				}

				strFaxMail = lpFaxNumbers[lpContabEntryID->email_offset].Value.lpszA;
			}

			strFaxMail += string("@") + lpszFaxDomain;
			if (strFaxMail[0] == '+') {
				strFaxMail = lpszFaxInternational + strFaxMail.substr(1, strFaxMail.length());
			}

			lpEmailAddress->Value.lpszA = (char*)strFaxMail.c_str();
			lpAddrType->Value.lpszA = (char*)"SMTP";
			// old value is stuck to the row allocation, so we can override it, but we also must free the new!
			ECCreateOneOff(lpEmailName->Value.lpszA, "SMTP", (char*)strFaxMail.c_str(), 0, &cbNewEntryID, &lpNewEntryID);
			lpEntryID->Value.bin.lpb = (LPBYTE)lpNewEntryID;
			lpEntryID->Value.bin.cb = cbNewEntryID;

			hr = lpMessage->ModifyRecipients(MODRECIP_MODIFY, (LPADRLIST)lpRowSet);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to set new FAX mail address for '%s' to '%s'", lpData->thread, lpEmailAddress->Value.lpszA, strFaxMail.c_str());
				goto nextfax;
			}

			lpLogger->Log(EC_LOGLEVEL_INFO, THREAD_DBG "Using new FAX mail address %s", lpData->thread, strFaxMail.c_str());

nextfax:
			hr = hrSuccess;

			if (lpNewEntryID)
				MAPIFreeBuffer(lpNewEntryID);
			lpNewEntryID = NULL;

			if (lpFaxMailuser)
				lpFaxMailuser->Release();
			lpFaxMailuser = NULL;

			if (lpFaxNumbers)
				MAPIFreeBuffer(lpFaxNumbers);
			lpFaxNumbers = NULL;
		}
	}

exit:
	if (lpNewEntryID)
		MAPIFreeBuffer(lpNewEntryID);

	if (lpFaxMailuser)
		lpFaxMailuser->Release();

	if (lpFaxNumbers)
		MAPIFreeBuffer(lpFaxNumbers);

	if (lpRecipColumns)
		MAPIFreeBuffer(lpRecipColumns);

	if (lpRowSet)
		FreeProws(lpRowSet);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT UniqueRecipients(struct thread_data *lpData, IMessage *lpMessage)
{
	HRESULT			hr = hrSuccess;
	IMAPITable		*lpTable = NULL;
	LPSRowSet		lpRowSet = NULL;
	LPSPropValue	lpEmailAddress = NULL;
	LPSPropValue	lpRecipType = NULL;
	string			strEmail;
	ULONG			ulRecipType = 0;

	SizedSPropTagArray(3, sptaColumns) = {
		3, {
			PR_ROWID,
			PR_SMTP_ADDRESS,
			PR_RECIPIENT_TYPE,
		}
	};

	SizedSSortOrderSet(2, sosOrder) = {
		2, 0, 0, {
			{ PR_SMTP_ADDRESS, TABLE_SORT_ASCEND },
			{ PR_RECIPIENT_TYPE, TABLE_SORT_ASCEND },
		}
	};

	hr = lpMessage->GetRecipientTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns((LPSPropTagArray)&sptaColumns, 0);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SortTable((LPSSortOrderSet)&sosOrder, 0);
	if (hr != hrSuccess)
		goto exit;

	while (TRUE) {
		if (lpRowSet) {
			FreeProws(lpRowSet);
			lpRowSet = NULL;
		}

		hr = lpTable->QueryRows(1, 0, &lpRowSet);
		if (hr != hrSuccess)
			goto exit;

		if (lpRowSet->cRows == 0)
			break;

		lpEmailAddress = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_SMTP_ADDRESS);
		lpRecipType = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_RECIPIENT_TYPE);

		if (!lpEmailAddress || !lpRecipType)
			continue;

		/* Filter To, Cc, Bcc individually */
		if (strEmail == lpEmailAddress->Value.lpszA && ulRecipType == lpRecipType->Value.ul) {
			hr = lpMessage->ModifyRecipients(MODRECIP_REMOVE, (LPADRLIST)lpRowSet);
			if (hr != hrSuccess)
				lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Failed to remove duplicate entry", lpData->thread);
		} else {
			strEmail = string(lpEmailAddress->Value.lpszA);
			ulRecipType = lpRecipType->Value.ul;
		}
	}

exit:
	if (lpRowSet)
		FreeProws(lpRowSet);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT SendUndeliverable(struct thread_data *lpData, ECSender *lpMailer, LPMDB lpStore, LPECUSER lpUserAdmin, LPMESSAGE lpMessage)
{
	HRESULT			hr;
	LPMAPIFOLDER	lpInbox = NULL;
	LPMESSAGE		lpErrorMsg = NULL;
	LPENTRYID		lpEntryID = NULL;
	ULONG			cbEntryID;
	ULONG			ulObjType;
	std::string		newbody;
	LPSPropValue	lpPropValue = NULL;
	unsigned int	ulPropPos = 0;
	FILETIME		ft;
	LPENTRYID		lpEntryIdSender	= NULL;
	ULONG			cbEntryIdSender;
	LPATTACH		lpAttach = NULL;
	LPMESSAGE		lpOrignalMessage = NULL;
	ULONG			cValuesOrignal = 0;
	LPSPropValue	lpPropArrayOrignal = NULL;
	LPSPropValue	lpPropValueAttach = NULL;
	unsigned int	ulPropModsPos;
	LPADRLIST		lpMods = NULL;
	LPSPropValue	lpPropValRecipientType = NULL; // Don't free this
	IMAPITable*		lpTableMods = NULL;
	SRestriction sRestriction;
	SPropValue sRestrictProp;
	LPSRowSet		lpRows = NULL;
	ULONG			ulRows = 0;
	ULONG			cEntries = 0;
	unsigned int	i, j;
	int				iRecip;
	std::string		strName, strType, strEmail;

	// CopyTo() var's
	unsigned int	ulPropAttachPos;
	ULONG			ulAttachNum;

	enum eORPos {
		OR_DISPLAY_TO, OR_DISPLAY_CC, OR_DISPLAY_BCC, OR_SEARCH_KEY, OR_SENDER_ADDRTYPE,
		OR_SENDER_EMAIL_ADDRESS, OR_SENDER_ENTRYID, OR_SENDER_NAME,
		OR_SENDER_SEARCH_KEY, OR_SENT_REPRESENTING_ADDRTYPE,
		OR_SENT_REPRESENTING_EMAIL_ADDRESS, OR_SENT_REPRESENTING_ENTRYID,
		OR_SENT_REPRESENTING_NAME, OR_SENT_REPRESENTING_SEARCH_KEY,
		OR_SUBJECT, OR_CLIENT_SUBMIT_TIME
	};

	SizedSPropTagArray(16, sPropsOrignal) = {
		16,
		{ PR_DISPLAY_TO, PR_DISPLAY_CC,
		  PR_DISPLAY_BCC, PR_SEARCH_KEY,
		  PR_SENDER_ADDRTYPE, PR_SENDER_EMAIL_ADDRESS,
		  PR_SENDER_ENTRYID, PR_SENDER_NAME,
		  PR_SENDER_SEARCH_KEY, PR_SENT_REPRESENTING_ADDRTYPE,
		  PR_SENT_REPRESENTING_EMAIL_ADDRESS, PR_SENT_REPRESENTING_ENTRYID,
		  PR_SENT_REPRESENTING_NAME, PR_SENT_REPRESENTING_SEARCH_KEY,
		  PR_SUBJECT, PR_CLIENT_SUBMIT_TIME }
	};

	SizedSPropTagArray(7, sPropTagRecipient) = {
		7,
		{ PR_RECIPIENT_TYPE, PR_DISPLAY_NAME, PR_DISPLAY_TYPE,
		  PR_ADDRTYPE, PR_EMAIL_ADDRESS,
		  PR_ENTRYID, PR_SEARCH_KEY }
	};

	// open inbox
	hr = lpStore->GetReceiveFolder("IPM", 0, &cbEntryID, &lpEntryID, NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to resolve incoming folder, error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}
	hr = lpStore->OpenEntry(cbEntryID, lpEntryID, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN *)&lpInbox);
	if (hr != hrSuccess || ulObjType != MAPI_FOLDER) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to open inbox folder, error code: 0x%08X", lpData->thread, hr);
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	// make new message in inbox
	hr = lpInbox->CreateMessage(NULL, 0, &lpErrorMsg);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to create undeliverable message, error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}

	// Get properties from the orinal message
	hr = lpMessage->GetProps((LPSPropTagArray)&sPropsOrignal, 0, &cValuesOrignal, &lpPropArrayOrignal);
	if (FAILED(hr))
		goto exit;

	// set props
	hr = MAPIAllocateBuffer(sizeof(SPropValue) * 50, (void **)&lpPropValue);
	if(hr != hrSuccess)
		goto exit;

	// Subject
	lpPropValue[ulPropPos].ulPropTag = PR_SUBJECT;
	lpPropValue[ulPropPos++].Value.lpszA = "Undelivered Mail Returned to Sender";

	// Message flags
	lpPropValue[ulPropPos].ulPropTag = PR_MESSAGE_FLAGS;
	lpPropValue[ulPropPos++].Value.ul = 0;

	// Message class
	lpPropValue[ulPropPos].ulPropTag = PR_MESSAGE_CLASS;
	lpPropValue[ulPropPos++].Value.lpszA = "REPORT.IPM.Note.NDR";

	// Get the time to add to the message as PR_CLIENT_SUBMIT_TIME
    GetSystemTimeAsFileTime(&ft);

	// Submit time
	lpPropValue[ulPropPos].ulPropTag = PR_CLIENT_SUBMIT_TIME;
	lpPropValue[ulPropPos++].Value.ft = ft;

	// Delivery time
	lpPropValue[ulPropPos].ulPropTag = PR_MESSAGE_DELIVERY_TIME;
	lpPropValue[ulPropPos++].Value.ft = ft;

	// FROM: set the properties PR_SENT_REPRESENTING_* and PR_SENDER_*
	if(lpUserAdmin->lpszFullName) {
		lpPropValue[ulPropPos].ulPropTag = PR_SENDER_NAME;
		lpPropValue[ulPropPos++].Value.lpszA = lpUserAdmin->lpszFullName;

		lpPropValue[ulPropPos].ulPropTag = PR_SENT_REPRESENTING_NAME;
		lpPropValue[ulPropPos++].Value.lpszA = lpUserAdmin->lpszFullName;
	}

	if(lpUserAdmin->lpszFullEmail) {
		// PR_SENDER_EMAIL_ADDRESS
		lpPropValue[ulPropPos].ulPropTag = PR_SENDER_EMAIL_ADDRESS;
		lpPropValue[ulPropPos++].Value.lpszA = lpUserAdmin->lpszFullEmail;

		// PR_SENT_REPRESENTING_EMAIL_ADDRESS
		lpPropValue[ulPropPos].ulPropTag = PR_SENT_REPRESENTING_EMAIL_ADDRESS;
		lpPropValue[ulPropPos++].Value.lpszA = lpUserAdmin->lpszFullEmail;

		// PR_SENDER_ADDRTYPE
		lpPropValue[ulPropPos].ulPropTag = PR_SENDER_ADDRTYPE;
		lpPropValue[ulPropPos++].Value.lpszA = "SMTP";

		// PR_SENT_REPRESENTING_ADDRTYPE
		lpPropValue[ulPropPos].ulPropTag = PR_SENT_REPRESENTING_ADDRTYPE;
		lpPropValue[ulPropPos++].Value.lpszA = "SMTP";

		// PR_SENDER_SEARCH_KEY
		lpPropValue[ulPropPos].ulPropTag = PR_SENDER_SEARCH_KEY;
		lpPropValue[ulPropPos].Value.bin.cb = 5 + strlen(lpUserAdmin->lpszFullEmail); // 5 = "SMTP:"

		MAPIAllocateMore(lpPropValue[ulPropPos].Value.bin.cb, lpPropValue, (void**)&lpPropValue[ulPropPos].Value.bin.lpb);
		memcpy(lpPropValue[ulPropPos].Value.bin.lpb, "SMTP:", 5);
		if(lpPropValue[ulPropPos].Value.bin.cb > 5)
			memcpy(lpPropValue[ulPropPos].Value.bin.lpb+5, lpUserAdmin->lpszFullEmail, lpPropValue[ulPropPos].Value.bin.cb - 5);
		ulPropPos++;

		// PR_SENT_REPRESENTING_SEARCH_KEY
		lpPropValue[ulPropPos].ulPropTag = PR_SENT_REPRESENTING_SEARCH_KEY;

		lpPropValue[ulPropPos].Value.bin.cb = 5 + strlen(lpUserAdmin->lpszFullEmail); // 5 = "SMTP:"

		MAPIAllocateMore(lpPropValue[ulPropPos].Value.bin.cb, lpPropValue, (void**)&lpPropValue[ulPropPos].Value.bin.lpb);
		memcpy(lpPropValue[ulPropPos].Value.bin.lpb, "SMTP:", 5);
		if(lpPropValue[ulPropPos].Value.bin.cb > 5)
			memcpy(lpPropValue[ulPropPos].Value.bin.lpb+5, lpUserAdmin->lpszFullEmail, lpPropValue[ulPropPos].Value.bin.cb - 5);
		ulPropPos++;

	}

	hr = ECCreateOneOff(lpUserAdmin->lpszFullName, (char*)"SMTP", lpUserAdmin->lpszFullEmail,
						MAPI_SEND_NO_RICH_INFO|MAPI_UNICODE, &cbEntryIdSender, &lpEntryIdSender);
	if(hr == hrSuccess)	{
		// PR_SENDER_ENTRYID
		lpPropValue[ulPropPos].ulPropTag		= PR_SENDER_ENTRYID;
		lpPropValue[ulPropPos].Value.bin.cb		= cbEntryIdSender;
		lpPropValue[ulPropPos++].Value.bin.lpb	= (LPBYTE)lpEntryIdSender;

		// PR_SENT_REPRESENTING_ENTRYID
		lpPropValue[ulPropPos].ulPropTag		= PR_SENT_REPRESENTING_ENTRYID;
		lpPropValue[ulPropPos].Value.bin.cb		= cbEntryIdSender;
		lpPropValue[ulPropPos++].Value.bin.lpb	= (LPBYTE)lpEntryIdSender;
	}else {
		hr = hrSuccess;
	}

	// Set the properties PR_RCVD_REPRESENTING_* and PR_RECEIVED_BY_* and
	// PR_ORIGINAL_SENDER_* and PR_ORIGINAL_SENT_*

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENDER_NAME].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_RECEIVED_BY_NAME;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENDER_NAME].Value.lpszA;

		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SENDER_NAME;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENDER_NAME].Value.lpszA;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENDER_EMAIL_ADDRESS].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_RECEIVED_BY_EMAIL_ADDRESS;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENDER_EMAIL_ADDRESS].Value.lpszA;

		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SENDER_EMAIL_ADDRESS;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENDER_EMAIL_ADDRESS].Value.lpszA;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENDER_ADDRTYPE].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_RECEIVED_BY_ADDRTYPE;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENDER_ADDRTYPE].Value.lpszA;

		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SENDER_ADDRTYPE;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENDER_ADDRTYPE].Value.lpszA;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENDER_SEARCH_KEY].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_RECEIVED_BY_SEARCH_KEY;
		lpPropValue[ulPropPos].Value.bin.cb = lpPropArrayOrignal[OR_SENDER_SEARCH_KEY].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb = lpPropArrayOrignal[OR_SENDER_SEARCH_KEY].Value.bin.lpb;

		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SENDER_SEARCH_KEY;
		lpPropValue[ulPropPos].Value.bin.cb = lpPropArrayOrignal[OR_SENDER_SEARCH_KEY].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb = lpPropArrayOrignal[OR_SENDER_SEARCH_KEY].Value.bin.lpb;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENDER_ENTRYID].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag		= PR_RECEIVED_BY_ENTRYID;
		lpPropValue[ulPropPos].Value.bin.cb		= lpPropArrayOrignal[OR_SENDER_ENTRYID].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb	= lpPropArrayOrignal[OR_SENDER_ENTRYID].Value.bin.lpb;

		lpPropValue[ulPropPos].ulPropTag		= PR_ORIGINAL_SENDER_ENTRYID;
		lpPropValue[ulPropPos].Value.bin.cb		= lpPropArrayOrignal[OR_SENDER_ENTRYID].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb	= lpPropArrayOrignal[OR_SENDER_ENTRYID].Value.bin.lpb;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENT_REPRESENTING_NAME].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_RCVD_REPRESENTING_NAME;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENT_REPRESENTING_NAME].Value.lpszA;

		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_NAME;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENT_REPRESENTING_NAME].Value.lpszA;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENT_REPRESENTING_EMAIL_ADDRESS].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_RCVD_REPRESENTING_EMAIL_ADDRESS;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENT_REPRESENTING_EMAIL_ADDRESS].Value.lpszA;

		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_EMAIL_ADDRESS;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENT_REPRESENTING_EMAIL_ADDRESS].Value.lpszA;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENT_REPRESENTING_ADDRTYPE].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_RCVD_REPRESENTING_ADDRTYPE;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENT_REPRESENTING_ADDRTYPE].Value.lpszA;

		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_ADDRTYPE;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_SENT_REPRESENTING_ADDRTYPE].Value.lpszA;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENT_REPRESENTING_SEARCH_KEY].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_RCVD_REPRESENTING_SEARCH_KEY;
		lpPropValue[ulPropPos].Value.bin.cb = lpPropArrayOrignal[OR_SENT_REPRESENTING_SEARCH_KEY].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb = lpPropArrayOrignal[OR_SENT_REPRESENTING_SEARCH_KEY].Value.bin.lpb;

		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_SEARCH_KEY;
		lpPropValue[ulPropPos].Value.bin.cb = lpPropArrayOrignal[OR_SENT_REPRESENTING_SEARCH_KEY].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb = lpPropArrayOrignal[OR_SENT_REPRESENTING_SEARCH_KEY].Value.bin.lpb;
	}

	if(PROP_TYPE(lpPropArrayOrignal[OR_SENT_REPRESENTING_ENTRYID].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag		= PR_RCVD_REPRESENTING_ENTRYID;
		lpPropValue[ulPropPos].Value.bin.cb		= lpPropArrayOrignal[OR_SENT_REPRESENTING_ENTRYID].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb	= lpPropArrayOrignal[OR_SENT_REPRESENTING_ENTRYID].Value.bin.lpb;

		lpPropValue[ulPropPos].ulPropTag		= PR_ORIGINAL_SENT_REPRESENTING_ENTRYID;
		lpPropValue[ulPropPos].Value.bin.cb		= lpPropArrayOrignal[OR_SENT_REPRESENTING_ENTRYID].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb	= lpPropArrayOrignal[OR_SENT_REPRESENTING_ENTRYID].Value.bin.lpb;
	}

	// Original display to
	if(PROP_TYPE(lpPropArrayOrignal[OR_DISPLAY_TO].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_DISPLAY_TO;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_DISPLAY_TO].Value.lpszA;
	}

	// Original display cc
	if(PROP_TYPE(lpPropArrayOrignal[OR_DISPLAY_CC].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_DISPLAY_CC;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_DISPLAY_CC].Value.lpszA;
	}

	// Original display bcc
	if(PROP_TYPE(lpPropArrayOrignal[OR_DISPLAY_BCC].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_DISPLAY_BCC;
		lpPropValue[ulPropPos++].Value.lpszA = lpPropArrayOrignal[OR_DISPLAY_BCC].Value.lpszA;
	}

	// Original subject
	if(PROP_TYPE(lpPropArrayOrignal[OR_SUBJECT].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SUBJECT;
		lpPropValue[ulPropPos++].Value.lpszA= lpPropArrayOrignal[OR_SUBJECT].Value.lpszA;
	}

	// Original submit time
	if(PROP_TYPE(lpPropArrayOrignal[OR_CLIENT_SUBMIT_TIME].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag = PR_ORIGINAL_SUBMIT_TIME;
		lpPropValue[ulPropPos++].Value.ft = lpPropArrayOrignal[OR_CLIENT_SUBMIT_TIME].Value.ft;
	}

	// Original searchkey
	if(PROP_TYPE(lpPropArrayOrignal[OR_SEARCH_KEY].ulPropTag) != PT_ERROR) {
		lpPropValue[ulPropPos].ulPropTag		= PR_ORIGINAL_SEARCH_KEY;
		lpPropValue[ulPropPos].Value.bin.cb		= lpPropArrayOrignal[OR_SEARCH_KEY].Value.bin.cb;
		lpPropValue[ulPropPos++].Value.bin.lpb	= lpPropArrayOrignal[OR_SEARCH_KEY].Value.bin.lpb;
	}


	// Add the orinal message into the errorMessage
	hr = lpErrorMsg->CreateAttach(NULL, 0, &ulAttachNum, &lpAttach);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to create attachment, error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}

	hr = lpAttach->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN *)&lpOrignalMessage);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMessage->CopyTo(0, NULL, NULL, 0, NULL, &IID_IMessage, (LPVOID)lpOrignalMessage, 0, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpOrignalMessage->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

	ulPropAttachPos = 0;
	MAPIAllocateBuffer(sizeof(SPropValue) * 4, (void**)&lpPropValueAttach);

	lpPropValueAttach[ulPropAttachPos].ulPropTag = PR_ATTACH_METHOD;
	lpPropValueAttach[ulPropAttachPos++].Value.ul = ATTACH_EMBEDDED_MSG;

	lpPropValueAttach[ulPropAttachPos].ulPropTag = PR_ATTACH_MIME_TAG;
	lpPropValueAttach[ulPropAttachPos++].Value.lpszA = "message/rfc822";

	if(PROP_TYPE(lpPropArrayOrignal[OR_SUBJECT].ulPropTag) != PT_ERROR) {
		lpPropValueAttach[ulPropAttachPos].ulPropTag = PR_DISPLAY_NAME;
		lpPropValueAttach[ulPropAttachPos++].Value.lpszA = lpPropArrayOrignal[OR_SUBJECT].Value.lpszA;
	}

	lpPropValueAttach[ulPropAttachPos].ulPropTag = PR_RENDERING_POSITION;
	lpPropValueAttach[ulPropAttachPos++].Value.ul = -1;

	hr = lpAttach->SetProps(ulPropAttachPos, lpPropValueAttach, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpAttach->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;


	// add failed recipients to error report

	hr = lpMessage->GetRecipientTable(0, &lpTableMods);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTableMods->SetColumns((LPSPropTagArray)&sPropTagRecipient, TBL_BATCH);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTableMods->GetRowCount(0, &ulRows);
	if (hr != hrSuccess)
		goto exit;

	if (ulRows == 0 || lpMailer->getRecipientErrorCount() == 0) {
		// If there's a pr_body, outlook will display that, and not the 'default' outlook error report

		// Message error
		newbody = "Unfortunately, I was unable to deliver your mail.\nThe error given was:\n\n";
		newbody.append(lpMailer->getErrorString());
		newbody.append("\n\nYou may need to contact your e-mail administrator to solve this problem.\n");

		lpPropValue[ulPropPos].ulPropTag = PR_BODY;
		lpPropValue[ulPropPos++].Value.lpszA = (char*)newbody.c_str();

		if (ulRows > 0) {
			// just copy recipient table if we did have recipients
			// for working resend button if the From was denied
			hr = lpTableMods->QueryRows(-1, 0, &lpRows);
			if (hr != hrSuccess)
				goto exit;

			hr = lpErrorMsg->ModifyRecipients(MODRECIP_ADD, (LPADRLIST)lpRows);
			if (hr != hrSuccess)
				goto exit;
		}
	}
	else if (ulRows > 0)
	{
		MAPIAllocateBuffer(CbNewADRLIST(ulRows), (void**)&lpMods);
		lpMods->cEntries = 0;
		// if this is a retry of a failed message with the same recipients, the items are double (with MAPI_TO and MAPI_P1 versions)
		// so we check for a table with P1 versions, use that, otherwise the complete table

		sRestriction.rt = RES_PROPERTY;
		sRestriction.res.resProperty.relop = RELOP_EQ;
		sRestriction.res.resProperty.ulPropTag = PR_RECIPIENT_TYPE;
		sRestriction.res.resProperty.lpProp = &sRestrictProp;

		sRestrictProp.ulPropTag = PR_RECIPIENT_TYPE;
		sRestrictProp.Value.ul = MAPI_P1;

		hr = lpTableMods->Restrict(&sRestriction, 0);
		if (hr != hrSuccess)
			goto exit;

		hr = lpTableMods->QueryRows(ulRows, 0, &lpRows);
		if (hr != hrSuccess)
			goto exit;

		if (lpRows->cRows == 0)
		{
			// no MAPI_P1 recipients, remove restriction and get again
			FreeProws(lpRows);
			lpRows = NULL;

			hr = lpTableMods->Restrict(NULL, 0);
			if (hr != hrSuccess)
				goto exit;

			hr = lpTableMods->QueryRows(ulRows, 0, &lpRows);
			if (hr != hrSuccess)
				goto exit;
		}

		cEntries = 0;
		for (i = 0; i < lpRows->cRows; i++)
		{
			if (HrGetAddress(lpData->lpAddrBook, lpRows->aRow[i].lpProps, lpRows->aRow[i].cValues,
							 PR_ENTRYID, PR_DISPLAY_NAME, PR_ADDRTYPE, PR_EMAIL_ADDRESS,
							 strName, strType, strEmail) != hrSuccess)
				continue;

			// find erroneous recipient
			iRecip = -1;
			for (j = 0; j < lpMailer->getRecipientErrorCount(); j++) {
				if (strEmail == lpMailer->getRecipientErrorEmailAddress(j)) {
					iRecip = j;
					break;
				}
			}
			if (iRecip == -1)
				continue;

			MAPIAllocateBuffer(sizeof(SPropValue) * (lpRows->aRow[i].cValues + 6), (void**)&lpMods->aEntries[cEntries].rgPropVals);
			ulPropModsPos = 0;
			lpMods->cEntries = cEntries;

			// Fast copy recipient properties
			for(j=0; j < lpRows->aRow[i].cValues; j++) // columns sPropTagRecipient
				lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos++] = lpRows->aRow[i].lpProps[j];

			// reset PR_RECIPIENT_TYPE to MAPI_TO for retries (unknown how this may become MAPI_CC or MAPI_BCC)
			lpPropValRecipientType = PpropFindProp(lpMods->aEntries[cEntries].rgPropVals, j, PR_RECIPIENT_TYPE);
			if (lpPropValRecipientType && lpPropValRecipientType->Value.ul == MAPI_P1)
				lpPropValRecipientType->Value.ul = MAPI_TO;

			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos].ulPropTag = PR_REPORT_TEXT;
			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos++].Value.lpszA = (char*)lpMailer->getRecipientErrorText(iRecip).c_str();

			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos].ulPropTag = PR_REPORT_TIME;
			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos++].Value.ft = ft;

			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos].ulPropTag = PR_TRANSMITABLE_DISPLAY_NAME;
			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos++].Value.lpszA = (char*)lpMailer->getRecipientErrorEmailAddress(iRecip).c_str();

			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos].ulPropTag = 0x0C200003;//PR_NDR_STATUS_CODE;
			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos++].Value.ul = lpMailer->getRecipientErrorSMTPCode(iRecip);

			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos].ulPropTag = PR_NDR_DIAG_CODE;
			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos++].Value.ul = MAPI_DIAG_MAIL_RECIPIENT_UNKNOWN;

			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos].ulPropTag = PR_NDR_REASON_CODE;
			lpMods->aEntries[cEntries].rgPropVals[ulPropModsPos++].Value.ul = MAPI_REASON_TRANSFER_FAILED;

			lpMods->aEntries[cEntries].cValues = ulPropModsPos;
			cEntries++;
		}

		lpMods->cEntries = cEntries;

		hr = lpErrorMsg->ModifyRecipients(MODRECIP_ADD, lpMods);
		if (hr != hrSuccess)
			goto exit;
	}

	// Add properties
	hr = lpErrorMsg->SetProps(ulPropPos, lpPropValue, NULL);
	if (hr != hrSuccess)
		goto exit;

	// save message
	hr = lpErrorMsg->SaveChanges(KEEP_OPEN_READONLY);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to commit message: 0x%08X", lpData->thread, hr);
		goto exit;
	}

	// New mail notification
	hr = HrNewMailNotification(lpStore, lpErrorMsg, lpLogger);

exit:
	if(lpOrignalMessage)
		lpOrignalMessage->Release();

	if(lpAttach)
		lpAttach->Release();

	if (lpInbox)
		lpInbox->Release();

	if (lpErrorMsg)
		lpErrorMsg->Release();

	if(lpTableMods)
		lpTableMods->Release();

	if (lpPropValue)
		MAPIFreeBuffer(lpPropValue);

	if(lpPropValueAttach)
		MAPIFreeBuffer(lpPropValueAttach);

	if(lpEntryIdSender)
		MAPIFreeBuffer(lpEntryIdSender);

	if(lpPropArrayOrignal)
		MAPIFreeBuffer(lpPropArrayOrignal);

	if(lpRows)
		FreeProws(lpRows);

	if(lpMods)
		FreePadrlist(lpMods);

	if(lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	return hr;
}

HRESULT ContactToZarafa(IMsgStore *lpUserStore, LPADRBOOK lpAddrBook, ULONG cbEntryId, LPENTRYID lpEntryId, ULONG *lpulZarafaEID, LPENTRYID *lppZarafaEID)
{
	HRESULT hr = hrSuccess;
	LPCONTAB_ENTRYID lpContabEntryID = (LPCONTAB_ENTRYID)lpEntryId;
	GUID* guid = (GUID*)&lpContabEntryID->muid;
	ULONG ulObjType;
	LPMAILUSER lpContact = NULL;
	ULONG cValues;
	LPSPropValue lpEntryIds = NULL;
	LPSPropTagArray lpPropTags = NULL;
	LPMAPINAMEID lpNames = NULL;
	LPMAPINAMEID *lppNames = NULL;

	if (sizeof(CONTAB_ENTRYID) > cbEntryId ||
		*guid != PSETID_CONTACT_FOLDER_RECIPIENT ||
		lpContabEntryID->email_offset > 2)
	{
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Invalid contact entryid in delegate/sendas");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = lpUserStore->OpenEntry(lpContabEntryID->cbeid, (LPENTRYID)lpContabEntryID->abeid, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpContact);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to open contact entryid: 0x%X",hr);
		goto exit;
	}

	hr = MAPIAllocateBuffer(sizeof(MAPINAMEID) * 3, (void**)&lpNames);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "No memory for named ids from contact");
		goto exit;
	}

	hr = MAPIAllocateBuffer(sizeof(LPMAPINAMEID) * 3, (void**)&lppNames);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "No memory for named ids from contact");
		goto exit;
	}

	// Email1EntryID
	lpNames[0].lpguid = (GUID*)&PSETID_Address;
	lpNames[0].ulKind = MNID_ID;
	lpNames[0].Kind.lID = 0x8085;
	lppNames[0] = &lpNames[0];

	// Email2EntryID
	lpNames[1].lpguid = (GUID*)&PSETID_Address;
	lpNames[1].ulKind = MNID_ID;
	lpNames[1].Kind.lID = 0x8095;
	lppNames[1] = &lpNames[1];

	// Email3EntryID
	lpNames[2].lpguid = (GUID*)&PSETID_Address;
	lpNames[2].ulKind = MNID_ID;
	lpNames[2].Kind.lID = 0x80A5;
	lppNames[2] = &lpNames[2];

	hr = lpContact->GetIDsFromNames(3, lppNames, 0, &lpPropTags);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while retrieving named data from contact: 0x%08X", hr);
		goto exit;
	}

	hr = lpContact->GetProps(lpPropTags, 0, &cValues, &lpEntryIds);
	if (FAILED(hr)) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get named properties: 0x%08X", hr);
		goto exit;
	}

	if (PROP_TYPE(lpEntryIds[lpContabEntryID->email_offset].ulPropTag) != PT_BINARY) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Offset %d not found in contact", lpContabEntryID->email_offset);
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = MAPIAllocateBuffer(lpEntryIds[lpContabEntryID->email_offset].Value.bin.cb, (void**)lppZarafaEID);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "No memory for contact eid");
		goto exit;
	}

	memcpy(*lppZarafaEID, lpEntryIds[lpContabEntryID->email_offset].Value.bin.lpb, lpEntryIds[lpContabEntryID->email_offset].Value.bin.cb);
	*lpulZarafaEID = lpEntryIds[lpContabEntryID->email_offset].Value.bin.cb;

exit:
	if (lpPropTags)
		MAPIFreeBuffer(lpPropTags);

	if (lppNames)
		MAPIFreeBuffer(lppNames);

	if (lpNames)
		MAPIFreeBuffer(lpNames);

	if (lpEntryIds)
		MAPIFreeBuffer(lpEntryIds);

	if (lpContact)
		lpContact->Release();

	return hr;
}

HRESULT SMTPToZarafa(LPADRBOOK lpAddrBook, ULONG ulSMTPEID, LPENTRYID lpSMTPEID, ULONG *lpulZarafaEID, LPENTRYID *lppZarafaEID)
{
	HRESULT hr = hrSuccess;
	string strName, strType, strEmailAddress;
	LPADRLIST lpAList = NULL;
	LPSPropValue lpSpoofEID = NULL;
	LPENTRYID lpSpoofBin = NULL;

	// representing entryid can also be a one off id, so search the user, and then get the entryid again ..
	// we then always should have yourself as the sender, otherwise: denied
	if (ECParseOneOff(lpSMTPEID, ulSMTPEID, strName, strType, strEmailAddress) == hrSuccess) {
		MAPIAllocateBuffer(CbNewADRLIST(1), (void**)&lpAList);
		lpAList->cEntries = 1;

		lpAList->aEntries[0].cValues = 1;
		MAPIAllocateBuffer(sizeof(SPropValue) * lpAList->aEntries[0].cValues, (void**)&lpAList->aEntries[0].rgPropVals);

		lpAList->aEntries[0].rgPropVals[0].ulPropTag = PR_DISPLAY_NAME;
		lpAList->aEntries[0].rgPropVals[0].Value.lpszA = (char*)strEmailAddress.c_str();
	
		hr = lpAddrBook->ResolveName(0, EMS_AB_ADDRESS_LOOKUP, NULL, lpAList);
		if (hr != hrSuccess)
			goto exit;

		lpSpoofEID = PpropFindProp(lpAList->aEntries[0].rgPropVals, lpAList->aEntries[0].cValues, PR_ENTRYID);
		if (!lpSpoofEID) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}

		hr = MAPIAllocateBuffer(lpSpoofEID->Value.bin.cb, (void**)&lpSpoofBin);
		if (hr != hrSuccess)
			goto exit;

		memcpy(lpSpoofBin, lpSpoofEID->Value.bin.lpb, lpSpoofEID->Value.bin.cb);
		*lppZarafaEID = lpSpoofBin;
		*lpulZarafaEID = lpSpoofEID->Value.bin.cb;
	} else {
		hr = MAPI_E_NOT_FOUND;
	}

exit:
	if (lpAList)
		FreePadrlist(lpAList);

	return hr;
}

HRESULT HrOpenRepresentStore(struct thread_data *lpData, ULONG ulRepresentCB, LPENTRYID lpRepresentEID, LPMDB *lppRepStore) {
	HRESULT hr = hrSuccess;
	ULONG ulObjType = 0;
	LPMAILUSER lpRepresenting = NULL;
	LPSPropValue lpRepAccount = NULL;
	LPEXCHANGEMANAGESTORE lpExchangeManageStore = NULL;
	ULONG ulRepStoreCB = 0;
	LPENTRYID lpRepStoreEID = NULL;
	LPMDB lpRepStore = NULL;

	if (lpData->lpAddrBook->OpenEntry(ulRepresentCB, lpRepresentEID, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpRepresenting) != hrSuccess) {
	    hr = MAPI_E_NOT_FOUND;
		goto exit;
    }

	if (HrGetOneProp(lpRepresenting, PR_ACCOUNT, &lpRepAccount) != hrSuccess) {
	    hr = MAPI_E_NOT_FOUND;
		goto exit;
    }

	hr = lpData->lpMDB->QueryInterface(IID_IExchangeManageStore, (void **)&lpExchangeManageStore);
	if (hr != hrSuccess)
		goto exit;

	hr = lpExchangeManageStore->CreateStoreEntryID(NULL, lpRepAccount->Value.lpszA, 0, &ulRepStoreCB, &lpRepStoreEID);
	if (hr != hrSuccess)
		goto exit;

	// Use the admin session to open the store, so we have full rights
	hr = lpData->lpAdminSession->OpenMsgStore(0, ulRepStoreCB, lpRepStoreEID, NULL, MAPI_BEST_ACCESS, &lpRepStore);
	if (hr != hrSuccess)
		goto exit;

	hr = lpRepStore->QueryInterface(IID_IMsgStore, (void**)lppRepStore);

exit:
	if (lpRepresenting)
		lpRepresenting->Release();

	if (lpRepAccount)
		MAPIFreeBuffer(lpRepAccount);

	if (lpExchangeManageStore)
		lpExchangeManageStore->Release();

	if (lpRepStoreEID)
		MAPIFreeBuffer(lpRepStoreEID);

	if (lpRepStore)
		lpRepStore->Release();

	return hr;
}

HRESULT CheckSendAs(struct thread_data *lpData, LPMDB lpUserStore, ECSender *lpMailer, ULONG ulOwnerCB, LPENTRYID lpOwnerEID, ULONG ulRepresentCB, LPENTRYID lpRepresentEID, bool *lpbAllowed, LPMDB *lppRepStore) {
	HRESULT hr = hrSuccess;
	bool bAllowed = false;
	ULONG ulObjType;
	LPMAILUSER lpMailboxOwner = NULL;
	LPMAILUSER lpRepresenting = NULL;
	LPSPropValue lpOwnerProps = NULL;
	LPSPropValue lpSpoofProps = NULL;
	SPropValue sSpoofEID = {0};
	ULONG ulCmpRes = 0;
	SizedSPropTagArray(2, sptaIDProps) = { 2, { PR_DISPLAY_NAME, PR_EC_SENDAS_USER_ENTRYIDS } };
	ULONG cValues = 0;


	hr = SMTPToZarafa(lpData->lpAddrBook, ulRepresentCB, lpRepresentEID, &sSpoofEID.Value.bin.cb, (LPENTRYID*)&sSpoofEID.Value.bin.lpb);
	if (hr != hrSuccess)
		hr = ContactToZarafa(lpUserStore, lpData->lpAddrBook, ulRepresentCB, lpRepresentEID, &sSpoofEID.Value.bin.cb, (LPENTRYID*)&sSpoofEID.Value.bin.lpb);
	if (hr == hrSuccess) {
		ulRepresentCB = sSpoofEID.Value.bin.cb;
		lpRepresentEID = (LPENTRYID)sSpoofEID.Value.bin.lpb;
	}

	// you can always send as yourself
	if (lpData->lpAddrBook->CompareEntryIDs(ulOwnerCB, lpOwnerEID, ulRepresentCB, lpRepresentEID, 0, &ulCmpRes)	== hrSuccess && ulCmpRes == TRUE)
	{
		bAllowed = true;
		goto exit;
	}


	hr = lpData->lpAddrBook->OpenEntry(ulOwnerCB, lpOwnerEID, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpMailboxOwner);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMailboxOwner->GetProps((LPSPropTagArray)&sptaIDProps, 0, &cValues, &lpOwnerProps);
	if (FAILED(hr))
		goto exit;

	// representing entryid is now always a Zarafa Entry ID. Open the user so we can log the display name
	if (lpData->lpAddrBook->OpenEntry(ulRepresentCB, lpRepresentEID, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpRepresenting) != hrSuccess)
		goto exit;

	sptaIDProps.aulPropTag[1] = PR_ENTRYID; // replace PR_EC_SENDAS_USER_ENTRYIDS for user PR_ENTRYID
	hr = lpRepresenting->GetProps((LPSPropTagArray)&sptaIDProps, 0, &cValues, &lpSpoofProps);
	if (FAILED(hr))
		goto exit;

    if(lpOwnerProps[1].ulPropTag != PR_EC_SENDAS_USER_ENTRYIDS) {
        // No sendas, therefore no sendas permissions, but we don't fail
        goto exit;
    }
    
    if(lpSpoofProps[1].ulPropTag != PR_ENTRYID) {
        // Should not happen
        hr = MAPI_E_NOT_FOUND;
        goto exit;
    }

	for (ULONG i = 0; i < lpOwnerProps[1].Value.MVbin.cValues; i++) {
		if (lpData->lpAddrBook->CompareEntryIDs(lpSpoofProps[1].Value.bin.cb, (LPENTRYID)lpSpoofProps[1].Value.bin.lpb,
												lpOwnerProps[1].Value.MVbin.lpbin[i].cb, (LPENTRYID)lpOwnerProps[1].Value.MVbin.lpbin[i].lpb,
												0, &ulCmpRes)
			!= hrSuccess)
			continue;

		if (ulCmpRes == TRUE) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Mail for user '%s' is send as user '%s'", lpData->thread, lpOwnerProps[0].Value.lpszA, lpSpoofProps[0].Value.lpszA);
			bAllowed = true;
			goto exit;
		}
	}

exit:
	if (!bAllowed) {
		if (lpSpoofProps && PROP_TYPE(lpSpoofProps[0].ulPropTag) != PT_ERROR)
			lpMailer->setError("You are not allowed to send as user "+string(lpSpoofProps[0].Value.lpszA));
		else
			lpMailer->setError("The user you try to send as could not be found.");

		lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "User '%s' is not allowed to send as user '%s'. "
					  "You may enable all outgoing addresses by enabling the always_send_delegates option.", lpData->thread,
					  (lpOwnerProps && PROP_TYPE(lpOwnerProps[0].ulPropTag) != PT_ERROR) ? lpOwnerProps[0].Value.lpszA : "<unknown>",
					  (lpSpoofProps && PROP_TYPE(lpSpoofProps[0].ulPropTag) != PT_ERROR) ? lpSpoofProps[0].Value.lpszA : "<unknown>");
	}

	if (bAllowed) {
		hr = HrOpenRepresentStore(lpData, ulRepresentCB, lpRepresentEID, lppRepStore);
		if (hr != hrSuccess)
			goto exit;
	}

	*lpbAllowed = bAllowed;

	if (sSpoofEID.Value.bin.lpb)
		MAPIFreeBuffer(sSpoofEID.Value.bin.lpb);

	if (lpOwnerProps)
		MAPIFreeBuffer(lpOwnerProps);

	if (lpSpoofProps)
		MAPIFreeBuffer(lpSpoofProps);

	if (lpMailboxOwner)
		lpMailboxOwner->Release();

	if (lpRepresenting)
		lpRepresenting->Release();

	return hr;
}

HRESULT CheckDelegate(struct thread_data *lpData, LPMDB lpUserStore, ULONG ulOwnerCB, LPENTRYID lpOwnerEID, ULONG ulRepresentCB, LPENTRYID lpRepresentEID, bool *lpbAllowed, LPMDB *lppRepStore)
{
	HRESULT hr = hrSuccess;
	bool bAllowed = false;
	ULONG ulObjType;
	LPMDB lpRepStore = NULL;
	LPMAPIFOLDER lpRepSubtree = NULL;
	LPSPropValue lpRepFBProp = NULL;
	LPMESSAGE lpRepFBMessage = NULL;
	LPSPropValue lpDelegates = NULL;
	ULONG ulCmpRes = 0;
	SPropValue sSpoofEID = {0};


	hr = SMTPToZarafa(lpData->lpAddrBook, ulRepresentCB, lpRepresentEID, &sSpoofEID.Value.bin.cb, (LPENTRYID*)&sSpoofEID.Value.bin.lpb);
	if (hr != hrSuccess)
		hr = ContactToZarafa(lpUserStore, lpData->lpAddrBook, ulRepresentCB, lpRepresentEID, &sSpoofEID.Value.bin.cb, (LPENTRYID*)&sSpoofEID.Value.bin.lpb);
	if (hr == hrSuccess) {
		ulRepresentCB = sSpoofEID.Value.bin.cb;
		lpRepresentEID = (LPENTRYID)sSpoofEID.Value.bin.lpb;
	}

	hr = HrOpenRepresentStore(lpData, ulRepresentCB, lpRepresentEID, &lpRepStore);
	if (hr != hrSuccess)
		goto exit;

	// open root container
	hr = lpRepStore->OpenEntry(0, NULL, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpRepSubtree);
	if (hr != hrSuccess)
		goto exit;


	hr = HrGetOneProp(lpRepSubtree, PR_FREEBUSY_ENTRYIDS, &lpRepFBProp);
	if (hr != hrSuccess)
		goto exit;

	if (lpRepFBProp->Value.MVbin.cValues < 2) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = lpRepSubtree->OpenEntry(lpRepFBProp->Value.MVbin.lpbin[1].cb, (LPENTRYID)lpRepFBProp->Value.MVbin.lpbin[1].lpb, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpRepFBMessage);
	if (hr != hrSuccess)
		goto exit;


	hr = HrGetOneProp(lpRepFBMessage, PR_SCHDINFO_DELEGATE_ENTRYIDS, &lpDelegates);
	if (hr != hrSuccess)
		goto exit;

	for (ULONG i = 0; i < lpDelegates->Value.MVbin.cValues; i++) {
		if (lpData->lpAddrBook->CompareEntryIDs(ulOwnerCB, lpOwnerEID,
												lpDelegates->Value.MVbin.lpbin[i].cb, (LPENTRYID)lpDelegates->Value.MVbin.lpbin[i].lpb,
												0, &ulCmpRes)
			!= hrSuccess)
			continue;

		if (ulCmpRes == TRUE) {
			bAllowed = true;
			goto exit;
		}
	}


exit:
	*lpbAllowed = bAllowed;
	// when any step failed, delegate is not setup correctly, so bAllowed == false
	hr = hrSuccess;

	if (bAllowed) {
		*lppRepStore = lpRepStore;
		lpRepStore = NULL;
	}

	if (sSpoofEID.Value.bin.lpb)
		MAPIFreeBuffer(sSpoofEID.Value.bin.lpb);

	if (lpRepStore)
		lpRepStore->Release();

	if (lpRepSubtree)
		lpRepSubtree->Release();

	if (lpRepFBProp)
		MAPIFreeBuffer(lpRepFBProp);

	if (lpRepFBMessage)
		lpRepFBMessage->Release();

	if (lpDelegates)
		MAPIFreeBuffer(lpDelegates);

	return hr;
}

HRESULT CopyDelegateMessageToSentItems(LPMESSAGE lpMessage, LPMDB lpRepStore, LPMESSAGE *lppRepMessage) {
	HRESULT hr = hrSuccess;
	LPSPropValue lpSentItemsEntryID = NULL;
	LPMAPIFOLDER lpSentItems = NULL;
	ULONG ulObjType;
	LPMESSAGE lpDestMsg = NULL;
	SPropValue sProp[1];

	hr = HrGetOneProp(lpRepStore, PR_IPM_SENTMAIL_ENTRYID, &lpSentItemsEntryID);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to find representee's sent items folder: error 0x%08X", pthread_self(), hr);
		goto exit;
	}

	hr = lpRepStore->OpenEntry(lpSentItemsEntryID->Value.bin.cb, (LPENTRYID)lpSentItemsEntryID->Value.bin.lpb,
							   &IID_IMAPIFolder, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN*)&lpSentItems);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to open representee's sent items folder: error 0x%08X", pthread_self(), hr);
		goto exit;
	}

	hr = lpSentItems->CreateMessage(NULL, 0, &lpDestMsg);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to create representee's message: error 0x%08X", pthread_self(), hr);
		goto exit;
	}

	hr = lpMessage->CopyTo(0, NULL, NULL, 0, NULL, &IID_IMessage, (LPVOID)lpDestMsg, 0, NULL);
	if (FAILED(hr)) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to copy representee's message: error 0x%08X", pthread_self(), hr);
		goto exit;
	}

	sProp[0].ulPropTag = PR_MESSAGE_FLAGS;
	sProp[0].Value.ul = MSGFLAG_READ;

	hr = lpDestMsg->SetProps(1, sProp, NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to edit representee's message: error 0x%08X", pthread_self(), hr);
		goto exit;
	}

	*lppRepMessage = lpDestMsg;
	lpDestMsg = NULL;

exit:
	if (lpSentItemsEntryID)
		MAPIFreeBuffer(lpSentItemsEntryID);

	if (lpSentItems)
		lpSentItems->Release();

	if (lpDestMsg)
		lpDestMsg->Release();

	return hr;
}

HRESULT ProcessMessage(struct thread_data *lpData, ECSender *lpMailer, LPECUSER lpUserAdmin, LPMESSAGE *lppMessage) {
	HRESULT 		hr 				= hrSuccess;
	LPMDB			lpStore			= NULL;
	LPMESSAGE		lpMessage		= NULL;
	ULONG			ulObjType		= 0;
	LPSPropValue	lpsPropObject	= NULL;
	IECSecurity		*lpSecurity		= NULL;
	IECUnknown		*lpECStore		= NULL;
	ULONG			cbOwner			= 0;
	LPENTRYID		lpOwner			= NULL;
	ECUSER			*lpUser			= NULL;
	SPropValue		sPropSender[4];

	SizedSPropTagArray(5, sptaMoveReprProps) = {
		5, { PR_SENT_REPRESENTING_NAME, PR_SENT_REPRESENTING_ADDRTYPE,
			 PR_SENT_REPRESENTING_EMAIL_ADDRESS, PR_SENT_REPRESENTING_ENTRYID, PR_SENT_REPRESENTING_SEARCH_KEY }
	};
	LPSPropValue	lpMoveReprProps = NULL;
	ULONG			cValuesMoveProps = 0;

	LPSPropValue	lpPropOwner = NULL;
	bool			bAllowSendAs = false;
	bool			bAllowDelegate = false;
	ULONG			ulCmpRes = 0;
	
	ULONG			cbEntryIdSender	= 0;
	LPENTRYID		lpEntryIdSender	= NULL;

	ULONG			cbEntryIdRepr	= 0;
	LPENTRYID		lpEntryIdRepr	= NULL;
	LPMDB			lpRepStore		= NULL;
	LPMESSAGE		lpRepMessage	= NULL;

	LPSPropValue	lpPropSRNam		= NULL;
	LPSPropValue	lpPropSREma		= NULL;
	LPSPropValue	lpPropSREid		= NULL;
	LPSPropValue	lpRepEntryID	= NULL;
	LPSPropValue	lpSubject		= NULL;
	LPSPropValue	lpMsgSize		= NULL;
	LPSPropValue	lpDeferSendTime	= NULL;
	LPSPropValue	lpAutoForward	= NULL;

	sending_options sopt;

	imopt_default_sending_options(&sopt);
	// optional force sending with TNEF
	sopt.force_tnef = parseBool(lpConfig->GetSetting("always_send_tnef"));

	// Get the owner of the store
	hr = lpData->lpMAPISession->OpenMsgStore(0, lpData->sStore.cb, (LPENTRYID)lpData->sStore.lpb, NULL, MAPI_BEST_ACCESS, &lpStore);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to open user store, error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}	

	hr = HrGetOneProp(lpStore, PR_EC_OBJECT, &lpsPropObject);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to get Zarafa internal object, error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}

	// NOTE: object is placed in Value.lpszA, not Value.x
	lpECStore = (IECUnknown *)lpsPropObject->Value.lpszA;
	hr = lpECStore->QueryInterface(IID_IECSecurity, (void **)&lpSecurity);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "IID_IECSecurity not supported by store, error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}

	hr = lpSecurity->GetOwner(&cbOwner, &lpOwner);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to get owner information, error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}

	// We now have the owner ID, get the owner information through the ServiceAdmin
	hr = lpData->lpServiceAdmin->GetUser(cbOwner, lpOwner, &lpUser);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to get user information from store, error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}

	// open the message we need to send
	hr = lpStore->OpenEntry(lpData->sMessage.cb, (LPENTRYID)lpData->sMessage.lpb,
							&IID_IMessage, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN*)&lpMessage);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Could not open message in store from user %s", lpData->thread, lpUser->lpszUsername);
		goto exit;
	}

	// get subject for logging
	HrGetOneProp(lpMessage, PR_SUBJECT, &lpSubject);
	HrGetOneProp(lpMessage, PR_MESSAGE_SIZE, &lpMsgSize);
	HrGetOneProp(lpMessage, PR_DEFERRED_SEND_TIME, &lpDeferSendTime);

	// do we need to send the message already?
	if (lpDeferSendTime) {
		// check time
		time_t now = time(NULL);
		time_t sendat;

		FileTimeToUnixTime(lpDeferSendTime->Value.ft, &sendat);

		if (now < sendat) {
			// should actually be logged just once .. but how?
			struct tm tmp;
			char timestring[256];

			localtime_r(&sendat, &tmp);
			strftime(timestring, 256, "%c", &tmp);

			lpLogger->Log(EC_LOGLEVEL_INFO, THREAD_DBG "E-mail for user %s, subject '%s', should be sent later at '%s'",
						  lpData->thread, lpUser->lpszUsername, lpSubject ? lpSubject->Value.lpszA : "<none>", timestring);

			hr = MAPI_E_WAIT;
			goto exit;
		}
	}

	// fatal, all other log messages are otherwise somewhat meaningless
	lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Sending e-mail for user %s, subject: '%s', size: %d",
				  lpData->thread, lpUser->lpszUsername, lpSubject ? lpSubject->Value.lpszA : "<none>",
				  lpMsgSize ? lpMsgSize->Value.ul : 0);

	/* 

	   PR_SENDER_* maps to Sender:
	   PR_SENT_REPRESENTING_* maps to From:

	   Sender: field is optional, From: is mandatory
	   PR_SENDER_* is mandatory, and always set by us (will be overwritten if was set)
	   PR_SENT_REPRESENTING_* is optional, and set by outlook when the user modifies the From in outlook.

	*/

    // Set PR_SENT_REPRESENTING, as this is set on all 'sent' items and is the column
    // that is shown by default in Outlook's 'sent items' folder
	if (HrGetOneProp(lpMessage, PR_SENT_REPRESENTING_ENTRYID, &lpRepEntryID) != hrSuccess) {
		// set new sender
		sPropSender[0].ulPropTag = PR_SENT_REPRESENTING_NAME;
		sPropSender[0].Value.lpszA = lpUser->lpszFullName;
		sPropSender[1].ulPropTag = PR_SENT_REPRESENTING_ADDRTYPE;
		sPropSender[1].Value.lpszA = "SMTP";
		sPropSender[2].ulPropTag = PR_SENT_REPRESENTING_EMAIL_ADDRESS;
		sPropSender[2].Value.lpszA = lpUser->lpszFullEmail;

		sPropSender[3].ulPropTag = PR_SENT_REPRESENTING_ENTRYID;
		if (ECCreateOneOff(lpUser->lpszFullName, (char*)"SMTP", lpUser->lpszFullEmail,
						   MAPI_SEND_NO_RICH_INFO|MAPI_UNICODE, &cbEntryIdRepr, &lpEntryIdRepr) != hrSuccess)
		{
			lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to create sender id for message", lpData->thread);
			goto exit;
		}
		sPropSender[3].Value.bin.cb = cbEntryIdRepr;
		sPropSender[3].Value.bin.lpb = (LPBYTE)lpEntryIdRepr;

		if (lpMessage->SetProps(4, sPropSender, NULL) != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to set sender id for message", lpData->thread);
			goto exit;
		}
	} else {
		// requested that mail is sent as somebody else
		// since we can have SMTP and ZARAFA entry id's, we'll open it, and get the

		// If this is a forwarded e-mail, then allow sending as the original sending e-mail address. Note that
		// this can be misused by MAPI client that just set PR_AUTO_FORWARDED. Since it would have been just as
		// easy for the client just to spoof their 'from' address via SMTP, we're allowing this for now. You can
		// completely turn it off via the 'allow_redirect_spoofing' setting.
		if (strcmp(lpConfig->GetSetting("allow_redirect_spoofing"), "yes") == 0 && HrGetOneProp(lpMessage, PR_AUTO_FORWARDED, &lpAutoForward) == hrSuccess && lpAutoForward->Value.b) {
            bAllowSendAs = true;
        } else {

            hr = HrGetOneProp(lpStore, PR_MAILBOX_OWNER_ENTRYID, &lpPropOwner);
            if (hr != hrSuccess) {
                lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to get Zarafa mailbox owner id, error code: 0x%08X", lpData->thread, hr);
                goto exit;
            }

            hr = lpData->lpAddrBook->CompareEntryIDs(lpPropOwner->Value.bin.cb, (LPENTRYID)lpPropOwner->Value.bin.lpb,
                                                     lpRepEntryID->Value.bin.cb, (LPENTRYID)lpRepEntryID->Value.bin.lpb, 0, &ulCmpRes);
            if (hr == hrSuccess && ulCmpRes == FALSE) {

                if (strcmp(lpConfig->GetSetting("always_send_delegates"), "yes") == 0) {
                    // pre 6.20 behaviour
                    bAllowDelegate = true;
					HrOpenRepresentStore(lpData, lpRepEntryID->Value.bin.cb, (LPENTRYID)lpRepEntryID->Value.bin.lpb, &lpRepStore);
					// ignore error if unable to open, just the copy of the mail might possibily not be done.
                } else
                    CheckDelegate(lpData, lpStore, lpPropOwner->Value.bin.cb, (LPENTRYID)lpPropOwner->Value.bin.lpb,
                                  lpRepEntryID->Value.bin.cb, (LPENTRYID)lpRepEntryID->Value.bin.lpb, &bAllowDelegate, &lpRepStore);

                if (!bAllowDelegate) {

                    CheckSendAs(lpData, lpStore, lpMailer, lpPropOwner->Value.bin.cb, (LPENTRYID)lpPropOwner->Value.bin.lpb,
                                lpRepEntryID->Value.bin.cb, (LPENTRYID)lpRepEntryID->Value.bin.lpb, &bAllowSendAs, &lpRepStore);

                    if (!bAllowSendAs) {

                        lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "E-mail for user %s may not be sent, notifying user",
                                      lpData->thread, lpUser->lpszUsername);
                        
                        if (SendUndeliverable(lpData, lpMailer, lpStore, lpUserAdmin, lpMessage) != hrSuccess) {
                            lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to create undeliverable message for user %s",
                                          lpData->thread, lpUser->lpszUsername);
                        }

                        goto exit;
                    }
                    // else {}: we are allowed to directly send
                }
                // else {}: allowed with 'on behalf of'
            }
            // else {}: owner and representing are the same, send as normal mail
        }
    }

	if (lpRepStore && parseBool(lpConfig->GetSetting("copy_delegate_mails",NULL,"yes"))) {
		// copy the original message, *before editing* the actual sender data
		// otherwise you might see the wrong sender data in the sent-items version
		// See ECXPLogon::SubmitMessage, which makes a copy to send.
		CopyDelegateMessageToSentItems(lpMessage, lpRepStore, &lpRepMessage);
		// error is logged in function.
	}
    
	if (!bAllowSendAs) {
		// put storeowner info in PR_SENDER_ props
		sPropSender[0].ulPropTag = PR_SENDER_NAME;
		sPropSender[0].Value.lpszA = lpUser->lpszFullName;
		sPropSender[1].ulPropTag = PR_SENDER_ADDRTYPE;
		sPropSender[1].Value.lpszA = "SMTP";
		sPropSender[2].ulPropTag = PR_SENDER_EMAIL_ADDRESS;
		sPropSender[2].Value.lpszA = lpUser->lpszFullEmail;

		sPropSender[3].ulPropTag = PR_SENDER_ENTRYID;
		hr = ECCreateOneOff(lpUser->lpszFullName, (char*)"SMTP", lpUser->lpszFullEmail,
							MAPI_SEND_NO_RICH_INFO|MAPI_UNICODE, &cbEntryIdSender, &lpEntryIdSender);
		// ignore error
		sPropSender[3].Value.bin.cb = cbEntryIdSender;
		sPropSender[3].Value.bin.lpb = (LPBYTE)lpEntryIdSender;

		hr = lpMessage->SetProps(4, sPropSender, NULL);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to update message with sender", lpData->thread);
			goto exit;
		}
	} else {
		// move PR_REPRESENTING to PR_SENDER_NAME
		hr = lpMessage->GetProps((LPSPropTagArray)&sptaMoveReprProps, 0, &cValuesMoveProps, &lpMoveReprProps);
		if (FAILED(hr)) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to find sender information", lpData->thread);
			goto exit;
		}

		hr = lpMessage->DeleteProps((LPSPropTagArray)&sptaMoveReprProps, NULL);
		if (FAILED(hr)) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to remove sender information", lpData->thread);
			goto exit;
		}

		lpMoveReprProps[0].ulPropTag = PROP_TAG(PROP_TYPE(lpMoveReprProps[0].ulPropTag), PROP_ID(PR_SENDER_NAME));
		lpMoveReprProps[1].ulPropTag = PROP_TAG(PROP_TYPE(lpMoveReprProps[1].ulPropTag), PROP_ID(PR_SENDER_ADDRTYPE));
		lpMoveReprProps[2].ulPropTag = PROP_TAG(PROP_TYPE(lpMoveReprProps[2].ulPropTag), PROP_ID(PR_SENDER_EMAIL_ADDRESS));
		lpMoveReprProps[3].ulPropTag = PROP_TAG(PROP_TYPE(lpMoveReprProps[3].ulPropTag), PROP_ID(PR_SENDER_ENTRYID));
		lpMoveReprProps[4].ulPropTag = PROP_TAG(PROP_TYPE(lpMoveReprProps[4].ulPropTag), PROP_ID(PR_SENDER_SEARCH_KEY));

		hr = lpMessage->SetProps(5, lpMoveReprProps, NULL);
		if (FAILED(hr)) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to update sender information", lpData->thread);
			goto exit;
		}
	}

	hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to save message before sending", lpData->thread);
		goto exit;
	}

	
	// Expand recipients with ADDRTYPE=ZARAFA to multiple ADDRTYPE=SMTP recipients
	hr = ExpandRecipients(lpData, lpMessage);
	if(hr != hrSuccess)
	    lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to expand message recipient groups", lpData->thread);

	hr = RewriteRecipients(lpData, lpMessage);
	if (hr != hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to rewrite recipients", lpData->thread);

	hr = UniqueRecipients(lpData, lpMessage);
	if (hr != hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Unable to remove duplicate recipients", lpData->thread);

	// Now hand message to library which will send it, inetmapi will handle addressbook
	hr = IMToINet(lpData->lpMAPISession, lpData->lpAddrBook, lpMessage, lpMailer, sopt, lpLogger);

	// log using fatal, all other log messages are otherwise somewhat meaningless
	if (hr == MAPI_W_NO_SERVICE) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to connect to SMTP server, retrying mail for user %s later",
					  lpData->thread, lpUser->lpszUsername);
		goto exit;
	} else if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "E-mail for user %s could not be sent, notifying user",
					  lpData->thread, lpUser->lpszUsername);
		if (SendUndeliverable(lpData, lpMailer, lpStore, lpUserAdmin, lpMessage) != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, THREAD_DBG "Unable to create undeliverable message for user %s",
						  lpData->thread, lpUser->lpszUsername);
		}
		goto exit;
	} else {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "E-mail for user %s was accepted by SMTP server",
					  lpData->thread, lpUser->lpszUsername);
	}

	// If we have a repsenting message, save that now in the sent-items of that user
	if (lpRepMessage) {
		if (lpRepMessage->SaveChanges(0) != hrSuccess)
			lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Representee's mail copy could not be saved: error 0x%08X", lpData->thread, hr);
	}

exit:
	*lppMessage = lpMessage;

	if (lpRepMessage)
		lpRepMessage->Release();

	if (lpRepStore)
		lpRepStore->Release();

	if (lpAutoForward)
	    MAPIFreeBuffer(lpAutoForward);

	if (lpOwner)
		MAPIFreeBuffer(lpOwner);

	if (lpPropOwner)
		MAPIFreeBuffer(lpPropOwner);

	if (lpMoveReprProps)
		MAPIFreeBuffer(lpMoveReprProps);

	//Free entry id's
	if (lpEntryIdRepr)
		MAPIFreeBuffer(lpEntryIdRepr);

	if (lpEntryIdSender)
		MAPIFreeBuffer(lpEntryIdSender);

	// Free checked properties
	if (lpPropSRNam)
		MAPIFreeBuffer(lpPropSRNam);

	if (lpPropSREma)
		MAPIFreeBuffer(lpPropSREma);

	if (lpPropSREid)
		MAPIFreeBuffer(lpPropSREid);

	if (lpRepEntryID)
		MAPIFreeBuffer(lpRepEntryID);

	// Free everything else..
	if (lpSecurity)
		lpSecurity->Release();

	if (lpUser)
		MAPIFreeBuffer(lpUser);

	if (lpsPropObject)
		MAPIFreeBuffer(lpsPropObject);

	if (lpStore)
		lpStore->Release();

	if (lpSubject)
		MAPIFreeBuffer(lpSubject);

	if (lpMsgSize)
		MAPIFreeBuffer(lpMsgSize);

	if (lpDeferSendTime)
		MAPIFreeBuffer(lpDeferSendTime);

	return hr;
}

void *ProcessMessageThread(void *threadid)
{
	HRESULT hr = hrSuccess;
	struct thread_data *lpData = (struct thread_data*)threadid;
	ECSender *lpMailer = NULL;
	ECUSER *lpUserAdmin = NULL;
	LPMESSAGE lpMessage = NULL;

	if (nQuit)
		goto exit;

	hr = InitThreadData(lpData);
	if (hr != hrSuccess)
		goto exit;

	lpMailer = CreateSender(lpLogger, lpData->lpszSMTP);

	hr = lpData->lpServiceAdmin->GetUser(g_cbDefaultEid, (LPENTRYID)g_lpDefaultEid, &lpUserAdmin);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, THREAD_DBG "Unable to get user admin information from store. error code: 0x%08X", lpData->thread, hr);
		goto exit;
	}

	hr = ProcessMessage(lpData, lpMailer, lpUserAdmin, &lpMessage);
	if (hr == hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_INFO, THREAD_DBG "Message successfully sent", lpData->thread);
	else if (hr == MAPI_E_WAIT || hr == MAPI_W_NO_SERVICE)
		goto exit; /* User indicated a particular date/time to send the message, or SMTP server not available */
	else
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Removing failed message from queue", lpData->thread);

	/* always remove message from queue, except when date was set by user when to send the mail */
	// we need to remove from the table before we do the DoSentMail() call, because the entry id will change due to a move action
	hr = lpData->lpSpooler->DeleteFromMasterOutgoingTable(lpData->sMessage.cb, (LPENTRYID)lpData->sMessage.lpb, lpData->ulFlags);
	if (hr != hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_WARNING, THREAD_DBG "Could not remove message from queue, error code: 0x%08X", lpData->thread, hr);

	// If the client wants us to do the DoSentMail(), then do it now
	if(lpMessage && (lpData->ulFlags & EC_SUBMIT_DOSENTMAIL)) {
		DoSentMail(lpData->lpMAPISession, 0, lpMessage);
		lpMessage = NULL; // DoSentMail frees the object (even on error!)
	}


exit:
	pthread_mutex_lock(&threadmap_lock);
	threadmap.erase(lpData->sMessage);
	pthread_mutex_unlock(&threadmap_lock);

	if (lpMessage)
		lpMessage->Release();

	if (lpMailer)
		delete lpMailer;

	if(lpUserAdmin)
		MAPIFreeBuffer(lpUserAdmin);

	if (lpData)
		FreeThreadData(lpData);

    // Do not pthread_exit() because linuxthreads is broken and will not free any objects
	// pthread_exit((void*)hr);
	return NULL;
}

HRESULT ProcessAllEntries(struct thread_data *lpAdminData, IMAPITable *lpTable, char *lpSMTP, char *szPath) {
	HRESULT 	hr				= hrSuccess;
	int			rc				= 0;
	unsigned int ulMaxThreads	= 0;
	unsigned int ulFreeThreads	= 0;
	ULONG		ulRowCount		= 0;
	LPSRowSet	lpsRowSet		= NULL;
	struct thread_data *lpData	= NULL;
	threaditer_t iterThread;
	map<SBinary, pthread_t>	localthreadmap;

	hr = lpTable->GetRowCount(0,&ulRowCount);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get outgoing queue count");
		goto exit;
	}

	lpLogger->Log(EC_LOGLEVEL_INFO, "Number of messages in the queue: %d", ulRowCount);

	ulMaxThreads = atoi(lpConfig->GetSetting("max_threads"));
	if (ulMaxThreads == 0)
		ulMaxThreads = 1;

    /* Copy the thread map now; since the processing thread removes itself from the outgoing queue before removing
     * itself from the threadmap, the threadmap here can only contain MORE threads than are present during the QueryRows() call
     * below. Since we're checking to see whether a message is being sent later, we can use this local list to make sure there is
     * no race between the QueryRows(1,..) call and the threadmap.find() call later
     */
    pthread_mutex_lock(&threadmap_lock);
    for(iterThread = threadmap.begin(); iterThread != threadmap.end(); iterThread++) {
        // We have to copy this data since it is freed by the thread
        SBinary bin;
        bin.cb = iterThread->first.cb;
        bin.lpb = new BYTE [bin.cb];
        memcpy(bin.lpb, iterThread->first.lpb, bin.cb);
        
        localthreadmap.insert(std::pair<SBinary, pthread_t>(bin, iterThread->second));
    }
    pthread_mutex_unlock(&threadmap_lock);

	while(!nQuit) {
		if (lpsRowSet) {
			FreeProws(lpsRowSet);
			lpsRowSet = NULL;
		}

		pthread_mutex_lock(&threadmap_lock);
		ulFreeThreads = ulMaxThreads - threadmap.size();
		pthread_mutex_unlock(&threadmap_lock);

		if (!ulFreeThreads || ulFreeThreads > ulMaxThreads) {
			Sleep(100);
			continue;	/* Continue looping until threads become available */
		}
		

		hr = lpTable->QueryRows(1, 0, &lpsRowSet);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to fetch data from table, error code: 0x%08X", hr);
			goto exit;
		}
		if (lpsRowSet->cRows == 0)		// All rows done
			goto exit;

		// Check whether the row contains the entryid and store id
		if (lpsRowSet->aRow[0].lpProps[0].ulPropTag != PR_EC_MAILBOX_OWNER_ACCOUNT ||
			lpsRowSet->aRow[0].lpProps[1].ulPropTag != PR_STORE_ENTRYID ||
		    lpsRowSet->aRow[0].lpProps[2].ulPropTag != PR_ENTRYID ||
		    lpsRowSet->aRow[0].lpProps[3].ulPropTag != PR_EC_OUTGOING_FLAGS) {
			/* Is this even possible? */
			lpLogger->Log(EC_LOGLEVEL_FATAL, "BUG: invalid data in OutgoingQueue");
			if (lpsRowSet->aRow[0].lpProps[2].ulPropTag == PR_ENTRYID && lpsRowSet->aRow[0].lpProps[3].ulPropTag == PR_EC_OUTGOING_FLAGS) {
				// we can remove this message
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Removing invalid entry from OutgoingQueue");

				hr = lpAdminData->lpSpooler->DeleteFromMasterOutgoingTable(lpsRowSet->aRow[0].lpProps[2].Value.bin.cb,
																	  (LPENTRYID)lpsRowSet->aRow[0].lpProps[2].Value.bin.lpb,
																	  lpsRowSet->aRow[0].lpProps[3].Value.ul);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_WARNING, "Could not remove invalid message from queue, error code: 0x%08X", hr);
					goto exit;
				}
			} else {
				lpLogger->Log(EC_LOGLEVEL_WARNING, "Could not remove invalid message from queue");
				// this error makes the spooler disconnect from the server, and reconnect again (nQuit still false)
				hr = MAPI_E_NOT_FOUND;
				goto exit;
			}
			continue;
		}

		/* Check if there is already an active thread for this message */
		if (localthreadmap.find(lpsRowSet->aRow[0].lpProps[2].Value.bin) != localthreadmap.end()) {
			continue;
		}

		hr = AllocThreadData(&lpData, szPath, lpAdminData->lpAdminSession, lpSMTP,
							lpsRowSet->aRow[0].lpProps[0].Value.lpszA,
							&lpsRowSet->aRow[0].lpProps[1].Value.bin,
							&lpsRowSet->aRow[0].lpProps[2].Value.bin,
							lpsRowSet->aRow[0].lpProps[3].Value.ul);
		if (hr != hrSuccess)
			goto exit;

		rc = pthread_create(&lpData->thread, &thread_attrib, ProcessMessageThread, lpData);
		if (rc) {
			/* lpData should only be cleaned up when the thread creation failed,
			 * otherwise the thread will clean lpData up. */
			FreeThreadData(lpData);
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		pthread_mutex_lock(&threadmap_lock);
		threadmap.insert(threadpair_t(lpData->sMessage, lpData->thread));
        // We have to copy this data since it is freed by the thread
        SBinary bin;
        bin.cb = lpsRowSet->aRow[0].lpProps[2].Value.bin.cb;
        bin.lpb = new BYTE [bin.cb];
        memcpy(bin.lpb, lpsRowSet->aRow[0].lpProps[2].Value.bin.lpb, bin.cb);
        
        localthreadmap.insert(std::pair<SBinary, pthread_t>(bin, iterThread->second));
		pthread_mutex_unlock(&threadmap_lock);
	}

	// Ok we're done
exit:
    for(iterThread = localthreadmap.begin(); iterThread != localthreadmap.end(); iterThread++) {
        delete [] iterThread->first.lpb;
    }

	if (lpsRowSet)
		FreeProws(lpsRowSet);

	return hr;
}

HRESULT ProcessQueue(char* lpSMTP, char *szPath)
{
	HRESULT				hr				= hrSuccess;
	struct thread_data	*lpData			= NULL;
	IMAPITable			*lpTable		= NULL;
	IMAPIAdviseSink		*lpAdviseSink	= NULL;
	ULONG				ulConnection	= 0;

	SizedSPropTagArray(4, sOutgoingCols) = {
		4, {
			PR_EC_MAILBOX_OWNER_ACCOUNT,
			PR_STORE_ENTRYID,
			PR_ENTRYID,
			PR_EC_OUTGOING_FLAGS,
		}
	};
	
	SSortOrderSet sSort = { 1, 0, 0, { { PR_EC_HIERARCHYID, TABLE_SORT_ASCEND } } };

	hr = AllocThreadData(&lpData, szPath);
	if (hr != hrSuccess)
		goto exit;

	// logs in to the server
	hr = InitThreadData(lpData);
	if (hr !=hrSuccess)
		goto exit;

	lpLogger->Log(EC_LOGLEVEL_WARNING, "Connection to Zarafa server succeeded"); // show up on default log level

	disconnects = 0;			// first call succeeded, assume all is well.

	// Request the master outgoing table
	hr = lpData->lpSpooler->GetMasterOutgoingTable(0, &lpTable);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Master outgoing queue not available");
		goto exit;
	}

	hr = lpTable->SetColumns((LPSPropTagArray)&sOutgoingCols, 0);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to setColumns() on OutgoingQueue");
		goto exit;
	}
	
	// Sort by ascending hierarchyid: first in, first out queue
	hr = lpTable->SortTable(&sSort, 0);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to SortTable() on OutgoingQueue");
		goto exit;
	}

	hr = HrAllocAdviseSink(AdviseCallback, NULL, &lpAdviseSink);	
	if (hr != hrSuccess)
		goto exit;

	// notify on new mail in the outgoing table
	hr = lpTable->Advise(fnevTableModified, lpAdviseSink, &ulConnection);

	while(!nQuit) {
		bMessagesWaiting = false;
		lpTable->SeekRow(BOOKMARK_BEGINNING, 0, NULL);

		hr = ProcessAllEntries(lpData, lpTable, lpSMTP, szPath);
		if(hr != hrSuccess)
			goto exit;

		// Exit signal, break the operation
		if(nQuit)
			break;

		pthread_mutex_lock(&hMutexMessagesWaiting);
		if(!bMessagesWaiting) {
			struct timespec timeout;
			struct timeval now;

			// Wait for max 60 sec, then run queue anyway
			gettimeofday(&now,NULL);
			timeout.tv_sec = now.tv_sec + 60;
			timeout.tv_nsec = now.tv_usec * 1000;

			pthread_cond_timedwait(&hCondMessagesWaiting, &hMutexMessagesWaiting, &timeout);
		}
		pthread_mutex_unlock(&hMutexMessagesWaiting);
	}

exit:
	if (ulConnection)
		lpTable->Unadvise(ulConnection);

	if (lpAdviseSink)
		lpAdviseSink->Release();

	if (lpTable)
		lpTable->Release();

	if (lpData)
		FreeThreadData(lpData);

	return hr;
}


void sighup(int signr) {
	// In Win32, the signal is sent in a seperate, special signal thread. So this test is
	// not needed or required.
	if (pthread_equal(pthread_self(), mainthread)==0)
		return;
	if (lpConfig)
		lpConfig->ReloadSettings();

	if (lpLogger) {
		if (lpConfig) {
			char *ll = lpConfig->GetSetting("log_level");
			int new_ll = ll ? atoi(ll) : 2;
			lpLogger->SetLoglevel(new_ll);
		}

		lpLogger->Reset();
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Log connection was reset");
	}
}

void sigusr2(int signr)
{
	if (pthread_equal(pthread_self(), mainthread)==0)
		return;

	pthread_mutex_lock(&threadmap_lock);

	lpLogger->Log(EC_LOGLEVEL_FATAL, "Spooler stats:");
	lpLogger->Log(EC_LOGLEVEL_FATAL, "Threads: %d", threadmap.size());
	lpLogger->Log(EC_LOGLEVEL_FATAL, "Disconnects: %d", disconnects);

	pthread_mutex_unlock(&threadmap_lock);
}

// SIGSEGV catcher
#include <execinfo.h>

void sigsegv(int signr)
{
	void *bt[64];
	int i, n;
	char **btsymbols;

	if(!lpLogger)
		goto exit;

	lpLogger->Log(EC_LOGLEVEL_FATAL, "Caught SIGSEGV (%d), traceback:", signr);

	n = backtrace(bt, 64);
        
	btsymbols = backtrace_symbols(bt, n);

	for (i = 0; i < n; i++) {
		if (btsymbols)
			lpLogger->Log(EC_LOGLEVEL_FATAL, "%016p %s", bt[i], btsymbols[i]);
		else
			lpLogger->Log(EC_LOGLEVEL_FATAL, "%016p", bt[i]);
	}

	lpLogger->Log(EC_LOGLEVEL_FATAL, "When reporting this traceback, please include Linux distribution name, system architecture and Zarafa version.");

exit:
	kill(0, SIGSEGV);
}


int main(int argc, char *argv[]) {

	HRESULT hr = hrSuccess;
	char *szConfig = "/etc/zarafa/spooler.cfg";
	char *szPath = NULL;
	char *szSMTP = NULL;
	int c;
	int daemonize = 1;

	// options
	enum {
		OPT_HELP,
		OPT_CONFIG,
		OPT_HOST,
		OPT_FOREGROUND
	};
	struct option long_options[] = {
		{ "help", 0, NULL, OPT_HELP },		// help text
		{ "config", 1, NULL, OPT_CONFIG },	// config file
		{ "host", 1, NULL, OPT_HELP },		// zarafa host location
		{ "foreground", 1, NULL, OPT_FOREGROUND },		// do not daemonize
		{ NULL, 0, NULL, 0 }
	};

	// Default settings
	const configsetting_t lpDefaults[] = {
		{ "smtp_server","localhost" },
		{ "server_socket", CLIENT_ADMIN_SOCKET },
		{ "run_as_user", "" },
		{ "run_as_group", "" },
		{ "pid_file", "/var/run/zarafa-spooler.pid" },
		{ "running_path", "/" },
		{ "log_method","file" },
		{ "log_file","-" },
		{ "log_level","2", CONFIGSETTING_RELOADABLE },
		{ "log_timestamp","1" },
		{ "sslkey_file", "" },
		{ "sslkey_pass", "" },
		{ "max_threads", "5", CONFIGSETTING_RELOADABLE },
		{ "fax_domain", "", CONFIGSETTING_RELOADABLE },
		{ "fax_international", "+", CONFIGSETTING_RELOADABLE },
		{ "always_send_delegates", "no", CONFIGSETTING_RELOADABLE },
		{ "always_send_tnef", "no", CONFIGSETTING_RELOADABLE },
		{ "allow_redirect_spoofing", "yes", CONFIGSETTING_RELOADABLE },
		{ "copy_delegate_mails", "yes", CONFIGSETTING_RELOADABLE },
		{ NULL, NULL },
	};

	while(1) {
		c = my_getopt_long(argc, argv, "c:h:iuVF", long_options, NULL);

		if(c == -1)
			break;

		switch(c) {
		case OPT_CONFIG:
		case 'c':
			szConfig = my_optarg;
			break;
		case OPT_HOST:
		case 'h':
			szPath = my_optarg;
			break;
		case 'i': // Install service
		case 'u': // Uninstall service
			break;
		case 'F':
		case OPT_FOREGROUND:
			daemonize = 0;
			break;
		case 'V':
			cout << "Product version:\t" <<  PROJECT_VERSION_SPOOLER_STR << endl
				 << "File version:\t\t" << PROJECT_SVN_REV_STR << endl;
			return 1;
		case OPT_HELP:
		default:
			print_help(argv[0]);
			return 1;
		}
	}

	lpConfig = new ECConfig(lpDefaults);
	if (szConfig) {
		if (!lpConfig->LoadSettings(szConfig)) {
			fprintf(stderr,"Unable to open configuration file %s\n", szConfig);
			fprintf(stderr,"Continuing with defaults\n");
		}
		if (lpConfig->HasErrors()) {
			lpLogger = new ECLogger_File(EC_LOGLEVEL_FATAL, 0, "-"); // create fatal logger without a timestamp to stderr
			LogConfigErrors(lpConfig, lpLogger);
			hr = E_FAIL;
			goto exit;
		}
	}

	// commandline overwrites spooler.cfg
	if(my_optind < argc) {
		szSMTP = argv[my_optind];
	} else {
		szSMTP = lpConfig->GetSetting("smtp_server");
	}

	// setup logging
	lpLogger = CreateLogger(lpConfig, argv[0], "ZarafaSpooler");

	if (lpConfig->HasWarnings())
		LogConfigErrors(lpConfig, lpLogger);

	// set socket filename
	if (!szPath)
		szPath = lpConfig->GetSetting("server_socket");

	mainthread = pthread_self();

	signal(SIGTERM, sighandle);
	signal(SIGINT, sighandle);
	signal(SIGHUP, sighup);
	/* SIGUSR1 signals are not being handled for some reason :S */
	signal(SIGUSR2, sigusr2);

    // SIGSEGV backtrace support
    stack_t st;
    struct sigaction act;

    memset(&st, 0, sizeof(st));
    memset(&act, 0, sizeof(act));

    st.ss_sp = malloc(65536);
    st.ss_flags = 0;
    st.ss_size = 65536;

    act.sa_handler = sigsegv;
    act.sa_flags = SA_ONSTACK | SA_RESETHAND;

    sigaltstack(&st, NULL);
    sigaction(SIGSEGV, &act, NULL);
    sigaction(SIGBUS, &act, NULL);
    sigaction(SIGABRT, &act, NULL);

// 	act.sa_handler = sighup;
// 	act.sa_flags = SA_RESTART | SA_NODEFER;
//     sigaction(SIGHUP, &act, NULL);

	nQuit = bMessagesWaiting = false;

	hr = MAPIInitialize(NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to initialize");
		goto exit;
	}

	// fork if needed and drop privileges as requested.
	// this must be done before we do anything with pthreads
	if (daemonize && unix_daemonize(lpConfig, lpLogger))
		goto exit;
	if (!daemonize)
		setsid();
	if (unix_create_pidfile(argv[0], lpConfig, lpLogger, false) < 0)
		goto exit;
	if (unix_runas(lpConfig, lpLogger))
		goto exit;
		hr = running_server(szSMTP, szPath);

	MAPIUninitialize();

exit:
	if (lpLogger)
		delete lpLogger;
	if (lpConfig)
		delete lpConfig;

	return hr == hrSuccess ? 0 : 1;
}

HRESULT running_server(char* szSMTP, char* szPath)
{
	HRESULT hr = hrSuccess;
	unsigned int ulThreads = 1;
	unsigned int ulCount = 0;

	pthread_mutex_init(&hMutexMessagesWaiting, NULL);
	pthread_cond_init(&hCondMessagesWaiting, NULL);

	pthread_mutex_init(&threadmap_lock, NULL);

	pthread_attr_init(&thread_attrib);
	pthread_attr_setdetachstate(&thread_attrib, PTHREAD_CREATE_DETACHED);

	lpLogger->Log(EC_LOGLEVEL_FATAL, "Starting zarafa-spooler version " PROJECT_VERSION_SPOOLER_STR " (" PROJECT_SVN_REV_STR "), pid %d", getpid());
	lpLogger->Log(EC_LOGLEVEL_FATAL, "using SMTP server: %s", szSMTP);

	disconnects = 0;

	while(1) {
		hr = ProcessQueue(szSMTP, szPath);

		if(nQuit || disconnects > MAX_DISCONNECTS)
			break;

		disconnects++;
		Sleep(3000);			// wait 3s until retry to connect
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Server connection lost. Reconnecting");
	}

	if (disconnects > MAX_DISCONNECTS) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Could not connect to server, exiting.");
		hr = 1;
	}

	/* Wait for all threads to finish, for 60 seconds */
	while (ulThreads > 0 && ulCount < 60) {
		pthread_mutex_lock(&threadmap_lock);
		ulThreads = threadmap.size();
		pthread_mutex_unlock(&threadmap_lock);

		if (ulThreads == 0)
			break;

		if ((ulCount % 5) == 0)
			lpLogger->Log(EC_LOGLEVEL_WARNING, "Still waiting for %d thread%c to exit.", ulThreads, ulThreads!=1?'s':' ');

		Sleep(1000);
		ulCount++;
	}

	lpLogger->Log(EC_LOGLEVEL_FATAL, "Spooler shutdown complete");

	pthread_mutex_destroy(&threadmap_lock);

	pthread_mutex_destroy(&hMutexMessagesWaiting);
	pthread_cond_destroy(&hCondMessagesWaiting);

	return hr;
}
