/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "ECDatabaseUtils.h"
#include "ECSessionManager.h"
#include "ECDBDef.h"
#include "ECDatabaseUpdate.h"
#include "ECLicenseClient.h"
#include "ECGuid.h"

#include "soapH.h"

#include <mapidefs.h>
#include <mapitags.h>

#include <sys/times.h>
#include <time.h>

#include <algorithm>
#include <sstream>
#include <set>
#include <stdio.h>

#include "ECTags.h"
#include "stringutil.h"
#include "SOAPUtils.h"
#include "ZarafaCode.h"
#include "Trace.h"
#include "ZarafaCmd.nsmap"

#include "ZarafaICS.h"

#include "Zarafa.h"
#include "ZarafaUtil.h"
#include "md5.h"

#include "ECAttachmentStorage.h"
#include "ECGenProps.h"
#include "ECUserManagement.h"
#include "ECSecurity.h"
#include "ECICS.h"
#include "ECMultiStoreTable.h"
#include "ECStatsCollector.h"

#include "ECDefs.h"
#include <EMSAbTag.h>
#include <edkmdb.h>
#include "ecversion.h"
#include "mapiext.h"

#include "../server/ECSoapServerConnection.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern ECSessionManager*	g_lpSessionManager;
extern ECStatsCollector*	g_lpStatsCollector;



// Above EC_TABLE_CHANGE_THRESHOLD, a TABLE_CHANGE notification is sent instead of individual notifications
#define EC_TABLE_CHANGE_THRESHOLD 10


ECRESULT CreateEntryId(GUID guidStore, unsigned int ulObjType, entryId** lppEntryId)
{
	ECRESULT	er = erSuccess;
	entryId*	lpEntryId = NULL;
	EID			eid;

	if(lppEntryId == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if(CoCreateGuid(&eid.uniqueId) != hrSuccess) {
		er = ZARAFA_E_CALL_FAILED;
		goto exit;
	}

	eid.guid = guidStore;
	eid.ulType = ulObjType;

	lpEntryId = new entryId;

	lpEntryId->__size = sizeof(EID);
	lpEntryId->__ptr = new unsigned char[lpEntryId->__size];

	memcpy(lpEntryId->__ptr, &eid, lpEntryId->__size);

	*lppEntryId = lpEntryId;

exit:

	return er;
}

ECRESULT GetSourceKey(unsigned int ulObjId, SOURCEKEY *lpSourceKey) {
	ECRESULT er = erSuccess;
	unsigned char *lpData = NULL;
	unsigned int cbData = 0;

	er = g_lpSessionManager->GetCacheManager()->GetPropFromObject(PROP_ID(PR_SOURCE_KEY), ulObjId, NULL, &cbData, &lpData);
	if(er != erSuccess)
		goto exit;

	*lpSourceKey = SOURCEKEY(cbData, (char *)lpData);

exit:
	if(lpData)
		delete [] lpData;
	return er;
}



ECRESULT GetLocalId(ECSession *lpecSession, entryId sUserId, unsigned int ulLegacyUserId, unsigned int *lpulUserId, std::string *lpstrExternId, bool *lpbHasLocalStore)
{
	ECRESULT 			er = erSuccess;
	unsigned int		ulUserId = 0;
	std::string			strExternId;
	unsigned int		ulUserObjType = 0;
	userobject_type_t	sUserObjType = USEROBJECT_TYPE_USER;
	objectdetails_t		sDetails;

	if (lpulUserId == NULL || (lpecSession == NULL && lpbHasLocalStore != NULL))
	{
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// If no entryid is present, use the 'current' user.
	if (ulLegacyUserId == 0 && sUserId.__size == 0) {
		// When lpstrExternId or lpbHasLocalStore are requested, the 'current' user will not be
		// requested in this way. However, to make sure a caller does expect a result in the future 
		// we'll return an error in that case.
		if (lpstrExternId != NULL || lpbHasLocalStore != NULL)
			er = ZARAFA_E_INVALID_PARAMETER;
		else
			*lpulUserId = 0;

		goto exit;
	}

	if (sUserId.__ptr) {
		// Extract the information from the entryid.
		er = ABEntryIDToID(&sUserId, &ulUserId, &strExternId, &ulUserObjType);
		if (er != erSuccess)
			goto exit;

		er = MAPITypeToType(ulUserObjType, &sUserObjType);
		if (er != erSuccess)
			goto exit;

		// If an extern id is present, we should get an object based on that.
		if (!strExternId.empty()) {
			er = g_lpSessionManager->GetCacheManager()->GetUserObject(strExternId, sUserObjType, NULL, &ulUserId, NULL);
		
			if (er != erSuccess && sUserObjType == USEROBJECT_TYPE_USER) {
				er = g_lpSessionManager->GetCacheManager()->GetUserObject(strExternId, USEROBJECT_TYPE_NONACTIVE, NULL, &ulUserId, NULL);
			}
			if (er != erSuccess && sUserObjType == USEROBJECT_TYPE_COMPANY) {
				er = g_lpSessionManager->GetCacheManager()->GetUserObject(strExternId, USEROBJECT_TYPE_ADDRESSLIST, NULL, &ulUserId, NULL);
			}
		}
	} else {
		// use user id from 6.20 and older clients
		ulUserId = ulLegacyUserId;
		er = g_lpSessionManager->GetCacheManager()->GetUserObject(ulLegacyUserId, &sUserObjType, NULL, &strExternId, NULL);
	}
	if (er != erSuccess)
		goto exit;

	// Optionaly check if the object has a local store. This only makes sense for
	// user objects. 
	if (lpbHasLocalStore) {
 		if (g_lpSessionManager->IsDistributedSupported()) {
            er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulUserId, &sDetails);
			if (er != erSuccess)
				goto exit;

			*lpbHasLocalStore = sDetails.GetPropString(OB_PROP_S_SERVERNAME).compare(g_lpSessionManager->GetConfig()->GetSetting("server_name")) == 0;
		} else 
			*lpbHasLocalStore = true;
	}

	*lpulUserId = ulUserId;
	if (lpstrExternId)
		*lpstrExternId = strExternId;

exit:
	return er;
}

ECRESULT GetABEntryID(unsigned int ulUserId, soap *lpSoap, entryId *lpUserId)
{
	ECRESULT			er = erSuccess;
	std::string			strExternId;
	entryId				sUserId = {0};
	PABEID				lpabeid = NULL;
	userobject_type_t	sUserObjType;
	unsigned int		ulType = 0;


	if (lpSoap == NULL)
	{
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}	

	if (ulUserId == ZARAFA_UID_SYSTEM) {
		ulType = MAPI_MAILUSER;
	} else if (ulUserId == ZARAFA_UID_EVERYONE) {
		ulType = MAPI_DISTLIST;
	} else {
		er = g_lpSessionManager->GetCacheManager()->GetUserExternId(ulUserId, &sUserObjType, &strExternId, NULL);
		if (er != erSuccess)
			goto exit;

		er = TypeToMAPIType(sUserObjType, &ulType);
		if (er != erSuccess)
			goto exit;
	}


	er = ABIDToEntryID(lpSoap, ulUserId, strExternId, ulType, &sUserId);
	if (er != erSuccess)
		goto exit;

	*lpUserId = sUserId;	// pointer (__ptr) is copied, not data

exit:
	return er;
}

ECRESULT PeerIsServer(struct soap *soap, const std::string &strServerName, const std::string &strHttpPath, const std::string &strSslPath, bool *lpbResult)
{
	ECRESULT		er = erSuccess;
	bool			bResult = false;
	std::string		strPath;
	std::string		strHost;
	int				ulPos = 0;
	struct hostent	*lpHostEnt = NULL;

	if (soap == NULL || lpbResult == NULL)
	{
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// First check if we're connecting through unix-socket/named-pipe and if the request url matches this server
	if (((SOAPINFO*)soap->user)->ulConnectionType == CONNECTION_TYPE_NAMED_PIPE &&
			strServerName.compare(g_lpSessionManager->GetConfig()->GetSetting("server_name")) == 0)
		bResult = true;

	else
	{
		strPath = strHttpPath;
		if (strPath.empty())
			strPath = strSslPath;
		if (!strPath.empty())
		{
			ulPos = strPath.find("://");
			strPath = strPath.substr(ulPos + 3);

			ulPos = strPath.find(":");
			strHost = strPath.substr(0, ulPos);
			
			lpHostEnt = gethostbyname(strHost.c_str());
			if (!lpHostEnt)
			{
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}

			if (lpHostEnt->h_addrtype != AF_INET)
			{
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}

			int i = 0;
			while (lpHostEnt->h_addr_list[i])
			{
				if (memcmp(lpHostEnt->h_addr_list[i], &soap->peer.sin_addr.s_addr, lpHostEnt->h_length) == 0)
				{
					bResult = true;
					break;
				}

				i++;
			}
		}
	}

	*lpbResult = bResult;

exit:
	return er;
}

ECRESULT GetBestServerPath(struct soap *soap, ECSession *lpecSession, const std::string &strServerName, std::string *lpstrServerPath)
{
	ECRESULT	er = erSuccess;
	std::string	strServerPath;
	bool		bConnectPipe = false;

	serverdetails_t	sServerDetails;
	std::string		strFilePath;
	std::string		strHttpPath;
	std::string		strSslPath;

	if (soap == NULL || soap->user == NULL || lpstrServerPath == NULL)
	{
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpecSession->GetUserManagement()->GetServerDetails(strServerName, &sServerDetails);
	if (er != erSuccess)
		goto exit;

	strFilePath = sServerDetails.GetFilePath();
	strHttpPath = sServerDetails.GetHttpPath();
	strSslPath = sServerDetails.GetSslPath();

	if (!strFilePath.empty())
	{
		er = PeerIsServer(soap, strServerName, strHttpPath, strSslPath, &bConnectPipe);
		if (er != erSuccess)
			goto exit;
	} else {
		// TODO: check if same server, and set strFilePath 'cause it's known

		bConnectPipe = false;
	}

	if (bConnectPipe)
		strServerPath = strFilePath;
	else
		switch (((SOAPINFO*)soap->user)->ulConnectionType)
		{
		case CONNECTION_TYPE_TCP:
			if (!strHttpPath.empty())
				strServerPath = strHttpPath;
			else if (!strSslPath.empty())
				strServerPath = strSslPath;
			break;

		case CONNECTION_TYPE_SSL:
			if (!strSslPath.empty())
				strServerPath = strSslPath;
			break;

		case CONNECTION_TYPE_NAMED_PIPE:
			if (!strSslPath.empty())
				strServerPath = strSslPath;
			else if (!strHttpPath.empty())
				strServerPath = strHttpPath;
			break;
		}

	if (strServerPath.empty())
	{
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	*lpstrServerPath = strServerPath;

exit:
	return er;
}


ECRESULT GetFolderCount(ECSession *lpecSession, unsigned int ulObjId, unsigned int *lpulContentCount, unsigned int *lpulUnreadCount,
						unsigned int *lpulAssocCount, unsigned int *lpulSubFolderCount, unsigned int *lpulDeleteMsgCount,
						unsigned int *lpulDeleteFolderCount, unsigned int *lpulDeleteAssocMsgCount);

// exception: This function does internal Begin + Commit/Rollback
ECRESULT MoveObjects(ECSession *lpSession, ECDatabase *lpDatabase, ECListInt* lplObjectIds, unsigned int ulDestFolderId, unsigned int ulSyncId);
// these functions don't do Begin + Commit/Rollback
ECRESULT CreateObject(ECSession *lpecSession, ECDatabase *lpDatabase, unsigned int ulParentObjId, unsigned int ulObjType, unsigned int ulFlags, unsigned int *lpulObjId);
ECRESULT WriteProps(struct soap *soap, ECSession *lpecSession, ECDatabase *lpDatabase, struct saveObject *lpsSaveObj, unsigned int ulObjId, bool fNewItem, unsigned int ulSyncId, struct saveObject *lpsReturnObj, bool *lpfHaveChangeKey);
ECRESULT DeleteObjects(ECSession *lpSession, ECDatabase *lpDatabase, ECListInt *lpsObjectList, unsigned int ulFlags, unsigned int ulSyncId, bool fNoBegin, bool bCheckPermission);
ECRESULT WriteLocalCommitTimeMax(struct soap *soap, ECDatabase *lpDatabase, unsigned int ulStoreId, unsigned int ulFolderId, propVal **ppvTime);

ECRESULT DoNotifySubscribe(ECSession *lpecSession, unsigned long long ulSessionId, struct notifySubscribe *notifySubscribe);
ECRESULT SaveLogonTime(ECSession *lpecSession, bool bLogon);

// this belongs to the DeleteObjects function
typedef struct {
	unsigned int ulId;
	unsigned int ulParent;
	bool fRoot;
	unsigned int ulObjType;
	unsigned int ulFlags;
	unsigned int ulSubmit;
	unsigned int ulStoreId;
	unsigned int ulObjSize;
	SOURCEKEY sSourceKey;
	SOURCEKEY sParentSourceKey;
	entryId sEntryId;
} DELETEITEM;

#define EC_DELETE_FOLDERS		0x00000001
#define EC_DELETE_MESSAGES		0x00000002
#define EC_DELETE_RECIPIENTS	0x00000004
#define EC_DELETE_ATTACHMENTS	0x00000008
#define EC_DELETE_CONTAINER		0x00000010
#define EC_DELETE_HARD_DELETE	0x00000020
#define	EC_DELETE_STORES		0x00000040

/**
 * logon: log on and create a session with provided credentials
 */
int ns__logon(struct soap *soap, char *user, char *pass, char *clientVersion, unsigned int clientCaps, struct xsd__base64Binary sLicenseRequest, unsigned long long ullSessionGroup, char *szClientApp, struct logonResponse *lpsResponse)
{
	ECRESULT	er = erSuccess;
	ECSession	*lpecSession = NULL;
	ECSESSIONID	sessionID;
	GUID		sServerGuid = {0};
    ECLicenseClient *lpLicenseClient = NULL;
    unsigned int ulLicenseResponse;
    unsigned char *lpLicenseResponse = NULL;

	// check username and password
	er = g_lpSessionManager->CreateSession(soap, user, pass, clientVersion, szClientApp, clientCaps, ullSessionGroup, &sessionID, &lpecSession, true);
	if(er != erSuccess){
		er = ZARAFA_E_LOGON_FAILED;
		goto exit;
	}


	lpsResponse->ulSessionId = sessionID;
	lpsResponse->lpszVersion = "0,"PROJECT_VERSION_SERVER_STR;
	lpsResponse->ulCapabilities = ZARAFA_CAP_CRYPT | ZARAFA_CAP_LICENSE_SERVER | ZARAFA_CAP_LOADPROP_ENTRYID;

	if (clientCaps & ZARAFA_CAP_COMPRESSION) {
		// client knows compression, then turn it on
		lpsResponse->ulCapabilities |= ZARAFA_CAP_COMPRESSION;
		// (ECSessionManager::ValidateSession() will do this for all other functions)
		soap_set_imode(soap, SOAP_ENC_ZLIB);	// also autodetected
		soap_set_omode(soap, SOAP_ENC_ZLIB | SOAP_IO_CHUNK);
	}

	if (clientCaps & ZARAFA_CAP_MULTI_SERVER)
		lpsResponse->ulCapabilities |= ZARAFA_CAP_MULTI_SERVER;

    if(sLicenseRequest.__size) {
        lpLicenseClient = new ECLicenseClient(g_lpSessionManager->GetConfig()->GetSetting("license_socket"));
        
        er = lpLicenseClient->Auth(sLicenseRequest.__ptr, sLicenseRequest.__size, &lpLicenseResponse, &ulLicenseResponse);

        // If the license server is not running, report this as no access.
        if(er == ZARAFA_E_NETWORK_ERROR)
            er = ZARAFA_E_NO_ACCESS;
        
        if(er != erSuccess) {
            g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Client requested license but zarafa-licensed could not be contacted");
            goto exit; // Note that Auth() succeeds even if the client request was denied. An error here is a real parsing error for example.
        }
        
        lpsResponse->sLicenseResponse.__size = ulLicenseResponse;
        lpsResponse->sLicenseResponse.__ptr = (unsigned char *)s_memcpy(soap, (const char *)lpLicenseResponse, ulLicenseResponse);
    }

	er = g_lpSessionManager->GetServerGUID(&sServerGuid);
	if (er != erSuccess)
		goto exit;

	lpsResponse->sServerGuid.__ptr = (unsigned char*)s_memcpy(soap, (char*)&sServerGuid, sizeof(sServerGuid));
	lpsResponse->sServerGuid.__size = sizeof(sServerGuid);

    // Only save logon if credentials were supplied by the user; otherwise the logon is probably automated
    if(lpecSession->GetAuthMethod() == ECSession::METHOD_USERPASSWORD || lpecSession->GetAuthMethod() == ECSession::METHOD_SSO)
        SaveLogonTime(lpecSession, true);
	
exit:
	if (lpecSession)
		lpecSession->Unlock();
        
    if(lpLicenseResponse)
        delete [] lpLicenseResponse;
        
    if(lpLicenseClient)
        delete lpLicenseClient;

	lpsResponse->er = er;

	return SOAP_OK;
}

/**
 * logon: log on and create a session with provided credentials
 */
int ns__ssoLogon(struct soap *soap, unsigned long long ulSessionId, char *szUsername, struct xsd__base64Binary *lpInput, char *szClientVersion, unsigned int clientCaps, struct xsd__base64Binary sLicenseRequest, unsigned long long ullSessionGroup, char *szClientApp, struct ssoLogonResponse *lpsResponse)
{
	ECRESULT		er = ZARAFA_E_LOGON_FAILED;
	ECAuthSession	*lpecAuthSession = NULL;
	ECSession		*lpecSession = NULL;
	ECSESSIONID		newSessionID = 0;
	GUID			sServerGuid = {0};
	xsd__base64Binary *lpOutput = NULL;
	char*			lpszEnabled = NULL;
	ECLicenseClient*lpLicenseClient = NULL;

	if (!lpInput || lpInput->__size == 0 || lpInput->__ptr == NULL || !szUsername || !szClientVersion)
		goto exit;

	lpszEnabled = g_lpSessionManager->GetConfig()->GetSetting("enable_sso_ntlmauth");
	if (!(lpszEnabled && stricmp(lpszEnabled, "yes") == 0))
		goto nosso;

	if (ulSessionId == 0) {
		// new auth session
		er = g_lpSessionManager->CreateAuthSession(soap, clientCaps, &newSessionID, &lpecAuthSession, true, true);
		if (er != erSuccess) {
			er = ZARAFA_E_LOGON_FAILED;
			goto exit;
		}
		// when the first validate fails, remove the correct sessionid
		ulSessionId = newSessionID;
	} else {
		er = g_lpSessionManager->ValidateSession(soap, ulSessionId, &lpecAuthSession, true);
		if (er != erSuccess)
			goto exit;
	}

	er = lpecAuthSession->ValidateSSOData(soap, szUsername, lpInput, &lpOutput);
	if (er == ZARAFA_E_SSO_CONTINUE) {
		// continue validation exchange
		lpsResponse->lpOutput = lpOutput;
	} else if (er == erSuccess) {
		// done and logged in
        if(sLicenseRequest.__size) {
            lpLicenseClient = new ECLicenseClient(g_lpSessionManager->GetConfig()->GetSetting("license_socket"));
            
            er = lpLicenseClient->Auth(sLicenseRequest.__ptr, sLicenseRequest.__size, &lpsResponse->sLicenseResponse.__ptr, (unsigned int *)&lpsResponse->sLicenseResponse.__size);
            // If the license server is not running, report this as no access.
            if(er == ZARAFA_E_NETWORK_ERROR)
                er = ZARAFA_E_NO_ACCESS;
            
            if(er != erSuccess) {
                g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Client requested license but zarafa-licensed could not be contacted");
                goto exit; // Note that Auth() succeeds even if the client request was denied. An error here is a real parsing error for example.
            }
        }

		// create ecsession from ecauthsession, and place in session map
		er = g_lpSessionManager->RegisterSession(lpecAuthSession, ullSessionGroup, szClientVersion, szClientApp, &newSessionID, &lpecSession, true);
		if (er != erSuccess) {
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "User authenticated, but failed to create session. Error 0x%08X", er);
			goto exit;
		}


		// delete authsession
		lpecAuthSession->Unlock();
		
		g_lpSessionManager->RemoveSession(ulSessionId);
		lpecAuthSession = NULL;

		// return ecsession number
		lpsResponse->lpOutput = NULL;

	} else {
		// delete authsession
		lpecAuthSession->Unlock();

		g_lpSessionManager->RemoveSession(ulSessionId);
		lpecAuthSession = NULL;

		er = ZARAFA_E_LOGON_FAILED;
		goto exit;
	}

	lpsResponse->ulSessionId = newSessionID;
	lpsResponse->lpszVersion = "0,"PROJECT_VERSION_SERVER_STR;
	lpsResponse->ulCapabilities = ZARAFA_CAP_CRYPT | ZARAFA_CAP_LICENSE_SERVER  | ZARAFA_CAP_LOADPROP_ENTRYID;

	if (clientCaps & ZARAFA_CAP_COMPRESSION) {
		// client knows compression, then turn it on
		lpsResponse->ulCapabilities |= ZARAFA_CAP_COMPRESSION;
		// (ECSessionManager::ValidateSession() will do this for all other functions)
		soap_set_imode(soap, SOAP_ENC_ZLIB);	// also autodetected
		soap_set_omode(soap, SOAP_ENC_ZLIB | SOAP_IO_CHUNK);
	}

	if (clientCaps & ZARAFA_CAP_MULTI_SERVER)
		lpsResponse->ulCapabilities |= ZARAFA_CAP_MULTI_SERVER;

    if(er != ZARAFA_E_SSO_CONTINUE) {
        // Don't reset er to erSuccess on SSO_CONTINUE, we don't need the server guid yet
    	er = g_lpSessionManager->GetServerGUID(&sServerGuid);
    	if (er != erSuccess)
    		goto exit;
    
    	lpsResponse->sServerGuid.__ptr = (unsigned char*)s_memcpy(soap, (char*)&sServerGuid, sizeof(sServerGuid));
    	lpsResponse->sServerGuid.__size = sizeof(sServerGuid);
    }
    
    if(lpecSession && (lpecSession->GetAuthMethod() == ECSession::METHOD_USERPASSWORD || lpecSession->GetAuthMethod() == ECSession::METHOD_SSO))
        SaveLogonTime(lpecSession, true);

exit:
    if (lpecAuthSession)
        lpecAuthSession->Unlock();

	if (lpecSession)
		lpecSession->Unlock();
        
    if (lpLicenseClient)
        delete lpLicenseClient;
        
	if (er == erSuccess)
		g_lpStatsCollector->Increment(SCN_LOGIN_SSO);
	else if (er != ZARAFA_E_SSO_CONTINUE)
		g_lpStatsCollector->Increment(SCN_LOGIN_DENIED);

nosso:
	lpsResponse->er = er;

	return SOAP_OK;
}

/**
 * logoff: invalidate the session and close all notifications and memory held by the session
 */
int ns__logoff(struct soap *soap, unsigned long long ulSessionId, unsigned int *result)
{
	ECRESULT	er = erSuccess;
	ECSession 	*lpecSession = NULL;

	er = g_lpSessionManager->ValidateSession(soap, ulSessionId, &lpecSession, false);
	if(er != erSuccess)
		goto exit;

    if(lpecSession->GetAuthMethod() == ECSession::METHOD_USERPASSWORD || lpecSession->GetAuthMethod() == ECSession::METHOD_SSO)
        SaveLogonTime(lpecSession, false);

    // lpecSession is discarded. It is not locked, so we can do that. We only did the 'validatesession'
    // call to see if the session id existed in the first place, and the request is coming from the correct
    // IP address. Another logoff() call called at the same time may remove the session *here*, in which case the following call
    // will fail. This makes sure people can't terminate each others sessions unless they have the same source IP.

	er = g_lpSessionManager->RemoveSession(ulSessionId);
exit:
    *result = er;

    return SOAP_OK;
}

double timespec2dbl(timespec t) {
    return (double)t.tv_sec + t.tv_nsec/1000000000.0;
}

double GetTimeOfDay();

#define SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname) \
    ECRESULT		er = erSuccess; \
    struct timespec	startTimes = {0}, endTimes = {0};	\
    double			dblStart = GetTimeOfDay(); \
    ECSession		*lpecSession = NULL; \
    unsigned int 	*lpResultVar = &resultvar; \
    clock_gettime(CLOCK_THREAD_CPUTIME_ID, &startTimes); \
    if(soap->user && ((SOAPINFO *)soap->user)->lpCallBack) { \
        ((SOAPINFO *)soap->user)->lpCallBack(pthread_self(), (std::string) "[" + PrettyIP(soap->ip) + "] " + #fname, ((SOAPINFO*)soap->user)->ulCallBackParam); \
    } \
	er = g_lpSessionManager->ValidateSession(soap, ulSessionId, &lpecSession, true);\
	if(er != erSuccess) \
		goto __soapentry_exit; \
	lpecSession->AddBusyState(pthread_self(), #fname);

#define SOAP_ENTRY_FUNCTION_FOOTER \
__soapentry_exit: \
    *lpResultVar = er; \
    clock_gettime(CLOCK_THREAD_CPUTIME_ID, &endTimes); \
    if(lpecSession) { \
    	lpecSession->AddClocks( timespec2dbl(endTimes) - timespec2dbl(startTimes), \
    	                        0, \
							    GetTimeOfDay() - dblStart); \
	lpecSession->RemoveBusyState(pthread_self()); \
        lpecSession->Unlock(); \
    } \
    if(soap->user && ((SOAPINFO *)soap->user)->lpCallBack) { \
        ((SOAPINFO *)soap->user)->lpCallBack(pthread_self(), (std::string) "[" + PrettyIP(soap->ip) + "] " + "Idle", ((SOAPINFO*)soap->user)->ulCallBackParam); \
    } \
    return SOAP_OK;


// This is the variadic macro we would like to use. Unfortunately, the win32 vc++ compiler
// doesn't understand variadic macro's, so we have to split them into the cases below...
#ifdef USE_VARIADIC_MACRO
#define SOAP_ENTRY_START(fname,resultvar,args...) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, ##args) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START1 SOAP_ENTRY_START
#define SOAP_ENTRY_START2 SOAP_ENTRY_START
#define SOAP_ENTRY_START3 SOAP_ENTRY_START
#define SOAP_ENTRY_START4 SOAP_ENTRY_START
#define SOAP_ENTRY_START5 SOAP_ENTRY_START
#define SOAP_ENTRY_START6 SOAP_ENTRY_START
#define SOAP_ENTRY_START7 SOAP_ENTRY_START
#define SOAP_ENTRY_START8 SOAP_ENTRY_START
#define SOAP_ENTRY_START9 SOAP_ENTRY_START
#else

#define SOAP_ENTRY_START1(fname, resultvar, arg1) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START2(fname, resultvar, arg1, arg2) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1, arg2) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START3(fname, resultvar, arg1, arg2, arg3) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1, arg2, arg3) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START4(fname, resultvar, arg1, arg2, arg3, arg4) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1, arg2, arg3, arg4) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START5(fname, resultvar, arg1, arg2, arg3, arg4, arg5) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1, arg2, arg3, arg4, arg5) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START6(fname, resultvar, arg1, arg2, arg3, arg4, arg5, arg6) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1, arg2, arg3, arg4, arg5, arg6) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START7(fname, resultvar, arg1, arg2, arg3, arg4, arg5, arg6, arg7) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1, arg2, arg3, arg4, arg5, arg6, arg7) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START8(fname, resultvar, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#define SOAP_ENTRY_START9(fname, resultvar, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9) \
int ns__##fname(struct soap *soap, unsigned long long ulSessionId, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9) \
{ \
    SOAP_ENTRY_FUNCTION_HEADER(resultvar, fname)

#endif // USE_VARIADIC_MACRO

#define SOAP_ENTRY_END() \
    SOAP_ENTRY_FUNCTION_FOOTER \
}


#define UNUSED __attribute__((unused))

#define ALLOC_DBRESULT() \
	DB_ROW 			UNUSED			lpDBRow = NULL; \
	DB_LENGTHS		UNUSED			lpDBLen = NULL; \
	DB_RESULT		UNUSED			lpDBResult = NULL; \
	std::string		UNUSED			strQuery;

#define USE_DATABASE() \
       ECDatabase*             lpDatabase = NULL; \
       ALLOC_DBRESULT(); \
    \
       lpDatabase = lpecSession->GetDatabase(); \
       if (!lpDatabase) { \
               er = ZARAFA_E_DATABASE_ERROR; \
               goto __soapentry_exit; \
       }

#define FREE_DBRESULT() \
    if(lpDBResult) { \
        lpDatabase->FreeResult(lpDBResult); \
        lpDBResult = NULL; \
    }

#define ROLLBACK_ON_ERROR() \
	if (lpDatabase && FAILED(er)) \
		lpDatabase->Rollback(); \


// Save the current time as the last logon time for the logged-on user of lpecSession
ECRESULT SaveLogonTime(ECSession *lpecSession, bool bLogon) 
{
    ECRESULT er = erSuccess;
    unsigned int ulUserId = 0;
    unsigned int ulStoreId = 0;
    time_t now = time(NULL);
    FILETIME ft;
    unsigned int ulProperty = bLogon ? PR_LAST_LOGON_TIME : PR_LAST_LOGOFF_TIME;
	ECDatabase *lpDatabase = NULL; 
	ALLOC_DBRESULT(); 
    
	lpDatabase = lpecSession->GetDatabase(); 
	if (!lpDatabase) { 
		er = ZARAFA_E_DATABASE_ERROR; 
		goto exit; 
	}
    
    UnixTimeToFileTime(now, &ft);
    
        
    ulUserId = lpecSession->GetSecurity()->GetUserId();
    
    strQuery = "SELECT hierarchy_id FROM stores WHERE stores.user_id=" + stringify(ulUserId);
    
    er = lpDatabase->DoSelect(strQuery, &lpDBResult);
    if(er != erSuccess)
        goto exit;
        
    if(lpDatabase->GetNumRows(lpDBResult) == 0)
        goto exit; // User has no store on this server
        
    lpDBRow = lpDatabase->FetchRow(lpDBResult);
    ulStoreId = atoi(lpDBRow[0]);
    
    strQuery = "REPLACE INTO properties (tag, type, hierarchyid, storeid, val_hi, val_lo) VALUES(" 
                + stringify(PROP_ID(ulProperty)) + ","  + stringify(PROP_TYPE(ulProperty)) + ","
                + stringify(ulStoreId) + "," + stringify(ulStoreId) + "," 
                + stringify(ft.dwHighDateTime) + "," + stringify(ft.dwLowDateTime) + ")";
                
    er = lpDatabase->DoInsert(strQuery);
    if(er != erSuccess)
        goto exit;
    
exit:
    ROLLBACK_ON_ERROR();
    FREE_DBRESULT();

    return er;
}

ECRESULT CreateAttachmentStorage(ECDatabase *lpDatabase, ECAttachmentStorage **lppAttachmentStorage)
{
	return ECAttachmentStorage::CreateAttachmentStorage(lpDatabase, g_lpSessionManager->GetConfig(), lppAttachmentStorage);
}

ECRESULT PurgeSoftDelete(ECSession *lpecSession, unsigned int ulLifetime, unsigned int *lpulMessages, unsigned int *lpulFolders)
{
    ECRESULT 		er = erSuccess;
	ECDatabase*		lpDatabase = NULL;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	std::string		strQuery;
	std::string		strStoreIds;
	FILETIME		ft;
	unsigned int	ulDeleteFlags = 0;
	time_t			ulTime = 0;
	ECListInt		lObjectIds;
	unsigned int	ulFolders = 0, ulMessages = 0;
    
	lpDatabase = lpecSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	GetSystemTimeAsFileTime(&ft);
	FileTimeToUnixTime(ft, &ulTime);

	ulTime -= ulLifetime;

	UnixTimeToFileTime(ulTime, &ft);

	// Select first all stores, find from hierarchy so we also see the deleted stores
	strQuery = "SELECT id FROM hierarchy WHERE parent IS NULL";
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
	{
		if(lpDBRow[0] == NULL)
			continue;
		if(!strStoreIds.empty())
			strStoreIds += ",";

		strStoreIds += lpDBRow[0];
	}
	
	if(lpDBResult){	lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL;}


	// First select softdeleted folders and store
	strQuery = "SELECT h.id FROM hierarchy AS h JOIN properties AS p ON p.storeid IN("+strStoreIds+") AND p.hierarchyid=h.id AND p.tag="+stringify(PROP_ID(PR_DELETED_ON))+" AND p.type="+stringify(PROP_TYPE(PR_DELETED_ON))+" WHERE (h.flags&"+stringify(MSGFLAG_DELETED)+")="+stringify(MSGFLAG_DELETED)+" AND p.val_hi<="+stringify(ft.dwHighDateTime)+" AND (h.type="+stringify(MAPI_FOLDER)+" OR h.type="+stringify(MAPI_STORE)+")";
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	ulFolders = lpDatabase->GetNumRows(lpDBResult);
	if(ulFolders > 0)
	{
		
		while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
		{
			if(lpDBRow == NULL || lpDBRow[0] == NULL)
				continue;

			lObjectIds.push_back(atoui(lpDBRow[0]));
		}

		ulDeleteFlags = EC_DELETE_CONTAINER | EC_DELETE_FOLDERS | EC_DELETE_MESSAGES | EC_DELETE_RECIPIENTS | EC_DELETE_ATTACHMENTS | EC_DELETE_HARD_DELETE;

		er = DeleteObjects(lpecSession, lpDatabase, &lObjectIds, ulDeleteFlags, 0, false, false);
		if(er != erSuccess)
			goto exit;

	}
	if(lpDBResult){	lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL;}

	// Second select softdeleted messages
	strQuery = "SELECT h.id FROM hierarchy AS h JOIN properties AS p ON p.storeid IN("+strStoreIds+") AND p.hierarchyid=h.id AND p.tag="+stringify(PROP_ID(PR_DELETED_ON))+" AND p.type="+stringify(PROP_TYPE(PR_DELETED_ON))+" WHERE (h.flags&"+stringify(MSGFLAG_DELETED)+")="+stringify(MSGFLAG_DELETED)+" AND h.type="+stringify(MAPI_MESSAGE)+" AND p.val_hi<="+stringify(ft.dwHighDateTime);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	ulMessages = lpDatabase->GetNumRows(lpDBResult);
	if(ulMessages > 0)
	{
		// Remove all items
		lObjectIds.clear();

		while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
		{
			if(lpDBRow == NULL || lpDBRow[0] == NULL)
				continue;

			lObjectIds.push_back(atoui(lpDBRow[0]));
		}

		ulDeleteFlags =EC_DELETE_ATTACHMENTS | EC_DELETE_RECIPIENTS | EC_DELETE_CONTAINER | EC_DELETE_MESSAGES | EC_DELETE_HARD_DELETE;

		er = DeleteObjects(lpecSession, lpDatabase, &lObjectIds, ulDeleteFlags, 0, false, false);
		if(er != erSuccess)
			goto exit;
	}
	
	if(lpulFolders)
	    *lpulFolders = ulFolders;
	    
    if(lpulMessages)
        *lpulMessages = ulMessages;

exit:
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}
    
/**
 * getPublicStore: get the root entryid, the store entryid and the store GUID for the public store.
 * FIXME, GUID is duplicate
 */
SOAP_ENTRY_START1(getPublicStore, lpsResponse->er, struct getStoreResponse *lpsResponse)
{
    unsigned int		ulCompanyId = 0;
	std::string			strStoreName;
	std::string			strStoreServer;
	std::string			strServerPath;
	std::string			strCompanyName;
	const std::string	strThisServer = g_lpSessionManager->GetConfig()->GetSetting("server_name", "", "Unknown");
	objectdetails_t		details;
    USE_DATABASE();

	if (lpecSession->GetSessionManager()->IsHostedSupported()) {
		/* Hosted support, Public store owner is company */
		er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
		if (er != erSuccess)
			goto exit;

        er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulCompanyId, &details);
        if(er != erSuccess)
            goto exit;
        
        strStoreServer = details.GetPropString(OB_PROP_S_SERVERNAME);
		strCompanyName = details.GetPropString(OB_PROP_S_COMPANYNAME);
	} else {
		ulCompanyId = ZARAFA_UID_EVERYONE; /* No hosted support, Public store owner is Everyone */

		if (lpecSession->GetSessionManager()->IsDistributedSupported()) {
			/*
			* GetObjectDetailsAndSync will return the group details for EVERYONE when called
			* with ZARAFA_UID_EVERYONE. But we want the pseudo company for that contains the
			* public store.
			*/
			er = lpecSession->GetUserManagement()->GetPublicStoreDetails(&details);
			if (er == ZARAFA_E_NO_SUPPORT) {
				/* Not supported: No MultiServer with this plugin, so we're good */
				strStoreServer = strThisServer;
				er = erSuccess;
			} else if (er == erSuccess)
				strStoreServer = details.GetPropString(OB_PROP_S_SERVERNAME);
			else
				goto exit;
		} else
			strStoreServer = strThisServer;
	}

	if (lpecSession->GetSessionManager()->IsDistributedSupported())
	{
		if (strStoreServer.empty()) {
			if (!strCompanyName.empty())
				g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_ERROR, "Company %s has no home server for its public store.", strCompanyName.c_str());
			else
				g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_ERROR, "Public store has no home server.", strCompanyName.c_str());
				
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
	
		/* Do we own the store? */
		if (strThisServer != strStoreServer)
		{
            er = GetBestServerPath(soap, lpecSession, strStoreServer, &strServerPath);
            if (er != erSuccess)
                goto exit;
            
            lpsResponse->lpszServerPath = s_strcpy(soap, strServerPath.c_str());
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Redirecting request to '%s'", lpsResponse->lpszServerPath);
			g_lpStatsCollector->Increment(SCN_REDIRECT_COUNT, 1);
			er = ZARAFA_E_UNABLE_TO_COMPLETE;
			goto exit;
		}
	}

    // Always return a pseudo URL
    lpsResponse->lpszServerPath = s_strcpy(soap, string("pseudo://" + strStoreServer).c_str());
    
	/*
	 * The public store is stored in the database with the companyid as owner.
	 */
	strQuery =
		"SELECT hierarchy.id, stores.guid, stores.hierarchy_id "
		"FROM stores "
		"JOIN hierarchy on stores.hierarchy_id=hierarchy.parent "
		"WHERE stores.user_id = " + stringify(ulCompanyId);

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) == 0) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if( lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL ||
		lpDBLen == NULL || lpDBLen[1] == 0)
	{
		er = ZARAFA_E_DATABASE_ERROR; // this should never happen
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(atoui(lpDBRow[0]), soap, &lpsResponse->sRootId);
	if(er != erSuccess)
	    goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(atoui(lpDBRow[2]), soap, &lpsResponse->sStoreId);
	if(er != erSuccess)
	    goto exit;

	lpsResponse->guid.__size= lpDBLen[1];
	lpsResponse->guid.__ptr = s_alloc<unsigned char>(soap, lpDBLen[1]);

	memcpy(lpsResponse->guid.__ptr, lpDBRow[1], lpDBLen[1]);

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

/**
 * getStore: get the root entryid, the store entryid and the GUID of a store specified with lpsEntryId
 * FIXME: output store entryid equals input store entryid ?
 * FIXME: output GUID is also in entryid
 */
SOAP_ENTRY_START2(getStore, lpsResponse->er, entryId* lpsEntryId, struct getStoreResponse *lpsResponse)
{
	unsigned int	ulStoreId = 0;
	unsigned int	ulUserId = 0;
	objectdetails_t	sUserDetails;
	string			strServerName;
	string			strServerPath;
    USE_DATABASE();

	if (!lpsEntryId) {
		ulUserId = lpecSession->GetSecurity()->GetUserId();

        // Check if the store should be available on this server
		if (lpecSession->GetSessionManager()->IsDistributedSupported() && 
            !lpecSession->GetUserManagement()->IsLocalItem(ulUserId)) {
			er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulUserId, &sUserDetails);
			if (er != erSuccess)
				goto exit;
            strServerName = sUserDetails.GetPropString(OB_PROP_S_SERVERNAME);
            if (strServerName.empty()) {
                er = ZARAFA_E_NOT_FOUND;
                goto exit;
            }
            if (strServerName.compare(g_lpSessionManager->GetConfig()->GetSetting("server_name")) != 0)  {
                er = GetBestServerPath(soap, lpecSession, strServerName, &strServerPath);
                if (er != erSuccess)
                    goto exit;
                
                lpsResponse->lpszServerPath = s_strcpy(soap, strServerPath.c_str());
                g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Redirecting request to '%s'", lpsResponse->lpszServerPath);
                g_lpStatsCollector->Increment(SCN_REDIRECT_COUNT, 1);
                er = ZARAFA_E_UNABLE_TO_COMPLETE;
                goto exit;
            }
		}
	}
    
    // If strServerName is empty, we're not running in distributed mode or we're dealing
    // with a local account. Just use the name from the configuration.
    if (strServerName.empty())
        strServerName = g_lpSessionManager->GetConfig()->GetSetting("server_name" ,"", "Unknown");
    
    // Always return a pseudo URL
    lpsResponse->lpszServerPath = s_strcpy(soap, string("pseudo://" + strServerName).c_str());

	strQuery = "SELECT hierarchy.id, stores.guid, stores.hierarchy_id "
	           "FROM stores join hierarchy on stores.hierarchy_id=hierarchy.parent ";

	if(lpsEntryId) {
		er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(lpsEntryId, &ulStoreId);
		if(er != erSuccess)
			goto exit;

		strQuery += "WHERE stores.hierarchy_id=" + stringify(ulStoreId) + " LIMIT 1";// FIXME: mysql query
	}else {
		strQuery += "WHERE stores.user_id=" + stringify(ulUserId);
	}

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) == 0) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if( lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL ||
		lpDBLen == NULL || lpDBLen[1] == 0 )
	{
		er = ZARAFA_E_DATABASE_ERROR; // this should never happen
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(atoui(lpDBRow[0]), soap, &lpsResponse->sRootId);
	if(er != erSuccess)
	    goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(atoui(lpDBRow[2]), soap, &lpsResponse->sStoreId);
	if(er != erSuccess)
	    goto exit;

	lpsResponse->guid.__size= lpDBLen[1];
	lpsResponse->guid.__ptr = s_alloc<unsigned char>(soap, lpDBLen[1]);

	memcpy(lpsResponse->guid.__ptr, lpDBRow[1], lpDBLen[1]);

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

/**
 * getStoreName: get the PR_DISPLAY_NAME of the store specified in sEntryId
 */
SOAP_ENTRY_START2(getStoreName, lpsResponse->er, entryId sEntryId, struct getStoreNameResponse *lpsResponse)
{
	unsigned int	ulObjId = 0;
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if(er != erSuccess)
	    goto exit;

	er = ECGenProps::GetStoreName(soap, lpecSession, ulObjId, ulObjId, &lpsResponse->lpszStoreName);
	if(er != erSuccess)
		goto exit;

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

/**
 * getStoreByUser: return the store ID for a specific user
 * FIXME: this is only called when doing zarafa-admin --delete-store <username>. There should be a better way
 * to do this.
 */
SOAP_ENTRY_START3(getStoreByUser, lpsResponse->er, unsigned int ulUserId, entryId sUserId, struct getStoreByUserResponse *lpsResponse)
{
	bool bHasLocalStore = false;

    USE_DATABASE();

    er = GetLocalId(lpecSession, sUserId, ulUserId, &ulUserId, NULL, &bHasLocalStore);
    if (er != erSuccess)
    	goto exit;

	if (!bHasLocalStore) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	strQuery = "SELECT id FROM stores WHERE stores.user_id=" + stringify(ulUserId);

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) == 0) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if( lpDBRow == NULL || lpDBRow[0] == NULL) {
		er = ZARAFA_E_DATABASE_ERROR; // this should never happen
		goto exit;
	}

	lpsResponse->ulStoreId = atoui(lpDBRow[0]);

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

typedef struct {
    DynamicPropValArray *lpPropVals;
    DynamicPropTagArray *lpPropTags;
} CHILDPROPS;

// Prepares child property data. This can be passed to ReadProps(). This allows the properties of child objects of object ulObjId to be
// retrieved with far less SQL queries, since this function bulk-receives the data. You may pass EITHER ulObjId OR ulParentId to retrieve an object itself, or
// children of an object.
ECRESULT PrepareReadProps(struct soap *soap, ECSession *lpecSession, unsigned int ulObjId, unsigned int ulParentId, std::map<unsigned int, CHILDPROPS> *lpChildProps)
{
    ECRESULT er = erSuccess;
	std::map<unsigned int, CHILDPROPS>::iterator iterChild;
	unsigned int ulSize;
	struct propVal sPropVal;
    unsigned int ulChildId;
	USE_DATABASE();

	if(ulObjId == 0 && ulParentId == 0) {
	    er = ZARAFA_E_INVALID_PARAMETER;
	    goto exit;
    }

	if(ulObjId)
        strQuery = "SELECT " + (std::string)PROPCOLORDER + ",hierarchyid FROM properties WHERE hierarchyid="+stringify(ulObjId);
    else
        strQuery = "SELECT " + (std::string)PROPCOLORDER + ",hierarchy.id FROM properties JOIN hierarchy ON hierarchy.id=properties.hierarchyid WHERE parent="+stringify(ulParentId);

    er = lpDatabase->DoSelect(strQuery, &lpDBResult);
    if(er != erSuccess)
        goto exit;

    while((lpDBRow = lpDatabase->FetchRow(lpDBResult)) != NULL) {
        unsigned int ulPropTag;
        
        lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

        if(lpDBLen == NULL) {
            er = ZARAFA_E_DATABASE_ERROR; // this should never happen
            goto exit;
        }

        ulPropTag = PROP_TAG(atoi(lpDBRow[FIELD_NR_TYPE]),atoi(lpDBRow[FIELD_NR_TAG]));
        ulChildId = atoui(lpDBRow[FIELD_NR_MAX]);

        iterChild = lpChildProps->find(ulChildId);
        
        if(iterChild == lpChildProps->end()) {
            CHILDPROPS sChild;
            
            sChild.lpPropTags = new DynamicPropTagArray(soap);
            sChild.lpPropVals = new DynamicPropValArray(soap, 20);
            
            // First property for this child
            lpChildProps->insert(std::pair<unsigned int, CHILDPROPS>(ulChildId, sChild));
            iterChild = lpChildProps->find(ulChildId);
        }
        
        er = iterChild->second.lpPropTags->AddPropTag(ulPropTag);
        if(er != erSuccess)
            goto exit;

        er = GetPropSize(lpDBRow, lpDBLen, &ulSize);

        if(er == erSuccess && ulSize < MAX_PROP_SIZE) {
            // the size of this property is small enough to send in the initial loading sequence
            
            er = CopyDatabasePropValToSOAPPropVal(soap, lpDBRow, lpDBLen, &sPropVal);
            if(er != erSuccess)
                continue;
                
            iterChild->second.lpPropVals->AddPropVal(sPropVal);
        }
    }

    FREE_DBRESULT();

    if(ulObjId)
        strQuery = "SELECT " + (std::string)MVPROPCOLORDER + ", hierarchyid FROM mvproperties WHERE hierarchyid="+stringify(ulObjId)+" GROUP BY hierarchyid, tag";
    else
        strQuery = "SELECT " + (std::string)MVPROPCOLORDER + ", hierarchyid FROM mvproperties JOIN hierarchy ON hierarchy.id=mvproperties.hierarchyid WHERE hierarchy.parent="+stringify(ulParentId)+" GROUP BY hierarchyid, tag";

    er = lpDatabase->DoSelect(strQuery, &lpDBResult);
    if(er != erSuccess)
        goto exit;

    // Do MV props
    while((lpDBRow = lpDatabase->FetchRow(lpDBResult)) != NULL) {
        lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

        if(lpDBLen == NULL) {
            er = ZARAFA_E_DATABASE_ERROR; // this should never happen
            goto exit;
        }

        ulChildId = atoui(lpDBRow[FIELD_NR_MAX]);

        iterChild = lpChildProps->find(ulChildId);
        
        if(iterChild == lpChildProps->end()) {
            CHILDPROPS sChild;
            
            sChild.lpPropTags = new DynamicPropTagArray(soap);
            sChild.lpPropVals = new DynamicPropValArray(soap, 20);
            
            // First property for this child
            lpChildProps->insert(std::pair<unsigned int, CHILDPROPS>(ulChildId, sChild));
            iterChild = lpChildProps->find(ulChildId);
        }
        
        er = iterChild->second.lpPropTags->AddPropTag( PROP_TAG(atoi(lpDBRow[FIELD_NR_TYPE]),atoi(lpDBRow[FIELD_NR_TAG])));
        if(er != erSuccess)
            continue;

        er = CopyDatabasePropValToSOAPPropVal(soap, lpDBRow, lpDBLen, &sPropVal);
        if(er != erSuccess)
            continue;
            
        iterChild->second.lpPropVals->AddPropVal(sPropVal);
    }

exit:
	FREE_DBRESULT();
__soapentry_exit:
	return er;
}

ECRESULT FreeChildProps(std::map<unsigned int, CHILDPROPS> *lpChildProps)
{
    std::map<unsigned int, CHILDPROPS>::iterator iterChild;
    
    for(iterChild = lpChildProps->begin(); iterChild != lpChildProps->end(); iterChild++)
    {
        // We need to delete the DynamicProp* objects themselves. This leaves the data that was
        // retrieved from those objects (via GetPropTagArray()) untouched though, since that is allocated
        // via soap_malloc and will be auto-freed when the SOAP call is over.
        delete iterChild->second.lpPropVals;
        delete iterChild->second.lpPropTags;
    }
    
    lpChildProps->clear();
    
    return erSuccess;
}

ECRESULT ReadProps(struct soap *soap, ECSession *lpecSession, unsigned int ulObjId, unsigned ulObjType, unsigned int ulObjTypeParent, CHILDPROPS sChildProps, struct propTagArray *lpsPropTag, struct propValArray *lpsPropVal)
{
	ECRESULT er = erSuccess;

	quotadetails_t	sDetails;
	unsigned int	ulCompanyId = 0;
	unsigned int	ulStoreOwner = 0;

	ECAttachmentStorage *lpAttachmentStorage = NULL;

	// Folder counters
	unsigned int ulContentCount = 0;
	unsigned int ulUnreadCount = 0;
	unsigned int ulAssocCount = 0;
	unsigned int ulSubFolderCount = 0;
	unsigned int ulDeleteMsgCount = 0;
	unsigned int ulDeleteFolderCount = 0;
	unsigned int ulDeleteAssocMsgCount = 0;

	struct propVal sPropVal;

	DB_RESULT		lpDBResultMV = NULL;
	USE_DATABASE();
        
	if(ulObjType == MAPI_STORE) //fimxe: except public stores
	{
		if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_USER_NAME, ulObjId, 0, ulObjId, &sPropVal) == erSuccess) {
		    sChildProps.lpPropTags->AddPropTag(PR_USER_NAME);
		    sChildProps.lpPropVals->AddPropVal(sPropVal);
        }

		if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_USER_ENTRYID, ulObjId, 0, ulObjId, &sPropVal) == erSuccess) {
		    sChildProps.lpPropTags->AddPropTag(PR_USER_ENTRYID);
		    sChildProps.lpPropVals->AddPropVal(sPropVal);
        }

        er = lpecSession->GetSecurity()->GetStoreOwner(ulObjId, &ulStoreOwner);
        if (er != erSuccess)
            goto exit;

		// Quota information
		if (lpecSession->GetSecurity()->GetUserQuota(ulStoreOwner, &sDetails) == erSuccess)
		{
			// PR_QUOTA_WARNING_THRESHOLD
			sPropVal.ulPropTag = PR_QUOTA_WARNING_THRESHOLD;
			sPropVal.__union = SOAP_UNION_propValData_ul;
			sPropVal.Value.ul = (unsigned long)(sDetails.llWarnSize / 1024);

			sChildProps.lpPropTags->AddPropTag(PR_QUOTA_WARNING_THRESHOLD);
			sChildProps.lpPropVals->AddPropVal(sPropVal);
			    
			// PR_QUOTA_SEND_THRESHOLD
			sPropVal.ulPropTag = PR_QUOTA_SEND_THRESHOLD;
			sPropVal.__union = SOAP_UNION_propValData_ul;
			sPropVal.Value.ul = (unsigned long)(sDetails.llSoftSize / 1024);

			sChildProps.lpPropTags->AddPropTag(PR_QUOTA_SEND_THRESHOLD);
            sChildProps.lpPropVals->AddPropVal(sPropVal);

			// PR_QUOTA_RECEIVE_THRESHOLD
			sPropVal.ulPropTag = PR_QUOTA_RECEIVE_THRESHOLD;
			sPropVal.__union = SOAP_UNION_propValData_ul;
			sPropVal.Value.ul = (unsigned long)(sDetails.llHardSize  / 1024);

			sChildProps.lpPropTags->AddPropTag(PR_QUOTA_RECEIVE_THRESHOLD);
            sChildProps.lpPropVals->AddPropVal(sPropVal);
		}

		if (lpecSession->GetCapabilities() & ZARAFA_CAP_MAILBOX_OWNER) {
			// get the companyid to which the logged in user belongs to.
			er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
			if (er != erSuccess)
				goto exit;

			// 5.0 client knows how to handle the PR_MAILBOX_OWNER_* properties
			if(ulStoreOwner != ZARAFA_UID_EVERYONE && ulStoreOwner != ulCompanyId )	{

				if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_MAILBOX_OWNER_NAME, ulObjId, 0, ulObjId, &sPropVal) == erSuccess) {
				    sChildProps.lpPropTags->AddPropTag(PR_MAILBOX_OWNER_NAME);
                    sChildProps.lpPropVals->AddPropVal(sPropVal);
				}

				// Add PR_MAILBOX_OWNER_ENTRYID
				if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_MAILBOX_OWNER_ENTRYID, ulObjId, 0, ulObjId, &sPropVal) == erSuccess) {
					sChildProps.lpPropTags->AddPropTag(PR_MAILBOX_OWNER_ENTRYID);
                    sChildProps.lpPropVals->AddPropVal(sPropVal);
				}
			}
		}

		// Add PR_DISPLAY_NAME
		if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_DISPLAY_NAME, ulObjId, 0, ulObjId, &sPropVal) == erSuccess) {
		    sChildProps.lpPropTags->AddPropTag(PR_DISPLAY_NAME);
		    sChildProps.lpPropVals->AddPropVal(sPropVal);
		}

		if(ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_MAPPING_SIGNATURE, ulObjId, 0, 0, &sPropVal) == erSuccess) {
		    sChildProps.lpPropTags->AddPropTag(PR_MAPPING_SIGNATURE);
		    sChildProps.lpPropVals->AddPropVal(sPropVal);
		}
	}

	if (ulObjType == MAPI_FOLDER)
	{
		er = GetFolderCount(lpecSession, ulObjId, &ulContentCount, &ulUnreadCount, &ulAssocCount,
							&ulSubFolderCount, &ulDeleteMsgCount, &ulDeleteFolderCount, &ulDeleteAssocMsgCount);
		if(er != erSuccess)
			goto exit;

		// PR_CONTENT_COUNT
		sPropVal.ulPropTag = PR_CONTENT_COUNT;
		sPropVal.__union = SOAP_UNION_propValData_ul;
		sPropVal.Value.ul = ulContentCount;

		sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		sChildProps.lpPropVals->AddPropVal(sPropVal);

		// PR_CONTENT_UNREAD
		sPropVal.ulPropTag = PR_CONTENT_UNREAD;
		sPropVal.__union = SOAP_UNION_propValData_ul;
		sPropVal.Value.ul = ulUnreadCount;

		sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		sChildProps.lpPropVals->AddPropVal(sPropVal);

		// PR_ASSOC_CONTENT_COUNT
		sPropVal.ulPropTag = PR_ASSOC_CONTENT_COUNT;
		sPropVal.__union = SOAP_UNION_propValData_ul;
		sPropVal.Value.ul = ulAssocCount;

		sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		sChildProps.lpPropVals->AddPropVal(sPropVal);

		// PR_SUBFOLDERS
		sPropVal.ulPropTag = PR_SUBFOLDERS;
		sPropVal.__union = SOAP_UNION_propValData_b;
		sPropVal.Value.b = (ulSubFolderCount > 0);

		sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		sChildProps.lpPropVals->AddPropVal(sPropVal);

		// PR_FOLDER_CHILD_COUNT
		sPropVal.ulPropTag = PR_FOLDER_CHILD_COUNT;
		sPropVal.__union = SOAP_UNION_propValData_ul;
		sPropVal.Value.ul = ulSubFolderCount;

		sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		sChildProps.lpPropVals->AddPropVal(sPropVal);

		// PR_DELETED_MSG_COUNT
		sPropVal.ulPropTag = PR_DELETED_MSG_COUNT;
		sPropVal.__union = SOAP_UNION_propValData_ul;
		sPropVal.Value.ul = ulDeleteMsgCount;

		sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		sChildProps.lpPropVals->AddPropVal(sPropVal);

		// PR_DELETED_FOLDER_COUNT
		sPropVal.ulPropTag = PR_DELETED_FOLDER_COUNT;
		sPropVal.__union = SOAP_UNION_propValData_ul;
		sPropVal.Value.ul = ulDeleteFolderCount;

		sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		sChildProps.lpPropVals->AddPropVal(sPropVal);

		// PR_DELETED_ASSOC_MSG_COUNT
		sPropVal.ulPropTag = PR_DELETED_ASSOC_MSG_COUNT;
		sPropVal.__union = SOAP_UNION_propValData_ul;
		sPropVal.Value.ul = ulDeleteAssocMsgCount;

		sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		sChildProps.lpPropVals->AddPropVal(sPropVal);
	}

	//PR_PARENT_SOURCE_KEY for folders and messages
	if(ulObjType == MAPI_FOLDER || (ulObjType == MAPI_MESSAGE && ulObjTypeParent == MAPI_FOLDER))
	{
		if(ECGenProps::GetPropComputed(soap, lpecSession, PR_PARENT_SOURCE_KEY, ulObjId, 0, 0, ulObjType, 0, 0, NULL, &sPropVal) == erSuccess)
		{
		    sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		    sChildProps.lpPropVals->AddPropVal(sPropVal);
		}

		if(ECGenProps::GetPropComputed(soap, lpecSession, PR_SOURCE_KEY, ulObjId, 0, 0, ulObjType, 0, 0, NULL, &sPropVal) == erSuccess)
		{
		    sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		    sChildProps.lpPropVals->AddPropVal(sPropVal);
		}
	}

	if (ulObjType == MAPI_ATTACH) {
		er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
		if (er != erSuccess)
			goto exit;

		if (lpAttachmentStorage->ExistAttachment(ulObjId))
		    sChildProps.lpPropTags->AddPropTag(PR_ATTACH_DATA_BIN);
	}

	if (ulObjType == MAPI_MAILUSER || ulObjType == MAPI_DISTLIST) {
		if(ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_INSTANCE_KEY, ulObjId, 0, 0, &sPropVal) == erSuccess) {
		    sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
		    sChildProps.lpPropVals->AddPropVal(sPropVal);
		}
	}

	// Set the PR_RECORD_KEY
	if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_RECORD_KEY, ulObjId, 0, 0, &sPropVal) == erSuccess) {
        sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
        sChildProps.lpPropVals->AddPropVal(sPropVal);
	}

	if (ulObjType == MAPI_FOLDER || ulObjType == MAPI_STORE || ulObjType == MAPI_MESSAGE) {
        // Get PARENT_ENTRYID, PR_MESSAGE_FLAGS|PR_FOLDER_TYPE, PR_LAST_MODIFICATION_TIME, PR_CREATION_TIME
        strQuery = "SELECT hierarchy.parent,hierarchy.flags,hierarchy.type,UNIX_TIMESTAMP(hierarchy.modtime),UNIX_TIMESTAMP(hierarchy.createtime) FROM hierarchy WHERE hierarchy.id="+stringify(ulObjId);
        er = lpDatabase->DoSelect(strQuery, &lpDBResult);
        if(er != erSuccess)
            goto exit;

        if(lpDatabase->GetNumRows(lpDBResult) > 0)
        {
            lpDBRow = lpDatabase->FetchRow(lpDBResult);

            if(lpDBRow && lpDBRow[0]) {

                if(ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_PARENT_ENTRYID, ulObjId, 0, 0, &sPropVal) == erSuccess)
                {
                    sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
                    sChildProps.lpPropVals->AddPropVal(sPropVal);
                }
            }

            if(lpDBRow && lpDBRow[1] && lpDBRow[2] && atoi(lpDBRow[2]) == 5) {
                DB_RESULT lpAttachRes;
                DB_ROW lpAttachRow;

                sChildProps.lpPropTags->AddPropTag(PR_MESSAGE_FLAGS);			

                sPropVal.ulPropTag = PR_MESSAGE_FLAGS;
                sPropVal.__union = SOAP_UNION_propValData_ul;
                sPropVal.Value.ul = atoi(lpDBRow[1]);
                
                // Find attachments
                strQuery = "SELECT id FROM hierarchy WHERE parent="+stringify(ulObjId)+" AND type=7 LIMIT 1";
                er = lpDatabase->DoSelect(strQuery, &lpAttachRes);
                if(er != erSuccess)
                    goto exit;

                lpAttachRow = lpDatabase->FetchRow(lpAttachRes);

                if(lpAttachRow && lpAttachRow[0] != NULL) {
                    // At least one attachment in message
                    sPropVal.Value.ul |= MSGFLAG_HASATTACH;
                }

                sChildProps.lpPropVals->AddPropVal(sPropVal);

                lpDatabase->FreeResult(lpAttachRes);
                lpAttachRes = NULL;
            }
            else if(lpDBRow && lpDBRow[1] && lpDBRow[2] && atoi(lpDBRow[2]) == 3) {
                // PR_FOLDER_TYPE
                sChildProps.lpPropTags->AddPropTag(PR_FOLDER_TYPE);

                sPropVal.ulPropTag = PR_FOLDER_TYPE;
                sPropVal.__union = SOAP_UNION_propValData_ul;
                sPropVal.Value.ul = atoui(lpDBRow[1])&(FOLDER_GENERIC|FOLDER_ROOT|FOLDER_SEARCH);

                sChildProps.lpPropVals->AddPropVal(sPropVal);

                // PR_RIGHTS
                if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_RIGHTS, ulObjId, 0, 0, &sPropVal) == erSuccess) {
                    sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
                    sChildProps.lpPropVals->AddPropVal(sPropVal);
                }
            }

            // Set the flags PR_ACCESS and PR_ACCESS_LEVEL
            if(lpDBRow && lpDBRow[0] && lpDBRow[2] && (atoi(lpDBRow[2]) == 3 || atoi(lpDBRow[2]) == 5) && lpDBRow[1])
            {
                if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_ACCESS, ulObjId, 0, 0, &sPropVal) == erSuccess) {
                    sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
                    sChildProps.lpPropVals->AddPropVal(sPropVal);
                }
            }

            if(lpDBRow && lpDBRow[2] && lpDBRow[1])
            {
                if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_ACCESS_LEVEL, ulObjId, 0, 0, &sPropVal) == erSuccess) {
                    sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
                    sChildProps.lpPropVals->AddPropVal(sPropVal);
                }
            }

            // Set PR_LAST_MODIFICATION_TIME and PR_CREATION_TIME
            if(lpDBRow && lpDBRow[3])
            {
                FILETIME ftMod;

                UnixTimeToFileTime(atoi(lpDBRow[3]), &ftMod);

                sPropVal.__union = SOAP_UNION_propValData_hilo;
                sPropVal.ulPropTag = PR_LAST_MODIFICATION_TIME;
                sPropVal.Value.hilo = s_alloc<hiloLong>(soap);

                sPropVal.Value.hilo->hi = ftMod.dwHighDateTime;
                sPropVal.Value.hilo->lo = ftMod.dwLowDateTime;

                sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
                sChildProps.lpPropVals->AddPropVal(sPropVal);
            }

            if(lpDBRow && lpDBRow[4])
            {
                FILETIME ftCreate;

                UnixTimeToFileTime(atoi(lpDBRow[4]), &ftCreate);

                sPropVal.__union = SOAP_UNION_propValData_hilo;
                sPropVal.ulPropTag = PR_CREATION_TIME;
                sPropVal.Value.hilo = s_alloc<hiloLong>(soap);

                sPropVal.Value.hilo->hi = ftCreate.dwHighDateTime;
                sPropVal.Value.hilo->lo = ftCreate.dwLowDateTime;

                sChildProps.lpPropTags->AddPropTag(sPropVal.ulPropTag);
                sChildProps.lpPropVals->AddPropVal(sPropVal);
            }
        }
    }
    
	er = sChildProps.lpPropTags->GetPropTagArray(lpsPropTag);
	if(er != erSuccess)
	    goto exit;
	    
	er = sChildProps.lpPropVals->GetPropValArray(lpsPropVal);
	if(er != erSuccess)
	    goto exit;


exit:
	FREE_DBRESULT();

	if (lpAttachmentStorage)
		delete lpAttachmentStorage;

	if (lpDBResultMV)
		lpDatabase->FreeResult(lpDBResultMV);

__soapentry_exit:
	return er;
}

/**
 * loadProp: Reads a single, large property from the database. No size limit.
 *   this can also be a complete attachment
 */
SOAP_ENTRY_START4(loadProp, lpsResponse->er, entryId sEntryId, unsigned int ulObjId, unsigned int ulPropTag, struct loadPropResponse *lpsResponse)
{
	unsigned int	ulStoreId;
	ECAttachmentStorage *lpAttachmentStorage = NULL;
	USE_DATABASE();

	if(ulObjId == 0) {
        er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
        if (er != erSuccess)
            goto exit;
            
    }
    
	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityRead);
	if(er != erSuccess)
		goto exit;

	er = lpDatabase->Begin();
	if(er != erSuccess)
		goto exit;

	if (ulPropTag != PR_ATTACH_DATA_BIN) {
		// Get store
		er = g_lpSessionManager->GetCacheManager()->GetStore(ulObjId, &ulStoreId, NULL);
		if(er != erSuccess)
			goto exit;

		if((ulPropTag&MV_FLAG) == MV_FLAG)
			strQuery = "SELECT " + (std::string)MVPROPCOLORDER + " FROM mvproperties WHERE hierarchyid="+stringify(ulObjId)+ " AND tag = " + stringify(PROP_ID(ulPropTag))+" AND storeid = " + stringify(ulStoreId) + " GROUP BY hierarchyid, tag";
		else
			strQuery = "SELECT " PROPCOLORDER " FROM properties WHERE hierarchyid = " + stringify(ulObjId) + " AND tag = " + stringify(PROP_ID(ulPropTag));
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		if(lpDatabase->GetNumRows(lpDBResult) != 1) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

		if (lpDBRow == NULL || lpDBLen == NULL)
		{
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		lpsResponse->lpPropVal = s_alloc<propVal>(soap);
		memset(lpsResponse->lpPropVal,0,sizeof(struct propVal));

		er = CopyDatabasePropValToSOAPPropVal(soap, lpDBRow, lpDBLen, lpsResponse->lpPropVal);
	} else {

		lpsResponse->lpPropVal = s_alloc<propVal>(soap);
		memset(lpsResponse->lpPropVal,0,sizeof(struct propVal));

		lpsResponse->lpPropVal->ulPropTag = ulPropTag;
		lpsResponse->lpPropVal->__union = SOAP_UNION_propValData_bin;
		lpsResponse->lpPropVal->Value.bin = s_alloc<struct xsd__base64Binary>(soap);
		lpsResponse->lpPropVal->Value.bin->__size = 0;

		er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
		if (er != erSuccess)
			goto exit;

		er = lpAttachmentStorage->LoadAttachment(soap, ulObjId, &lpsResponse->lpPropVal->Value.bin->__size, &lpsResponse->lpPropVal->Value.bin->__ptr);
		if (er != erSuccess)
			goto exit;
	}

	er = lpDatabase->Commit();
	if(er != erSuccess)
		goto exit;

exit:
	if (lpAttachmentStorage)
		delete lpAttachmentStorage;

	ROLLBACK_ON_ERROR();

    FREE_DBRESULT();
}
SOAP_ENTRY_END()

//TODO: flag to get size of normal folder or deleted folders
ECRESULT GetFolderSize(ECDatabase* lpDatabase, unsigned int ulStoreId, unsigned int ulFolderId, long long* lpllFolderSize)
{

	ECRESULT		er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	long long		llSize = 0;
	long long		llSubSize = 0;
	std::string		strQuery;

	// sum size of all messages in a folder
	strQuery = "SELECT SUM(p.val_ulong) FROM hierarchy AS h JOIN properties AS p ON p.hierarchyid=h.id AND p.tag="+stringify(PROP_ID(PR_MESSAGE_SIZE))+" AND p.type="+stringify(PROP_TYPE(PR_MESSAGE_SIZE))+" AND p.storeid=" + stringify(ulStoreId) + " WHERE h.parent="+stringify(ulFolderId)+" AND h.type="+stringify(MAPI_MESSAGE);
	// except the deleted items!
	strQuery += " AND h.flags & " + stringify(MSGFLAG_DELETED)+ "=0";

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if(lpDBRow == NULL || lpDBRow[0] == NULL)
		llSize = 0;
	else
		llSize = _atoi64(lpDBRow[0]);

	// Free results
	if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

	// Get the subfolders
	strQuery = "SELECT id FROM hierarchy WHERE parent=" + stringify(ulFolderId) + " AND type="+stringify(MAPI_FOLDER);
	// except the deleted items!
	strQuery += " AND flags & " + stringify(MSGFLAG_DELETED)+ "=0";

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) > 0)
	{
		// Walk through the folder list
		while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
		{
			if(lpDBRow[0] == NULL)
				continue; //Skip item

			er = GetFolderSize(lpDatabase, atoi(lpDBRow[0]), ulStoreId, &llSubSize);
			if(er != erSuccess)
				goto exit;

			llSize += llSubSize;
		}
	}

	// Free results
	if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

	*lpllFolderSize = llSize;

exit:
	// Free results
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT CalculateObjectSize(ECDatabase* lpDatabase, unsigned int objid, unsigned int ulObjType, unsigned int ulStoreId, unsigned int* lpulSize)
{
	ECRESULT		er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	ULONG			ulAttachSize = 0;
	unsigned int	ulSize = 0;
	std::string		strQuery;
	ECAttachmentStorage *lpAttachmentStorage = NULL;

	*lpulSize = 0;
	//	strQuery = "SELECT SUM(16 + IF(val_ulong, 4, 0)+IF(val_double||val_longint||val_hi||val_lo, 8, 0)+ LENGTH(IFNULL(val_string, 0))+length(IFNULL(val_binary, 0))) FROM properties WHERE hierarchyid=" + stringify(objid);

	// SQLite doesn't support IF-type statements, so we're now using a slightly simpler construct ..
	strQuery = "SELECT (SELECT SUM(20 + LENGTH(IFNULL(val_string, 0))+length(IFNULL(val_binary, 0))) FROM properties WHERE hierarchyid=" + stringify(objid) + ") + IFNULL( (SELECT SUM(LENGTH(lob.val_binary)) FROM `lob` JOIN `singleinstances` ON singleinstances.instanceid = lob.instanceid WHERE singleinstances.hierarchyid=" + stringify(objid) + "), 0)";
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if(lpDBRow == NULL || lpDBRow[0] == NULL)
		ulSize = 0;
	else
		ulSize = atoui(lpDBRow[0])+ 28;// + hierarchy size

	er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
	if (er != erSuccess)
		goto exit;

	if (lpAttachmentStorage->IsFileStorage()) {
		er = lpAttachmentStorage->GetSize(objid, &ulAttachSize);
		if (er != erSuccess)
			goto exit;

		ulSize += ulAttachSize;
	}

	// Free results
	if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

	// Calculate also mv-props
	strQuery = "SELECT SUM(20 + LENGTH(IFNULL(val_string, 0))+length(IFNULL(val_binary, 0))) FROM mvproperties WHERE hierarchyid=" + stringify(objid);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if(lpDBRow != NULL && lpDBRow[0] != NULL)
		ulSize += atoui(lpDBRow[0]); // Add the size

	// Free results
	if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

	// Get parent sizes
	strQuery = "SELECT SUM(IFNULL(p.val_ulong, 0)) FROM hierarchy as h JOIN properties AS p ON hierarchyid=h.id WHERE h.parent=" + stringify(objid)+ " AND ( (p.tag="+stringify(PROP_ID(PR_MESSAGE_SIZE))+" AND p.type="+stringify(PROP_TYPE(PR_MESSAGE_SIZE)) + ") || (p.tag="+stringify(PROP_ID(PR_ATTACH_SIZE))+" AND p.type="+stringify(PROP_TYPE(PR_ATTACH_SIZE))+ ") AND p.storeid=" + stringify(ulStoreId) + ")";
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if(lpDBRow != NULL && lpDBRow[0] != NULL)
		ulSize += atoui(lpDBRow[0]); // Add the size

	// Free results
	if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

	*lpulSize = ulSize;

exit:
	if (lpAttachmentStorage)
		delete lpAttachmentStorage;

	// Free results
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

ECRESULT UpdateStoreIDRecursive(ECDatabase *lpDatabase, unsigned int ulFolderId, unsigned int ulStoreId, bool bUpdateRoot)
{
    ECRESULT er = erSuccess;
    DB_RESULT lpResult = NULL;
    DB_ROW lpRow = NULL;
    std::string strQuery;
    std::list<unsigned int> lstSubObjects;
    std::list<unsigned int>::iterator iterSubObjects;

    // Update folder itself
    if(bUpdateRoot) {
        strQuery = "UPDATE properties SET storeid=" + stringify(ulStoreId) + " WHERE hierarchyid=" + stringify(ulFolderId);

        er = lpDatabase->DoUpdate(strQuery);
        if(er != erSuccess)
            goto exit;

        g_lpSessionManager->GetCacheManager()->Update(fnevObjectDeleted, ulFolderId);
		// HACK: we're not actualy changing the data here, but we know it's going to change
		g_lpSessionManager->GetCacheManager()->RemoveIndexData(ulFolderId);
    }

    // Get list of subobjects
    strQuery = "SELECT id FROM hierarchy WHERE parent=" + stringify(ulFolderId);

    er = lpDatabase->DoSelect(strQuery, &lpResult);
    if(er != erSuccess)
        goto exit;

    while(1) {
        lpRow = lpDatabase->FetchRow(lpResult);

        if(lpRow == NULL)
            break;

        if(lpRow[0] == NULL)
            continue;

        lstSubObjects.push_back(atoi(lpRow[0]));
    }

	if(!lstSubObjects.empty()) {
        // Update subobjects
        strQuery = "UPDATE properties SET storeid=" + stringify(ulStoreId) + " WHERE hierarchyid IN(";

        for(iterSubObjects = lstSubObjects.begin(); iterSubObjects != lstSubObjects.end(); iterSubObjects++) {
            strQuery += stringify(*iterSubObjects);
            strQuery += ",";

            g_lpSessionManager->GetCacheManager()->Update(fnevObjectDeleted, *iterSubObjects);
			// HACK: we're not actualy changing the data here, but we know it's going to change
			g_lpSessionManager->GetCacheManager()->RemoveIndexData(*iterSubObjects);
        }

        // Remove trailing comma
        strQuery = strQuery.substr(0,strQuery.size()-1);

        strQuery += ")";

        // Update database
		//FIXME: uhmm check error??
        lpDatabase->DoUpdate(strQuery);

    }

    // continue down in the tree recursively
    for(iterSubObjects = lstSubObjects.begin(); iterSubObjects != lstSubObjects.end(); iterSubObjects++) {
        er = UpdateStoreIDRecursive(lpDatabase, *iterSubObjects, ulStoreId, false); // the root has already been updated above
		if(er != erSuccess)
			goto exit;
    }
exit:
    if(lpResult)
        lpDatabase->FreeResult(lpResult);

    return er;
}

enum eSizeUpdateAction{ UPDATE_NEW, UPDATE_ADD, UPDATE_SUB };
ECRESULT UpdateObjectSize(ECDatabase* lpDatabase, unsigned int ulStoreId, unsigned int ulObjId, unsigned int ulObjType, bool fNewItem, eSizeUpdateAction updateAction, long long llSize)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulPropTag = 0;
	unsigned int	ulAffRows = 0;
	std::string		strQuery;
	std::string		strField;

	if(ulObjType == MAPI_ATTACH) {
		ulPropTag = PR_ATTACH_SIZE;
		strField = "val_ulong";
	}else if(ulObjType == MAPI_STORE) {
		ulPropTag = PR_MESSAGE_SIZE_EXTENDED;
		strField = "val_longint";
	}else {
		ulPropTag = PR_MESSAGE_SIZE;
		strField = "val_ulong";
	}

	if(fNewItem) {
		llSize+=20; //size of PR_MESSAGE_SIZE eg
		strQuery = "REPLACE INTO properties(hierarchyid, tag, type, "+strField+", storeid) VALUES(" + stringify(ulObjId) + "," + stringify(PROP_ID(ulPropTag)) + "," + stringify(PROP_TYPE(ulPropTag)) + "," + stringify_int64(llSize) + "," + stringify(ulStoreId) + ")";
		er = lpDatabase->DoInsert(strQuery);
	}else {

		strQuery = "UPDATE properties SET "+strField+"=";

		if(updateAction == UPDATE_ADD)
			strQuery += strField+"+";
		else if(updateAction == UPDATE_SUB)
			strQuery += strField+"-";

		strQuery += stringify_int64(llSize) +" WHERE tag="+stringify(PROP_ID(ulPropTag))+" AND type="+stringify(PROP_TYPE(ulPropTag)) + " AND hierarchyid="+stringify(ulObjId) + " AND storeid=" + stringify(ulStoreId);
		if(updateAction == UPDATE_SUB)
			strQuery += " AND "+strField+" >="+stringify_int64(llSize);

		er = lpDatabase->DoUpdate(strQuery, &ulAffRows);

	}// if(fNewItem)

	return er;
}

/* Get the size of an object, PR_MESSAGE_SIZE or PR_ATTACH_SIZE */
ECRESULT GetObjectSize(ECDatabase* lpDatabase, unsigned int ulStoreId, unsigned int ulObjId, unsigned int* lpulSize)
{
	ECRESULT		er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	unsigned int	ulSize = 0;
	std::string		strQuery;

	strQuery = "SELECT val_ulong FROM properties WHERE hierarchyid="+stringify(ulObjId)+" AND ((tag="+stringify(PROP_ID(PR_MESSAGE_SIZE))+" AND type="+stringify(PROP_TYPE(PR_MESSAGE_SIZE))+") OR (tag="+stringify(PROP_ID(PR_ATTACH_SIZE))+" AND type="+stringify(PROP_TYPE(PR_ATTACH_SIZE))+") ) AND properties.storeid=" + stringify(ulStoreId);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) != 1) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if(lpDBRow == NULL || lpDBRow[0] == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	ulSize = atoi(lpDBRow[0]);

	// Free results
	if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

	*lpulSize = ulSize;
exit:
	// Free results
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

// sync id only to saveObject !
// You need to check the permissions and quota before you call this function
ECRESULT WriteProps(struct soap *soap, ECSession *lpecSession, ECDatabase *lpDatabase, struct saveObject *lpsSaveObj, unsigned int ulObjId, bool fNewItem, unsigned int ulSyncId, struct saveObject *lpsReturnObj, bool *lpfHaveChangeKey)
{
	ECRESULT		er;
    std::string		strInsertQuery;
	std::string		strColName;
	std::string		strColData;
	std::string		strQuery;

	struct propValArray *lpPropValArray = &lpsSaveObj->modProps;

	unsigned int	ulParent = 0;
	unsigned int	ulGrandParent = 0;
	unsigned int	ulParentType = 0;
	unsigned int	ulObjType = 0;
	unsigned int	ulOwner = 0;
	unsigned int	ulFlags = 0;
	unsigned int	ulAffected = 0;
	unsigned int	ulTemp = 0;
	unsigned int	j;
	int				i;
	unsigned int	ulPropInserts = 0;

	ULONG			nMVItems;
	unsigned int	ulStoreId = 0;
	unsigned long long ullIMAP = 0;

	std::set<unsigned int>	setInserted;
	std::set<unsigned int>::iterator iterInserted;

	GUID sGuidServer;
	ECAttachmentStorage *lpAttachmentStorage = NULL;
	std::list<ULONG> lstObjIds;
	ULONG ulInstanceId = 0;
	ULONG ulInstanceTag = 0;
	bool bAttachmentStored = false;
	entryId sUserId;
	std::string strUsername;
	
	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey;

	DB_RESULT		lpDBResult = NULL;

	if (lpfHaveChangeKey)
		*lpfHaveChangeKey = false;

	er = g_lpSessionManager->GetServerGUID(&sGuidServer);
	if (er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetStore(ulObjId, &ulStoreId, NULL);
	if(er != erSuccess)
	    goto exit;

    er = g_lpSessionManager->GetCacheManager()->GetObject(ulObjId, &ulParent, &ulOwner, &ulFlags, &ulObjType);
    if(er != erSuccess)
        goto exit;

	if(ulObjType != MAPI_STORE){
		er = g_lpSessionManager->GetCacheManager()->GetObject(ulParent, &ulGrandParent, NULL, NULL, &ulParentType);
		if(er != erSuccess)
			goto exit;
	}

	if(ulObjType == MAPI_FOLDER && !ulSyncId)
	{
		for(i=0;i<lpPropValArray->__size;i++) {

			// Check whether the requested folder name already exists
			if(lpPropValArray->__ptr[i].ulPropTag == PR_DISPLAY_NAME )
			{
				if(lpPropValArray->__ptr[i].Value.lpszA == NULL)
					break; // Name property found, but name isn't present. This is broken, so skip this.

				strQuery = "SELECT hierarchy.id FROM hierarchy JOIN properties ON hierarchy.id = properties.hierarchyid WHERE hierarchy.parent=" + stringify(ulParent) + " AND hierarchy.type="+stringify(MAPI_FOLDER)+" AND hierarchy.flags & " + stringify(MSGFLAG_DELETED)+ "=0 AND properties.tag=" + stringify(ZARAFA_TAG_DISPLAY_NAME) + " AND properties.val_string = '" + lpDatabase->Escape(lpPropValArray->__ptr[i].Value.lpszA) + "' AND properties.storeid=" + stringify(ulStoreId) +" AND hierarchy.id!=" + stringify(ulObjId);
				er = lpDatabase->DoSelect(strQuery, &lpDBResult);
				if(er != erSuccess)
					goto exit;

				if(lpDatabase->GetNumRows(lpDBResult) > 0) {
					er = ZARAFA_E_COLLISION;
					goto exit;
				}

				// Free results
				if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

				break;
			}
		}// for(...)
	}


	if(fNewItem) {
		// If this is a new item, update the createtime and modtime to NOW
		strQuery = "UPDATE hierarchy SET modtime=CURRENT_TIMESTAMP, createtime=CURRENT_TIMESTAMP WHERE id=" + stringify(ulObjId);

		er = lpDatabase->DoUpdate(strQuery);
		if(er != erSuccess)
			goto exit;

	} else {
		// Just an update, so write the modtime
		strQuery = "UPDATE hierarchy SET modtime=CURRENT_TIMESTAMP";
		if(ulObjType == MAPI_MESSAGE) // message is modified
			strQuery+= ", flags=flags&"+stringify(~MSGFLAG_UNMODIFIED);

		strQuery+= " WHERE id=" + stringify(ulObjId);

		er = lpDatabase->DoUpdate(strQuery);
		if(er != erSuccess)
			 goto exit;
	}

	// If we have a recipient, remove the old one. The client does not send a diff of props because ECMemTable cannot do this
	if (!fNewItem && (ulObjType == MAPI_MAILUSER || ulObjType == MAPI_DISTLIST)) {
		strQuery = "DELETE FROM properties WHERE hierarchyid=" + stringify(ulObjId);

		er = lpDatabase->DoDelete(strQuery);
		if(er != erSuccess)
			goto exit;

		strQuery = "DELETE FROM mvproperties WHERE hierarchyid=" + stringify (ulObjId);

		er = lpDatabase->DoDelete(strQuery);
		if(er != erSuccess)
			goto exit;
	}

	/* FIXME: Support multiple InstanceIds */
	if (lpsSaveObj->lpInstanceIds && lpsSaveObj->lpInstanceIds->__size) {
		GUID sGuidTmp;

		if (lpsSaveObj->lpInstanceIds->__size > 1) {
			er = ZARAFA_E_UNKNOWN_INSTANCE_ID;
			goto exit;
		}

		er = SIEntryIDToID(&lpsSaveObj->lpInstanceIds->__ptr[0], &sGuidTmp, (unsigned int*)&ulInstanceId, (unsigned int*)&ulInstanceTag);
		if (er != erSuccess)
			goto exit;

		/* Server GUID must always match */
		if (memcmp(&sGuidTmp, &sGuidServer, sizeof(sGuidTmp)) != 0) {
			er = ZARAFA_E_UNKNOWN_INSTANCE_ID;
			goto exit;
		}

		er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
		if (er != erSuccess)
			goto exit;

		/* The attachment should at least exist */
		if (!lpAttachmentStorage->ExistAttachmentInstance(ulInstanceId)) {
			er = ZARAFA_E_UNKNOWN_INSTANCE_ID;
			goto exit;
		}

		/*
		 * Check if we have access the the Instance which is being referenced,
		 * an user has access to an Instance when he is administrator or owns at
		 * least one reference to the Instance.
		 * For security we won't announce the difference between not finding the instance
		 * or not having access to it.
		 */
		if (lpecSession->GetSecurity()->GetAdminLevel() != ADMIN_LEVEL_SYSADMIN) {
			er = lpAttachmentStorage->GetSingleInstanceParents(ulInstanceId, &lstObjIds);
			if (er != erSuccess)
				goto exit;

			er = ZARAFA_E_UNKNOWN_INSTANCE_ID;
			for (std::list<ULONG>::iterator i = lstObjIds.begin(); i != lstObjIds.end(); i++) {
				if (lpecSession->GetSecurity()->CheckPermission(*i, ecSecurityRead) == erSuccess) {
						er = erSuccess;
						break;
				}
			}

			if (er != erSuccess)
				goto exit;
		}

		er = lpAttachmentStorage->SaveAttachment(ulObjId, !fNewItem, ulInstanceId, &ulInstanceId);
		if (er != erSuccess)
			goto exit;

		lpsReturnObj->lpInstanceIds = s_alloc<entryList>(soap);
		lpsReturnObj->lpInstanceIds->__size = 1;
		lpsReturnObj->lpInstanceIds->__ptr = s_alloc<entryId>(soap, lpsReturnObj->lpInstanceIds->__size);

		er = SIIDToEntryID(soap, &sGuidServer, ulInstanceId, ulInstanceTag, &lpsReturnObj->lpInstanceIds->__ptr[0]);
		if (er != erSuccess) {
			lpsReturnObj->lpInstanceIds = NULL;
			goto exit;
		}

		/* Either by instanceid or attachment data, we have stored the attachment */
		bAttachmentStored = true;
	}

	// Write the properties
	for(i=0;i<lpPropValArray->__size;i++) {
	    // Check if we already inserted this property tag. We only accept the first.
	    iterInserted = setInserted.find(lpPropValArray->__ptr[i].ulPropTag);
	    if(iterInserted != setInserted.end())
	        continue;

        // Check if we actually need to write this property. The client may send us properties
        // that we generate, so we don't need to save them
        if(ECGenProps::IsPropRedundant(lpPropValArray->__ptr[i].ulPropTag, ulObjType) == erSuccess)
            continue;

		// The client may only override the PR_LAST_MODIFICATION_TIME on its FIRST SaveChanges()
		if(lpPropValArray->__ptr[i].ulPropTag == PR_LAST_MODIFICATION_TIME) {
			if(!fNewItem)
				continue;
			else {
				// Update the mod. time in hierarchy table
				if(lpPropValArray->__ptr[i].Value.hilo == NULL)
					continue;

				strQuery = "UPDATE hierarchy SET modtime='" + stringify_datetime(FileTimeToUnixTime(lpPropValArray->__ptr[i].Value.hilo->hi, lpPropValArray->__ptr[i].Value.hilo->lo)) + "' WHERE id=" + stringify(ulObjId);

				er = lpDatabase->DoUpdate(strQuery);
				if(er != erSuccess)
					goto exit;
			}
			continue;
		}

		if(lpPropValArray->__ptr[i].ulPropTag == PR_LAST_MODIFIER_NAME || lpPropValArray->__ptr[i].ulPropTag == PR_LAST_MODIFIER_ENTRYID) {
			if(!fNewItem && ulSyncId == 0)
                continue;
		}

		// Same goes for flags in PR_MESSAGE_FLAGS
		if(lpPropValArray->__ptr[i].ulPropTag == PR_MESSAGE_FLAGS) {
			if(!fNewItem)
				continue;
			else {
			    ulFlags = (lpPropValArray->__ptr[i].Value.ul & (MSGFLAG_SETTABLE_BY_USER | MSGFLAG_SETTABLE_BY_SPOOLER)) | ulFlags;

				// Update the flags in hierarchy table
				strQuery = "UPDATE hierarchy SET flags=" + stringify(ulFlags)  + " WHERE id=" + stringify(ulObjId);

				er = lpDatabase->DoUpdate(strQuery);
				if(er != erSuccess)
					goto exit;
			}
			continue;
		}
		// Same goes for PR_SEARCH_KEY
		if(lpPropValArray->__ptr[i].ulPropTag == PR_SEARCH_KEY)
		{
			if(!fNewItem)
				continue;
		}

		if(lpPropValArray->__ptr[i].ulPropTag == PR_SOURCE_KEY)
		{
			if(!(fNewItem && ulParentType == MAPI_FOLDER))
				continue;

			//don't use the sourcekey if found.
			if(ZARAFA_E_NOT_FOUND != g_lpSessionManager->GetCacheManager()->GetObjectFromProp(PROP_ID(PR_SOURCE_KEY), lpPropValArray->__ptr[i].Value.bin->__size, lpPropValArray->__ptr[i].Value.bin->__ptr, &ulTemp))
				continue;

			// Insert sourcekey, use REPLACE because createfolder already created a sourcekey.
			// Because there is a non-primary unique key on the
			// val_binary part of the table, it will fail if the source key is duplicate.
			strQuery = "REPLACE INTO indexedproperties(hierarchyid,tag,val_binary) VALUES ("+stringify(ulObjId)+","+stringify(PROP_ID(PR_SOURCE_KEY))+","+lpDatabase->EscapeBinary(lpPropValArray->__ptr[i].Value.bin->__ptr, lpPropValArray->__ptr[i].Value.bin->__size)+")";
			er = lpDatabase->DoInsert(strQuery);
			if(er != erSuccess)
				goto exit;

			setInserted.insert(lpPropValArray->__ptr[i].ulPropTag);

			// Remember the source key in the cache
			g_lpSessionManager->GetCacheManager()->SetObjectProp(PROP_ID(PR_SOURCE_KEY), lpPropValArray->__ptr[i].Value.bin->__size, lpPropValArray->__ptr[i].Value.bin->__ptr, ulObjId);
			continue;
		}

		// attachments are in the blob too
		if (lpPropValArray->__ptr[i].ulPropTag == PR_ATTACH_DATA_BIN) {
			/* 
			 * bAttachmentStored indicates we already processed the attachment.
			 * this could happen when the user provided the instance ID but as
			 * backup also send the PR_ATTACH_DATA_BIN.
			 */
			if (bAttachmentStored)
				continue;

			if (!lpAttachmentStorage) {
				er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
				if (er != erSuccess)
					goto exit;
			}

			er = lpAttachmentStorage->SaveAttachment(ulObjId, !fNewItem,
													 lpPropValArray->__ptr[i].Value.bin->__size,
													 lpPropValArray->__ptr[i].Value.bin->__ptr,
													 &ulInstanceId);
			if (er != erSuccess)
				goto exit;

			lpsReturnObj->lpInstanceIds = s_alloc<entryList>(soap);
			lpsReturnObj->lpInstanceIds->__size = 1;
			lpsReturnObj->lpInstanceIds->__ptr = s_alloc<entryId>(soap, lpsReturnObj->lpInstanceIds->__size);

			er = SIIDToEntryID(soap, &sGuidServer, ulInstanceId, lpPropValArray->__ptr[i].ulPropTag, &lpsReturnObj->lpInstanceIds->__ptr[0]);
			if (er != erSuccess) {
				lpsReturnObj->lpInstanceIds = NULL;
				goto exit;
			}

			continue;
		}
		
		if (lpfHaveChangeKey && lpPropValArray->__ptr[i].ulPropTag == PR_CHANGE_KEY)
			*lpfHaveChangeKey = true;


		if((PROP_TYPE(lpPropValArray->__ptr[i].ulPropTag) & MV_FLAG) == MV_FLAG) {
			//Write mv properties
			if(!fNewItem) {

				strQuery = "DELETE FROM mvproperties WHERE hierarchyid=" + stringify (ulObjId) +
							" AND tag=" + stringify(PROP_ID(lpPropValArray->__ptr[i].ulPropTag)) +
							" AND type=" + stringify(PROP_TYPE(lpPropValArray->__ptr[i].ulPropTag));

				er = lpDatabase->DoDelete(strQuery);
				if(er != erSuccess)
					goto exit;
			}

			nMVItems = GetMVItemCount(&lpPropValArray->__ptr[i]);
			for(j=0; j < nMVItems; j++)
			{
				er = CopySOAPPropValToDatabaseMVPropVal(&lpPropValArray->__ptr[i], j, strColName, strColData, lpDatabase);
				if(er != erSuccess)
					continue;

				strQuery = "INSERT INTO mvproperties(hierarchyid,orderid,tag,type," + strColName + ") VALUES(" + stringify(ulObjId) + "," + stringify(j) + "," + stringify(PROP_ID(lpPropValArray->__ptr[i].ulPropTag)) + "," + stringify(PROP_TYPE(lpPropValArray->__ptr[i].ulPropTag)) + "," + strColData + ")";
				er = lpDatabase->DoInsert(strQuery, NULL, &ulAffected);
				if(er != erSuccess)
					goto exit;

				if(ulAffected != 1) {
					er = ZARAFA_E_DATABASE_ERROR;
					goto exit;
				}
			}
		} else {
			//Write basic properties
			unsigned int ulColId = 0;
			sObjectTableKey key(ulObjId,0);
			
			g_lpSessionManager->GetCacheManager()->SetCell(&key, lpPropValArray->__ptr[i].ulPropTag, &lpPropValArray->__ptr[i]);
			
			er = CopySOAPPropValToDatabasePropVal(&lpPropValArray->__ptr[i], &ulColId, strColData, lpDatabase);
			if(er != erSuccess) {
				er = erSuccess;
				continue;
			}

			if(!strInsertQuery.empty() && strInsertQuery.size() + strColData.size()*2 > lpDatabase->GetMaxAllowedPacket()) {
			    // The packet may become larger than the max allowed packet size, assuming worst-case encoding size of size*2. Try
			    // to write the packet by flushing all data until now and starting a new query.
			    er = lpDatabase->DoInsert(strInsertQuery, NULL, NULL);
				if(er != erSuccess)
					goto exit;

				strInsertQuery.clear();
			}

			if(strInsertQuery.empty()) {
				strInsertQuery = "REPLACE INTO properties (hierarchyid,tag,type,storeid," + (std::string)PROPCOLVALUEORDER + ") VALUES";
				ulPropInserts = 0;
            }

			if(ulPropInserts > 0)
				strInsertQuery += ",";

			strInsertQuery += "(" + stringify(ulObjId) + "," +
									stringify(PROP_ID(lpPropValArray->__ptr[i].ulPropTag)) + "," +
									stringify(PROP_TYPE(lpPropValArray->__ptr[i].ulPropTag)) + "," +
									stringify(ulStoreId) + ",";

			for(unsigned int k=0; k<VALUE_NR_MAX; k++) {
				if(k==ulColId) {
					strInsertQuery += strColData;
				} else {
					if(k == VALUE_NR_HILO)
						strInsertQuery += "null,null";
					else
						strInsertQuery += "null";
				}
				if(k != VALUE_NR_MAX-1) {
					strInsertQuery += ",";
				}
			}

			strInsertQuery += ")";
			ulPropInserts++;
		}

		setInserted.insert(lpPropValArray->__ptr[i].ulPropTag);
	} // for(i=0;i<lpPropValArray->__size;i++)

	// ignore error from failed properties
	er = erSuccess;

	// Insert the properties
	if(ulPropInserts > 0) {
		er = lpDatabase->DoInsert(strInsertQuery, NULL, NULL);
		if(er != erSuccess)
			goto exit;
	}
	
	if(ulObjType == MAPI_MESSAGE) {
		iterInserted = setInserted.find(PR_LAST_MODIFIER_NAME);
		
		// update the PR_LAST_MODIFIER_NAME and PR_LAST_MODIFIER_ENTRYID
		if(iterInserted == setInserted.end()) {
			er = GetABEntryID(lpecSession->GetSecurity()->GetUserId(), soap, &sUserId);
			if(er != erSuccess)
				goto exit;
			
			lpecSession->GetSecurity()->GetUsername(&strUsername); //Ignore errors

			strQuery = "REPLACE INTO properties(hierarchyid, tag, type, storeid, val_string, val_binary) VALUES(" +
						stringify(ulObjId) + "," +
						stringify(PROP_ID(PR_LAST_MODIFIER_NAME)) + "," +
						stringify(PROP_TYPE(PR_LAST_MODIFIER_NAME)) + "," +
						stringify(ulStoreId) + ",\"" +
						lpDatabase->Escape(strUsername) +
						"\", NULL), (" +
						stringify(ulObjId) + "," +
						stringify(PROP_ID(PR_LAST_MODIFIER_ENTRYID)) + "," +
						stringify(PROP_TYPE(PR_LAST_MODIFIER_ENTRYID)) + "," +
						stringify(ulStoreId) + ", NULL, " +
						lpDatabase->EscapeBinary(sUserId.__ptr, sUserId.__size) + ")";

			er = lpDatabase->DoInsert(strQuery);
			if(er != erSuccess)
				goto exit;

			sObjectTableKey key(ulObjId,0);
			struct propVal	sPropVal;

			sPropVal.ulPropTag = PR_LAST_MODIFIER_NAME;
			sPropVal.Value.lpszA = (char*)strUsername.c_str();
			sPropVal.__union = SOAP_UNION_propValData_lpszA;
            g_lpSessionManager->GetCacheManager()->SetCell(&key, PR_LAST_MODIFIER_NAME, &sPropVal);

			sPropVal.ulPropTag = PR_LAST_MODIFIER_ENTRYID;
			sPropVal.Value.bin = &sUserId;
			sPropVal.__union = SOAP_UNION_propValData_bin;
			g_lpSessionManager->GetCacheManager()->SetCell(&key, PR_LAST_MODIFIER_ENTRYID, &sPropVal);
		}
	}

	if(fNewItem && ulObjType == MAPI_MESSAGE) {
        // Add PR_SOURCE_KEY to new messages without a given PR_SOURCE_KEY
        // This isn't for folders, this done in the createfolder function
		iterInserted = setInserted.find(PR_SOURCE_KEY);

		if(iterInserted == setInserted.end()) {
			lpecSession->GetNewSourceKey(1, &sSourceKey);

			strQuery = "INSERT INTO indexedproperties(hierarchyid,tag,val_binary) VALUES(" + stringify(ulObjId) + "," + stringify(PROP_ID(PR_SOURCE_KEY)) + "," + lpDatabase->EscapeBinary(sSourceKey, sSourceKey.size()) + ")";
			er = lpDatabase->DoInsert(strQuery);

			if(er != erSuccess)
				goto exit;

			g_lpSessionManager->GetCacheManager()->SetObjectProp(PROP_ID(PR_SOURCE_KEY), sSourceKey.size(), sSourceKey, ulObjId);
		}


		if(ulParentType == MAPI_FOLDER) {
			// Add a PR_EC_IMAP_ID to the newly created message
			er = g_lpSessionManager->GetNewSequence(ECSessionManager::SEQ_IMAP, 1, &ullIMAP);
			if(er != erSuccess)
				goto exit;

			strQuery = "INSERT INTO properties(hierarchyid, tag, type, storeid, val_ulong) VALUES(" +
						stringify(ulObjId) + "," +
						stringify(PROP_ID(PR_EC_IMAP_ID)) + "," +
						stringify(PROP_TYPE(PR_EC_IMAP_ID)) + "," +
						stringify(ulStoreId) + "," +
						stringify(ullIMAP) +
						")";

			er = lpDatabase->DoInsert(strQuery);
			if(er != erSuccess)
				goto exit;
		}
	}

	// Invalidate anything cached for this object
	if(!fNewItem)
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulObjId);

	// We know the values for the object cache, so add them here
	g_lpSessionManager->GetCacheManager()->SetObject(ulObjId, ulParent, ulOwner, ulFlags, ulObjType);
exit:
	if (lpAttachmentStorage)
		delete lpAttachmentStorage;

    FREE_DBRESULT();
	return er;
}

// You need to check the permissions before you call this function
ECRESULT DeleteProps(ECSession *lpecSession, ECDatabase *lpDatabase, ULONG ulObjId, struct propTagArray *lpsPropTags) {
	ECRESULT er = erSuccess;
	int				i;
	unsigned int	ulStoreId = 0;
	std::string		strQuery;

	er = g_lpSessionManager->GetCacheManager()->GetStore(ulObjId, &ulStoreId, NULL);
	if(er != erSuccess)
	    goto exit;

	// Delete one or more properties of an object
	for(i=0;i<lpsPropTags->__size;i++) {
		if((lpsPropTags->__ptr[i]&MV_FLAG) == 0)
			strQuery = "DELETE FROM properties WHERE hierarchyid="+stringify(ulObjId)+" AND tag="+stringify(PROP_ID(lpsPropTags->__ptr[i])) + " AND storeid= " + stringify(ulStoreId);
		else // mvprops
			strQuery = "DELETE FROM mvproperties WHERE hierarchyid="+stringify(ulObjId)+" AND tag="+stringify(PROP_ID(lpsPropTags->__ptr[i]) );

		er = lpDatabase->DoDelete(strQuery);
		if(er != erSuccess)
			goto exit;
	}

	g_lpSessionManager->GetCacheManager()->Update(fnevObjectDeleted, ulObjId);

exit:
	return er;
}

unsigned int SaveObject(struct soap *soap, ECSession *lpecSession, ECDatabase *lpDatabase, unsigned int ulStoreId, unsigned int ulParentObjId, unsigned int ulParentType,
						unsigned int ulFlags, unsigned int ulSyncId, struct saveObject *lpsSaveObj, struct saveObject *lpsReturnObj, bool *lpfHaveChangeKey = NULL)
{
	ECRESULT er = erSuccess;
	ALLOC_DBRESULT();
	int n;
	unsigned int ulStore = 0;
	unsigned int ulParentObjType = 0;
	unsigned int ulSize = 0;
	unsigned int ulObjId = 0;
	bool fNewItem = false;

	// reset return object
	lpsReturnObj->__size = 0;
	lpsReturnObj->__ptr = NULL;
	lpsReturnObj->delProps.__size = 0;
	lpsReturnObj->delProps.__ptr = NULL;
	lpsReturnObj->modProps.__size = 0;
	lpsReturnObj->modProps.__ptr = NULL;
	lpsReturnObj->bDelete = false;
	lpsReturnObj->ulClientId = lpsSaveObj->ulClientId;
	lpsReturnObj->ulServerId = lpsSaveObj->ulServerId;
	lpsReturnObj->ulObjType = lpsSaveObj->ulObjType;
	lpsReturnObj->lpInstanceIds = NULL;

	if (lpsSaveObj->ulServerId == 0) {
		if (ulParentObjId == 0) {
			er = ZARAFA_E_INVALID_PARAMETER;
			goto exit;
		}

		er = CreateObject(lpecSession, lpDatabase, ulParentObjId, lpsSaveObj->ulObjType, ulFlags, &lpsReturnObj->ulServerId);
		if (er != erSuccess)
			goto exit;

		fNewItem = true;

		ulObjId = lpsReturnObj->ulServerId;
	} else {
	    ulObjId = lpsSaveObj->ulServerId;
    }

	if (lpsSaveObj->bDelete) {
		// make list of all children object id's in std::list<int> ?
		ECListInt lstDel;

		lstDel.push_back(lpsSaveObj->ulServerId);

		// we always hard delete, because we can only delete submessages here
		// make sure we also delete message-in-message attachments, so all message related flags are on
		er = DeleteObjects(lpecSession, lpDatabase, &lstDel, EC_DELETE_CONTAINER | EC_DELETE_MESSAGES | EC_DELETE_RECIPIENTS | EC_DELETE_ATTACHMENTS | EC_DELETE_HARD_DELETE, ulSyncId, true, false);

		// object deleted, so we're done
		goto exit;
	}

	// ------
	// the following code is only for objects that still exist
	// ------
	{

	    // Don't delete properties if this is a new object: this avoids any delete queries that cause unneccessary locks on the tables
		if (lpsSaveObj->delProps.__size > 0 && !fNewItem) {
			er = DeleteProps(lpecSession, lpDatabase, lpsReturnObj->ulServerId, &lpsSaveObj->delProps);
			if (er != erSuccess)
				goto exit;
		}

		if (lpsSaveObj->modProps.__size > 0) {
			er = WriteProps(soap, lpecSession, lpDatabase, lpsSaveObj, lpsReturnObj->ulServerId, fNewItem, ulSyncId, lpsReturnObj, lpfHaveChangeKey);
			if (er != erSuccess)
				goto exit;
		}

		// check children
		if (lpsSaveObj->__size > 0) {
			lpsReturnObj->__size = lpsSaveObj->__size;
			lpsReturnObj->__ptr = s_alloc<struct saveObject>(soap, lpsReturnObj->__size);

			for (int i = 0; i < lpsSaveObj->__size; i++) {
				er = SaveObject(soap, lpecSession, lpDatabase, ulStoreId, /*myself as parent*/lpsReturnObj->ulServerId, lpsReturnObj->ulObjType, 0, ulSyncId, &lpsSaveObj->__ptr[i], &lpsReturnObj->__ptr[i]);
				if (er != erSuccess)
					goto exit;
			}
		}

		// 1. calc size of object, now that all children are saved.

		if (lpsReturnObj->ulObjType == MAPI_MESSAGE || lpsReturnObj->ulObjType == MAPI_ATTACH) {
			// Remove old size
			if (fNewItem != true && lpsReturnObj->ulObjType == MAPI_MESSAGE && ulParentType == MAPI_FOLDER) {
				if (GetObjectSize(lpDatabase, ulStoreId, lpsReturnObj->ulServerId, &ulSize) == erSuccess)
					UpdateObjectSize(lpDatabase, ulStoreId, ulStoreId, MAPI_STORE, false, UPDATE_SUB, ulSize);
			}

			// Add new size
			if (CalculateObjectSize(lpDatabase, lpsReturnObj->ulServerId, lpsReturnObj->ulObjType, ulStoreId, &ulSize) == erSuccess) {
				UpdateObjectSize(lpDatabase, ulStoreId, lpsReturnObj->ulServerId, lpsReturnObj->ulObjType, fNewItem, UPDATE_NEW, ulSize);

				if (lpsReturnObj->ulObjType == MAPI_MESSAGE && ulParentType == MAPI_FOLDER) {
					UpdateObjectSize(lpDatabase, ulStoreId, ulStoreId, MAPI_STORE, false, UPDATE_ADD, ulSize);
				}

			} else {
				ASSERT(FALSE);
			}
		}

		// 2. find props to return

		// the server returns the following 4 properties, when the item is new:
		//   PR_CREATION_TIME, (PR_PARENT_SOURCE_KEY, PR_SOURCE_KEY /type==5|3) (PR_RECORD_KEY /type==7|5|3)
		// TODO: recipients: PR_INSTANCE_KEY
		// it always sends the following property:
		//   PR_LAST_MODIFICATION_TIME
		// currently, it always sends them all
		// we also can't send the PR_MESSAGE_SIZE and PR_MESSAGE_FLAGS, since the recursion is the wrong way around: attachments come later than the actual message
		// we can skip PR_ACCESS and PR_ACCESS_LEVEL because the client already inherited those from the parent
		// we need to alloc 2 properties for PR_CHANGE_KEY and PR_PREDECESSOR_CHANGE_LIST
		lpsReturnObj->delProps.__size = 8;
		lpsReturnObj->delProps.__ptr = s_alloc<unsigned int>(soap, lpsReturnObj->delProps.__size);
		lpsReturnObj->modProps.__size = 8;
		lpsReturnObj->modProps.__ptr = s_alloc<struct propVal>(soap, lpsReturnObj->modProps.__size);

		n = 0;

		// set the PR_RECORD_KEY
		if (lpsSaveObj->ulObjType == MAPI_ATTACH || lpsSaveObj->ulObjType == MAPI_MESSAGE || lpsSaveObj->ulObjType == MAPI_FOLDER) {
			if (ECGenProps::GetPropComputedUncached(soap, lpecSession, PR_RECORD_KEY, lpsSaveObj->ulServerId, 0, 0, &lpsReturnObj->modProps.__ptr[n]) == erSuccess) {
				lpsReturnObj->delProps.__ptr[n] = PR_RECORD_KEY;
				n++;
			}
		}

		if (lpsSaveObj->ulObjType != MAPI_STORE) {
			er = g_lpSessionManager->GetCacheManager()->GetObject(ulParentObjId, NULL, NULL, NULL, &ulParentObjType);
			if(er != erSuccess)
				goto exit;
		}

		//PR_PARENT_SOURCE_KEY for folders and messages
		if (lpsSaveObj->ulObjType == MAPI_FOLDER || (lpsSaveObj->ulObjType == MAPI_MESSAGE && ulParentObjType == MAPI_FOLDER))
		{
			if(ECGenProps::GetPropComputed(soap, lpecSession, PR_PARENT_SOURCE_KEY, ulObjId, 0, 0, lpsSaveObj->ulObjType, 0, ulStore, NULL, &lpsReturnObj->modProps.__ptr[n]) == erSuccess)
			{
				lpsReturnObj->delProps.__ptr[n] = PR_PARENT_SOURCE_KEY;
				n++;
			}

			if(ECGenProps::GetPropComputed(soap, lpecSession, PR_SOURCE_KEY, ulObjId, 0, 0, lpsSaveObj->ulObjType, 0, ulStore, NULL, &lpsReturnObj->modProps.__ptr[n]) == erSuccess)
			{
				lpsReturnObj->delProps.__ptr[n] = PR_SOURCE_KEY;
				n++;
			}
		}

		// Get PR_LAST_MODIFICATION_TIME, PR_CREATION_TIME
		strQuery = "SELECT UNIX_TIMESTAMP(hierarchy.modtime), UNIX_TIMESTAMP(hierarchy.createtime) FROM hierarchy WHERE hierarchy.id="+stringify(lpsReturnObj->ulServerId);
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		if (lpDBRow == NULL) {
			// this would be really bad, since we just wrote it
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		if (lpDBRow[0])
		{
			lpsReturnObj->delProps.__ptr[n] = PR_LAST_MODIFICATION_TIME;

			FILETIME ftMod;

			UnixTimeToFileTime(atoi(lpDBRow[0]), &ftMod);

			lpsReturnObj->modProps.__ptr[n].__union = SOAP_UNION_propValData_hilo;
			lpsReturnObj->modProps.__ptr[n].ulPropTag = PR_LAST_MODIFICATION_TIME;
			lpsReturnObj->modProps.__ptr[n].Value.hilo = s_alloc<hiloLong>(soap);

			lpsReturnObj->modProps.__ptr[n].Value.hilo->hi = ftMod.dwHighDateTime;
			lpsReturnObj->modProps.__ptr[n].Value.hilo->lo = ftMod.dwLowDateTime;

			n++;
		}

		if (lpDBRow[1])
		{
			lpsReturnObj->delProps.__ptr[n] = PR_CREATION_TIME;

			FILETIME ftCreate;

			UnixTimeToFileTime(atoi(lpDBRow[1]), &ftCreate);

			lpsReturnObj->modProps.__ptr[n].__union = SOAP_UNION_propValData_hilo;
			lpsReturnObj->modProps.__ptr[n].ulPropTag = PR_CREATION_TIME;
			lpsReturnObj->modProps.__ptr[n].Value.hilo = s_alloc<hiloLong>(soap);

			lpsReturnObj->modProps.__ptr[n].Value.hilo->hi = ftCreate.dwHighDateTime;
			lpsReturnObj->modProps.__ptr[n].Value.hilo->lo = ftCreate.dwLowDateTime;

			n++;
		}


		// set actual array size
		lpsReturnObj->delProps.__size = n;
		lpsReturnObj->modProps.__size = n;
	}

exit:
    FREE_DBRESULT();

	return er;
}

SOAP_ENTRY_START6(saveObject, lpsLoadObjectResponse->er, entryId sParentEntryId, entryId sEntryId, struct saveObject *lpsSaveObj, unsigned int ulFlags, unsigned int ulSyncId, struct loadObjectResponse *lpsLoadObjectResponse)
{
	USE_DATABASE();
	unsigned int	ulStoreId = 0;
	unsigned int	ulGrandParent = 0;
	unsigned int	ulGrandParentType = 0;
	unsigned int	ulParentObjId = 0;
	unsigned int	ulParentObjType = 0;
	unsigned int	ulObjFlags = 0;
	unsigned int	ulParentFlags = 0;
	unsigned int	ulObjType = lpsSaveObj->ulObjType;
	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey;
	struct saveObject sReturnObject;
	std::string		strChangeKey;
	std::string		strChangeList;
	long long		llStoreSize = 0;
	eQuotaStatus	QuotaStatus;

	BOOL			fNewItem = false;
	bool			fHaveChangeKey = false;
	unsigned int	ulObjId = 0;
	unsigned int	ulPrevSubmitFlag = 0;
	struct propVal  *lpPropMessageFlags = NULL; // non-free
	struct propVal	*pvCommitTime = NULL;

	er = lpDatabase->Begin();
	if (er != erSuccess)
		goto exit;

	if (!sParentEntryId.__ptr) {
		// saveObject is called on the store itself (doesn't have a parent)
		ulParentObjType = MAPI_STORE;
		ulStoreId = lpsSaveObj->ulServerId;
	} else {
		if (lpsSaveObj->ulServerId == 0) {
			// new object, parent entry id given by client
			er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sParentEntryId, &ulParentObjId);
			if (er != erSuccess)
				goto exit;

			fNewItem = true;
		} else {
			// existing item, search parent ourselves cause the client just sent it's store entryid (see ECMsgStore::OpenEntry())
			er = g_lpSessionManager->GetCacheManager()->GetObject(lpsSaveObj->ulServerId, &ulParentObjId, NULL, NULL, &ulObjType);
			if (er != erSuccess)
				goto exit;

			fNewItem = false;
		}

		if (ulParentObjId != CACHE_NO_PARENT) {
			er = g_lpSessionManager->GetCacheManager()->GetObject(ulParentObjId, NULL, NULL, NULL, &ulParentObjType);
			if(er != erSuccess)
				goto exit;

			er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulParentObjId, &ulStoreId, NULL);
			if(er != erSuccess)
				goto exit;
		} else {
			// we get here on create store, but I'm not exactly sure why :|
			ulParentObjType = MAPI_STORE;
			ulStoreId = lpsSaveObj->ulServerId;
		}
	}


	// Check permissions
	if(!fNewItem && ulObjType == MAPI_FOLDER) {
		er = lpecSession->GetSecurity()->CheckPermission(lpsSaveObj->ulServerId, ecSecurityFolderAccess);
	}else if(fNewItem) {
		er = lpecSession->GetSecurity()->CheckPermission(ulParentObjId, ecSecurityCreate);
	} else {
		er = lpecSession->GetSecurity()->CheckPermission(lpsSaveObj->ulServerId, ecSecurityEdit);
	}
	if(er != erSuccess)
		goto exit;
	
	/////////////////////////////////////////////////
	// Quota check
	if(ulObjType == MAPI_MESSAGE) {
    	er = lpecSession->GetSecurity()->GetStoreSize(ulStoreId, &llStoreSize);
    	if(er != erSuccess)
		    goto exit;

    	er = lpecSession->GetSecurity()->CheckQuota(ulStoreId, llStoreSize, &QuotaStatus);
    	if(er != erSuccess)
    		goto exit;

    	if(QuotaStatus == QUOTA_HARDLIMIT) {
    		er = ZARAFA_E_STORE_FULL;
    		goto exit;
    	}
    }
	
	// update PR_LOCAL_COMMIT_TIME_MAX for disconnected clients who want to know if the folder contents changed
	if (ulObjType == MAPI_MESSAGE && ulParentObjType == MAPI_FOLDER) {
		er = WriteLocalCommitTimeMax(soap, lpDatabase, ulStoreId, ulParentObjId, &pvCommitTime);
		if(er != erSuccess)
			goto exit;
	}

	er = SaveObject(soap, lpecSession, lpDatabase, ulStoreId, ulParentObjId, ulParentObjType, ulFlags, ulSyncId, lpsSaveObj, &sReturnObject, &fHaveChangeKey);
	if (er != erSuccess)
		goto exit;

	if (lpsSaveObj->ulServerId == 0) {
		// map entryId <-> ulObjId on new items
		//0x0FFF = PR_ENTRYID
		strQuery = "INSERT INTO indexedproperties (hierarchyid,tag,val_binary) VALUES("+stringify(sReturnObject.ulServerId)+", 0x0FFF, "+lpDatabase->EscapeBinary(sEntryId.__ptr, sEntryId.__size)+")";
		er = lpDatabase->DoInsert(strQuery);
		if(er != erSuccess)
			goto exit;

		g_lpSessionManager->GetCacheManager()->SetObjectEntryId(&sEntryId, sReturnObject.ulServerId);

		ulObjId = sReturnObject.ulServerId;
	} else {
		ulObjId = lpsSaveObj->ulServerId;
	}

	// 3. pr_source_key magic
	if ((sReturnObject.ulObjType == MAPI_MESSAGE && ulParentObjType == MAPI_FOLDER) ||
		(sReturnObject.ulObjType == MAPI_FOLDER && !(ulFlags & FOLDER_SEARCH)))
	{
		GetSourceKey(sReturnObject.ulServerId, &sSourceKey);
		GetSourceKey(ulParentObjId, &sParentSourceKey);

		if (sReturnObject.ulObjType == MAPI_MESSAGE && ulParentObjType == MAPI_FOLDER) {
			if (lpsSaveObj->ulServerId == 0) {
				AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_MESSAGE_NEW, NULL, 0, !fHaveChangeKey, &strChangeKey, &strChangeList);
			} else {
				AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_MESSAGE_CHANGE, NULL, 0, !fHaveChangeKey, &strChangeKey, &strChangeList);
			}
		} else if (lpsSaveObj->ulObjType == MAPI_FOLDER && !(ulFlags & FOLDER_SEARCH)) {
			AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_FOLDER_CHANGE, NULL, 0, !fHaveChangeKey, &strChangeKey, &strChangeList);
		}

		if(!strChangeKey.empty()){
			sReturnObject.delProps.__ptr[sReturnObject.delProps.__size] = PR_CHANGE_KEY;
			sReturnObject.delProps.__size++;
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].ulPropTag = PR_CHANGE_KEY;
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].__union = SOAP_UNION_propValData_bin;
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].Value.bin = s_alloc<struct xsd__base64Binary>(soap);
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].Value.bin->__size = strChangeKey.size();
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].Value.bin->__ptr = s_alloc<unsigned char>(soap, strChangeKey.size());
			memcpy(sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].Value.bin->__ptr, strChangeKey.c_str(), strChangeKey.size());
			sReturnObject.modProps.__size++;
		}

		if(!strChangeList.empty()){
			sReturnObject.delProps.__ptr[sReturnObject.delProps.__size] = PR_PREDECESSOR_CHANGE_LIST;
			sReturnObject.delProps.__size++;
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].ulPropTag = PR_PREDECESSOR_CHANGE_LIST;
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].__union = SOAP_UNION_propValData_bin;
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].Value.bin = s_alloc<struct xsd__base64Binary>(soap);
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].Value.bin->__size = strChangeList.size();
			sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].Value.bin->__ptr = s_alloc<unsigned char>(soap, strChangeList.size());
			memcpy(sReturnObject.modProps.__ptr[sReturnObject.modProps.__size].Value.bin->__ptr, strChangeList.c_str(), strChangeList.size());
			sReturnObject.modProps.__size++;
		}
	}


	er = g_lpSessionManager->SaveSourceKeyAutoIncrement();
	if (er != erSuccess)
		goto exit;


	// 5. TODO: ulSyncId updates sync tables

	er = lpDatabase->Commit();
	if (er != erSuccess)
		goto exit;

	// 6. process MSGFLAG_SUBMIT if needed

	// If the messages was saved by an ICS syncer, then we need to sync the PR_MESSAGE_FLAGS for MSGFLAG_SUBMIT if it
	// was included in the save.
	lpPropMessageFlags = FindProp(&lpsSaveObj->modProps, PR_MESSAGE_FLAGS);

	if(ulSyncId > 0 && lpPropMessageFlags) {
		if(fNewItem) {
			// Item is new, so it's not in the queue at the moment
			ulPrevSubmitFlag = 0;
		} else {
			// Existing item. Check it's current submit flag by looking at the outgoing queue
			strQuery = "SELECT hierarchy_id FROM outgoingqueue WHERE hierarchy_id=" + stringify(ulObjId) + " AND flags & " + stringify(EC_SUBMIT_MASTER) + " = 0";
			er = lpDatabase->DoSelect(strQuery, &lpDBResult);

			if(er != erSuccess)
				goto exit;

			if(lpDatabase->GetNumRows(lpDBResult) > 0) {
				// Item is in the outgoing queue at the moment
				ulPrevSubmitFlag = 1;
			} else {
				// Item is not in the queue at the moment
				ulPrevSubmitFlag = 0;
			}

			lpDatabase->FreeResult(lpDBResult);
			lpDBResult = NULL;

		}

		if((lpPropMessageFlags->Value.ul & MSGFLAG_SUBMIT) && ulPrevSubmitFlag == 0) {
			// Message was previously not submitted, but it is now, so add it to the outgoing queue and set the correct flags.

			strQuery = "INSERT INTO outgoingqueue (store_id, hierarchy_id, flags) VALUES("+stringify(ulStoreId)+", "+stringify(ulObjId)+"," + stringify(EC_SUBMIT_LOCAL) + ")";
			er = lpDatabase->DoInsert(strQuery);

			if(er != erSuccess)
				goto exit;

			strQuery = "UPDATE hierarchy SET flags = flags | " + stringify(MSGFLAG_SUBMIT) + " WHERE id = " + stringify(ulObjId);
			er = lpDatabase->DoUpdate(strQuery);

			if(er != erSuccess)
				goto exit;

			// The object has changed, update the cache.
		    g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulObjId);
			// Update in-memory outgoing tables
			g_lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_ADD, ulStoreId, ulObjId, EC_SUBMIT_LOCAL, MAPI_MESSAGE);

		} else if(((lpPropMessageFlags->Value.ul & MSGFLAG_SUBMIT) == 0) && ulPrevSubmitFlag == 1) {
			// Message was previously submitted, but is not submitted any more. Remove it from the outgoing queue and remove the flags.

			strQuery = "DELETE FROM outgoingqueue WHERE hierarchy_id = " + stringify(ulObjId);
			er = lpDatabase->DoDelete(strQuery);

			if(er != erSuccess)
				goto exit;

			strQuery = "UPDATE hierarchy SET flags = flags & ~" + stringify(MSGFLAG_SUBMIT) + " WHERE id = " + stringify(ulObjId);
			er = lpDatabase->DoUpdate(strQuery);

			if(er != erSuccess)
				goto exit;

			// The object has changed, update the cache.
		    g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulObjId);
			// Update in-memory outgoing tables
			g_lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_DELETE, ulStoreId, ulObjId, EC_SUBMIT_LOCAL, MAPI_MESSAGE);
		}
	}

	// 7. notification
	// Only Notify on MAPI_MESSAGE, MAPI_FOLDER and MAPI_STORE
	// but don't nofity if parent object is a store and object type is attachment or message
	if (ulParentObjId != 0 && ulParentObjType != MAPI_STORE) {
		er = g_lpSessionManager->GetCacheManager()->GetObject(ulParentObjId, &ulGrandParent, NULL, NULL, &ulGrandParentType);
		if(er != erSuccess)
			goto exit;
	}

	if(!((sReturnObject.ulObjType == MAPI_ATTACH || sReturnObject.ulObjType == MAPI_MESSAGE) && ulParentObjType == MAPI_STORE) &&
		(sReturnObject.ulObjType == MAPI_MESSAGE || sReturnObject.ulObjType == MAPI_FOLDER || sReturnObject.ulObjType == MAPI_STORE))

	{
		g_lpSessionManager->GetCacheManager()->GetObjectFlags(sReturnObject.ulServerId, &ulObjFlags);

		// update PR_LOCAL_COMMIT_TIME_MAX in cache for disconnected clients who want to know if the folder contents changed
		if (pvCommitTime) {
			sObjectTableKey key(ulParentObjId, 0);
			g_lpSessionManager->GetCacheManager()->SetCell(&key, PR_LOCAL_COMMIT_TIME_MAX, pvCommitTime);
		}

		if(fNewItem == true) {
			// Notify that the message has been created

			g_lpSessionManager->NotificationCreated(sReturnObject.ulObjType, sReturnObject.ulServerId, ulParentObjId);

			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_ADD, ulObjFlags & MSGFLAG_NOTIFY_FLAGS, ulParentObjId, sReturnObject.ulServerId, sReturnObject.ulObjType);

			// Notify that the folder in which the message resided has changed (PR_CONTENT_COUNT, PR_CONTENT_UNREAD)
			
			g_lpSessionManager->GetCacheManager()->UpdateCell(ulParentObjId, PR_CONTENT_COUNT, 1);
			if(lpPropMessageFlags && (lpPropMessageFlags->Value.ul & MSGFLAG_READ) == 0) {
			    // Unread message
    			g_lpSessionManager->GetCacheManager()->UpdateCell(ulParentObjId, PR_CONTENT_UNREAD, 1);
			}
			
			g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulParentObjId);
			if(ulGrandParent) {
				g_lpSessionManager->GetCacheManager()->GetObjectFlags(ulParentObjId, &ulParentFlags);
				g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, ulParentFlags & MSGFLAG_NOTIFY_FLAGS, ulGrandParent, ulParentObjId, MAPI_FOLDER);
			}

		} else if(sReturnObject.ulObjType == MAPI_STORE) {
			g_lpSessionManager->NotificationModified(sReturnObject.ulObjType, sReturnObject.ulServerId);
		} else {
			// Notify that the message has been modified
			if(sReturnObject.ulObjType == MAPI_MESSAGE)
				g_lpSessionManager->NotificationModified(sReturnObject.ulObjType, sReturnObject.ulServerId, ulParentObjId);
			else
				g_lpSessionManager->NotificationModified(sReturnObject.ulObjType, sReturnObject.ulServerId);

			if(ulParentObjId)
				g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, ulObjFlags & MSGFLAG_NOTIFY_FLAGS, ulParentObjId, sReturnObject.ulServerId, sReturnObject.ulObjType);

			// Notify that the folder in which the message resided has changed (PR_CONTENT_COUNT, PR_SUBFOLDERS and maybe unread count)
			if(ulParentObjId) {
				g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulParentObjId);
				g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulParentObjId);
			}
			if(ulGrandParent && ulParentObjId) {
				g_lpSessionManager->GetCacheManager()->GetObjectFlags(ulParentObjId, &ulParentFlags);
				g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, ulParentFlags & MSGFLAG_NOTIFY_FLAGS, ulGrandParent, ulParentObjId, MAPI_FOLDER);
			}
		}
	}

	lpsLoadObjectResponse->sSaveObject = sReturnObject;

exit:
	ROLLBACK_ON_ERROR();
}
SOAP_ENTRY_END()

ECRESULT LoadObject(struct soap *soap, ECSession *lpecSession, unsigned int ulObjId, unsigned int ulObjType, unsigned int ulParentObjType, struct saveObject *lpsSaveObj, std::map<unsigned int, CHILDPROPS> *lpChildProps)
{
	ECRESULT 		er = erSuccess;
	ECAttachmentStorage *lpAttachmentStorage = NULL;
	ULONG			ulInstanceId = 0;
	ULONG			ulInstanceTag = 0;
	struct saveObject sSavedObject;
	int i;
	GUID			sGuidServer;
	std::map<unsigned int, CHILDPROPS> mapChildProps;
	std::map<unsigned int, CHILDPROPS>::iterator iterProps;
	USE_DATABASE();
	CHILDPROPS		sEmptyProps;
	
	sEmptyProps.lpPropVals = new DynamicPropValArray(soap);
	sEmptyProps.lpPropTags = new DynamicPropTagArray(soap);

	memset(&sSavedObject, 0, sizeof(saveObject));

	// Check permission
	if (ulObjType == MAPI_STORE || (ulObjType == MAPI_FOLDER && ulParentObjType == MAPI_STORE))
		// Always read rights on the store and the root folder
		er = erSuccess;
	else if (ulObjType == MAPI_FOLDER)
		er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityFolderVisible);
	else if (ulObjType == MAPI_MESSAGE)
		er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityRead);
		
    // Allow reading MAPI_MAILUSER and MAPI_ATTACH since that is only called internally

	if (er != erSuccess)
		goto exit;

	sSavedObject.ulClientId = 0;
	sSavedObject.ulServerId = ulObjId;
	sSavedObject.ulObjType = ulObjType;
	
	if(lpChildProps == NULL) {
	    // We were not provided with a property list for this object, get our own now.
	    er = PrepareReadProps(soap, lpecSession, ulObjId, 0, &mapChildProps);
	    if(er != erSuccess)
	        goto exit;
	        
        lpChildProps = &mapChildProps;
    }
	
	iterProps = lpChildProps->find(ulObjId);
	if(iterProps == lpChildProps->end())
    	er = ReadProps(soap, lpecSession, ulObjId, ulObjType, ulParentObjType, sEmptyProps, &sSavedObject.delProps, &sSavedObject.modProps);
    else
    	er = ReadProps(soap, lpecSession, ulObjId, ulObjType, ulParentObjType, iterProps->second, &sSavedObject.delProps, &sSavedObject.modProps);

	if (er != erSuccess)
		goto exit;
		
    FreeChildProps(&mapChildProps);
    

	if (ulObjType == MAPI_MESSAGE || ulObjType == MAPI_ATTACH) {
        // Pre-load *all* properties of *all* subobjects for fast accessibility
        er = PrepareReadProps(soap, lpecSession, 0, ulObjId, &mapChildProps);
        if (er != erSuccess)
            goto exit;

		// find subobjects
		strQuery = "SELECT id, type FROM hierarchy WHERE parent="+stringify(ulObjId);
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		sSavedObject.__size = lpDatabase->GetNumRows(lpDBResult);
		sSavedObject.__ptr = s_alloc<saveObject>(soap, sSavedObject.__size);
		memset(sSavedObject.__ptr, 0, sizeof(saveObject) * sSavedObject.__size);

		for (i = 0; i < sSavedObject.__size; i++) {
		    unsigned int ulChildId;
			lpDBRow = lpDatabase->FetchRow(lpDBResult);
			lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

			if(lpDBRow == NULL || lpDBLen == NULL) {
				er = ZARAFA_E_DATABASE_ERROR; // this should never happen
				goto exit;
			}
			
			ulChildId = atoi(lpDBRow[0]);
			
   			LoadObject(soap, lpecSession, atoi(lpDBRow[0]), atoi(lpDBRow[1]), ulObjType, &sSavedObject.__ptr[i], &mapChildProps);
		}
    	FreeChildProps(&mapChildProps);
	}

	if (ulObjType == MAPI_ATTACH) {
		er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
		if (er != erSuccess)
			goto exit;

		er = g_lpSessionManager->GetServerGUID(&sGuidServer);
		if (er != erSuccess)
			goto exit;

		/* Not having a single instance ID is not critical, we might create it at a later time */
		if (lpAttachmentStorage->GetSingleInstanceId(ulObjId, &ulInstanceId, &ulInstanceTag) == erSuccess) {
			sSavedObject.lpInstanceIds = s_alloc<entryList>(soap);
			sSavedObject.lpInstanceIds->__size = 1;
			sSavedObject.lpInstanceIds->__ptr = s_alloc<entryId>(soap, sSavedObject.lpInstanceIds->__size);

			er = SIIDToEntryID(soap, &sGuidServer, ulInstanceId, ulInstanceTag, &sSavedObject.lpInstanceIds->__ptr[0]);
			if (er != erSuccess) {
				sSavedObject.lpInstanceIds->__size = NULL;
				goto exit;
			}
		}
	}

	*lpsSaveObj = sSavedObject;

exit:
	if (lpAttachmentStorage)
		delete lpAttachmentStorage;
		
    if (sEmptyProps.lpPropVals)
        delete sEmptyProps.lpPropVals;
    if (sEmptyProps.lpPropTags)
        delete sEmptyProps.lpPropTags;

	FREE_DBRESULT();
__soapentry_exit:
	return er;
}

SOAP_ENTRY_START4(loadObject, lpsLoadObjectResponse->er, entryId sEntryId, struct notifySubscribe *lpsNotSubscribe, unsigned int ulFlags, struct loadObjectResponse *lpsLoadObjectResponse)
{
	unsigned int	ulObjId = 0;
	unsigned int	ulObjFlags = 0;
	unsigned int	ulObjType = 0;
	unsigned int	ulParentId = 0;
	unsigned int	ulParentObjType = 0;
	
	struct saveObject sSavedObject;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if (er != erSuccess)
		goto exit;
		
    er = g_lpSessionManager->GetCacheManager()->GetObject(ulObjId, &ulParentId, NULL, &ulObjFlags, &ulObjType);
    if (er != erSuccess)
        goto exit;
    
    if(ulObjType == MAPI_STORE)
        ulParentObjType = 0;
    else if(ulObjType == MAPI_MESSAGE)
        ulParentObjType = MAPI_FOLDER;
    else if(ulObjType == MAPI_FOLDER) {

		// Disable searchfolder for delegate stores without admin permissions
		if ( (ulObjFlags&FOLDER_SEARCH) && lpecSession->GetSecurity()->GetAdminLevel() == 0 && 
			lpecSession->GetSecurity()->IsStoreOwner(ulObjId) != erSuccess)
		{
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

        er = g_lpSessionManager->GetCacheManager()->GetObject(ulParentId, NULL, NULL, NULL, &ulParentObjType);
		if (er != erSuccess)
			goto exit;
    }
        
    if(ulFlags & 0x80000000) {
        // Flags passed by client, check object flags
        ulFlags = ulFlags & ~0x80000000;
        
        if((ulObjFlags & MSGFLAG_DELETED) != ulFlags) {
            er = ZARAFA_E_NOT_FOUND;
            goto exit;
        }
    }

	// Subscribe for notification
	if (lpsNotSubscribe) {
		er = DoNotifySubscribe(lpecSession, ulSessionId, lpsNotSubscribe);
		if (er != erSuccess)
			goto exit;
	}

	er = LoadObject(soap, lpecSession, ulObjId, ulObjType, ulParentObjType, &sSavedObject, NULL);
	if (er != erSuccess)
		goto exit;

	lpsLoadObjectResponse->sSaveObject = sSavedObject;

exit:
	;
}
SOAP_ENTRY_END()

// if lpsNewEntryId is NULL this function create a new entryid
// if lpsOrigSourceKey is NULL this function creates a new sourcekey
ECRESULT CreateFolder(ECSession *lpecSession, ECDatabase *lpDatabase, unsigned int ulParentId, entryId *lpsNewEntryId, unsigned int type, char *name, char *comment, bool openifexists, bool bNotify, unsigned int ulSyncId, struct xsd__base64Binary *lpsOrigSourceKey, unsigned int *lpFolderId, bool *lpbExist)
{
	ECRESULT		er = erSuccess;
	ALLOC_DBRESULT();

	unsigned int	ulFolderId = 0;
	unsigned int	ulLastId = 0;
	unsigned int	ulStoreId = 0;
	unsigned int	ulGrandParent = 0;
	bool			bExist = false;
	bool			bFreeNewEntryId = false;
	GUID			guid;
	SOURCEKEY		sSourceKey;

	// You're only allowed to create search folders in your own store
	if(type == FOLDER_SEARCH) {
	    if(lpecSession->GetSecurity()->IsObjectAdmin(ulParentId) != erSuccess && (lpecSession->GetSecurity()->IsStoreOwner(ulParentId) != erSuccess)) {
	        er = ZARAFA_E_NO_ACCESS;
	        goto exit;
        }
    }

	er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulParentId, &ulStoreId, &guid);
	if(er != erSuccess)
	    goto exit;

	// Check whether the requested name already exists

	strQuery = "SELECT hierarchy.id FROM hierarchy JOIN properties ON hierarchy.id = properties.hierarchyid WHERE hierarchy.parent=" + stringify(ulParentId) + " AND hierarchy.type="+stringify(MAPI_FOLDER)+" AND hierarchy.flags & " + stringify(MSGFLAG_DELETED)+ "=0 AND properties.tag=" + stringify(ZARAFA_TAG_DISPLAY_NAME) + " AND properties.val_string = '" + lpDatabase->Escape(name) + "' AND properties.storeid=" + stringify(ulStoreId);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) > 0 && !ulSyncId) {
		// Check folder read access
		er = lpecSession->GetSecurity()->CheckPermission(ulParentId, ecSecurityFolderVisible);
		if(er != erSuccess)
			goto exit;

		// Object exists
		if(openifexists) {
			lpDBRow = lpDatabase->FetchRow(lpDBResult);

			if(lpDBRow == NULL || lpDBRow[0] == NULL) {
				er = ZARAFA_E_DATABASE_ERROR;
				goto exit;
			}

			ulFolderId = atoi(lpDBRow[0]);
			bExist = true;

		} else {
			er = ZARAFA_E_COLLISION;
			goto exit;
		}
	} else {
		// Check write permission of the folder destination
		er = lpecSession->GetSecurity()->CheckPermission(ulParentId, ecSecurityCreateFolder);
		if(er != erSuccess)
			goto exit;

		// Create folder
		strQuery = "INSERT INTO hierarchy (parent, type, createtime, modtime, flags, owner) values(" + stringify(ulParentId) + "," + stringify(ZARAFA_OBJTYPE_FOLDER) + ", CURRENT_TIMESTAMP, CURRENT_TIMESTAMP, " + stringify(type) + ", "+stringify(lpecSession->GetSecurity()->GetUserId(ulParentId))+")";
		er = lpDatabase->DoInsert(strQuery, &ulLastId);
		if(er != erSuccess)
			goto exit;

		if(lpsNewEntryId == NULL) {
			er = CreateEntryId(guid, MAPI_FOLDER, &lpsNewEntryId);
			if(er != erSuccess)
				goto exit;

			bFreeNewEntryId = true;
		}

		//Create entryid, 0x0FFF = PR_ENTRYID
		strQuery = "INSERT INTO indexedproperties (hierarchyid,tag,val_binary) VALUES("+stringify(ulLastId)+", 0x0FFF, "+lpDatabase->EscapeBinary(lpsNewEntryId->__ptr, lpsNewEntryId->__size)+")";
		er = lpDatabase->DoInsert(strQuery);
		if(er != erSuccess)
			goto exit;

		// Create Displayname
		strQuery = "INSERT INTO properties (hierarchyid, tag, type, storeid, val_string) values(" + stringify(ulLastId) + "," + stringify(ZARAFA_TAG_DISPLAY_NAME) + "," + stringify(PT_STRING8) + "," + stringify(ulStoreId) + ",'" + lpDatabase->Escape(name) + "')";
		er = lpDatabase->DoInsert(strQuery);
		if(er != erSuccess)
			goto exit;

		if (comment) {
			strQuery = "INSERT INTO properties (hierarchyid, tag, type, storeid, val_string) values(" + stringify(ulLastId) + "," + stringify(PROP_ID(PR_COMMENT)) + "," + stringify(PT_STRING8) + "," + stringify(ulStoreId) + ",'" + lpDatabase->Escape(comment) + "')";
			er = lpDatabase->DoInsert(strQuery);
			if(er != erSuccess)
				goto exit;
		}

		// Create SourceKey
		if (lpsOrigSourceKey && lpsOrigSourceKey->__size > (int)sizeof(GUID) && lpsOrigSourceKey->__ptr){
			sSourceKey = SOURCEKEY(lpsOrigSourceKey->__size, (char*)lpsOrigSourceKey->__ptr);
		}else{
			er = lpecSession->GetNewSourceKey(1, &sSourceKey);
			if(er != erSuccess)
				goto exit;
		}

		strQuery = "INSERT INTO indexedproperties(hierarchyid,tag,val_binary) VALUES(" + stringify(ulLastId) + "," + stringify(PROP_ID(PR_SOURCE_KEY)) + "," + lpDatabase->EscapeBinary(sSourceKey, sSourceKey.size()) + ")";
		er = lpDatabase->DoInsert(strQuery);
		if(er != erSuccess)
			goto exit;

		ulFolderId = ulLastId;
	}

	if(bExist == false && !(type & FOLDER_SEARCH)){
		SOURCEKEY sParentSourceKey;
		
		GetSourceKey(ulParentId, &sParentSourceKey);
		AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_FOLDER_CHANGE);
	}

	// Notify that the folder has been created
	if(bExist == false && bNotify == true) {
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulParentId);
		g_lpSessionManager->NotificationCreated(MAPI_FOLDER, ulFolderId, ulParentId);
		g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulParentId);

		// Update all tables viewing this folder
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_ADD, 0, ulParentId, ulFolderId, MAPI_FOLDER);


		// Update notification, grandparent of the mainfolder
		g_lpSessionManager->GetCacheManager()->GetParent(ulParentId, &ulGrandParent);
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParent, ulParentId, MAPI_FOLDER);
	}

	if(lpFolderId)
		*lpFolderId = ulFolderId;

	if(lpbExist)
		*lpbExist = bExist;

	er = g_lpSessionManager->SaveSourceKeyAutoIncrement();
	if (er != erSuccess)
		goto exit;

exit:
	FREE_DBRESULT();

	if(bFreeNewEntryId == true && lpsNewEntryId)
		FreeEntryId(lpsNewEntryId, true);

	return er;
}

/**
 * createFolder: Create a folder object in the hierarchy table, and add a 'PR_DISPLAY_NAME' property.
 *
 * The data model actually supports multiple folders having the same PR_DISPLAY_NAME, however, MAPI does not,
 * so we have to give the engine the knowledge of the PR_DISPLAY_NAME property here, one of the few properties
 * that the backend engine actually knows about.
 *
 * Of course, the frontend could also enforce this constraint, but that would require 2 server accesses (1. check
 * existing, 2. create folder), and we're trying to keep the amount of server accesses as low as possible.
 *
 */

SOAP_ENTRY_START9(createFolder, lpsResponse->er, entryId sParentId, entryId* lpsNewEntryId, unsigned int ulType, char *szName, char *szComment, bool fOpenIfExists, unsigned int ulSyncId, struct xsd__base64Binary sOrigSourceKey, struct createFolderResponse *lpsResponse)
{
	unsigned int	ulParentId = 0;
	unsigned int	ulFolderId = 0;
	USE_DATABASE();

	if(szName == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpDatabase->Begin();
	if (er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sParentId, &ulParentId);
	if (er != erSuccess)
		goto exit;

	er = CreateFolder(lpecSession, lpDatabase, ulParentId, lpsNewEntryId, ulType, szName, szComment, fOpenIfExists, true, ulSyncId, &sOrigSourceKey, &ulFolderId, NULL);
	if (er != erSuccess)
		goto exit;

	er = lpDatabase->Commit();
	if (er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(ulFolderId, soap, &lpsResponse->sEntryId);
	if (er != erSuccess)
		goto exit;

exit:
	ROLLBACK_ON_ERROR();
}
SOAP_ENTRY_END()


ECRESULT GetFolderCount(ECSession *lpecSession, unsigned int ulObjId, unsigned int *lpulContentCount, unsigned int *lpulUnreadCount,
						unsigned int *lpulAssocCount, unsigned int *lpulSubFolderCount, unsigned int *lpulDeleteMsgCount,
						unsigned int *lpulDeleteFolderCount, unsigned int *lpulDeleteAssocMsgCount)
{
	ECRESULT		er = erSuccess;
	ECDatabase		*lpDatabase = NULL;
	unsigned int	ulFlags = 0;
	ALLOC_DBRESULT();

	lpDatabase = lpecSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	// Get folder type
	er = lpecSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, NULL, NULL, &ulFlags, NULL);
	if(er != erSuccess)
		goto exit;

	strQuery = "SELECT ";
	if (ulFlags == 2) { // Searchfolders

		// Get the amount of messages in the folder (PR_CONTENT_COUNT)
		strQuery += "count(*) as cc, " ; 

		// Get the amount of unread messages in the folder (PR_CONTENT_UNREAD)
		strQuery += "count(if(searchresults.flags & " + stringify(MSGFLAG_DELETED | MSGFLAG_ASSOCIATED | MSGFLAG_READ) + " = 0,1,null)) AS cu,";

		// The following properties are always 0 for searchfolders
		strQuery += "0 AS cac,";	// PR_ASSOC_CONTENT_COUNT
		strQuery += "0 AS sf,";		// PR_SUBFOLDERS
		strQuery += "0 AS dmc,";	// PR_DELETED_MSG_COUNT
		strQuery += "0 AS dfc,";	// PR_DELETED_FOLDER_COUNT
		strQuery += "0 AS damc";	// PR_DELETED_ASSOC_MSG_COUNT
		strQuery += " FROM searchresults WHERE searchresults.folderid=" + stringify(ulObjId);
	}else {
		// Get the amount of messages in the folder (PR_CONTENT_COUNT)
		strQuery += "count(if(hierarchy.type=5 AND hierarchy.flags & " + stringify(MSGFLAG_ASSOCIATED | MSGFLAG_DELETED) + " = 0,1,null)) AS cc,";

		// Get the amount of unread messages in the folder (PR_CONTENT_UNREAD)
		strQuery += "count(if(hierarchy.type=5 AND hierarchy.flags & " + stringify(MSGFLAG_DELETED | MSGFLAG_ASSOCIATED | MSGFLAG_READ) + " = 0,1,null)) AS cu,";

		// Get the amount of messages with flags '0x40' in the folder (associated messages)(PR_ASSOC_CONTENT_COUNT)
        strQuery += "count(if(hierarchy.type=5 AND hierarchy.flags & " + stringify(MSGFLAG_ASSOCIATED) + " = " + stringify(MSGFLAG_ASSOCIATED) + " AND hierarchy.flags & " + stringify(MSGFLAG_DELETED) + " = 0,1,null)) AS cac,";

		// Get the amount of folders in the folder (subfolders) (PR_SUBFOLDERS)
        strQuery += "count(if(hierarchy.type=3 AND hierarchy.flags & " + stringify(MSGFLAG_DELETED) + " = 0,1,null)) AS sf,";

        // Get the amount of deleted messages in the folder (PR_DELETED_MSG_COUNT)
        strQuery += "count(if(hierarchy.type=5 AND hierarchy.flags & " + stringify(MSGFLAG_ASSOCIATED) + " = 0 AND hierarchy.flags & " + stringify(MSGFLAG_DELETED) + " = " + stringify(MSGFLAG_DELETED)+",1,null)) AS dmc,";

        // Get the amount of deleted folders (PR_DELETED_FOLDER_COUNT)
        strQuery += "count(if(hierarchy.type=3 AND hierarchy.flags & " + stringify(MSGFLAG_DELETED) + " = " + stringify(MSGFLAG_DELETED)+",1,null)) AS dfc,";

        // Get the amount of messages with flags '0x440' in the folder (deleted associated messages) (PR_DELETED_ASSOC_MSG_COUNT)
        strQuery += "count(if(hierarchy.type=5 AND hierarchy.flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " = " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED)+",1,null)) AS damc";
        
        strQuery += " FROM hierarchy WHERE hierarchy.parent=" + stringify(ulObjId);
	}

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) !=1 ) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);

	if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL || lpDBRow[3] == NULL ||
		lpDBRow[4] == NULL || lpDBRow[5] == NULL || lpDBRow[6] == NULL)
	{
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	*lpulContentCount = atoui(lpDBRow[0]);
	*lpulUnreadCount = atoui(lpDBRow[1]);
	*lpulAssocCount = atoui(lpDBRow[2]);
	*lpulSubFolderCount = atoui(lpDBRow[3]);
	*lpulDeleteMsgCount = atoui(lpDBRow[4]);
	*lpulDeleteFolderCount = atoui(lpDBRow[5]);
	*lpulDeleteAssocMsgCount = atoui(lpDBRow[6]);

	//Free database results
	FREE_DBRESULT();
exit:

	//Free database results
	FREE_DBRESULT();

	return er;
}

/**
 * getFolderCount: Get various counts for the folder specified by sEntryId. The counts are the amount of messsages in the folder. On
 * search folders, some counts are unavailable, like deleted messages count and associated messages.
 */
SOAP_ENTRY_START2(getFolderCount, lpsResponse->er, entryId sEntryId, struct getFolderCountResponse *lpsResponse)
{
	unsigned int	ulObjId = 0;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if(er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityFolderVisible);
	if(er != erSuccess)
		goto exit;

	memset(lpsResponse, 0, sizeof(struct getFolderCountResponse) );


	er = GetFolderCount(lpecSession, ulObjId, &lpsResponse->ulContentCount, &lpsResponse->ulUnreadCount,
						&lpsResponse->ulAssocCount, &lpsResponse->ulSubFolderCount, &lpsResponse->ulDeleteMsgCount,
						&lpsResponse->ulDeleteFolderCount, &lpsResponse->ulDeleteAssocMsgCount);
	if(er != erSuccess)
		goto exit;


exit:
	;
}
SOAP_ENTRY_END()


/**
 tableOpen: Open a mapi table

ulTableType
	TABLETYPE_MS
		For all tables of a messagestore.
	TABLETYPE_AB
		For the addressbook tables.
	TABLETYPE_SPOOLER
		For the spooler tables, the sEntryId must always the store entryid.
		ulType and ulFlags are ignored, reserved for future use.
ulType
	The type of the object you want to open.
*/
ECRESULT OpenTable(ECSession *lpecSession, entryId sEntryId, unsigned int ulTableType, unsigned int ulType, unsigned int ulFlags, unsigned int *lpulTableId)
{
    ECRESULT er = erSuccess;
	unsigned int	ulTableId = 0;
	unsigned int	ulId = 0;
	unsigned int	ulTypeId = 0;

	switch (ulTableType) {
		case TABLETYPE_MS:
			if(ulFlags & SHOW_SOFT_DELETES)
			{
				ulFlags &=~SHOW_SOFT_DELETES;
				ulFlags |= MSGFLAG_DELETED;
			}

			er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulId);
			if(er != erSuccess)
				goto exit;

			er = lpecSession->GetTableManager()->OpenGenericTable(ulId, ulType, ulFlags, &ulTableId);
			if( er != erSuccess)
				goto exit;

			break;
		case TABLETYPE_AB:
			er = ABEntryIDToID(&sEntryId, &ulId, &ulTypeId);
			if(er != erSuccess)
				goto exit;

			er = lpecSession->GetTableManager()->OpenABTable(ulId, ulTypeId, ulType, ulFlags, &ulTableId);
			if( er != erSuccess)
				goto exit;

			break;
		case TABLETYPE_SPOOLER:
			// sEntryId must be a store entryid or zero for all stores
			if(sEntryId.__size > 0 ) {
				er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulId);
				if(er != erSuccess)
					goto exit;
			}else
				ulId = 0; //All stores

			er = lpecSession->GetTableManager()->OpenOutgoingQueueTable(ulId, &ulTableId);
			if( er != erSuccess)
				goto exit;

			break;
		case TABLETYPE_MULTISTORE:
			er = lpecSession->GetTableManager()->OpenMultiStoreTable(ulType, ulFlags, &ulTableId);
			if( er != erSuccess)
				goto exit;
			break;
		case TABLETYPE_USERSTORES:
			er = lpecSession->GetTableManager()->OpenUserStoresTable(ulFlags, &ulTableId);
			if( er != erSuccess)
				goto exit;
			break;
		case TABLETYPE_STATS_SYSTEM:
		case TABLETYPE_STATS_SESSIONS:
		case TABLETYPE_STATS_USERS:
		case TABLETYPE_STATS_COMPANY:
			er = lpecSession->GetTableManager()->OpenStatsTable(ulTableType, &ulTableId);
			if (er != erSuccess)
				goto exit;
			break;
		default:
			er = ZARAFA_E_BAD_VALUE;
			goto exit;
			break; //Happy compiler
	} // switch (ulTableType)

	*lpulTableId = ulTableId;
exit:
	return er;
}

SOAP_ENTRY_START5(tableOpen, lpsTableOpenResponse->er, entryId sEntryId, unsigned int ulTableType, unsigned ulType, unsigned int ulFlags, struct tableOpenResponse *lpsTableOpenResponse)
{
    unsigned int ulTableId = 0;
    
    er = OpenTable(lpecSession, sEntryId, ulTableType, ulType, ulFlags, &ulTableId);
    if(er != erSuccess)
        goto exit;
        
	lpsTableOpenResponse->ulTableId = ulTableId;
exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableClose: close the table with the specified table ID
 */
SOAP_ENTRY_START2(tableClose, *result, unsigned int ulTableId, unsigned int *result)
{
	er = lpecSession->GetTableManager()->CloseTable(ulTableId);
}
SOAP_ENTRY_END()

/**
 * tableSetSearchCritieria: set search criteria for a searchfolder
 */
SOAP_ENTRY_START5(tableSetSearchCriteria, *result, entryId sEntryId, struct restrictTable *lpRestrict, struct entryList *lpFolders, unsigned int ulFlags, unsigned int *result)
{
	unsigned int	ulStoreId = 0;
	unsigned int	ulParent = 0;

	if(!(ulFlags & STOP_SEARCH) && (lpRestrict == NULL || lpFolders == NULL)) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulParent);
	if(er != erSuccess)
	    goto exit;

    er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulParent, &ulStoreId, NULL);
    if(er != erSuccess)
        goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulParent, ecSecurityEdit);
	if(er != erSuccess)
		goto exit;

	// If a STOP was requested, then that's all we need to do
	if(ulFlags & STOP_SEARCH) {
		er = lpecSession->GetSessionManager()->GetSearchFolders()->SetSearchCriteria(ulStoreId, ulParent, NULL);
	} else {
		struct searchCriteria sSearchCriteria;

		sSearchCriteria.lpRestrict = lpRestrict;
		sSearchCriteria.lpFolders = lpFolders;
		sSearchCriteria.ulFlags = ulFlags;

		er = lpecSession->GetSessionManager()->GetSearchFolders()->SetSearchCriteria(ulStoreId, ulParent, &sSearchCriteria);
	}

exit:
	;
}
SOAP_ENTRY_END()

/**
 * tableGetSearchCriteria: get the search criteria for a searchfolder previously called with tableSetSearchCriteria
 */
SOAP_ENTRY_START2(tableGetSearchCriteria, lpsResponse->er,  entryId sEntryId, struct tableGetSearchCriteriaResponse *lpsResponse)
{
	unsigned int	ulFlags = 0;
	unsigned int	ulStoreId = 0;
	unsigned int	ulId = 0;
	struct searchCriteria *lpSearchCriteria = NULL;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulId);
	if(er != erSuccess)
	    goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetStore(ulId, &ulStoreId, NULL);
	if(er != erSuccess)
	    goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulId, ecSecurityRead);
	if(er != erSuccess)
		goto exit;

	er = lpecSession->GetSessionManager()->GetSearchFolders()->GetSearchCriteria(ulStoreId, ulId, &lpSearchCriteria, &ulFlags);
	if(er != erSuccess)
		goto exit;

	er = CopyRestrictTable(soap, lpSearchCriteria->lpRestrict, &lpsResponse->lpRestrict);
	if(er != erSuccess)
		goto exit;

	er = CopyEntryList(soap, lpSearchCriteria->lpFolders, &lpsResponse->lpFolderIDs);
	if(er != erSuccess)
		goto exit;

	lpsResponse->ulFlags = ulFlags;

exit:
	if(lpSearchCriteria)
		FreeSearchCriteria(lpSearchCriteria);
}
SOAP_ENTRY_END()

/**
 * tableSetColumns: called from IMAPITable::SetColumns()
 */
SOAP_ENTRY_START3(tableSetColumns, *result, unsigned int ulTableId, struct propTagArray *aPropTag, unsigned int *result)
{
	ECGenericObjectTable	*lpTable = NULL;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);
	if(er != erSuccess)
		goto exit;

	er = lpTable->SetColumns(aPropTag);

exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableQueryColumns: called from IMAPITable::GetColumns()
 */
SOAP_ENTRY_START3(tableQueryColumns, lpsResponse->er, unsigned int ulTableId, unsigned int ulFlags, struct tableQueryColumnsResponse *lpsResponse)
{
	ECGenericObjectTable	*lpTable = NULL;

	struct propTagArray *lpPropTags = NULL;

	// Init
	lpsResponse->sPropTagArray.__size = 0;
	lpsResponse->sPropTagArray.__ptr = NULL;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	er = lpTable->GetColumns(soap, ulFlags, &lpPropTags);

	if(er != erSuccess)
		goto exit;

	lpsResponse->sPropTagArray.__size = lpPropTags->__size;
	lpsResponse->sPropTagArray.__ptr = lpPropTags->__ptr;

exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableRestrict: called from IMAPITable::Restrict()
 */
SOAP_ENTRY_START3(tableRestrict, *result, unsigned int ulTableId, struct restrictTable *lpsRestrict, unsigned int *result)
{
	ECGenericObjectTable	*lpTable = NULL;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	// NOTE: This is a workaround for version 4.1 backward compatibility
	if(lpsRestrict)
		BackwardCompRestrict4_1(lpsRestrict);

	er = lpTable->Restrict(lpsRestrict);

exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableSort: called from IMAPITable::Sort()
 */
SOAP_ENTRY_START5(tableSort, *result, unsigned int ulTableId, struct sortOrderArray *lpSortOrder, unsigned int ulCategories, unsigned int ulExpanded, unsigned int *result)
{
	ECGenericObjectTable	*lpTable = NULL;

	if(lpSortOrder == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	er = lpTable->SetSortOrder(lpSortOrder, ulCategories, ulExpanded);

exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableQueryRows: called from IMAPITable::QueryRows()
 */
SOAP_ENTRY_START4(tableQueryRows, lpsResponse->er, unsigned int ulTableId, unsigned int ulRowCount, unsigned int ulFlags, struct tableQueryRowsResponse *lpsResponse)
{
	ECGenericObjectTable	*lpTable = NULL;
	struct rowSet	*lpRowSet = NULL;

	lpsResponse->sRowSet.__ptr = NULL;
	lpsResponse->sRowSet.__size = 0;

	// Get the table
	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);
	if(er != erSuccess)
		goto exit;

	// FIXME: Check permission

	er = lpTable->QueryRows(soap, ulRowCount, ulFlags, &lpRowSet);
	if(er != erSuccess)
		goto exit;

	lpsResponse->sRowSet.__ptr = lpRowSet->__ptr;
	lpsResponse->sRowSet.__size = lpRowSet->__size;
exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableGetRowCount: called from IMAPITable::GetRowCount()
 */
SOAP_ENTRY_START2(tableGetRowCount, lpsResponse->er, unsigned int ulTableId, struct tableGetRowCountResponse *lpsResponse)
{
	ECGenericObjectTable	*lpTable = NULL;

	//FIXME: security? give rowcount 0 is failed ?
	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	er = lpTable->GetRowCount(&lpsResponse->ulCount, &lpsResponse->ulRow);

exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableSeekRow: called from IMAPITable::SeekRow()
 */
SOAP_ENTRY_START4(tableSeekRow, lpsResponse->er, unsigned int ulTableId , unsigned int ulBookmark, int lRows, struct tableSeekRowResponse *lpsResponse)
{
	ECGenericObjectTable	*lpTable = NULL;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	er = lpTable->SeekRow(ulBookmark, lRows, &lpsResponse->lRowsSought);

exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableFindRow: called from IMAPITable::FindRow()
 */
SOAP_ENTRY_START5(tableFindRow, *result, unsigned int ulTableId ,unsigned int ulBookmark, unsigned int ulFlags, struct restrictTable *lpsRestrict, unsigned int *result)
{
	ECGenericObjectTable	*lpTable = NULL;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	// NOTE: This is a workaround for version 4.1 backward compatibility
	if(lpsRestrict)
		BackwardCompRestrict4_1(lpsRestrict);

	er = lpTable->FindRow(lpsRestrict, ulBookmark, ulFlags);

exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableCreateBookmark: called from IMAPITable::CreateBookmark()
 */
SOAP_ENTRY_START2(tableCreateBookmark, lpsResponse->er, unsigned int ulTableId, struct tableBookmarkResponse *lpsResponse)
{
	ECGenericObjectTable	*lpTable = NULL;
	unsigned int ulbkPosition = 0;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	er = lpTable->CreateBookmark(&ulbkPosition);
	if(er != erSuccess)
		goto exit;

	lpsResponse->ulbkPosition = ulbkPosition;

exit:
	;
}
SOAP_ENTRY_END()

/**
 * tableCreateBookmark: called from IMAPITable::FreeBookmark()
 */
SOAP_ENTRY_START3(tableFreeBookmark, *result, unsigned int ulTableId, unsigned int ulbkPosition, unsigned int *result)
{
	ECGenericObjectTable	*lpTable = NULL;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	er = lpTable->FreeBookmark(ulbkPosition);
	if(er != erSuccess)
		goto exit;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(tableExpandRow, lpsResponse->er, unsigned int ulTableId, xsd__base64Binary sInstanceKey, unsigned int ulRowCount, unsigned int ulFlags, tableExpandRowResponse* lpsResponse)
{
	ECGenericObjectTable	*lpTable = NULL;
	struct rowSet	*lpRowSet = NULL;
	unsigned int ulMoreRows = 0;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	er = lpTable->ExpandRow(soap, sInstanceKey, ulRowCount, ulFlags, &lpRowSet, &ulMoreRows);
	if(er != erSuccess)
		goto exit;

    lpsResponse->ulMoreRows = ulMoreRows;
    lpsResponse->rowSet = *lpRowSet;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START4(tableCollapseRow, lpsResponse->er, unsigned int ulTableId, xsd__base64Binary sInstanceKey, unsigned int ulFlags, tableCollapseRowResponse* lpsResponse)
{
	ECGenericObjectTable	*lpTable = NULL;
	unsigned int ulRows = 0;

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);

	if(er != erSuccess)
		goto exit;

	er = lpTable->CollapseRow(sInstanceKey, ulFlags, &ulRows);
	if(er != erSuccess)
		goto exit;

    lpsResponse->ulRows = ulRows;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(tableGetCollapseState, lpsResponse->er, unsigned int ulTableId, struct xsd__base64Binary sBookmark, tableGetCollapseStateResponse *lpsResponse)
{
    ECGenericObjectTable *lpTable = NULL;

    er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);
    if(er != erSuccess)
        goto exit;

    er = lpTable->GetCollapseState(soap, sBookmark, &lpsResponse->sCollapseState);
    if(er != erSuccess)
        goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(tableSetCollapseState, lpsResponse->er, unsigned int ulTableId, struct xsd__base64Binary sCollapseState, struct tableSetCollapseStateResponse *lpsResponse);
{
    ECGenericObjectTable *lpTable = NULL;

    er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);
    if(er != erSuccess)
        goto exit;

    er = lpTable->SetCollapseState(sCollapseState, &lpsResponse->ulBookmark);
    if(er != erSuccess)
        goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

/**
 * tableSetMultiStoreEntryIDs: client sets content for MultiStoreTable
 */
SOAP_ENTRY_START3(tableSetMultiStoreEntryIDs, *result, unsigned int ulTableId, struct entryList *lpEntryList, unsigned int *result)
{
	ECGenericObjectTable	*lpTable = NULL;
	ECMultiStoreTable		*lpMultiStoreTable = NULL;
	ECListInt	lObjectList;


	if(lpEntryList == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);
	if (er != erSuccess)
		goto exit;

	// ignore errors
	g_lpSessionManager->GetCacheManager()->GetEntryListToObjectList(lpEntryList, &lObjectList);

	lpMultiStoreTable = dynamic_cast<ECMultiStoreTable*>(lpTable);
	if (!lpMultiStoreTable) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpMultiStoreTable->SetEntryIDs(&lObjectList);
	if(er != erSuccess)
	    goto exit;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(tableMulti, lpsResponse->er, struct tableMultiRequest sRequest, struct tableMultiResponse *lpsResponse)
{
    unsigned int ulTableId = sRequest.ulTableId;
    ECGenericObjectTable *lpTable = NULL;
    struct rowSet *lpRowSet = NULL;
    
    if(sRequest.lpOpen) {
        er = OpenTable(lpecSession, sRequest.lpOpen->sEntryId, sRequest.lpOpen->ulTableType, sRequest.lpOpen->ulType, sRequest.lpOpen->ulFlags, &lpsResponse->ulTableId);
        ulTableId = lpsResponse->ulTableId;
    }
    
	er = lpecSession->GetTableManager()->GetTable(ulTableId, &lpTable);
	if(er != erSuccess)
		goto exit;

    if(sRequest.lpSort) {
        er = lpTable->SetSortOrder(&sRequest.lpSort->sSortOrder, sRequest.lpSort->ulCategories, sRequest.lpSort->ulExpanded);
        if(er != erSuccess)
            goto exit;
    }

    if(sRequest.lpSetColumns) {
        er = lpTable->SetColumns(sRequest.lpSetColumns);
        if(er != erSuccess)
            goto exit;
    }
    
    if(sRequest.lpRestrict) {
        er = lpTable->Restrict(sRequest.lpRestrict);
        if(er != erSuccess)
            goto exit;
    }
    
    if(sRequest.lpQueryRows) {
        er = lpTable->QueryRows(soap, sRequest.lpQueryRows->ulCount, sRequest.lpQueryRows->ulFlags, &lpRowSet);
        if(er != erSuccess)
            goto exit;
                
        lpsResponse->sRowSet.__ptr = lpRowSet->__ptr;
        lpsResponse->sRowSet.__size = lpRowSet->__size;
    }

exit:
    ;
}
SOAP_ENTRY_END()

ECRESULT CreateObject(ECSession *lpecSession, ECDatabase *lpDatabase, unsigned int ulParentObjId, unsigned int ulObjType, unsigned int ulFlags, unsigned int *lpulObjId) {
	ECRESULT		er;
	unsigned int	ulNewObjId = 0;
	unsigned int	ulAffected = 0;
	unsigned int	ulParentType = 0;
	unsigned int	ulOwner = 0;
	std::string		strQuery;


	//
    // We skip quota checking because we do this during writeProps.

	// Get object type
	er =  g_lpSessionManager->GetCacheManager()->GetObject(ulParentObjId, NULL, NULL, NULL, &ulParentType);
	if(er != erSuccess)
		goto exit;

	// Check permission, Skip if objecttype is an attachment with the parent MAPI_STORE
	if(ulParentType == MAPI_STORE && (ulObjType == MAPI_ATTACH || ulObjType == MAPI_MESSAGE) )
	{
		// All access rights
// FIXME: check store owner, or admin
	}else
	{
		er = lpecSession->GetSecurity()->CheckPermission(ulParentObjId, ecSecurityCreate);
		if(er != erSuccess)
			goto exit;
	}

	ulOwner = lpecSession->GetSecurity()->GetUserId(ulParentObjId); // Owner of object is either the current user or the owner of the folder

	// Create object
	strQuery = "INSERT INTO hierarchy (parent, type, flags, owner) values("+stringify(ulParentObjId)+", "+stringify(ulObjType)+", "+stringify(ulFlags)+", "+stringify(ulOwner)+")";
	er = lpDatabase->DoInsert(strQuery, &ulNewObjId, &ulAffected);
	if(er != erSuccess)
		goto exit;

	if(ulAffected != 1) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}


	// No notification is sent now, as the notification is sent directly after writeProps

	// Save this item in the cache, as there is a very high probability that this data will be required very soon (almost 100% sure)
	g_lpSessionManager->GetCacheManager()->SetObject(ulNewObjId, ulParentObjId, ulOwner, ulFlags, ulObjType);

	// return new object id to saveObject
	if (lpulObjId)
		*lpulObjId = ulNewObjId;

exit:
	return er;
}

// update PR_LOCAL_COMMIT_TIME_MAX on a folder which contents has changed
// FIXME: this function should be called when the contents of a folder change
// affected: emptyFolder, copyObjects, copyFolder, deleteObjects
ECRESULT WriteLocalCommitTimeMax(struct soap *soap, ECDatabase *lpDatabase, unsigned int ulStoreId, unsigned int ulFolderId, propVal **ppvTime)
{
	ECRESULT er = erSuccess;
	FILETIME ftNow;
	std::string strQuery;
	propVal *pvTime = NULL;

	if (ppvTime == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	GetSystemTimeAsFileTime(&ftNow);

	pvTime = s_alloc<propVal>(soap);
	pvTime->ulPropTag = PR_LOCAL_COMMIT_TIME_MAX;
	pvTime->__union = SOAP_UNION_propValData_hilo;
	pvTime->Value.hilo = s_alloc<hiloLong>(soap);
	pvTime->Value.hilo->hi = ftNow.dwHighDateTime;
	pvTime->Value.hilo->lo = ftNow.dwLowDateTime;

	strQuery = "REPLACE INTO properties (storeid, hierarchyid, tag, type, val_hi, val_lo) VALUES ("
		+stringify(ulStoreId)+","+stringify(ulFolderId)+","+stringify(PROP_ID(PR_LOCAL_COMMIT_TIME_MAX))+","+stringify(PROP_TYPE(PR_LOCAL_COMMIT_TIME_MAX))+","
		+stringify(ftNow.dwHighDateTime)+","+stringify(ftNow.dwLowDateTime)+")";

	er = lpDatabase->DoInsert(strQuery);
	if (er != erSuccess)
		goto exit;

	*ppvTime = pvTime;

exit:
	return er;
}

/*
 * This is a generic delete function that is called from
 *
 * ns__deleteObjects
 * ns__emptyFolder
 * ns__deleteFolder
 * ns__saveObject
 *
 * It does a recursive delete of objects in the hierarchytable, according to the flags given
 * which is any combination of
 *
 * EC_DELETE_FOLDERS		- Delete subfolders
 * EC_DELETE_MESSAGES		- Delete messages
 * EC_DELETE_RECIPIENTS		- Delete recipients of messages
 * EC_DELETE_ATTACHMENTS	- Delete attachments of messages
 * EC_DELETE_CONTAINER		- Delete the container specified in the first place (otherwise only subobjects)
 *
 * This is done by first recusively retrieving the object ID's, then checking the types of objects in that
 * list. If there is any subobject that has *not* been specified for deletion, the function fails. Else,
 * all properties in the subobjects and the subobjects themselves are deleted. If EC_DELETE_CONTAINER
 * is specified, then the objects passed in lpEntryList are also deleted (together with their properties).
 *
 */



/* This could be optimised by making the amount of SQL queries smaller, and getting
 * the data we want with a couple of larger resultsets
 *
 * This function can be delete objects from different stores.
 *
 * DeleteObjects() may be deleting a large amount of items. It therefore keeps its own SQL transaction state. Every
 * 32 items are grouped together and deleted in a single transaction. The caller should not start / commit a transaction
 * before calling this function, unless fNoBegin is TRUE, in which case DeleteObjects() will do no transaction handling.
 */

ECRESULT DeleteObjects(ECSession *lpSession, ECDatabase *lpDatabase, ECListInt *lpsObjectList, unsigned int ulFlags, unsigned int ulSyncId, bool fNoBegin, bool bCheckPermission)
{
	ECRESULT		er = erSuccess;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;
	std::string		strQuery;
	ECListIntIterator	iListObjectId;

	std::string		strInclause;
	bool			bPartialComplete = false;
	bool			fFirstClause = true;
	unsigned int	ulGrandParent = 0;
	size_t			cDeleteditems = 0;

	ECAttachmentStorage *lpAttachmentStorage = NULL;
	std::list<ULONG>	lstDeleteAttachments; // is also going to contain non-attachment object id's

	std::list<DELETEITEM>				lstDeleteItems;
	std::list<DELETEITEM>				lstDeleted;
	std::list<DELETEITEM>				lstToBeDeleted;
	std::list<unsigned int>				lstParent;
	std::list<DELETEITEM>::iterator		iterDeleteItems;
	std::list<unsigned int>::iterator	iterParent;
	map<unsigned int, long long>	mapStoreSize;
	map<unsigned int, long long>::iterator	iterStoreSizeItems;

	DELETEITEM sItem;
	FILETIME ft;

    if(!fNoBegin)
    	lpDatabase->Begin();

	// First, put all the root objects in the list
	for(iListObjectId = lpsObjectList->begin(); iListObjectId != lpsObjectList->end(); iListObjectId++) {

		//Free database results
		if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

		sItem.fRoot = true;
		sItem.ulId = *iListObjectId;

		strQuery = "SELECT parent,type, flags FROM hierarchy WHERE id=" + stringify(sItem.ulId);
		if((ulFlags & EC_DELETE_HARD_DELETE) != EC_DELETE_HARD_DELETE)
			strQuery += " AND (flags&"+stringify(MSGFLAG_DELETED)+") !="+stringify(MSGFLAG_DELETED);

		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		if(lpDatabase->GetNumRows(lpDBResult) == 0){ // object doesn't exist, ignore it.
			// Entry id is missing (by folders)
			//er = ZARAFA_E_INVALID_ENTRYID;
			//goto exit;
			// FIXME: skip to the messages part ?
			continue;
		}

		lpDBRow = lpDatabase->FetchRow(lpDBResult);

		if(lpDBRow == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL) // no parent folder or no type??
			continue;

		//lpDBRow[0] == NULL <- only happens when deleting a store
		if(lpDBRow[0] == NULL && atoi(lpDBRow[1]) != MAPI_STORE)
			continue;

		sItem.ulParent = (lpDBRow[0])?atoi(lpDBRow[0]) : 0;
		sItem.ulObjType = atoi(lpDBRow[1]);
		sItem.ulFlags = atoui(lpDBRow[2]) & (MSGFLAG_ASSOCIATED|MSGFLAG_DELETED);
		sItem.ulSubmit = atoui(lpDBRow[2]) & MSGFLAG_SUBMIT;

		// TODO: Join with properties to get PR_MESSAGE_SIZE/PR_ATTACH_SIZE, so it faster ?
		if(g_lpSessionManager->GetCacheManager()->GetStore(sItem.ulId, &sItem.ulStoreId , NULL) != erSuccess)
			sItem.ulStoreId = 0;

		if(sItem.ulObjType == MAPI_STORE || sItem.ulObjType == MAPI_FOLDER ||
			GetObjectSize(lpDatabase, sItem.ulStoreId, sItem.ulId, &sItem.ulObjSize) != erSuccess)
			sItem.ulObjSize = 0;


		GetSourceKey(sItem.ulId, &sItem.sSourceKey);
		GetSourceKey(sItem.ulParent, &sItem.sParentSourceKey);

		if(g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(sItem.ulId, NULL, &sItem.sEntryId) != erSuccess) {
			sItem.sEntryId.__size = 0;
			sItem.sEntryId.__ptr = NULL;
		}

		lstDeleteItems.push_back(sItem);

	}

	// Now, run through the list, adding children to the bottom of the list. This means
	// we're actually running width-first, and don't have to do anything recursive.
	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) {

		//Free database results
		if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

		strQuery = "SELECT id,type FROM hierarchy WHERE parent=" + stringify(iterDeleteItems->ulId);
		if((ulFlags & EC_DELETE_HARD_DELETE) != EC_DELETE_HARD_DELETE)
			strQuery += " AND (flags&"+stringify(MSGFLAG_DELETED)+") !="+stringify(MSGFLAG_DELETED);

		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		if(lpDatabase->GetNumRows(lpDBResult) == 0) {
			// Now parent objects
			continue;
		}


		while((lpDBRow = lpDatabase->FetchRow(lpDBResult)) != NULL ) {

			if(lpDBRow[0] == NULL || lpDBRow[1] == NULL) // no parent folder or no type??
				continue;

			// Add this object as a node to the end of the list
			sItem.fRoot = false;
			sItem.ulId = atoi(lpDBRow[0]);
			sItem.ulObjType = atoi(lpDBRow[1]);
			sItem.ulParent = iterDeleteItems->ulId;
			sItem.ulFlags = 0;
			// TODO: Join with properties to get PR_MESSAGE_SIZE/PR_ATTACH_SIZE, so it faster ?
			if(g_lpSessionManager->GetCacheManager()->GetStore(sItem.ulId, &sItem.ulStoreId , NULL) != erSuccess)
				sItem.ulStoreId = 0;

			if( sItem.ulObjType == MAPI_STORE || sItem.ulObjType == MAPI_FOLDER ||
				GetObjectSize(lpDatabase, sItem.ulStoreId, sItem.ulId, &sItem.ulObjSize) != erSuccess)
				sItem.ulObjSize = 0;

			GetSourceKey(sItem.ulId, &sItem.sSourceKey);
			GetSourceKey(sItem.ulParent, &sItem.sParentSourceKey);

			if(g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(sItem.ulId, NULL, &sItem.sEntryId) != erSuccess) {
				sItem.sEntryId.__size = 0;
				sItem.sEntryId.__ptr = NULL;
			}


			lstDeleteItems.push_back(sItem);
		}

	}

	// We now have a complete list of all the objects that would have to be deleted if the specified
	// objects were to be deleted

	// Check whether there are any object types that we're not supposed to delete, or are not
	// allowed to be deleted
	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) {

		// Check permission for each folder and messages etc..
		if (bCheckPermission) {
			er = lpSession->GetSecurity()->CheckPermission(iterDeleteItems->ulId, ecSecurityDelete);
			if(er != erSuccess)
				goto exit;
		}

		if(iterDeleteItems->fRoot)
			continue;		// always allow root folders that were passed

		switch(iterDeleteItems->ulObjType) {
		case MAPI_MESSAGE:
			if(! (ulFlags & EC_DELETE_MESSAGES) ) {
				er = ZARAFA_E_HAS_MESSAGES;
				goto exit;
			}
			break;
		case MAPI_FOLDER:
			if(! (ulFlags & EC_DELETE_FOLDERS) ) {
				er = ZARAFA_E_HAS_FOLDERS;
				goto exit;
			}
			break;
		case MAPI_MAILUSER:
		case MAPI_DISTLIST:
			if(! (ulFlags & EC_DELETE_RECIPIENTS) ) {
				er = ZARAFA_E_HAS_RECIPIENTS;
				goto exit;
			}
			break;
		case MAPI_ATTACH:
			if(! (ulFlags & EC_DELETE_ATTACHMENTS) ) {
				er = ZARAFA_E_HAS_ATTACHMENTS;
				goto exit;
			}
			break;
		case MAPI_STORE: // only admins can delete a store
			if(!(ulFlags & EC_DELETE_STORES) || lpSession->GetSecurity()->IsObjectAdmin(iterDeleteItems->ulStoreId) != erSuccess)
			{
				er = ZARAFA_E_NOT_FOUND;
				goto exit;
			}
			break;
		default:
			// Unknown object type? We'll delete it anyway so we don't get frustrating non-deletable items

			break;
		}

	}

	// Remove search results for deleted folders
	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) {
		if(iterDeleteItems->ulObjType == MAPI_FOLDER) {
			if(g_lpSessionManager->GetSearchFolders()->IsSearchFolder(iterDeleteItems->ulStoreId, iterDeleteItems->ulId) == erSuccess) {
				g_lpSessionManager->GetSearchFolders()->RemoveSearchFolder(iterDeleteItems->ulStoreId, iterDeleteItems->ulId);
			}
		}
	}

	/* From this point, we're going to start deleting. Each item that was successfully deleted and committed to the database
	 * we're adding to a list. We need the list because if the delete fails halfway, we still need to update the store size, etc.
	 */

	if(ulFlags & EC_DELETE_HARD_DELETE)
	{
	    /* Perform a hard delete.
	     *
	     * This means we should be really deleting the actual data from the database. We do this in batches of 32 items each because
	     * deleting records is generally fairly slow. Also, large delete batches originating from the SoftDeleteRemover() can be very
	     * large, taking up to an hour to process. We don't want to have a transaction lasting an hour because it would cause lots of
	     * locking problems
	     */
		er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
		if (er != erSuccess)
			goto exit;

		// If we get here, all the objects have been OK'd for deletion, so start the deletion ..
		for (iterDeleteItems = lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); ) {
			int i;

			strInclause = "IN (";
			fFirstClause = true;
			i = 0;

			lstDeleteAttachments.clear();
            lstToBeDeleted.clear();

			// Delete max 32 items per query
			while (i<32 && iterDeleteItems != lstDeleteItems.end()) {
				if((ulFlags & EC_DELETE_CONTAINER) || !iterDeleteItems->fRoot) {
					if(!fFirstClause)
						strInclause += ",";

					strInclause += stringify(iterDeleteItems->ulId);

					// make new list for attachment deletes
					lstDeleteAttachments.push_back(iterDeleteItems->ulId);

					lstToBeDeleted.push_back(*iterDeleteItems);

					fFirstClause = false;
					i++;
				}
				iterDeleteItems++;
			}

			strInclause += ")";

			// list may contain non-attachment object id's!
			er = lpAttachmentStorage->DeleteAttachments(lstDeleteAttachments);
			if (er != erSuccess)
				goto fail;

			if(fFirstClause == false) {
				// First, the hierarchy entries of all the objects
				strQuery = "DELETE FROM hierarchy WHERE id " + strInclause;
				er = lpDatabase->DoDelete(strQuery);
				if(er!= erSuccess)
					goto fail;

				// Then, the properties for the objects we just deleted
				strQuery = "DELETE FROM properties WHERE hierarchyid " + strInclause;
				er = lpDatabase->DoDelete(strQuery);
				if(er!= erSuccess)
					goto fail;

				// Then, the MVproperties for the objects we just deleted
				strQuery = "DELETE FROM mvproperties WHERE hierarchyid " + strInclause;
				er = lpDatabase->DoDelete(strQuery);
				if(er!= erSuccess)
					goto fail;

				// Then, the acls for the objects we just deleted (if exist)
				strQuery = "DELETE FROM acl WHERE hierarchy_id " + strInclause;
				er = lpDatabase->DoDelete(strQuery);
				if(er != erSuccess)
					goto fail;

				// remove indexedproperties
				strQuery = "DELETE FROM indexedproperties WHERE hierarchyid " + strInclause;
				er = lpDatabase->DoDelete(strQuery);
				if(er != erSuccess)
					goto fail;

			}

			// Commit and start new transaction
			if(!fNoBegin) {
    			er = lpDatabase->Commit();
	    		if(er != erSuccess)
		    	    goto fail;
            }

            // Deletes have been committed, add the deleted items to the list of items we have deleted
            lstDeleted.splice(lstDeleted.end(),lstToBeDeleted);

		    if(!fNoBegin)
			    lpDatabase->Begin();

		} // while iterDeleteItems != end()

	} else {
	    /* Do a soft delete
	     *
	     * Since this is a fairly simple operation, we are doing soft deletes in a single transaction. Once the SQL has gone OK, we know that
	     * all items were successfully deleted and we can therefore add all soft-deleted items into the 'lstDeleted' list at once
	     */
		strInclause = "IN (";
		fFirstClause = true;
		for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) {

			if( ( ((ulFlags & EC_DELETE_CONTAINER)== EC_DELETE_CONTAINER && iterDeleteItems->fRoot == true) || ((ulFlags & EC_DELETE_CONTAINER)== 0 && iterDeleteItems->fRoot == false) ) && (iterDeleteItems->ulObjType == MAPI_MESSAGE || iterDeleteItems->ulObjType == MAPI_FOLDER  || iterDeleteItems->ulObjType == MAPI_STORE) ) {
				if(!fFirstClause)
					strInclause += ",";

				strInclause += stringify(iterDeleteItems->ulId);

				fFirstClause = false;

			}

		}

		strInclause += ")";

		if(fFirstClause == false)
		{
			// First, the hierarchy entries of all the objects
			strQuery = "UPDATE hierarchy SET flags=flags|"+stringify(MSGFLAG_DELETED)+" WHERE id " + strInclause;
			er = lpDatabase->DoUpdate(strQuery);
			if(er!= erSuccess)
				goto exit;

			// Add properties: PR_DELETED_ON
			GetSystemTimeAsFileTime(&ft);

			for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++)
			{
				if( ( ((ulFlags & EC_DELETE_CONTAINER)== EC_DELETE_CONTAINER && iterDeleteItems->fRoot == true) || ((ulFlags & EC_DELETE_CONTAINER)== 0 && iterDeleteItems->fRoot == false) ) && (iterDeleteItems->ulObjType == MAPI_MESSAGE || iterDeleteItems->ulObjType == MAPI_FOLDER || iterDeleteItems->ulObjType == MAPI_STORE) )
				{
					strQuery = "REPLACE INTO properties(hierarchyid, tag, type, storeid, val_lo, val_hi) VALUES("+stringify(iterDeleteItems->ulId)+","+stringify(PROP_ID(PR_DELETED_ON))+","+stringify(PROP_TYPE(PR_DELETED_ON))+","+stringify(iterDeleteItems->ulStoreId)+","+stringify(ft.dwLowDateTime)+","+stringify(ft.dwHighDateTime)+")";
					er = lpDatabase->DoUpdate(strQuery);
					if(er!= erSuccess)
						goto exit;
				}
			}

			// Add all items from 'lstDeleteItems' to the list of 'deleted' objects. Note that not
			// *all* of these objects are marked for deletion; when you delete a folder, only the top-level
			// item is actually marked for deletion, but we have to take into account the size of all items, not
			// just the top-level folder.
			lstDeleted = lstDeleteItems;

		}else{
			er = erSuccess;// No items to delete ?
			goto exit;
		}
	}

fail:
	// Size of the deleted item list
	cDeleteditems = lstDeleted.size();

    if(er != erSuccess) {
        if(cDeleteditems == 0)
            goto exit; // Nothing has been deleted, completely fail

        // Otherwise, some items have successfully been deleted. We should therefore continue and return partial completion.
        bPartialComplete = true;
        er = erSuccess;
    }

	///////////////////////////////
	// remove the size for quota
	
	for(iterDeleteItems=lstDeleted.begin(); iterDeleteItems != lstDeleted.end(); iterDeleteItems++) {
		if(((ulFlags & EC_DELETE_CONTAINER) || !iterDeleteItems->fRoot) &&
			(iterDeleteItems->ulFlags&MSGFLAG_DELETED) != MSGFLAG_DELETED &&
			iterDeleteItems->ulStoreId > 0 && iterDeleteItems->ulObjSize > 0 &&
			iterDeleteItems->ulObjType == MAPI_MESSAGE)
		{
			if (mapStoreSize.find(iterDeleteItems->ulStoreId) != mapStoreSize.end() )
				mapStoreSize[iterDeleteItems->ulStoreId] += iterDeleteItems->ulObjSize;
			else
				mapStoreSize[iterDeleteItems->ulStoreId] = iterDeleteItems->ulObjSize;
		}
	}

	// Update store size for each store
	for(iterStoreSizeItems=mapStoreSize.begin(); iterStoreSizeItems != mapStoreSize.end(); iterStoreSizeItems++) {
		UpdateObjectSize(lpDatabase, iterStoreSizeItems->first, iterStoreSizeItems->first, MAPI_STORE, false, UPDATE_SUB, iterStoreSizeItems->second);
	}

	
	for(iterDeleteItems=lstDeleted.begin(); iterDeleteItems != lstDeleted.end(); iterDeleteItems++)
	{
		// Remove any entries in the outgoing queue for deleted items
		if(iterDeleteItems->ulSubmit && iterDeleteItems->ulObjType == MAPI_MESSAGE) {
			strQuery = "DELETE FROM outgoingqueue WHERE hierarchy_id = " + stringify(iterDeleteItems->ulId);
			er = lpDatabase->DoDelete(strQuery);
			if( er != erSuccess )
				goto exit;

			// Make sure the SUBMIT flag is no longer set on a soft deleted item. If the item is restored, the submit flag will
			// be OFF. This will do nothing on a hard delete.
			if((ulFlags & EC_DELETE_HARD_DELETE) == 0) {
				strQuery = "UPDATE hierarchy SET flags = flags &~" + stringify(MSGFLAG_SUBMIT) + " WHERE id = " + stringify(iterDeleteItems->ulId);
				er = lpDatabase->DoUpdate(strQuery);
				if( er != erSuccess )
					goto exit;
			}
		}

		// ICS update
		if ((ulFlags & EC_DELETE_CONTAINER) || !iterDeleteItems->fRoot) {
			if(iterDeleteItems->ulObjType == MAPI_MESSAGE)
				AddChange(lpSession, ulSyncId, iterDeleteItems->sSourceKey, iterDeleteItems->sParentSourceKey, ulFlags & EC_DELETE_HARD_DELETE ? ICS_MESSAGE_HARD_DELETE : ICS_MESSAGE_SOFT_DELETE);
			else if(iterDeleteItems->ulObjType == MAPI_FOLDER && !(iterDeleteItems->ulFlags & FOLDER_SEARCH)){
				AddChange(lpSession, ulSyncId, iterDeleteItems->sSourceKey, iterDeleteItems->sParentSourceKey, ulFlags & EC_DELETE_HARD_DELETE ? ICS_FOLDER_HARD_DELETE : ICS_FOLDER_SOFT_DELETE);
			}
		}
	}

	// Finish transaction
	if(!fNoBegin)
    	lpDatabase->Commit(); //FIXME: check error

	for(iterDeleteItems=lstDeleted.begin(); iterDeleteItems != lstDeleted.end(); iterDeleteItems++) {

		// update the cache
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectDeleted, iterDeleteItems->ulId);

		// Update cache, Remove index properties
		if((ulFlags & EC_DELETE_HARD_DELETE) == EC_DELETE_HARD_DELETE)
			g_lpSessionManager->GetCacheManager()->RemoveIndexData(iterDeleteItems->ulId);

		// Remove the item from both the local and master outgoing queues
		if(iterDeleteItems->ulSubmit && iterDeleteItems->ulObjType == MAPI_MESSAGE) {
			g_lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_DELETE, iterDeleteItems->ulStoreId, iterDeleteItems->ulId, EC_SUBMIT_LOCAL, MAPI_MESSAGE);
			g_lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_DELETE, iterDeleteItems->ulStoreId, iterDeleteItems->ulId, EC_SUBMIT_MASTER, MAPI_MESSAGE);
		}
	}

	// Now, send the notifications for MAPI_MESSAGE and MAPI_FOLDER types
	for(iterDeleteItems=lstDeleted.begin(); iterDeleteItems != lstDeleted.end(); iterDeleteItems++) {
		if( ((ulFlags & EC_DELETE_CONTAINER) || !iterDeleteItems->fRoot) && (iterDeleteItems->ulObjType == MAPI_MESSAGE || iterDeleteItems->ulObjType == MAPI_FOLDER) ) {
			// Notify that the message has been deleted

			g_lpSessionManager->NotificationDeleted(iterDeleteItems->ulObjType, iterDeleteItems->ulId, iterDeleteItems->ulStoreId, &iterDeleteItems->sEntryId, iterDeleteItems->ulParent, iterDeleteItems->ulFlags);

			// Update all tables viewing this message
			if(cDeleteditems < EC_TABLE_CHANGE_THRESHOLD)
    			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_DELETE, iterDeleteItems->ulFlags, iterDeleteItems->ulParent, iterDeleteItems->ulId, iterDeleteItems->ulObjType);

			if((ulFlags & EC_DELETE_HARD_DELETE) != EC_DELETE_HARD_DELETE && cDeleteditems < EC_TABLE_CHANGE_THRESHOLD)
			{
				// Update all tables viewing this message
				g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_ADD, MSGFLAG_DELETED, iterDeleteItems->ulParent, iterDeleteItems->ulId, iterDeleteItems->ulObjType);
			}

			if(iterDeleteItems->fRoot == true)
				lstParent.push_back(iterDeleteItems->ulParent);
		}

		if((ulFlags & EC_DELETE_CONTAINER) == 0 && iterDeleteItems->fRoot == true && iterDeleteItems->ulObjType == MAPI_FOLDER)
			lstParent.push_back(iterDeleteItems->ulId);
	}

	// We have a list of all the folders in which something was deleted, so get a unique list
	lstParent.sort();
	lstParent.unique();

	// Now, send each parent folder a notification that it has been altered and send
	// its parent a notification (ie the grandparent of the deleted object) that its
	// hierarchy table has been changed.
	for(iterParent = lstParent.begin(); iterParent != lstParent.end(); iterParent++) {
	    if(cDeleteditems >= EC_TABLE_CHANGE_THRESHOLD) {
	        // FIXME: Don't know the object type so try both ..
			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, (ulFlags & EC_DELETE_HARD_DELETE)?MSGFLAG_DELETED:0, *iterParent, 0, MAPI_MESSAGE);
			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, (ulFlags & EC_DELETE_HARD_DELETE)?MSGFLAG_DELETED:0, *iterParent, 0, MAPI_FOLDER);

			if((ulFlags & EC_DELETE_HARD_DELETE) != EC_DELETE_HARD_DELETE) {
				g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, MSGFLAG_DELETED, *iterParent, 0, MAPI_MESSAGE);
				g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, MSGFLAG_DELETED, *iterParent, 0, MAPI_FOLDER);
			}
        }

        g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, *iterParent);
		g_lpSessionManager->NotificationModified(MAPI_FOLDER, *iterParent);

		if(g_lpSessionManager->GetCacheManager()->GetParent(*iterParent, &ulGrandParent) == erSuccess)
			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParent, *iterParent, MAPI_FOLDER);
	}

	if(bPartialComplete)
	    er = ZARAFA_W_PARTIAL_COMPLETION;

exit:
    if(er != erSuccess && !fNoBegin && er != ZARAFA_W_PARTIAL_COMPLETION)
        lpDatabase->Rollback();

	if (lpAttachmentStorage)
		delete lpAttachmentStorage;

	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) {
		if(iterDeleteItems->sEntryId.__ptr)
			delete [] iterDeleteItems->sEntryId.__ptr;
	}

	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

// Delete a set of messages, recipients, or attachments
SOAP_ENTRY_START4(deleteObjects, *result, unsigned int ulFlags, struct entryList *lpEntryList, unsigned int ulSyncId, unsigned int *result)
{
	ECListInt	lObjectList;
	unsigned int ulDeleteFlags = EC_DELETE_ATTACHMENTS | EC_DELETE_RECIPIENTS | EC_DELETE_CONTAINER | EC_DELETE_MESSAGES;
	USE_DATABASE();

	if(lpEntryList == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if(ulFlags & DELETE_HARD_DELETE)
		ulDeleteFlags |= EC_DELETE_HARD_DELETE;

	// ignore errors
	g_lpSessionManager->GetCacheManager()->GetEntryListToObjectList(lpEntryList, &lObjectList);

	er = DeleteObjects(lpecSession, lpDatabase, &lObjectList, ulDeleteFlags, ulSyncId, false, true);
	if(er != erSuccess)
		goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

// Delete everything in a folder, but not the folder itself
// Quirk: this works with messages also, deleting attachments and recipients, but not the message itself.
//FIXME: michel? what with associated messages ?
SOAP_ENTRY_START4(emptyFolder, *result, entryId sEntryId, unsigned int ulFlags, unsigned int ulSyncId, unsigned int *result)
{
	unsigned int		ulDeleteFlags = 0;
	unsigned int		ulId = 0;
	ECListInt			lObjectIds;
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulId);
	if(er != erSuccess)
	    goto exit;

	// Check Rights set permission
	er = lpecSession->GetSecurity()->CheckPermission(ulId, ecSecurityDelete);
	if(er != erSuccess)
		goto exit;

	// Add object into the list
	lObjectIds.push_back(ulId);

	ulDeleteFlags = EC_DELETE_MESSAGES | EC_DELETE_FOLDERS | EC_DELETE_RECIPIENTS | EC_DELETE_ATTACHMENTS;

	if((ulFlags & DELETE_HARD_DELETE) == DELETE_HARD_DELETE)
		ulDeleteFlags |= EC_DELETE_HARD_DELETE;

	er = DeleteObjects(lpecSession, lpDatabase, &lObjectIds, ulDeleteFlags, ulSyncId, false, true);
	if (er != erSuccess)
		goto exit;

exit:
	ROLLBACK_ON_ERROR();
}
SOAP_ENTRY_END()

/* FIXME
 *
 * Currently, when deleteFolders is called with DEL_FOLDERS but without DEL_MESSAGES, it will return an error
 * when a subfolder of the specified folder contains messages. I don't think this is up to spec. DeleteObjects
 * should therefore be changed so that the check is only done against messages and folders directly under the
 * top-level object.
 */

// Deletes a complete folder, with optional recursive subfolder and submessage deletion
SOAP_ENTRY_START4(deleteFolder, *result,  entryId sEntryId, unsigned int ulFlags, unsigned int ulSyncId, unsigned int *result)
{
	unsigned int		ulDeleteFlags = 0;
	unsigned int		ulId = 0;
	unsigned int		ulFolderFlags = 0;
	ECListInt			lObjectIds;
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulId);
	if(er != erSuccess)
	    goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulId, ecSecurityFolderAccess);
	if(er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFlags(ulId, &ulFolderFlags);
	if(er != erSuccess)
		goto exit;

	// insert objectid into the delete list
	lObjectIds.push_back(ulId);

	ulDeleteFlags = EC_DELETE_CONTAINER;

	if(ulFlags & DEL_FOLDERS)
		ulDeleteFlags |= EC_DELETE_FOLDERS;

	if(ulFlags & DEL_MESSAGES)
		ulDeleteFlags |= EC_DELETE_MESSAGES | EC_DELETE_RECIPIENTS | EC_DELETE_ATTACHMENTS;

	if( (ulFlags & DELETE_HARD_DELETE) || ulFolderFlags == FOLDER_SEARCH)
		ulDeleteFlags |= EC_DELETE_HARD_DELETE;

	er = DeleteObjects(lpecSession, lpDatabase, &lObjectIds, ulDeleteFlags, ulSyncId, false, true);
	if (er != erSuccess)
		goto exit;

exit:
	ROLLBACK_ON_ERROR();
}
SOAP_ENTRY_END()

ECRESULT DoNotifySubscribe(ECSession *lpecSession, unsigned long long ulSessionId, struct notifySubscribe *notifySubscribe)
{
	ECRESULT er = erSuccess;
	unsigned int ulKey = 0;
	ECGenericObjectTable *lpTable = NULL;

	//NOTE: An sKey with size 4 is a table notification id
	if(notifySubscribe->sKey.__size == 4) {
		 memcpy(&ulKey, notifySubscribe->sKey.__ptr, 4);
	}else {
		er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&notifySubscribe->sKey, &ulKey);
		if(er != erSuccess)
			goto exit;
	}
	
	if(notifySubscribe->ulEventMask & fnevTableModified) {
	    // An advise has been done on a table. The table ID is in 'ulKey' in this case. When this is done
	    // we have to populate the table first since row modifications would otherwise be wrong until the
	    // table is populated; if the table is unpopulated and a row changes, the row will be added into the table
	    // whenever it is modified, producing a TABLE_ROW_ADDED for that row instead of the correct TABLE_ROW_MODIFIED.
	    er = lpecSession->GetTableManager()->GetTable(ulKey, &lpTable);
	    if(er != erSuccess)
	        goto exit;
	        
        er = lpTable->Populate();
        if(er != erSuccess)
            goto exit;
	}

	er = lpecSession->AddAdvise(notifySubscribe->ulConnection, ulKey, notifySubscribe->ulEventMask);
	if(er == erSuccess){
		TRACE_SOAP(TRACE_INFO, "ns__notifySubscribe", "connectionId: %d SessionId: %d Mask: %d",notifySubscribe->ulConnection, ulSessionId, notifySubscribe->ulEventMask);
	}

exit:
	return er;
}

SOAP_ENTRY_START2(notifySubscribe, *result,  struct notifySubscribe *notifySubscribe, unsigned int *result)
{
	er = DoNotifySubscribe(lpecSession, ulSessionId, notifySubscribe);
	if (er != erSuccess)
		goto exit;
exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(notifyUnSubscribe, *result,  unsigned int ulConnection, unsigned int *result)
{
	er = lpecSession->DelAdvise(ulConnection);
	if (er != erSuccess)
		goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

/*
 * Get the notifications of a session.
 * This fucntion keep the session alive by timeout every "WAIT_TIMEOUT",
 * the client reconnect automaticly. so the session will be alive.
 */
SOAP_ENTRY_START1(notifyGetItems, notifications->er, struct notifyResponse *notifications)
{
	// Blocking call
	er = lpecSession->GetNotifyItems(soap, notifications);
	if(FAILED(er))
		goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getRights, lpsRightResponse->er, entryId sEntryId, int ulType, struct rightsResponse *lpsRightResponse)
{
	unsigned int	ulobjid = 0;

	struct rightsArray *lpsRightArray = NULL;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulobjid);
	if(er != erSuccess)
	    goto exit;

	lpsRightArray = new struct rightsArray;
	memset(lpsRightArray, 0, sizeof(struct rightsArray));

	er = lpecSession->GetSecurity()->GetRights(ulobjid, ulType, lpsRightArray);
	if(er != erSuccess)
		goto exit;

	er = CopyRightsArrayToSoap(soap, lpsRightArray, &lpsRightResponse->pRightsArray);
	if (er != erSuccess)
		goto exit;

exit:
	if (lpsRightArray) {
		for (unsigned int i = 0; i < lpsRightArray->__size; i++)
			if (lpsRightArray->__ptr[i].sUserId.__ptr)
				delete [] lpsRightArray->__ptr[i].sUserId.__ptr;

		if (lpsRightArray->__size > 0)
			delete[] lpsRightArray->__ptr;

		delete lpsRightArray;
	}
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(setRights, *result, entryId sEntryId, struct rightsArray *lpsRightsArray, unsigned int *result)
{
	unsigned int	ulObjId = 0;

	if(lpsRightsArray == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if(er != erSuccess)
	    goto exit;

	// Check Rights set permission
	er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityFolderAccess);
	if(er != erSuccess)
		goto exit;

	er = lpecSession->GetSecurity()->SetRights(ulObjId, lpsRightsArray);
	if(er != erSuccess)
		goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START4(getUserObjectList, lpsUserObjectResponse->er, unsigned int ulCompanyId, entryId sCompanyId, int ulType, struct userobjectResponse *lpsUserObjectResponse)
{
	std::list<localobjectdetails_t> *lpObjects = NULL;
	std::list<localobjectdetails_t>::iterator iterObjects;
	entryId			sUserEid = {0};
	
	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	/* Input check, if ulCompanyId is 0, we want the user's company,
	 * otherwise we must check if the requested company is visible for the user. */
	if (ulCompanyId == 0) {
		er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
		if (er != erSuccess)
			goto exit;
	} else {
		er = lpecSession->GetSecurity()->IsUserObjectVisible(ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	// Get the userlist
	er = lpecSession->GetUserManagement()->GetMergedObjectListAndSync(ulType, ulCompanyId, &lpObjects, 0);
	if(er != erSuccess)
		goto exit;

	lpsUserObjectResponse->pUserObjectArray = s_alloc<userobjectArray>(soap);
	lpsUserObjectResponse->pUserObjectArray->__size = 0;
	lpsUserObjectResponse->pUserObjectArray->__ptr = s_alloc<userobject>(soap, lpObjects->size());

	for(iterObjects = lpObjects->begin(); iterObjects!= lpObjects->end(); iterObjects++) {
		er = GetABEntryID(iterObjects->ulId, soap, &sUserEid);
		if (er != erSuccess)
			goto exit;

		er = CopyUserObjectDetailsToSoap(iterObjects->ulId, &sUserEid, *iterObjects, soap, &lpsUserObjectResponse->pUserObjectArray->__ptr[lpsUserObjectResponse->pUserObjectArray->__size]);
		if (er != erSuccess)
			goto exit;

		lpsUserObjectResponse->pUserObjectArray->__size++;
	}

exit:
	if(lpObjects)
		delete lpObjects;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(getOwner, lpsResponse->er, entryId sEntryId, struct getOwnerResponse *lpsResponse)
{
	unsigned int	ulobjid = 0;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulobjid);
	if(er != erSuccess)
	    goto exit;

	er = lpecSession->GetSecurity()->GetOwner(ulobjid, &lpsResponse->ulOwner);
	if(er != erSuccess)
		goto exit;

	er = GetABEntryID(lpsResponse->ulOwner, soap, &lpsResponse->sOwner);
	if (er != erSuccess)
		goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getIDsFromNames, lpsResponse->er,  struct namedPropArray *lpsNamedProps, unsigned int ulFlags, struct getIDsFromNamesResponse *lpsResponse)
{
	unsigned int	i;
	std::string		strEscapedString;
	std::string		strEscapedGUID;
	unsigned int	ulLastId = 0;
    USE_DATABASE();

	if(lpsNamedProps == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpDatabase->Begin();
	if(er != erSuccess)
		goto exit;

	lpsResponse->lpsPropTags.__ptr = s_alloc<unsigned int>(soap, lpsNamedProps->__size);
	lpsResponse->lpsPropTags.__size = 0;

	// One query per named property (too slow ?) FIXME could be faster if brought down to less SQL queries
	for(i=0;i<lpsNamedProps->__size;i++) {
		strQuery = "SELECT id FROM names WHERE ";

		// ID, then add ID where clause
		if(lpsNamedProps->__ptr[i].lpId != NULL) {
			strQuery += "nameid=" + stringify(*lpsNamedProps->__ptr[i].lpId) + " ";
		}

		// String, then add STRING where clause
		else if(lpsNamedProps->__ptr[i].lpString != NULL) {
			strEscapedString = lpDatabase->Escape(lpsNamedProps->__ptr[i].lpString);

			strQuery += "namestring='" + strEscapedString + "' ";
		}

		// Add a GUID specifier if there
		if(lpsNamedProps->__ptr[i].lpguid != NULL) {
			strEscapedGUID = lpDatabase->EscapeBinary(lpsNamedProps->__ptr[i].lpguid->__ptr, lpsNamedProps->__ptr[i].lpguid->__size);

			strQuery += "AND guid=" + strEscapedGUID;
		}

		// Run the query
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		if(lpDatabase->GetNumRows(lpDBResult) == 0) {
			// No rows found, so the named property has not been registered yet

			if(ulFlags & MAPI_CREATE) {
				// Create requested ? then add a new named property

				// GUID must be specified with create
				if(lpsNamedProps->__ptr[i].lpguid == NULL) {
					er = ZARAFA_E_NO_ACCESS;
					goto exit;
				}

				strQuery = "INSERT INTO names (nameid, namestring, guid) VALUES(";

				if(lpsNamedProps->__ptr[i].lpId) {
					strQuery += stringify(*lpsNamedProps->__ptr[i].lpId);
				} else {
					strQuery += "null";
				}

				strQuery += ",";

				if(lpsNamedProps->__ptr[i].lpString) {
					strQuery += "'" + strEscapedString + "'";
				} else {
					strQuery += "null";
				}

				strQuery += ",";

				if(lpsNamedProps->__ptr[i].lpguid) {
					strQuery += strEscapedGUID;
				} else {
					strQuery += "null";
				}

				strQuery += ")";

				er = lpDatabase->DoInsert(strQuery, &ulLastId);
				if(er != erSuccess)
					goto exit;

				lpsResponse->lpsPropTags.__ptr[i] = ulLastId+1; // offset one because 0 is 'not found'
			} else {
				// No create ? Then not found
				lpsResponse->lpsPropTags.__ptr[i] = 0;
			}
		} else {
			// found it
			lpDBRow = lpDatabase->FetchRow(lpDBResult);

			if(lpDBRow!= NULL && lpDBRow[0] != NULL)
				lpsResponse->lpsPropTags.__ptr[i] = atoi(lpDBRow[0])+1;
			else
				lpsResponse->lpsPropTags.__ptr[i] = 0;
		}

		//Free database results
		FREE_DBRESULT();
	}

	// Everything is done, now set the size
	lpsResponse->lpsPropTags.__size = lpsNamedProps->__size;

	er = lpDatabase->Commit();
	if(er != erSuccess)
		goto exit;
exit:
    FREE_DBRESULT();
	ROLLBACK_ON_ERROR();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(getNamesFromIDs, lpsResponse->er, struct propTagArray *lpPropTags, struct getNamesFromIDsResponse *lpsResponse)
{
	int				i;
	USE_DATABASE();

	if(lpPropTags == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// Allocate memory for return object
	lpsResponse->lpsNames.__ptr = s_alloc<namedProp>(soap, lpPropTags->__size);
	lpsResponse->lpsNames.__size = lpPropTags->__size;
	memset(lpsResponse->lpsNames.__ptr, 0, sizeof(struct namedProp) * lpPropTags->__size);

	for(i=0;i<lpPropTags->__size;i++) {
		strQuery = "SELECT nameid, namestring, guid FROM names WHERE id=" + stringify(lpPropTags->__ptr[i]-1);

		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		if(lpDatabase->GetNumRows(lpDBResult) == 1) {

			lpDBRow = lpDatabase->FetchRow(lpDBResult);
			lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

			if(lpDBRow != NULL) {
				if(lpDBRow[0] != NULL) {
					// It's an ID type
					lpsResponse->lpsNames.__ptr[i].lpId = s_alloc<unsigned int>(soap);
					*lpsResponse->lpsNames.__ptr[i].lpId = atoi(lpDBRow[0]);
				} else if(lpDBRow[1] != NULL) {
					// It's a String type
					lpsResponse->lpsNames.__ptr[i].lpString = s_alloc<char>(soap, strlen(lpDBRow[1])+1);
					strcpy(lpsResponse->lpsNames.__ptr[i].lpString, lpDBRow[1]);
				}

				if(lpDBRow[2] != NULL) {
					// Got a GUID (should always do so ...)
					lpsResponse->lpsNames.__ptr[i].lpguid = s_alloc<struct xsd__base64Binary>(soap);
					lpsResponse->lpsNames.__ptr[i].lpguid->__size = lpDBLen[2];
					lpsResponse->lpsNames.__ptr[i].lpguid->__ptr = s_alloc<unsigned char>(soap, lpDBLen[2]);
					memcpy(lpsResponse->lpsNames.__ptr[i].lpguid->__ptr, lpDBRow[2], lpDBLen[2]);
				}
			} else {
				er = ZARAFA_E_DATABASE_ERROR;
				goto exit;
			}
		} else {
			// No entry
			lpsResponse->lpsNames.__ptr[i].lpguid = NULL;
			lpsResponse->lpsNames.__ptr[i].lpId = NULL;
			lpsResponse->lpsNames.__ptr[i].lpString = NULL;
		}

		//Free database results
        FREE_DBRESULT();
	}


exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getReceiveFolder, lpsReceiveFolder->er, entryId sStoreId, char* lpszMessageClass, struct receiveFolderResponse *lpsReceiveFolder)
{
	unsigned int	ulStoreid = 0;

	int				nLen = 0;
	char			*lpDest;
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sStoreId, &ulStoreid);
	if(er != erSuccess)
	    goto exit;

	// Check for default store
	if(lpszMessageClass == NULL)
		lpszMessageClass = "";

	strQuery = "SELECT objid, messageclass FROM receivefolder WHERE storeid="+stringify(ulStoreid)+" AND (";

	strQuery += "messageclass='"+lpDatabase->Escape(lpszMessageClass)+"'";

	lpDest = lpszMessageClass;
	do {
		lpDest = strchr(lpDest, '.');

		if(lpDest){
			strQuery += " OR messageclass='"+lpDatabase->Escape(string(lpszMessageClass, lpDest-lpszMessageClass))+"'";
			lpDest++;
		}

	}while(lpDest);

	if(lstrlen(lpszMessageClass) != 0)
		strQuery += " OR messageclass=''";

	strQuery += ") ORDER BY length(messageclass) DESC LIMIT 1";

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;


	if(lpDatabase->GetNumRows(lpDBResult) == 1) {
		lpDBRow = lpDatabase->FetchRow(lpDBResult);

		if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL){
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(atoui(lpDBRow[0]), soap, &lpsReceiveFolder->sReceiveFolder.sEntryId);
		if(er != erSuccess)
			goto exit;

		nLen = strlen(lpDBRow[1])+1;

		lpsReceiveFolder->sReceiveFolder.lpszAExplicitClass = s_alloc<char>(soap, nLen);

		memcpy(lpsReceiveFolder->sReceiveFolder.lpszAExplicitClass, lpDBRow[1], nLen);
	}else{
		//items not found
		er = ZARAFA_E_NOT_FOUND;
	}


exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

// FIXME: should be able to delete an entry too
SOAP_ENTRY_START4(setReceiveFolder, *result, entryId sStoreId, entryId* lpsEntryId, char* lpszMessageClass, unsigned int *result)
{
	bool			bIsUpdate = false;
	unsigned int	ulCheckStoreId = 0;
	unsigned int	ulStoreid = 0;
	unsigned int	ulId = 0;
	USE_DATABASE();

	er = lpDatabase->Begin();
	if (er != erSuccess)
		goto exit;

	// Check, lpsEntryId and lpszMessageClass can't both be empty or 0
	if(lpsEntryId == NULL && (lpszMessageClass == NULL || lpszMessageClass == '\0') ){
		er = ZARAFA_E_INVALID_TYPE;
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sStoreId, &ulStoreid);
	if(er != erSuccess)
	    goto exit;

	// an empty lpszMessageClass is the default folder
	if(lpszMessageClass == NULL)
		lpszMessageClass = "";

	// If the lpsEntryId parameter is set to NULL then replace the current receive folder with the message store's default.
	if(lpsEntryId)
	{
		// Check if object really exist and the relation between storeid and ulId

		er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(lpsEntryId, &ulId);
		if(er != erSuccess)
			goto exit;

		// Check if storeid and ulId have a relation
		er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulId, &ulCheckStoreId, NULL);
		if(er != erSuccess){
			er = ZARAFA_E_INVALID_ENTRYID;
			goto exit;
		}

		if(ulStoreid != ulCheckStoreId) {
			er = ZARAFA_E_INVALID_ENTRYID;
			goto exit;
		}

	} else {
		// Set MessageClass with the default of the store (that's the empty MessageClass)
		strQuery = "SELECT objid FROM receivefolder WHERE storeid="+stringify(ulStoreid)+" AND messageclass=''";
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		if(lpDatabase->GetNumRows(lpDBResult) == 1)
		{
			lpDBRow = lpDatabase->FetchRow(lpDBResult);

			if(lpDBRow == NULL || lpDBRow[0] == NULL){
				er = ZARAFA_E_DATABASE_ERROR;
				goto exit;
			}

			//Set the default folder
			ulId = atoi(lpDBRow[0]);
		}else{
			er = ZARAFA_E_DATABASE_ERROR; //FIXME: no default error ?
			goto exit;
		}

		FREE_DBRESULT();
	}


	strQuery = "SELECT objid, id FROM receivefolder WHERE storeid="+stringify(ulStoreid)+" AND messageclass='"+lpDatabase->Escape(lpszMessageClass)+"'";
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	bIsUpdate = false;
	// If ok, item already exists, return ok
	if(lpDatabase->GetNumRows(lpDBResult) == 1){
		lpDBRow = lpDatabase->FetchRow(lpDBResult);

		if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL){
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		// Item exists
		if(ulId == atoui(lpDBRow[0])){
			er = erSuccess;
			goto exit;
		}

		bIsUpdate = true;
	}

	//Free database results
	FREE_DBRESULT();

	// Check permission
	//FIXME: also on delete?
	if(bIsUpdate)
		er = lpecSession->GetSecurity()->CheckPermission(ulStoreid, ecSecurityEdit);
	else
		er = lpecSession->GetSecurity()->CheckPermission(ulStoreid, ecSecurityCreate);
	if(er != erSuccess)
		goto exit;


	if(bIsUpdate) {
		strQuery = "UPDATE receivefolder SET objid="+stringify(ulId);
		strQuery+= " WHERE storeid="+stringify(ulStoreid)+" AND messageclass='"+lpDatabase->Escape(lpszMessageClass)+"'";
	}else{
		strQuery = "INSERT INTO receivefolder (storeid, objid, messageclass) VALUES (";
		strQuery += stringify(ulStoreid)+", "+stringify(ulId)+", '"+lpDatabase->Escape(lpszMessageClass)+"')";
	}

	if(bIsUpdate)
		er = lpDatabase->DoUpdate(strQuery);
	else
		er = lpDatabase->DoUpdate(strQuery);

	if(er != erSuccess)
		goto exit;

	er = lpDatabase->Commit();
	if (er != erSuccess)
		goto exit;

exit:
    FREE_DBRESULT();
	ROLLBACK_ON_ERROR();
}
SOAP_ENTRY_END()

/*
 * WARNING
 *
 * lpsEntryID != NULL && lpMessageList != NULL: messages in lpMessageList must be set, lpsEntryId MUST BE IGNORED (may be entryid of search folder)
 * lpsEntryID == NULL && lpMessageList != NULL: called from IMessage::SetReadFlag, lpMessageList->__size == 1
 * lpsEntryID != NULL && lpMessageList == NULL: 'mark all messages as (un)read'
 *
 */
SOAP_ENTRY_START5(setReadFlags, *result, unsigned int ulFlags, entryId* lpsEntryId, struct entryList *lpMessageList, unsigned int ulSyncId, unsigned int *result)
{
	ECListInt		lObjectIds;
	ECListIntIterator iObjectid;
	std::string		strQueryCache;
	USE_DATABASE();

	unsigned int	i = 0;
	unsigned int	ulParent = 0;
	unsigned int	ulGrandParent = 0;
	unsigned int	ulFlagsNotify = 0;
	unsigned int	ulFlagsRemove = 0;
	unsigned int	ulFlagsAdd = 0;
	unsigned int	ulFolderId = 0;
	size_t			cObjectSize = 0;
	
	// List of unique parents
	std::set<unsigned int> setParents;
	std::set<unsigned int>::iterator iterParents;

	//NOTE: either lpMessageList may be NULL or lpsEntryId may be NULL

	if(ulFlags & GENERATE_RECEIPT_ONLY)
		goto exit;

    if(lpMessageList == NULL && lpsEntryId == NULL) {
        // Bad input
        er = ZARAFA_E_INVALID_PARAMETER;
        goto exit;
    }

	if(lpMessageList) {
		// Ignore errors
		g_lpSessionManager->GetCacheManager()->GetEntryListToObjectList(lpMessageList, &lObjectIds);
	}

	strQuery = "UPDATE hierarchy SET ";

	if ((ulFlags & CLEAR_NRN_PENDING) || (ulFlags & SUPPRESS_RECEIPT) || (ulFlags & GENERATE_RECEIPT_ONLY) )
		ulFlagsRemove |= MSGFLAG_NRN_PENDING;

	if ((ulFlags & CLEAR_RN_PENDING) || (ulFlags & SUPPRESS_RECEIPT) || (ulFlags & GENERATE_RECEIPT_ONLY) )
		ulFlagsRemove |= MSGFLAG_RN_PENDING;

    if (!(ulFlags & GENERATE_RECEIPT_ONLY) && (ulFlags & CLEAR_READ_FLAG))
        ulFlagsRemove |= MSGFLAG_READ;
	else if( !(ulFlags & GENERATE_RECEIPT_ONLY) )
        ulFlagsAdd |= MSGFLAG_READ;

	if(ulFlagsRemove != 0)
		strQuery += "flags=flags & ~" + stringify(ulFlagsRemove);

	if(ulFlagsAdd != 0) {
		strQuery += (ulFlagsRemove!=0)?",":"";
		strQuery += "flags=flags | " + stringify(ulFlagsAdd);
	}

	if(ulFlagsRemove == 0 && ulFlagsAdd == 0)
	{
		// Nothing to update
		goto exit;
	}

	er = lpDatabase->Begin();
	if(er != hrSuccess)
		goto exit;

	if(lpMessageList == NULL) {
	    // No message list passed, so 'mark all items (un)read'
        er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(lpsEntryId, &ulFolderId);
        if(er != erSuccess)
            goto exit;

        // Check permission
        er = lpecSession->GetSecurity()->CheckPermission(ulFolderId, ecSecurityRead);
        if(er != erSuccess)
            goto exit;

		// Get all items MAPI_MESSAGE exclude items with flags MSGFLAG_DELETED AND MSGFLAG_ASSOCIATED of which we will be changing flags
		strQueryCache = "SELECT id FROM hierarchy WHERE parent="+ stringify(ulFolderId) + " AND hierarchy.type=5 AND flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " = 0 AND (flags & " + stringify(ulFlagsRemove) + " OR flags & " + stringify(ulFlagsAdd) + " != " + stringify(ulFlagsAdd) + ")";
		er = lpDatabase->DoSelect(strQueryCache, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
		{
			if(lpDBRow[0] == NULL){
				er = ZARAFA_E_DATABASE_ERROR;
				goto exit;
			}

			lObjectIds.push_back( atoui(lpDBRow[0]) );
            i++;
		}

        FREE_DBRESULT();
	} else {
	    // Because the messagelist can contain messages from all over the place, we have to check permissions for all the parent folders of the items
	    // we are setting 'read' or 'unread'
		for(iObjectid = lObjectIds.begin(); iObjectid != lObjectIds.end(); iObjectid++)
		{

			// Get the parent object. Note that the cache will hold this information so the loop below with GetObject() will
			// be done directly from the cache (assuming it's not too large)
			if(g_lpSessionManager->GetCacheManager()->GetObject(*iObjectid, &ulParent, NULL, NULL) != erSuccess) {
			    continue;
			}

            // Check permission (may be the same folder over and over)
            er = lpecSession->GetSecurity()->CheckPermission(ulParent, ecSecurityRead);
            if(er != erSuccess)
                goto exit;
        }
	}

	// Security passed, do the query.

    strQuery += " WHERE hierarchy.id IN(";

    for(iObjectid = lObjectIds.begin(); iObjectid != lObjectIds.end(); iObjectid++) {

        if(iObjectid != lObjectIds.begin())
            strQuery += ",";

        strQuery += stringify(*iObjectid);
    }
    strQuery += ")";

	// Make sure we only update MAPI_MESSAGE types
	strQuery += " AND hierarchy.type=5 AND hierarchy.flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " = 0";

	if(! lObjectIds.empty()) {
    	// Update the database
    	er = lpDatabase->DoUpdate(strQuery);
    	if(er != erSuccess)
	    	goto exit;
    }

    // Add changes to ICS
    for(iObjectid = lObjectIds.begin(); iObjectid != lObjectIds.end(); iObjectid++) {
        if( (ulFlagsRemove & MSGFLAG_READ) || (ulFlagsAdd & MSGFLAG_READ) ) {
            // Only save ICS change when the actual readflag has changed
            SOURCEKEY		sSourceKey;
            SOURCEKEY		sParentSourceKey;

			if(g_lpSessionManager->GetCacheManager()->GetObject(*iObjectid, &ulParent, NULL, NULL) != erSuccess) {
			    continue;
			}

            GetSourceKey(*iObjectid, &sSourceKey);
            GetSourceKey(ulParent, &sParentSourceKey);

            // Because we know that ulFlagsRemove && MSGFLAG_READ || ulFlagsAdd & MSGFLAG_READ and we assume
            // that they are never both TRUE, we can ignore ulFlagsRemove and just look at ulFlagsAdd for the new
            // readflag state
            AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_MESSAGE_FLAG, NULL, ulFlagsAdd & MSGFLAG_READ);
        }
    }

	er = lpDatabase->Commit();
    if(er != erSuccess)
	    goto exit;

	// Now, update cache and send the notifications

	cObjectSize = lObjectIds.size();

    // Loop through the messages, updating each
    for(iObjectid = lObjectIds.begin(); iObjectid != lObjectIds.end(); iObjectid++)
    {
		// Remove the item from the cache (FIXME: USING fnevObjectDeleted instead of fnevObjectModified BECAUSE FLAGS ARE CHANGING, SO OBJECT CACHE MUST BE INVALIDATED)
        g_lpSessionManager->GetCacheManager()->Update(fnevObjectDeleted, *iObjectid);

        if(g_lpSessionManager->GetCacheManager()->GetObject(*iObjectid, &ulParent, NULL, &ulFlagsNotify) != erSuccess) {
            ulParent = 0;
            ulFlagsNotify = 0;
		} else {
			// Group parents for notification
			setParents.insert(ulParent);
		}

        // Update the message itself in tables and object notification
        g_lpSessionManager->NotificationModified(MAPI_MESSAGE, *iObjectid, ulParent);

        if(ulParent &&  cObjectSize < EC_TABLE_CHANGE_THRESHOLD)
            g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, ulFlagsNotify&MSGFLAG_NOTIFY_FLAGS, ulParent, *iObjectid, MAPI_MESSAGE);
    }

    // Loop through all the parent folders of the objects, sending notifications for them
    for(iterParents = setParents.begin(); iterParents != setParents.end(); iterParents++) {
        // The parent has changed its PR_CONTENT_UNREAD
        g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, *iterParents);
        g_lpSessionManager->NotificationModified(MAPI_FOLDER, *iterParents);

        // The grand parent's table view of the parent has changed
        if(g_lpSessionManager->GetCacheManager()->GetObject(*iterParents, &ulGrandParent, NULL, &ulFlagsNotify) == erSuccess)
            g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, ulFlagsNotify&MSGFLAG_NOTIFY_FLAGS, ulGrandParent, *iterParents, MAPI_FOLDER);

        if(cObjectSize >= EC_TABLE_CHANGE_THRESHOLD)
            g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, ulFlagsNotify&MSGFLAG_NOTIFY_FLAGS, *iterParents, 0, MAPI_MESSAGE);
    }

exit:
    ROLLBACK_ON_ERROR();
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(createUser, lpsUserSetResponse->er, struct user *lpsUser, struct setUserResponse *lpsUserSetResponse)
{
	unsigned int		ulUserId = 0;
	unsigned int		ulCompanyId = 0;
	userobject_type_t	ulType;
	objectdetails_t		details;

	if (lpsUser == NULL || lpsUser->lpszUsername == NULL || lpsUser->lpszFullName == NULL || lpsUser->lpszMailAddress == NULL ||
		(lpsUser->lpszPassword == NULL && lpsUser->ulIsNonActive == 0)) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (lpsUser->ulIsNonActive)
		ulType = USEROBJECT_TYPE_NONACTIVE;
	else
		ulType = USEROBJECT_TYPE_USER;

	details = objectdetails_t(ulType);

	er = CopyUserDetailsFromSoap(lpsUser, NULL, &details, soap);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->ConvertLoginToUserAndCompany(&details);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY,
																(char *)details.GetPropString(OB_PROP_S_COMPANYNAME).c_str(),
																&ulCompanyId);
	if (er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyId);
	if(er != erSuccess)
		goto exit;

    // Create user and sync
	er = lpecSession->GetUserManagement()->CreateObjectAndSync(ulType, details, &ulUserId);
	if(er != erSuccess)
	    goto exit;

	er = GetABEntryID(ulUserId, soap, &lpsUserSetResponse->sUserId);
	if (er != erSuccess)
		goto exit;

	lpsUserSetResponse->ulUserId = ulUserId;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(setUser, *result, struct user *lpsUser, unsigned int *result)
{
	objectdetails_t		details;
	objectdetails_t		oldDetails;
	unsigned int		ulUserId = 0;
	std::string			strExternId;
	userobject_type_t	sUserObjType = USEROBJECT_TYPE_USER;

	if(lpsUser == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (lpsUser->sUserId.__size > 0 && lpsUser->sUserId.__ptr != NULL)
	{
		er = GetLocalId(NULL, lpsUser->sUserId, lpsUser->ulUserId, &ulUserId, &strExternId, NULL);
		if (er != erSuccess) 
			goto exit;
	}
	else
		ulUserId = lpsUser->ulUserId;

	if(ulUserId) {
		er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulUserId, &oldDetails);
		if (er != erSuccess)
			goto exit;
	}

	// Check security
	if (lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserId) == erSuccess) {
		// admins can update anything of a user
		// FIXME: prevent the user from removing admin rights from itself?
		er = erSuccess;
	} else if (lpecSession->GetSecurity()->GetUserId() == ulUserId) {
		// you're only allowed to set your password, force the lpsUser struct to only contain that update
		if (lpsUser->lpszUsername && oldDetails.GetPropString(OB_PROP_S_LOGIN) != lpsUser->lpszUsername) {
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Disallowing user %s to update its username to %s",
												 oldDetails.GetPropString(OB_PROP_S_LOGIN).c_str(), lpsUser->lpszUsername);
			lpsUser->lpszUsername = NULL;
		}

		// leave lpszPassword

		if (lpsUser->lpszMailAddress && oldDetails.GetPropString(OB_PROP_S_EMAIL) != lpsUser->lpszMailAddress) {
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Disallowing user %s to update its mail address to %s",
												 oldDetails.GetPropString(OB_PROP_S_LOGIN).c_str(), lpsUser->lpszMailAddress);
			lpsUser->lpszMailAddress = NULL;
		}

		if (lpsUser->lpszFullName && oldDetails.GetPropString(OB_PROP_S_FULLNAME) != lpsUser->lpszFullName) {
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Disallowing user %s to update its fullname to %s",
												 oldDetails.GetPropString(OB_PROP_S_LOGIN).c_str(), lpsUser->lpszFullName);
			lpsUser->lpszFullName = NULL;
		}

		if (lpsUser->lpszServername && oldDetails.GetPropString(OB_PROP_S_SERVERNAME) != lpsUser->lpszServername) {
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Disallowing user %s to update its home server to %s",
												 oldDetails.GetPropString(OB_PROP_S_LOGIN).c_str(), lpsUser->lpszServername);
			lpsUser->lpszServername = NULL;
		}

		if (lpsUser->ulIsNonActive != (ULONG)-1 && oldDetails.GetPropBool(OB_PROP_B_NONACTIVE) != lpsUser->ulIsNonActive) {
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Disallowing user %s to update its active flag to %d",
												 oldDetails.GetPropString(OB_PROP_S_LOGIN).c_str(), lpsUser->ulIsNonActive);
			lpsUser->ulIsNonActive = (ULONG)-1;
		}

		if (lpsUser->ulIsAdmin != (ULONG)-1 && oldDetails.GetPropInt(OB_PROP_I_ADMINLEVEL) != lpsUser->ulIsAdmin) {
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Disallowing user %s to update its admin flag to %d",
												 oldDetails.GetPropString(OB_PROP_S_LOGIN).c_str(), lpsUser->ulIsAdmin);
			lpsUser->ulIsAdmin = (ULONG)-1;
		}
	} else {
		// you cannot set any details if you're not an admin or not yourself
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Disallowing user %s to update details of user %s",
											 oldDetails.GetPropString(OB_PROP_S_LOGIN).c_str(), lpsUser->lpszUsername);
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	if (lpsUser->ulIsNonActive)
		sUserObjType = USEROBJECT_TYPE_NONACTIVE;
	else
		sUserObjType = USEROBJECT_TYPE_USER;

	details = objectdetails_t(sUserObjType);

	// construct new details
	er = CopyUserDetailsFromSoap(lpsUser, &strExternId, &details, soap);
	if (er != erSuccess)
		goto exit;

	if(lpsUser->lpszUsername){
		er = lpecSession->GetUserManagement()->ConvertLoginToUserAndCompany(&details);
		if (er != erSuccess)
			goto exit;
	}

	er = lpecSession->GetUserManagement()->SetObjectDetailsAndSync(ulUserId , details, NULL);

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getUser, lpsGetUserResponse->er, unsigned int ulUserId, entryId sUserId, struct getUserResponse *lpsGetUserResponse)
{
	objectdetails_t	details;
	entryId			sTmpUserId = {0};

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	/* Check if we are able to view the returned userobject */
	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulUserId);
	if (er != erSuccess)
		goto exit;

	lpsGetUserResponse->lpsUser = s_alloc<user>(soap);
	memset(lpsGetUserResponse->lpsUser, 0, sizeof(struct user));

	if(ulUserId == 0)
        ulUserId = lpecSession->GetSecurity()->GetUserId();

    er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulUserId, &details);
    if(er != erSuccess)
        goto exit;

	if (details.GetType() != USEROBJECT_TYPE_USER && details.GetType() != USEROBJECT_TYPE_NONACTIVE) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	er = GetABEntryID(ulUserId, soap, &sTmpUserId);
	if (er == erSuccess)
		er = CopyUserDetailsToSoap(ulUserId, &sTmpUserId, details, soap, lpsGetUserResponse->lpsUser);
	if (er != erSuccess)
		goto exit;

exit:;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getUserList, lpsUserList->er, unsigned int ulCompanyId, entryId sCompanyId, struct userListResponse *lpsUserList)
{
	std::list<localobjectdetails_t> *lpUsers = NULL;
	std::list<localobjectdetails_t>::iterator iterUsers;

	std::string	strExternId;
	entryId		sUserEid = {0};

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	/* Input check, if ulCompanyId is 0, we want the user's company,
	 * otherwise we must check if the requested company is visible for the user. */
	if (ulCompanyId == 0) {
		er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
		if (er != erSuccess)
			goto exit;
	} else {
		er = lpecSession->GetSecurity()->IsUserObjectVisible(ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	er = lpecSession->GetUserManagement()->GetMergedObjectListAndSync(ECUSEROBJECT_LIST_USERS_ACTIVE | ECUSEROBJECT_LIST_USERS_NONACTIVE,
																	  ulCompanyId, &lpUsers);
	if(er != erSuccess)
		goto exit;

    lpsUserList->sUserArray.__size = 0;
    lpsUserList->sUserArray.__ptr = s_alloc<user>(soap, lpUsers->size());

    for(iterUsers = lpUsers->begin(); iterUsers != lpUsers->end(); iterUsers++) {
		er = GetABEntryID(iterUsers->ulId, soap, &sUserEid);
		if (er != erSuccess)
			goto exit;

		er = CopyUserDetailsToSoap(iterUsers->ulId, &sUserEid, *iterUsers, soap, &lpsUserList->sUserArray.__ptr[lpsUserList->sUserArray.__size]);
		if (er != erSuccess)
			goto exit;

        lpsUserList->sUserArray.__size++;
    }

exit:
    if(lpUsers)
        delete lpUsers;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getSendAsList, lpsUserList->er, unsigned int ulUserId, entryId sUserId, struct userListResponse *lpsUserList)
{
	objectdetails_t userDetails, senderDetails;
	list<string> userIds;
	list<string>::iterator iterUserIds;
	unsigned int ulSenderId;
	entryId sSenderEid = {0};

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulUserId);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulUserId, &userDetails);
	if (er != erSuccess)
		goto exit;

	userIds = userDetails.GetPropList(OB_PROP_L_SENDAS);

	lpsUserList->sUserArray.__size = 0;
	lpsUserList->sUserArray.__ptr = s_alloc<user>(soap, userIds.size());

	for (iterUserIds = userIds.begin(); iterUserIds != userIds.end(); iterUserIds++) {
		ulSenderId = atoi(iterUserIds->c_str());

		er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulSenderId, &senderDetails);
		if (er == ZARAFA_E_NOT_FOUND)
			continue;
		if (er != erSuccess)
			goto exit;

		er = GetABEntryID(ulSenderId, soap, &sSenderEid);
		if (er != erSuccess)
			goto exit;

		er = CopyUserDetailsToSoap(ulSenderId, &sSenderEid, senderDetails, soap, &lpsUserList->sUserArray.__ptr[lpsUserList->sUserArray.__size]);
		if (er != erSuccess)
			goto exit;

        lpsUserList->sUserArray.__size++;
	}
	er = erSuccess;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(addSendAsUser, *result, unsigned int ulUserId, entryId sUserId, unsigned int ulSenderId, entryId sSenderId, unsigned int *result)
{

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sSenderId, ulSenderId, &ulSenderId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	if (ulUserId == ulSenderId) {
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	// Check security, only admins can set sendas users, not the user itself
	if(lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserId) != erSuccess) {
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	// needed?
	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulUserId);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->AddSubObjectToObjectAndSync(OBJECTRELATION_USER_SENDAS, ulUserId, ulSenderId);
	
exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(delSendAsUser, *result, unsigned int ulUserId, entryId sUserId, unsigned int ulSenderId, entryId sSenderId, unsigned int *result)
{

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sSenderId, ulSenderId, &ulSenderId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	if (ulUserId == ulSenderId) {
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	// Check security, only admins can set sendas users, not the user itself
	if(lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserId) != erSuccess) {
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	// needed ?
	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulUserId);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->DeleteSubObjectFromObjectAndSync(OBJECTRELATION_USER_SENDAS, ulUserId, ulSenderId);

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(purgeSoftDelete, *result, unsigned int ulDays, unsigned int *result)
{
    unsigned int	ulFolders = 0;
    unsigned int	ulMessages = 0;

    // Only system-admins may run this
    if(lpecSession->GetSecurity()->GetAdminLevel() < ADMIN_LEVEL_SYSADMIN) {
        er = ZARAFA_E_NO_ACCESS;
        goto exit;
    }

    g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Start forced softdelete clean up");

    er = PurgeSoftDelete(lpecSession, ulDays * 24 * 60 * 60, &ulMessages, &ulFolders);

    if (er == erSuccess)
        g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Softdelete done: removed folders: %d, removed messages: %d", ulFolders, ulMessages);
    else
        g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Softdelete failed: removed folders: %d, removed messages: %d", ulFolders, ulMessages);

exit:
    ;
}
SOAP_ENTRY_END()

//Create a store
// Info: Userid can also be a group id ('everyone' for public store)
SOAP_ENTRY_START6(createStore, *result, unsigned int ulStoreType, unsigned int ulUserId, entryId sUserId, entryId sStoreId, entryId sRootId, unsigned int *result)
{
	unsigned int	ulStoreId = 0;
	unsigned int	ulRootMapId = 0;
	objectdetails_t userDetails;
	unsigned int	ulCompanyId = 0;
	bool			bHasLocalStore = false;

	SOURCEKEY		sSourceKey;
	GUID			guidStore;

	struct rightsArray srightsArray;
	USE_DATABASE();

	memset(&srightsArray, 0 , sizeof(srightsArray));


	er = GetLocalId(lpecSession, sUserId, ulUserId, &ulUserId, NULL, &bHasLocalStore);
	if (er != erSuccess)
		goto exit;

	if (!bHasLocalStore) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Started to create store (userid=%d)", ulUserId);

	er = lpDatabase->Begin();
	if (er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserId);
	if (er != erSuccess)
		goto exit;

	// Get object details, and resolve company
	er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulUserId, &userDetails);
	if (er != erSuccess)
		goto exit;

	if (lpecSession->GetSessionManager()->IsHostedSupported()) {
		er = lpecSession->GetUserManagement()->ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY,
																	(char *)userDetails.GetPropString(OB_PROP_S_COMPANYNAME).c_str(),
																	&ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	// Validate store entryid
	if(ValidateZarafaEntryId(sStoreId.__size, sStoreId.__ptr, MAPI_STORE) == false) {
		er = ZARAFA_E_INVALID_ENTRYID;
		goto exit;
	}

	// Validate root entryid
	if(ValidateZarafaEntryId(sRootId.__size, sRootId.__ptr, MAPI_FOLDER) == false) {
		er = ZARAFA_E_INVALID_ENTRYID;
		goto exit;
	}

	er = GetStoreGuidFromEntryId(sStoreId.__size, sStoreId.__ptr, &guidStore);
	if(er != erSuccess)
		goto exit;

	// Check if already a store on the user or group
	strQuery = "SELECT id FROM stores WHERE user_id="+stringify(ulUserId)+" OR guid="+lpDatabase->EscapeBinary((unsigned char*)&guidStore , sizeof(GUID));
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) > 0) {
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	FREE_DBRESULT();

	// Create Toplevel of the store
	strQuery = "INSERT INTO hierarchy(parent, modtime, createtime, type, owner) VALUES(NULL, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP, "+stringify(MAPI_STORE)+", "+ stringify(ulUserId)+")";
	er = lpDatabase->DoInsert(strQuery, &ulStoreId);
	if(er != erSuccess)
		goto exit;

	// Create the rootfolder of a store
	strQuery = "INSERT INTO hierarchy(parent, modtime, createtime, type, owner) VALUES("+stringify(ulStoreId)+", CURRENT_TIMESTAMP, CURRENT_TIMESTAMP, "+stringify(MAPI_FOLDER)+ ", "+ stringify(ulUserId)+")";
	er = lpDatabase->DoInsert(strQuery, &ulRootMapId);
	if(er != erSuccess)
		goto exit;

	//Init storesize
	UpdateObjectSize(lpDatabase, ulStoreId, ulStoreId, MAPI_STORE, true, UPDATE_NEW, 0);


	// Add SourceKey

	er = lpecSession->GetNewSourceKey(1, &sSourceKey);
	if(er != erSuccess)
		goto exit;

	strQuery = "INSERT INTO indexedproperties(hierarchyid,tag,val_binary) VALUES(" + stringify(ulRootMapId) + "," + stringify(PROP_ID(PR_SOURCE_KEY)) + "," + lpDatabase->EscapeBinary(sSourceKey, sSourceKey.size()) + ")";
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

	// Add store entryid: 0x0FFF = PR_ENTRYID
	strQuery = "INSERT INTO indexedproperties (hierarchyid,tag,val_binary) VALUES("+stringify(ulStoreId)+", 0x0FFF, "+lpDatabase->EscapeBinary(sStoreId.__ptr, sStoreId.__size)+")";
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

	// Add rootfolder entryid: 0x0FFF = PR_ENTRYID
	strQuery = "INSERT INTO indexedproperties (hierarchyid,tag,val_binary) VALUES("+stringify(ulRootMapId)+", 0x0FFF, "+lpDatabase->EscapeBinary(sRootId.__ptr, sRootId.__size)+")";
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

	// Couple store with user
	strQuery = "INSERT INTO stores(hierarchy_id, user_id, user_name, company, guid) VALUES(" +
		stringify(ulStoreId) + ", " +
		stringify(ulUserId) + ", " +
		"'" + lpDatabase->Escape(userDetails.GetPropString(OB_PROP_S_LOGIN)) + "', " +
		stringify(ulCompanyId) + ", " +
		lpDatabase->EscapeBinary((unsigned char*)&guidStore , sizeof(GUID))+")";
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;


	// Set ACL's on public store
	if(ulStoreType == ECSTORE_TYPE_PUBLIC) {
		// ulUserId == a group
		// ulUserId 1 = group everyone

		srightsArray.__ptr = new rights[1];
		srightsArray.__ptr[0].ulRights = ecRightsDefaultPublic;
		srightsArray.__ptr[0].ulUserid = ulUserId;
		srightsArray.__ptr[0].ulState = RIGHT_NEW|RIGHT_AUTOUPDATE_DENIED;
		srightsArray.__ptr[0].ulType = ACCESS_TYPE_GRANT;
		memset(&srightsArray.__ptr[0].sUserId, 0, sizeof(entryId));

		srightsArray.__size = 1;
		er = lpecSession->GetSecurity()->SetRights(ulStoreId, &srightsArray);
		if(er != erSuccess)
			goto exit;
	}

	er = lpDatabase->Commit();
	if (er != erSuccess)
		goto exit;

	g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Finished create store (userid=%d, storeid=%d)", ulUserId, ulStoreId);

	er = g_lpSessionManager->SaveSourceKeyAutoIncrement();
	if (er != erSuccess)
		goto exit;

exit:
	if(er == ZARAFA_E_NO_ACCESS)
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Failed to create store access denied");
	else if(er != erSuccess)
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Failed to create store (id=%d), errorcode=0x%08X", ulUserId, er);

	if(srightsArray.__size > 0)
		delete [] srightsArray.__ptr;

    FREE_DBRESULT();
	ROLLBACK_ON_ERROR();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(createGroup, lpsSetGroupResponse->er, struct group *lpsGroup, struct setGroupResponse *lpsSetGroupResponse)
{
	unsigned int			ulGroupId = 0;
	unsigned int 			ulCompanyId = 0;
	objectdetails_t			details;

	if (lpsGroup == NULL || lpsGroup->lpszGroupname == NULL || lpsGroup->lpszFullname == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	details = objectdetails_t(USEROBJECT_TYPE_GROUP);

	er = CopyGroupDetailsFromSoap(lpsGroup, NULL, &details, soap);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->ConvertLoginToUserAndCompany(&details);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY,
																(char *)details.GetPropString(OB_PROP_S_COMPANYNAME).c_str(),
																&ulCompanyId);
	if (er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyId);
	if(er != erSuccess)
		goto exit;

    er = lpecSession->GetUserManagement()->CreateObjectAndSync(USEROBJECT_TYPE_GROUP, details, &ulGroupId);
    if(er != erSuccess)
        goto exit;

	er = GetABEntryID(ulGroupId, soap, &lpsSetGroupResponse->sGroupId);
	if (er != erSuccess)
		goto exit;

	lpsSetGroupResponse->ulGroupId = ulGroupId;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(setGroup, *result, struct group *lpsGroup, unsigned int *result)
{
	objectdetails_t details;
	unsigned int	ulGroupId = 0;
	std::string		strExternId;
	
	if(lpsGroup == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (lpsGroup->sGroupId.__size > 0 && lpsGroup->sGroupId.__ptr != NULL)
	{
		er = GetLocalId(NULL, lpsGroup->sGroupId, lpsGroup->ulGroupId, &ulGroupId, &strExternId, NULL);
		if (er != erSuccess) 
			goto exit;
	}
	else
		ulGroupId = lpsGroup->ulGroupId;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulGroupId);
	if(er != erSuccess)
		goto exit;

	details = objectdetails_t(USEROBJECT_TYPE_GROUP);

	er = CopyGroupDetailsFromSoap(lpsGroup, &strExternId, &details, soap);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->ConvertLoginToUserAndCompany(&details);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->SetObjectDetailsAndSync(ulGroupId, details, NULL);
	if(er != erSuccess)
	    goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getGroup, lpsResponse->er, unsigned int ulGroupId, entryId sGroupId, struct getGroupResponse *lpsResponse)
{
	objectdetails_t details;
	entryId			sTmpGroupId = {0};

	er = GetLocalId(NULL, sGroupId, ulGroupId, &ulGroupId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	/* Check if we are able to view the returned userobject */
	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulGroupId);
	if (er != erSuccess)
		goto exit;

    er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulGroupId, &details);
    if(er != erSuccess)
        goto exit;

	if (details.GetType() != USEROBJECT_TYPE_GROUP) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lpsResponse->lpsGroup = s_alloc<group>(soap);

	er = GetABEntryID(ulGroupId, soap, &sTmpGroupId);
	if (er == erSuccess)
		er = CopyGroupDetailsToSoap(ulGroupId, &sTmpGroupId, details, soap, lpsResponse->lpsGroup);
	if (er != erSuccess)
		goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getGroupList, lpsGroupList->er, unsigned int ulCompanyId, entryId sCompanyId, struct groupListResponse *lpsGroupList)
{
	std::list<localobjectdetails_t> *lpGroups = NULL;
	std::list<localobjectdetails_t>::iterator iterGroups;

	entryId	sGroupEid = {0};

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	/* Input check, if ulCompanyId is 0, we want the user's company,
	 * otherwise we must check if the requested company is visible for the user. */
	if (ulCompanyId == 0) {
		er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
		if (er != erSuccess)
			goto exit;
	} else {
		er = lpecSession->GetSecurity()->IsUserObjectVisible(ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	er = lpecSession->GetUserManagement()->GetObjectListAndSync(USEROBJECT_TYPE_GROUP, ulCompanyId, &lpGroups);
	if (er != erSuccess)
		goto exit;

	lpsGroupList->sGroupArray.__size = 0;
	lpsGroupList->sGroupArray.__ptr = s_alloc<group>(soap, lpGroups->size());
	for(iterGroups = lpGroups->begin(); iterGroups != lpGroups->end(); iterGroups++)
	{
		er = GetABEntryID(iterGroups->ulId, soap, &sGroupEid);
		if (er != erSuccess)
			goto exit;

		er = CopyGroupDetailsToSoap(iterGroups->ulId, &sGroupEid, *iterGroups, soap, &lpsGroupList->sGroupArray.__ptr[lpsGroupList->sGroupArray.__size]);
		if (er != erSuccess)
			goto exit;

	    lpsGroupList->sGroupArray.__size++;
	}

exit:
    if (lpGroups)
        delete lpGroups;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(groupDelete, *result, unsigned int ulGroupId, entryId sGroupId, unsigned int *result)
{
	er = GetLocalId(NULL, sGroupId, ulGroupId, &ulGroupId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulGroupId);
	if(er != erSuccess)
		goto exit;

    er = lpecSession->GetUserManagement()->DeleteObjectAndSync(ulGroupId);

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(resolveUsername, lpsResponse->er, char *lpszUsername, struct resolveUserResponse *lpsResponse)
{
	unsigned int		ulUserId = 0;
	userobject_type_t	sObjType = USEROBJECT_TYPE_USER;

	if(lpszUsername == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpecSession->GetUserManagement()->ResolveObjectAndSync(sObjType, lpszUsername, &ulUserId);
	if (er == ZARAFA_E_NOT_FOUND) {
		sObjType = USEROBJECT_TYPE_NONACTIVE;
        er = lpecSession->GetUserManagement()->ResolveObjectAndSync(sObjType, lpszUsername, &ulUserId);
    }
    if(er != erSuccess)
        goto exit;

	/* Check if we are able to view the returned userobject */
	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulUserId);
	if (er != erSuccess)
		goto exit;
		
	er = GetABEntryID(ulUserId, soap, &lpsResponse->sUserId);
	if (er != erSuccess)
		goto exit;

	lpsResponse->ulUserId = ulUserId;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(resolveGroupname, lpsResponse->er, char *lpszGroupname, struct resolveGroupResponse *lpsResponse)
{
	unsigned int	ulGroupId = 0;

	if(lpszGroupname == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

    er = lpecSession->GetUserManagement()->ResolveObjectAndSync(USEROBJECT_TYPE_GROUP, lpszGroupname, &ulGroupId);
    if(er != erSuccess)
        goto exit;

	/* Check if we are able to view the returned userobject */
	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulGroupId);
	if (er != erSuccess)
		goto exit;

	er = GetABEntryID(ulGroupId, soap, &lpsResponse->sGroupId);
	if (er != erSuccess)
		goto exit;

	lpsResponse->ulGroupId = ulGroupId;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(deleteGroupUser, *result, unsigned int ulGroupId, entryId sGroupId, unsigned int ulUserId, entryId sUserId, unsigned int *result)
{
	er = GetLocalId(NULL, sGroupId, ulGroupId, &ulGroupId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulGroupId);
	if(er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

    er = lpecSession->GetUserManagement()->DeleteSubObjectFromObjectAndSync(OBJECTRELATION_GROUP_MEMBER, ulGroupId, ulUserId);

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(addGroupUser, *result, unsigned int ulGroupId, entryId sGroupId, unsigned int ulUserId, entryId sUserId, unsigned int *result)
{
	er = GetLocalId(NULL, sGroupId, ulGroupId, &ulGroupId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulGroupId);
	if(er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

    er = lpecSession->GetUserManagement()->AddSubObjectToObjectAndSync(OBJECTRELATION_GROUP_MEMBER, ulGroupId, ulUserId);

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getUserListOfGroup, lpsUserList->er, unsigned int ulGroupId, entryId sGroupId, struct userListResponse *lpsUserList)
{
	std::list<localobjectdetails_t> *lpUsers = NULL;
	std::list<localobjectdetails_t>::iterator iterUsers;
	std::list<localobjectdetails_t> *lpNonactives = NULL;
	std::list<std::list<localobjectdetails_t> *> lstSources;
	std::list<std::list<localobjectdetails_t> *>::iterator iterSources;

	std::string	strExternId;
	entryId		sUserEid = {0};

	er = GetLocalId(NULL, sGroupId, ulGroupId, &ulGroupId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulGroupId);
	if (er != erSuccess)
		goto exit;

    er = lpecSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(OBJECTRELATION_GROUP_MEMBER, USEROBJECT_TYPE_USER, ulGroupId, &lpUsers);
    if(er != erSuccess)
        goto exit;

    er = lpecSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(OBJECTRELATION_GROUP_MEMBER, USEROBJECT_TYPE_NONACTIVE, ulGroupId, &lpNonactives);
    if(er != erSuccess)
        goto exit;

    lpsUserList->sUserArray.__size = 0;
    lpsUserList->sUserArray.__ptr = s_alloc<user>(soap, lpUsers->size() + lpNonactives->size());

    // Just a construction so we can loop through everything in one go
    lstSources.push_back(lpUsers);
    lstSources.push_back(lpNonactives);

    for(iterSources = lstSources.begin(); iterSources != lstSources.end(); iterSources++) {
        for(iterUsers = (*iterSources)->begin(); iterUsers != (*iterSources)->end(); iterUsers++) {
			if (lpecSession->GetSecurity()->IsUserObjectVisible(iterUsers->ulId) != erSuccess) {
				continue;
			}

			er = GetABEntryID(iterUsers->ulId, soap, &sUserEid);
			if (er != erSuccess)
				goto exit;

			er = CopyUserDetailsToSoap(iterUsers->ulId, &sUserEid, *iterUsers, soap, &lpsUserList->sUserArray.__ptr[lpsUserList->sUserArray.__size]);
			if (er != erSuccess)
				goto exit;

            lpsUserList->sUserArray.__size++;
        }
    }

exit:
    if(lpUsers)
        delete lpUsers;

    if(lpNonactives)
        delete lpNonactives;

}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getGroupListOfUser, lpsGroupList->er, unsigned int ulUserId, entryId sUserId, struct groupListResponse *lpsGroupList)
{
	std::list<localobjectdetails_t> *lpGroups = NULL;
	std::list<localobjectdetails_t>::iterator iterGroups;

	entryId sGroupEid = {0};

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulUserId);
	if (er != erSuccess)
		goto exit;

    er = lpecSession->GetUserManagement()->GetParentObjectsOfObjectAndSync(OBJECTRELATION_GROUP_MEMBER, USEROBJECT_TYPE_GROUP, ulUserId, &lpGroups);
    if(er != erSuccess)
        goto exit;

	lpsGroupList->sGroupArray.__size = 0;
	lpsGroupList->sGroupArray.__ptr = s_alloc<group>(soap, lpGroups->size());
	for(iterGroups = lpGroups->begin(); iterGroups != lpGroups->end(); iterGroups++) {
		if (lpecSession->GetSecurity()->IsUserObjectVisible(iterGroups->ulId) != erSuccess) {
			continue;
		}

		er = GetABEntryID(iterGroups->ulId, soap, &sGroupEid);
		if (er != erSuccess)
			goto exit;

		er = CopyGroupDetailsToSoap(iterGroups->ulId, &sGroupEid, *iterGroups, soap, &lpsGroupList->sGroupArray.__ptr[lpsGroupList->sGroupArray.__size]);
		if (er != erSuccess)
			goto exit;

		lpsGroupList->sGroupArray.__size++;
	}

exit:
    if(lpGroups)
        delete lpGroups;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(createCompany, lpsResponse->er, struct company *lpsCompany, struct setCompanyResponse *lpsResponse)
{
	unsigned int ulCompanyId = 0;
	objectdetails_t details;

	if (lpsCompany == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	// Check permission, only the system user is allowed to create or delete a company
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ZARAFA_UID_SYSTEM);
	if(er != erSuccess)
		goto exit;

	er = CopyCompanyDetailsFromSoap(lpsCompany, NULL, "", &details, soap);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->CreateObjectAndSync(USEROBJECT_TYPE_COMPANY, details, &ulCompanyId);
	if(er != erSuccess)
		goto exit;

	er = GetABEntryID(ulCompanyId, soap, &lpsResponse->sCompanyId);
	if (er != erSuccess)
		goto exit;

	lpsResponse->ulCompanyId = ulCompanyId;

exit:
;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(deleteCompany, *result, unsigned int ulCompanyId, entryId sCompanyId, unsigned int *result)
{
	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	// Check permission, only the system user is allowed to create or delete a company
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ZARAFA_UID_SYSTEM);
	if(er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->DeleteObjectAndSync(ulCompanyId);

exit:
;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(setCompany, *result, struct company *lpsCompany, unsigned int *result)
{
	objectdetails_t details;
	objectdetails_t userdetails;
	unsigned int	ulCompanyId = 0;
	unsigned int	ulAdministrator = 0;
	std::string		strExternId;
	
	if(lpsCompany == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	if (lpsCompany->sCompanyId.__size > 0 && lpsCompany->sCompanyId.__ptr != NULL)
	{
		er = GetLocalId(NULL, lpsCompany->sCompanyId, lpsCompany->ulCompanyId, &ulCompanyId, &strExternId, NULL);
		if (er != erSuccess) 
			goto exit;
	}
	else
		ulCompanyId = lpsCompany->ulCompanyId;

	er = GetLocalId(NULL, lpsCompany->sAdministrator, lpsCompany->ulAdministrator, &ulAdministrator, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyId);
	if(er != erSuccess)
		goto exit;

	details = objectdetails_t(USEROBJECT_TYPE_COMPANY);

	if(ulAdministrator) {
		er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulAdministrator, &userdetails);
		if (er != erSuccess)
			goto exit;
	}

	er = CopyCompanyDetailsFromSoap(lpsCompany, &strExternId, (char*)userdetails.GetPropString(OB_PROP_S_LOGIN).c_str(), &details, soap);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->SetObjectDetailsAndSync(ulCompanyId, details, NULL);
	if(er != erSuccess)
		goto exit;

exit:
;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getCompany, lpsResponse->er, unsigned int ulCompanyId, entryId sCompanyId, struct getCompanyResponse *lpsResponse)
{
	objectdetails_t details;
	unsigned int ulAdmin = 0;

	entryId sAdminEid = {0};
	entryId sTmpCompanyId = {0};

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	/* Input check, if ulCompanyId is 0, we want the user's company,
	 * otherwise we must check if the requested company is visible for the user. */
	if (ulCompanyId == 0) {
		er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
		if (er != erSuccess)
			goto exit;
	} else {
		er = lpecSession->GetSecurity()->IsUserObjectVisible(ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulCompanyId, &details);
	if(er != erSuccess)
		goto exit;

	if (details.GetType() != USEROBJECT_TYPE_COMPANY) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lpsResponse->lpsCompany = s_alloc<company>(soap);

	er = lpecSession->GetUserManagement()->SearchObjectAndSync(USEROBJECT_TYPE_USER,
															   (char *)details.GetPropString(OB_PROP_S_SYSADMIN).c_str(),
															   EMS_AB_ADDRESS_LOOKUP, &ulAdmin);
	if (er != erSuccess) {
		er = erSuccess;
		ulAdmin = ZARAFA_UID_SYSTEM;
	}
	
	er = GetABEntryID(ulAdmin, soap, &sAdminEid);
	if (er != erSuccess)
		goto exit;

	er = GetABEntryID(ulCompanyId, soap, &sTmpCompanyId);
	if (er == erSuccess)
		er = CopyCompanyDetailsToSoap(ulCompanyId, &sTmpCompanyId, ulAdmin, &sAdminEid, details, soap, lpsResponse->lpsCompany);
	if (er != erSuccess)
		goto exit;

exit:
;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(resolveCompanyname, lpsResponse->er, char *lpszCompanyname, struct resolveCompanyResponse *lpsResponse)
{
	unsigned int ulCompanyId = 0;

	if(lpszCompanyname == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = lpecSession->GetUserManagement()->ResolveObjectAndSync(USEROBJECT_TYPE_COMPANY, lpszCompanyname, &ulCompanyId);
	if(er != erSuccess)
		goto exit;

	/* Check if we are able to view the returned userobject */
	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulCompanyId);
	if (er != erSuccess)
		goto exit;

	er = GetABEntryID(ulCompanyId, soap, &lpsResponse->sCompanyId);
	if (er != erSuccess)
		goto exit;
	
	lpsResponse->ulCompanyId = ulCompanyId;

exit:
;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START1(getCompanyList, lpsCompanyList->er, struct companyListResponse *lpsCompanyList)
{
	unsigned int	ulAdmin = 0;
	entryId			sCompanyEid = {0};
	entryId			sAdminEid = {0};
	std::list<localobjectdetails_t> *lpCompanies = NULL;
	std::list<localobjectdetails_t>::iterator iterCompanies;

	er = lpecSession->GetUserManagement()->GetObjectListAndSync(USEROBJECT_TYPE_COMPANY, 0, &lpCompanies);
	if(er != erSuccess)
		goto exit;

	lpsCompanyList->sCompanyArray.__size = 0;
	lpsCompanyList->sCompanyArray.__ptr = s_alloc<company>(soap, lpCompanies->size());
	for(iterCompanies = lpCompanies->begin(); iterCompanies != lpCompanies->end(); iterCompanies++) {
		er = lpecSession->GetUserManagement()->SearchObjectAndSync(USEROBJECT_TYPE_USER,
																   (char *)iterCompanies->GetPropString(OB_PROP_S_SYSADMIN).c_str(),
																   EMS_AB_ADDRESS_LOOKUP, &ulAdmin);
		if (er != erSuccess) {
			er = erSuccess;
			ulAdmin = ZARAFA_UID_SYSTEM;
		}

		er = GetABEntryID(iterCompanies->ulId, soap, &sCompanyEid);
		if (er != erSuccess)
			goto exit;
			
		er = GetABEntryID(ulAdmin, soap, &sAdminEid);
		if (er != erSuccess)
			goto exit;

		er = CopyCompanyDetailsToSoap(iterCompanies->ulId, &sCompanyEid, ulAdmin, &sAdminEid, *iterCompanies, soap, &lpsCompanyList->sCompanyArray.__ptr[lpsCompanyList->sCompanyArray.__size]);
		if (er != erSuccess)
			goto exit;

		lpsCompanyList->sCompanyArray.__size++;
	}

exit:
	if(lpCompanies)
		delete lpCompanies;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(addCompanyToRemoteViewList, *result, unsigned int ulSetCompanyId, entryId sSetCompanyId, unsigned int ulCompanyId, entryId sCompanyId, unsigned int *result)
{
	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyId);
	if(er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sSetCompanyId, ulSetCompanyId, &ulSetCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->AddSubObjectToObjectAndSync(OBJECTRELATION_COMPANY_VIEW, ulCompanyId, ulSetCompanyId);
	if(er != erSuccess)
		goto exit;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(delCompanyFromRemoteViewList, *result, unsigned int ulSetCompanyId, entryId sSetCompanyId, unsigned int ulCompanyId, entryId sCompanyId, unsigned int *result)
{
	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyId);
	if(er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sSetCompanyId, ulSetCompanyId, &ulSetCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->DeleteSubObjectFromObjectAndSync(OBJECTRELATION_COMPANY_VIEW, ulCompanyId, ulSetCompanyId);
	if(er != erSuccess)
		goto exit;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getRemoteViewList, lpsCompanyList->er, unsigned int ulCompanyId, entryId sCompanyId, struct companyListResponse *lpsCompanyList)
{
	unsigned int	ulAdmin = 0;
	entryId			sCompanyEid = {0};
	entryId			sAdminEid = {0};

	std::list<localobjectdetails_t> *lpCompanies = NULL;
	std::list<localobjectdetails_t>::iterator iterCompanies;

	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	/* Input check, if ulCompanyId is 0, we want the user's company,
	 * otherwise we must check if the requested company is visible for the user. */
	if (ulCompanyId == 0) {
		er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
		if (er != erSuccess)
			goto exit;
	} else {
		er = lpecSession->GetSecurity()->IsUserObjectVisible(ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	er = lpecSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(OBJECTRELATION_COMPANY_VIEW, USEROBJECT_TYPE_COMPANY,
																		ulCompanyId, &lpCompanies);
	if(er != erSuccess)
		goto exit;

	lpsCompanyList->sCompanyArray.__size = 0;
	lpsCompanyList->sCompanyArray.__ptr = s_alloc<company>(soap, lpCompanies->size());

	for(iterCompanies = lpCompanies->begin(); iterCompanies != lpCompanies->end(); iterCompanies++) {
		if (lpecSession->GetSecurity()->IsUserObjectVisible(iterCompanies->ulId) != erSuccess) {
			continue;
		}

		er = lpecSession->GetUserManagement()->SearchObjectAndSync(USEROBJECT_TYPE_USER,
																   (char *)iterCompanies->GetPropString(OB_PROP_S_SYSADMIN).c_str(),
																   EMS_AB_ADDRESS_LOOKUP, &ulAdmin);
		if (er != erSuccess) {
			er = erSuccess;
			ulAdmin = ZARAFA_UID_SYSTEM;
		}

		er = GetABEntryID(iterCompanies->ulId, soap, &sCompanyEid);
		if (er != erSuccess)
			goto exit;
			
		er = GetABEntryID(ulAdmin, soap, &sAdminEid);
		if (er != erSuccess)
			goto exit;

		er = CopyCompanyDetailsToSoap(iterCompanies->ulId, &sCompanyEid, ulAdmin, &sAdminEid, *iterCompanies, soap, &lpsCompanyList->sCompanyArray.__ptr[lpsCompanyList->sCompanyArray.__size]);
		if (er != erSuccess)
			goto exit;

		lpsCompanyList->sCompanyArray.__size++;
	}

exit:
	if(lpCompanies)
		delete lpCompanies;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(addUserToRemoteAdminList, *result, unsigned int ulUserId, entryId sUserId, unsigned int ulCompanyId, entryId sCompanyId, unsigned int *result)
{
	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyId);
	if(er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->AddSubObjectToObjectAndSync(OBJECTRELATION_COMPANY_ADMIN, ulCompanyId, ulUserId);
	if(er != erSuccess)
		goto exit;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(delUserFromRemoteAdminList, *result, unsigned int ulUserId, entryId sUserId, unsigned int ulCompanyId, entryId sCompanyId, unsigned int *result)
{
	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyId);
	if(er != erSuccess)
		goto exit;

	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->DeleteSubObjectFromObjectAndSync(OBJECTRELATION_COMPANY_ADMIN, ulCompanyId, ulUserId);
	if(er != erSuccess)
		goto exit;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getRemoteAdminList, lpsUserList->er, unsigned int ulCompanyId, entryId sCompanyId, struct userListResponse *lpsUserList)
{
	std::list<localobjectdetails_t> *lpUsers = NULL;
	std::list<localobjectdetails_t>::iterator iterUsers;

	std::string	strExternId;
	entryId		sUserEid = {0};

	if (!g_lpSessionManager->IsHostedSupported()) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	/* Input check, if ulCompanyId is 0, we want the user's company,
	 * otherwise we must check if the requested company is visible for the user. */
	if (ulCompanyId == 0) {
		er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
		if (er != erSuccess)
			goto exit;
	} else {
		er = lpecSession->GetSecurity()->IsUserObjectVisible(ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	// only users can be admins, nonactive users make no sense.
	er = lpecSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(OBJECTRELATION_COMPANY_ADMIN, USEROBJECT_TYPE_USER,
																		ulCompanyId, &lpUsers);
	if(er != erSuccess)
		goto exit;

	lpsUserList->sUserArray.__size = 0;
	lpsUserList->sUserArray.__ptr = s_alloc<user>(soap, lpUsers->size());

	for(iterUsers = lpUsers->begin(); iterUsers != lpUsers->end(); iterUsers++) {
		if (lpecSession->GetSecurity()->IsUserObjectVisible(iterUsers->ulId) != erSuccess) {
			continue;
		}

		er = GetABEntryID(iterUsers->ulId, soap, &sUserEid);
		if (er != erSuccess)
			goto exit;

		er = CopyUserDetailsToSoap(iterUsers->ulId, &sUserEid, *iterUsers, soap, &lpsUserList->sUserArray.__ptr[lpsUserList->sUserArray.__size]);
		if (er != erSuccess)
			goto exit;

		lpsUserList->sUserArray.__size++;

		if (sUserEid.__ptr)
		{
			// sUserEid is placed in userdetails, no need to free
			sUserEid.__ptr = NULL;
			sUserEid.__size = 0;
		}
	}

exit:
	if(lpUsers)
		delete lpUsers;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(submitMessage, *result, entryId sEntryId, unsigned int ulFlags, unsigned int *result)
{
	unsigned int	ulParentId	= 0;
	unsigned int	ulObjId = 0;
	unsigned int	ulMsgFlags 	= 0;
	unsigned int	ulStoreId = 0;
	unsigned int	ulStoreOwner = 0;
	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey;
	bool			bMessageChanged = false;

	eQuotaStatus	QuotaStatus;
	long long		llStoreSize = 0;
	objectdetails_t details;
	
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if(er != erSuccess)
	    goto exit;

	er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulObjId, &ulStoreId, NULL);
	if(er != erSuccess)
	    goto exit;

    er = lpecSession->GetSessionManager()->GetCacheManager()->GetObject(ulObjId, &ulParentId, NULL, &ulMsgFlags, NULL);
    if(er != erSuccess)
        goto exit;
        
    er = lpecSession->GetSessionManager()->GetCacheManager()->GetObject(ulStoreId, NULL, &ulStoreOwner, NULL, NULL);
    if(er != erSuccess)
        goto exit;
     
    er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulStoreOwner, &details);
    if(er != erSuccess)
        goto exit;
         
    // Cannot submit a message in a public store   
    if(details.GetType() != USEROBJECT_TYPE_USER && details.GetType() != USEROBJECT_TYPE_NONACTIVE) {
        er = ZARAFA_E_NO_ACCESS;
        goto exit;
    }

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulStoreId, ecSecurityOwner);
	if(er != erSuccess)
		goto exit;

	/////////////////////////////////////////////////
	// Quota check
	er = lpecSession->GetSecurity()->GetStoreSize(ulStoreId, &llStoreSize);
	if(er != erSuccess)
		goto exit;

	er = lpecSession->GetSecurity()->CheckQuota(ulStoreId, llStoreSize, &QuotaStatus);
	if(er != erSuccess)
		goto exit;

	if(QuotaStatus == QUOTA_SOFTLIMIT || QuotaStatus == QUOTA_HARDLIMIT) {
		er = ZARAFA_E_STORE_FULL;
		goto exit;
	}

	er = lpDatabase->Begin();
	if(er != hrSuccess)
		goto exit;

	// Set PR_MESSAGE_FLAGS to MSGFLAG_SUBMIT|MSGFLAG_UNSENT
	if(!(ulFlags & EC_SUBMIT_MASTER)) {
	    // Set the submit flag (because it has just been submitted), and set it to UNSENT, as it has definitely
	    // not been sent if the user has just submitted it.
		strQuery = "UPDATE hierarchy SET flags=flags|"+stringify(MSGFLAG_SUBMIT|MSGFLAG_UNSENT)+" where id="+stringify(ulObjId);
		er = lpDatabase->DoUpdate(strQuery);
		if(er != erSuccess)
			goto exit;

		// Add change to ICS
		GetSourceKey(ulObjId, &sSourceKey);
		GetSourceKey(ulParentId, &sParentSourceKey);

		AddChange(lpecSession, 0, sSourceKey, sParentSourceKey, ICS_MESSAGE_CHANGE);

		// Mask for notification
		bMessageChanged = true;
	}

	// Insert the message into the outgoing queue
	strQuery = "INSERT INTO outgoingqueue (store_id, hierarchy_id, flags) VALUES("+stringify(ulStoreId)+", "+stringify(ulObjId)+","+stringify(ulFlags)+")";
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

	er = lpDatabase->Commit();
	if(er != erSuccess)
		goto exit;

	if (bMessageChanged) {
		// Update cache
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulObjId);
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulParentId);

		// Notify
		g_lpSessionManager->NotificationModified(MAPI_MESSAGE, ulObjId, ulParentId);
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulParentId, ulObjId, MAPI_MESSAGE);
	}
	
	g_lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_ADD, ulStoreId, ulObjId, ulFlags, MAPI_MESSAGE);

exit:
    ROLLBACK_ON_ERROR();
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(finishedMessage, *result,  entryId sEntryId, unsigned int ulFlags, unsigned int *result)
{
	unsigned int	ulParentId	= 0;
	unsigned int	ulGrandParentId = 0;
	unsigned int	ulAffectedRows = 0;
	unsigned int	ulStoreId = (unsigned int)-1; // not 0 security issue
	unsigned int	ulObjId = 0;
	bool			bMessageChanged = false;
	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey;

	USE_DATABASE();

	er = lpDatabase->Begin();
	if(er != hrSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if(er != erSuccess)
	    goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetParent(ulObjId, &ulParentId);
	if(er != erSuccess)
		goto exit;

	//Get storeid
	er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulObjId, &ulStoreId, NULL);
	switch (er) {
	case erSuccess:
		break;
	case ZARAFA_E_NOT_FOUND:
		// ulObjId should be in outgoingtable, but the ulStoreId cannot be retrieved
		// because ulObjId does not exist in the hierarchy table, so we remove the message
		// fix table and notify and pass error to caller

		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_WARNING, "Unable to find store for hierarchy id %d", ulObjId);

		ulStoreId = 0;
		goto table;
	default:
		goto exit;				// database error
	}

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulStoreId, ecSecurityOwner);
	if(er != erSuccess)
		goto exit;

	strQuery = "UPDATE hierarchy ";

	if(!(ulFlags & EC_SUBMIT_MASTER)) {
        // Removing from local queue; remove submit flag and unsent flag
	    strQuery += " SET flags=flags&~"+stringify(MSGFLAG_SUBMIT|MSGFLAG_UNSENT);
    } else {
        // Removing from master queue
    	if(ulFlags & EC_SUBMIT_DOSENTMAIL) {
            // Spooler sent message and moved, remove submit flag and unsent flag
    	    strQuery += " SET flags=flags&~" +stringify(MSGFLAG_SUBMIT|MSGFLAG_UNSENT);
        } else {
            // Spooler only sent message
            strQuery += " SET flags=flags&~" +stringify(MSGFLAG_UNSENT);
        }
    }

    // Always set message read
    strQuery += ", flags=flags|" + stringify(MSGFLAG_READ) + " WHERE ID="+stringify(ulObjId);

   	er = lpDatabase->DoUpdate(strQuery);
   	if(er != erSuccess)
   		goto exit;

	GetSourceKey(ulObjId, &sSourceKey);
	GetSourceKey(ulParentId, &sParentSourceKey);

	AddChange(lpecSession, 0, sSourceKey, sParentSourceKey, ICS_MESSAGE_CHANGE);
    
	// NOTE: Unlock message is done in client

	// Mark for notification
	bMessageChanged = true;

table:

	// delete the message from the outgoing queue
	strQuery = "DELETE FROM outgoingqueue WHERE hierarchy_id="+stringify(ulObjId) + " AND flags & 1=" + stringify(ulFlags & 1);
	er = lpDatabase->DoDelete(strQuery, &ulAffectedRows);
	if(er != erSuccess)
		goto exit;

	er = lpDatabase->Commit();
	if(er != erSuccess)
		goto exit;


	// Remove messge from the outgoing queue
	g_lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_DELETE, ulStoreId, ulObjId, ulFlags, MAPI_MESSAGE);

	// The flags have changed, so we have to send a modified 
	if (bMessageChanged) {
		lpecSession->GetSessionManager()->GetCacheManager()->Update(fnevObjectModified, ulObjId);
		g_lpSessionManager->NotificationModified(MAPI_MESSAGE, ulObjId, ulParentId);

		if(g_lpSessionManager->GetCacheManager()->GetParent(ulObjId, &ulParentId) == erSuccess) {
			g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulParentId);
			g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulParentId);
	        
			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulParentId, ulObjId, MAPI_MESSAGE);
			if(g_lpSessionManager->GetCacheManager()->GetParent(ulParentId, &ulGrandParentId) == erSuccess) {
				g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParentId, ulParentId, MAPI_FOLDER);
			}
		}
	}

exit:
    ROLLBACK_ON_ERROR();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(abortSubmit, *result, entryId sEntryId, unsigned int *result)
{
	unsigned int	ulParentId	= 0;
	unsigned int	ulStoreId	= 0;
	unsigned int	ulSubmitFlags = 0;
	unsigned int	ulGrandParentId = 0;
	unsigned int	ulObjId = 0;
	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey;

	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if(er != erSuccess)
	    goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetParent(ulObjId, &ulParentId);
	if(er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityOwner);
	if(er != erSuccess)
		goto exit;

	er = lpDatabase->Begin();
	if(er != hrSuccess)
		goto exit;

	// Get storeid and check if the message into the queue
	strQuery = "SELECT store_id, flags FROM outgoingqueue WHERE hierarchy_id="+stringify(ulObjId);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;
// FIXME: can be also more than 2??
	if(lpDatabase->GetNumRows(lpDBResult) != 1){
		er = ZARAFA_E_NOT_IN_QUEUE;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	ulStoreId = atoui(lpDBRow[0]);
	ulSubmitFlags = atoi(lpDBRow[1]);

	// delete the message from the outgoing queue
	strQuery = "DELETE FROM outgoingqueue WHERE hierarchy_id="+stringify(ulObjId);
	er = lpDatabase->DoDelete(strQuery);
	if(er != erSuccess)
		goto exit;

	// remove in property PR_MESSAGE_FLAGS the MSGFLAG_SUBMIT flag
	strQuery = "UPDATE hierarchy SET flags=flags& ~"+stringify(MSGFLAG_SUBMIT)+" WHERE id="+stringify(ulObjId);
	er = lpDatabase->DoUpdate(strQuery);
	if(er != erSuccess)
		goto exit;
	
	// Update ICS system
	GetSourceKey(ulObjId, &sSourceKey);
	GetSourceKey(ulParentId, &sParentSourceKey);

	AddChange(lpecSession, 0, sSourceKey, sParentSourceKey, ICS_MESSAGE_CHANGE);
    
	er = lpDatabase->Commit();
	if(er != erSuccess)
		goto exit;

	g_lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_DELETE, ulStoreId, ulObjId, ulSubmitFlags, MAPI_MESSAGE);

	if(g_lpSessionManager->GetCacheManager()->GetParent(ulObjId, &ulParentId) == erSuccess) {

        g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulObjId);
		g_lpSessionManager->NotificationModified(MAPI_MESSAGE, ulObjId, ulParentId);
        g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulParentId);
		g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulParentId);


		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulParentId, ulObjId, MAPI_MESSAGE);
		if(g_lpSessionManager->GetCacheManager()->GetParent(ulParentId, &ulGrandParentId) == erSuccess) {
			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParentId, ulParentId, MAPI_FOLDER);
		}
	}

exit:
    ROLLBACK_ON_ERROR();
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(isMessageInQueue, *result, entryId sEntryId, unsigned int *result)
{
	unsigned int	ulObjId = 0;
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if(er != erSuccess)
	    goto exit;

	// Checks if message is unsent
	strQuery = "SELECT hierarchy_id FROM outgoingqueue WHERE hierarchy_id=" + stringify(ulObjId) + " AND flags & " + stringify(EC_SUBMIT_MASTER);

	if(lpDatabase->DoSelect(strQuery, &lpDBResult) != erSuccess) {
	    er = ZARAFA_E_DATABASE_ERROR;
	    goto exit;
    }

    if(lpDatabase->GetNumRows(lpDBResult) == 0) {
        er = ZARAFA_E_NOT_FOUND;
        goto exit;
    }

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(resolveStore, lpsResponse->er, unsigned int ulStoreId, struct resolveUserStoreResponse *lpsResponse)
{
	USE_DATABASE();

	if (!ulStoreId) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	strQuery =
		"SELECT u.id, s.hierarchy_id, s.guid, s.company "
		"FROM stores AS s "
		"LEFT JOIN users AS u "
			"ON s.user_id = u.id "
		"WHERE s.id=" + stringify(ulStoreId);
	if(lpDatabase->DoSelect(strQuery, &lpDBResult) != erSuccess) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if(lpDatabase->GetNumRows(lpDBResult) != 1) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if(lpDBRow == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL || lpDBRow[3] == NULL || lpDBLen == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	if (lpDBRow[0] == NULL) {
		// check if we're admin over the store object
		er = lpecSession->GetSecurity()->IsUserObjectAdmin(atoi(lpDBRow[3]));
		if (er != erSuccess)
			goto exit;

		lpsResponse->ulUserId = 0;
		lpsResponse->sUserId.__size = 0;
		lpsResponse->sUserId.__ptr = NULL;
	} else {
		lpsResponse->ulUserId = atoi(lpDBRow[0]);

		er = GetABEntryID(lpsResponse->ulUserId, soap, &lpsResponse->sUserId);
		if (er != erSuccess)
			goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(atoui(lpDBRow[1]), soap, &lpsResponse->sStoreId);
	if(er != erSuccess)
		goto exit;

	lpsResponse->guid.__size = lpDBLen[2];
	lpsResponse->guid.__ptr = s_alloc<unsigned char>(soap, lpDBLen[2]);
	memcpy(lpsResponse->guid.__ptr, lpDBRow[2], lpDBLen[2]);

exit:
	FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(resolveUserStore, lpsResponse->er, char *szUserName, struct resolveUserStoreResponse *lpsResponse)
{
	unsigned int		ulObjectId = 0;
	userobject_type_t	ulObjectTypes[] = {
		USEROBJECT_TYPE_USER,
		USEROBJECT_TYPE_NONACTIVE,
		USEROBJECT_TYPE_COMPANY,
	};
	objectdetails_t		sUserDetails;
	string				strServerName;
    string              strServerPath;

	USE_DATABASE();

	if (szUserName == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	for (unsigned int i = 0; i < (sizeof(ulObjectTypes) / sizeof(ulObjectTypes[0])); i++) {
		er = lpecSession->GetUserManagement()->ResolveObjectAndSync(ulObjectTypes[i], szUserName, &ulObjectId);
		if (er == erSuccess)
			break;
	}
	if (er != erSuccess)
		goto exit;

	/* If we are allowed to view the user, we are allowed to know the store exists */
	er = lpecSession->GetSecurity()->IsUserObjectVisible(ulObjectId);
	if (er != erSuccess)
		goto exit;
		
	er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulObjectId, &sUserDetails);
	if (er != erSuccess)
		goto exit;

	if (lpecSession->GetSessionManager()->IsDistributedSupported() &&
		!lpecSession->GetUserManagement()->IsLocalItem(ulObjectId)) {
		
		/* We found an object ID. Check if this is the correct server for its store */
		strServerName = sUserDetails.GetPropString(OB_PROP_S_SERVERNAME);
		if (strServerName.empty()) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}			
		if (strServerName.compare(g_lpSessionManager->GetConfig()->GetSetting("server_name")) != 0) {
			er = GetBestServerPath(soap, lpecSession, strServerName, &strServerPath);
			if (er != erSuccess)
				goto exit;
			
			lpsResponse->lpszServerPath = s_strcpy(soap, strServerPath.c_str());
			g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Redirecting request to '%s'", lpsResponse->lpszServerPath);
			g_lpStatsCollector->Increment(SCN_REDIRECT_COUNT, 1);
			er = ZARAFA_E_UNABLE_TO_COMPLETE;
			goto exit;
		} 
	}
	
	strQuery = "SELECT hierarchy_id, guid FROM stores WHERE user_id = " + stringify(ulObjectId);
	if(lpDatabase->DoSelect(strQuery, &lpDBResult) != erSuccess) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if (lpDBRow == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

    /* We found the store, so we don't need to check if this is the correct server. */
    strServerName = g_lpSessionManager->GetConfig()->GetSetting("server_name", "", "Unknown");
    // Always return the pseudo URL.
    lpsResponse->lpszServerPath = s_strcpy(soap, string("pseudo://" + strServerName).c_str());

	er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(atoui(lpDBRow[0]), soap, &lpsResponse->sStoreId);
	if(er != erSuccess)
	    goto exit;

	er = GetABEntryID(ulObjectId, soap, &lpsResponse->sUserId);
	if (er != erSuccess)
		goto exit;

	lpsResponse->ulUserId = ulObjectId;
	lpsResponse->guid.__size = lpDBLen[1];
	lpsResponse->guid.__ptr = s_alloc<unsigned char>(soap, lpDBLen[1]);
	memcpy(lpsResponse->guid.__ptr, lpDBRow[1], lpDBLen[1]);

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

typedef struct{
	unsigned int ulId;
	unsigned int ulType;
	unsigned int ulParent;
	unsigned int ulNewId;
	unsigned int ulFlags;
	SOURCEKEY 	 sSourceKey;
	SOURCEKEY	 sParentSourceKey;
}COPYITEM;

// Move one or more messages and/or moved a softdeleted message to a normal message
ECRESULT MoveObjects(ECSession *lpSession, ECDatabase *lpDatabase, ECListInt* lplObjectIds, unsigned int ulDestFolderId, unsigned int ulSyncId)
{
	ECRESULT		er = erSuccess;
	bool			bPartialCompletion = false;
	unsigned int	ulGrandParent = 0;
	COPYITEM		sItem;
	unsigned int	ulItemSize = 0;
	unsigned int	ulSourceStoreId = 0;
	unsigned int	ulDestStoreId = 0;
	long long		llStoreSize;
	eQuotaStatus	QuotaStatus;
	bool			bUpdateDeletedSize = false;
	ECListIntIterator	iObjectId;
	size_t			cCopyItems = 0;

	unsigned int	ulSourceId = 0;
	unsigned long long ullIMAP = 0;

	std::list<unsigned int> lstParent;
	std::list<unsigned int>::iterator iterParent;
	std::list<unsigned int> lstGrandParent;
	std::list<unsigned int>::iterator iterGrandParent;
	std::list<COPYITEM> lstCopyItems;
	std::list<COPYITEM>::iterator iterCopyItems;

	SOURCEKEY	sSourceKey;
	SOURCEKEY	sDestFolderSourceKey;

	entryId*	lpsNewEntryId = NULL;
	entryId*	lpsOldEntryId = NULL;
	GUID		guidStore;

	ALLOC_DBRESULT();

	if(lplObjectIds == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if(lplObjectIds->empty())
		goto exit; // Nothing to do

	// Check permission, Destination folder
	er = lpSession->GetSecurity()->CheckPermission(ulDestFolderId, ecSecurityCreate);
	if(er != erSuccess)
		goto exit;

	er = lpSession->GetSessionManager()->GetCacheManager()->GetStore(ulDestFolderId, &ulDestStoreId, &guidStore);
	if(er != erSuccess)
		goto exit;

	// Get from the first object the store id
	ulSourceId = *lplObjectIds->begin();

	GetSourceKey(ulDestFolderId, &sDestFolderSourceKey);

	// Get all items for the object list
	strQuery = "SELECT h.id, h.parent, h.type, h.flags, p.val_ulong FROM hierarchy AS h LEFT JOIN properties AS p ON p.hierarchyid=h.id AND p.tag="+stringify(PROP_ID(PR_MESSAGE_SIZE))+" AND p.type="+stringify(PROP_TYPE(PR_MESSAGE_SIZE)) + " WHERE h.id IN(";

	for(iObjectId = lplObjectIds->begin(); iObjectId != lplObjectIds->end(); iObjectId++) {
		if(iObjectId != lplObjectIds->begin())
			strQuery += ",";

		strQuery += stringify(*iObjectId);
	}
	strQuery += ")";

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	// object doesn't exist, ignore it. FIXME we should issue a warning to the client
	//if(lpDatabase->GetNumRows(lpDBResult) != lpEntryList->__size)
	// Then do something

	// First, put all the root objects in the list
	while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) != NULL)
	{
		if(lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL || lpDBRow[3] == NULL) // no id, type or parent folder?
			continue;

		sItem.ulId		= atoi(lpDBRow[0]);
		sItem.ulParent	= atoi(lpDBRow[1]);
		sItem.ulType	= atoi(lpDBRow[2]);
		sItem.ulFlags	= atoi(lpDBRow[3]);

		if (sItem.ulType != MAPI_MESSAGE) {
			bPartialCompletion = true;
			continue;
		}

		GetSourceKey(sItem.ulId, &sItem.sSourceKey);
		GetSourceKey(sItem.ulParent, &sItem.sParentSourceKey);

		// Check permission, source messages
		er = lpSession->GetSecurity()->CheckPermission(sItem.ulId, ecSecurityDelete);
		if(er == erSuccess) {

			// Check if the source and dest the same store
		    er = lpSession->GetSessionManager()->GetCacheManager()->GetStore(sItem.ulId, &ulSourceStoreId, NULL);
			if(er != erSuccess || ulSourceStoreId != ulDestStoreId) {
				bPartialCompletion = true;
				er = erSuccess;
				continue;
			}

			lstCopyItems.push_back(sItem);
			ulItemSize += (lpDBRow[4] != NULL)? atoi(lpDBRow[4]) : 0;

			// check if it a deleted item
			if(lpDBRow[3] != NULL && (atoi(lpDBRow[3])&MSGFLAG_DELETED ) == MSGFLAG_DELETED)
				bUpdateDeletedSize = true;
		}else {
			bPartialCompletion = true;
			er = erSuccess;
		}
	}

	// Free database results
	if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL;}

	// Check the quota size when the item is a softdelete item
	if(bUpdateDeletedSize == true)
	{
		/////////////////////////////////////////////////
		// Quota check
		er = lpSession->GetSecurity()->GetStoreSize(ulDestFolderId, &llStoreSize);
		if(er != erSuccess)
			goto exit;

		// substract itemsize and check
		llStoreSize -= (llStoreSize >= (long long)ulItemSize)?(long long)ulItemSize:0;
		er = lpSession->GetSecurity()->CheckQuota(ulDestFolderId, llStoreSize, &QuotaStatus);
		if(er != erSuccess)
			goto exit;

		if(QuotaStatus == QUOTA_HARDLIMIT) {
			er = ZARAFA_E_STORE_FULL;
			goto exit;
		}

	}

	cCopyItems = lstCopyItems.size();

	// Move the messages to an other folder
	for(iterCopyItems=lstCopyItems.begin(); iterCopyItems != lstCopyItems.end(); iterCopyItems++) {

		// Check or it is a move to the same parent, skip them
		if(iterCopyItems->ulParent == ulDestFolderId && (iterCopyItems->ulFlags&MSGFLAG_DELETED) == 0)
			continue;

		er = lpDatabase->Begin();
		if(er != hrSuccess)
			goto exit;

		FreeEntryId(lpsOldEntryId, true);
		lpsOldEntryId = NULL;

		er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(iterCopyItems->ulId, NULL, &lpsOldEntryId);
		if(er != erSuccess) {
			bPartialCompletion = true;
			er = erSuccess;
			// FIXME: Delete from list: iterCopyItems
			continue;
		}

		FreeEntryId(lpsNewEntryId, true);
		lpsNewEntryId = NULL;

		er = CreateEntryId(guidStore, MAPI_MESSAGE, &lpsNewEntryId);
		if(er != erSuccess)
			goto exit;

        // Update entryid (changes on move)
		strQuery = "REPLACE INTO indexedproperties(hierarchyid,tag,val_binary) VALUES (" + stringify(iterCopyItems->ulId) + ", 0x0FFF," + lpDatabase->EscapeBinary(lpsNewEntryId->__ptr, lpsNewEntryId->__size) + ")";
		er = lpDatabase->DoUpdate(strQuery);
		if(er != erSuccess)
			goto exit;

		er = lpSession->GetNewSourceKey(1, &sSourceKey);
		if(er != erSuccess)
			goto exit;

        // Update source key (changes on move)
		strQuery = "REPLACE INTO indexedproperties(hierarchyid,tag,val_binary) VALUES (" + stringify(iterCopyItems->ulId) + "," + stringify(PROP_ID(PR_SOURCE_KEY)) + "," + lpDatabase->EscapeBinary(sSourceKey, sSourceKey.size()) + ")";
		er = lpDatabase->DoUpdate(strQuery);
		if(er != erSuccess)
			goto exit;

        er = g_lpSessionManager->GetNewSequence(ECSessionManager::SEQ_IMAP, 1, &ullIMAP);
        if(er != erSuccess)
            goto exit;

        // Update IMAP ID (changes on move)
        strQuery = "REPLACE INTO properties(hierarchyid, tag, type, storeid, val_ulong) VALUES(" +
                    stringify(iterCopyItems->ulId) + "," +
                    stringify(PROP_ID(PR_EC_IMAP_ID)) + "," +
                    stringify(PROP_TYPE(PR_EC_IMAP_ID)) + "," +
                    stringify(ulDestStoreId) + "," +
                    stringify(ullIMAP) +
                    ")";

        er = lpDatabase->DoInsert(strQuery);
        if(er != erSuccess)
            goto exit;

		strQuery = "UPDATE hierarchy SET modtime=CURRENT_TIMESTAMP, parent="+stringify(ulDestFolderId)+", flags=flags&"+stringify(~MSGFLAG_DELETED)+" WHERE id="+stringify(iterCopyItems->ulId);

		er = lpDatabase->DoUpdate(strQuery);
		if(er != erSuccess) {
			bPartialCompletion = true;
			er = erSuccess;
			// FIXME: Delete from list: iterCopyItems
			continue;
		}

		// remove PR_DELETED_ON, This is on a softdeleted message
		strQuery = "DELETE FROM properties WHERE hierarchyid="+stringify(iterCopyItems->ulId)+" AND tag="+stringify(PROP_ID(PR_DELETED_ON))+" AND type="+stringify(PROP_TYPE(PR_DELETED_ON)) + " AND storeid=" + stringify(ulDestStoreId);
		er = lpDatabase->DoDelete(strQuery);
		if(er != erSuccess) {
			bPartialCompletion = true;
			er = erSuccess;//ignore error
		}

		// a move is a delete in the originating folder and a new in the destination folder except for softdelete that is a change
		if(iterCopyItems->ulParent != ulDestFolderId){
			AddChange(lpSession, ulSyncId, iterCopyItems->sSourceKey, iterCopyItems->sParentSourceKey, ICS_MESSAGE_HARD_DELETE);
			AddChange(lpSession, ulSyncId, sSourceKey, sDestFolderSourceKey, ICS_MESSAGE_NEW);
		}else if(iterCopyItems->ulFlags & MSGFLAG_DELETED) {
			// Restore a softdeleted message
			AddChange(lpSession, ulSyncId, iterCopyItems->sSourceKey, sDestFolderSourceKey, ICS_MESSAGE_NEW);
		}

		er = lpDatabase->Commit();
		if(er != erSuccess)
			goto exit;

		// Cache update for objects, fnevObjectDeleted is used to allow update object cache
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectDeleted, iterCopyItems->ulId);

		// Remove old sourcekey and entryid and add them
		g_lpSessionManager->GetCacheManager()->RemoveIndexData(iterCopyItems->ulId);

		g_lpSessionManager->GetCacheManager()->SetObjectProp(PROP_ID(PR_SOURCE_KEY), sSourceKey.size(), sSourceKey, iterCopyItems->ulId);

		g_lpSessionManager->GetCacheManager()->SetObjectProp(PROP_ID(PR_ENTRYID), lpsNewEntryId->__size, lpsNewEntryId->__ptr, iterCopyItems->ulId);

		// update destenation folder after PR_ENTRYID update
		if(cCopyItems < EC_TABLE_CHANGE_THRESHOLD) {
			//Update messages
			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_DELETE, 0, iterCopyItems->ulParent, iterCopyItems->ulId, iterCopyItems->ulType);
			//Update destenation folder
			g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_ADD, 0, ulDestFolderId, iterCopyItems->ulId, iterCopyItems->ulType);
		}

		// Update Store object
		g_lpSessionManager->NotificationMoved(iterCopyItems->ulType, iterCopyItems->ulId, ulDestFolderId, iterCopyItems->ulParent, lpsOldEntryId);

		lstParent.push_back(iterCopyItems->ulParent);

	}


	// change the size if it is a soft delete item
	if(bUpdateDeletedSize == true)
		UpdateObjectSize(lpDatabase, ulDestStoreId, ulDestStoreId, MAPI_STORE, false, UPDATE_ADD, ulItemSize);


	lstParent.sort();
	lstParent.unique();

	//Update message folders
	for(iterParent = lstParent.begin(); iterParent != lstParent.end(); iterParent++) {
        if(cCopyItems >= EC_TABLE_CHANGE_THRESHOLD) 
            g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, 0, *iterParent, 0, MAPI_MESSAGE);

		// Get the grandparent
		g_lpSessionManager->GetCacheManager()->GetParent(*iterParent, &ulGrandParent);

        g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, *iterParent);
		g_lpSessionManager->NotificationModified(MAPI_FOLDER, *iterParent);

		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParent, *iterParent, MAPI_FOLDER);

	}

    if(cCopyItems >= EC_TABLE_CHANGE_THRESHOLD) 
        g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, 0, ulDestFolderId, 0, MAPI_MESSAGE);

	//Update destination folder
    g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulDestFolderId);
	g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulDestFolderId);

	// Update the grandfolder of dest. folder
	g_lpSessionManager->GetCacheManager()->GetParent(ulDestFolderId, &ulGrandParent);
	g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParent, ulDestFolderId, MAPI_FOLDER);


	er = g_lpSessionManager->SaveSourceKeyAutoIncrement();
	if(er != erSuccess)
		goto exit;

	if(bPartialCompletion && er == erSuccess)
		er = ZARAFA_W_PARTIAL_COMPLETION;
exit:

	if(lpDatabase && er != hrSuccess && er != ZARAFA_W_PARTIAL_COMPLETION)
		lpDatabase->Rollback();

	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	if(lpsNewEntryId)
		FreeEntryId(lpsNewEntryId, true);

	if(lpsOldEntryId)
		FreeEntryId(lpsOldEntryId, true);

	return er;
}

// Copy one message with his parent data like attachments and recipient
//FIXME: Now notifications before commit, must be changed
ECRESULT CopyObject(ECSession *lpecSession, unsigned int ulObjId, unsigned int ulDestFolderId, bool bIsRoot, bool bDoNotification, bool bDoTableNotification, unsigned int ulSyncId)
{
	ECRESULT		er = erSuccess;
	ECDatabase		*lpDatabase = NULL;
	std::string		strQuery;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow;
	unsigned int	ulNewObjectId = 0;
	std::string		strExclude;
	long long		llStoreSize;
	unsigned int	ulStoreId = 0;
	unsigned int	ulSize;
	unsigned int	ulObjType;
	unsigned int	ulParent = 0;
	GUID			guidStore;
	eQuotaStatus QuotaStatus;
	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey;
	entryId*		lpsNewEntryId = NULL;
	ECAttachmentStorage *lpAttachmentStorage = NULL;
	unsigned int	ulRollbackAttachment = 0;
	unsigned long long ullIMAP = 0;

	lpDatabase = lpecSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulDestFolderId, &ulStoreId, &guidStore);
	if(er != erSuccess)
	    goto exit;

	// Check permission
	if(bIsRoot == true)
	{
		er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityRead);
		if(er != erSuccess)
			goto exit;

		/////////////////////////////////////////////////
		// Quota check
		er = lpecSession->GetSecurity()->GetStoreSize(ulDestFolderId, &llStoreSize);
		if(er != erSuccess)
			goto exit;

		er = lpecSession->GetSecurity()->CheckQuota(ulDestFolderId, llStoreSize, &QuotaStatus);
		if(er != erSuccess)
			goto exit;

		if(QuotaStatus == QUOTA_HARDLIMIT) {
			er = ZARAFA_E_STORE_FULL;
			goto exit;
		}
	}

	// Get the hierarchy messageroot but not the deleted items
	strQuery = "SELECT parent, modtime, createtime, type, flags FROM hierarchy WHERE flags & " + stringify(MSGFLAG_DELETED) + " = 0 AND id="+stringify(ulObjId);

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) < 1) {
		er = ZARAFA_E_NOT_FOUND;// FIXME: right error?
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if( lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[3] == NULL || lpDBRow[4] == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	ulObjType		= atoui(lpDBRow[3]);
	ulParent		= atoui(lpDBRow[0]);

	if (bIsRoot == true && ulObjType != MAPI_MESSAGE) {
		er = ZARAFA_E_INVALID_ENTRYID;
		goto exit;
	}

	//FIXME: Why do we always use the mod and create time of the old object? Create time can always be NOW
	//Create new message (filter flag MSGFLAG_DELETED)
	strQuery = "INSERT INTO hierarchy(parent, modtime, createtime, type, flags, owner) VALUES("+stringify(ulDestFolderId)+", "+((lpDBRow[1] == NULL)?"NULL": ("'"+(string)lpDBRow[1]+"'"))+", "+((lpDBRow[2] == NULL)?"NULL":("'"+(string)lpDBRow[2]+"'"))+", "+(string)lpDBRow[3]+", "+(string)lpDBRow[4]+"&"+stringify(~MSGFLAG_DELETED)+","+stringify(lpecSession->GetSecurity()->GetUserId())+") ";
	er = lpDatabase->DoInsert(strQuery, &ulNewObjectId);
	if(er != erSuccess)
		goto exit;


	// Create message entry
	er = CreateEntryId(guidStore, MAPI_MESSAGE, &lpsNewEntryId);
	if(er != erSuccess)
		goto exit;

	//0x0FFF = PR_ENTRYID
	strQuery = "INSERT INTO indexedproperties (hierarchyid,tag,val_binary) VALUES("+stringify(ulNewObjectId)+", 0x0FFF, "+lpDatabase->EscapeBinary(lpsNewEntryId->__ptr, lpsNewEntryId->__size)+")";
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

    // Add a PR_EC_IMAP_ID
    er = g_lpSessionManager->GetNewSequence(ECSessionManager::SEQ_IMAP, 1, &ullIMAP);
    if(er != erSuccess)
        goto exit;

    strQuery = "INSERT INTO properties(hierarchyid, tag, type, storeid, val_ulong) VALUES(" +
                stringify(ulNewObjectId) + "," +
                stringify(PROP_ID(PR_EC_IMAP_ID)) + "," +
                stringify(PROP_TYPE(PR_EC_IMAP_ID)) + "," +
                stringify(ulStoreId) + "," +
                stringify(ullIMAP) +
                ")";

    er = lpDatabase->DoInsert(strQuery);
    if(er != erSuccess)
        goto exit;

	//Free Results
	if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }
	if(lpsNewEntryId) { FreeEntryId(lpsNewEntryId, true); lpsNewEntryId = NULL; }


	// Get child items of the message like , attachment, recipient...
	strQuery = "SELECT id FROM hierarchy WHERE parent="+stringify(ulObjId);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) > 0) {

		while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
		{
			if(lpDBRow[0] == NULL)
				continue; // FIXME: Skip, give an error/warning ?

			er = CopyObject(lpecSession, atoui(lpDBRow[0]), ulNewObjectId, false, false, false, ulSyncId);
			if(er != erSuccess && er != ZARAFA_E_NOT_FOUND)
				goto exit;
			else
				er = erSuccess;

		}

	}

	// Exclude properties
	// PR_DELETED_ON
	strExclude = " AND NOT (tag="+stringify(PROP_ID(PR_DELETED_ON))+" AND type="+stringify(PROP_TYPE(PR_DELETED_ON))+")";

	//Exclude PR_SOURCE_KEY, PR_CHANGE_KEY, PR_PREDECESSOR_CHANGE_LIST
	strExclude += " AND NOT (tag="+stringify(PROP_ID(PR_SOURCE_KEY))+" AND type="+stringify(PROP_TYPE(PR_SOURCE_KEY))+")";
	strExclude += " AND NOT (tag="+stringify(PROP_ID(PR_CHANGE_KEY))+" AND type="+stringify(PROP_TYPE(PR_CHANGE_KEY))+")";
	strExclude += " AND NOT (tag="+stringify(PROP_ID(PR_PREDECESSOR_CHANGE_LIST))+" AND type="+stringify(PROP_TYPE(PR_PREDECESSOR_CHANGE_LIST))+")";
	strExclude += " AND NOT (tag="+stringify(PROP_ID(PR_EC_IMAP_ID))+" AND type="+stringify(PROP_TYPE(PR_EC_IMAP_ID))+")";

	// Copy properties...
	strQuery = "INSERT INTO properties (hierarchyid, tag, type, storeid, val_ulong, val_string, val_binary,val_double,val_longint,val_hi,val_lo) SELECT "+stringify(ulNewObjectId)+", tag,type," + stringify(ulStoreId) + ",val_ulong,val_string,val_binary,val_double,val_longint,val_hi,val_lo FROM properties WHERE hierarchyid ="+stringify(ulObjId)+strExclude;
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

	// Copy MVproperties...
	strQuery = "INSERT INTO mvproperties (hierarchyid, orderid, tag, type, val_ulong, val_string, val_binary,val_double,val_longint,val_hi,val_lo) SELECT "+stringify(ulNewObjectId)+", orderid, tag,type,val_ulong,val_string,val_binary,val_double,val_longint,val_hi,val_lo FROM mvproperties WHERE hierarchyid ="+stringify(ulObjId);
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

	// Copy large objects... if present
	er = lpAttachmentStorage->CopyAttachment(ulObjId, ulNewObjectId);
	if(er != erSuccess && er != ZARAFA_E_NOT_FOUND)
		goto exit;
	er = erSuccess;
	ulRollbackAttachment = ulNewObjectId;

	if(bIsRoot == true)
	{
		// Create indexedproperties, Add new PR_SOURCE_KEY
		lpecSession->GetNewSourceKey(1, &sSourceKey);

		strQuery = "INSERT INTO indexedproperties(hierarchyid,tag,val_binary) VALUES(" + stringify(ulNewObjectId) + "," + stringify(PROP_ID(PR_SOURCE_KEY)) + "," + lpDatabase->EscapeBinary(sSourceKey, sSourceKey.size()) + ")";
		er = lpDatabase->DoInsert(strQuery);
		if(er != erSuccess)
			goto exit;

		g_lpSessionManager->GetCacheManager()->SetObjectProp(PROP_ID(PR_SOURCE_KEY), sSourceKey.size(), sSourceKey, ulNewObjectId);

		// Update Size
		if(GetObjectSize(lpDatabase, ulStoreId, ulNewObjectId, &ulSize) == erSuccess)
		{
			if(lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulNewObjectId, &ulStoreId, NULL) == erSuccess)
				UpdateObjectSize(lpDatabase, ulStoreId, ulStoreId, MAPI_STORE, false, UPDATE_ADD, ulSize);
		}
	}

	g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulDestFolderId);

	if(bDoNotification){
		// Update destenation folder
    	if(bDoTableNotification)
        	g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_ADD, 0, ulDestFolderId, ulNewObjectId, MAPI_MESSAGE);

		g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulDestFolderId);

		// Notify object is copied
		g_lpSessionManager->NotificationCopied(MAPI_MESSAGE, ulNewObjectId, ulDestFolderId, ulObjId, ulParent);
	}

	// Update ICS system
	if(bIsRoot == true){
		GetSourceKey(ulDestFolderId, &sParentSourceKey);
		AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_MESSAGE_NEW);
	}

	er = g_lpSessionManager->SaveSourceKeyAutoIncrement();
	if(er != erSuccess)
		goto exit;

exit:
	if (er != erSuccess && ulRollbackAttachment != 0 && lpAttachmentStorage->IsFileStorage())
		lpAttachmentStorage->DeleteAttachment(ulRollbackAttachment);

	//Free Results
	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	if (lpAttachmentStorage)
		delete lpAttachmentStorage;

	if(lpsNewEntryId)
		FreeEntryId(lpsNewEntryId, true);

	return er;
}

/**
 * Copy folder and his childs
 *
 * @note please check the object type before you call this function, the type should be MAPI_FOLDER
 */
ECRESULT CopyFolderObjects(struct soap *soap, ECSession *lpecSession, unsigned int ulFolderFrom, unsigned int ulDestFolderId, char *lpszNewFolderName, bool bCopySubFolder, unsigned int ulSyncId)
{
	ECRESULT		er = erSuccess;
	ECDatabase		*lpDatabase = NULL;
	std::string		strQuery, strSubQuery, strExclude;
	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow;

	unsigned int	ulNewDestFolderId = 0;
	unsigned int	ulItems = 0;
	unsigned int	ulGrandParent = 0;

	unsigned int	ulDestStoreId = 0;
	unsigned int	ulSourceStoreId = 0;

	bool			bPartialCompletion = false;
	long long		llStoreSize = 0;
	eQuotaStatus QuotaStatus;

	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey;
	ECAttachmentStorage *lpAttachmentStorage = NULL;
	unsigned int	ulRollbackAttachment = 0;


	if(lpszNewFolderName == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	lpDatabase = lpecSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	er = CreateAttachmentStorage(lpDatabase, &lpAttachmentStorage);
	if(er != erSuccess)
	    goto exit;

	er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulDestFolderId, &ulDestStoreId, NULL);
	if(er != erSuccess)
	    goto exit;

	er = lpecSession->GetSessionManager()->GetCacheManager()->GetStore(ulFolderFrom, &ulSourceStoreId, NULL);
	if(er != erSuccess)
	    goto exit;

	er = lpDatabase->Begin();
	if(er != hrSuccess)
		goto exit;

	/////////////////////////////////////////////////
	// Quota check
	er = lpecSession->GetSecurity()->GetStoreSize(ulDestFolderId, &llStoreSize);
	if(er != erSuccess)
		goto exit;

	er = lpecSession->GetSecurity()->CheckQuota(ulDestFolderId, llStoreSize, &QuotaStatus);
	if(er != erSuccess)
		goto exit;

	if(QuotaStatus == QUOTA_HARDLIMIT) {
		er = ZARAFA_E_STORE_FULL;
		goto exit;
	}

	// Create folder (with a sourcekey)
	er = CreateFolder(lpecSession, lpDatabase, ulDestFolderId, NULL, FOLDER_GENERIC, lpszNewFolderName, NULL, false, true, ulSyncId, NULL, &ulNewDestFolderId, NULL);
	if(er != erSuccess)
		goto exit;

	strExclude = " AND NOT (tag="+stringify(PROP_ID(PR_DELETED_ON))+" AND type="+stringify(PROP_TYPE(PR_DELETED_ON))+")";
	strExclude += " AND NOT (tag="+stringify(PROP_ID(PR_DISPLAY_NAME))+" AND type="+stringify(PROP_TYPE(PR_DISPLAY_NAME))+")";

	//Exclude PR_SOURCE_KEY, PR_CHANGE_KEY, PR_PREDECESSOR_CHANGE_LIST
	strExclude += " AND NOT (tag="+stringify(PROP_ID(PR_SOURCE_KEY))+" AND type="+stringify(PROP_TYPE(PR_SOURCE_KEY))+")";
	strExclude += " AND NOT (tag="+stringify(PROP_ID(PR_CHANGE_KEY))+" AND type="+stringify(PROP_TYPE(PR_CHANGE_KEY))+")";
	strExclude += " AND NOT (tag="+stringify(PROP_ID(PR_PREDECESSOR_CHANGE_LIST))+" AND type="+stringify(PROP_TYPE(PR_PREDECESSOR_CHANGE_LIST))+")";

	// Copy properties...
	strQuery = "INSERT INTO properties (hierarchyid, tag, type, storeid, val_ulong, val_string, val_binary,val_double,val_longint,val_hi,val_lo) SELECT "+stringify(ulNewDestFolderId)+", tag,type,"+stringify(ulDestStoreId)+",val_ulong,val_string,val_binary,val_double,val_longint,val_hi,val_lo FROM properties WHERE hierarchyid ="+stringify(ulFolderFrom)+strExclude;
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

	// Copy MVproperties...
	strQuery = "INSERT INTO mvproperties (hierarchyid, orderid, tag, type, val_ulong, val_string, val_binary,val_double,val_longint,val_hi,val_lo) SELECT "+stringify(ulNewDestFolderId)+", orderid, tag,type,val_ulong,val_string,val_binary,val_double,val_longint,val_hi,val_lo FROM mvproperties WHERE hierarchyid ="+stringify(ulFolderFrom);
	er = lpDatabase->DoInsert(strQuery);
	if(er != erSuccess)
		goto exit;

	// Copy large objects... if present .. probably not, on a folder
	er = lpAttachmentStorage->CopyAttachment(ulFolderFrom, ulNewDestFolderId);
	if(er != erSuccess && er != ZARAFA_E_NOT_FOUND)
		goto exit;
	er = erSuccess;
	ulRollbackAttachment = ulNewDestFolderId;

	// update ICS sytem with a change
	GetSourceKey(ulDestFolderId, &sParentSourceKey);
	GetSourceKey(ulNewDestFolderId, &sSourceKey);

	AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_FOLDER_CHANGE);


	//Select all Messages of the home folder
	// Skip deleted and associated items
	strQuery = "SELECT id FROM hierarchy WHERE parent="+stringify(ulFolderFrom)+ " AND type="+stringify(MAPI_MESSAGE)+" AND flags & " + stringify(MSGFLAG_DELETED|MSGFLAG_ASSOCIATED) + " = 0";
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	ulItems = lpDatabase->GetNumRows(lpDBResult);

	// Walk through the messages list
	while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
	{
		if(lpDBRow[0] == NULL)
			continue; //FIXME: error show ???

		er = CopyObject(lpecSession, atoui(lpDBRow[0]), ulNewDestFolderId, true, false, false, ulSyncId);
		// FIXME: handle ZARAFA_E_STORE_FULL
		if(er == ZARAFA_E_NOT_FOUND) {
			bPartialCompletion = true;
		} else if(er != erSuccess)
			goto exit;
	}

	if(lpDBResult) {
		lpDatabase->FreeResult(lpDBResult);
		lpDBResult = NULL;
	}

	er = lpDatabase->Commit();
	if (er != erSuccess)
		goto exit;

	// Notifications
	if(ulItems > 0)
	{
		//Update destenation folder
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, 0, ulNewDestFolderId, 0, MAPI_MESSAGE);

		// Update the grandfolder of dest. folder
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulDestFolderId, ulNewDestFolderId, MAPI_FOLDER);

		//Update destination folder
        g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulNewDestFolderId);
		g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulNewDestFolderId);

	}

	g_lpSessionManager->GetCacheManager()->GetParent(ulFolderFrom ,&ulGrandParent);
	g_lpSessionManager->NotificationCopied(MAPI_FOLDER, ulNewDestFolderId, ulDestFolderId, ulFolderFrom, ulGrandParent);

	if(bCopySubFolder) {
		//Select all folders of the home folder
		// Skip deleted folders
		strQuery = "SELECT hierarchy.id, properties.val_string FROM hierarchy JOIN properties ON hierarchy.id = properties.hierarchyid WHERE hierarchy.parent=" + stringify(ulFolderFrom) +" AND hierarchy.type="+stringify(MAPI_FOLDER)+" AND (flags & " + stringify(MSGFLAG_DELETED) + ") = 0 AND properties.tag=" + stringify(ZARAFA_TAG_DISPLAY_NAME) + " AND properties.storeid=" + stringify(ulSourceStoreId);
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		if(lpDatabase->GetNumRows(lpDBResult) > 0) {

			// Walk through the folder list
			while( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
			{
				if(lpDBRow[0] == NULL || lpDBRow[1] == NULL)
					continue; // ignore

				// Create SubFolder with messages. This object type checking is done in the where of the query
				er = CopyFolderObjects(soap, lpecSession, atoui(lpDBRow[0]), ulNewDestFolderId, lpDBRow[1], true, ulSyncId);
				if(er == ZARAFA_W_PARTIAL_COMPLETION)
					bPartialCompletion = true;
				else if(er != erSuccess)
					goto exit;
			}
		}
	}

	if(bPartialCompletion && er == erSuccess)
		er = ZARAFA_W_PARTIAL_COMPLETION;

exit:
    if(lpDatabase && er != erSuccess && er != ZARAFA_W_PARTIAL_COMPLETION) {
        lpDatabase->Rollback();
		if (ulRollbackAttachment != 0 && lpAttachmentStorage->IsFileStorage())
			lpAttachmentStorage->DeleteAttachment(ulRollbackAttachment);
	}

	if (lpAttachmentStorage)
		delete lpAttachmentStorage;

	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;

}

/**
 * Copy one or more messages to a destenation
 */
SOAP_ENTRY_START5(copyObjects, *result, struct entryList *aMessages, entryId sDestFolderId, unsigned int ulFlags, unsigned int ulSyncId, unsigned int *result)
{
	bool			bPartialCompletion = false;
	unsigned int	ulGrandParent=0;
	unsigned int	ulDestFolderId = 0;
	ECListInt			lObjectIds;
	ECListIntIterator	iObjectId;
	size_t				cObjectItems = 0;

	USE_DATABASE();

	if(aMessages == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sDestFolderId, &ulDestFolderId);
	if(er != erSuccess)
	    goto exit;

	// Check permission, Destination folder
	er = lpecSession->GetSecurity()->CheckPermission(ulDestFolderId, ecSecurityCreate);
	if(er != erSuccess)
		goto exit;

	if(g_lpSessionManager->GetCacheManager()->GetEntryListToObjectList(aMessages, &lObjectIds) != erSuccess)
		bPartialCompletion = true;

	// @note The object type checking wille be done in MoveObjects or CopyObject

	//check copy or a move
	if(ulFlags & FOLDER_MOVE ) { // A move
		er = MoveObjects(lpecSession, lpDatabase, &lObjectIds, ulDestFolderId, ulSyncId);
		if(er != erSuccess)
			goto exit;

	}else { // A copy

		cObjectItems = lObjectIds.size();
		for(iObjectId = lObjectIds.begin(); iObjectId != lObjectIds.end(); iObjectId++)
		{
			er = CopyObject(lpecSession, *iObjectId, ulDestFolderId, true, true, cObjectItems < EC_TABLE_CHANGE_THRESHOLD, ulSyncId);
			if(er != erSuccess) {
				bPartialCompletion = true;
				er = erSuccess;
			}
		}

		if(cObjectItems >= EC_TABLE_CHANGE_THRESHOLD)
		    g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, 0, ulDestFolderId, 0, MAPI_MESSAGE);

		// Update the grandfolder of dest. folder
		g_lpSessionManager->GetCacheManager()->GetParent(ulDestFolderId, &ulGrandParent);
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParent, ulDestFolderId, MAPI_FOLDER);

		//Update destination folder
        g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulDestFolderId);
		g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulDestFolderId);

	}

	if(bPartialCompletion && er == erSuccess)
		er = ZARAFA_W_PARTIAL_COMPLETION;
exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START6(copyFolder, *result, entryId sEntryId, entryId sDestFolderId, char *lpszNewFolderName, unsigned int ulFlags, unsigned int ulSyncId, unsigned int *result)
{
	unsigned int	ulAffRows = 0;
	unsigned int	ulOldParent = 0;
	unsigned int	ulGrandParent = 0;
	unsigned int	ulParentCycle = 0;
	unsigned int	ulDestStoreId = 0;
	unsigned int	ulSourceStoreId = 0;
	unsigned int	ulObjFlags = 0;
	unsigned int	ulOldGrandParent = 0;
	unsigned int	ulFolderId = 0;
	unsigned int	ulDestFolderId = 0;
	unsigned int	ulSourceType = 0;
	unsigned int	ulDestType = 0;
	long long		llFolderSize = 0;

	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey; // Old parent
	SOURCEKEY		sDestSourceKey; // New parent

	std::string 	strSubQuery;
	USE_DATABASE();

	// NOTE: lpszNewFolderName can be NULL

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulFolderId);
	if(er != erSuccess)
	    goto exit;

	// Get source store
    er = g_lpSessionManager->GetCacheManager()->GetStore(ulFolderId, &ulSourceStoreId, NULL);
    if(er != erSuccess)
        goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sDestFolderId, &ulDestFolderId);
	if(er != erSuccess)
	    goto exit;

	// Get dest store
    er = g_lpSessionManager->GetCacheManager()->GetStore(ulDestFolderId, &ulDestStoreId, NULL);
    if(er != erSuccess)
        goto exit;

	if(ulDestStoreId != ulSourceStoreId) {
		ASSERT(FALSE);
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulDestFolderId, ecSecurityCreateFolder);
	if(er != erSuccess)
		goto exit;

	if(ulFlags & FOLDER_MOVE ) // is the folder editable?
		er = lpecSession->GetSecurity()->CheckPermission(ulFolderId, ecSecurityFolderAccess);
	else // is the folder readable
		er = lpecSession->GetSecurity()->CheckPermission(ulFolderId, ecSecurityRead);
	if(er != erSuccess)
		goto exit;

	// Check MAPI_E_FOLDER_CYCLE
	if(ulFolderId == ulDestFolderId) {
		er = ZARAFA_E_FOLDER_CYCLE;
		goto exit;
	}

	// Get the parent id, for notification and copy
	er = g_lpSessionManager->GetCacheManager()->GetObject(ulFolderId, &ulOldParent, NULL, &ulObjFlags, &ulSourceType);
	if(er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObject(ulDestFolderId, NULL, NULL, NULL, &ulDestType);
	if(er != erSuccess)
		goto exit;

	if (ulSourceType != MAPI_FOLDER || ulDestType != MAPI_FOLDER) {
		er = ZARAFA_E_INVALID_ENTRYID;
		goto exit;
	}

	// Check folder and dest folder are the same
	if(!(ulObjFlags & MSGFLAG_DELETED) && (ulFlags & FOLDER_MOVE) && (ulDestFolderId == ulOldParent))
		goto exit; // Do nothing... folder already on the right place

	ulParentCycle = ulDestFolderId;
	while(g_lpSessionManager->GetCacheManager()->GetParent(ulParentCycle, &ulParentCycle) == erSuccess)
	{
		if(ulFolderId == ulParentCycle)
		{
			er = ZARAFA_E_FOLDER_CYCLE;
			goto exit;
		}
	}

	// Check whether the requested name already exists
	strQuery = "SELECT hierarchy.id FROM hierarchy JOIN properties ON hierarchy.id = properties.hierarchyid WHERE parent=" + stringify(ulDestFolderId) + " AND (hierarchy.flags & " + stringify(MSGFLAG_DELETED) + ") = 0 AND hierarchy.type="+stringify(MAPI_FOLDER)+" AND properties.tag=" + stringify(ZARAFA_TAG_DISPLAY_NAME)+" AND properties.storeid=" + stringify(ulDestStoreId);
	if(lpszNewFolderName) {
		strQuery+= " AND properties.val_string = '" + lpDatabase->Escape(lpszNewFolderName) + "'";
	} else {
		strSubQuery = "SELECT properties.val_string FROM hierarchy JOIN properties ON hierarchy.id = properties.hierarchyid WHERE hierarchy.id=" + stringify(ulFolderId) + " AND properties.tag=" + stringify(ZARAFA_TAG_DISPLAY_NAME) + " AND properties.storeid=" + stringify(ulSourceStoreId) + " AND properties.type=" + stringify(PT_STRING8);
		strQuery+= " AND properties.val_string = ("+strSubQuery+")";
	}

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) > 0 && !ulSyncId) {
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	FREE_DBRESULT();

	if(lpszNewFolderName == NULL)
	{
		strQuery = "SELECT properties.val_string FROM hierarchy JOIN properties ON hierarchy.id = properties.hierarchyid WHERE hierarchy.id=" + stringify(ulFolderId) + " AND properties.tag=" + stringify(ZARAFA_TAG_DISPLAY_NAME) + " AND properties.type=" + stringify(PT_STRING8) + " AND properties.storeid=" + stringify(ulDestStoreId);
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		if( lpDBRow == NULL || lpDBRow[0] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		lpszNewFolderName = s_alloc<char>(soap, strlen(lpDBRow[0])+1);
		memcpy(lpszNewFolderName, lpDBRow[0], strlen(lpDBRow[0])+1);

		FREE_DBRESULT();
	}

	//check copy or a move
	if(ulFlags & FOLDER_MOVE ) {

        if(ulObjFlags & MSGFLAG_DELETED) {
            // The folder we're moving used to be deleted. This effictively makes this call an un-delete. We need to get the folder size
            // for quota management
            er = GetFolderSize(lpDatabase, ulSourceStoreId, ulFolderId, &llFolderSize);
            if(er != erSuccess)
                goto exit;
        }

		// Get grandParent of the old folder
		g_lpSessionManager->GetCacheManager()->GetParent(ulOldParent, &ulOldGrandParent);

		er = lpDatabase->Begin();
		if(er != hrSuccess)
			goto exit;

		// Move the folder to the dest. folder
		strQuery = "UPDATE hierarchy SET modtime=CURRENT_TIMESTAMP, parent="+stringify(ulDestFolderId)+", flags=flags&"+stringify(~MSGFLAG_DELETED)+" WHERE id="+stringify(ulFolderId);
		if(lpDatabase->DoUpdate(strQuery, &ulAffRows) != erSuccess) {
		    lpDatabase->Rollback();
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		if(ulAffRows != 1) {
		    lpDatabase->Rollback();
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		// Update the folder to the destination folder
		//Info: Always an update, It's not faster first check and than update/or not
		strQuery = "UPDATE properties SET val_string = '" + lpDatabase->Escape(lpszNewFolderName) + "' WHERE tag=" + stringify(ZARAFA_TAG_DISPLAY_NAME) + " AND hierarchyid="+stringify(ulFolderId) + " AND type=" + stringify(PT_STRING8) + " AND storeid=" + stringify(ulDestStoreId);
		if(lpDatabase->DoUpdate(strQuery, &ulAffRows) != erSuccess) {
		    lpDatabase->Rollback();
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		// remove PR_DELETED_ON, as the folder is a softdelete folder
		strQuery = "DELETE FROM properties WHERE hierarchyid="+stringify(ulFolderId)+" AND tag="+stringify(PROP_ID(PR_DELETED_ON))+" AND type="+stringify(PROP_TYPE(PR_DELETED_ON)) + " AND storeid=" + stringify(ulDestStoreId);
		er = lpDatabase->DoDelete(strQuery);
		if(er != erSuccess) {
		    lpDatabase->Rollback();
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		// Update the store size if we did an undelete. Note ulSourceStoreId == ulDestStoreId.
		if(llFolderSize > 0)
    		UpdateObjectSize(lpDatabase, ulSourceStoreId, ulSourceStoreId, MAPI_STORE, false, UPDATE_ADD, llFolderSize);

		// ICS
		GetSourceKey(ulFolderId, &sSourceKey);
		GetSourceKey(ulDestFolderId, &sDestSourceKey);
		GetSourceKey(ulOldParent, &sParentSourceKey);
			
		AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_FOLDER_CHANGE, &sDestSourceKey);

		er = lpDatabase->Commit();
		if(er != erSuccess) {
			lpDatabase->Rollback();
			goto exit;
		}

		// Cache update for objects
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectMoved, ulFolderId);
		// Notify that the folder has moved
		g_lpSessionManager->NotificationMoved(MAPI_FOLDER, ulFolderId, ulDestFolderId, ulOldParent);

		// Update the old folder
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulOldParent);
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_DELETE, 0, ulOldParent, ulFolderId, MAPI_FOLDER);
		g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulOldParent);

		// Update the old folder's parent
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulOldGrandParent, ulOldParent, MAPI_FOLDER);

		// Update the destination folder
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulDestFolderId);
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_ADD, 0, ulDestFolderId, ulFolderId, MAPI_FOLDER);
		g_lpSessionManager->NotificationModified(MAPI_FOLDER, ulDestFolderId);

		// Update the destination's parent
		g_lpSessionManager->GetCacheManager()->GetParent(ulDestFolderId, &ulGrandParent);
		g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParent, ulDestFolderId, MAPI_FOLDER);

	}else {// a copy

		er = CopyFolderObjects(soap, lpecSession, ulFolderId, ulDestFolderId, lpszNewFolderName, !!(ulFlags&COPY_SUBFOLDERS), ulSyncId);
		if(er != erSuccess)
			goto exit;

	}

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(notify, *result, struct notification sNotification, unsigned int *result)
{
	unsigned int ulKey = 0;
	USE_DATABASE();

	// You are only allowed to send newmail notifications at the moment. This could currently
	// only be misused to send other users new mail popup notification for e-mails that aren't
	// new at all ...
	if (sNotification.ulEventType != fnevNewMail) {
	    er = ZARAFA_E_NO_ACCESS;
	    goto exit;
    }
    
    if (sNotification.newmail == NULL || sNotification.newmail->pParentId == NULL || sNotification.newmail->pEntryId == NULL) {
        er = ZARAFA_E_INVALID_PARAMETER;
        goto exit;
    }

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(sNotification.newmail->pParentId, &ulKey);
	if(er != erSuccess)
	    goto exit;

	sNotification.ulConnection = ulKey;

	er = g_lpSessionManager->AddNotification(&sNotification, ulKey);

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(getReceiveFolderTable, lpsReceiveFolderTable->er, entryId sStoreId, struct receiveFolderTableResponse *lpsReceiveFolderTable)
{
	size_t			size = 0;
	int				ulRows = 0;
	int				i;
	unsigned int	ulStoreid = 0;
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sStoreId, &ulStoreid);
	if(er != erSuccess)
	    goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->CheckPermission(ulStoreid, ecSecurityRead);
	if(er != erSuccess)
		goto exit;

	strQuery = "SELECT objid, messageclass FROM receivefolder WHERE storeid="+stringify(ulStoreid);

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;


	ulRows = lpDatabase->GetNumRows(lpDBResult);

	lpsReceiveFolderTable->sFolderArray.__ptr = s_alloc<receiveFolder>(soap, ulRows);
	lpsReceiveFolderTable->sFolderArray.__size = 0;

	i = 0;
	while((lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
	{

		if(lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL){
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(atoui(lpDBRow[0]), soap, &lpsReceiveFolderTable->sFolderArray.__ptr[i].sEntryId);
		if(er != erSuccess){
			er = erSuccess;
			continue;
		}

		size = strlen(lpDBRow[1])+1;
		lpsReceiveFolderTable->sFolderArray.__ptr[i].lpszAExplicitClass = s_alloc<char>(soap, size);
		memcpy(lpsReceiveFolderTable->sFolderArray.__ptr[i].lpszAExplicitClass, lpDBRow[1], size);

		i++;
	}

	lpsReceiveFolderTable->sFolderArray.__size = i;

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(deleteUser, *result, unsigned int ulUserId, entryId sUserId, unsigned int *result)
{
	er = GetLocalId(NULL, sUserId, ulUserId, &ulUserId, NULL, NULL);
	if (er != erSuccess)
		goto exit;

	// Check permission
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserId);
	if(er != erSuccess)
		goto exit;

    er = lpecSession->GetUserManagement()->DeleteObjectAndSync(ulUserId);
    if(er != erSuccess)
        goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START4(hookStore, *result, entryId sUserId, struct xsd__base64Binary sStoreGuid, unsigned int ulSyncId, unsigned int *result)
{
	unsigned int ulUserId = 0;
	userobject_type_t ulType;
	unsigned int ulMapiType = 0;
	unsigned int ulAffected = 0;
	objectdetails_t sUserDetails;

	USE_DATABASE();

	er = ABEntryIDToID(&sUserId, &ulUserId, NULL, &ulMapiType);
	if(er != erSuccess)
		goto exit;

	er = MAPITypeToType(ulMapiType, &ulType);
	if(er != erSuccess)
		goto exit;

	if (ulUserId == 0 || ulUserId == ZARAFA_UID_SYSTEM || (ulType != USEROBJECT_TYPE_USER && ulType != USEROBJECT_TYPE_NONACTIVE)) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// get user details, see if this is the correct server
	er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulUserId, &sUserDetails);
	if (er != erSuccess)
		goto exit;

	if (g_lpSessionManager->IsDistributedSupported()) {
		if (sUserDetails.GetPropString(OB_PROP_S_SERVERNAME).compare(lpecSession->GetSessionManager()->GetConfig()->GetSetting("server_name")) != 0) {
			// TODO: redirect?
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
	}

	// check if store currently is owned
	strQuery = "SELECT users.id, stores.id, stores.user_id, stores.hierarchy_id FROM stores LEFT JOIN users ON stores.user_id = users.id WHERE guid = ";
	strQuery += lpDatabase->EscapeBinary(sStoreGuid.__ptr, sStoreGuid.__size);
	
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if(lpDBRow == NULL || lpDBLen == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	if (lpDBRow[0]) {
		// this store already belongs to a user
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Hooking store %s to user %d", lpDBRow[1], ulUserId);

	// lpDBRow[2] is the old user id, which is now orphaned. We'll use this id to make the other store orphaned, so we "trade" user id's.

	// update user with new store id
	er = lpDatabase->Begin();
	if (er != erSuccess)
		goto exit;

	// remove previous user of store
	strQuery = "UPDATE stores SET user_id = " + string(lpDBRow[2]) + " WHERE user_id = " + stringify(ulUserId);

	er = lpDatabase->DoUpdate(strQuery, &ulAffected);
	if (er != erSuccess)
		goto exit;

	// ulAffected == 0: The user was already orphaned 
	// ulAffected == 1: correctly disowned previous owner of store
	if (ulAffected > 1) {
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	// set new store
	strQuery = "UPDATE stores SET user_id = " + stringify(ulUserId) + " WHERE guid = ";
	strQuery += lpDatabase->EscapeBinary(sStoreGuid.__ptr, sStoreGuid.__size);

	er = lpDatabase->DoUpdate(strQuery, &ulAffected);
	if (er != erSuccess)
		goto exit;

	// we can't have one store being owned by multiple users
	if (ulAffected != 1) {
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	// update owner of store
	strQuery = "UPDATE hierarchy SET owner = " + stringify(ulUserId) + " WHERE id = " + string(lpDBRow[3]);

	er = lpDatabase->DoUpdate(strQuery, &ulAffected);
	if (er != erSuccess)
		goto exit;

	// one store has only one entry point in the hierarchy
	// (may be zero, when the user returns to it's original store, so the owner field stays the same)
	if (ulAffected > 1) {
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	er = lpDatabase->Commit();
	if (er != erSuccess)
		goto exit;

	// remove store cache item
	g_lpSessionManager->GetCacheManager()->Update(fnevObjectMoved, atoi(lpDBRow[3]));

exit:
	if (er != erSuccess)
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Hook of store failed: 0x%x", er);
		
	ROLLBACK_ON_ERROR();
	FREE_DBRESULT();
}
SOAP_ENTRY_END()

// Used to move the store to the public, but this isn't possible due to multi-server setups.
// And is was super slow because of the storeid column in the properties table.
SOAP_ENTRY_START3(deleteStore, *result, unsigned int ulStoreId, unsigned int ulSyncId, unsigned int *result)
{
	er = ZARAFA_E_NOT_IMPLEMENTED;
}
SOAP_ENTRY_END()

// softdelete the store from the database, so this function returns quickly
SOAP_ENTRY_START3(removeStore, *result, struct xsd__base64Binary sStoreGuid, unsigned int ulSyncId, unsigned int *result)
{
	unsigned int	ulCompanyId = 0;
	unsigned int	ulStoreId = 0;
	unsigned int	ulStoreHierarchyId = 0;
	ECListInt		lObjectIds;
	objectdetails_t sObjectDetails;
	std::string		strUsername;

	USE_DATABASE();

	// find store id and company of guid
	strQuery = "SELECT users.id, stores.id, stores.hierarchy_id, stores.company, stores.user_name FROM stores LEFT JOIN users ON stores.user_id = users.id WHERE stores.guid = ";
	strQuery += lpDatabase->EscapeBinary(sStoreGuid.__ptr, sStoreGuid.__size);

	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if(lpDBRow == NULL || lpDBLen == NULL) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	// if users.id != NULL, user still present .. log a warning admin is doing something that might not have been the action it wanted to do.
	if (lpDBRow[0] != NULL) {
		// trying to remove store from existing user
		strUsername = lpDBRow[4];
		if (lpecSession->GetUserManagement()->GetObjectDetailsAndSync(atoi(lpDBRow[0]), &sObjectDetails) == erSuccess)
			strUsername = sObjectDetails.GetPropString(OB_PROP_S_LOGIN); // fullname?

		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Unable to remove store: store is in use by user %s", strUsername.c_str());
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	// these are all 'not null' columns
	ulStoreId = atoi(lpDBRow[1]);
	ulStoreHierarchyId = atoi(lpDBRow[2]);
	ulCompanyId = atoi(lpDBRow[3]);

	// Must be administrator over the company to be able to remove the store
	er = lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyId);
	if (er != erSuccess)
		goto exit;

	g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Started to remove store (%d) with storename %s", ulStoreId, lpDBRow[4]);

	er = lpDatabase->Begin();
	if(er != hrSuccess)
		goto exit;

	// Soft delete store with parent etc..
	lObjectIds.push_back(ulStoreHierarchyId);

	er = DeleteObjects(lpecSession, lpDatabase, &lObjectIds, EC_DELETE_CONTAINER | EC_DELETE_FOLDERS | EC_DELETE_MESSAGES | EC_DELETE_RECIPIENTS | EC_DELETE_ATTACHMENTS | EC_DELETE_STORES, ulSyncId, true, false);
	if(er != erSuccess)
		goto exit;

	// Remove the store entry
	strQuery = "DELETE FROM stores WHERE id="+stringify(ulStoreId);
	er = lpDatabase->DoDelete(strQuery);
	if(er != erSuccess)
		goto exit;

	// Remove receivefolder entries
	strQuery = "DELETE FROM receivefolder WHERE storeid="+stringify(ulStoreHierarchyId);
	er = lpDatabase->DoDelete(strQuery);
	if(er != erSuccess)
		goto exit;

	// Remove the acls
	// this aint enough .. i think ?
	strQuery = "DELETE FROM acl WHERE hierarchy_id="+stringify(ulStoreHierarchyId);
	er = lpDatabase->DoDelete(strQuery);
	if(er != erSuccess)
		goto exit;
	// TODO: acl cache!

	er = lpDatabase->Commit();
	if(er != erSuccess)
		goto exit;

	g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Finished remove store (%d)", ulStoreId);

exit:
	FREE_DBRESULT();

	if(er == ZARAFA_E_NO_ACCESS)
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Failed to remove store access denied");
	else if(er != erSuccess) {

		if(lpDatabase)
			lpDatabase->Rollback();

		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_FATAL, "Failed to remove store (%d), errorcode=0x%08X", ulStoreId, er);
	}
}
SOAP_ENTRY_END()

void* SoftDeleteRemover(void* lpTmpMain)
{
	ECRESULT		er = erSuccess;
	ECRESULT*		lper = NULL;
	char*			lpszSetting = NULL;
	unsigned int	ulDeleteTime = 0;
	unsigned int	ulFolders = 0;
	unsigned int	ulMessages = 0;
	ECSESSIONID		sessionID = 0;
	ECSession		*lpecSession = NULL;

	lpszSetting = g_lpSessionManager->GetConfig()->GetSetting("softdelete_lifetime");

	if(lpszSetting)
		ulDeleteTime = atoi(lpszSetting) * 24 * 60 * 60;

	if(ulDeleteTime == 0)
		goto exit;

	er = g_lpSessionManager->CreateSessionInternal(&sessionID, &lpecSession);
	if(er != erSuccess)
		goto exit;

	// Lock the session
	lpecSession->Lock();

	g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Start scheduled softdelete clean up");

	er = PurgeSoftDelete(lpecSession, ulDeleteTime, &ulMessages, &ulFolders);

exit:
	if(ulDeleteTime > 0 && er == erSuccess)
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Softdelete done: removed folders: %d, removed messages: %d", ulFolders, ulMessages);
	else if(ulDeleteTime > 0)
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Softdelete failed: removed folders: %d, removed messages: %d", ulFolders, ulMessages);

	if(lpecSession)
		lpecSession->Unlock(); 

	if(sessionID > 0)
		g_lpSessionManager->RemoveSession(sessionID);

	// Exit with the error result
	lper = new ECRESULT;
	*lper = er;

	// Do not pthread_exit() because linuxthreads is broken and will not free any objects
	// pthread_exit((void*)lper);

	return (void *)lper;
}


SOAP_ENTRY_START3(checkExistObject, *result, entryId sEntryId, unsigned int ulFlags, unsigned int *result)
{
	unsigned int	ulObjId = 0;
	unsigned int	ulObjType = 0;
	unsigned int	ulDBFlags = 0;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulObjId);
	if(er != erSuccess)
	    goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObject(ulObjId, NULL, NULL, &ulDBFlags, &ulObjType);
	if(er != erSuccess)
	    goto exit;

	if(ulFlags & SHOW_SOFT_DELETES) {
		if(!(ulDBFlags & MSGFLAG_DELETED)) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
	} else {
		if(ulDBFlags & MSGFLAG_DELETED) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
	}

	// Deny open to non-store-owners of search folders
	if(ulObjType == MAPI_FOLDER) {
	    if(lpecSession->GetSecurity()->IsObjectAdmin(ulObjId) != erSuccess && lpecSession->GetSecurity()->IsStoreOwner(ulObjId) != hrSuccess) {
	    	if(ulDBFlags & FOLDER_SEARCH) {
	    		er = ZARAFA_E_NOT_FOUND;
	    		goto exit;
			}
        }
    }
exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(readABProps, readPropsResponse->er, entryId sEntryId, struct readPropsResponse *readPropsResponse)
{
	// FIXME: when props are PT_ERROR, they shouldn't be send to the client
	// now we have properties in the client which are MAPI_E_NOT_ENOUGH_MEMORY,
	// while they shouldn't be present (or atleast MAPI_E_NOT_FOUND)

	unsigned int	sProps[] = {
		/* Don't touch the order of the first 7 elements!!! */
		PR_ENTRYID, PR_CONTAINER_FLAGS, PR_DEPTH, PR_EMS_AB_CONTAINERID, PR_DISPLAY_NAME, PR_EMS_AB_IS_MASTER, PR_EMS_AB_PARENT_ENTRYID,
		PR_EMAIL_ADDRESS, PR_OBJECT_TYPE, PR_DISPLAY_TYPE, PR_SEARCH_KEY, PR_PARENT_ENTRYID, PR_ADDRTYPE, PR_RECORD_KEY, PR_ACCOUNT,
		PR_SMTP_ADDRESS, PR_TRANSMITABLE_DISPLAY_NAME, 0x8006001E, 0x8007001E, 0x800F101E, PR_EC_ADMINISTRATOR, PR_EC_NONACTIVE, PR_COMPANY_NAME,
		PR_EMS_AB_X509_CERT, PR_AB_PROVIDER_ID, PR_EMS_AB_HIERARCHY_PATH, PR_EC_SENDAS_USER_ENTRYIDS, PR_EC_EXCHANGE_DN, PR_EC_HOMESERVER_NAME
	};

	unsigned int	sPropsContainerRoot[] = {
		/* Don't touch the order of the first 7 elements!!! */
		PR_ENTRYID, PR_CONTAINER_FLAGS, PR_DEPTH, PR_EMS_AB_CONTAINERID, PR_DISPLAY_NAME, PR_EMS_AB_IS_MASTER, PR_EMS_AB_PARENT_ENTRYID,
		PR_OBJECT_TYPE, PR_DISPLAY_TYPE, PR_SEARCH_KEY, PR_RECORD_KEY, PR_PARENT_ENTRYID, PR_AB_PROVIDER_ID, PR_EMS_AB_HIERARCHY_PATH, PR_ACCOUNT,
		PR_EC_HOMESERVER_NAME, PR_COMPANY_NAME};

	unsigned int		ulId = 0;
	std::string			strExternId;
	unsigned int    	ulTypeId = 0;
	ECDatabase*			lpDatabase = NULL;
	userobject_type_t	sUserObjType = USEROBJECT_TYPE_USER;

	lpDatabase = lpecSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	er = ABEntryIDToID(&sEntryId, &ulId, &strExternId, &ulTypeId);
	if(er != erSuccess)
		goto exit;

	// A v1 EntryID would return a non-empty extern id string.
	if (!strExternId.empty())
	{
		er = MAPITypeToType(ulTypeId, &sUserObjType);
		if (er != erSuccess)
			goto exit;

		er = lpecSession->GetSessionManager()->GetCacheManager()->GetUserObject(strExternId, sUserObjType, NULL, &ulId, NULL);
		if (er != erSuccess && sUserObjType == USEROBJECT_TYPE_USER) {
    		er = g_lpSessionManager->GetCacheManager()->GetUserObject(strExternId, USEROBJECT_TYPE_NONACTIVE, NULL, &ulId, NULL);
        }
		if (er != erSuccess && sUserObjType == USEROBJECT_TYPE_COMPANY) {
    		er = g_lpSessionManager->GetCacheManager()->GetUserObject(strExternId, USEROBJECT_TYPE_ADDRESSLIST, NULL, &ulId, NULL);
        }
		if (er != erSuccess)
			goto exit;
	}

	// Get properties of user or container
	if (ulTypeId == MAPI_ABCONT) {
		// Build property tag array
		readPropsResponse->aPropTag.__size = sizeof(sPropsContainerRoot) / sizeof(sPropsContainerRoot[0]);
		readPropsResponse->aPropTag.__ptr = s_alloc<unsigned int>(soap, readPropsResponse->aPropTag.__size);
		memcpy(readPropsResponse->aPropTag.__ptr, sPropsContainerRoot, readPropsResponse->aPropTag.__size * sizeof(unsigned int));

		er = lpecSession->GetUserManagement()->GetContainerProps(soap, ulId, &readPropsResponse->aPropTag, &readPropsResponse->aPropVal);
	} else {
		struct propTagArray ptaProps;
		er = lpecSession->GetSecurity()->IsUserObjectVisible(ulId);
		if (er != erSuccess)
			goto exit;

		ptaProps.__size = sizeof(sProps) / sizeof(sProps[0]);
		ptaProps.__ptr = sProps;
		er = lpecSession->GetUserManagement()->GetProps(soap, ulId, &ptaProps, &readPropsResponse->aPropVal);
		if(er != erSuccess)
		    goto exit;
		    
		// Build property tag array
		readPropsResponse->aPropTag.__ptr = s_alloc<unsigned int>(soap, readPropsResponse->aPropVal.__size);

		int n = 0;
		for(int i=0;i<readPropsResponse->aPropVal.__size;i++) {
		    if(PROP_TYPE( readPropsResponse->aPropVal.__ptr[i].ulPropTag) != PT_ERROR) {
    		    readPropsResponse->aPropTag.__ptr[n] = readPropsResponse->aPropVal.__ptr[i].ulPropTag;
    		    n++;
            }
		}
		readPropsResponse->aPropTag.__size = n;
	}
	if(er != erSuccess)
		goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(writeABProps, *result, entryId sEntryId, struct propValArray *aPropVal, unsigned int *result)
{
    er = ZARAFA_E_NOT_FOUND;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(deleteABProps, *result, entryId sEntryId, struct propTagArray *lpsPropTags, unsigned int *result)
{
    er = ZARAFA_E_NOT_FOUND;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(loadABProp, lpsResponse->er, entryId sEntryId, unsigned int ulPropTag, struct loadPropResponse *lpsResponse)
{
	return SOAP_OK;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(abResolveNames, lpsABResolveNames->er, struct propTagArray* lpaPropTag, struct rowSet* lpsRowSet, struct flagArray* lpaFlags, unsigned int ulFlags, struct abResolveNamesResponse* lpsABResolveNames)
{
	unsigned int	ulFlag;
	unsigned int    ulObjectId;
	userobject_type_t objects[] = {
		USEROBJECT_TYPE_USER,
		USEROBJECT_TYPE_NONACTIVE,
		USEROBJECT_TYPE_GROUP,
		USEROBJECT_TYPE_COMPANY,
	};

	struct propValArray sPropValArrayDst;

	lpsABResolveNames->aFlags.__size = lpaFlags->__size;
	lpsABResolveNames->aFlags.__ptr = s_alloc<unsigned int>(soap, lpaFlags->__size);
	memset(lpsABResolveNames->aFlags.__ptr, 0, sizeof(unsigned int) * lpaFlags->__size);

	lpsABResolveNames->sRowSet.__size = lpsRowSet->__size;
	lpsABResolveNames->sRowSet.__ptr = s_alloc<propValArray>(soap, lpsRowSet->__size);
	memset(lpsABResolveNames->sRowSet.__ptr, 0, sizeof(struct propValArray) * lpsRowSet->__size);

	for(int i =0; i < lpsRowSet->__size; i++)
	{
		lpsABResolveNames->aFlags.__ptr[i] = lpaFlags->__ptr[i];

		if(lpaFlags->__ptr[i] == MAPI_RESOLVED)
			continue; // Client knows the information

		struct propVal *lpDisplayName = FindProp(&lpsRowSet->__ptr[i], PR_DISPLAY_NAME);
		if(lpDisplayName == NULL)
			continue; // No display name

		ulFlag = MAPI_UNRESOLVED;
		for (unsigned int j = 0; j < (sizeof(objects) / sizeof(objects[0])); j++) {
			/* Don't search for companies when hosted is disabled */
			if (objects[j] == USEROBJECT_TYPE_COMPANY && !g_lpSessionManager->IsHostedSupported())
				continue;
				
            /* Don't search for companies or groups if EMS_AB_ADDRESS_LOOKUP is specified */
            if ((ulFlags & EMS_AB_ADDRESS_LOOKUP) && objects[j] != USEROBJECT_TYPE_USER && objects[j] != USEROBJECT_TYPE_NONACTIVE)
                continue;
				
			/* Blackberry likes it to put a '=' in front of the username */
            char *search = lpDisplayName->Value.lpszA;
            if(strrchr(lpDisplayName->Value.lpszA, '='))
                search = strrchr(lpDisplayName->Value.lpszA, '=')+1;

			switch (lpecSession->GetUserManagement()->SearchObjectAndSync(objects[j], search, ulFlags, &ulObjectId)) {
			case ZARAFA_E_COLLISION:
				ulFlag = MAPI_AMBIGUOUS;
				break;
			case erSuccess:
				/* NOTE: ECUserManagement is responsible for calling ECSecurity::IsUserObjectVisible(ulObjectId) for found objects */
				if(ulFlag == MAPI_RESOLVED)
					ulFlag = MAPI_AMBIGUOUS;
				else if(ulFlag == MAPI_UNRESOLVED)
					ulFlag = MAPI_RESOLVED;
				else
					ulFlag = MAPI_AMBIGUOUS;
				break;
			}

			if (ulFlag == MAPI_AMBIGUOUS)
				break;
		}

   		lpsABResolveNames->aFlags.__ptr[i] = ulFlag;

		if(lpsABResolveNames->aFlags.__ptr[i] == MAPI_RESOLVED) {
			er = lpecSession->GetUserManagement()->GetProps(soap, ulObjectId, lpaPropTag, &sPropValArrayDst);
			if(er != erSuccess)
				goto exit;
			er = MergePropValArray(soap, &lpsRowSet->__ptr[i], &sPropValArrayDst, &lpsABResolveNames->sRowSet.__ptr[i]);
			if(er != erSuccess)
				goto exit;
		}
	}

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(syncUsers, *result, unsigned int ulCompanyId, entryId sCompanyId, unsigned int *result)
{
	std::list<localobjectdetails_t> *lstUserObjects = NULL;

	er = GetLocalId(NULL, sCompanyId, ulCompanyId, &ulCompanyId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	/* Input check, if ulCompanyId is 0, we want the user's company,
	 * otherwise we must check if the requested company is visible for the user. */
	if (ulCompanyId == 0) {
		er = lpecSession->GetSecurity()->GetUserCompany(&ulCompanyId);
		if (er != erSuccess)
			goto exit;
	} else {
		er = lpecSession->GetSecurity()->IsUserObjectVisible(ulCompanyId);
		if (er != erSuccess)
			goto exit;
	}

	/*
	 * When syncing the users we first start emptying the cache, this makes sure the
	 * second step won't be accidently "optimized" by caching.
	 * The second step is requesting all user objects from the plugin, ECUserManagement
	 * will then sync all results into the user database. And because the cache was
	 * cleared all signatures in the database will be checked against the signatures
	 * from the plugin.
	 * When this function has completed we can be sure that the cache has been repopulated
	 * with the userobject types, external ids and signatures of all user objects. This
	 * means that we have only "lost" the user details which will be repopulated later.
	 */
	er = g_lpSessionManager->GetCacheManager()->PurgeUserCache();
	if (er != erSuccess)
		goto exit;

	er = lpecSession->GetUserManagement()->GetMergedObjectListAndSync(USEROBJECT_TYPE_USER | USEROBJECT_TYPE_GROUP |
																	  USEROBJECT_TYPE_NONACTIVE | USEROBJECT_TYPE_COMPANY,
																	  ulCompanyId, &lstUserObjects, USERMANAGEMENT_IDS_ONLY);
	if (er != erSuccess)
		goto exit;

exit:
	if (lstUserObjects)
		delete lstUserObjects;
}
SOAP_ENTRY_END()

///////////////////
// Quota
//

SOAP_ENTRY_START3(GetQuota, lpsQuota->er, unsigned int ulUserid, entryId sUserId, struct quotaResponse* lpsQuota)
{
	quotadetails_t	quotadetails;

	er = GetLocalId(NULL, sUserId, ulUserid, &ulUserid, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	// Check permission
	if(lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserid) != erSuccess &&
		(lpecSession->GetSecurity()->GetUserId() != ulUserid))
	{
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	er = lpecSession->GetSecurity()->GetUserQuota(ulUserid, &quotadetails);
	if(er != erSuccess)
	    goto exit;

	lpsQuota->sQuota.bUseDefaultQuota = quotadetails.bUseDefaultQuota;
	lpsQuota->sQuota.bIsUserDefaultQuota = quotadetails.bIsUserDefaultQuota;
	lpsQuota->sQuota.llHardSize = quotadetails.llHardSize;
	lpsQuota->sQuota.llSoftSize = quotadetails.llSoftSize;
	lpsQuota->sQuota.llWarnSize = quotadetails.llWarnSize;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START4(SetQuota, *result, unsigned int ulUserid, entryId sUserId, struct quota* lpsQuota, unsigned int *result)
{
	quotadetails_t	quotadetails;

	er = GetLocalId(NULL, sUserId, ulUserid, &ulUserid, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	// Check permission
	if(lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserid) != erSuccess &&
		(lpecSession->GetSecurity()->GetUserId() != ulUserid))
	{
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	quotadetails.bUseDefaultQuota = lpsQuota->bUseDefaultQuota;
	quotadetails.bIsUserDefaultQuota = lpsQuota->bIsUserDefaultQuota;
	quotadetails.llHardSize = lpsQuota->llHardSize;
	quotadetails.llSoftSize = lpsQuota->llSoftSize;
	quotadetails.llWarnSize = lpsQuota->llWarnSize;

	er = lpecSession->GetUserManagement()->SetQuotaDetailsAndSync(ulUserid, quotadetails);

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START6(AddQuotaRecipient, *result, unsigned int ulCompanyid, entryId sCompanyId, unsigned int ulRecipientId, entryId sRecipientId, unsigned int ulType, unsigned int *result);
{
	er = GetLocalId(NULL, sCompanyId, ulCompanyid, &ulCompanyid, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	if (lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyid) != erSuccess) {
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	er = GetLocalId(NULL, sRecipientId, ulRecipientId, &ulRecipientId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	if (ulType == USEROBJECT_TYPE_USER)
		er = lpecSession->GetUserManagement()->AddSubObjectToObjectAndSync(OBJECTRELATION_QUOTA_USERRECIPIENT, ulCompanyid, ulRecipientId);
	else if (ulType == USEROBJECT_TYPE_COMPANY)
		er = lpecSession->GetUserManagement()->AddSubObjectToObjectAndSync(OBJECTRELATION_QUOTA_COMPANYRECIPIENT, ulCompanyid, ulRecipientId);
	else
		er = ZARAFA_E_INVALID_TYPE;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START6(DeleteQuotaRecipient, *result, unsigned int ulCompanyid, entryId sCompanyId, unsigned int ulRecipientId, entryId sRecipientId, unsigned int ulType, unsigned int *result);
{
	er = GetLocalId(NULL, sCompanyId, ulCompanyid, &ulCompanyid, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	if (lpecSession->GetSecurity()->IsUserObjectAdmin(ulCompanyid) != erSuccess) {
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	er = GetLocalId(NULL, sRecipientId, ulRecipientId, &ulRecipientId, NULL, NULL);
	if (er != hrSuccess)
		goto exit;

	if (ulType == USEROBJECT_TYPE_USER)
		er = lpecSession->GetUserManagement()->DeleteSubObjectFromObjectAndSync(OBJECTRELATION_QUOTA_USERRECIPIENT, ulCompanyid, ulRecipientId);
	else if (ulType == USEROBJECT_TYPE_COMPANY)
		er = lpecSession->GetUserManagement()->DeleteSubObjectFromObjectAndSync(OBJECTRELATION_QUOTA_COMPANYRECIPIENT, ulCompanyid, ulRecipientId);
	else
		er = ZARAFA_E_INVALID_TYPE;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(GetQuotaRecipients, lpsUserList->er, unsigned int ulUserid, entryId sUserId, struct userListResponse *lpsUserList)
{
	std::list<localobjectdetails_t> *lpUsers = NULL;
	std::list<localobjectdetails_t> *lpNonActive = NULL;
	std::list<localobjectdetails_t>::iterator iterUsers;
	objectdetails_t details;
	userobject_relation_t relation;
	unsigned int ulCompanyId;
	bool bHasLocalStore = false;

	std::string	strExternId;
	entryId		sUserEid = {0};

	er = GetLocalId(lpecSession, sUserId, ulUserid, &ulUserid, NULL, &bHasLocalStore);
	if (er != hrSuccess)
		goto exit;

	if (!bHasLocalStore) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	//Check permission
	if (lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserid) != erSuccess) {
			er = ZARAFA_E_NO_ACCESS;
			goto exit;
	}

	er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulUserid, &details);
	if (er != erSuccess)
		goto exit;

	if (details.GetType() == USEROBJECT_TYPE_USER || details.GetType() == USEROBJECT_TYPE_NONACTIVE) {
		er = lpecSession->GetUserManagement()->GetItemCompany(ulUserid, details.GetType(), &ulCompanyId);
		if (er != erSuccess)
			goto exit;

		relation = OBJECTRELATION_QUOTA_USERRECIPIENT;
	} else if (details.GetType() == USEROBJECT_TYPE_COMPANY) {
		ulCompanyId = ulUserid;
		relation = OBJECTRELATION_QUOTA_COMPANYRECIPIENT;
	} else {
		er = ZARAFA_E_INVALID_TYPE;
		goto exit;
	}

	/* When uLCompanyId is 0 then there are no recipient relations we could request,
	 * in that case we should manually allocate the list so it is safe to add the user
	 * to the list. */
	if (ulCompanyId != 0) {
		er = lpecSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(relation, USEROBJECT_TYPE_USER, ulCompanyId, &lpUsers);
		if (er != erSuccess)
			goto exit;

		er = lpecSession->GetUserManagement()->GetSubObjectsOfObjectAndSync(relation, USEROBJECT_TYPE_NONACTIVE, ulCompanyId, &lpNonActive);
		if (er != erSuccess)
			goto exit;

		lpUsers->merge(*lpNonActive);
	} else
		lpUsers = new std::list<localobjectdetails_t>();

	if (details.GetType() == USEROBJECT_TYPE_USER || details.GetType() == USEROBJECT_TYPE_NONACTIVE) {
		/* The main recipient (the user over quota) must be the first entry */
		lpUsers->push_front(localobjectdetails_t(ulUserid, details));
	} else if (details.GetType() == USEROBJECT_TYPE_COMPANY) {
		/* Append the system administrator for the company */
		unsigned int ulSystem;
		objectdetails_t systemdetails;

		er = lpecSession->GetUserManagement()->SearchObjectAndSync(USEROBJECT_TYPE_USER,
																   (char *)details.GetPropString(OB_PROP_S_SYSADMIN).c_str(),
																   EMS_AB_ADDRESS_LOOKUP, &ulSystem);
		if (er != erSuccess) {
			er = erSuccess;
			ulSystem = ZARAFA_UID_SYSTEM;
		}

		er = lpecSession->GetUserManagement()->GetObjectDetailsAndSync(ulSystem, &systemdetails);
		if (er != erSuccess)
			goto exit;

		lpUsers->push_front(localobjectdetails_t(ulSystem, systemdetails));

		/* The main recipient (the company's public store) must be the first entry */
		lpUsers->push_front(localobjectdetails_t(ulUserid, details));
	}

	lpsUserList->sUserArray.__size = 0;
	lpsUserList->sUserArray.__ptr = s_alloc<user>(soap, lpUsers->size());

	for(iterUsers = lpUsers->begin(); iterUsers != lpUsers->end(); iterUsers++) {
		if (lpecSession->GetSecurity()->IsUserObjectVisible(iterUsers->ulId) != erSuccess)
			continue;

		er = GetABEntryID(iterUsers->ulId, soap, &sUserEid);
		if (er != erSuccess)
			goto exit;

		er = CopyUserDetailsToSoap(iterUsers->ulId, &sUserEid, *iterUsers, soap, &lpsUserList->sUserArray.__ptr[lpsUserList->sUserArray.__size]);
		if (er != erSuccess)
			goto exit;

		lpsUserList->sUserArray.__size++;

		if (sUserEid.__ptr)
		{
			// sUserEid is placed in userdetails, no need to free
			sUserEid.__ptr = NULL;
			sUserEid.__size = 0;
		}
	}

exit:
	if(lpUsers)
		delete lpUsers;

	if(lpNonActive)
		delete lpNonActive;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(GetQuotaStatus, lpsQuotaStatus->er, unsigned int ulUserid, entryId sUserId, struct quotaStatus* lpsQuotaStatus)
{
	quotadetails_t	quotadetails;
	long long		llStoreSize = 0;
	userobject_type_t ulType;
	bool			bHasLocalStore = false;

	eQuotaStatus QuotaStatus;

	//Set defaults
	lpsQuotaStatus->llStoreSize = 0;

	er = GetLocalId(lpecSession, sUserId, ulUserid, &ulUserid, NULL, &bHasLocalStore);
	if (er != hrSuccess)
		goto exit;

	if (!bHasLocalStore) {
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	// Check permission
	if(lpecSession->GetSecurity()->IsUserObjectAdmin(ulUserid) != erSuccess &&
		(lpecSession->GetSecurity()->GetUserId() != ulUserid))
	{
		er = ZARAFA_E_NO_ACCESS;
		goto exit;
	}

	er = lpecSession->GetUserManagement()->GetItemObjectType(ulUserid, &ulType);
	if (er != erSuccess)
		goto exit;

	if (ulType == USEROBJECT_TYPE_USER || ulType == USEROBJECT_TYPE_NONACTIVE) {
		er = lpecSession->GetSecurity()->GetUserSize(ulUserid, &llStoreSize);
		if(er != erSuccess)
			goto exit;
	} else if (ulType == USEROBJECT_TYPE_COMPANY) {
		er = lpecSession->GetSecurity()->GetCompanySize(ulUserid, &llStoreSize);
		if (er != erSuccess)
			goto exit;
	} else {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// check the store quota status
	er = lpecSession->GetSecurity()->CheckUserQuota(ulUserid, llStoreSize, &QuotaStatus);
	if(er != erSuccess)
		goto exit;

	lpsQuotaStatus->llStoreSize = llStoreSize;
	lpsQuotaStatus->ulQuotaStatus = (unsigned int)QuotaStatus;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getMessageStatus, lpsStatus->er, entryId sEntryId, unsigned int ulFlags, struct messageStatus* lpsStatus)
{
	unsigned int	ulMsgStatus = 0;
	unsigned int	ulStore = 0;
	unsigned int	ulId = 0;
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulId);
	if(er != erSuccess)
	    goto exit;

	//Check security
	er = lpecSession->GetSecurity()->CheckPermission(ulId, ecSecurityRead);
	if(er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetStore(ulId, &ulStore, NULL);
	if(er != erSuccess)
		goto exit;

	// Get the old flags
	strQuery = "SELECT val_ulong FROM properties WHERE storeid="+stringify(ulStore)+" AND hierarchyid="+stringify(ulId)+" AND tag=3607 AND type=3";
	if(lpDatabase->DoSelect(strQuery, &lpDBResult) != erSuccess) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if (lpDatabase->GetNumRows(lpDBResult) == 1) {
		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		if(lpDBRow == NULL || lpDBRow[0] == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		ulMsgStatus = atoui(lpDBRow[0]);
	}

	lpsStatus->ulMessageStatus = ulMsgStatus;

exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START5(setMessageStatus, lpsOldStatus->er, entryId sEntryId, unsigned int ulNewStatus, unsigned int ulNewStatusMask, unsigned int ulSyncId, struct messageStatus* lpsOldStatus)
{
	unsigned int	ulOldMsgStatus = 0;
	unsigned int	ulNewMsgStatus = 0;
	unsigned int	ulStore = 0;
	unsigned int	ulRows = 0;
	unsigned int	ulId = 0;
	unsigned int	ulParent = 0;
	unsigned int	ulObjFlags = 0;
	
	SOURCEKEY		sSourceKey;
	SOURCEKEY		sParentSourceKey;

	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sEntryId, &ulId);
	if(er != erSuccess)
	    goto exit;

	//Check security
	er = lpecSession->GetSecurity()->CheckPermission(ulId, ecSecurityEdit);
	if(er != erSuccess)
		goto exit;

	er = lpDatabase->Begin();
	if(er != hrSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetStore(ulId, &ulStore, NULL);
	if(er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObject(ulId, &ulParent, NULL, &ulObjFlags);
	if(er != erSuccess)
		goto exit;

	// Get the old flags
	strQuery = "SELECT val_ulong FROM properties WHERE storeid="+stringify(ulStore)+" AND hierarchyid="+stringify(ulId)+" AND tag=3607 AND type=3";
	if(lpDatabase->DoSelect(strQuery, &lpDBResult) != erSuccess) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	ulRows = lpDatabase->GetNumRows(lpDBResult);
	if (ulRows == 1) {
		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		if(lpDBRow == NULL || lpDBRow[0] == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		ulOldMsgStatus = atoui(lpDBRow[0]);
	}

	// Set the new flags
	ulNewMsgStatus = (ulOldMsgStatus &~ulNewStatusMask) | (ulNewStatusMask & ulNewStatus);

	if(ulRows > 0){
		strQuery = "UPDATE properties SET val_ulong="+stringify(ulNewMsgStatus)+" WHERE storeid="+stringify(ulStore)+" AND hierarchyid="+stringify(ulId)+" AND tag=3607 AND type=3";
		er = lpDatabase->DoUpdate(strQuery);
	}else {
		strQuery = "INSERT INTO properties(storeid, hierarchyid, tag, type, val_ulong) VALUES("+stringify(ulStore)+","+stringify(ulId)+", 3607, 3,"+stringify(ulNewMsgStatus)+")";
		er = lpDatabase->DoInsert(strQuery);
	}

	if(er != erSuccess) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	lpsOldStatus->ulMessageStatus = ulOldMsgStatus;

	GetSourceKey(ulId, &sSourceKey);
	GetSourceKey(ulParent, &sParentSourceKey);

	AddChange(lpecSession, ulSyncId, sSourceKey, sParentSourceKey, ICS_MESSAGE_CHANGE);

	er = lpDatabase->Commit();
	if(er != hrSuccess)
		goto exit;

	// Now, send the notifications
	g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulId);
	g_lpSessionManager->NotificationModified(MAPI_MESSAGE, ulId, ulParent);
	g_lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, ulObjFlags&MSGFLAG_NOTIFY_FLAGS, ulParent, ulId, MAPI_MESSAGE);

exit:
    ROLLBACK_ON_ERROR();
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START7(getChanges, lpsChangesResponse->er, struct xsd__base64Binary sSourceKeyFolder, unsigned int ulSyncId, unsigned int ulChangeId, unsigned int ulChangeType, unsigned int ulFlags, struct restrictTable *lpsRestrict, struct icsChangeResponse* lpsChangesResponse)
{
	icsChangesArray *lpChanges = NULL;
	SOURCEKEY		sSourceKey(sSourceKeyFolder.__size, (char *)sSourceKeyFolder.__ptr);

	er = GetChanges(soap, lpecSession, sSourceKey, ulSyncId, ulChangeId, ulChangeType, ulFlags, lpsRestrict, &lpsChangesResponse->ulMaxChangeId, &lpChanges);
	if(er != erSuccess)
		goto exit;

	lpsChangesResponse->sChangesArray = *lpChanges;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(getChangesMulti, lpsResponse->er, struct syncStateArray sStateArray, struct getChangesMultiResponse* lpsResponse)
{
	lpsResponse->sResponses.__ptr = (struct getChangesMultiResponsePart *)soap_malloc(soap, sizeof(struct getChangesMultiResponsePart) * sStateArray.__size);
	lpsResponse->sResponses.__size = sStateArray.__size;
	
	for(unsigned int i=0;i<sStateArray.__size;i++) {
    	er = GetChanges(soap, lpecSession, SOURCEKEY(0,NULL), sStateArray.__ptr[i].ulSyncId, sStateArray.__ptr[i].ulChangeId, sStateArray.__ptr[i].ulType, sStateArray.__ptr[i].ulFlags, sStateArray.__ptr[i].lpsRestriction, &lpsResponse->sResponses.__ptr[i].ulMaxChangeId, &lpsResponse->sResponses.__ptr[i].lpsChanges);
    	if(er != erSuccess) {
    	    // Simply return 0 changes on error
    	    lpsResponse->sResponses.__ptr[i].ulMaxChangeId = 0;
    	    lpsResponse->sResponses.__ptr[i].lpsChanges = NULL;
    	    er = erSuccess;
        }
    }
exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START6(setSyncStatus, lpsResponse->er, struct xsd__base64Binary sSourceKeyFolder, unsigned int ulSyncId, unsigned int ulChangeId, unsigned int ulChangeType, unsigned int ulFlags, struct setSyncStatusResponse *lpsResponse)
{
	SOURCEKEY		sSourceKey(sSourceKeyFolder.__size, (char *)sSourceKeyFolder.__ptr);
	unsigned int	ulFolderId = 0;
	USE_DATABASE();

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromProp(PROP_ID(PR_SOURCE_KEY), sSourceKey.size(), sSourceKey, &ulFolderId);
	if(er != erSuccess)
	    goto exit;

	//Check security
	if(ulChangeType == ICS_SYNC_CONTENTS){
		er = lpecSession->GetSecurity()->CheckPermission(ulFolderId, ecSecurityRead);
	}else if(ulChangeType == ICS_SYNC_HIERARCHY){
		er = lpecSession->GetSecurity()->CheckPermission(ulFolderId, ecSecurityFolderVisible);
	}else{
		er = ZARAFA_E_INVALID_TYPE;
	}
	if(er != erSuccess)
		goto exit;

	if(ulSyncId == 0){
	    // SyncID is 0, which means the client will be requesting an initial sync from this new sync. The change
	    // ID we save for the sync is therefore the current highest change ID. This is the ID at which 'normal'
	    // change processing will resume after the initial sync is done.
		strQuery = "INSERT INTO syncs (change_id, sourcekey, sync_type, sync_time) VALUES (IFNULL((SELECT MAX(id) FROM changes), 1), "+lpDatabase->EscapeBinary(sSourceKey, sSourceKey.size())+", '" + stringify(ulChangeType) + "', NOW())";
		er = lpDatabase->DoInsert(strQuery, &ulSyncId);
		lpsResponse->ulSyncId = ulSyncId;
		goto exit;
	}

	strQuery = "SELECT sourcekey, change_id, sync_type FROM syncs WHERE id ="+stringify(ulSyncId);
	//TODO check existing sync
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if(er != erSuccess)
		goto exit;

	if(lpDatabase->GetNumRows(lpDBResult) == 0){
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

	if( lpDBRow == NULL || lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL){
		er = ZARAFA_E_DATABASE_ERROR; // this should never happen
		goto exit;
	}

	if(lpDBLen[0] != sSourceKey.size() || memcmp(lpDBRow[0], sSourceKey, sSourceKey.size()) != 0){
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	if(atoui(lpDBRow[2]) != ulChangeType){
		er = ZARAFA_E_COLLISION;
		goto exit;
	}

	strQuery = "UPDATE syncs SET change_id = "+stringify(ulChangeId)+", sync_time = NOW() WHERE id = "+stringify(ulSyncId);
	er = lpDatabase->DoUpdate(strQuery);
	if(er != erSuccess)
		goto exit;

	lpsResponse->ulSyncId = ulSyncId;
exit:
    FREE_DBRESULT();
}
SOAP_ENTRY_END()

SOAP_ENTRY_START4(getEntryIDFromSourceKey, lpsResponse->er, entryId sStoreId, struct xsd__base64Binary folderSourceKey, struct xsd__base64Binary messageSourceKey, struct getEntryIDFromSourceKeyResponse *lpsResponse)
{
	unsigned int	ulObjType = 0;
	unsigned int	ulObjId = 0;
	unsigned int	ulMessageId = 0;
	unsigned int	ulFolderId = 0;
	unsigned int	ulParent = 0;
	unsigned int	ulStoreId = 0;
	EID				eid;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromEntryId(&sStoreId, &ulStoreId);
	if(er != erSuccess)
	    goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObjectFromProp(PROP_ID(PR_SOURCE_KEY), folderSourceKey.__size, folderSourceKey.__ptr, &ulFolderId);
	if(er != erSuccess)
		goto exit;

	if(messageSourceKey.__size != 0) {
		er = g_lpSessionManager->GetCacheManager()->GetObjectFromProp(PROP_ID(PR_SOURCE_KEY), messageSourceKey.__size, messageSourceKey.__ptr, &ulMessageId);
		if(er != erSuccess)
			goto exit;

		er = g_lpSessionManager->GetCacheManager()->GetParent(ulMessageId, &ulParent);
		if(er != erSuccess || ulFolderId != ulParent) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		ulObjId = ulMessageId;
		ulObjType = MAPI_MESSAGE;

	} else {
		ulObjId = ulFolderId;
		ulObjType = MAPI_FOLDER;
	}


	// Check security
	if(ulObjType == MAPI_FOLDER){
		er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityFolderVisible);
	}else{
		er = lpecSession->GetSecurity()->CheckPermission(ulObjId, ecSecurityRead);
	}
	if(er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetEntryIdFromObject(ulObjId, soap, &lpsResponse->sEntryId);
	if(er != erSuccess)
		goto exit;

exit:
    ;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(getLicenseAuth, lpsResponse->er, struct xsd__base64Binary sAuthData, struct getLicenseAuthResponse *lpsResponse)
{
    ECLicenseClient *lpLicenseClient = new ECLicenseClient(g_lpSessionManager->GetConfig()->GetSetting("license_socket"));
	unsigned char *data = NULL;

    er = lpLicenseClient->Auth(sAuthData.__ptr, sAuthData.__size, &data, (unsigned int *)&lpsResponse->sAuthResponse.__size);
    if(er != erSuccess)
        goto exit;

	lpsResponse->sAuthResponse.__ptr = s_alloc<unsigned char>(soap, lpsResponse->sAuthResponse.__size);
	memcpy(lpsResponse->sAuthResponse.__ptr, data, lpsResponse->sAuthResponse.__size);

exit:
	if (data)
		delete [] data;

    delete lpLicenseClient;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START1(getLicenseCapa, lpsResponse->er, struct getLicenseCapaResponse *lpsResponse)
{
    ECLicenseClient *lpLicenseClient = new ECLicenseClient(g_lpSessionManager->GetConfig()->GetSetting("license_socket"));
    std::vector<std::string> lstCapabilities;
    
    er = lpLicenseClient->GetCapabilities(lstCapabilities);
    if(er != erSuccess)
        goto exit;
        
    lpsResponse->sCapabilities.__size = lstCapabilities.size();
    lpsResponse->sCapabilities.__ptr = s_alloc<char *>(soap, lstCapabilities.size());
    
    for(unsigned int i=0; i<lstCapabilities.size(); i++) {
        lpsResponse->sCapabilities.__ptr[i] = s_strcpy(soap, lstCapabilities[i].c_str());
    }
exit:

    delete lpLicenseClient;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(resolvePseudoUrl, lpsResponse->er, char *lpszPseudoUrl, struct resolvePseudoUrlResponse* lpsResponse)
{
	std::string		strServerPath;

	if (!lpecSession->GetSessionManager()->IsDistributedSupported())
	{
		/**
         * Non distributed environments do issue pseudo url's, but merely to 
         * make upgrading later possible. For now we'll just return
         * ZARAFA_E_NOT_FOUND to have the client use the server path from its
         * profile.
         **/
		er = ZARAFA_E_NOT_FOUND;
		goto exit;
	}

	if (strncmp(lpszPseudoUrl, "pseudo://", 9))
	{
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}
	
	er = GetBestServerPath(soap, lpecSession, lpszPseudoUrl + 9, &strServerPath);
	if (er != erSuccess)
		goto exit;

	lpsResponse->lpszServerPath = s_strcpy(soap, strServerPath.c_str());
	lpsResponse->bIsPeer = strcmp(g_lpSessionManager->GetConfig()->GetSetting("server_name"), lpszPseudoUrl + 9) == 0;

exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START3(getServerDetails, lpsResponse->er, struct mv_string8 szaSvrNameList, unsigned int ulFlags, struct getServerDetailsResponse* lpsResponse)
{
	serverdetails_t	sDetails;
	std::string		strServerPath;
	
	if (!lpecSession->GetSessionManager()->IsDistributedSupported())
	{
		/**
		 * We want to pretend the method doesn't exist. The best way to do that is to return
		 * SOAP_NO_METHOD. But that doesn't fit well in the macro famework. And since the
		 * client would translate that to a ZARAFA_E_NETWORK_ERROR anyway, we'll just return
		 * that instead.
		 **/
		er = ZARAFA_E_NETWORK_ERROR;
		goto exit;
	}
	
	if (ulFlags == 0)
		ulFlags = EC_SERVERDETAIL_FILEPATH|EC_SERVERDETAIL_HTTPPATH|EC_SERVERDETAIL_SSLPATH	|EC_SERVERDETAIL_PREFEREDPATH;
	
	if (szaSvrNameList.__size > 0 && szaSvrNameList.__ptr != NULL) {
		lpsResponse->sServerList.__size = szaSvrNameList.__size;
		lpsResponse->sServerList.__ptr = s_alloc<struct server>(soap, szaSvrNameList.__size);
		memset(lpsResponse->sServerList.__ptr, 0, szaSvrNameList.__size * sizeof *lpsResponse->sServerList.__ptr);
		
		for (int i = 0; i < szaSvrNameList.__size; ++i) {
			er = lpecSession->GetUserManagement()->GetServerDetails(szaSvrNameList.__ptr[i], &sDetails);
			if (er != erSuccess)
				goto exit;

			
			if (sDetails.GetServerName().compare(g_lpSessionManager->GetConfig()->GetSetting("server_name")) == 0)
				lpsResponse->sServerList.__ptr[i].ulFlags |= EC_SDFLAG_IS_PEER;
				
			if ((ulFlags & EC_SERVERDETAIL_NO_NAME) != EC_SERVERDETAIL_NO_NAME)
				lpsResponse->sServerList.__ptr[i].lpszName = s_strcpy(soap, sDetails.GetServerName().c_str());
				
			if ((ulFlags & EC_SERVERDETAIL_FILEPATH) == EC_SERVERDETAIL_FILEPATH)
				lpsResponse->sServerList.__ptr[i].lpszFilePath = s_strcpy(soap, sDetails.GetFilePath().c_str());
				
			if ((ulFlags & EC_SERVERDETAIL_HTTPPATH) == EC_SERVERDETAIL_HTTPPATH)
				lpsResponse->sServerList.__ptr[i].lpszHttpPath = s_strcpy(soap, sDetails.GetHttpPath().c_str());
				
			if ((ulFlags & EC_SERVERDETAIL_SSLPATH) == EC_SERVERDETAIL_SSLPATH)
				lpsResponse->sServerList.__ptr[i].lpszSslPath = s_strcpy(soap, sDetails.GetSslPath().c_str());
				
			if ((ulFlags & EC_SERVERDETAIL_PREFEREDPATH) == EC_SERVERDETAIL_PREFEREDPATH) {
				if (GetBestServerPath(soap, lpecSession, sDetails.GetServerName(), &strServerPath) == erSuccess)
					lpsResponse->sServerList.__ptr[i].lpszPreferedPath = s_strcpy(soap, strServerPath.c_str());
			}
		}
	}
exit:
	;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START1(getServerBehavior, lpsResponse->er ,struct getServerBehaviorResponse* lpsResponse)
{
	lpsResponse->ulBehavior = EC_BF_ABEID_V1;
}
SOAP_ENTRY_END()

SOAP_ENTRY_START2(setServerBehavior, *result, unsigned int ulBehavior, unsigned int *result)
{
	er = ZARAFA_E_NO_SUPPORT;
}
SOAP_ENTRY_END()
